<?php

namespace App\Module\Cadastre\WebService\Controller;

use App\Module\Cadastre\WebService\Entity\Commune;
use App\Module\Vitis\WebService\Controller\GenericController;
use App\Module\Vitis\WebService\Repository\GenericRepository;
use Symfony\Component\HttpFoundation\Response;
use Doctrine\ORM\EntityManager;

/**
 * Controller abstrait pour déterminer les droits sur la commune
 * @author Alizée Buckel <alizee.buckel@veremes.com>
 */

abstract class AbstractCadastreController extends GenericController {
  /**
   * @var string[]
   */
  private const INTERSECTION_FORBIDEN_KEY = ['title', 'type', 'intersect', 'geom_column'];

  /**
   * @param string $id_com
   * @return bool
   * Cette méthode permet de controler les droits de l'utilisateur sur la "commune"
   */
  protected function haveCommuneRights(string $id_com): bool {
    // Vérifie si on a les droits sur la commune
    $oEntityManager = $this->oDatabaseConnection->getUserEntityManager();
    $aRepository = $oEntityManager->getRepository(Commune::class);
    if (!$aRepository instanceof GenericRepository) {
      throw new \Exception("Repository is not a GenericRepository", 1);
    };
    $aCommunes = $aRepository->findById($id_com);
    return (count($aCommunes) > 0);
  }

  /**
   * fonction utilisé dans un callback pour array_filter
   * @param mixed $val
   * @param string $key
   * @return bool true pour garder l'élément, false pour le supprimer
   */
  protected function filterValueForAttribute($val, string $key): bool {
    return !in_array($key, self::INTERSECTION_FORBIDEN_KEY);
  }

  /**
   * fonction utilisé dans un callback pour array_filter
   * @param mixed $val
   * @param string $key
   * @return bool true pour garder l'élément, false pour le supprimer
   */
  protected function filterUnusedKeyInValue($val, string $key): bool {
    return !in_array($key, ['schema', 'table']);
  }

  /**
   * Retourne un objet contenant les intersections
   * @param string $sSchema : le schéma des données
   * @param string $sTable : la table des données
   * @param string $geom : la géométrie
   * @param array $aAttributs : la structure des données
   * @param integer $iTolerance : la distance entre deux surfaces tolérées
   * @param string $sGeomColumn "geom"
   * @param string $sGeomType : le type de géométrie
   * @param EntityManager|null $oEm : connexion à utiliser pour réquêter la base
   * @return array un tableau contenant les données concernant les intersections
   */
  protected function vmapIntersectGet($sSchema, $sTable, $geom, $aAttributs, $iTolerance = 0, $sGeomColumn = 'geom', $sGeomType = 'POLYGON', $oEm = null) {
    // Initialisation des variables contenant les données
    $column = [];
    $aIntersect = null;

    $oEm = $oEm ?? $this->oDatabaseConnection->getUserEntityManager();

    // Ajout de column
    foreach ($aAttributs['columns'] as $col) {
      if ($col['name'] != 'intersect') {
        array_push($column, $col['name']);
      }
    }

    // Preparation de la requete SQL pour récupérer le % d'intersection
    if (in_array($sGeomType, ['POLYGON', 'MULTIPOLYGON'])) {
      $sIntersectPercentage = 'round(';
      $sIntersectPercentage .= 'ST_Area(';
      $sIntersectPercentage .= 'ST_Intersection(';
      $sIntersectPercentage .= 'ST_Transform(';
      $sIntersectPercentage .= 'ST_GeomFromEWKT([GEOM]),';
      $sIntersectPercentage .= 'ST_SRID([GEOM_COL])';
      $sIntersectPercentage .= '), ';
      $sIntersectPercentage .= '[GEOM_COL])';
      $sIntersectPercentage .= ')/ST_Area(';
      $sIntersectPercentage .= 'ST_Transform(';
      $sIntersectPercentage .= 'ST_GeomFromEWKT([GEOM]),';
      $sIntersectPercentage .= 'ST_SRID([GEOM_COL])';
      $sIntersectPercentage .= ')';
      $sIntersectPercentage .= ') * 100) as intersect';
    } else if (in_array($sGeomType, ['LINE', 'LINESTRING', 'MULTILINESTRING'])) {
      $sIntersectPercentage = 'round(';
      $sIntersectPercentage .= 'ST_Length(';
      $sIntersectPercentage .= 'ST_Intersection(';
      $sIntersectPercentage .= 'ST_Transform(';
      $sIntersectPercentage .= 'ST_GeomFromEWKT([GEOM]),';
      $sIntersectPercentage .= 'ST_SRID([GEOM_COL])';
      $sIntersectPercentage .= '), ';
      $sIntersectPercentage .= '[GEOM_COL])';
      $sIntersectPercentage .= ')) as intersect';
    } else {
      $sIntersectPercentage = '100 as intersect';
    }

    // Requête SQL pour obtenir le % d'intersection
    $qb = $this->oDatabaseConnection->getQueryAccessor($oEm)->createQueryBuilder();

    array_unshift($column, $sIntersectPercentage);

    $qb->select($column)
      ->from('[SCHEMA].[TABLE]')
      ->where('ST_Intersects([GEOM_COL], ST_GeomFromEWKT([GEOM]))');

    $aParams = [
      'SCHEMA' => ['value' => $sSchema, 'type' => 'schema_name'],
      'TABLE' => ['value' => $sTable, 'type' => 'table_name'],
      'GEOM_COL' => ['value' => $sGeomColumn, 'type' => 'column_name'],
      'GEOM' => ['value' => $geom, 'type' => 'string'],
    ];

    try {
      $aIntersect = $qb->executeWithParams($aParams)->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('The data could not be retrieved.' . $e->getMessage());
      $aReturn = ['status' => 0];
      $aReturn['message'] = 'The data could not be retrieved. ' . $e->getMessage();
      $this->formatResponse($aReturn, Response::HTTP_NOT_ACCEPTABLE);
      return $aReturn;
    }

    // Ajout de la tolerance et mise à forme de chaque columns
    $aTolerancedRecords = [];
    foreach ($aIntersect as $oIntersect) {
      if ($oIntersect['intersect'] >= $iTolerance) {
        if (in_array($sGeomType, ['POLYGON', 'MULTIPOLYGON'])) {
          $oIntersect['intersect'] .= '%';
        } else if (in_array($sGeomType, ['LINE', 'LINESTRING', 'MULTILINESTRING'])) {
          $oIntersect['intersect'] .= 'm';
        } else {
          $oIntersect['intersect'] = true;
        }
        $aTolerancedRecords[] = $oIntersect;
      }
    }
    return $aTolerancedRecords;
  }

  /**
   * Récupère la géométrie d'une parcelle
   *
   * @param  string     $sIdPar
   * @return string|null
   */
  protected function findParcelleGeom(string $sIdPar): ?string{
    // Recuperation de la géométrie de la parcelle
    $sSchema = $this->oProperties->get('cadastre_schema_cadastre');
    $qb = $this->oDatabaseConnection->getQueryAccessor()->createQueryBuilder();

    $aParams = [
      'SCHEMA' => ['value' => $sSchema, 'type' => 'schema_name'],
      'TABLE' => ['value' => $this->oProperties->get('cadastre_views:parcelle'), 'type' => 'table_name'],
      'ID_PAR' => ['value' => $sIdPar, 'type' => 'string'],
    ];
    if (empty($aData['result_srid'])) {
      $qb->select('ST_AsEWKT(m.geom) as geom');
    } else {
      $qb->select('ST_AsEWKT(ST_Transform(m.geom, [SRID]::int)) as geom');
      $aParams['SRID'] = ['value' => $aData['result_srid'], 'type' => 'string'];
    }
    $qb->from('[SCHEMA].[TABLE]', 'm')
      ->where($qb->expr()->eq('m.id_par', '[ID_PAR]'));
    try {
      $oGeom = $qb->executeWithParams($aParams)->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('The data could not be retrieved.' . $e->getMessage());
      return null;
    }
    return $oGeom[0]['geom'];
  }


    /**
     * récupère les intersections
   * @param bool $bInterectIdCom cadastre_fiche_urba_intersect_id_com dans les properties
   * @param string $geom géométrie de la parcelle
   * @param array $aIntersectionViews cadastre_fiche_urba_intersect_tables dans les properties
   * @param integer $iTolerance cadastre_fiche_urba_intersect_tolerance dans le sproperties
   * @param string $idPar paramètres de la route
   * @return array
   */
  protected function getGeomIntersections($bInterectIdCom, $geom, $aIntersectionViews, $iTolerance, $idPar): array{
    $intersections = [];

    $oEm = $this->oDatabaseConnection->getUserEntityManager();
    if ($this->oProperties->get('cadastre_database_intersect') !== $this->oProperties->get('db_name')){
      $oEm = $this->oDatabaseConnection->createExternalEntityManager($this->oProperties->get('cadastre_database_intersect'));
    }

    // Récupère les intersections
    foreach ($aIntersectionViews as $value) {
      $sSchema = $this->oProperties->get('cadastre_schema_cadastre');
      $sTable = $this->oProperties->get('cadastre_views:parcelle');
      // Creation de l'infrastructure pour chaque intersection (schema, table, title, type, columns)
      $sSchema = $value['schema'] ?? 'public';
      $sTable = $value['table'] ?? '';

      // Ajout de columns dans les attributs
      $aAttributs = array_filter($value, [$this, 'filterValueForAttribute'], ARRAY_FILTER_USE_BOTH);

      // Récuperation de geom
      $sGeomColumn = (!empty($value['geom_column'])) ? $value['geom_column'] : 'geom';
      // Récuperation du type de géometrie
      $sGeomType = (!empty($value['type'])) ? $value['type'] : 'POLYGON';

      // Creation des columns pour chaque intersection
      $aIntersect = $this->vmapIntersectGet($sSchema, $sTable, $geom, $aAttributs, $iTolerance, $sGeomColumn, $sGeomType, $oEm);

      // Supprime les intersections qui ne sont pas de le même commune
      if ($bInterectIdCom == true) {
        foreach ($aIntersect as $key2 => $value2) {
          if (!empty($value2['id_com'])) {
            if ($value2['id_com'] != substr($idPar, 0, 5)) {
              unset($aIntersect[$key2]);
            }
          }
        }
      }

      // Ajout de data
      $value['data'] = $aIntersect;
      // suppression de schema table
      $value = array_filter($value, [$this, 'filterUnusedKeyInValue'], ARRAY_FILTER_USE_BOTH);
      array_push($intersections, $value);
    }

    return $intersections;
  }
}
