<?php

namespace App\Module\Cadastre\WebService\Controller;

use App\Module\Cadastre\WebService\Entity\FicheUrbanisme\FicheUrbanisme;
use App\Module\Cadastre\WebService\Repository\FicheUrbanismeRepository;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\PrintService\PrintTemplate;
use App\Module\Vitis\WebService\Service\VitisRequest;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller permettant la génération de la fiche d'urbanisme d'une parcelle
 * @author Alizée Buckel <alizee.buckel@veremes.com>
 */

class FicheUrbanismeController extends AbstractCadastreController {
  /**
   * @var PrintTemplate
   */
  protected $oPrintService;

  /**
   * Initialise le controler et instancie les services requis
   * @param Properties $oPropertiesSrvc Service d'accès aux properties
   * @param VitisRequest $oVitisRequest Service d'abstraction de la requête Symfony
   * @param Route $oRouteConfig Service d'accès à la configuration de la route
   * @param LoggerManager $oLoggerManager Service d'utilsation des logs
   * @param DatabaseConnection $oDatabaseConnection Service d'accès à la base de données
   * @param FileManager $oFileManager Service de gestion des fichiers
   * @param PrintTemplate $oTemplate Service de manipulation des token JWE
   */
  public function __construct(Properties $oPropertiesSrvc, VitisRequest $oVitisRequest, Route $oRouteConfig, LoggerManager $oLoggerManager, DatabaseConnection $oDatabaseConnection, FileManager $oFileManager, PrintTemplate $oTemplate) {
    parent::__construct($oPropertiesSrvc, $oVitisRequest, $oRouteConfig, $oLoggerManager, $oDatabaseConnection, $oFileManager);
    $this->oPrintService = $oTemplate;
  }

  /**
   * Retourne les infos de la fiche urbanisme
   * @return Response
   */
  public function getDatas(): Response {
    $oParcelleInfos = $this->getParcelleInfos();

    // Gestion du retour d'erreur
    if (!is_array($oParcelleInfos)) {
      return $oParcelleInfos;
    }

    return $this->formatResponse($oParcelleInfos);
  }

  /**
   * @param array $aData paramètres de la route
   * récupère la géométrie de la vue v_vmap_parcelle_all_geom
   * @return string|null|Response
   */
  public function getGeom($aData) {
    $idPar = $aData['ID_PAR'];
    // Recuperation de la géométrie de la parcelle
    $oGeom = $this->findParcelleGeom($idPar);

    if (empty($oGeom)) {
      $aReturn = ['status' => 0];
      $aReturn['message'] = 'The data could not be retrieved';
      return $this->formatResponse($aReturn, Response::HTTP_NOT_ACCEPTABLE);
    }

    return $oGeom;
  }

  /**
   * récupère les intersections
   * @param bool $bInterectIdCom cadastre_fiche_urba_intersect_id_com dans les properties
   * @param string $geom géométrie de la parcelle
   * @param array $aIntersectionViews cadastre_fiche_urba_intersect_tables dans les properties
   * @param integer $iTolerance cadastre_fiche_urba_intersect_tolerance dans le sproperties
   * @param string $idPar paramètres de la route
   * @return array
   */
  protected function getIntersection($bInterectIdCom, $geom, $aIntersectionViews, $iTolerance, $idPar): array{
    return $this->getGeomIntersections($bInterectIdCom, $geom, $aIntersectionViews, $iTolerance, $idPar);
  }

  /**
   * Imprime la fiche d'urbanisme
   */
  public function printFicheUrba(): Response {
    // Récupération des infos
    $aData = $this->requestToArray();
    $idPar = $aData['ID_PAR'];
    $oParcelleInfos = $this->getParcelleInfos();

    // Gestion du retour d'erreur
    if (!is_array($oParcelleInfos)) {
      return $oParcelleInfos;
    }

    /* Génération du dossier */
    $oFile = $this->oFileManager->getLocalFileAccessor();
    $sDirPath = '@prop(cadastre_report_dir)/fiche_urbanisme/';
    $sDirPath = $this->oProperties->replacePropertiesInString($sDirPath);
    $sPath = $sDirPath . $idPar . '.pdf';
    if (!$oFile->file_exists($sDirPath)) {
      $oFile->mkdir($sDirPath, 0777, true);
    }

    // Données
    $scope = $this->getScope($oParcelleInfos, $aData);

    // Cartes
    $aMaps = $this->getMaps(json_decode($aData['maps_json'], true));

    /* Génération du PDF*/
    // HTML par default -> dans utils/template
    if ($this->oProperties->get('cadastre_fiche_urba_print_template_id') == null) { 
      $sTemplatePath = $this->oProperties->replacePropertiesInString('@prop(cadastre_template_dir)/fiche_urbanisme.html'); 
      $aReturn = $this->transformPDF($sTemplatePath, null, 'A4', 'pdf', 'paysage', $scope, [], $aMaps, $sPath, $oFile, true); 
      // HTML custom -> provient des modèles d'impression 
    } else { 
      $oResult = $this->findTemplateById($this->oProperties->get('cadastre_fiche_urba_print_template_id')); 
      $aReturn = $this->transformPDF(null, $oResult[0]['definition'], $oResult[0]['rt_format_id'], 'pdf', $oResult[0]['rt_orientation_id'], $scope, [], $aMaps, $sPath, $oFile, true); 
    } 

    if ($aReturn instanceof Response) {
      return $aReturn;
    } else {
      return $this->formatResponse($aReturn);
    }
  }

  /**
   * Retourne le template à utiliser pour la fiche urba
   */
  public function getFicheUrbaTemplate(): Response {
    if ($this->oProperties->get('cadastre_fiche_urba_print_template_id') == null) {
      $sTemplatePath = $this->oProperties->replacePropertiesInString('@prop(cadastre_template_dir)/fiche_urbanisme.html');
      $sPrintTemplate = $this->oFileManager->getLocalFileAccessor()->file_get_contents($sTemplatePath);
    } else {
      $oResult = $this->findTemplateById($this->oProperties->get('cadastre_fiche_urba_print_template_id'));
      $sPrintTemplate = $oResult[0]['definition'];
    }

    // $sPrintTemplate = $this->getFicheUrbaTemplateAsString();
    $aReturn = [
      'data' => $sPrintTemplate,
    ];
    return $this->formatResponse($aReturn);
  }

  /**
   * Trouve les infos de la parcelle
   * Retourne un tableau ou une response en cas d'erreur
   * @return array|Response
   */
  private function getParcelleInfos() : array|Response {
    $oParcelleInfos = [];
    $aData = $this->requestToArray();
    $aUrba = [];
    $idPar = $aData['ID_PAR'];
    $sIdCom = substr($idPar, 0, 5);
    if (!$this->haveCommuneRights($sIdCom)) {
      $aReturn = [
        'message' => 'Erreur lors de la génération du fichier PDF de la fiche d\'urbanisme (Vous n\'avez pas les droits sur la commune).',
        'errorMessage' => 'ERROR_GENERIC_CONTROLLER_GET',
        'status' => 0];
      return $this->formatResponse($aReturn, Response::HTTP_UNAUTHORIZED);
    }

    // chargement de la connexion à la base de donnée
    $oEntityManager = $this->oDatabaseConnection->getUserEntityManager();
    $oUrbaRepository = $oEntityManager->getRepository(FicheUrbanisme::class);
    if (!$oUrbaRepository instanceof FicheUrbanismeRepository) {
      $aReturn = [
        'message' => 'Erreur lors de la génération du fichier PDF de la fiche d\'urbanisme (Vous n\'avez pas les droits sur la commune).',
        'errorMessage' => 'ERROR_GENERIC_CONTROLLER_GET',
        'status' => 0];
      return $this->formatResponse($aReturn, Response::HTTP_UNAUTHORIZED);
    }
    $aFicheUrba = $oUrbaRepository->findByIdPar($idPar);
    if (count($aFicheUrba) == 0) {
      $aReturn = [
        'message' => 'Erreur lors de la génération du fichier PDF de la fiche d\'urbanisme (Absence de Fiche d\'urbanisme).',
        'errorMessage' => 'ERROR_GENERIC_CONTROLLER_GET',
        'status' => 0];
      return $this->formatResponse($aReturn, Response::HTTP_NOT_FOUND);
    }
    foreach ($aFicheUrba as $urba) {
      $aUrba[] = $urba->getMajicData();
    }
    $oParcelleInfos['data'] = $aUrba;
    $aGeom = $this->getGeom($aData);
    $oParcelleInfos['data']['geometry'] = $aGeom;
    $geom = null;
    if ($aGeom != null) {
      $geom = $aGeom;
    }

    // Récupère les intersections
    $oParcelleInfos['data']['aIntersections'] = [];
    $oParcelleInfos['data']['aIntersectionsArray'] = [];

    // Infos complémentaires (paramétrées par l'administrateur)
    $iTolerance = $this->oProperties->get('cadastre_fiche_urba_intersect_tolerance');
    $bInterectIdCom = $this->oProperties->get('cadastre_fiche_urba_intersect_id_com');
    $aIntersectionViews = $this->oProperties->get('cadastre_fiche_urba_intersect_tables');
    $intersection = [];

    if (!empty($aIntersectionViews)) {
      $intersection = $this->getIntersection($bInterectIdCom, $geom, $aIntersectionViews, $iTolerance, $idPar);
    }
    $oParcelleInfos['data']['aIntersections'] = $intersection;
    return $oParcelleInfos;
  }

  /**
   * @param string $sTemplatePath le chemin du template
   * @param string $sTemplate définition du template (si sTemplatePath est vide)
   * @param string $size A4, A3
   * @param string $format .pdf
   * @param string $orientation portait ou paysage
   * @param array $aData les données
   * @param array $aIncludes images à inclure dans l'HTML
   * @param array $aMaps flux cartographique
   * @param string $sPath le chemin du PDF
   * @param Object $oFile management des fichiers
   * @param bool $bPaged
   * Methode pour générer l'excel et retourner le lien de téléchargement
   * @return Response|array
   */
  private function transformPDF(string|null $sTemplatePath, string|null $sTemplate, string $size, string $format, string $orientation, array $aData, array $aIncludes, array $aMaps, string $sPath, mixed $oFile, bool $bPaged = false) {
    if ($sTemplatePath) {
      if ($this->oFileManager->getLocalFileAccessor()->file_exists($sTemplatePath)) {
        $sTemplate = $oFile->file_get_contents($sTemplatePath);
      } else {
      $aReturn = [
        'message' => 'Le template n\'a pas été trouvé.',
        'errorMessage' => 'ERROR_GENERIC_CONTROLLER_GET',
        'status' => 0];
        return $this->formatResponse($aReturn, Response::HTTP_NOT_FOUND);
      }
    }
    $this->oPrintService->print($sTemplate, $size, $format, $orientation, $aData, $aIncludes, $aMaps, $sPath, $bPaged);
    /* Retourne le lien de téléchargement de l'Excel */
    $aReturn = [
      'data' => [
        'rapport' => $oFile->getProxyPassUrl($sPath),
      ],
      'status' => 1,
    ];
    return $aReturn;
  }

  /**
   * Traite les infos et retourne le scope utilisé dans le template
   * @param array $oParcelleInfos
   * @param array $aData
   * @return array
   */
  private function getScope(array $oParcelleInfos, array $aData) : array {
    $scope = [];

    // Scope envoyé par le front
    if (array_key_exists('scope_json', $aData) && is_string($aData['scope_json'])) {
      $scope = json_decode($aData['scope_json'], true);
    }

    // Infos parcelle
    $scope['ficheurb'] = $oParcelleInfos;

    // Proprietaires
    $scope['proprietaires'] = [];
    foreach ($oParcelleInfos['data'] as $key => $value) {
      if ($key !== 'geometry' && $key !== 'aIntersections' && $key !== 'aIntersectionsArray') {
        $prop = [];
        $prop['nom'] = $oParcelleInfos['data'][$key]['DDENOM'];
        $prop['adresse'] = $oParcelleInfos['data'][$key]['DLIGN4'];
        $prop['ville'] = $oParcelleInfos['data'][$key]['DLIGN6'];
        array_push($scope['proprietaires'], $prop);
      }
    }

    // Intersections
    $scope['intersections'] = [];
    $scope['intersectionsArray'] = [];
    foreach ($oParcelleInfos['data']['aIntersections'] as $key => $intersection) {
      $inter = [];
      $inter['title'] = $intersection['title'];
      foreach ($intersection['columns'] as $node) {
        $nod = [];
        $nod['label'] = $node['label'];
        foreach ($intersection['data'] as $intersectionData) {
          $nod['value'] = $intersectionData[$node['name']];
        }
        array_push($inter, $nod);
      }
      array_push($scope['intersections'], $inter);
    }

    if (!empty($this->oProperties->get('cadastre_fiche_urba_logo'))) {
      $scope['logo'] = $this->oProperties->get('cadastre_fiche_urba_logo');
    }
    if (!empty($this->oProperties->get('cadastre_fiche_urba_company'))) {
      $scope['company'] = $this->oProperties->get('cadastre_fiche_urba_company');
    }

    return $scope;
  }

  /**
   * Parse la liste des cartes à envoyer au service d'impression
   * @param array $aMapsJSON
   * @return array
   */
  private function getMaps(array $aMapsJSON) : array {
    foreach ($aMapsJSON as $key => $value) {
      if (array_key_exists('map_definition', $aMapsJSON[$key]) &&
        !empty($aMapsJSON[$key]['map_definition'])) {
        // Token
        $aMapsJSON[$key]['map_definition']['token'] = $this->oVitisRequest->getToken();

        // Feature zoom
        if (!empty($this->oProperties->get('cadastre_fiche_urba_feature_zoom'))) {
          $aMapsJSON[$key]['map_definition']['features_zoom'] = $this->oProperties->get('cadastre_fiche_urba_feature_zoom');
        } else {
          $aMapsJSON[$key]['map_definition']['features_zoom'] = 100;
        }
      }
    }

    return $aMapsJSON;
  }

  /**
   * @param string $sIdTemplate id du template
   * Cette méthode permet d'obtenir le template créé par l'utilisateur dans les modèles d'impression
   *  @return array
   */
  private function findTemplateById(string $sIdTemplate) : array {
    $qb = $this->oDatabaseConnection->getQueryAccessor()->createQueryBuilder();
    $qb->select('*')
      ->from('s_vmap_2.printtemplate', 'a')
      ->where('a.printtemplate_id=:sIdTemplate')
      ->setParameter('sIdTemplate', $sIdTemplate, 'string');
    try {
      $result = $qb->execute()->fetchAllAssociative();
    } catch (\Exception $e) {
      $result = [
        'message' => 'Erreur lors de la recherche du template:' . $e->getMessage(),
        'errorMessage' => 'ERROR_GENERIC_CONTROLLER_GET',
        'status' => 0];
    }
    return $result;
  }
}
