<?php

namespace App\Module\Cadastre\WebService\Controller;

use App\Module\Cadastre\WebService\Entity\Majic\CompteCommunal;
use App\Module\Cadastre\WebService\Entity\Majic\AbstractCompteCommunal;
use App\Module\Cadastre\WebService\Entity\Majic\CompteCommunalParcelle;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\FileAccessor\LocalAccessor;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\PrintService\PrintTemplate;
use App\Module\Vitis\WebService\Service\VitisRequest;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller permettant la génération du relevé de propriété
 * @author Alizée Buckel <alizee.buckel@veremes.com>
 */

class ReleveDeProprieteController extends AbstractCadastreController {
  /**
   * sous-dossier et nom du template par type de rapport
   *
   * @var array
   */
  protected const VAR_BY_RP_TYPE = [
    'standard' => [
      'folder' => '@prop(cadastre_report_dir)/@annee/standard',
      'template' => '@prop(cadastre_template_dir)/rp_standard.html',
    ],
    'tiers' => [
      'folder' => '@prop(cadastre_report_dir)/@annee/tiers',
      'template' => '@prop(cadastre_template_dir)/rp_tiers.html',
    ],
    'parcelle' => [
      'folder' => '@prop(cadastre_report_dir)/@annee/parcelle',
      'template' => '@prop(cadastre_template_dir)/rp_standard.html',
    ],
    'parcelle_tiers' => [
      'folder' => '@prop(cadastre_report_dir)/@annee/parcelle_tiers',
      'template' => '@prop(cadastre_template_dir)/rp_parcelle_tiers.html',
    ],
  ];
  /**
   * @var PrintTemplate
   */
  protected $oPrintService;

  /**
   * Initialise le controler et instancie les services requis
   * @param Properties $oProSrvc Service d'accès aux properties
   * @param VitisRequest $oRequest Service d'abstraction de la requête Symfony
   * @param Route $oRoute Service d'accès à la configuration de la route
   * @param LoggerManager $oLogger Service d'utilsation des logs
   * @param DatabaseConnection $oDbConnection Service d'accès à la base de données
   * @param FileManager $oFileManager Service de gestion des fichiers
   * @param PrintTemplate $oTemplate Service de manipulation des token JWE
   */
  public function __construct(
    Properties $oPropSrvc,
    VitisRequest $oRequest,
    Route $oRoute,
    LoggerManager $oLogger,
    DatabaseConnection $oDbConnection,
    FileManager $oFileManager
  ) {
    parent::__construct($oPropSrvc, $oRequest, $oRoute, $oLogger, $oDbConnection, $oFileManager);
    $this->oPrintService = new PrintTemplate($oLogger, $oPropSrvc, $oFileManager);
  }

  /**
   * génére le relevé de propriété par id_dnupro
   * Correspond a l'ancien standard et tiers
   * @param array $idDnupro la liste des idDnupro
   * @param string $sIdCom liste des IdCom
   * @return array|CompteCommunal
   */
  protected function getRPByDnupro($idDnupro, $sIdCom): ?CompteCommunal {
    if (!$this->haveCommuneRights($sIdCom)) {
      $this->formatErrorMissingRightforRP('Vous n\'avez pas les droits sur la commune', Response::HTTP_UNAUTHORIZED);
      return null;
    }
    // chargement de la connexion à la base de donnée
    $oEntityManager = $this->oDatabaseConnection->getAdminEntityManager();
    $oCCRepository = $oEntityManager->getRepository(CompteCommunal::class);
    $aCompteCom = $oCCRepository?->findById($idDnupro);
    if (empty($aCompteCom)) {
      $this->formatErrorMissingRightforRP('Absence de Compte Communal', Response::HTTP_NOT_FOUND);
      return null;
    }
    return $aCompteCom[0];
  }

  /**
   * récupère les 5 premier caractère de l'identifiant ud compte
   *
   * @param  string $sIdCompte dnupro ou id_par
   * @return string idcom
   */
  protected function extractIdComFromCompteId(string $sIdCompte) : string {
    return substr($sIdCompte, 0, 5);
  }

  /**
   * Formatage de l'erreur pour l agénération du RP
   *
   * @param  string   $sMessage
   * @param  int      $iCode
   * @return Response
   */
  protected function formatErrorMissingRightforRP(string $sMessage, int $iCode): Response {
    $sDfMessage = 'Erreur lors de la génération du fichier PDF de relevé de propriété';
    if (!empty($sMessage)) {
      $sDfMessage .= '(' . $sMessage . ')';
    }

    return $this->formatGenericError('ERROR_GENERIC_CONTROLLER_GET', '', $sDfMessage, $iCode);
  }

  /**
   * génére le relevé de propriété par id_par
   * Correspond a l'ancien parcelle et parcelle tiers
   * @param array $idPar la liste des idPar
   * @param string $sIdCom liste des IdCom
   * @return array|CompteCommunalParcelle
   */
  protected function getRPByPar($idPar, $sIdCom): ?CompteCommunalParcelle {
    if (!$this->haveCommuneRights($sIdCom)) {
      $this->formatErrorMissingRightforRP('Vous n\'avez pas les droits sur la commune', Response::HTTP_UNAUTHORIZED);
      return null;
    }
    // chargement de la connexion à la base de donnée
    $oEntityManager = $this->oDatabaseConnection->getAdminEntityManager();
    $oCCRepository = $oEntityManager->getRepository(CompteCommunalParcelle::class);
    $aCompteCom = $oCCRepository?->findById($idPar);
    if (empty($aCompteCom)) {
      $this->formatErrorMissingRightforRP('Absence de Compte Communal', Response::HTTP_NOT_FOUND);
      return null;
    }
    return $aCompteCom[0];
  }

  /**
   * génére un pdf pour le relevé de propriété standard
   * Correspond a l'ancien standard
   * @return Response
   */
  public function GetRPStandard(): Response {
    $aData = $this->requestToArray();
    if (!array_key_exists('IDDNUPRO', $aData)) {
      return $this->formatErrorMissingParameter('IDDNUPRO is missing');
    }
    $sDestPath = $this->generateRpByDnupro_($aData['IDDNUPRO'], 'standard');
    // en erreur, la réponse est déjà formaté on envoi
    if (empty($sDestPath)) {
      return $this->oResponse;
    }
    return $this->formatRpResponse($sDestPath);
  }

  /**
   * Trouve le comptecommunal lié au propriétaire et lance la génération du RP avec
   *
   * @param  string          $idDnupro
   * @return string|null
   */
  protected function generateRpByDnupro_(string $idDnupro, string $sReporType) : ?string {
    $sIdCom = $this->extractIdComFromCompteId($idDnupro);

    /* Récupération des données*/
    $oCompteCom = $this->getRPByDnupro($idDnupro, $sIdCom);
    if ($oCompteCom == null) {
      return null;
    }
    $sFileName = $idDnupro . '.pdf';
    return $this->generateRpWithCC_($oCompteCom, $sReporType, $sFileName);
  }
  /**
   * Trouve le comptecommunal lié au propriétaire et lance la génération du RP avec
   *
   * @param  string          $idDnupro
   * @return string|null
   */
  protected function generateRpByIdPar_(string $idPar, string $sReporType) : ?string {
    $sIdCom = $this->extractIdComFromCompteId($idPar);
    /* Récupération des données*/
    $oCompteCom = $this->getRPByPar($idPar, $sIdCom);
    if ($oCompteCom == null) {
      return null;
    }
    $sFileName = $idPar . '.pdf';
    return $this->generateRpWithCC_($oCompteCom, $sReporType, $sFileName);
  }

  /**
   * Lance la génération du RP à parti d'un compte communal de la Majic lib
   *
   * @param  AbstractCompteCommunal $oCompteCom
   * @param  string                 $subfolder
   * @param  string                 $sTplName
   * @param  string                 $sFilename
   * @return string|null
   */
  private function generateRpWithCC_(AbstractCompteCommunal $oCompteCom, string $sReporType, string $sFilename) : ?string {
    $aData = $oCompteCom->getMajicData();
    $subfolder = static::VAR_BY_RP_TYPE[$sReporType]['folder'];
    $sTplName = static::VAR_BY_RP_TYPE[$sReporType]['template'];

    /* Génération du dossier */
    $sPath = $this->formatRpDestPath($oCompteCom->getAnneeReference(), $subfolder, $sFilename);
    $sTemplatePath = $this->getTemplatePath($sTplName);
    /* Génération du PDF*/
    return $this->printRpPdf($sTemplatePath, $aData, $sPath);

  }

  /**
   * permet de mettre en forme le chemin et de tester l'existence du template
   *
   * @param  string      $sTemplateName
   * @return string|null
   */
  protected function getTemplatePath(string $sTemplateName) : ?string {
    $sPath = $this->oProperties->replacePropertiesInString($sTemplateName);
    $oFa = $this->oFileManager->getLocalFileAccessor();
    return ($oFa->file_exists($sPath)) ?  $sPath : null;
  }

  /**
   * génére un pdf pour le relevé de propriété tiers
   * Correspond a l'ancien tiers
   * @return Response
   */
  public function GetRPTiers(): Response {
    $aData = $this->requestToArray();
    if (!array_key_exists('IDDNUPRO', $aData)) {
      return $this->formatErrorMissingParameter('IDDNUPRO is missing');
    }
    $sDestPath = $this->generateRpByDnupro_($aData['IDDNUPRO'], 'tiers');
    // en erreur, la réponse est déjà formaté on envoi
    if (empty($sDestPath)) {
      return $this->oResponse;
    }
    return $this->formatRpResponse($sDestPath);
  }

  /**
   * Formate le chemin destination
   * @param  string|int $sAnneeRef année de référence
   * @param  string     $sSubfolder sous dossier à utiliser pour le type de rapport
   * @param  string     $sFilename nom du fichier final
   * @return string
   */
  protected function formatRpDestPath (string|int $sAnneeRef, string $sFolder, string $sFilename) : string {
    $sBindPath = $this->oProperties->replacePropertiesInString($sFolder);
    $sPath = str_replace('@annee', $sAnneeRef, $sBindPath);
    return $sPath . '/' . $sFilename;
  }

  /**
   * génére un pdf pour le relevé de propriété parcelle
   * Correspond a l'ancien parcelle
   * @return Response
   */
  public function GetRPParcelle(): Response {
    $aData = $this->requestToArray();
    if (!array_key_exists('ID_PAR', $aData)) {
      return $this->formatErrorMissingParameter('ID_PAR is missing');
    }
    $sDestPath = $this->generateRpByIdPar_($aData['ID_PAR'], 'parcelle');
    // en erreur, la réponse est déjà formaté on envoi
    if (empty($sDestPath)) {
      return $this->oResponse;
    }
    return $this->formatRpResponse($sDestPath);
  }

  /**
   * génére un pdf pour le relevé de propriété parcelle tiers
   * Correspond a l'ancien parcelle tiers
   * @return Response
   */
  public function GetRPParcelleTiers(): Response {
    $aData = $this->requestToArray();
    // check si l'id par est présent dans la requête
    if (!array_key_exists('ID_PAR', $aData)) {
      return $this->formatErrorMissingParameter('ID_PAR is missing');
    }
    $sDestPath = $this->generateRpByIdPar_($aData['ID_PAR'], 'parcelle_tiers');
    // en erreur, la réponse est déjà formaté on envoi
    if (empty($sDestPath)) {
      return $this->oResponse;
    }
    return $this->formatRpResponse($sDestPath);
  }

  /**
   * Methode pour générer l'excel et retourner le lien de téléchargement
   *
   * @param  string      $sTemplatePath le chemin du template
   * @param  array       $aData les données
   * @param  string      $sPath le chemin du PDF
   * @param  bool        $bPaged
   * @return string|null $sDestPath si tout va bien null sinon, formate l'erreur si retourne null
   */
  protected function printRpPdf(string $sTemplatePath, array $aData, string $sPath, bool $bPaged = false) : ?string {
    $oLocalFa = $this->oFileManager->getLocalFileAccessor();
    // Gestion cache, le fichier est déjà généré
    if ($oLocalFa->file_exists($sPath)) {
      return $sPath;
    }
    if (empty($sTemplatePath) || !$oLocalFa->file_exists($sTemplatePath)){
      $this->formatGenericError('ERROR_GENERIC_CONTROLLER_GET', '', 'Le template n\'a pas été trouvé.');
      return null;
    }
    // génération du dossier parent
    $sDirPath = \dirname($sPath);
    if (!$oLocalFa->file_exists($sDirPath)) {
      $oLocalFa->mkdir($sDirPath, LocalAccessor::DEFAULT_LINUX_PERMISSION, true);
    }
    $sTemplate = $oLocalFa->file_get_contents($sTemplatePath);
    $this->oPrintService->print($sTemplate, 'A4', 'pdf', 'paysage', $aData, [], [], $sPath, $bPaged);
    // si le fichier destination existe alors on retourne sinon on part en erreur
    if ($oLocalFa->file_exists($sPath)) {
      return $sPath;
    }
    $this->formatGenericError('ERROR_GENERIC_CONTROLLER_GET', '', 'Le Relevé n\'a pas pu être généré.');
    return null;
  }

  /**
   * génère l'url vers le file Accessor et formate la réponse pour l'API
   *
   * @param  string   $sTplPath
   * @param  array    $aData
   * @param  string   $sPath
   * @param  bool     $bPaged
   * @return Response
   */
  protected function formatRpResponse(string $sPath) : Response {
    $oLocalFa = $this->oFileManager->getLocalFileAccessor();
    return $this->formatResponse([
      'data' => [
        'rapport' => $oLocalFa->getProxyPassUrl($sPath),
      ],
      'status' => 1,
    ]);
  }
}
