<?php

namespace App\Module\Cadastre\WebService\Entity\Majic;

use App\Module\Cadastre\WebService\Collection\BatiCollection;
use App\Module\Cadastre\WebService\Collection\ParcelleCollection;
use App\Module\Cadastre\WebService\Collection\ProprietaireCollection;
use App\Module\Cadastre\WebService\Entity\Majic\Bati;
use App\Module\Cadastre\WebService\Entity\Majic\Parcelle;
use App\Module\Cadastre\WebService\Entity\Majic\Proprietaire;
use App\Module\Cadastre\WebService\Repository\MajicCommuneRepository;
use App\Module\Cadastre\WebService\Repository\ProprietaireRepository;
use App\Module\Cadastre\WebService\Service\MajicUtils;
use App\Module\Vitis\WebService\Entity\AbstractEntity;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use Doctrine\ORM\EntityManager;
use Doctrine\ORM\Mapping\Column;
use Doctrine\ORM\Mapping\Entity;
use Doctrine\ORM\Mapping\Id;
use Doctrine\ORM\Mapping\MappedSuperclass;
use Doctrine\ORM\Mapping\Table;
use Doctrine\Persistence\Mapping\ClassMetadata;
use Doctrine\Persistence\ObjectManager;

/**
 *  Objet abstrait pour CompteCommunal et CompteCommunalParcelle
 *
 * @author Alizée Buckel <alizee.buckel@veremes.com>
 *
 * @MappedSuperclass
 */
abstract class AbstractCompteCommunal extends AbstractEntity implements MajicInterface {
  /**
   * @var EntityManager|null
   */
  protected $oEntityManager;

  /**
   * @var LoggerManager|null
   */
  protected $oLoggerManager;

  /**
   * @var string
   * @Column(type="string", name="ccodep")
   */
  protected $ccodep;

  /**
   * @var string
   * @Column(type="string", name="ccocom")
   */
  protected $ccocom;

  /**
   * @var string
   * @Column(type="string", name="libcom")
   */
  protected $libcom;

  /**
   * @var string
   * @Column(type="string", name="dnupro")
   */
  protected $dnupro;

  /**
   * @var string
   * @Column(type="string", name="ccodir")
   */
  protected $ccodir;

  /**
   * @var string
   * @Column(type="string", name="id_dnupro")
   * @Id
   */
  protected $id_dnupro;

  /**
   * @var ProprietaireCollection
   */
  protected $aProp;

  /**
   * @var BatiCollection
   */
  protected $aBati;

  /**
   * @var ParcelleCollection
   */
  protected $aParcelle;

  public function __construct() {
    $this->aProp = new ProprietaireCollection();
    $this->aBati = new BatiCollection();
    $this->aParcelle = new ParcelleCollection();
  }

  /**
   * remplis la liste des proprietaires si necessaire
   * @return ProprietaireCollection|array
   */
  private function fillProprietaireArray() {
    $aProp = [];
    if ($this->aProp instanceof ProprietaireCollection && !$this->aProp->isEmpty()) { 
      return $this->aProp;
    }
    if (!isset($this->oEntityManager)) {
      return [];
    }
    try {
      $proprietaireRepository = $this->oEntityManager->getRepository(Proprietaire::class);
      if (!$proprietaireRepository instanceof ProprietaireRepository) {
        throw new \Exception("Repository is not a ProprietaireRepository", 1);
      }
      $aProp = $proprietaireRepository->findByIdDnupro($this->id_dnupro);
      $this->aProp = new ProprietaireCollection();
      foreach ($aProp as $prop) {
        $this->aProp->add($prop);
      }
      return $this->aProp;
    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('Cannot get proprietaire list for compte communal ', [$this, $e]);
      return [];
    }
  }

  /**
   * Cette méthode permet de récupérer les Proprietaires
   * @return ProprietaireCollection|null
   */
  public function getProprietaireCollection(): ?ProprietaireCollection {
    return $this->fillProprietaireArray();
  }

  /**
   * Cette méthode permet de récupérer la liste des bâtis
   * @return BatiCollection|null
   */
  public function getBatiCollection(): ?BatiCollection {
    return $this->fillBatiArray();
  }

  /**
   * Cette méthode permet de récupérer les Parcelles
   * @return ParcelleCollection|null
   */
  public function getParcelleCollection(): ?ParcelleCollection {
    return $this->fillParcelleArray();
  }

  /**
   * @return string|null
   * Cette méthode permet de récupérer l'année de référence pour une commune
   * Méthode modifiée
   */
  public function getAnneeReference(): ?string {
    if (!isset($this->oEntityManager)) {
      return null;
    }
    try {
      $MajicCommuneRepository = $this->oEntityManager->getRepository(Commune::class);
      $sIdCom = $this->ccodep . $this->ccocom;
      if (!$MajicCommuneRepository instanceof MajicCommuneRepository) {
        throw new \Exception("Repository is not a MajicCommuneRepository", 1);
      }
      return $MajicCommuneRepository->getAnneeReference($sIdCom);
    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('Cannot get bati list for compte communal ', [$this, $e]);
      return null;
    }
  }

  /**
   * @return string
   */
  public function getDepdir() {
    return $this->ccodep . ' ' . $this->ccodir;
  }

  /**
   * @return string
   */
  public function getCom() {
    return $this->ccocom . ' ' . $this->libcom;
  }

  /**
   * @return integer
   * Cette méthode permet de calculer le revenu imposable total du bâti
   * Méthode dans BatiCollection
   */
  public function getRevImpoBati(): int {
    return $this->getBatiCollection()->getRevImpoBati();
  }

  /**
   * @return float
   * Cette méthode permet de récupérer la valeur locative totale du bâti
   * Méthode dans BatiCollection
   */
  public function getValeurLocativeBati(): float {
    return $this->getBatiCollection()->getValeurLocativeBati();
  }

  /**
   * @return integer
   * Cette méthode permet de récupérer la valeur de l'attribut 'contenance' pour le compte communal avec exonération dans le tableau aParcelle
   * Méthode dans ParcelleCollection
   */
  public function getContenance(): int {
    return $this->getParcelleCollection()->getContenance();
  }

  /**
   * @return integer
   * Cette méthode permet de calculer la contenance totale des parcelles
   * Méthode dans ParcelleCollection
   */
  public function getTotalContenance(): int {
    return $this->getParcelleCollection()->getTotalContenance();
  }

  /**
   * @return integer
   * Cette méthode permet de calculer la contenance totale des parcelles en mettre carré
   * Méthode dans ParcelleCollection
   */
  public function getTotalContenanceM2(): int {
    return $this->getParcelleCollection()->getTotalContenanceM2();
  }

  /**
   * @return string|null
   * Cette méthode permet de récupérer les caractères correspondants à la colonne HA depuis la méthode contenance
   */
  public function getContenanceHA(): ?string{
    $sContenanceHA = strval(intval($this->getContenance() / 10000));
    if ($sContenanceHA == 0) {
      return null;
    }
    return $sContenanceHA;
  }

  /**
   * @return string|null
   * Cette méthode permet de récupérer les caractères correspondants à la colonne A depuis depuis la méthode contenance
   */
  public function getContenanceA(): ?string{
    $sContenanceA = strval(intval(($this->getContenance() % 10000) / 100));
    if ($this->getContenanceHA() == null && $sContenanceA == 0) {
      return null;
    } else {
      if ($this->getContenanceHA() == null && strlen($sContenanceA) == 2 && substr($sContenanceA, 0, 1) == 0) {
        return null . substr($sContenanceA, 1, 1);
      } else {
        if ($this->getContenanceHA() != null && strlen($sContenanceA) == 1) {
          return 0 . $sContenanceA;
        }
      }
    }
    return $sContenanceA;
  }

  /**
   * @return string
   * Cette méthode permet de récupérer les caractères correspondants à la colonne CA depuis la méthode contenance
   */
  public function getContenanceCA(): string {
    $sContenanceCA = strval(intval($this->getContenance() % 100));
    if ($this->getContenanceA() != 0) {
      if (strlen($sContenanceCA) == 1) {
        return 0 . $sContenanceCA;
      }
    }
    return $sContenanceCA;
  }

  /**
   * @return integer
   * Cette méthode permet de calculer l'exonération communale totale du bâti
   * Méthode dans BatiCollection
   * Méthode corrigée
   */
  public function getExoComBati(): int {
    return $this->getBatiCollection()->getExoComBati($this->getProprietaireCollection()->getLCcodem());
  }

  /**
   * @return integer
   * Cette méthode permet de calculer le revenu imposable communal du bâti en fonction du revenu imposable total (bâti) et de l'exonération communale totale (bâti)
   * Métode modifiée
   */
  public function getRevImpoComBati(): int {
    return $this->getRevImpoBati() - $this->getExoComBati();
  }

  /**
   * @return integer
   * Cette méthode permet de calculer l'exonération départementale totale du bâti
   * Méthode dans BatiCollection
   * Méthode corrigée
   */
  public function getExoDepBati(): int {
    return $this->getBatiCollection()->getExoDepBati($this->getProprietaireCollection()->getLCcodem());
  }

  /**
   * @return integer
   * Cette méthode permet de calculer le revenu imposable départemental du bâti en fonction du revenu imposable total (bâti) et de l'exonération départemantale totale (bâti)
   */
  public function getRevImpoDepBati(): int {
    return $this->getRevImpoBati() - $this->getExoDepBati();
  }

  /**
   * @return integer
   * Cette méthode permet de calculer l'exonération régionale totale du bâti
   * Méthode dans BatiCollection
   */
  public function getExoRegBati(): int {
    return $this->getBatiCollection()->getExoRegBati($this->getProprietaireCollection()->getLCcodem());
  }

  /**
   * @return integer
   * Cette méthode permet de calculer le revenu imposable régional du bâti en fonction du revenu imposable total (bâti) et de l'exonération régionale totale (bâti)
   */
  public function getRevImpoRegBati(): int {
    return $this->getRevImpoBati() - $this->getExoRegBati();
  }

  /**
   * @return integer
   * Cette méthode permet de calculer l'éxonération communale totale des parcelles
   * Méthode dans ParcelleCollection
   */
  public function getExoComPar(): int {
    return $this->getParcelleCollection()->getExoComPar();
  }

  /**
   * @return integer
   * Cette méthode permet de calculer le revenu imposable total des parcelles
   * Méthode dans ParcelleCollection
   */
  public function getRevImpoPar(): int {
    return $this->getParcelleCollection()->getRevImpoPar();
  }

  /**
   * @return integer
   * Cette méthode permet de calculer le revenu imposable communal des parcelles en fonction du revenu imposable total (parcelles) et de l'exonération communale totale (parcelles)
   */
  public function getRevImpoComPar(): int {
    return $this->getRevImpoPar() - $this->getExoComPar();
  }

  /**
   * @return integer
   * Cette méthode permet de calculer l'exonération départementale totale des parcelles
   * Méthode dans ParcelleCollection
   */
  public function getExoDepPar(): int {
    return $this->getParcelleCollection()->getExoDepPar();
  }

  /**
   * @return integer
   * Cette méthode permet de calculer le revenu imposable départemental des parcelles en fonction du revenu imposable total (parcelles) et de l'exonération (départementale et régionale) totale (parcelles)
   */
  public function getRevImpoDepPar(): int {
    return $this->getRevImpoPar() - $this->getExoDepPar();
  }

  /**
   * @return integer
   * Cette méthode permet de calculer l'exonération régionale totale des parcelles
   * Méthode dans ParcelleCollection
   */
  public function getExoRegPar(): int {
    return $this->getParcelleCollection()->getExoRegPar();
  }

  /**
   * @return integer
   * Cette méthode permet de calculer le revenu imposable régional des parcelles en fonction du revenu imposable total (parcelles) et de l'exonération (départementale et régionale) totale (parcelles)
   */
  public function getRevImpoRegPar(): int {
    return $this->getRevImpoPar() - $this->getExoRegPar();
  }

  /**
   * @return float
   * Cette méthode permet de calculer le revenu imposable total du bâti
   * Méthode dans BatiCollection
   */
  public function getRevImpoBatifloat(): float {
    return $this->getBatiCollection()->getRevImpoBatifloat();
  }

  /**
   * @return float
   * Cette méthode permet de calculer le revenu imposable total des parcelles
   * Méthode dans ParcelleCollection
   */
  public function getRevImpoParfloat(): float {
    return $this->getParcelleCollection()->getRevImpoParfloat();
  }

  /**
   * @return float
   * Cette méthode permet de calculer l'exonération communale totale du bâti
   * Méthode dans BatiCollection
   * Méthode corrigée
   */
  public function getExoComBatifloat(): float {
    return $this->getBatiCollection()->getExoComBatifloat(($this->getProprietaireCollection()->getLCcodem()));
  }

  /**
   * @return float
   * Cette méthode permet de calculer le revenu imposable communal du bâti en fonction du revenu imposable total (bâti) et de l'exonération communale totale (bâti)
   * Méthode corrigée
   */
  public function getRevImpoComBatifloat(): float {
    return $this->getRevImpoBatifloat() - $this->getExoComBatifloat();
  }

  /**
   * @return float
   * Cette méthode permet de calculer l'exonération départementale totale du bâti
   * Méthode dans BatiCollection
   * Méthode corrigée
   */
  public function getExoDepBatifloat(): float {
    return $this->getBatiCollection()->getExoDepBatifloat(($this->getProprietaireCollection()->getLCcodem()));
  }

  /**
   * @return float
   * Cette méthode permet de calculer le revenu imposable départemental du bâti en fonction du revenu imposable total (bâti) et de l'exonération départemantale totale (bâti)
   */
  public function getRevImpoDepBatifloat(): float {
    return $this->getRevImpoBatifloat() - $this->getExoDepBatifloat();
  }

  /**
   * @return float
   * Cette méthode permet de calculer l'exonération régionale totale du bâti
   * Méthode dans BatiCollection
   */
  public function getExoRegBatifloat(): float {
    return $this->getBatiCollection()->getExoRegBatifloat(($this->getProprietaireCollection()->getLCcodem()));
  }

  /**
   * @return float
   * Cette méthode permet de récupérer le revenu cadastral total des parcelles
   * Méthode dans ParcelleCollection
   */
  public function getRevenuCadastralTotal(): float {
    return $this->getParcelleCollection()->getRevenuCadastralTotal();
  }

  /**
   * @return float
   * Cette méthode permet de calculer le revenu imposable régional du bâti en fonction du revenu imposable total (bâti) et de l'exonération régionale totale (bâti)
   */
  public function getRevImpoRegBatifloat(): float {
    return $this->getRevImpoBatifloat() - $this->getExoRegBatifloat();
  }

  /**
   * @return float
   * Cette méthode permet de calculer l'éxonération communale totale des parcelles
   * Méthode dans ParcelleCollection
   * Méthode modifiée
   */
  public function getExoComParfloat(): float {
    return $this->getParcelleCollection()->getExoComParfloat();
  }

  /**
   * @return float
   * Cette méthode permet de calculer le revenu imposable communal des parcelles en fonction du revenu imposable total (parcelles) et de l'exonération communale totale (parcelles)
   * Métode modifiée
   */
  public function getRevImpoComParfloat(): float {
    $fRevImpoComPar = MajicUtils::setFloatToText($this->getRevImpoParfloat() - $this->getExoComParfloat());
    return floatval($fRevImpoComPar);
  }

  /**
   * @return float
   * Cette méthode permet de calculer l'exonération départementale totale des parcelles
   * Méthode dans ParcelleCollection
   */
  public function getExoDepParfloat(): float {
    return $this->getParcelleCollection()->getExoDepParfloat();
  }

  /**
   * @return float
   * Cette méthode permet de calculer le revenu imposable départemental des parcelles en fonction du revenu imposable total (parcelles) et de l'exonération (départementale et régionale) totale (parcelles)
   * Métode modifiée
   */
  public function getRevImpoDepParfloat(): float {
    $fRevImpoDepPar = MajicUtils::setFloatToText($this->getRevImpoParfloat() - $this->getExoDepParfloat());
    return floatval($fRevImpoDepPar);
  }

  /**
   * @return float
   * Cette méthode permet de calculer l'exonération régionale totale des parcelles
   * Méthode dans ParcelleCollection
   */
  public function getExoRegParfloat(): float {
    return $this->getParcelleCollection()->getExoRegParfloat();
  }

  /**
   * @return float
   * Cette méthode permet de calculer le revenu imposable régional des parcelles en fonction du revenu imposable total (parcelles) et de l'exonération (départementale et régionale) totale (parcelles)
   * Métode modifiée
   */
  public function getRevImpoRegParfloat(): float {
    return $this->getRevImpoParfloat() - $this->getExoRegParfloat();
  }

  /**
   * @return float
   * Cette méthode permet de calculer l'éxonération des taxes additionnelles totale des parcelles
   * Méthode dans ParcelleCollection
   */
  public function getExoTaxeADParfloat(): float {
    return $this->getParcelleCollection()->getExoTaxeADParfloat();
  }

  /**
   * @return float
   * Cette méthode permet de calculer le revenu imposable des taxes additionnelles des parcelles en fonction du revenu imposable total (parcelles) et de l'exonération communale totale (parcelles)
   * Métode modifiée
   */
  public function getRevImpoTaxeADParfloat(): float {
    $fRevImpoTaxeADPar = MajicUtils::setFloatToText($this->getRevImpoParfloat() - $this->getExoTaxeADParfloat());
    return floatval($fRevImpoTaxeADPar);

  }

  /**
   * @return float
   * Cette méthode permet de calculer le revenu imposable des taxes additionnelles des parcelles en fonction du revenu imposable total (parcelles) et de l'exonération communale totale (parcelles)
   * Méthode dans ParcelleCollection
   * Méthode modifiée
   */
  public function getExoTaxeADPar(): float {
    return $this->getParcelleCollection()->getExoTaxeADPar();
  }

  /**
   * @return void
   * Cette méthode permet de récupérer l'année de référence
   * Méthode dans ParcelleCollection
   */
  public function getEmprises(EntityManager $oEntityManager): void {
    $this->getParcelleCollection()->getEmprises($oEntityManager);
  }

  /**
   * @return integer
   * Cette méthode permet de calculer le revenu imposable des taxes additionnelles des parcelles en fonction du revenu imposable total (parcelles) et de l'exonération communale totale (parcelles)
   * Métode modifiée
   */
  public function getRevImpoTaxeADPar(): int {
    return intval($this->getRevImpoPar() - $this->getExoTaxeADPar());
  }

  /**
   * @return float
   * Cette méthode permet de calculer l'emprise bâtie totale des parcelles.
   * Méthode dans ParcelleCollection
   * Métode modifiée
   */
  public function getTotalContenanceBati(): float {
    return $this->getParcelleCollection()->getTotalContenanceBati();
  }

  /**
   * @return array
   * Cette méthode permet de construire un tableau correspondant à CompteCommunal dans le relevé de propriété
   * Il contient Propriétaire, Bati et Parcelle
   */
  public function getMajicData(): array{
    $result = [
      'aProp' => $this->getProprietaireCollection()->getMajicData(),
      'aBat' => $this->getBatiCollection()->getMajicData(),
      'aPar' => $this->getParcelleCollection()->getMajicData(),
      'codeinsee' => null, // ABS
      'annee_de_maj' => $this->getAnneeReference(),
      'dep_dir' => $this->getDepdir(),
      'code_dep' => $this->ccodep,
      'code_com' => $this->ccocom,
      'com' => $this->getCom(),
      'libcom' => $this->libcom,
      'role' => null,
      'vue' => null,
      'numero_communal' => $this->dnupro,
      'rev_impo_bati' => $this->getRevImpoBati(),
      'rev_impo_par' => $this->getRevImpoPar(),
      'total_contenance' => $this->getTotalContenance(),
      'total_contenance_m2' => $this->getTotalContenanceM2(),
      'total_revenu_cadastral' => $this->getRevenuCadastralTotal(),
      'total_valeur_locative' => $this->getValeurLocativeBati(),
      'emprise_totale_intersect' => $this->getTotalContenanceBati(), // Lié aux intersections, chargées que pour la fiche d'urbanisme
      'ha' => $this->getContenanceHA(),
      'a' => $this->getContenanceA(),
      'ca' => $this->getContenanceCA(),
      'exo_com_bati' => $this->getExoComBati(),
      'rev_impo_com_bati' => $this->getRevImpoComBati(),
      'exo_dep_bati' => $this->getExoDepBati(),
      'rev_impo_dep_bati' => $this->getRevImpoDepBati(),
      'exo_reg_bati' => $this->getExoRegBati(),
      'rev_impo_reg_bati' => $this->getRevImpoRegBati(),
      'exo_com_par' => $this->getExoComPar(),
      'rev_impo_com_par' => $this->getRevImpoComPar(),
      'exo_dep_par' => $this->getExoDepPar(),
      'rev_impo_dep_par' => $this->getRevImpoDepPar(),
      'exo_reg_par' => $this->getExoRegPar(),
      'rev_impo_reg_par' => $this->getRevImpoRegPar(),
      'rev_impo_bati_float' => $this->getRevImpoBatifloat(),
      'rev_impo_par_float' => $this->getRevImpoParfloat(),
      'exo_com_bati_float' => $this->getExoComBatifloat(),
      'rev_impo_com_bati_float' => $this->getRevImpoComBatifloat(),
      'exo_dep_bati_float' => $this->getExoDepBatifloat(),
      'rev_impo_dep_bati_float' => $this->getRevImpoDepBatifloat(),
      'exo_reg_bati_float' => $this->getExoRegBatifloat(),
      'rev_impo_reg_bati_float' => $this->getRevImpoRegBatifloat(),
      'exo_com_par_float' => $this->getExoComParfloat(),
      'rev_impo_com_par_float' => $this->getRevImpoComParfloat(),
      'exo_dep_par_float' => $this->getExoDepParfloat(),
      'rev_impo_dep_par_float' => $this->getRevImpoDepParfloat(),
      'exo_reg_par_float' => $this->getExoRegParfloat(),
      'rev_impo_reg_par_float' => $this->getRevImpoRegParfloat(),
      'exo_tax_ad_par_float' => $this->getExoTaxeADParfloat(),
      'rev_impo_tax_ad_par_float' => $this->getRevImpoTaxeADParfloat(),
      'exo_tax_ad_par' => $this->getExoTaxeADPar(),
      'rev_impo_tax_ad_par' => $this->getRevImpoTaxeADPar(),
    ];
    return $result;
  }

  /**
   * remplis la liste des batis si necessaire
   * @return BatiCollection|array
   */
  abstract protected function fillBatiArray();

  /**
   * remplis la liste des parcelles si necessaire
   * @return ParcelleCollection|array
   */
  abstract protected function fillParcelleArray();

  /* ---------------------------------------------------------------- */
  /* -------------------------- Interfaces -------------------------- */
  /* ---------------------------------------------------------------- */

  /**
   * Permet la récupération d'un gestionnaire d'entité pour interagir avec la base
   * @return void
   */
  public function injectObjectManager(ObjectManager $objectManager, ClassMetadata $classMetadata): void {
    if ($objectManager instanceof EntityManager) {
      $this->oEntityManager = $objectManager;
    }
  }

  /**
   * Permet au gestionnaire d'event de doctrine de charger le service des logger lors du chargement
   * @return void
   */
  public function setLoggerManagerService(LoggerManager $oService): void {
    $this->oLoggerManager = $oService;
  }
}
