<?php

namespace App\Module\Cadastre\WebService\Entity\Majic;

use Doctrine\ORM\Mapping\Column;
use Doctrine\ORM\Mapping\Entity;
use Doctrine\ORM\Mapping\EntityListeners;
use Doctrine\ORM\Mapping\Id;
use Doctrine\ORM\Mapping\Table;

/**
 *  Objet ORM pour cadastre_views.emprise
 *  Cette entité permet de sélectionner les attributs nécessaires afin de traiter la partie Emprise
 *
 * @author Alizée Buckel <alizee.buckel@veremes.com>
 *
 * @Entity(repositoryClass="App\Module\Cadastre\WebService\Repository\EmpriseRepository")
 * @EntityListeners({"App\Module\Vitis\WebService\EventListener\VitisEntityListener"})
 * @Table(schema="@prop(cadastre_schema_majic)",name="@prop(cadastre_views:emprise)")
 */
class Emprise implements MajicInterface {
  /**
   * @var string
   * @Column(type="string", name="id_par")
   * @Id
   */
  private $id_par;

  /**
   * @var float
   * @Column(type="float", name="suf_parcelle")
   */
  private $suf_parcelle;

  /**
   * @var array|null
   */
  private $attribute;

  /**
   * @var array|null
   */
  private $definition;

/**
 * @var string
 * @Column(type="geometry", name="geom")
 */
  private $geom;

  /**
   * @var integer
   * @Column(type="integer", name="srid_parcelle")
   */
  private $srid_parcelle;

  /**
   * permet de cloner une emprise
   *
   * @return void
   */
  public function __clone() {
    $this->setAttribute(null);
    $this->setDefinition(null);
  }

  /**
   * @return string
   */
  public function getIdPar(): string {
    return $this->id_par;
  }

  /**
   * @return string
   */
  public function getGeom(): string {
    return $this->geom;
  }

  /**
   * @return float
   */
  public function getSufParcelle(): float {
    return $this->suf_parcelle;
  }

  /**
   * @param array|null $attribute
   * @return self
   */
  public function setAttribute(?array $attribute): self {
    $this->attribute = $attribute;
    return $this;
  }

  /**
   * @return array|null
   */
  public function getAttribute(): ?array{
    return $this->attribute;
  }

  /**
   * @return array
   */
  private function getDataAttributes(): array{
    return array_filter($this->attribute, [$this, 'filterValueForAttribute'], ARRAY_FILTER_USE_BOTH);
  }

  /**
   * @param array|null $definition
   * @return self
   */
  public function setDefinition(?array $definition): self {
    $this->definition = $definition;
    return $this;
  }

  /**
   * @return array|null
   */
  public function getDefinition(): ?array{
    return $this->definition;
  }

  /**
   * fonction utilisé dans un callback pour array_filter
   * @param mixed $val
   * @param string $key
   * @return bool true pour garder l'élément, false pour le supprimer
   */
  protected function filterValueForAttribute($val, string $key): bool {
    if (!isset($this->definition['columns'])) {
      return false;
    }
    foreach ($this->definition['columns'] as $column) {
      if (!isset($column['name'])) {
        continue;
      }
      if ($column['name'] == $key) {
        return true;
      }
    }
    return false;
  }

  /**
   * @return int|null
   */
  public function getSrid(): ?int {
    return $this->srid_parcelle;
  }

  /**
   * @return string
   * Cette méthode permet d'obtenir la valeur de l'intersection
   * Méthode modifiée
   */
  public function getIntersect(): string {
    return $this->attribute['intersect'];
  }

  /**
   * @return string
   * Cette méthode permet d'obtenir la mesure de l'intersection
   * Méthode modifiée
   */
  private function getIntersectLabel(): string {
    $intersectLabel = null;
    switch (strtoupper($this->definition["type"])) {
      case "LINE":
        $intersectLabel = $this->attribute['intersect'] . ' m';
        break;
      case "POLYGON":
        $intersectLabel = $this->attribute['intersect'] . ' m²';
        break;
    }
    return $intersectLabel;
  }

  /**
   * @return float|null
   * Cette méthode permet d'obtenir le ratio
   * Méthode modifiée
   */
  public function getRatio(): ?float {
    if (strtoupper($this->definition["type"]) == "POLYGON") {
      return (round(($this->attribute['intersect'] / $this->suf_parcelle), 2) * 100);
    }
    return null;
  }

  /**
   * @return string
   * Cette méthode permet d'obtenir la mesure du ratio
   * Méthode modifiée
   */
  private function getRatioLabel(): string {
    return $this->getRatio() . ' %';
  }

  /**
   * @return string
   * Cette méthode permet de générer une chaine de caractère reunissant divers informations et utilisée dans les excels
   * Méthode ajoutée
   */
  private function getDataSummary(): string {
    $sum = 'Pourcentage de la parcelle intersectée: ' . $this->getRatioLabel() . ', ';
    foreach ($this->definition['columns'] as $column) {
      foreach ($this->getDataAttributes() as $key => $val) {
        if ($key == $column['name']) {
          if ($column['name'] == 'intersect') {
            $sum = $sum . $column['label'] . ': ' . $val . 'm²' . ', ';
          } else {
            $sum = $sum . $column['label'] . ': ' . $val . ', ';
          }
        }
      }
    }
    return substr_replace($sum, "", -2);
  }

  /**
   * {@inheritDoc}
   * Cette méthode permet de construire un tableau correspondant à emprise pour parcelle
   */
  public function getMajicData(): array{
    $aData = [
      'emprise_intersect' => $this->attribute['intersect'],
      'emprise_intersect_label' => $this->getIntersectLabel(),
      'summary' => $this->getDataSummary(),
    ];
    $sRatio = $this->getRatio();
    if (!empty($sRatio)) {
      $aData['emprise_ratio'] = $sRatio;
      $aData['emprise_ratio_label'] = $this->getRatioLabel();
    }
    $aData['aData'] = $this->getDataAttributes();
    return $aData;
  }
}
