import { Component, OnDestroy, OnInit, ViewChild } from '@angular/core';
import { VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { DatasourceTree } from '../../models/DatasourceTree.model';
import { AdminDatasourceService } from '../../services/admin-datasource.service';
import { ActivatedRoute } from '@angular/router';
import { DataGridComponent } from 'src/app/vitis/data-grid/data-grid.component';
import { DataGridOptions } from 'src/app/vitis/data-grid/data-grid.model';
import { Subscription } from 'rxjs';
import { WorkspaceListOptions } from 'src/app/vitis/workspacelist/workspacelist.model';

@Component({
  selector: 'app-admin-datasource-data-grid',
  templateUrl: './admin-datasource-data-grid.component.html',
  styleUrls: ['./admin-datasource-data-grid.component.scss']
})
export class AdminDatasourceDataGridComponent implements OnInit, OnDestroy {

  @ViewChild(DataGridComponent, { static: true }) oDataGrid: DataGridComponent;

  private aTree: DatasourceTree;
  private oParams: any;
  private oOptions: any;
  private isInit: boolean = false;

  // Options de la grille de données.
  public oDataGridOptions: DataGridOptions = {
    datatableOptions: {},
    datagridOptions: WorkspaceListOptions
  };

  private subscriptions: Subscription[] = [];

  /**
   * Constructeur
   */
  constructor(
    private route: ActivatedRoute,
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private adminDatasourceService: AdminDatasourceService
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit(): void {

    this.subscriptions.push(this.route.data.subscribe(oOptions => {
      this.oOptions = oOptions;
    }));

    this.subscriptions.push(this.route.params.subscribe(oParams => {
      // Avant la réasignation de this.oParams
      // Si un objet était précédement selectionné on enlève la sélection (pour un objet sur la même branche)
      this.adminDatasourceService.setObjectSelected(false, this.oParams, this.aTree);

      this.oParams = oParams;

      // Pour que initform soit lancé il faut l'arbre ainsi que l'id
      if (this.doInitGrid_()) {
        this.initDataGrid();
        this.isInit = true;
      }
    }));

    this.subscriptions.push(this.adminDatasourceService.getTreeObs().subscribe((aTree) => {
      this.aTree = aTree;

      // Pour que initform soit lancé il faut l'arbre ainsi que l'id
      if (this.doInitGrid_()) {
        this.initDataGrid();
        this.isInit = true;
      }
    }));
  }

  /**
   * Destruction du composant
   */
  ngOnDestroy() {
    // Enlève la sélection quand on change de branche
    this.adminDatasourceService.setObjectSelected(false, this.oParams, this.aTree);
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
  }

  /**
   * Initialise la grille de données.
   * @param {oDataGrid} DataGridComponent
   * @returns void
   */
  private initDataGrid() {
    this.getColumns().then((columns) => {
      // Options de la grille de données.
      const datatableOptions: any = {
        select: false,
        ajax: (data: any, callback: any, settings: any) => {
          const oResponse = {
            recordsTotal: 0,
            recordsFiltered: 0,
            data: []
          };

          this.getValues().then((oResults) => {
            if (oResults.length > 0) {
              oResponse.data = oResults;
              oResponse.recordsTotal = oResults.length;
              oResponse.recordsFiltered = oResults.length;
            }

            callback(oResponse);
          }).catch(() => callback(oResponse));
        }
      };

      this.oDataGridOptions.datagridOptions.columns = columns;

      this.oDataGridOptions.datatableOptions = datatableOptions;

      // Création de la grille.
      setTimeout(() => {
        this.oDataGrid.initTable(this.oDataGridOptions);
      });
    });
  }

  /**
   * Récupère les valeurs (et eventuellement les colonnes) pour la grille
   */
  private getColumns(): Promise<any> {
    return new Promise((resolve, reject) => {
      const oObject = this.adminDatasourceService.getObjectById(this.oParams.treeid);
      if (this.vitisCheck.isDefAndNotNull(oObject)) {
        if (this.vitisCheck.isDefAndNotNull(this.oOptions.get_url_func)) {
          if (this.vitisCheck.isFunction(oObject[this.oOptions.get_url_func])) {
            const oUrlAndParams = oObject[this.oOptions.get_url_func]();

            const params: any = oUrlAndParams.params;
            params.get_columns = true;
            params.limit = 0;
            this.vitisRequest.ajaxRequest({
              method: 'GET',
              url: oUrlAndParams.url,
              params: params,
              success: (resp: any) => {
                if (this.vitisCheck.isDefAndNotNull(resp.data)) {
                  if (resp.data.status == 1 && this.vitisCheck.isDefAndNotNull(resp.data.columns)) {
                    if (this.vitisCheck.isDefAndNotNull(resp.data.data)) {
                      resolve(resp.data.columns);
                    }
                  }
                }
              },
              catch: () => {
                resolve([]);
              }
            });
          } else {
            console.error('La fonction ' + this.oOptions.get_url_func + ' (param get_url_func) n\'existe pas dans l\'objet ', oObject);
          }
        }
      }
    });
  }

  /**
   * Récupère les valeurs (et eventuellement les colonnes) pour la grille
   */
  private getValues(): Promise<any> {
    return new Promise((resolve, reject) => {
      const oObject = this.adminDatasourceService.getObjectById(this.oParams.treeid);
      if (this.vitisCheck.isDefAndNotNull(oObject)) {
        if (this.vitisCheck.isDefAndNotNull(this.oOptions.get_url_func)) {
          if (this.vitisCheck.isFunction(oObject[this.oOptions.get_url_func])) {
            const oUrlAndParams = oObject[this.oOptions.get_url_func]();

            const params: any = oUrlAndParams.params;
            params.get_columns = true;
            params.limit = 100;
            this.vitisRequest.ajaxRequest({
              method: 'GET',
              url: oUrlAndParams.url,
              params: params,
              success: async (resp: any) => {
                if (this.vitisCheck.isDefAndNotNull(resp.data)) {
                  if (resp.data.status == 1) {
                    if (this.vitisCheck.isDefAndNotNull(resp.data.data)) {
                      resolve(resp.data.data);
                    }
                  }
                }
              }
            });
          } else {
            console.error('La fonction ' + this.oOptions.get_url_func + ' (param get_url_func) n\'existe pas dans l\'objet ', oObject);
          }
        }
      }
    });
  }

  /**
   * Si tout les paramètres requis pour initialisé la grille sont chargés.
   * @param oDataGrid
   * @returns boolean
   */
  private doInitGrid_(oDataGrid?: any): boolean {
    return this.vitisCheck.isDefAndNotNull(this.aTree) &&
      this.vitisCheck.isDefAndNotNull(this.oParams) &&
      this.vitisCheck.isDefAndNotNull(this.oParams.treeid) &&
      (this.vitisCheck.isDefAndNotNull(this.oDataGrid) || this.vitisCheck.isDefAndNotNull(oDataGrid)) &&
      !this.isInit;
  }
}
