import { VitisCheck, VitisRequest, vitisAlert } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';
import { DatasourceTreeviewFile } from './DatasourceTreeviewFile';
import { Injector } from '@angular/core';
import { AdminDatasourceComponent } from '../../admin-datasource.component';
import { TranslateService } from '@ngx-translate/core';
import { AdminDatasourceService } from '../../services/admin-datasource.service';

export class DatasourceTreeviewFilesConnection {
  uId: number;
  icon: string = 'icon-vmap-bxs-folder';
  subDir: boolean;
  name: string;
  description: string;
  directoryPath: string;
  datasourceId: number;
  pg_public: boolean;
  collapsed: boolean = false;
  aChildrenTypes = [DatasourceTreeviewFile, DatasourceTreeviewFilesConnection];
  aChildrenObjects = [];
  aRightClickMenu = [];
  route: string = DatasourceTreeviewFilesConnection.getRoute();

  translate: TranslateService;
  adminDatasourceService: AdminDatasourceService;

  extentions = ['ecw', 'jp2', 'tif', 'shp'];

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private injector: Injector,
    datas: any,
  ) {
    this.translate = this.injector.get(TranslateService);
    this.adminDatasourceService = this.injector.get(AdminDatasourceService);
    if (this.vitisCheck.isDefAndNotNull(datas)) {
      // Si ce n'est pas la connexion par défaut (privé) ou un sous dossier
      if (((!this.vitisCheck.isDefAndNotNull(datas.pg_public) || datas.pg_public === true)) &&
        (!this.vitisCheck.isDefAndNotNull(datas.subDir) || datas.subDir === false)) {
        // on ajoute le clique droit pour la modification
        this.aRightClickMenu.push({
          icon: 'icon-vmap-wms',
          label: 'ADMIN_DATASOURCE_BTN_UPDATE_FILES_CONNECTION',
          event: 'updateFilesConnection',
        });

        // on ajoute le clique droit pour la suppression
        this.aRightClickMenu.push({
          icon: 'fa-trash',
          label: 'ADMIN_DATASOURCE_BTN_DELETE_FILES_CONNECTION',
          event: 'deleteFilesConnection',
        });
      }
      this.setData(datas);
    }
  }

  /**
   * Retourne la route
   * @returns
   */
  static getRoute() {
    return 'filesconnection/info';
  }

  /**
   * Rafraichit les données de l'objet
   * @param datas
   */
  public refreshData() {
    const sUrl = `${environment.properties.app_api_url}/vmap/datasources/${this.datasourceId}`;
    this.vitisRequest.ajaxRequestPromise({
      url: sUrl,
      params: {
        attributs: 'datasource_id|name|description|directory_path|pg_public'
      },
      method: 'GET'
    }).then((resp: any) => {
      if (this.vitisCheck.isObject(resp.data)) {
        const oResult = resp.data;
        if (oResult.status == 1 && oResult.data) {
          this.setData(oResult.data);
        }
      }
    });

  }

  /**
   * Met les valeurs du webservice dans l'objet
   * @param datas
   */
  private setData(datas: any) {
    this.name = datas.name;
    this.description = datas.description;
    this.directoryPath = datas.directory_path;
    this.datasourceId = datas.datasource_id;
    this.subDir = datas.subDir;
    this.pg_public = datas.pg_public;
    if (this.vitisCheck.isDefAndNotNull(datas.icon)) {
      this.icon = datas.icon;
    }
  }

  /**
   * Charge les objets enfants
   * @returns {Promise}
   */
  public loadChildrens(): Promise<any> {
    return new Promise((resolve, reject) => {
      this.aChildrenObjects = [];
      const sUrl = `${environment.properties.app_api_url}/vmap/datasourcetreeview/filestree`;
      this.vitisRequest.ajaxRequestPromise({
        url: sUrl,
        params: {
          path: this.directoryPath,
          depth: 1
        },
        method: 'GET'
      }).then((resp: any) => {
        if (this.vitisCheck.isObject(resp.data)) {
          const oResult = resp.data;
          if (oResult.status == 1 && oResult.data) {
            if (oResult.data.content && oResult.data.content.length > 0) {
              for (let i = 0; i < oResult.data.content.length; i++) {
                const file = oResult.data.content[i];
                const datas: any = {};
                // Dossier
                if (this.vitisCheck.isDefAndNotNull(file.content)) {
                  datas.name = file.filename;
                  datas.directory_path = this.directoryPath + '/' + file.filename;
                  datas.description = '';
                  datas.datasource_id = this.datasourceId;
                  datas.subDir = true;
                  datas.pg_public = this.pg_public;
                  const child = new DatasourceTreeviewFilesConnection(this.vitisCheck, this.vitisRequest, this.injector, datas);
                  this.aChildrenObjects.push(child);
                } else if (this.extentions.includes(file.extention.toLowerCase())) { // fichier
                  file.directory_path = this.directoryPath;
                  const child = new DatasourceTreeviewFile(this.vitisCheck, this.vitisRequest, this.injector, this, file);
                  this.aChildrenObjects.push(child);
                }
              }
            }

            resolve(this.aChildrenObjects);
          } else {
            reject(oResult);
          }
        } else {
          reject(resp);
        }
      });
    });
  }

  /**
   * Returne la liste des objets enfants
   * @returns {Array<any>}
   */
  public getChildrens(): Array<any> {
    return this.aChildrenObjects;
  }

  /**
   * Récupère les valeurs pour le formulaire
   */
  public getFormValues(): DatasourceFilesConnectionForm {
    const oFormValues: DatasourceFilesConnectionForm = {
      name: this.name,
      description: this.description,
      directory_path: this.directoryPath,
      datasource_id: this.datasourceId,
      pg_public: this.pg_public
    };

    return oFormValues;
  }

  /**
   * Lance le formulaire de mise à jour de l'objet
   * @param {AdminDatasourceComponent} scope
   */
  public updateFilesConnection(scope: AdminDatasourceComponent) {
    scope.openRoute({
      route: DatasourceTreeviewFilesConnection.getRoute(),
      formName: 'update',
      uId: this.uId
    });
  }

  /**
   * Lance le formulaire de mise à jour de l'objet
   * @param {AdminDatasourceComponent} scope
   */
  public deleteFilesConnection(scope: AdminDatasourceComponent) {
    // Vérifie qu'il n'y ai pas de couches relier à ce datasourceId
    this.vitisRequest.ajaxRequestPromise({
      url: `${environment.properties.app_api_url}/vmap/layers`,
      method: 'GET',
      params: {
        attributs: 'layer_id',
        filter: { 'relation': 'AND', 'operators': [{ 'column': 'datasource_id', 'compare_operator': '=', 'value': this.datasourceId }] },
      },
    }).then((resp: any) => {
      if (resp && resp.data) {
        if (resp.data.total_row_number > 0) {
          this.translate.get('ADMIN_DATASOURCE_BTN_DELETE_CONNECTION_WARNING').subscribe((translation) => {
            vitisAlert.alert({
              type: 'error',
              text: translation,
              showCancelButton: true
            });
          });
        } else {
          this.adminDatasourceService.deleteDatasource(this.datasourceId, this.uId);
        }
      }
    });
  }

  /**
   * Initialise les valeurs du formulaire insert enfant
   */
  public getChildrenInitialFormValues(): any {
    return new Promise((resolve, reject) => {
      const oFormValues: any = {
        datasource_id: this.datasourceId
      };
      resolve(oFormValues);
    });
  }
}

export class DatasourceFilesConnectionForm {
  name: string;
  description: string;
  directory_path: string;
  datasource_id: number;
  pg_public: boolean;
}
