import { VitisCheck, VitisRequest, vitisAlert } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';
import { AdminDatasourceComponent } from '../../admin-datasource.component';
import { DatasourceTreeviewPgDatabase } from './DatasourceTreeviewPgDatabase';
import { Injector } from '@angular/core';
import { AdminDatasourceService } from '../../services/admin-datasource.service';
import { TranslateService } from '@ngx-translate/core';

export class DatasourceTreeviewPgConnection {

  public uId: number;
  public icon: string = 'icon-vmap-link-1';
  public name: string;
  public description: string;
  public database: string;
  public host: string;
  public port: string;
  public user: string;
  public password: string;
  public pgPublic: boolean;
  public datasourceId: number;
  public collapsed: boolean = false;
  public aChildrenTypes = [DatasourceTreeviewPgDatabase];
  public aChildrenObjects = [];
  public   aRightClickMenu = [{
    icon: 'fa fa-refresh',
    label: 'ADMIN_DATASOURCE_BTN_REFRESH',
    event: 'refreshPgConnection',
  }];
  public route: string = DatasourceTreeviewPgConnection.getRoute();
  public adminDatasourceService: AdminDatasourceService;
  public translate: TranslateService;

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private injector: Injector,
    datas: any,
  ) {
    if (this.vitisCheck.isDefAndNotNull(datas)) {

      // Si ce n'est pas la connexion par défaut (privé) on ajoute le clique droit pour la modification et suppression
      if (!this.vitisCheck.isDefAndNotNull(datas.pg_public) || datas.pg_public === true) {
        this.aRightClickMenu.push({
          icon: 'icon-vmap-link-1',
          label: 'ADMIN_DATASOURCE_BTN_UPDATE_PG_CONNECTION',
          event: 'updatePgConnection',
        });
        this.aRightClickMenu.push({
          icon: 'fa-trash',
          label: 'ADMIN_DATASOURCE_BTN_DELETE_PG_CONNECTION',
          event: 'deletePgConnection',
        });
      }

      if (this.vitisCheck.isDefAndNotNull(datas.icon)) {
        this.icon = datas.icon;
      }
    }
    this.adminDatasourceService = this.injector.get(AdminDatasourceService);
    this.translate = this.injector.get(TranslateService);
    this.setData(datas);
  }

  /**
   * Retourne la route en cours
   * @returns
   */
  static getRoute() {
    return 'pgconnection';
  }

  /**
   * Rafraichit les données de l'objet
   * @param datas
   */
  public refreshData() {
    const sUrl = `${environment.properties.app_api_url}/vmap/datasources/${this.datasourceId}`;
    this.vitisRequest.ajaxRequestPromise({
      url: sUrl,
      params: {
        attributs: 'datasource_id|name|description|pg_server|pg_port|pg_login|pg_password|pg_database|pg_public'
      },
      method: 'GET'
    }).then((resp: any) => {
      if (this.vitisCheck.isObject(resp.data)) {
        const oResult = resp.data;
        if (oResult.status == 1 && oResult.data) {
          this.setData(oResult.data);
        }
      }
    });

  }

  /**
   * Met les valeurs du webservice dans l'objet
   * @param datas
   */
  private setData(datas: any) {
    if (this.vitisCheck.isDefAndNotNull(datas)) {
      this.name = datas.name;
      this.description = datas.description;
      this.host = datas.pg_server;
      this.port = datas.pg_port;
      this.user = datas.pg_login;
      this.password = datas.pg_password;
      this.database = datas.pg_database;
      this.datasourceId = datas.datasource_id;
      this.pgPublic = datas.pg_public;
    }
  }

  /**
   * Charge les objets enfants
   * @returns {Promise}
   */
  public loadChildrens(): Promise<any> {
    return new Promise((resolve, reject) => {
      this.aChildrenObjects = [];
      const sUrl = `${environment.properties.app_api_url}/vmap/datasourcetreeview/databases`;
      const params: any = {
        filter: {
          'relation': 'AND',
          'operators': [
            { 'column': 'datname', 'compare_operator': 'NOT IN', 'value': ['template0', 'template1'] }
          ]
        },
        order_by: 'database',
        datasource_id: this.datasourceId
      };
      if (this.vitisCheck.isDefAndNotNull(this.database)) {
        params.database = this.database;
      }

      this.vitisRequest.ajaxRequestPromise({
        url: sUrl,
        params: params,
        method: 'GET'
      }).then((resp: any) => {
        if (this.vitisCheck.isObject(resp.data)) {
          const oResult = resp.data;
          if (oResult.status == 1 && oResult.data) {
            if (oResult.data && oResult.data.length > 0) {
              // Ajoute les connexions issue de la base
              for (let i = 0; i < oResult.data.length; i++) {
                const database = oResult.data[i].database;

                const child = new DatasourceTreeviewPgDatabase(this.vitisCheck, this.vitisRequest, this.injector, database);
                this.aChildrenObjects.push(child);
              }
            }

            resolve(this.aChildrenObjects);
          } else {
            reject(oResult);
          }
        } else {
          reject(resp);
        }
      });
    });
  }

  /**
   * Returne la liste des objets enfants
   * @returns {Array<any>}
   */
  public getChildrens(): Array<any> {
    return this.aChildrenObjects;
  }

  /**
   * Récupère les valeurs pour le formulaire
   */
  public getFormValues(): DatasourcePgConnectionForm {
    const oFormValues: DatasourcePgConnectionForm = {
      name: this.name,
      description: this.description,
      pg_server: this.host,
      pg_port: this.port,
      pg_database: this.database,
      pg_login: this.user,
      pg_password: this.password,
      datasource_id: this.datasourceId
    };

    return oFormValues;
  }

  /**
   * Lance le formulaire de mise à jour de l'objet
   * @param {AdminDatasourceComponent} scope
   */
  public updatePgConnection(scope: AdminDatasourceComponent) {
    scope.openRoute({
      route: DatasourceTreeviewPgConnection.getRoute(),
      formName: 'update',
      uId: this.uId
    });
  }

  /**
   * Exécute la requête de suppression d'une source de donnée postgres/postgis
   * @param scope
   */
  public deletePgConnection(scope: AdminDatasourceComponent) {
    this.vitisRequest.ajaxRequestPromise({
      url: `${environment.properties.app_api_url}/vmap/layers`,
      method: 'GET',
      params: {
        filter: { 'relation': 'AND', 'operators': [{ 'column': 'datasource_id', 'compare_operator': '=', 'value': this.datasourceId }] },
      },
    }).then(respLayers => {
      if (respLayers.data && respLayers.data.total_row_number) {
        this.translate.get('ADMIN_DATASOURCE_BTN_DELETE_CONNECTION_WARNING').subscribe((translation) => {
          vitisAlert.alert({
            type: 'error',
            text: translation,
            showCancelButton: true
          });
        });
      } else {
        this.adminDatasourceService.deleteDatasource(this.datasourceId, this.uId);
      }
    }).catch(() => {
      this.translate.get('ADMIN_DATASOURCE_BTN_DELETE_PG_CONNECTION_ERROR').subscribe((translation) => {
        vitisAlert.alert({
          type: 'error',
          text: translation,
          showCancelButton: true
        });
      });
    });
  }

  /**
   * Rafraichit la liste des bases de données.
   * @param scope
   */
  public refreshPgConnection(scope: AdminDatasourceComponent) {
    // Rechargement de la connexion Postgis.
    this.aChildrenObjects.length = 0;
    scope.getChildrens(this);
    // Page d'accueil de la connexion Postgis.
    scope.openRoute({
      route: this.route,
      formName: 'display',
      uId: this.uId
    });
  }
}

export class DatasourcePgConnectionForm {
  name: string;
  description: string;
  pg_server: string;
  pg_port: string;
  pg_database: string;
  pg_login: string;
  pg_password: string;
  datasource_id: number;
}
