import { VitisCheck, VitisRequest, vitisAlert } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';
import { AdminDatasourceComponent } from '../../admin-datasource.component';
import { Injector } from '@angular/core';
import { AdminDatasourceService } from '../../services/admin-datasource.service';
import { DatasourceTreeviewVectortilesLayer } from './DatasourceTreeviewVectortilesLayer';
import { TranslateService } from '@ngx-translate/core';

export class DatasourceTreeviewVectortilesConnection {

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private injector: Injector,
    datas: any,
  ) {
    this.adminDatasourceService = this.injector.get(AdminDatasourceService);
    this.translate = this.injector.get(TranslateService);
    this.setData(datas);
  }

  uId: number;
  icon: string = 'icon-vmap-wms';
  // Services
  adminDatasourceService: AdminDatasourceService;
  translate: TranslateService;

  datasourceId: number;
  name: string;
  description: string;
  flux_login: string;
  flux_password: string;
  headers: string;
  collapsed: boolean = false;

  aChildrenTypes = [DatasourceTreeviewVectortilesLayer];
  aChildrenObjects = [];
  aRightClickMenu = [{
    icon: 'icon-vmap-wms',
    label: 'ADMIN_DATASOURCE_BTN_UPDATE_VECTORTILES_CONNECTION',
    event: 'updateVectortilesConnection',
  }, {
    icon: 'fa-trash',
    label: 'ADMIN_DATASOURCE_BTN_DELETE_VECTORTILES_CONNECTION',
    event: 'deleteVectortilesConnection',
  }, {
    icon: 'icon-vmap-layers-half',
    label: 'ADMIN_DATASOURCE_BTN_ADD_VECTORTILES_LAYER',
    event: 'insertVectortilesLayer',
  }];
  route: string = DatasourceTreeviewVectortilesConnection.getRoute();

  /**
   * Retourne la route
   * @returns
   */
  static getRoute() {
    return 'vectortilesconnection/info';
  }

  /**
   * Rafraichit les données de l'objet
   * @param datas
   */
  public refreshData() {
    const sUrl = `${environment.properties.app_api_url}/vmap/datasources/${this.datasourceId}`;
    this.vitisRequest.ajaxRequestPromise({
      url: sUrl,
      params: {
        attributs: 'datasource_id|name|description|flux_url|flux_type|flux_version|flux_login|flux_password|headers'
      },
      method: 'GET'
    }).then((resp: any) => {
      if (this.vitisCheck.isObject(resp.data)) {
        const oResult = resp.data;
        if (oResult.status == 1 && oResult.data) {
          this.setData(oResult.data);
        }
      }
    });

  }

  /**
   * Met les valeurs du webservice dans l'objet
   * @param datas
   */
  private setData(datas: any) {
    this.datasourceId = datas.datasource_id;
    this.name = datas.name;
    this.description = datas.description;
    this.headers = datas.headers;
    this.flux_login = datas.flux_login;
    this.flux_password = datas.flux_password;
  }

  /**
   * Charge les objets enfants
   * @returns {Promise}
   */
  public loadChildrens(): Promise<any> {
    return new Promise((resolve, reject) => {
      this.aChildrenObjects = [];
      const sUrl = `${environment.properties.app_api_url}/vmap/datasourcetreeview/vectortileslayers`;
      this.vitisRequest.ajaxRequestPromise({
        url: sUrl,
        params: {
          filter: { 'relation': 'AND', 'operators': [{ 'column': 'datasource_id', 'compare_operator': '=', 'value': this.datasourceId }] },
          order_by: 'name'
        },
        method: 'GET'
      }).then((resp: any) => {
        if (this.vitisCheck.isObject(resp.data)) {
          const oResult = resp.data;
          if (oResult.status == 1 && oResult.data) {
            if (oResult.data && oResult.data.length > 0) {
              for (let i = 0; i < oResult.data.length; i++) {
                const service = oResult.data[i];
                this.aChildrenObjects.push(new DatasourceTreeviewVectortilesLayer(this.vitisCheck, this.vitisRequest, this.injector, service));
              }
            }

            resolve(this.aChildrenObjects);
          } else {
            reject(oResult);
          }
        } else {
          reject(resp);
        }
      });
    });
  }

  /**
   * Returne la liste des objets enfants
   * @returns {Array<any>}
   */
  public getChildrens(): Array<any> {
    return this.aChildrenObjects;
  }

  /**
   * Récupère les valeurs pour le formulaire
   */
  public getFormValues(): DatasourceVectortilesConnectionForm {
    const oFormValues: DatasourceVectortilesConnectionForm = {
      name: this.name,
      description: this.description,
      datasource_id: this.datasourceId,
      headers: this.headers,
      flux_login: this.flux_login,
      flux_password: this.flux_password,
    };

    return oFormValues;
  }

  /**
   * Lance le formulaire de mise à jour de l'objet
   * @param {AdminDatasourceComponent} scope
   */
  public updateVectortilesConnection(scope: AdminDatasourceComponent) {
    scope.openRoute({
      route: DatasourceTreeviewVectortilesConnection.getRoute(),
      formName: 'update',
      uId: this.uId
    });
  }

  /**
   * Lance le formulaire de mise à jour de l'objet
   * @param {AdminDatasourceComponent} scope
   */
  public deleteVectortilesConnection(scope: AdminDatasourceComponent) {
    this.loadChildrens().then(resp => {
      if (resp.length) {
        this.translate.get('ADMIN_DATASOURCE_BTN_DELETE_CONNECTION_WARNING').subscribe((translation) => {
          vitisAlert.alert({
            type: 'error',
            text: translation,
            showCancelButton: true
          });
        });
      } else {
        this.adminDatasourceService.deleteDatasource(this.datasourceId, this.uId);
      }
    });
  }

  /**
   * Lance le formulaire d'ajout d'une couche Vectortiles
   * @param {AdminDatasourceComponent} scope
   */
  public insertVectortilesLayer(scope: AdminDatasourceComponent) {
    scope.openRoute({
      route: DatasourceTreeviewVectortilesLayer.getRoute(),
      formName: 'insert',
      uId: this.uId
    });
  }

  /**
   * Initialise les valeurs du formulaire insert enfant
   */
  public getChildrenInitialFormValues(): any {
    return new Promise((resolve, reject) => {
      const oFormValues: any = {
        datasource_id: this.datasourceId
      };
      resolve(oFormValues);
    });
  }
}

export class DatasourceVectortilesConnectionForm {
  name: string;
  description: string;
  datasource_id: number;
  headers: string;
  flux_login: string;
  flux_password: string;
}
