import { VitisCheck, VitisRequest, vitisAlert } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';
import { AdminDatasourceComponent } from '../../admin-datasource.component';
import { DatasourceTreeviewWMSLayer } from './DatasourceTreeviewWMSLayer';
import { AdminDatasourceService } from '../../services/admin-datasource.service';
import { WmsService } from '../../services/wms.service';
import { Injector } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';

export class DatasourceTreeviewWMSConnection {

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private injector: Injector,
    datas: any,
  ) {
    this.wmsService = this.injector.get(WmsService);
    this.adminDatasourceService = this.injector.get(AdminDatasourceService);
    this.translate = this.injector.get(TranslateService);
    this.setData(datas);
  }

  uId: number;
  icon: string = 'icon-vmap-wms';
  // Services
  wmsService: WmsService;
  adminDatasourceService: AdminDatasourceService;
  translate: TranslateService;

  datasourceId: number;
  name: string;
  description: string;
  flux_url: string;
  headers: string;
  flux_version: string;
  flux_options: string;
  flux_login: string;
  flux_password: string;
  collapsed: boolean = false;

  oCapabilities: any;
  aChildrenTypes = [DatasourceTreeviewWMSLayer];
  aChildrenObjects = [];
  aRightClickMenu = [{
    icon: 'icon-vmap-wms',
    label: 'ADMIN_DATASOURCE_BTN_UPDATE_WMS_CONNECTION',
    event: 'updateWMSConnection',
  }, {
    icon: 'fa-trash',
    label: 'ADMIN_DATASOURCE_BTN_DELETE_WMS_CONNECTION',
    event: 'deleteWMSConnection',
  }, {
    icon: 'icon-vmap-layers-half',
    label: 'ADMIN_DATASOURCE_BTN_ADD_WMS_LAYER',
    event: 'insertWMSLayer',
  }];
  route: string = DatasourceTreeviewWMSConnection.getRoute();

  /**
   * Retourne la route
   * @returns
   */
  static getRoute() {
    return 'wmsconnection/info';
  }

  /**
   * Rafraichit les données de l'objet
   * @param datas
   */
  public refreshData() {
    const sUrl = `${environment.properties.app_api_url}/vmap/datasources/${this.datasourceId}`;
    this.vitisRequest.ajaxRequestPromise({
      url: sUrl,
      params: {
        attributs: 'datasource_id|name|description|flux_url|headers|flux_type|flux_version|flux_login|flux_password'
      },
      method: 'GET'
    }).then((resp: any) => {
      if (this.vitisCheck.isObject(resp.data)) {
        const oResult = resp.data;
        if (oResult.status == 1 && oResult.data) {
          this.setData(oResult.data);
        }
      }
    });

  }

  /**
   * Met les valeurs du webservice dans l'objet
   * @param datas
   */
  private setData(datas: any) {
    this.datasourceId = datas.datasource_id;
    this.name = datas.name;
    this.description = datas.description;
    this.flux_url = datas.flux_url;
    this.headers = datas.headers;
    this.flux_version = datas.flux_version;
    this.flux_options = datas.flux_options;
    this.flux_login = datas.flux_login;
    this.flux_password = datas.flux_password;
  }

  /**
   * Charge les objets enfants
   * @returns {Promise}
   */
  public loadChildrens(): Promise<any> {
    return new Promise((resolve, reject) => {
      this.aChildrenObjects = [];
      const sUrl = `${environment.properties.app_api_url}/vmap/datasourcetreeview/wmslayers`;
      this.vitisRequest.ajaxRequestPromise({
        url: sUrl,
        params: {
          filter: { 'relation': 'AND', 'operators': [{ 'column': 'datasource_id', 'compare_operator': '=', 'value': this.datasourceId }] },
          order_by: 'name'
        },
        method: 'GET'
      }).then((resp: any) => {
        if (this.vitisCheck.isObject(resp.data)) {
          const oResult = resp.data;
          if (oResult.status == 1 && oResult.data) {
            if (oResult.data && oResult.data.length > 0) {
              for (let i = 0; i < oResult.data.length; i++) {
                const service = oResult.data[i];
                this.aChildrenObjects.push(new DatasourceTreeviewWMSLayer(this.vitisCheck, this.vitisRequest, this.injector, service));
              }
            }

            resolve(this.aChildrenObjects);
          } else {
            reject(oResult);
          }
        } else {
          reject(resp);
        }
      });
    });
  }

  /**
   * Returne la liste des objets enfants
   * @returns {Array<any>}
   */
  public getChildrens(): Array<any> {
    return this.aChildrenObjects;
  }

  /**
   * Récupère les valeurs pour le formulaire
   */
  public getFormValues(): DatasourceWMSConnectionForm {
    const oFormValues: DatasourceWMSConnectionForm = {
      name: this.name,
      description: this.description,
      flux_url: this.flux_url,
      headers: this.headers,
      flux_version: this.flux_version,
      flux_options: this.flux_options,
      flux_login: this.flux_login,
      flux_password: this.flux_password,
      datasource_id: this.datasourceId
    };

    return oFormValues;
  }

  /**
   * Lance le formulaire de mise à jour de l'objet
   * @param {AdminDatasourceComponent} scope
   */
  public updateWMSConnection(scope: AdminDatasourceComponent) {
    scope.openRoute({
      route: DatasourceTreeviewWMSConnection.getRoute(),
      formName: 'update',
      uId: this.uId
    });
  }

  /**
   * Lance le formulaire de mise à jour de l'objet
   * @param {AdminDatasourceComponent} scope
   */
  public deleteWMSConnection(scope: AdminDatasourceComponent) {
    this.loadChildrens().then(resp => {
      if (resp.length) {
        this.translate.get('ADMIN_DATASOURCE_BTN_DELETE_CONNECTION_WARNING').subscribe((translation) => {
          vitisAlert.alert({
            type: 'error',
            text: translation,
            showCancelButton: true
          });
        });
      } else {
        this.adminDatasourceService.deleteDatasource(this.datasourceId, this.uId);
      }
    });
  }

  /**
   * Lance le formulaire d'ajout d'une couche
   * @param {AdminDatasourceComponent} scope
   */
  public insertWMSLayer(scope: AdminDatasourceComponent) {
    scope.openRoute({
      route: DatasourceTreeviewWMSLayer.getRoute(),
      formName: 'insert',
      uId: this.uId
    });
  }

  /**
   * Utilisé comme getFormValues pour l'onglet avec la carte, pour attendre que le oCapabilities soit chargé avant d'ouvrir le formulaire
   */
  public beforeInitForm(): Promise<any> {
    return new Promise((resolve, reject) => {
      if (this.vitisCheck.isDefAndNotNull(this.getChildrens()) && this.getChildrens().length > 0) {
        this.getCapabilitiesWMS().then((oXmlCapabilities: any) => {
          resolve(oXmlCapabilities);
        });
      } else {
        // Si besoin charge les enfants pour pouvoir vérifier si les couches sont déjà chargé
        this.loadChildrens().then((aChildrens) => {
          this.adminDatasourceService.setChildrensMetadata(aChildrens, this);
          // Récupère le oCapabilities
          this.getCapabilitiesWMS().then((oXmlCapabilities: any) => {
            resolve(oXmlCapabilities);
          });
        });
      }
    });
  }

  /**
   * Utilisé comme getFormValues pour l'onglet avec la carte, pour attendre que le oCapabilities soit chargé avant d'ouvrir le formulaire
   */
  public getCapabilitiesWMS(): Promise<any> {
    return new Promise((resolve, reject) => {
      // Récupère le oCapabilities
      if (this.vitisCheck.isDefAndNotNull(this.oCapabilities)) {
        resolve({
          get_capabilities: this.oCapabilities.xml,
          json_capabilities: this.oCapabilities.json
        });
      } else {
        this.wmsService.getAjaxCapabilitiesWMS(this).then((oCapabilities) => {
          if (oCapabilities !== false) {
            this.oCapabilities = oCapabilities;
            resolve({
              get_capabilities: this.oCapabilities.xml,
              json_capabilities: this.oCapabilities.json
            });
          } else {
            resolve(false);
          }
        });
      }
    });
  }

  /**
   * Initialise les valeurs du formulaire insert enfant
   */
  public getChildrenInitialFormValues(): Promise<any> {
    return new Promise((resolve, reject) => {
      this.getCapabilitiesWMS().then((oCapabilities) => {
        if (oCapabilities !== false) {
          const oFormValues = {
            layer_options: oCapabilities.json_capabilities,
            datasource_id: this.datasourceId,
            datasource: this.getFormValues()
          };
          resolve(oFormValues);
        } else {
          resolve(false);
        }
      });
    });
  }
}

export class DatasourceWMSConnectionForm {
  name: string;
  description: string;
  flux_url: string;
  headers: string;
  flux_version: string;
  flux_options: string;
  flux_login: string;
  flux_password: string;
  datasource_id: number;
}
