import { VitisCheck, VitisRequest, vitisAlert } from 'src/app/vitis/core/services/utils/vitis-utils';
import { Injectable } from '@angular/core';
import { BehaviorSubject, Subject } from 'rxjs';
import { DatasourceTree } from '../models/DatasourceTree.model';
import { environment } from 'src/environments/environment';
import { TranslateService } from '@ngx-translate/core';
import { Router } from '@angular/router';

class CombinaisonObject {
  label: string;
  value: Array<any>;
}

@Injectable({
  providedIn: 'root'
})
export class AdminDatasourceService {

  public obsTree = new BehaviorSubject<DatasourceTree>(null);
  public obsTreeObjectUpdated = new BehaviorSubject<boolean>(null);
  private uId: number = 0;
  public showRightComponent: boolean = true;
  public childrenParams: any;

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private translate: TranslateService,
    private router: Router
  ) { }

  /**
   * obsTree Getter
   */
  public getTreeObs(): BehaviorSubject<DatasourceTree> {
    return this.obsTree;
  }

  /**
   * obsTree Setter
   */
  public setTree(aTree: DatasourceTree) {
    this.obsTree.next(aTree);
  }

  /**
   * obsTreeObjectUpdated Getter
   */
  public getTreeObjectUpdatedObs(): BehaviorSubject<boolean> {
    return this.obsTreeObjectUpdated;
  }

  /**
   * obsTreeObjectUpdated Setter
   */
  public setTreeObjectUpdated(isUpdated: boolean) {
    this.obsTreeObjectUpdated.next(isUpdated);
  }

  /**
   * Set uId et oParent à un objet
   * @param oObject
   * @param oParent
   */
  public setObjectMetadata(oObject: any, oParent?: any): any {

    // Set oParent
    if (oParent && !oObject.oParent) {
      oObject.oParent = oParent;
    }

    // Set uId
    if (!oObject.uId) {
      oObject.uId = this.uId;
      this.uId++;
    }

    if (!this.vitisCheck.isDefAndNotNull(oObject.selected)) {
      oObject.selected = false;
    }

    return oObject;
  }

  /**
   * retourne un des objets de l'arbre par son identifiant
   * @param uId
   */
  public getObjectById(uId: number): any {
    const aTree: any = this.getTreeObs().getValue();
    if (this.vitisCheck.isDefAndNotNull(aTree) && this.vitisCheck.isDefAndNotNull(uId)) {
      return this.recursiveFindObjectById_(aTree, uId);
    } else {
      return null;
    }
  }

  /**
   * Fonction récursive permettant de rechercher un objet dans l'arbre par son uId
   * @param aTree
   * @param uId
   */
  private recursiveFindObjectById_(aTree: Array<any>, uId: number): any {
    let returnedObject = null;
    for (let i = 0; i < aTree.length; i++) {
      const oObject = aTree[i];
      if (oObject.uId == uId) { // Attention double égal obligatoire
        return oObject;
      }
      if (this.vitisCheck.isDefAndNotNull(oObject.aChildrenObjects) && oObject.aChildrenObjects.length > 0) {
        returnedObject = this.recursiveFindObjectById_(oObject.aChildrenObjects, uId);
      }

      if (this.vitisCheck.isDefAndNotNull(returnedObject)) {
        return returnedObject;
      }
    }
  }

  /**
   * Si l'objet existe change son état, sélectionné ou non
   * @param isSelected
   * @param oParams
   * @param aTree
   */
  public setObjectSelected(isSelected: boolean, oParams: any, aTree: DatasourceTree) {
    if (this.vitisCheck.isDefAndNotNull(oParams) && this.vitisCheck.isDefAndNotNull(aTree)) {
      const oObject = this.getObjectById(oParams.treeid);
      if (this.vitisCheck.isDefAndNotNull(oObject)) {
        oObject.selected = isSelected;
        this.setTreeObjectUpdated(true);
      }
    }
  }

  /**
   * Vérifie si la condition est rempli
   * @param sCondition
   * @param oObject
   */
  public isFulfilledCondition(sCondition: string, oObject: any) {
    if (this.isCustomTernaryString_(sCondition, true, oObject)) {
      return eval(this.translateCustomTernaryString_(sCondition, true, oObject));
    }
    return false;
  }

  /**
   * isCustomTernaryString function.
   * La chaine de caractère passée est une expression ternaire définie par l'administrateur
   * ex : "= {{puissance}} == 500"
   * @param {string} sString Chaine de caractère.
   * @param {boolean} bIsCondition True si il s'agis d'une condition true/false
   * @param  {any} oObject Objet contenant les valeurs
   * @return {boolean}
   */
  private isCustomTernaryString_(sString: string, bIsCondition: boolean, oObject: any): boolean {
    if (this.vitisCheck.isDefAndNotNull(this.translateCustomTernaryString_(sString, bIsCondition, oObject))) {
      return true;
    } else {
      return false;
    }
  }

  /**
   * translateCustomTernaryString function
   * Traduit une chaine ternaire customisée par une chaine ternaire exploitable
   * ex : "= {{puissance}} == 500" devient "oObject.value[puissance] == 500 ? true : false"
   * @param  {string} sString
   * @param  {boolean} bIsCondition
   * @param  {any} oObject Objet contenant les valeurs
   * @return {string}
   */
  private translateCustomTernaryString_(sString: string, bIsCondition: boolean, oObject: any): string {

    // Verif type
    if (!this.vitisCheck.isString(sString)) {
      return null;
    }

    // Vérification chaine
    if (sString.substring(0, 1) !== '=') {
      return null;
    }

    // Enlève le premier signe "="
    sString = sString.substring(1);

    // Parse le template
    sString = this.parseCustomTernaryString_(
      sString,
      // S'il y a des symboles d'interpolation
      sString.indexOf('{{') >= 0 && sString.indexOf('}}') >= 0,
      oObject
    );

    // Ajoute le retour true/false
    if (bIsCondition) {
      sString = '(' + sString + ') ? true : false';
    }

    return sString;
  }

  /**
   * parseCustomTernaryString_ function
   * Parse une chaine ternaire customisée par une chaine ternaire exploitable
   * ex : "= oProperties.enableEncryptionForUsers == true" devient "true == true"
   * @param  {string} sString
   * @param  {boolean} bMustache
   * @param  {any} oObject Objet contenant les valeurs
   * @return {string}
   */
  private parseCustomTernaryString_(sString: string, bMustache: boolean, oObject: any): string {
    // S'il y a des crochets on remplace par oObject
    if (bMustache) {
      // Suppression d'espaces entre crochets
      sString = sString.replace(/{{ */g, '{{');
      sString = sString.replace(/ *}}/g, '}}');
      const aKey = sString.match(/{{(\w+(\.\w+)*)}}/g);
      // Parcours tous les attributs trouvé
      aKey.forEach((sKey: string) => {
        // Remplace toutes les occurances trouvées
        sString = sString.replace(RegExp(sKey, 'g'), 'oObject.' + sKey.replace(/[{}]+/g, ''));
      });
    }
    // Trouve tout les mots(variables) dans le template
    const aFinded = sString.match(/(\w+(\.\w+)*)/g);
    aFinded.forEach((sFinded: string) => {
      let replacementValue: any = null;
      // Decoupe le mot trouvé
      const aSplited = sFinded.split('.');
      // Remplace la première partie par ça valeur
      switch (aSplited[0]) {
        case 'oObject':
          replacementValue = oObject;
          break;
        case 'oProperties':
          replacementValue = environment.properties;
          break;
      }
      // si autre chose est bind on bypass
      if (replacementValue !== null && aSplited.length > 1) {
        // Remplace les attributes par leur valeur
        for (let idx = 1; idx < aSplited.length; idx++) {
          if (this.vitisCheck.isDefAndNotNull(replacementValue[aSplited[idx]])) {
            replacementValue = replacementValue[aSplited[idx]];
          } else {
            replacementValue = '';
            break;
          }
        }
        let sValue = (replacementValue === null) ? '""' : replacementValue;

        if (typeof (replacementValue) === 'string') {
          sValue = '\'' + replacementValue + '\'';
        }
        // Remplace la variable trouvé par sa valeur
        sString = sString.replace(sFinded, sValue);
      }
    });
    return sString;
  }

  /**
   * Set uId et oParent à chacun des enfants
   * @param aChildrens
   * @param oParent
   */
  public setChildrensMetadata(aChildrens: Array<any>, oParent?: any): Array<any> {
    for (let i = 0; i < aChildrens.length; i++) {
      aChildrens[i] = this.setObjectMetadata(aChildrens[i], oParent);
    }

    return aChildrens;
  }

  /**
   * Rafraichit le composant à droite de l'arbre si un composant est ouvert et que la demande de rafraichissement se fait sur l'objet ouvert
   * @param iTreeObjectId Identifiant de l'objet de l'arbre
   */
  public refreshRightComponent(iTreeObjectId: number) {
    if (this.vitisCheck.isDefAndNotNull(this.childrenParams) && this.vitisCheck.isDefAndNotNull(this.childrenParams.treeid)) {
      if (this.childrenParams.treeid == iTreeObjectId) {
        this.showRightComponent = false;
        setTimeout(() => {
          this.showRightComponent = true;
        }, 100);
      }
    }
  }

  /**
   * Function qui supprime une source de donnée dans l'arbre
   * @param datasourceId
   * @param uId Id de la source de donnée dans l'arbre
   * @returns
   */
  public deleteDatasource(datasourceId: number, uId: number) {
    const oObject = this.getObjectById(uId);
    return new Promise<void>((resolve, reject) => {
      this.vitisRequest.ajaxRequestPromise({
        url: `${environment.properties.app_api_url}/vmap/datasources/${datasourceId}`,
        method: 'DELETE'
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status === 1) {
            this.translate.get('ADMIN_DATASOURCE_BTN_DELETE_CONNECTION_SUCCESS').subscribe((sTranslate: string) => {
              vitisAlert.fireToast({
                type: 'success',
                title: sTranslate
              });
            });
            oObject.oParent.loadChildrens().then((aChildrens: Array<any>) => {
              this.setChildrensMetadata(aChildrens, oObject.oParent);
              this.setTree(this.getTreeObs().getValue());
              const previousParent = this.previousRouteParent(oObject.oParent);
              this.router.navigate([`/dashboard/datasource/${previousParent.route}/display/${previousParent.uId}`]);
            });
            resolve();
          } else {
            reject();
          }
        }
      });
    });
  }

  /**
   * Function qui supprime une couche dans l'arbre
   * @param layer_id
   * @param uId Id de la couche dans l'arbre
   * @returns
   */
  public deleteLayer(layer_id: number, uId: number) {
    return new Promise<void>((resolve, reject) => {
      this.vitisRequest.ajaxRequestPromise({
        url: `${environment.properties.app_api_url}/vmap/maps/layers/getAll`,
        method: 'GET',
        params: {
          filter: { 'relation': 'AND', 'operators': [{ 'column': 'layer_id', 'compare_operator': '=', 'value': layer_id }] },
        },
      }).then((resp: any) => {
        if (resp && resp.data && resp.data.data) {
          const data = resp.data.data;
          if (data.length) {
            this.translate.get('ADMIN_DATASOURCE_BTN_DELETE_LAYER_WARNING').subscribe((translation) => {
              vitisAlert.alert({
                type: 'error',
                text: translation,
                showCancelButton: true
              });
            });
          } else {
            const oObject = this.getObjectById(uId);
            this.vitisRequest.ajaxRequestPromise({
              url: `${environment.properties.app_api_url}/vmap/layers/${layer_id}`,
              method: 'DELETE'
            }).then((response) => {
              if (response.data) {
                if (response.data.status === 1) {
                  this.translate.get('ADMIN_DATASOURCE_BTN_DELETE_LAYER_SUCCESS').subscribe((sTranslate: string) => {
                    vitisAlert.fireToast({
                      type: 'success',
                      title: sTranslate
                    });
                  });
                  oObject.oParent.loadChildrens().then((aChildrens: Array<any>) => {
                    this.setChildrensMetadata(aChildrens, oObject.oParent);
                    this.setTree(this.getTreeObs().getValue());
                    const previousParent = this.previousRouteParent(oObject.oParent);
                    this.router.navigate([`/dashboard/datasource/${previousParent.route}/display/${previousParent.uId}`]);
                  });
                  resolve();
                } else {
                  reject();
                }
              }
            });
          }
        }
      });
    });
  }

  /**
   * Supprime une publication de l'arbre
   * @param publication_id
   * @param uId
   */
  public deletePublication(publication_id: number, uId: number) {
    return new Promise<void>((resolve, reject) => {
      const oObject = this.getObjectById(uId);
      this.vitisRequest.ajaxRequestPromise({
        url: `${environment.properties.app_api_url}/vmap/publications/${publication_id}`,
        method: 'DELETE'
      }).then((response) => {
        if (response.data) {
          if (response.data.status === 1) {
            this.translate.get('ADMIN_DATASOURCE_BTN_DELETE_LAYER_SUCCESS').subscribe((sTranslate: string) => {
              vitisAlert.fireToast({
                type: 'success',
                title: sTranslate
              });
            });
            oObject.oParent.loadChildrens().then((aChildrens: Array<any>) => {
              this.setChildrensMetadata(aChildrens, oObject.oParent);
              this.setTree(this.getTreeObs().getValue());
              const previousParent = this.previousRouteParent(oObject.oParent);
              this.router.navigate([`/dashboard/datasource/${previousParent.route}/display/${previousParent.uId}`]);
            });
            resolve();
          } else {
            reject();
          }
        }
      });
    });
  }

  /**
 * Retourne le dernier parent accessible par son attribut route
 * @param oParent
 * @returns
 */
  private previousRouteParent(oParent) {
    if (this.vitisCheck.isDefAndNotNull(oParent.route)) {
      return oParent;
    } else if (oParent.oParent) {
      return this.previousRouteParent(oParent.oParent);
    }
    return { route: 'postgis', uId: 0 };
  }

  /**
   * Retourne toutes les combinaison possible sous forme d'object de 3 arrays
   * Exemple :
   * array1 = [{ label: 'style', value: ["default"]}]
   * array1 = [{ label: 'format', value: ["image/png", "image/jpeg"]}]
   * [{ style: "default", format: "image/png" },{ style: "default", format: "image/jpeg" }]
   * @param array1
   * @param array2
   * @param array3
   * @returns
   */
  public buildCombinationsObjects(array1: CombinaisonObject, array2: CombinaisonObject, array3?: CombinaisonObject) {
    if (array1.value.length === 0) { // Cas ou array1 est vide
      array1 = { ...array2 };
      array2.value = [];
    }
    // Combine N fois l'array2 par l'array1
    const combinArr1Arr2 = array1.value.reduce((acc, a) => {
      const obj = [];
      if (array2.value.length === 0) { // Cas ou array2 est vide
        obj.push({ [array1.label]: a });
      } else {
        array2.value.map(b => {
          obj.push({ [array1.label]: a, [array2.label]: b });
        });
      }
      return [...acc, ...obj];
    }, []);

    // Reproduit N fois l'array3 en la combinaison de l'array1 et l'array2
    if (!array3 || array3.value.length === 0) { return combinArr1Arr2; } // Cas ou array3 est vide
    if (combinArr1Arr2.length === 0) {
      return array3.value.map(b => {
        return { [array3.label]: b };
      });
    }
    return combinArr1Arr2.reduce((acc, curr) => {
      return [...acc, ...array3.value.map(b => {
        return { ...curr, [array3.label]: b };
      })];
    }, []);
  }
}
