import { Component, Inject, OnDestroy, OnInit, Optional, ViewChild } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { Location } from '@angular/common';
import Map from 'ol/Map';
import { Vector as VectorSource } from 'ol/source.js';
import { Observable, Subject, Subscription } from 'rxjs';
import { LocalStorage, SessionStorage, vitisAlert, VitisCheck, VitisClone, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { MapCoreComponent } from 'src/app/vitis/map-core/map-core.component';
import { MapCatalog, MapDefinition, MapJSON } from 'src/app/vitis/map-core/services/map-models.model';
import { MapCoreManagerService } from 'src/app/vitis/map-core/services/map-core-manager/map-core-manager.service';
import { TranslateService } from '@ngx-translate/core';
import { environment } from 'src/environments/environment';
import { DashboardComponent } from 'src/app/vitis/core/components/dashboard/dashboard.component';
import { MapCorePopupOverlayService } from 'src/app/vitis/map-core/services/map-core-popup-overlay/map-core-popup-overlay.service';
import { MapCoreDrawService } from 'src/app/vitis/map-core/services/map-core-draw/map-core-draw.service';
import { PrintMeasureService } from './services/vmap-tools/vmap-tools-print/vmap-tools-print-measure.service';
import { VmapToolsService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools.service';
import { VmapSelectService } from 'src/modules/vmap/components/carto/services/vmap-select/vmap-select.service';
import { transformExtent } from 'ol/proj';
import { MapCorePopupService } from 'src/app/vitis/map-core/services/map-core-popup/map-core-popup.service';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import { Feature } from 'ol';
import { Geometry } from 'ol/geom';
import { VmapToolsGeolocationService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools-geolocation/vmap-tools-geolocation.service';
import { IsSmallScreenService } from 'src/app/vitis/core/services/utils/is-small-screen/is-small-screen.service';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';

export class PrintMapOptions {
  token?: string;
  apiUrl?: string;
  mapJSON?: MapDefinition;
  measureFeatures?: any;
  extent?: string;
  features?: string;
  featuresZoom?: number;
  resolutionCoeff?: number;
  properties?: any;
}

@Component({
  selector: 'app-carto',
  templateUrl: './carto.component.html',
  styleUrls: ['./carto.component.scss']
})
export class CartoComponent implements OnInit, OnDestroy {

  @ViewChild(MapCoreComponent) mapCoreComponent: MapCoreComponent;

  // Mode en cours
  public mode: any;
  public modeOptions: any;

  // Cate en cours
  public olMap: Map;
  public oMapCatalog: MapCatalog;

  // Id de la div OL3
  public targetId: string;
  public targetCompareId: string;
  public tileSize: Array<number>;

  // Permet de communiquer lorsque la carte est chargée
  public initMapObs = new Subject<Map>();

  // Variables de session
  private localStorageValues_: any;
  private sessionStorageValues_: any;

  // Def de la carte
  public mapJSON: MapJSON;
  public haveControls = false;
  public haveTools = false;
  public haveLayersPanel = false;
  public haveMapsPanel = false;
  public haveHelperTool = false;
  public isWidget = false;
  public widgetInitialized = false;
  public isPrintWidget = false;

  // url params
  private layers_tool = false;
  private maps_tool = false;
  private helper_tool = false;

  // Permet de rafraîchir la carte
  private reloadMapTrigger = new Subject<void>();

  public selectLayerMobile: any;

  public isSmallScreen;
  private subscribeSelectLayerMobile;
  private sActiveRightPanel: string;
  private sActiveLeftPanel: string;

  // Apramètres présents dans l'URL
  private oUrlParams: any;

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private route: ActivatedRoute,
    private router: Router,
    private location: Location,
    private localStorage: LocalStorage,
    private sessionStorage: SessionStorage,
    private mapCoreManagerService: MapCoreManagerService,
    private mapCorePopupService: MapCorePopupService,
    private mapCorePopupOverlayService: MapCorePopupOverlayService,
    private mapCoreDrawService: MapCoreDrawService,
    private mapCoreUtils: MapCoreUtilsService,
    private translate: TranslateService,
    private printMeasureService: PrintMeasureService,
    private vmapToolsService: VmapToolsService,
    private mapSelectService: VmapSelectService,
    private isSmallScreenService: IsSmallScreenService,
    private geolocationSrvc: VmapToolsGeolocationService,
    private vitisRequest: VitisRequest,
    private vitiClone: VitisClone,
    private auth: AuthService,
    @Optional() @Inject(DashboardComponent) private dashboardComponent: DashboardComponent,
  ) {
    // @ts-ignore
    window.initPrintMap = (oOptions: any) => {
      return this.initPrintMap_(oOptions);
    };
  }

  /**
   * Retourne true si le tracking GPS est ctif
   */
  get trackingGpsActive() {
    return this.geolocationSrvc.isTrackingEnable();
  }

  /**
   * Retourne le nom de la couche en cours de sélection
   */
  get selectLayerMobileName() {
    if (this.selectLayerMobile && this.selectLayerMobile.name !== 'Premier plan') {
      return this.selectLayerMobile.name;
    }
    return '';
  }

  /**
   * Chargement du composant
   */
  ngOnInit(): void {

    this.isSmallScreen = this.isSmallScreenService.getIsSmallScreen();
    this.getSelectLayerMobile(this.isSmallScreen);

    this.subscriptions.push(this.isSmallScreenService.isSmallScreenObservable.subscribe(isSmallScreen => {
      if (!this.isPrintWidget) {  // Condition pour éviter les erreurs pendant l'impression, cette dernière change la taille de la fenêtre au moment du screenshot
        this.isSmallScreen = isSmallScreen;
        this.getSelectLayerMobile(isSmallScreen);
      }
    }));

    // Traductions (obligatoire sinon ne fonctionne pas dans le module en lazy loading)
    this.translate.addLangs(['en', 'fr']);
    this.translate.setDefaultLang('en');
    this.translate.use(environment.properties.language ? environment.properties.language : 'en');

    setTimeout(() => { // Attend que le composant carte soit chargé

      // Récupère targetId et tileSize du composant MapCore
      this.targetId = this.mapCoreComponent.targetId;
      this.targetCompareId = this.mapCoreComponent.targetCompareId;
      this.tileSize = this.mapCoreComponent.tileSize;

      // Mode en cours d'utilisation
      this.subscriptions.push(this.route.data.subscribe(mode => {
        this.mode = mode;

        // options app.json
        if (mode.options) {
          this.modeOptions = mode.options;
        }

        if (mode.name === 'vmap_widget') {
          this.isWidget = true;
          this.initWidgetInteractions();
        }

        if (mode.name === 'vmap_print_widget') {
          this.isPrintWidget = true;
        }

        // Initialise localStorageValues_ et sessionStorageValues_
        this.initCookies_();
      }));

      // Surveille le paramètre pour le changement de carte
      setTimeout(() => {
        this.subscriptions.push(this.route.queryParams.subscribe(oParams => {
          this.oUrlParams = oParams;

          // Étendue
          const extent = this.getExtentFromParams_(oParams);
          // Map id
          const mapIdParam = this.getMapIdFromParams_(oParams);

          if (this.isWidget === true) {
            if (this.widgetInitialized === false) {
              this.initWidget(oParams, extent);
            } else if (mapIdParam && Number(mapIdParam)) {
              this.initFromFormMapId_(mapIdParam, extent);
            }
          } else {
            // Type d'initialisation
            if (mapIdParam && Number(mapIdParam)) {
              // map_id dans l'url
              this.initFromFormMapId_(mapIdParam, extent);
            } else if (this.isPrintWidget === false) {
              // Utilisation avec le catalogue
              this.initFromCatalog_();
            }
          }
        }));
      });
    });

    if (this.vitisCheck.isDefAndNotNull(this.dashboardComponent)) {
      this.subscriptions.push(this.dashboardComponent.sidenavOpenChange.subscribe((sidenavOpen: any) => {
        if (this.olMap) {
          this.mapCoreManagerService.resizeMap(this.olMap);
        }
        if (this.olMap.get('map_compare')) {
          this.mapCoreManagerService.resizeMap(this.olMap.get('map_compare'));
        }
      }));
    }

    // Permet le rechargement de la carte
    this.subscriptions.push(this.reloadMapTrigger.subscribe(() => {
      this.reloadMap_();
    }));

    this.subscriptions.push(this.vmapToolsService.getActiveRightPanelEmiter().subscribe(tool => {
      this.sActiveRightPanel = tool;
    }));
    this.subscriptions.push(this.vmapToolsService.getActiveLeftPanelEmiter().subscribe(tool => {
      this.sActiveLeftPanel = tool;
    }));
  }

  /**
   * À la destruction
   */
  ngOnDestroy() {

    // Vide les paramètres de l'URL
    this.untrackUrlParams_();

    // Vide les couches vectorielles
    this.clearMapLayers_();

    // Ferme tous les menus
    this.vmapToolsService.closeAllPanels();

    this.subscriptions.forEach(subscription => subscription.unsubscribe());
  }

  /**
   * Initialise la carte en mode widget
   * @param oParams
   * @param extent
   */
  private initWidget(oParams: any, extent: any) {
    if (oParams.widget_id) {
      if (this.widgetInitialized === false) {
        this.vitisRequest.ajaxRequestPromise({
          method: 'GET',
          url: environment.properties.app_api_url + `/vitis/widget/${oParams.widget_id}`,
          params: {
            params_url: oParams,
          }
        }).then((resp: any) => {
          if (resp.data.status === 1) {
            const oWidget = resp.data.data;

            if (this.vitisCheck.isDefAndNotNull(oWidget)) {

              // Gestion du mode mobile en mode widget
              environment.properties.smallscreen_limit = oWidget.mobile_mode === true ? 768 : 1;

              const comp_data = JSON.parse(oWidget.comp_data);

              if (comp_data.layers_tool === true) {
                this.layers_tool = true;
              }

              if (comp_data.maps_tool === true) {
                this.maps_tool = true;
              }

              if (comp_data.helper_tool === true) {
                this.helper_tool = true;
              }

              // Évite de relancer la requête
              this.widgetInitialized = true;

              this.initFromFormMapId_(comp_data.map_id, extent);
            }
          }

        })
          .catch((resp) => {
            if (this.auth.isAuthenticated() === false &&
              resp.data &&
              (resp.data.errorMessage === 'ERROR_SESSION_EXPIRED' ||
                resp.data.errorMessage === 'ERROR_MISSING_AUTHORIZATION_HEADER')) {
              vitisAlert.alert({
                type: 'error',
                html: this.translate.instant('ERROR_SESSION_EXPIRED'),
              }).then(() => {
                const sDefaultLoginEndpoint: string = environment.properties.default_login_endpoint || 'login';
                const afterConnectUrl = this.vitiClone.clone(this.router.url);
                this.router.navigate([sDefaultLoginEndpoint]).then(() => {
                  this.auth.disconnect(afterConnectUrl);
                });
              });
            } else {
              vitisAlert.alert({
                type: 'error',
                title: this.translate.instant('FORM_VALIDATION_ERROR'),
                html: this.translate.instant('COMPONENT_MAP_ERROR_GET_MAP_CATALOG_UNAUTHORIZED'),
              });
            }
          });
      } else {
        // Ancien système de widget
        console.error('old widget format');
        setTimeout(() => {
          vitisAlert.alert({
            type: 'warning',
            title: this.translate.instant('MAP_WIDGET_OLD_FORMAT_TITLE'),
            text: this.translate.instant('MAP_WIDGET_OLD_FORMAT_DESCRIPTION'),
          });
        }, 1000);
      }
    }
  }

  /**
   * Attibut selectLayerMobile si on est en version mobile
   * @param isSmallScreen
   */
  private getSelectLayerMobile(isSmallScreen) {
    if (isSmallScreen) {
      this.subscribeSelectLayerMobile = this.mapSelectService.getSelectLayerObs().subscribe(layer => {
        if (this.vitisCheck.isDefAndNotNull(layer) && this.vitisCheck.isDefAndNotNull(layer.name)) {
          this.selectLayerMobile = layer;
        }
      });
    } else {
      if (this.subscribeSelectLayerMobile) {
        this.subscribeSelectLayerMobile.unsubscribe();
      }
    }
  }

  /**
   * Initialise localStorageValues_ et sessionStorageValues_
   */
  private initCookies_() {
    this.localStorageValues_ = this.localStorage.get(this.mode.name);
    this.sessionStorageValues_ = this.sessionStorage.get(this.mode.name);
    if (!this.vitisCheck.isDefAndNotNull(this.localStorageValues_)) {
      this.localStorageValues_ = {};
    }
    if (!this.vitisCheck.isDefAndNotNull(this.sessionStorageValues_)) {
      this.sessionStorageValues_ = {};
    }


    // Set les paramètres du sessionstorage dans l'URL
    const queryParams: any = {};
    if (this.vitisCheck.isDefAndNotNull(this.sessionStorageValues_.map_id) && !this.vitisCheck.isDefAndNotNull(this.route.snapshot.queryParamMap.get('map_id'))) {
      queryParams.map_id = this.sessionStorageValues_.map_id;
    }
    if (this.vitisCheck.isDefAndNotNull(this.sessionStorageValues_.layer_id) && !this.vitisCheck.isDefAndNotNull(this.route.snapshot.queryParamMap.get('layer_id'))) {
      queryParams.layer_id = this.sessionStorageValues_.layer_id;
    }
    if (this.vitisCheck.isDefAndNotNull(this.sessionStorageValues_.id) && !this.vitisCheck.isDefAndNotNull(this.route.snapshot.queryParamMap.get('id'))) {
      queryParams.id = this.sessionStorageValues_.id;
    }
    if (queryParams && Object.keys(queryParams).length > 0) {
      this.router.navigate(
        [],
        {
          relativeTo: this.route,
          queryParams,
          queryParamsHandling: 'merge',
        });
    }
  }

  /**
   * Initialisation à partir d'un map_id
   */
  private initFromFormMapId_(sMapId: string, extent?: Array<any>) {
    const mapId: number = Number(sMapId);

    // Si la carte n'est pas chargé ou que le map_id est différent de celui de la carte courante
    if (!this.vitisCheck.isDefAndNotNull(this.olMap) ||
      (this.vitisCheck.isDefAndNotNull(this.olMap) &&
        !this.vitisCheck.isUndefined(this.olMap.get('map_id')) &&
        this.olMap.get('map_id') !== mapId)
    ) {

      // Chargement carte
      if (this.oMapCatalog) {

        // Si l'utilisateur a une étendue par défaut
        let extentProj: string;
        if (!this.vitisCheck.isDefAndNotNull(extent) &&
          this.vitisCheck.isString(this.oMapCatalog.defaultExtent) &&
          this.oMapCatalog.defaultExtent.length > 0) {
          extent = this.getExtentByQuery_(this.oMapCatalog.defaultExtent);
          extentProj = 'EPSG:2154';
        }

        // Charge la carte
        this.initMapById_(mapId, extent, extentProj);
      } else {

        // Charge le catalogue puis la carte
        this.subscriptions.push(this.mapCoreManagerService.getMapCatalog().subscribe((oMapCatalog) => {

          // Enregistre le catalogue
          this.oMapCatalog = oMapCatalog;

          // Si l'utilisateur a une étendue par défaut
          let extentProj: string;
          if (!this.vitisCheck.isDefAndNotNull(extent) &&
            this.vitisCheck.isString(this.oMapCatalog.defaultExtent) &&
            this.oMapCatalog.defaultExtent.length > 0) {
            extent = this.getExtentByQuery_(this.oMapCatalog.defaultExtent);
            extentProj = 'EPSG:2154';
          }

          // Charge la carte
          this.initMapById_(mapId, extent, extentProj);
        }));
      }
    }
  }

  /**
   * Utilisation vMap
   */
  private initFromCatalog_() {

    // Charge la carte par défaut dans l'URL
    this.subscriptions.push(this.mapCoreManagerService.getMapCatalog().subscribe((oMapCatalog) => {
      if (oMapCatalog) {

        // Enregistre le catalogue
        this.oMapCatalog = oMapCatalog;

        // Trouve la carte par défaut et l'ajoute dans l'URL
        const defaultMapId = this.mapCoreManagerService.getDefaultMapFromCatalog(oMapCatalog);
        if (defaultMapId) {
          this.router.navigate(
            [],
            {
              relativeTo: this.route,
              queryParams: {
                map_id: defaultMapId
              },
              queryParamsHandling: 'merge',
            });
        } else {
          this.translate.get(['COMPONENT_MAP_ERROR_GET_MAP_CATALOG_NO_MAP', 'COMPONENT_MAP_ERROR_GET_MAP_CATALOG_NO_MAP_DESCRIPTION']).subscribe((translations: string) => {
            vitisAlert.alert({
              type: 'error',
              title: translations['COMPONENT_MAP_ERROR_GET_MAP_CATALOG_NO_MAP'],
              html: translations['COMPONENT_MAP_ERROR_GET_MAP_CATALOG_NO_MAP_DESCRIPTION']
            });
          });
        }
      } else {
        this.translate.get(['COMPONENT_MAP_ERROR_GET_MAP_CATALOG', 'COMPONENT_MAP_ERROR_GET_MAP_CATALOG_DESCRIPTION']).subscribe((translations: string) => {
          vitisAlert.alert({
            type: 'error',
            title: translations['COMPONENT_MAP_ERROR_GET_MAP_CATALOG'],
            html: translations['COMPONENT_MAP_ERROR_GET_MAP_CATALOG_DESCRIPTION']
          });
        });
      }
    }));
  }

  /**
   * Initialise la carte depuis un map_id
   * @param iMapId
   */
  private initMapById_(iMapId: number, extent?: Array<any>, extentProj?: string): Observable<any> {

    const oReturn = new Subject<Map>();

    this.subscriptions.push(this.mapCoreManagerService.initOlMap(iMapId, this.targetId, this.tileSize, this.olMap).subscribe((olMap: any) => {

      if (olMap.error) {
        if (olMap.error === 'unauthorized') {
          // Recharge la première carte du catalogue
          if (this.oMapCatalog && this.oMapCatalog.maps.length) {
            this.router.navigate(
              [],
              {
                relativeTo: this.route,
                queryParams: {
                  map_id: this.oMapCatalog.maps[0].map_id
                },
              });
          }
        }
      } else {
        // Enregistre l'objet Map
        this.olMap = olMap;

        // Enregistre l'identifiant de carte utilisé
        this.olMap.set('map_id', iMapId);

        // Enregistre l'identifiant de l'élement utilisé
        this.olMap.set('target_id', this.targetId);

        // Enregistre le catalogue de cartes
        this.olMap.set('map_catalog', this.oMapCatalog);

        // Fit l'étendue
        if (this.vitisCheck.isDefAndNotNull(extentProj) &&
          this.vitisCheck.isString(extentProj) &&
          extentProj.length > 0) {
          const sCurrentProjCode = this.olMap.getView().getProjection().getCode();
          extent = transformExtent(extent, extentProj, sCurrentProjCode);
        }
        if (extent) {
          this.olMap.getView().fit(extent, {
            nearest: true
          });
        }

        // Bouchon MapJSON
        if (this.olMap && this.olMap.get('map_definition')) {
          const oMapDef = this.olMap.get('map_definition');
          if (oMapDef.maptools) {
            if (!this.mapJSON) {
              this.mapJSON = {
                type: 'vmap',
                controls: {},
                interactions: {},
              };
            }
            this.mapJSON.controls = oMapDef.maptools.controls;
          }
          this.olMap.set('map_json', this.mapJSON);
        }

        // Attribue les autres paramètres (draw_layer...)
        this.setMapAttributes_(olMap);

        // Affiche les composants de la carte
        this.displaySubComponents_();

        // Remplit automatiquement l'URL en fonction de la carte
        this.trackUrlParams_();

        // Permet de rafraîchir la carte
        this.handleReloadMap_();

        this.initMapObs.next(this.olMap);

        oReturn.next(this.olMap);
      }
    }));
    return oReturn;
  }

  /**
   * Initialise une carte pour une impression
   * @param oOptions
   */
  private initPrintMap_(oOptions: PrintMapOptions) {
    if (oOptions.mapJSON) {

      let olMap = this.mapCoreManagerService.initMapFromDef(oOptions.mapJSON, this.targetId, this.tileSize, this.olMap, oOptions.measureFeatures);
      if (oOptions.measureFeatures && oOptions.measureFeatures.length > 0) {
        olMap = this.printMeasureService.initPrintableMeasures(olMap, oOptions.resolutionCoeff);
      } else {
        this.mapCorePopupService.closePopups();
      }

      // Enregistre l'objet Map
      this.olMap = olMap;

      // Enregistre l'identifiant de l'élement utilisé
      this.olMap.set('target_id', this.targetId);

      // Fit l'étendue
      if (this.vitisCheck.isString(oOptions.extent)) {
        const aExtent: any[] = oOptions.extent.split('|');
        for (let i = 0; i < aExtent.length; i++) {
          aExtent[i] = Number(aExtent[i]);
        }
        this.olMap.getView().fit(aExtent, {
          nearest: true
        });
      }

      // Fit features
      if (oOptions.features && oOptions.features.length > 0) {
        let aFeatures: Array<Feature<Geometry>> = [];
        const aEWKTFeatures = oOptions.features.split('|');

        for (const sEWKTFeature of aEWKTFeatures) {
          aFeatures = aFeatures.concat(this.mapCoreUtils.getFeaturesFromGeom(sEWKTFeature, olMap));
        }

        // Ajout de la ou les features à la carte
        const olLocationPopupLayer = this.mapCorePopupOverlayService.initPopupOverlay(this.olMap);
        olLocationPopupLayer.getSource().addFeatures(aFeatures);

        // Récupère les géométries
        const aGeometries: Array<Geometry> = [];

        for (const oFeature of aFeatures) {
          aGeometries.push(oFeature.getGeometry());
        }

        // Zoom sur la ou les géométries
        const featuresZoom = oOptions.featuresZoom ? Number(oOptions.featuresZoom) : 200;
        this.mapCoreUtils.centerMapOnGeoms(olMap, aGeometries, [featuresZoom, featuresZoom, featuresZoom, featuresZoom]);
      }
    }

    return {
      scale: Math.round(this.mapCoreUtils.getScale(this.olMap, { pretty: false }))
    };
  }

  /**
   * Retourne le map_id selon session_storage et l'URL
   * @param oParams
   */
  private getMapIdFromParams_(oParams: any): string {
    let mapId: string;

    if (this.sessionStorageValues_.map_id) {
      mapId = this.sessionStorageValues_.map_id;
    }
    if (oParams.map_id) {
      mapId = oParams.map_id;
    }
    return mapId;
  }

  /**
   * Retourne le extent selon session_storage et l'URL
   * @param oParams
   */
  private getExtentFromParams_(oParams: any): Array<any> {
    let extent: Array<number>;
    if (this.sessionStorageValues_.map_ext) {
      extent = this.sessionStorageValues_.map_ext;
    }
    if (oParams.map_ext) {
      extent = this.getExtentByQuery_(oParams.map_ext);
    }
    if (this.olMap &&
      this.olMap.get('map_id')) {
      if (this.olMap.get('map_id') !== oParams.map_id &&
        environment.properties.keep_map_extent !== true) {
        extent = null;
      }
    }
    return extent;
  }

  /**
   * Retourne une étendue à partir d'une chaine séparée par des |
   * @param sExtent
   */
  private getExtentByQuery_(sExtent: string): Array<number> {

    const aReturn = [];

    if (this.vitisCheck.isString(sExtent)) {
      const aExtent = sExtent.split('|');
      for (let i = 0; i < aExtent.length; i++) {
        aReturn.push(Number(aExtent[i]));
      }
    }

    if (aReturn.length === 4) {
      return aReturn;
    } else {
      return null;
    }
  }

  /**
   * Initialise les attributs Vitis de l'objet Map
   * @param olMap
   */
  public setMapAttributes_(olMap: Map) {

    // Couche de dessin
    if (!this.vitisCheck.isDefAndNotNull(this.olMap.get('draw_layer'))) {
      this.olMap.set('draw_layer', this.mapCoreDrawService.initDrawLayer(this.olMap));
    }

    // Couche de popup
    if (!this.vitisCheck.isDefAndNotNull(this.olMap.get('popup_layer'))) {
      this.olMap.set('popup_layer', this.mapCorePopupOverlayService.initPopupOverlay(this.olMap));
    }

    // Type de template (fromreader / fullscreen)
    this.olMap.set('template_type', 'fullscreen');

    // Nom de la carte
    if (this.vitisCheck.isDefAndNotNull(this.olMap.get('map_catalog')) &&
      this.vitisCheck.isDefAndNotNull(this.olMap.get('map_catalog').maps)) {
      for (let i = 0; i < this.olMap.get('map_catalog').maps.length; i++) {
        if (this.olMap.get('map_catalog').maps[i].map_id == this.olMap.get('map_id')) {
          this.olMap.set('name', this.olMap.get('map_catalog').maps[i]['name']);
        }
      }
    }
  }

  /**
   * Remplit automatiquement l'URL sans déclencher de router event
   */
  private trackUrlParams_() {
    if (this.olMap) {
      const olView = this.olMap.getView();
      if (olView) {

        // Étendue
        this.olMap.on('moveend', () => {
          const extent = this.getQueryByExtent(olView.calculateExtent());
          if (extent) {

            // Mémorisation de l'étendue dans l'URL
            const urlTree = this.router.parseUrl(this.router.url);
            urlTree.queryParams.map_ext = extent;
            this.location.replaceState(urlTree.toString());

            // Mémorisation de l'étendue dans la session
            this.sessionStorage.add(this.mode.name, {
              map_ext: olView.calculateExtent()
            });
          }
        });
      }
    }
  }

  /**
   * Vide l'URL des paramètres
   */
  private untrackUrlParams_() {

    // Mémorise les paramètres dans la session (sauf map_ext qui est enregistré à la volée)
    const oSessionParams = {
      map_id: null,
      layer_id: null,
      id: null,
      ids: null,
    };
    if (this.oUrlParams && this.oUrlParams.map_id) {
      oSessionParams.map_id = this.oUrlParams.map_id;
    }
    if (this.oUrlParams && this.oUrlParams.layer_id) {
      oSessionParams.layer_id = this.oUrlParams.layer_id;
    }
    if (this.oUrlParams && this.oUrlParams.id) {
      oSessionParams.id = this.oUrlParams.id;
    }
    if (this.oUrlParams && this.oUrlParams.ids) {
      oSessionParams.ids = this.oUrlParams.ids;
    }
    this.sessionStorage.add(this.mode.name, oSessionParams);

    // Vide l'URL
    this.router.navigate(
      [],
      {
        relativeTo: this.route,
        queryParams: {
          map_id: null,
          map_ext: null,
          layer_id: null,
          ids: null,
          id: null,
        },
        queryParamsHandling: 'merge',
      }
    );
  }

  /**
   * Retourne une chaine URL à partir d'une étendue
   * @param extent
   */
  private getQueryByExtent(extent: Array<number>): any {

    if (extent) {
      let sQuery = '';
      for (let i = 0; i < extent.length; i++) {
        if (i > 0) {
          sQuery += '|';
        }
        sQuery += extent[i].toPrecision(10);
      }
      return sQuery;
    } else {
      return null;
    }
  }

  /**
   * Affiche les sous-composants
   */
  private displaySubComponents_() {
    // Boutons zoom, projection, échelle etc.
    this.haveControls = true;

    // Couches de la carte
    if (!this.isWidget || this.layers_tool) {
      this.haveLayersPanel = true;

      setTimeout(() => {
        const oEnabledControls = this.olMap.get('enabled_controls');
        if (oEnabledControls &&
          oEnabledControls.layers_tree_tab === false &&
          oEnabledControls.layers_list_tab === false &&
          oEnabledControls.layers_legend_tab === false) {
          this.haveLayersPanel = false;
        }
      });
    }

    // Menu changement de carte
    if (!this.isWidget || this.maps_tool) {
      this.haveMapsPanel = true;
    }

    if (this.isWidget && this.helper_tool) {
      this.haveHelperTool = true;
    }

    // outils de la carte
    this.haveTools = true;
  }

  /**
   * Permet le rechargement de la carte
   */
  private handleReloadMap_() {
    this.olMap.set('reloadMapTrigger', this.reloadMapTrigger);
  }

  /**
   * Recharge la carte
   */
  private reloadMap_() {
    if (this.vitisCheck.isDefAndNotNull(this.olMap) &&
      !this.vitisCheck.isUndefined(this.olMap.get('map_id'))
    ) {
      this.clearMapLayers_();
      setTimeout(() => {
        this.initMapById_(this.olMap.get('map_id'));
      });
    }
  }

  /**
   * Vide les overlays et les couches vectorielles
   */
  private clearMapLayers_() {
    const aLayers = this.olMap.getLayers().getArray();
    for (let i = 0; i < aLayers.length; i++) {
      const olLayer: any = aLayers[i];
      const olSource = olLayer.getSource();

      if (olSource instanceof VectorSource) {
        olSource.clear();
      }
    }

    this.olMap.getOverlays().clear();
  }

  /**
   * Ouvre les modes de la carte version mobile
   * @param panel
   */
  public openMobilePanel(panel: string) {
    if (panel === this.sActiveLeftPanel) {
      this.vmapToolsService.setLeftPanel(0, '');
    } else {
      this.vmapToolsService.setLeftPanel(1, panel);
    }
    this.vmapToolsService.setPanelBtnClicked(panel);
  }

  /**
   * Ouvre les outils de la carte version mobile
   * @param panel
   */
  public openMobilePanelTools(panel: string) {
    if (panel === this.sActiveRightPanel) {
      this.vmapToolsService.setRightPanel(0, '');
    } else {
      this.vmapToolsService.setRightPanel(1, panel);
    }
    this.vmapToolsService.setPanelBtnClicked(panel);
  }

  /**
   * Initialise les interactions widget
   */
  private initWidgetInteractions() {
    window.onmessage = (e: any) => {
      switch (e.data) {
        case 'zoomMap':
          break;
        case 'filterMap':
          break;
        default:
          break;
      }
    };
  }
}
