import { Component, Input, OnDestroy, OnInit, SecurityContext } from '@angular/core';
import { Map } from 'ol';
import { MousePosition, ScaleLine, Attribution } from 'ol/control';
import { format } from 'ol/coordinate';
import { VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import { VmapSelectService } from 'src/modules/vmap/components/carto/services/vmap-select/vmap-select.service';
import { environment } from 'src/environments/environment';
import { DomSanitizer } from '@angular/platform-browser';
import { TranslateService } from '@ngx-translate/core';
import { Subscription } from 'rxjs';
import { MapCoreManagerService } from 'src/app/vitis/map-core/services/map-core-manager/map-core-manager.service';

@Component({
  selector: 'app-vmap-controls-map-info',
  templateUrl: './vmap-controls-map-info.component.html',
  styleUrls: ['./vmap-controls-map-info.component.scss']
})
export class VmapControlsMapInfoComponent implements OnInit, OnDestroy {

  @Input() olMap: Map;
  @Input() isSmallScreen: boolean;

  public layerName = '';
  public mapProjection: any;
  public mapScale: number;
  public mousePosition: MousePosition;
  public scaleLine: ScaleLine;
  public map_projection: boolean = true;
  public numeric_scale: boolean = true;
  public graphic_scale: boolean = true;
  public mouse_coordinates: boolean = true;

  private oProjections: any;

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private mapSelectService: VmapSelectService,
    private mapUtils: MapCoreUtilsService,
    private translate: TranslateService,
    private sanitizer: DomSanitizer,
    private mapCoreManagerService: MapCoreManagerService,
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit(): void {
    if (this.olMap.get('enabled_controls')) {
      this.map_projection = this.olMap.get('enabled_controls').map_projection;
      this.numeric_scale = this.olMap.get('enabled_controls').numeric_scale;
      this.graphic_scale = this.olMap.get('enabled_controls').graphic_scale;
      this.mouse_coordinates = this.olMap.get('enabled_controls').mouse_coordinates;
    }
    this.loadProjections_().then(() => {

      this.initOlControls_();
      this.reloadMapProjection_();
      this.refreshScale_();

      this.olMap.on('change', () => {
        setTimeout(() => {
          if (this.olMap.get('enabled_controls')) {
            this.map_projection = this.olMap.get('enabled_controls').map_projection;
            this.numeric_scale = this.olMap.get('enabled_controls').numeric_scale;
            this.graphic_scale = this.olMap.get('enabled_controls').graphic_scale;
            this.mouse_coordinates = this.olMap.get('enabled_controls').mouse_coordinates;
          }

          this.reloadMapProjection_();
          this.refreshOlControls_();
          this.refreshAttributions_();
        });
      });

      // @ts-ignore
      this.olMap.on('vectortiles_loaded', () => {
        this.refreshAttributions_();
      });

      this.olMap.on('moveend', () => {
        this.refreshScale_();
      });
    });

    this.subscriptions.push(this.mapSelectService.getSelectLayerObs().subscribe(oLayer => {
      this.layerName = oLayer ? oLayer.name : '';
    }));
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
  }

  /**
   * Récupère les données de projection provenant d'un JSON et les pose dans this.oProjections
   * @returns Promise
   */
  private loadProjections_(): Promise<any> {
    return new Promise((resolve, reject) => {

      const url = 'assets/data/proj.json';

      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data)) {
          this.oProjections = resp.data;
          resolve(resp.data);
        } else {
          console.error('Error getting JSON form');
          reject(false);
        }
      });
    });
  }


  /**
   * Initialisation des controles OL3
   */
  private initOlControls_() {
    const mapJSON = this.olMap.get('map_json');

    // Mouse Position
    const sMouseProjection = environment.properties.mouse_coordinate_projection;
    mapJSON.coord_accuracy = sMouseProjection === 'EPSG:4326' ? 5 : 0;
    this.mousePosition = new MousePosition({
      coordinateFormat: (coordinate: any) => {
        const template = '<div>X : {x}</div><div>Y : {y}</div>';
        return format(coordinate, template, mapJSON.coord_accuracy);
      },
      placeholder: false,
      target: document.getElementById('map-coordinates')
    });
    this.olMap.set('mouse_position', this.mousePosition);

    // Projection des coordonnées de la souris (properties).
    if (this.vitisCheck.isDefAndNotNull(sMouseProjection) && sMouseProjection != '') {
      this.mousePosition.setProjection(sMouseProjection);
    }

    // Scale line
    this.scaleLine = new ScaleLine({
      target: document.getElementById('map-scale-line')
    });

    this.olMap.addControl(this.scaleLine);
    this.olMap.set('scale_line', this.scaleLine);


    // Controles
    this.refreshOlControls_();

    // Attributions
    this.refreshAttributions_();
  }

  /**
   * Recharge le contenu de projection
   */
  private reloadMapProjection_() {
    const sCurrentProjCode = this.olMap.getView().getProjection().getCode();
    const oCurrentProjection = this.oProjections.projections.find((proj: { code: any; }) => proj.code == sCurrentProjCode);

    if (oCurrentProjection !== -1) {
      this.mapProjection = oCurrentProjection.name;
    } else {
      this.mapProjection = sCurrentProjCode;
    }
  }

  /**
   * Refraichit le texte de l'échelle
   */
  private refreshScale_() {
    this.mapScale = Math.round(this.mapUtils.getScale(this.olMap, { pretty: false }));
  }

  /**
   * Crée ou rafraîchir les l'outil sources des couches de la carte
   */
  private refreshAttributions_() {
    const aLayers = this.mapCoreManagerService.getLayersList(this.olMap);
    if (this.vitisCheck.isArray(aLayers)) {
      aLayers.forEach(vmapLayer => {
        if (vmapLayer.olLayer) {
          const olLayer = vmapLayer.olLayer;
          const attributionSource = olLayer.values.attribution;
          if (this.vitisCheck.isDefAndNotNull(attributionSource) &&
            this.vitisCheck.isString(attributionSource.name) &&
            attributionSource.name.length > 0) {
            if (this.vitisCheck.isString(attributionSource.url) && attributionSource.url.length > 0) {
              olLayer.getSource().setAttributions(` <a href="${this.sanitizer.sanitize(SecurityContext.HTML, attributionSource.url)}" target="_blank">${this.sanitizer.sanitize(SecurityContext.HTML, attributionSource.name)}</a>.`);
            } else {
              olLayer.getSource().setAttributions(` ${this.sanitizer.sanitize(SecurityContext.HTML, attributionSource.name)}.`);
            }
          }
        }
      });
    }

    const attributionOptions = {
      collapsible: true,
      collapsed: false,
      tipLabel: this.translate.instant('SOURCE_MAP')
    };


    if (this.vitisCheck.isBoolean(environment.properties.layer_sources_collapsed) &&
      environment.properties.layer_sources_collapsed === true) {
      attributionOptions.collapsed = environment.properties.layer_sources_collapsed;
    } else if (this.isSmallScreen === true) {
      attributionOptions.collapsed = true;
    } else {
      attributionOptions.collapsed = false;
    }

    if (this.olMap.get('attribution')) {
      this.olMap.removeControl(this.olMap.get('attribution'));
    }

    const attribution = new Attribution(attributionOptions);
    attribution.setTarget(document.getElementById('map-attributions'));
    this.olMap.addControl(attribution);
    this.olMap.set('attribution', attribution);
  }

  /**
   * Active / desactive les controles OL
   */
  private refreshOlControls_() {

    // Coord souris
    if (this.mouse_coordinates === true) {
      this.olMap.addControl(this.mousePosition);
    } else {
      this.olMap.removeControl(this.mousePosition);
    }

    // Échelle graphique
    if (this.graphic_scale === true) {
      this.olMap.addControl(this.scaleLine);
    } else {
      this.olMap.removeControl(this.scaleLine);
    }
  }

  /**
   * Définit s'il y a au moins un outil à afficher
   * @returns
   */
  public mapInfoDisplay(): boolean {
    return this.map_projection || this.numeric_scale || this.graphic_scale || this.mouse_coordinates;
  }
}
