import { Component, Input, OnDestroy, OnInit } from '@angular/core';
import { Map } from 'ol';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { MapJSON } from 'src/app/vitis/map-core/services/map-models.model';
import { VmapToolsService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools.service';
import { availableControls, availableControlsMobile } from 'src/app/vitis/map-core/services/map-models.model';
import { Subscription } from 'rxjs';
import { environment } from 'src/environments/environment';
import { IsSmallScreenService } from 'src/app/vitis/core/services/utils/is-small-screen/is-small-screen.service';

@Component({
  selector: 'app-vmap-controls',
  templateUrl: './vmap-controls.component.html',
  styleUrls: ['./vmap-controls.component.scss']
})
export class VmapControlsComponent implements OnInit, OnDestroy {

  private currentControlsAvailable;
  public enabledControls: any = {};

  public mapJSON: MapJSON;
  public bMapsPanelOpen: boolean = false;
  public bLayersPanelOpen: boolean = false;

  public iLeftPanelOpen: number = 0; // 0, 1, 2, 3
  public iRightPanelOpen: number = 0; // 0, 1, 2, 3
  public iBottomPanelOpen: number = 0; // 0, 1, 2, 3

  @Input() olMap: Map;
  @Input() isWidget: boolean;
  @Input() haveLayersPanel: boolean;
  @Input() haveMapsPanel: boolean;

  public olMapCompare: Map;

  public isSmallScreen;

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];

  /**
   * Constructeur
   */
  constructor(
    private vmapToolsService: VmapToolsService,
    private vitisCheck: VitisCheck,
    private isSmallScreenService: IsSmallScreenService
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit(): void {

    this.isSmallScreen = this.isSmallScreenService.getIsSmallScreen();
    this.subscriptions.push(this.isSmallScreenService.isSmallScreenObservable.subscribe(isSmallScreen => {
      this.isSmallScreen = isSmallScreen;
    }));

    this.onInit();

    this.olMap.on('change', (event) => {
      // Ferme les panneaux
      this.vmapToolsService.closeAllPanels();

      setTimeout(() => {
        // reset les valeurs de la carte
        this.onInit();
        this.olMapCompare = event.target.get('map_compare') || null;
      });
    });

    // Observe l'état des bandeaux
    this.subscriptions.push(this.vmapToolsService.getLeftPanelEmiter().subscribe((next: number) => {
      if (this.isSmallScreen !== true) {
        this.iLeftPanelOpen = next;
      }
    }));
    this.subscriptions.push(this.vmapToolsService.getRightPanelEmiter().subscribe((next: number) => {
      if (this.isSmallScreen !== true) {
        this.iRightPanelOpen = next;
      }
    }));
    this.subscriptions.push(this.vmapToolsService.getBottomPanelEmiter().subscribe((next: number) => {
      this.iBottomPanelOpen = next;
    }));
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
  }

  /**
   * Appelé à l'initialisation
   */
  private onInit() {
    this.currentControlsAvailable = this.isSmallScreen ? availableControlsMobile : availableControls;
    this.enabledControls = this.currentControlsAvailable.reduce((acc, curr) => {
      return { ...acc, [curr]: false };
    }, {});
    this.mapJSON = this.olMap.get('map_json');

    // retrocompatibilité
    this.enabledControls.layers_tree_tab = null;
    this.enabledControls.layers_list_tab = null;
    this.enabledControls.layers_legend_tab = null;

    if (this.mapJSON) {
      this.initMapControls_(this.mapJSON);
      this.olMap.set('enabled_controls', this.enabledControls);
    }
  }

  /**
   * Initialise les controles de la carte
   * @param  {any} oControls
   */
  private initMapControls_(mapJSON: any) {
    const oControls = mapJSON.controls;

    if (this.vitisCheck.isDefAndNotNull(oControls)) {
      this.currentControlsAvailable.forEach(control => {
        if (oControls[control]) {
          this.enabledControls[control] = true;
        } else if (oControls[control] === false) {
          this.enabledControls[control] = false;
        }
      });
    }
  }
}
