import { Component, Input, OnDestroy, OnInit } from '@angular/core';
import { Map } from 'ol';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { VmapToolsService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools.service';
import { MapCoreManagerService } from 'src/app/vitis/map-core/services/map-core-manager/map-core-manager.service';
import { Subscription } from 'rxjs';

@Component({
  selector: 'app-vmap-layers-list',
  templateUrl: './vmap-layers-list.component.html',
  styleUrls: ['./vmap-layers-list.component.scss']
})
export class VmapLayersListComponent implements OnInit, OnDestroy {

  @Input() public olMap: Map;

  public aLayersList: any;
  public sDragElementID: string;
  public placeholder: HTMLDivElement;
  public sVisibleParametersID: string = '';

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private mapCoreManagerService: MapCoreManagerService,
    private vmapToolsService: VmapToolsService,
  ) { }

  /**
   * Retourne la liste des couches qui ne sont pas is_map_hidden
   */
  get layersListNotMapHidden() {
    if (this.aLayersList && this.aLayersList.length) {
      return this.aLayersList.filter(layer => !layer.is_map_hidden);
    } else {
      return [];
    }
  }

  /**
   * Chargement du composant
   */
  ngOnInit(): void {

    // initialise l'arbre
    this.initLayersList_();
    this.olMap.on('change', () => {
      this.initLayersList_();
      // Rechargement au bout de x secondes pour régler les problèmes de chargement asynchrone pour les vectortiles
      setTimeout(() => {
        this.initLayersList_();
      }, 3000);
    });

    this.subscriptions.push(this.vmapToolsService.getLeftPanelEmiter().subscribe((iNbCols: number) => {
      if (iNbCols === 0) {
        this.sVisibleParametersID = '';
      }
    }));
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
  }

  /**
   * Fonction déclenchée lors d'un début de drag and drop sur une couche
   * @param event
   */
  public onDragStart(event) {
    // Tout dans le setTimeout (sinon bug après scroll sur chrome).
    setTimeout(() => {
      // On récupère la div element-item en cours de drag
      const draggingElement = event.target;

      // On lui applique quelques styles pour qu'il ne soit plus visible pendant le drag
      draggingElement.style.position = 'absolute';
      draggingElement.style.zIndex = -10;

      // On récupère tout le groupe de l'élément
      const element = draggingElement.parentNode;
      this.sDragElementID = element.id;

      // On crée la div de placeholder
      this.placeholder = document.createElement('div');
      this.placeholder.classList.add('placeholder');

      // Insertion du placeholder
      element.parentNode.insertBefore(
        this.placeholder,
        element.nextSibling
      );
    });
  }

  /**
   * Fonction déclenchée lors du survol d'un élément en cours de drag and drop
   * @param event
   */
  public onDragOver(event) {
    let focusedElement: any;
    const sortableNeighbors = this.getSortableNeighbors(this.sDragElementID);
    const element = event.target.closest('.element');

    // Drag au dessus d'un element
    if (this.vitisCheck.isDefAndNotNull(element) && sortableNeighbors.some(e => e.id === element.id)) {
      focusedElement = element;
    }

    const { bottom, top } = focusedElement.getBoundingClientRect();
    const elHeigth = bottom - top;
    const offsetY = bottom - event.clientY;

    if (offsetY > elHeigth / 2) {
      focusedElement.parentNode.insertBefore(
        this.placeholder,
        focusedElement
      );
    } else {
      focusedElement.parentNode.insertBefore(
        this.placeholder,
        focusedElement.nextSibling
      );
    }
  }

  /**
   * Fonction déclenchée lors du drop d'un élément
   * @param event
   */
  public onDragEnd(event) {
    const draggingElement = event.target;
    draggingElement.style.removeProperty('position');
    draggingElement.style.removeProperty('z-index');

    // On met l'élément que l'on déplacait à la place du placeholder
    this.placeholder.parentNode.replaceChild(draggingElement.parentNode, this.placeholder);

    // Récupération de l'ancien ordre des couches
    const aOldLayersOrder = this.mapCoreManagerService.getLayersList(this.olMap).map(e => e.layer_id);

    // Ordre actuel des couches
    const aCurrentLayersOrder = this.getCurrentLayersOrder_().reverse();

    // Si l'ordre des couches est différent, réorganiser l'ordre sur la carte
    if (JSON.stringify(aOldLayersOrder) !== JSON.stringify(aCurrentLayersOrder)) {
      this.mapCoreManagerService.setLayersOrder(aCurrentLayersOrder, this.olMap);
    }
  }

  /**
   * Retourne les layer_id des couches selon leur ordre dans le HTML
   * @returns
   */
  private getCurrentLayersOrder_() {
    const aCurrentLayersOrder = [];

    const listElement = document.getElementById('layer-list');
    const aElements = Array.from(listElement.querySelectorAll('.element')).map(e => e.id);

    for (const element of aElements) {
      for (const children of this.aLayersList) {
        if (this.vitisCheck.isDefAndNotNull(children.layer_id)) {
          if (children.id === element) {
            aCurrentLayersOrder.push(children.layer_id);
          }
        }

        if (this.vitisCheck.isDefAndNotNull(children.children)) {
          const result = children.children.find(e => e.id === element);

          if (result) {
            aCurrentLayersOrder.push(result.layer_id);
          }
        }
      }
    }

    return aCurrentLayersOrder;
  }

  /**
   * Renvoie la liste des voisins avec lesquels un élément peut se réorganiser
   * @param nodeID
   * @returns
   */
  public getSortableNeighbors(nodeID) {
    const element = document.getElementById(nodeID);
    const sortableNeighbors = [];

    const neighbors = element.parentNode.querySelectorAll(':scope > .element');
    const groupNeighbors = element.parentNode.querySelectorAll(':scope > .layer-group');

    for (let i = 0; i < neighbors.length; i++) {
      sortableNeighbors.push({
        id: neighbors[i].id,
        type: 'element'
      });
    }

    for (let j = 0; j < groupNeighbors.length; j++) {
      sortableNeighbors.push({
        id: groupNeighbors[j].id,
        type: 'group'
      });
    }

    return sortableNeighbors;
  }

  /**
   * Retourne la visibilité de la couche dans l'arbre
   * @param layerElement couche dans l'arbre
   * @returns
   */
  public getLayerVisibility(layerElement: any): boolean {
    if (layerElement.layer_id) {
      return layerElement.olLayer.get('visible');
    }

    if (layerElement.children) {
      let visible = false;

      for (const children of layerElement.children) {
        const childrenVisibility = this.getLayerVisibility(children);

        if (childrenVisibility) {
          visible = true;
        }
      }

      return visible;
    }
  }

  /**
   * Set la visibilité d'une couche
   * @param layerElement
   * @param visibility
   */
  private setLayerVisibility(layerElement: any, visibility: boolean) {
    if (layerElement.layer_id && layerElement.is_map_hidden !== true) {
      this.mapCoreManagerService.setLayerVisible(layerElement, this.olMap, visibility);
    }

    if (layerElement.children) {
      for (const children of layerElement.children) {
        this.setLayerVisibility(children, visibility);
      }
    }
  }

  /**
   * Bascule l'état de visibilité d'une couche
   * @param id
   */
  public toggleLayerVisibility(element: any) {
    const currentVisibility = this.getLayerVisibility(element);
    this.setLayerVisibility(element, !currentVisibility);
  }

  /**
   * Bascule l'état de visibilité de la boîte de paramètres d'une couche
   * @param id
   */
  public toggleVisibleParameters(id) {
    if (id === this.sVisibleParametersID) {
      this.sVisibleParametersID = '';
    } else {
      this.sVisibleParametersID = id;
    }
  }

  /**
   * Initialise l'arbre
   */
  private initLayersList_() {
    const aLayersList = this.mapCoreManagerService.getLayersList(this.olMap);
    if (this.vitisCheck.isArray(aLayersList)) {
      this.aLayersList = aLayersList.reverse();
    }
  }
}
