import { Component, OnInit, Input, Output, EventEmitter, OnDestroy, OnChanges, SimpleChanges, AfterViewInit } from '@angular/core';
import { Feature } from 'ol';
import { Geometry, Point, Polygon, LineString, Circle } from 'ol/geom';
import VectorLayer from 'ol/layer/Vector';
import VectorSource from 'ol/source/Vector';
import { MapCoreDrawService } from 'src/app/vitis/map-core/services/map-core-draw/map-core-draw.service';
import { LayerToolsService } from '../../../services/layer-tools/layer-tools.service';
import { getLength } from 'ol/sphere';
import { VmapToolsMeasureService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools-measure/vmap-tools-measure.service';
import { Fill, Stroke, Style } from 'ol/style';
import CircleStyle from 'ol/style/Circle';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { VmapToolsGeolocationService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools-geolocation/vmap-tools-geolocation.service';
import { MapCorePopupService, Popup } from 'src/app/vitis/map-core/services/map-core-popup/map-core-popup.service';

@Component({
  selector: 'app-vmap-add-geometry-mobile',
  templateUrl: './vmap-add-geometry-mobile.component.html',
  styleUrls: ['./vmap-add-geometry-mobile.component.scss']
})
export class VmapAddGeometryMobileComponent implements OnInit, OnDestroy, OnChanges, AfterViewInit {

  @Input() olMap;
  @Input() oLayer;
  @Input() updateFeature: Feature<Geometry>;
  @Input() measureMode = false;

  @Output() cancelInsertGeometry = new EventEmitter();
  @Output() addFeature = new EventEmitter();

  public geomType = 'POINT';
  public isGeometryGeomType = false;

  private prevHistoric: any[] = [];
  private indexHistoric = -1;

  private feature: Feature<Geometry>;
  private vectorSource: VectorSource<Geometry>;
  private vectorLayer: VectorLayer<VectorSource<Geometry>>;
  private firstPointFeature: Feature<Geometry>;

  private locationPopup: Popup;

  public menuListId = 'menuListId_' + Date.now();
  public listGeomtypesId = 'listGeomtypesId_' + Date.now();

  public geometries = ['POINT', 'LINE', 'POLYGON'];

  private olFeatureStyle = new Style({
    fill: new Fill({
      color: 'rgba(255, 255, 255, 0.2)',
    }),
    stroke: new Stroke({
      color: '#ffcc33',
      width: 2,
    }),
    image: new CircleStyle({
      radius: 7,
      fill: new Fill({
        color: '#ffcc33',
      }),
    })
  });

  public modalGeomtypeSelection = 'modalGeomtypeSelection_' + Date.now();

  /**
   * Retourne true s'il y a un historique suivant
   */
  get historicHasNextAction() {
    return this.vitisCheck.isDefAndNotNull(this.prevHistoric[this.indexHistoric + 1]);
  }

  /**
   * Retourne true s'il y a une action précédente dans l'historique
   */
  get historicHasPreviousAction() {
    return this.indexHistoric > 0;
  }

  /**
   * Retourne true si this.vectorSource possède des features
   */
  get isFeatureExsist() {
    return this.vitisCheck.isDefAndNotNull(this.vectorSource) && this.vectorSource.getFeatures().length > 0;
  }

  /**
   * Constructeur
   */
  constructor(
    private mapCoreDrawService: MapCoreDrawService,
    private layerToolsService: LayerToolsService,
    private vmapToolsMeasureService: VmapToolsMeasureService,
    private geolocationSrvc: VmapToolsGeolocationService,
    private vitisCheck: VitisCheck,
    private mapPopup: MapCorePopupService
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit(): void {
    // Autorise le dessin sur la couche vector layer
    this.vectorSource = new VectorSource();
    this.vectorLayer = this.mapCoreDrawService.initDrawLayer(this.olMap, this.vectorSource, this.olFeatureStyle);

    // Initialise la feature avec le geomtype de la couche
    if (this.vitisCheck.isDefAndNotNull(this.oLayer)) {
      this.geomType = this.layerToolsService.getLayerGeomType(this.oLayer);
      this.isGeometryGeomType = this.oLayer['ms_geomtype'] && this.oLayer['ms_geomtype']['label'] === 'GEOMETRY';
    } else {
      this.isGeometryGeomType = true;
    }
    this.feature = this.layerToolsService.getInitFeature(this.geomType);

    if (this.measureMode) {
      this.geometries = this.vmapToolsMeasureService.measuresTypes;
    }
  }

  /**
   * Destruction du composant
   */
  ngOnDestroy(): void {
    this.cancelInsert();
  }

  /**
   * Appelé lorsqu'il y a des changements sur les input
   */
  ngOnChanges(changes: SimpleChanges): void {
    // Utilisé pour redessiner une feature après avoir fermer le formulaire sans validation
    if (changes.updateFeature && changes.updateFeature.currentValue && !this.vectorLayer) {
      this.feature = new Feature({ geometry: changes.updateFeature.currentValue.getGeometry() });
      this.vectorSource.addFeature(this.feature);
      this.vectorLayer = this.mapCoreDrawService.initDrawLayer(this.olMap, this.vectorSource, this.olFeatureStyle);
    }
  }

  /**
   * Appelé après le chargement du DOM
   */
  ngAfterViewInit() {
    if (this.isGeometryGeomType) {
      document.getElementById(this.modalGeomtypeSelection).classList.toggle('show');
    }
  }

  /**
   * Supprime la couche de dessin
   */
  private cancelCurrentDraw() {
    this.feature = null;
    this.vectorSource.clear();
    if (this.vitisCheck.isDefAndNotNull(this.vectorLayer)) {
      this.olMap.removeLayer(this.vectorLayer);
      this.vectorLayer = null;
    }
  }

  /**
   * Ouvre la liste des éléments du menu
   */
  public openDropdownMenuList() {
    document.getElementById(this.menuListId).classList.toggle('show');
  }

  /**
   * Annule la saisie
   */
  public cancelInsert() {
    if (this.vitisCheck.isDefAndNotNull(this.locationPopup)) {
      this.locationPopup.close();
    }
    this.removeFeatures();
    this.cancelCurrentDraw();
    this.cancelInsertGeometry.emit();
  }

  /**
   * Annule la dernière action
   */
  public previousAction() {
    if (this.historicHasPreviousAction) {
      this.indexHistoric--;
      this.updateHistoric();
    }
  }

  /**
   * Renvoie à la dernière action annulée
   */
  public nextAction() {
    if (this.historicHasNextAction) {
      this.indexHistoric++;
      this.updateHistoric();
    }
  }

  /**
   * Execute les changements provoqué par l'historique
   */
  private updateHistoric() {
    const historic = this.prevHistoric[this.indexHistoric];
    if (historic.type === 'feature') {
      if (this.feature.getGeometry().getType() !== historic.feature.getGeometry().getType()) {
        this.changeGeomType(historic.feature.getGeometry().getType().toUpperCase());
      }
      const currGeometry: any = this.feature.getGeometry();
      const historicGeometry: any = historic.feature.getGeometry();
      switch (currGeometry.getType()) {
        case 'Circle':
          currGeometry.setCenterAndRadius(historicGeometry.getCenter(), historicGeometry.getRadius());
          break;
        default:
          currGeometry.setCoordinates(historicGeometry.getCoordinates());
          break;
      }
      this.feature = new Feature({ geometry: currGeometry });
      if (!this.vectorSource.getFeatures().length) {
        this.vectorSource.addFeature(this.feature);
      }
      if (historic.hasOwnProperty('firstPointFeature')) {
        this.firstPointFeature = historic.firstPointFeature;
        this.vectorSource.addFeature(this.firstPointFeature);
      } else {
        if (this.vitisCheck.isDefAndNotNull(this.firstPointFeature) && this.vectorSource.hasFeature(this.firstPointFeature)) {
          this.vectorSource.removeFeature(this.firstPointFeature);
        }
      }
    }
  }

  /**
   * Ajoute un historique
   * @param feature
   */
  private addHistoric(feature: Feature<Geometry>, firstPointFeature?: Feature) {
    if (this.indexHistoric < this.prevHistoric.length) {
      this.prevHistoric = [...this.prevHistoric.slice(0, this.indexHistoric + 1)];
    }
    this.indexHistoric = this.prevHistoric.length;
    if (this.vitisCheck.isDefAndNotNull(firstPointFeature)) {
      this.prevHistoric.push({ type: 'feature', feature: feature.clone(), firstPointFeature });
    } else {
      this.prevHistoric.push({ type: 'feature', feature: feature.clone() });
    }
  }

  /**
   * Renvoie la feature dessiner
   * @returns
   */
  public appendFeature() {
    if (this.vitisCheck.isDefAndNotNull(this.feature)) {
      if (this.measureMode) {
        this.vmapToolsMeasureService.infosMeasure(this.feature, this.olMap);
      }
      this.addFeature.emit(this.feature);
      this.removeFeatures();
      this.cancelCurrentDraw();
    }
  }

  /**
   * Supprime toutes les features du vector source
   */
  private removeFeatures() {
    const features = this.vectorSource.getFeatures();
    if (features.length > 0) {
      features.forEach(removeFeature => this.vectorSource.removeFeature(removeFeature));
    }
  }

  /**
   * Centre la carte sur la localisation actuelle
   */
  public localisationGPS() {
    const position = this.geolocationSrvc.getPosition();
    if (position) {
      this.olMap.getView().setCenter(position);
    }
  }

  /**
   * Ajoute une géométrie
   */
  public addGeometry() {
    if (!this.vectorLayer) {
      this.vectorLayer = this.mapCoreDrawService.initDrawLayer(this.olMap, this.vectorSource, this.olFeatureStyle);
    }
    const centerPoint: Point = this.getCenterMapCross();
    if (this.geomType == 'CIRCLE') {
      this.addFeatureCircle(centerPoint);
    } else {
      this.addPoint(centerPoint);
    }
  }

  /**
   * Modifie la géométrie en cours ou la supprime si c'est une géométrie simple
   */
  public addPoint(centerPoint) {
    let firstPoint = false;
    switch (this.geomType) {
      case 'POINT':
        this.addFeaturePoint(centerPoint);
        break;
      case 'LINE':
      case 'LINESTRING':
        firstPoint = this.addFeatureLineString(centerPoint);
        break;
      case 'POLYGON':
        firstPoint = this.addFeaturePolygon(centerPoint);
        break;
    }
    if (firstPoint) {
      this.firstPointFeature = new Feature(centerPoint);
      this.vectorSource.addFeature(this.firstPointFeature);
      this.addHistoric(this.feature.clone(), this.firstPointFeature);
    } else {
      if (this.vitisCheck.isDefAndNotNull(this.firstPointFeature) && this.vectorSource.hasFeature(this.firstPointFeature)) {
        this.vectorSource.removeFeature(this.firstPointFeature);
      }
      this.addHistoric(this.feature.clone());
    }
  }

  /**
   * Ajout d'une feature de géométrie point sur la carte
   * @param point
   */
  private addFeaturePoint(point: Point) {
    this.removeFeatures();
    this.feature = new Feature({ geometry: point });
    this.vectorSource.addFeature(this.feature);
  }

  /**
   * Ajoute un point à la feature de géométrie linestring
   * @param point
   */
  private addFeatureLineString(point: Point): boolean {
    const currLine: any = this.feature ? this.feature.getGeometry() : new LineString([]);
    const currCoordinates = currLine.getCoordinates();

    currLine.appendCoordinate(point.getCoordinates());
    this.feature = new Feature({ geometry: currLine });

    if (currCoordinates.length === 0) {
      this.vectorSource.addFeature(this.feature);
      return true;
    }
    return false;
  }

  /**
   * Ajoute un point à la feature de géométrie polygon
   * @param point
   */
  private addFeaturePolygon(point: Point): boolean {
    const currPolygon: any = this.feature ? this.feature.getGeometry() : new Polygon([]);
    let currCoordinates = currPolygon.getCoordinates();
    if (currCoordinates.length === 0) {
      currCoordinates = [[point.getCoordinates(), point.getCoordinates()]];
      currPolygon.setCoordinates(currCoordinates);
      this.feature = new Feature({ geometry: currPolygon });
      this.vectorSource.addFeature(this.feature);
      return true;
    } else {
      currCoordinates[0].splice(currCoordinates[0].length - 1, 0, point.getCoordinates());
      currPolygon.setCoordinates(currCoordinates);
      this.feature = new Feature({ geometry: currPolygon });
    }
    return false;
  }

  /**
   * Ajoute une feature de géométrie cercle sur la carte
   * @param point
   */
  private addFeatureCircle(point: Point): boolean {

    this.removeFeatures();
    this.feature = new Feature({ geometry: new Circle(point.getCoordinates()) });
    this.vectorSource.addFeature(this.feature);
    document.getElementById('vertical_line').style.visibility = 'hidden';
    document.getElementById('horizontal_line').style.visibility = 'hidden';
    this.locationPopup = this.mapPopup.displayPopup('<div><div class="mb-2">Rayon (en m) : </div><input type="text" class="form-control form-control-sm" id="set-radius-circle"></input></div>', this.feature, this.olMap, {
      hideCloseBtn: false, displayFeature: false, buttons: [
        {
          type: 'btn',
          position: 'right',
          text: 'Ok',
          class: 'btn-primary',
          bo: this,
          event:
            () => {
              const input: HTMLInputElement = document.getElementById('set-radius-circle') as HTMLInputElement;
              const circle = this.feature.getGeometry();
              if (!(circle instanceof Circle)) {
                console.error('La géométrie n\'est pas un cercle');
              } else {
                document.getElementById('vertical_line').style.visibility = 'visible';
                document.getElementById('horizontal_line').style.visibility = 'visible';
                circle.setRadius(parseInt(input.value, 10));
                this.feature = new Feature({ geometry: circle });
                this.vectorSource.addFeature(this.feature);
                this.addHistoric(this.feature.clone());
                this.locationPopup.close();
              }
            }
        }]
    });
    return true;
  }

  /**
   * Retourne la position du centre de la croix
   * @returns
   */
  private getCenterMapCross(): Point {
    const center = this.olMap.getView().getCenter();
    return new Point(center);
  }

  /**
   * Ouvre la liste des geomtypes pour les dessiner
   */
  public openDropdownGeomtypesList() {
    document.getElementById(this.listGeomtypesId).classList.toggle('show');
  }

  /**
   * Change le geomtype choisit pour le dessin
   * @param geomType
   */
  public changeGeomType(geomType) {
    document.getElementById(this.listGeomtypesId).classList.remove('show');
    this.geomType = geomType;
    this.removeFeatures();
    this.feature = this.layerToolsService.getInitFeature(this.geomType);
  }

  /**
   * Initialise le type de géométrie à insérer quand la couche avec plusieurs géométrie
   * @param event
   */
  public initGeomType(event) {
    const geomType = event.target.value;
    this.changeGeomType(geomType);
  }

  /**
  * Ferme la modal d'initialisation du type de géométrie
  */
  public closeInitGeomType() {
    document.getElementById(this.modalGeomtypeSelection).classList.remove('show');
  }

  /**
   * Supprime toute les mesures sur la carte
   */
  public deleteMeasures() {
    if (this.measureMode) {
      // Suppression des popups
      this.vmapToolsMeasureService.emptiedInfosMeasures();
      // Suppression des mesures
      if (this.vitisCheck.isDefAndNotNull(this.olMap.get('measure_layer'))) {
        const mesure_layer_source = this.olMap.get('measure_layer').getSource();
        const features = mesure_layer_source.getFeatures();
        if (features.length) {
          features.forEach(removeFeature => mesure_layer_source.removeFeature(removeFeature));
        }
        mesure_layer_source.clear();
      }
    }
  }
}
