import { Component, OnInit, Input, OnDestroy } from '@angular/core';
import { Map, Feature } from 'ol';
import { Geometry, Point, Polygon } from 'ol/geom';
import { VmapToolsService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools.service';
import { Circle as CircleStyle, Fill, Stroke, Style } from 'ol/style';
import { Vector as VectorSource } from 'ol/source';
import { Vector as VectorLayer } from 'ol/layer';
import { unByKey } from 'ol/Observable';
import { Subscription } from 'rxjs';
import { EventsKey } from 'ol/events';
import { VmapToolsGeolocationService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools-geolocation/vmap-tools-geolocation.service';
import { MapCoreDrawService } from 'src/app/vitis/map-core/services/map-core-draw/map-core-draw.service';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import { environment } from 'src/environments/environment';
import { vitisAlert } from 'src/app/vitis/core/services/utils/vitis-utils';
import { TranslateService } from '@ngx-translate/core';
@Component({
  selector: 'app-vmap-tools-geolocation-mobile',
  templateUrl: './vmap-tools-geolocation-mobile.component.html',
  styleUrls: ['./vmap-tools-geolocation-mobile.component.scss']
})
export class VmapToolsGeolocationMobileComponent implements OnInit, OnDestroy {

  @Input() public olMap: Map;

  private currentPosition: number[];
  private currentAccuracy: any;
  private accuracyFeature: Feature<Geometry>;
  private positionFeature: Feature<Geometry>;
  private vectorLayer: VectorLayer<VectorSource<Geometry>>;
  private vectorSource: VectorSource<Geometry>;

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];
  private mapSubscriptions: EventsKey[] = [];

  /**
   * Retourne true si le tracking GPS est actif
   */
  get trackingGps() {
    return this.geolocationSrvc.isTrackingEnable();
  }

  /**
   * Retourne true s'il faut afficher la précision
   */
  get accuracyGps() {
    return this.geolocationSrvc.isAccuracyEnable();
  }

  /**
   * Constructeur
   */
  constructor(
    private vmapToolsService: VmapToolsService,
    private geolocationSrvc: VmapToolsGeolocationService,
    private mapCoreDrawService: MapCoreDrawService,
    private mapUtils: MapCoreUtilsService,
    private translate: TranslateService,

  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit() {
    // Initialise les features et les ajoutes à la carte
    this.initGPSFeature();

    // Écoute l'ouverture des menus pour lancer le tracking au click
    this.subscriptions.push(this.vmapToolsService.getPanelBtnClickedEmiter().subscribe((sTool: string) => {
      if (sTool === 'geolocation-mobile') {
        this.centerMapPosition();
      }
    }));

    if (!this.geolocationSrvc.getPosition().length) {
      this.geolocationSrvc.initGeolocation(this.olMap.getView().getProjection());
    }

    // Event du tracking de la précision de la position gps
    this.subscriptions.push(this.geolocationSrvc.getTrackingAccuracyGeometry().subscribe((accuracyGeometry: any) => {
      if (accuracyGeometry && accuracyGeometry.length !== 0) {
        this.currentAccuracy = accuracyGeometry;
        if (this.accuracyFeature) { // Met à jour la précision
          this.setAccuracyFeature();

          // Affiche la feature sur la carte si accuracyGps est activé
          if (this.accuracyGps && !this.vectorSource.hasFeature(this.accuracyFeature)) {
            this.vectorSource.addFeature(this.accuracyFeature);

          }
        }
      }
    }));

    // Event du tracking de la position gps
    this.subscriptions.push(this.geolocationSrvc.getTrackingPosition().subscribe(position => {
      if (position && position.length !== 0) {
        this.currentPosition = position;
        if (this.positionFeature) { // Met à jour la position
          this.setPositionFeature();

          // Si le tracking est activé, centra la carte sur la position
          if (this.trackingGps) {
            this.centerMapPosition(true);
          }
        }
      }
    }));

    // Click sur le bouton afficher la précision
    this.subscriptions.push(this.geolocationSrvc.getSubscribeAccuraryGps().subscribe(enable => {
      if (this.positionFeature) {
        if (enable) {
          this.vectorSource.addFeature(this.accuracyFeature);
        } else {
          this.vectorSource.removeFeature(this.accuracyFeature);
        }
      }
    }));

    // Gestion du changement de carte sur une autre projection
    this.mapSubscriptions.push(this.olMap.on('change', () => {
      this.geolocationSrvc.initGeolocation(this.olMap.getView().getProjection());
    }));
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
    this.mapSubscriptions.forEach(mapSubscription => unByKey(mapSubscription));
    this.cancelGPSFeature();
    this.currentPosition = null;
    this.currentAccuracy = null;
  }

  /**
   * Initialise les features et les ajoutes à la carte
   */
  private initGPSFeature() {
    this.accuracyFeature = new Feature(new Polygon([]));
    this.positionFeature = new Feature(new Point([]));

    this.positionFeature.setStyle(
      new Style({
        image: new CircleStyle({
          radius: 6,
          fill: new Fill({
            color: '#3399CC',
          }),
          stroke: new Stroke({
            color: '#fff',
            width: 2,
          }),
        }),
      })
    );

    this.vectorSource = new VectorSource();
    this.vectorLayer = this.mapCoreDrawService.initDrawLayer(this.olMap, this.vectorSource);
    this.vectorSource.addFeature(this.positionFeature);
  }

  /**
   * Annule l'outil Gps
   */
  private cancelGPSFeature() {
    this.positionFeature = null;
    this.accuracyFeature = null;
    if (this.vectorSource) {
      this.vectorSource.clear();
    }
    if (this.vectorLayer) {
      this.olMap.removeLayer(this.vectorLayer);
      this.vectorLayer = null;
    }
  }

  /**
   * Set la feature accuracy
   */
  private setAccuracyFeature() {
    const accuracy: any = this.accuracyFeature.getGeometry();
    accuracy.setCoordinates(this.currentAccuracy.getCoordinates());
  }

  /**
   * Set la feature position
   */
  private setPositionFeature() {
    const positionGeometry: any = this.positionFeature.getGeometry();
    positionGeometry.setCoordinates(this.currentPosition);
  }

  /**
   * Centre la map sur la position
   */
  private centerMapPosition(bForceNoZoom = false) {

    if (this.currentPosition != null ) { // On vérifie que l'utilisateur a bien autorisé la localisation
      if (bForceNoZoom !== true && environment.properties.map_zoom_threshold_for_gps_location === '23') {
        this.mapUtils.zoomLocationGPS(this.olMap, this.accuracyFeature.getGeometry(), [200, 0, 200, 0], bForceNoZoom);
      } else {
        this.mapUtils.zoomLocationGPS(this.olMap, this.positionFeature.getGeometry(), [200, 0, 200, 0], bForceNoZoom);
      }
    }else {
      this.translate.get('COMPONENT_VMAP_MOBILE_TOOLS_LOCATION_DISABLED').subscribe((sTranslate: string) => {
        const sHtml = '<p>' + sTranslate + '</p>' ;
        vitisAlert.alert({
          'type': 'warning',
          'width': '400',
          'html': sHtml,
        });
      });
    }
  }
}
