import { Component, OnInit, ViewChild } from '@angular/core';
import { FormBuilder, FormGroup } from '@angular/forms';
import { TranslateService } from '@ngx-translate/core';
import { vitisAlert, VitisCheck, VitisClone, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { DataGridOptions } from 'src/app/vitis/data-grid/data-grid.model';
import { FormreaderComponent } from 'src/app/vitis/formreader/formreader.component';
import { FormreaderService } from 'src/app/vitis/formreader/services/formreader.service';
import { VmapLayer } from 'src/app/vitis/map-core/services/map-models.model';
import { environment } from 'src/environments/environment';
import { GtfWorkspace } from '../../../../services/vmap-tools/vmap-tools.service';

@Component({
  selector: 'app-vmap-gtf-workspace-modal',
  templateUrl: './vmap-gtf-workspace-modal.component.html',
  styleUrls: ['./vmap-gtf-workspace-modal.component.scss']
})
export class VmapGtfWorkspaceModalComponent implements OnInit {

  public JSONform: any;
  public workspace: GtfWorkspace;
  public layer: VmapLayer;
  public entity: any; // quand il est lancé depuis une entité
  public dt: any; // quand il est lancé depuis le requêteur
  public parentDataGridOptions: DataGridOptions; // quand il est lancé depuis le requêteur
  public closeModal: any; // Fonction pour fermer la modale
  public sSelectionMode: string; // requestor / entity

  @ViewChild(FormreaderComponent) public GtfFormModalFormReader: FormreaderComponent;

  public workspaceForm: FormGroup;

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private vitisClone: VitisClone,
    private vitisRequest: VitisRequest,
    private translate: TranslateService,
    private formReaderService: FormreaderService,
    private formBuilder: FormBuilder,
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit(): void {
    // formulaire d'exportation des données du requestor.
    this.workspaceForm = this.formBuilder.group({
      export_type: ['selected_rows'],
    });
  }

  /**
   * Initialise le formulaire
   * @param sFormName
   */
  public initForm(sFormName: string) {
    this.sSelectionMode = this.entity ? 'entity' : 'requestor';
    this.GtfFormModalFormReader.initForm(JSON.parse(this.JSONform), sFormName);
  }

  /**
   * Crée la demande GTF
   */
  public createGtfOrder() {
    const oFormValues: any = this.GtfFormModalFormReader.getFormValues(true);
    const wkParameters: any = this.getWkParams_(oFormValues);
    const sExportType: string = this.workspaceForm.value.export_type;

    if (this.GtfFormModalFormReader.isFormValid()) {
      this.getDataForExport(sExportType).then((data) => {
        const jsonParams = JSON.parse(wkParameters['sWkParams']);
        jsonParams['VMAP_REQUESTOR_DATA'] = JSON.stringify(data);
        jsonParams['VMAP_REQUESTOR_LAYER'] = this.layer.layer_id;
        wkParameters['sWkParams'] = JSON.stringify(jsonParams);

        this.createOrder(this.workspace.workspace_id, wkParameters, {}).then(() => {
          // Message de succès.
          vitisAlert.alert({
            type: 'success',
            title: this.translate.instant('GTF_MODAL_PROJECT_RUN_ORDER_SUCCESS'),
            text: this.translate.instant('GTF_MODAL_PROJECT_RUN_ORDER_SUCCESS_2'),
          });
          this.closeModal();
        }).catch(() => {
          this.translate.get('GTF_MODAL_PROJECT_RUN_ORDER_FAILURE').subscribe((translation: string) => {
            vitisAlert.fireToast({
              icon: 'error',
              title: translation
            });
          });
          this.closeModal();
        });
      });
    }
  }

  /**
   * Ferme la modale
   */
  public callCloseModal() {
    if (this.vitisCheck.isFunction(this.closeModal)) {
      this.closeModal();
    }
  }

  /**
   * Retourne les wk_params d'apprès les résultats du formulaire
   */
  private getWkParams_(oFormValues: any) {
    const oWkParams = {};
    const oWkFiles = {};
    const ignoredKeys = ['vitis_thumbnail_image_files', 'widget_email'];

    for (const key in oFormValues) {
      if (oFormValues.hasOwnProperty(key)) {

        const oField = this.GtfFormModalFormReader.getFormField(key, JSON.parse(this.JSONform), 'insert');

        // Champs à ne pas envoyer
        if (ignoredKeys.indexOf(key) !== -1) {
          continue;
        } else if (oFormValues[key] instanceof File) {
          const sFileField = 'vitis_file_' + key;
          oWkParams[key] = {
            type: 'file',
            name: oFormValues[key].name,
            size: oFormValues[key].size,
            field: sFileField
          };
          oWkFiles[sFileField] = oFormValues[key];
        } else if (oFormValues[key] instanceof Array && oField.type === 'upload') {
          const sFileField = 'vitis_file_' + key;
          oWkParams[key] = {
            type: 'multifile',
            field: sFileField,
            files: []
          };
          oWkFiles[sFileField] = [];
          for (let i = 0; i < oFormValues[key].length; i++) {
            if (oFormValues[key][i] instanceof File) {
              oWkParams[key]['files'].push({
                type: 'file',
                name: oFormValues[key][i].name,
                size: oFormValues[key][i].size,
                field: sFileField
              });
              oWkFiles[sFileField].push(oFormValues[key][i]);
            } else {
              oWkParams[key].push(oFormValues[key][i]);
            }
          }
        } else {
          oWkParams[key] = oFormValues[key];
        }
      }
    }

    return {
      sWkParams: JSON.stringify(oWkParams),
      oWkFiles
    };
  }

  /**
   * Récupère la donnée pour le lancement du traitement
   * @param sExportType
   * @returns
   */
  private getDataForExport(sExportType: string): Promise<any> {
    return new Promise((resolve, reject) => {
      if (this.sSelectionMode === 'entity') {
        if (this.entity) {
          if (this.vitisCheck.isDefAndNotNull(this.entity[this.layer.idfield])) {
            resolve([this.entity[this.layer.idfield]]);
          }
        }
      } else if (this.sSelectionMode = 'requestor') {
        let iNbRows: number;
        switch (sExportType) {
          case 'current_filter':
            this.getDataGridIds(this.workspace.selection_max_entities).then(aData => {
              resolve(aData);
            });
            break;
          case 'all_rows':
            vitisAlert.alert({
              text: this.translate.instant('GTF_MODAL_PROJECT_TYPE_ALL_DATA_WARNING'),
              type: 'warning',
              showCancelButton: true,
              confirmButtonText: this.translate.instant('GTF_MODAL_PROJECT_TYPE_ALL_DATA_WARNING_OK'),
              cancelButtonText: this.translate.instant('GTF_MODAL_PROJECT_TYPE_ALL_DATA_WARNING_CANCEL'),
            }).then((result) => {
              if (result.isDismissed === false) {
                resolve('all');
              }
            });
            break;
          case 'selected_rows':
            const aSelectedRows = this.dt.getSelectedRows_();
            if (aSelectedRows.length > this.workspace.selection_max_entities) {
              vitisAlert.fireToast({
                icon: 'error',
                title: this.translate.instant('GTF_MODAL_PROJECT_SELECT_MAX_ENTITIES_ERROR').replace('[limit]', this.workspace.selection_max_entities)
              });
              reject();
              return;
            }
            if (aSelectedRows.length === 0) {
              vitisAlert.fireToast({
                icon: 'error',
                title: this.translate.instant('GTF_MODAL_PROJECT_MISSING_SELECT')
              });
              reject();
              return;
            }
            const aIds = [];
            aSelectedRows.each((row: any) => {
              if (this.vitisCheck.isDefAndNotNull(row[this.layer.idfield])) {
                aIds.push(row[this.layer.idfield]);
              }
            });
            resolve(aIds);
            break;
          default:
            vitisAlert.fireToast({
              icon: 'error',
              title: this.translate.instant('GTF_MODAL_PROJECT_SELECT_METHOD_NOT_VALID')
            });
            reject();
            break;
        }
      }
    });
  }

  /**
   * Retourne l'ensemble des ID de la page en cours
   * @param iNbRows permet de limiter le nombre de lignes
   * @returns
   */
  private getDataGridIds(iNbRows?: number): Promise<any[]> {
    return new Promise((resolve, reject) => {
      const oDataOption = this.vitisClone.clone(this.dt.ajaxParams);

      if (iNbRows) {
        oDataOption.length = iNbRows;
      } else {
        oDataOption.length = this.dt.oDataGridOptionsRecordsTotal;
      }

      // Ne retient que la colonne des id
      oDataOption.columns = [{
        data: this.layer.idfield,
        name: this.layer.idfield,
      }];

      this.parentDataGridOptions.datatableOptions.ajax(oDataOption, (oReturn: any) => {

        // Vérification limite d'enregistrements
        if (oReturn.recordsTotal > this.workspace.selection_max_entities) {
          vitisAlert.fireToast({
            icon: 'error',
            title: this.translate.instant('GTF_MODAL_PROJECT_SELECT_MAX_ENTITIES_ERROR').replace('[limit]', this.workspace.selection_max_entities)
          });
          reject();
          return;
        }

        const aIds = [];
        oReturn.data.forEach((row: any) => {
          if (this.vitisCheck.isDefAndNotNull(row[this.layer.idfield])) {
            aIds.push(row[this.layer.idfield]);
          }
        });
        resolve(aIds);
      }, null);
    });
  }

  /**
   * Crée la demande GTF
   * @param workspaceId
   * @param wkParams contient sWkParams et oWkFiles
   * @param options
   */
  private createOrder(workspaceId: any, wkParams?: any, options?: any): Promise<void> {
    return new Promise((resolve, reject) => {

      // workspace_id
      let params: any = {
        workspace_id: workspaceId
      };

      // wk_params
      if (this.vitisCheck.isDefAndNotNull(wkParams.sWkParams)) {
        params.wk_params = wkParams.sWkParams;
      }
      if (this.vitisCheck.isDefAndNotNull(wkParams.oWkFiles)) {
        for (const key in wkParams.oWkFiles) {
          if (wkParams.oWkFiles.hasOwnProperty(key)) {
            params[key] = wkParams.oWkFiles[key];
          }
        }
      }

      // Autres paramètres facultatifs
      if (this.vitisCheck.isObject(options)) {
        params = {
          ...params,
          ...options
        };
      }
      this.formReaderService.manageLargesFiles(params).then((fileValues: any) => {
        this.vitisRequest.ajaxRequestPromise({
          method: 'POST',
          url: `${environment.properties.app_api_url}/vmap/layers/${this.layer.layer_id}/fme-userworkspaces/${workspaceId}/createorder`,
          sendToken: true,
          data: this.formReaderService.transformToDataForm(fileValues)
        }).then((resp: any) => {
          if (resp.data &&
            resp.data.data) {
            resolve();
          } else {
            console.error('could not create order', resp);
            reject();
          }
        }).catch((err) => {
          console.error('could not create order', err);
          reject();
        });
      });
    });
  }
}
