import { Injectable } from '@angular/core';
import { VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import { FormreaderService } from 'src/app/vitis/formreader/services/formreader.service';
import { Map } from 'ol';

@Injectable({
  providedIn: 'root'
})

export class VmapToolsCadastreService {

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private auth: AuthService,
    private formreaderSrvc: FormreaderService) { }

  /**
  * Permet de récupérer le privilège du cadastre le plus élevé
  * @returns privilege string
  */
  public getPrivilege(): string {
    const privileges = this.auth.getPrivileges();
    let privilege: string;
    if (privileges) {
      for (let i = 0; i < privileges.length; i++) {
        if (privileges[i] == 'vmap_cadastre_user') {
          privilege = privileges[i];
        }

        if (privileges[i] == 'vmap_cadastre_medium_user') {
          if (privilege != 'vmap_cadastre_user') {
            privilege = privileges[i];
          }
        }

        if (privileges[i] == 'vmap_cadastre_light_user') {
          if (privilege != 'vmap_cadastre_user' && privilege != 'vmap_cadastre_medium_user') {
            privilege = privileges[i];
          }
        }
      }
      if (privilege == null) {
        console.error('There are no access privileges for the cadastre');
      }
    } else {
      console.error('There are no privileges');
    }
    return privilege;
  }


  /**
  * Récupère les données des départements
  * @returns Promise<any>
  */
  public loadDepartements_(): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/departements`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        params: {
          order_by: 'dep_code',
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting list of departements');
          reject(false);
        }
      });
    });
  }

  /**
  * Récupère les données des communes
  * @returns Promise<any>
  */
  public loadCommunes_(map?: Map): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/communes`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        params: {
          order_by: 'nom',
          srid: this.vitisCheck.isDefAndNotNull(map) ? map.getView().getProjection().getCode().substring(5) : ''
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting list of communes');
          reject(false);
        }
      });
    });
  }

  /**
* Récupère les données des communes par département
* @param oDepartement string département
* @returns Promise<any>
*/
  public loadCommunesParDepartement_(oDepartement: string, map?: Map): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/communes`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        params: {
          order_by: 'nom',
          srid: this.vitisCheck.isDefAndNotNull(map) ? map.getView().getProjection().getCode().substring(5) : '',
          filter: {
            relation: 'AND',
            operators: [{
              column: 'dep_code',
              compare_operator: '=',
              value: oDepartement
            }]
          }
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting list of communes');
          reject(false);
        }
      });
    });
  }

  /**
  * Récupère les données des sections par commune
  * @param oCommune string commune
  * @returns Promise<any>
  */
  public loadSections_(oCommune: string, map?: Map): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/sections`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        params: {
          order_by: 'section',
          srid: this.vitisCheck.isDefAndNotNull(map) ? map.getView().getProjection().getCode().substring(5) : '',
          filter: {
            relation: 'AND',
            operators: [{
              column: 'id_com',
              compare_operator: '=',
              value: oCommune
            }]
          }
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting list of sections');
          reject(false);
        }
      });
    });
  }

  /**
 * Récupère les données des lieux-dits par commune
 * @param oCommune string commune
 * @returns Promise<any>
 */
  public loadLieuxDits_(oCommune: string, map?: Map): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/lieudits`;

      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        params: {
          order_by: 'tex',
          srid: this.vitisCheck.isDefAndNotNull(map) ? map.getView().getProjection().getCode().substring(5) : '',
          filter: {
            relation: 'AND',
            operators: [{
              column: 'id_com',
              compare_operator: '=',
              value: oCommune
            }]
          }
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting list of locality');
          reject(false);
        }
      });
    });
  }

  /**
   * Récupère les données des propriétaires par commune et propriétaire
   * @param oCommune string commune
   * @param oProprietaire string premières lettres du propriétaire
   * @returns Promise<any>
   */
  public loadProprietaires_(oCommune: string, oProprietaire: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/proprietaires`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        params: {
          order_by: 'DDENOM',
          attributs: 'DDENOM|ID_COM',
          distinct: true,
          limit: 500,
          filter: {
            relation: 'AND',
            operators: [{
              column: 'ID_COM',
              compare_operator: '=',
              value: oCommune,
            }, {
              column: 'DDENOM',
              compare_operator: 'LIKE',
              value: '%' + oProprietaire + '%',
              compare_operator_options: {
                case_insensitive: true
              }
            }]
          }
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting list of owners');
          reject(false);
        }
      });
    });
  }

  /**
   * Récupère les données des comptes par commune et propriétaire
   * @param oCommune string commune
   * @param oProprietaire string nom du propriétaire
   * @returns Promise<any>
   */
  public loadComptesParProprietaires_(oCommune: string, oProprietaire: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/proprietaires`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        params: {
          order_by: 'DNUPRO',
          attributs: 'ID_COM|DNUPRO',
          filter: {
            relation: 'AND',
            operators: [{
              column: 'ID_COM',
              compare_operator: '=',
              value: oCommune,
            }, {
              column: 'DDENOM',
              compare_operator: '=',
              value: oProprietaire,
            }]
          }
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting list of accounts by owner');
          reject(false);
        }
      });
    });
  }

  /**
  * Récupère les numéros des parcelles par commune et comptes
  * @param oCommune string commune
  * @param oCompte Array<string> liste des comptes sélectionnés
  * @returns Promise<any>
  */
  public loadDescriptionParcelleParProprietaires_(oCommune: string, oCompte: Array<string>): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/descriptionparcelles`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        params: {
          order_by: 'ID_PAR',
          attributs: 'ID_PAR',
          filter: {
            relation: 'AND',
            operators: [{
              column: 'ID_COM',
              compare_operator: '=',
              value: oCommune,
            }, {
              column: 'DNUPRO',
              compare_operator: 'IN',
              value: oCompte,
            }]
          }
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting list of parcels by owner');
          reject(false);
        }
      });
    });
  }

  /**
  * Récupère les données des invariants par comptes
  * @param oCompte Array<string> liste des comptes sélectionnés
  * @returns Promise<any>
  */
  public loadInvariants_(oCompte: Array<any>): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/invariants`;
      // Récupère le formulaire
      const test: any = [];
      for (let i = 0; i < oCompte.length; i++) {
        test.push({
          relation: 'AND',
          operators: [{
            column: 'DNUPRO',
            compare_operator: '=',
            value: oCompte[i].DNUPRO
          }, {
            column: 'ID_COM',
            compare_operator: '=',
            value: oCompte[i].ID_COM
          }]
        });
      }
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        params: {
          order_by: 'INVAR',
          attributs: 'INVAR|ID_PAR|id_local',
          filter: {
            relation: 'OR',
            operators: test
          }
        }
      }).then((resp) => {
        if (resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting list of invariants');
          reject(false);
        }
      });
    });
  }

  /**
   * Récupère les données des parcelles par l'identifiant de la parcelle
   * @param oParcelle Array<string> liste des parcelles sélectionnées
   * @param map Map carte OpenLayer (optionnelle)
   * @returns Promise<any>
   */
  public loadParcelles_(oParcelle: Array<string>, map?: Map): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/parcelles`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        params: {
          order_by: 'parcelle',
          srid: this.vitisCheck.isDefAndNotNull(map) ? map.getView().getProjection().getCode().substring(5) : '',
          filter: {
            relation: 'AND',
            operators: [{
              column: 'id_par',
              compare_operator: 'IN',
              value: oParcelle
            }]
          }
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting list of parcels');
          reject(false);
        }
      });
    });
  }

  /**
  * Récupère les données des parcelles par section et commune
  * @param oCommune string commune
  * @param oSection string section
  * @returns Promise<any>
  */
  public loadParcellesBySection_(oCommune: string, oSection: string, map?: Map): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/parcelles`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        params: {
          order_by: 'parcelle',
          srid: this.vitisCheck.isDefAndNotNull(map) ? map.getView().getProjection().getCode().substring(5) : '',
          filter: {
            relation: 'AND',
            operators: [{
              column: 'id_com',
              compare_operator: '=',
              value: oCommune,
            },
            {
              column: 'id_sec',
              compare_operator: '=',
              value: oSection,
            }]
          }
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting list of parcels by sections');
          reject(false);
        }
      });
    });
  }

  /**
  * Récupère les données des voies par commune
  * @param oCommune string commune
  * @returns Promise<any>
  */
  public loadVoiesByCommunes_(oCommune: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/voies`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        params: {
          order_by: 'DVOILIB',
          filter: {
            relation: 'AND',
            operators: [{
              column: 'ID_COM',
              compare_operator: '=',
              value: oCommune,
            }]
          }
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting list of ways');
          reject(false);
        }
      });
    });
  }

  /**
  * Récupère les données des adresses par voies
  * @param oVoie string numéro de la voie
  * @returns Promise<any>
  */
  public loadAdressesByVoie_(oVoie: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/adresses`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        params: {
          order_by: 'DNVOIRI',
          filter: {
            relation: 'AND',
            operators: [{
              column: 'ID_RIVOLI',
              compare_operator: '=',
              value: oVoie,
            }]
          }
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting list of adress');
          reject(false);
        }
      });
    });
  }

  /**
  * Récupère les données des parcelles par géometrie
  * @param sEWKT string coordonnées sous format EWKT
  * @param buffer number zone tampon autour de l'élement
  * @param typeOfGeom string type de géométrie (point | linestring | polygon)
  * @returns Promise<any>
  */
  public loadParcellesByGeom_(sEWKT: string, buffer: number, typeOfGeom: string, projection: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/parcelles`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        httpMethodOverride: true,
        params: {
          order_by: 'parcelle',
          result_srid: this.vitisCheck.isDefAndNotNull(projection) ? projection : '',
          filter: {
            relation: 'AND',
            operators: [{
              column: 'geom',
              compare_operator: 'intersect',
              value: sEWKT,
              compare_operator_options: {
                source_proj: this.getProjection(),
                intersect_buffer: buffer,
                intersect_buffer_geom_type: typeOfGeom
              }
            }]
          }
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting list of parcels by geometry');
          reject(false);
        }
      });
    });
  }

  /**
  * Récupère les données des communes par géometrie
  * @param sEWKT string coordonnées sous format EWKT
  * @param buffer number zone tampon autour de l'élement
  * @param typeOfGeom string type de géométrie (point | linestring | polygon)
  * @returns Promise<any>
  */
  public loadCommunesByGeom_(sEWKT: string, buffer: number, typeOfGeom: string, projection: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/communes`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        httpMethodOverride: true,
        params: {
          order_by: 'nom',
          result_srid: this.vitisCheck.isDefAndNotNull(projection) ? projection : '',
          filter: {
            relation: 'AND',
            operators: [{
              column: 'geom',
              compare_operator: 'intersect',
              value: sEWKT,
              compare_operator_options: {
                source_proj: this.getProjection(),
                intersect_buffer: buffer,
                intersect_buffer_geom_type: typeOfGeom
              }
            }]
          }
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting list of communes by geometry');
          reject(false);
        }
      });
    });
  }

  /**
  * Récupère les données des sections par géometrie
  * @param sEWKT string coordonnées sous format EWKT
  * @param buffer number zone tampon autour de l'élement
  * @param typeOfGeom string type de géométrie (point | linestring | polygon)
  * @returns Promise<any>
  */
  public loadSectionsByGeom_(sEWKT: string, buffer: number, typeOfGeom: string, projection: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/sections`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        httpMethodOverride: true,
        params: {
          order_by: 'nom',
          result_srid: this.vitisCheck.isDefAndNotNull(projection) ? projection : '',
          filter: {
            relation: 'AND',
            operators: [{
              column: 'geom',
              compare_operator: 'intersect',
              value: sEWKT,
              compare_operator_options: {
                source_proj: this.getProjection(),
                intersect_buffer: buffer,
                intersect_buffer_geom_type: typeOfGeom
              }
            }]
          }
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting list of sections by geometry');
          reject(false);
        }
      });
    });
  }

  /**
  * Récupère les données des lieux-dits par géometrie
  * @param sEWKT string coordonnées sous format EWKT
  * @param buffer number zone tampon autour de l'élement
  * @param typeOfGeom string type de géométrie (point | linestring | polygon)
  * @returns Promise<any>
  */
  public loadLieuxDitsByGeom_(sEWKT: string, buffer: number, typeOfGeom: string, projection: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/lieudits`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        httpMethodOverride: true,
        params: {
          order_by: 'tex',
          result_srid: this.vitisCheck.isDefAndNotNull(projection) ? projection : '',
          filter: {
            relation: 'AND',
            operators: [{
              column: 'geom',
              compare_operator: 'intersect',
              value: sEWKT,
              compare_operator_options: {
                source_proj: this.getProjection(),
                intersect_buffer: buffer,
                intersect_buffer_geom_type: typeOfGeom
              }
            }]
          }
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting list of lieu-dit by geometry');
          reject(false);
        }
      });
    });
  }

  /**
  * Récupère les descriptions des parcelles en fonction de l'identifiant de la parcelle
  * @param idPar string l'identifiant de la parcelle
  * @returns Promise<any>
  */
  public loadDescriptionParcelleByIdPar_(idPar: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/descriptionparcelles`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        httpMethodOverride: true,
        params: {
          filter: {
            relation: 'AND',
            operators: [{
              column: 'ID_PAR',
              compare_operator: '=',
              value: idPar
            }]
          }
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1 &&
          this.vitisCheck.isArray(resp.data.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting description of the parcels by id_par');
          reject(false);
        }
      });
    });
  }

  /**
* Récupère les données de la fiche descriptive en fonction de l'identifiant de la parcelle
* @param idPar string l'identifiant de la parcelle
* @returns Promise<any>
*/
  public loadFicheDescriptiveByIdPar_(idPar: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/fichedescriptiveparcelle/${idPar}`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting descriptive sheet by id_par');
          reject(false);
        }
      });
    });
  }

  /**
* Récupère les données de l'invariant de la fiche descriptive en fonction de l'identifiant du bâti
* @param idBat string l'identifiant du bâti
* Nécessite l'existence d'un bâti
* @returns Promise<any>
*/
  public loadInvariantByIdBat_(idBat: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/fichedescriptiveinvariant/${idBat}`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting invariant sheet by id_bat');
          reject(false);
        }
      });
    });
  }


  /**
  * Retourne le lien de téléchargement du PDF du relevé de propriété standard
  * @param idDnupro string compte communal du propriétaire de la parcelle
  * @returns Promise<any>
  */
  public downloadRPStandard_(idDnupro: number): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/relevedepropriete/standard`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        httpMethodOverride: true,
        params: {
          IDDNUPRO: idDnupro
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting download standard property survey');
          reject(false);
        }
      });
    });
  }

  /**
  * Retourne le lien de téléchargement du PDF du relevé de propriété de parcelle
  * @param idPar number l'identifiant de la parcelle
  * @returns Promise<any>
  */
  public downloadRPParcelle_(idPar: number): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/relevedepropriete/parcelle`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        httpMethodOverride: true,
        params: {
          ID_PAR: idPar
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting download property survey of parcels');
          reject(false);
        }
      });
    });
  }

  /**
  * Retourne le lien de téléchargement du PDF du relevé de propriété de parcelle_tiers
  * @param idPar number l'identifiant de la parcelle
  * @returns Promise<any>
  */
  public downloadRPParcelleTier_(idPar: number): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/relevedepropriete/parcelle_tiers`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        httpMethodOverride: true,
        params: {
          ID_PAR: idPar
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting download property survey of parcels third party');
          reject(false);
        }
      });
    });
  }

  /**
  * Retourne le lien de téléchargement du PDF du relevé de propriété des tiers
  * @param idDnupro string compte communal du propriétaire de la parcelle
  * @returns Promise<any>
  */
  public downloadRPTier_(idDnupro: number): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/relevedepropriete/tiers`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        httpMethodOverride: true,
        params: {
          IDDNUPRO: idDnupro
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting download property survey of third party');
          reject(false);
        }
      });
    });
  }


  /**
  * Retourne le lien de téléchargement du PDF de la fiche d'urbanisme
  * @param idPar number l'identifiant de la parcelle
  * @returns Promise<any>
  */
  public loadFicheUrbaByIdPar_(idPar: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/ficheurbanisme`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        httpMethodOverride: true,
        params: {
          ID_PAR: idPar
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting download of urbanism sheet');
          reject(false);
        }
      });
    });
  }

  /**
  * Retourne le lien de téléchargement du PDF de la fiche d'urbanisme
  * @returns Promise<any>
  */
  public downloadFicheUrba_(data: any): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/printficheurbanisme`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'POST',
        url: url,
        data: this.formreaderSrvc.transformToDataForm(data, true),
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data)) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting download of urbanism sheet');
          reject(false);
        }
      });
    });
  }

  /**
  * Retourne le lien de téléchargement de l'excel des locaux
  * @param idPar string Id des parcelles
  * @returns Promise<any>
  */
  public downloadExcelLocaux_(idPar: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/rapportscadastreparcelle/locaux`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        httpMethodOverride: true,
        params: {
          parcelles: idPar
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting download property survey of parcels third party');
          reject(false);
        }
      });
    });
  }

  /**
  * Retourne le lien de téléchargement de l'excel de toutes les informations
  * @param idPar string Id des parcelles
  * @returns Promise<any>
  */
  public downloadExcelToutesInfos_(idPar: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/rapportscadastreparcelle/toutesInfos`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        httpMethodOverride: true,
        params: {
          parcelles: idPar
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting download Excel all informations');
          reject(false);
        }
      });
    });
  }

  /**
  * Retourne le lien de téléchargement de l'excel des baties et non baties
  * @param idPar string Id des parcelles
  * @returns Promise<any>
  */
  public downloadExcelBatiesNonBatie_(idPar: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/rapportscadastreparcelle/batieNonBatie`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        httpMethodOverride: true,
        params: {
          parcelles: idPar
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting download property survey of parcels third party');
          reject(false);
        }
      });
    });
  }

  /**
  * Retourne le lien de téléchargement de l'excel des proprietaires
  * @param idPar string Id des parcelles
  * @returns Promise<any>
  */
  public downloadExcelProprietaires_(idPar: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/rapportscadastreparcelle/proprietaire`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        httpMethodOverride: true,
        params: {
          parcelles: idPar
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting download property survey of parcels third party');
          reject(false);
        }
      });
    });
  }

  /**
  * Retourne le lien de téléchargement de l'excel des emprises posPlu
  * @param idPar string Id des parcelles
  * @returns Promise<any>
  */
  public downloadExcelEmprises_(idPar: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/rapportscadastreparcelle/emprisePosPlu`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        httpMethodOverride: true,
        params: {
          parcelles: idPar
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting download property survey of parcels third party');
          reject(false);
        }
      });
    });
  }

  /**
  * Retourne le lien de téléchargement de l'excel des proprietaires des locaux
  * @param idPar string Id des parcelles
  * @returns Promise<any>
  */
  public downloadExcelProprietairesLocaux_(idPar: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/rapportscadastreparcelle/proprietaireLocaux`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        httpMethodOverride: true,
        params: {
          parcelles: idPar
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting download property survey of parcels third party');
          reject(false);
        }
      });
    });
  }

  /**
  * Retourne le lien de téléchargement de l'excel des subdivisions fiscales
  * @param idPar string Id des parcelles
  * @returns Promise<any>
  */
  public downloadExcelSubdFiscales_(idPar: string): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = `${environment.properties.app_api_url}/cadastre/rapportscadastreparcelle/subdivisionFiscale`;
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
        httpMethodOverride: true,
        params: {
          parcelles: idPar
        }
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data) &&
          resp.data.status === 1) {
          resolve(resp.data.data);
        } else {
          console.error('Error getting download property survey of parcels third party');
          reject(false);
        }
      });
    });
  }

  /**
   * Retourne la projection utilisée en dans la BDD
   * @returns
   */
  private getProjection(): string {
    let projection: string;
    if (!environment.properties.cadastre_database_projection) {
      projection = '2154';
    } else {
      projection = environment.properties.cadastre_database_projection.substring(5, 9);
    }
    return projection;
  }
}
