import { Component, Input, OnInit, OnDestroy } from '@angular/core';
import { Map } from 'ol';
import { MapToolDef } from '../../../map-tools.model';
import { VmapToolsCadastreMapService } from '../service/vmap-tools-cadastre-map.service';
import { VmapToolsCadastreService } from '../service/vmap-tools-cadastre.service';
import { Subscription } from 'rxjs';

@Component({
  selector: 'app-vmap-cadastre-tools-releve-de-propriete',
  templateUrl: './vmap-cadastre-tools-releve-de-propriete.component.html',
  styleUrls: ['./vmap-cadastre-tools-releve-de-propriete.component.scss']
})
export class VmapCadastreToolsReleveDeProprieteComponent implements OnInit, OnDestroy {

  @Input() olMap: Map;
  @Input() oTool: MapToolDef;

  public bToolOpen: boolean = false;

  public privilege: string;
  public aParcelle: any;

  public rapportTypes: Array<string> = ['Relevé de propriété de la parcelle tiers', 'Relevé de propriété de la parcelle', 'Relevé de propriété pour un tiers', 'Relevé de propriété standard'];
  public bSelectingRapport: string = null;

  // Subscription de l'élement courant
  private subscriptionCurrentElement: Subscription;

  /**
   * Constructeur
   */
  constructor(
    private vmapToolsCadastreMapService: VmapToolsCadastreMapService,
    private vmapToolsCadastreService: VmapToolsCadastreService) { }

  /**
   * Chargement du composant
   */
  ngOnInit(): void {
    // Ouverture de l'outil
    this.oTool.onOpen.subscribe(() => {
      this.bToolOpen = true;
      this.privilege = this.vmapToolsCadastreService.getPrivilege();
    });

    // Fermeture de l'outil
    this.oTool.onClose.subscribe(() => {
      this.bToolOpen = false;
      this.bSelectingRapport = null;
      this.unsubscribeCurrentElement();
      this.vmapToolsCadastreMapService.removeInteractions(this.olMap);
    });
  }

  /**
   * Destruction du composant
   */
  ngOnDestroy(): void {
    this.unsubscribeCurrentElement();
    this.bToolOpen = false;
  }

  /**
  * Permet de ne plus observer les changements de données géographique
  */
  private unsubscribeCurrentElement(): void {
    // désinscription de la fiche descriptive si subscription
    if (this.subscriptionCurrentElement) {
      this.subscriptionCurrentElement.unsubscribe();
    }
  }

  /**
  * Permet de faire une sélection cartographique
  * @returns Promise
  */
  private async locateCadastreOnMap(type: string): Promise<void> {
    try {
      this.aParcelle = await this.vmapToolsCadastreMapService.getGeom(this.olMap, 'point', type);
      // Ecoute les changements de coordonnées géographiques
      this.subscriptionCurrentElement = this.vmapToolsCadastreMapService.getCurrentElement().subscribe((next: any) => {
        this.aParcelle = next;
      });
    } catch (error) {
      console.error(error);
    }
  }

  /**
  * Télécharge le relevé de propriété standard
  * @returns Promise
  */
  public async getRPStandard(): Promise<void> {
    this.locateCadastreOnMap('releveDeProprieteStandard');
  }

  /**
  * Télécharge le relevé de propriété de la parcelle
  * @returns Promise
  */
  public async getRPParcelle(): Promise<void> {
    this.locateCadastreOnMap('releveDeProprieteParcelle');
  }

  /**
  * Télécharge le relevé de propriété de la parcelle tier
  * @returns Promise
  */
  public async getParcelleTier(): Promise<void> {
    this.locateCadastreOnMap('releveDeProprieteParcelleTier');
  }

  /**
  * Télécharge le relevé de propriété d'un tier
  * @returns Promise
  */
  public async getRPTier(): Promise<void> {
    this.locateCadastreOnMap('releveDeProprieteTier');
  }

  /**
  * Sélectionne le bon relevé de propriété
  * @param rapport string type du relevé de propriété
  */
  public selectedRapport(rapport: string): void {
    this.bSelectingRapport = rapport;
    this.vmapToolsCadastreMapService.removeInteractions(this.olMap);

    if (rapport === 'Relevé de propriété de la parcelle tiers') {
      this.getParcelleTier();

    } else if (rapport === 'Relevé de propriété de la parcelle') {
      this.getRPParcelle();

    } else if (rapport === 'Relevé de propriété pour un tiers') {
      this.getRPTier();

    } else if (rapport === 'Relevé de propriété standard') {
      this.getRPStandard();
    }
  }
}
