import { Component, Input, OnDestroy, OnInit } from '@angular/core';
import { Map, Overlay } from 'ol';
import { MapToolDef } from '../../map-tools.model';
import { MapCoreManagerService } from 'src/app/vitis/map-core/services/map-core-manager/map-core-manager.service';
import { unByKey } from 'ol/Observable';
import { transform } from 'node_modules/ol/proj.js';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import { Subscription } from 'rxjs';
import { EventsKey } from 'ol/events';

declare let $: any;

@Component({
  selector: 'app-vmap-tools-compare-map',
  templateUrl: './vmap-tools-compare-map.component.html',
  styleUrls: ['./vmap-tools-compare-map.component.scss']
})
export class VmapToolsCompareMapComponent implements OnInit, OnDestroy {

  @Input() olMap: Map;
  @Input() oTool: MapToolDef;
  // Id de la div OL3
  @Input() targetCompareId: string;
  @Input() tileSize: Array<number>;

  public bToolOpen;

  public mapId: number;
  // Carte de comparaison en cours
  public mapCompareId: number | null = null;
  private olMapCompare: Map;
  private targetOverlay;

  // Subscribe mouse moves
  private keyChangeMouseEnterMap;
  private keyChangeMouseLeaveMap;
  private keyChangeMouseEnterMapCompare;
  private keyChangeMouseLeaveMapCompare;

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];
  private mapChangeSubcriptions: EventsKey[] = [];
  private mouseMovesMap: EventsKey[] = [];
  private mouseMovesMapCompare: EventsKey[] = [];

  /**
   * Constructeur
   */
  constructor(
    private mapCoreManagerService: MapCoreManagerService,
    private mapUtils: MapCoreUtilsService,
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit(): void {
    this.mapId = this.olMap.get('map_id');
    // Ouverture de l'outil
    this.subscriptions.push(this.oTool.onOpen.subscribe(() => {
      this.bToolOpen = true;
    }));

    // Fermeture de l'outil
    this.subscriptions.push(this.oTool.onClose.subscribe(() => {
      this.bToolOpen = false;
    }));

    this.olMap.on('change', (event: any) => {
      setTimeout(() => {
        if (this.mapId != event.target.get('map_id')) {
          this.mapId = event.target.get('map_id');
          if (this.olMapCompare || this.mapCompareId || this.olMap.get('map_compare')) {
            this.selectedCompareMap(this.mapCompareId);
          }
        }
      });
    });
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.resetKeyEvent();
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
    this.mapChangeSubcriptions.forEach(key => unByKey(key));
    this.mouseMovesMap.forEach(key => unByKey(key));
    this.mouseMovesMapCompare.forEach(key => unByKey(key));
    this.olMap.unset('map_compare');
  }

  /**
   * Change le mode de comparaison
   * @param compareMode true | false
   */
  private onChangeCompareMode(compareMode: boolean) {
    const mapContainer = document.getElementsByClassName('mapContainer');
    const mapCompare = document.getElementsByClassName('mapCompare');
    const controls = document.getElementsByClassName('controls-container');
    const controlsTopCenter = controls[0].getElementsByClassName('top-center');
    const controlsBottomRight = controls[0].getElementsByClassName('bottom-right');

    // Si le mode de comparaison est activé
    if (compareMode === true) {
      mapContainer[0]['style']['display'] = 'flex';
      mapCompare[0]['style']['display'] = 'block';
      controlsTopCenter[0].classList.add('map-compare');
      controlsBottomRight[0].classList.add('map-compare');
    } else { // Si le mode de comparaison est désactivé
      mapContainer[0]['style']['display'] = 'block';
      mapCompare[0]['style']['display'] = 'none';
      controlsTopCenter[0].classList.remove('map-compare');
      controlsBottomRight[0].classList.remove('map-compare');
      document.getElementsByClassName('mapCompare')[0].innerHTML = '';

      // Unset la variable de la carte
      if (this.olMap.get('map_compare')) {
        this.olMap.unset('map_compare');
      }

      // Supprime le pointeur carto
      if (this.targetOverlay) {
        this.targetOverlay.setPosition(null);
        this.olMap.removeOverlay(this.targetOverlay);
        this.olMapCompare.removeOverlay(this.targetOverlay);
        this.targetOverlay = null;
      }

      this.resetKeyEvent();
      this.mapChangeSubcriptions.forEach(key => unByKey(key));
      this.mouseMovesMap.forEach(key => unByKey(key));
      this.mouseMovesMapCompare.forEach(key => unByKey(key));
      this.olMapCompare = null;
      this.mapCompareId = null;
    }
  }

  /**
   * Détruit les events des cartes
   */
  private resetKeyEvent() {
    if (this.keyChangeMouseEnterMap) {
      $(this.olMap.getTargetElement()).off('mouseenter');
    }
    if (this.keyChangeMouseLeaveMap) {
      $(this.olMap.getTargetElement()).off('mouseleave');
    }
    if (this.keyChangeMouseEnterMapCompare) {
      $(this.olMapCompare.getTargetElement()).off('mouseenter');
    }
    if (this.keyChangeMouseLeaveMapCompare) {
      $(this.olMapCompare.getTargetElement()).off('mouseleave');
    }
  }

  /**
   * Sélection de la carte de comparaison
   * @param mapId
   */
  public selectedCompareMap(mapCompareId) {
    this.mapCompareId = this.mapCompareId == mapCompareId ? null : mapCompareId;
    if (this.mapCompareId) {
      this.onChangeCompareMode(true);

      this.initCompareMapId(this.mapCompareId).then(() => {
        this.olMap.set('map_compare', this.olMapCompare);
        this.olMap.updateSize();
        this.olMap.changed();

        // Synchronisation de l'extent des views
        this.keyChangeMouseEnterMap = $(this.olMap.getTargetElement()).on('mouseenter', (e) => {
          this.mouseMovesMap.push(this.olMap.on('moveend', (event: any) => {
            this.setMapCompareCenter(this.olMap.getView());
            this.setMapCompareScale();
          }));
        });

        this.keyChangeMouseLeaveMap = $(this.olMap.getTargetElement()).on('mouseleave', (e) => {
          this.mouseMovesMap.forEach(key => unByKey(key));
        });

        this.keyChangeMouseEnterMapCompare = $(this.olMapCompare.getTargetElement()).on('mouseenter', (e) => {
          this.mouseMovesMapCompare.push(this.olMapCompare.on('moveend', (event: any) => {
            this.setMapCenter(this.olMapCompare.getView());
            this.setMapScale();
          }));
        });

        this.keyChangeMouseLeaveMapCompare = $(this.olMapCompare.getTargetElement()).on('mouseleave', (e) => {
          this.mouseMovesMapCompare.forEach(key => unByKey(key));
        });

        // Synchronisation du déplacement sur les carte
        this.mapChangeSubcriptions.push(this.olMap.on('pointerdrag', (e) => {
          this.setMapCompareCenter(this.olMap.getView());
        }));

        this.mapChangeSubcriptions.push(this.olMapCompare.on('pointerdrag', (e) => {
          this.setMapCenter(this.olMapCompare.getView());
        }));

        // Synchronisation des pointers sur les cartes
        this.mapChangeSubcriptions.push(this.olMap.on('pointermove', (e) => {
          if (document.getElementsByClassName('over-map-compare').length > 0) {
            this.targetOverlay.setPosition(null);
            this.olMap.removeOverlay(this.targetOverlay);
            this.targetOverlay = null;
          }
          if (!this.targetOverlay) {
            const element = document.createElement('div');
            element.className = 'ol-target';
            this.targetOverlay = new Overlay({
              element: element,
              position: [0, 0],
              positioning: 'center-center',
            });
            this.olMapCompare.addOverlay(this.targetOverlay);
            if (element.parentElement) {
              element.parentElement.classList.add('ol-target-overlay');
              element.parentElement.classList.add('over-map');
            }
          }
          const coordinate = transform(e.coordinate, this.olMap.getView().getProjection(), this.olMapCompare.getView().getProjection());
          this.targetOverlay.setPosition(coordinate);
        }));

        this.mapChangeSubcriptions.push(this.olMapCompare.on('pointermove', (e) => {
          if (document.getElementsByClassName('over-map').length > 0) {
            this.targetOverlay.setPosition(null);
            this.olMapCompare.removeOverlay(this.targetOverlay);
            this.targetOverlay = null;
          }
          if (!this.targetOverlay) {
            const element = document.createElement('div');
            element.className = 'ol-target';
            this.targetOverlay = new Overlay({
              element: element,
              position: [0, 0],
              positioning: 'center-center',
            });
            this.olMap.addOverlay(this.targetOverlay);
            if (element.parentElement) {
              element.parentElement.classList.add('ol-target-overlay');
              element.parentElement.classList.add('over-map-compare');
            }
          }
          const coordinate = transform(e.coordinate, this.olMapCompare.getView().getProjection(), this.olMap.getView().getProjection());
          this.targetOverlay.setPosition(coordinate);
        }));
      });
    } else {
      this.closeCompareMode();
    }
  }

  /**
   * Stoppe le mode comparaison et ferme le bandeau
   */
  public closeCompareMode() {
    this.onChangeCompareMode(false);
    this.olMap.updateSize();
    this.olMap.changed();
  }

  /**
   * Met à jour le centre de la carte de comparaison
   * @param olMapView
   */
  private setMapCompareCenter(olMapView) {
    const center = olMapView.getCenter();
    const centerMapCompare = transform(center, olMapView.getProjection(), this.olMapCompare.getView().getProjection());
    this.olMapCompare.getView().setCenter(centerMapCompare);
  }

  /**
   * Met à jour le centre de la carte principale
   * @param olMapCompareView
   */
  private setMapCenter(olMapCompareView) {
    const center = olMapCompareView.getCenter();
    const centerMap = transform(center, olMapCompareView.getProjection(), this.olMap.getView().getProjection());
    this.olMap.getView().setCenter(centerMap);
  }

  /**
   * Met à jour l'étendue de la carte de comparaison
   */
  private setMapCompareScale() {
    this.mapUtils.setScale(this.olMapCompare, this.mapUtils.getScale(this.olMap));
  }

  /**
   * Met à jour l'étendue de la carte principale
   */
  private setMapScale() {
    this.mapUtils.setScale(this.olMap, this.mapUtils.getScale(this.olMapCompare));
  }

  /**
   * Initialise la carte de comparaison
   * @param mapId
   * @returns
   */
  private initCompareMapId(mapId): Promise<void> {
    return new Promise((resolve, reject) => {
      this.subscriptions.push(this.mapCoreManagerService.initOlMap(mapId, this.targetCompareId, this.tileSize, this.olMapCompare).subscribe((mapCompare: any) => {

        //  // Enregistre l'objet Map
        this.olMapCompare = mapCompare;

        // Enregistre l'identifiant de carte utilisé
        this.olMapCompare.set('map_id', mapId);

        // Enregistre l'identifiant de l'élement utilisé
        this.olMapCompare.set('target_id', this.targetCompareId);

        // Fit l'étendue
        this.setMapCompareScale();
        this.setMapCompareCenter(this.olMap.getView());

        // Enregistre le nom de la carte utilisé
        this.olMapCompare.set('name', this.olMapCompare.get('map_definition')['name']);

        this.olMapCompare.changed();

        resolve();
      }));
    });
  }

}
