import { Component, Input, OnDestroy, OnInit, Injectable } from '@angular/core';
import { Map } from 'ol';
import { Subscription } from 'rxjs';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { MapCoreManagerService } from 'src/app/vitis/map-core/services/map-core-manager/map-core-manager.service';
import { VmapLayer } from 'src/app/vitis/map-core/services/map-models.model';
import { MapToolDef } from 'src/modules/vmap/components/carto/components/vmap-tools/map-tools.model';
import { VmapToolsService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools.service';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import moment from 'moment-timezone';
import {NgbDateStruct, NgbDateAdapter,NgbDateParserFormatter} from '@ng-bootstrap/ng-bootstrap';

/**
 * Ce service gère la façon dont la date est représentée dans les scripts (ngModel).
 */
@Injectable()
export class CustomAdapter extends NgbDateAdapter<string> {
  readonly DELIMITER = '-';
  /**
   * Constructeur
   * @param {VmapToolsFilterComponent} vmapToolsFilterComponent.
   */
  constructor(private vmapToolsFilterComponent: VmapToolsFilterComponent) {
    super();
  }

  /**
   * Converts a NgbDateStruct value into NgbDateStruct value
   * @param {string|null} value.
   * @return {NgbDateStruct|null}
   */
  fromModel(value: string | null): NgbDateStruct | null {
    if (value) {
      const date = moment(new Date(parseInt(value + '000', 10)));
      return {
        day: date.date(),
        month: date.month() + 1,
        year: date.year(),
      };
    }
    return null;
  }

  /**
   * Converts a NgbDateStruct value into NgbDateStruct value
   * @param {NgbDateStruct|null} date.
   * @return {string|null}
   */
  toModel(date: NgbDateStruct | null): string | null {
    return date ? moment().date(date.day).month(date.month - 1).year(date.year).format('X') : null;
  }
}
/**
 * Ce service gère le formatage de la date sélectionnée dans le champ <input>.
 */
@Injectable()
export class CustomDateParserFormatter extends NgbDateParserFormatter {
  readonly DELIMITER = '-';
  /**
   * Constructeur
   * @param {VmapToolsFilterComponent} vmapToolsFilterComponent.
   */
  constructor(private vmapToolsFilterComponent: VmapToolsFilterComponent) {
    super();
  }

  /**
   * Parses the given `string` to an `NgbDateStruct`
   * @param {string} value.
   * @return {NgbDateStruct|null}
   */
  parse(value: string): NgbDateStruct | null {
    return null;
  }

  /**
   * Formats the given `NgbDateStruct` to a `string`.
   * @param {NgbDateStruct|null} date.
   * @return {string}
   */
  format(date: NgbDateStruct | null): string {
    return date ? moment().date(date.day).month(date.month - 1).year(date.year).format(this.vmapToolsFilterComponent.sDateFormat) : null;
  }
}

class LayerFilterOperator {
  column: string;
  value: string;
  compare_operator: string;
  sql_operator: string;
  oSelectedColumn?: LayerColumn;

  /**
   * Constructeur
   */
  constructor(obj: any = {}) {
    this.column = obj.column || null;
    this.value = obj.value || null;
    this.compare_operator = obj.compare_operator || null;
    this.sql_operator = obj.sql_operator || null;
  }
}

interface LayerColumn {
  column: string;
  filter: boolean;
  index: string;
  name: string;
  requestor: boolean;
  type?: string;
  filter_type?: string; // Suite à une évolution le type est désormais contenu dans filter_type, pour des raisons de retrocompatibilité on continue de faire fontionner type
  requestor_type?: string;
}

@Component({
  selector: 'app-vmap-tools-filter',
  templateUrl: './vmap-tools-filter.component.html',
  styleUrls: ['./vmap-tools-filter.component.scss'],
  providers: [
    { provide: NgbDateAdapter, useClass: CustomAdapter },
    { provide: NgbDateParserFormatter, useClass: CustomDateParserFormatter },
  ]
})
export class VmapToolsFilterComponent implements OnInit, OnDestroy {

  @Input() olMap: Map;
  @Input() oTool: MapToolDef;

  public aAvailableLayers: Array<VmapLayer> = [];
  public oSelectedLayer: VmapLayer;
  public bToolOpen: boolean;
  public aActiveFilter: Array<LayerFilterOperator> = [new LayerFilterOperator()];
  public aLayerColumns: Array<LayerColumn> | null;
  public isOpenDropDown: boolean = false;

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];

  // Timezone de l'utilisateur
  private sUserTimezone: string;

  // Format d'affichage de la date
  public sDateFormat: string;

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private mapManagerService: MapCoreManagerService,
    private vmapToolsService: VmapToolsService,
    private authService: AuthService
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit(): void {

    // Ouverture de l'outil
    this.subscriptions.push(this.oTool.onOpen.subscribe(() => {
      this.initTool();
      this.bToolOpen = true;
    }));

    // Fermeture de l'outil
    this.subscriptions.push(this.oTool.onClose.subscribe(() => {
      this.bToolOpen = false;
    }));

    // Timezone de l'utilisateur
    const oUser = this.authService.getUserInfos();
    this.sUserTimezone = oUser.timezone_id;
    if (this.sUserTimezone) {
      moment.tz.setDefault(this.sUserTimezone);
    }

    // Format d'affichage de la date
    this.sDateFormat = this.getDateFormat();
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.unsetGreenCircle_();
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
  }

  /**
   * Initialise l'outil
   */
  private initTool() {
    this.aAvailableLayers = this.getAvailableLayers_(this.olMap);
    const layerExists = this.aAvailableLayers.find(layer => {
      if (this.oSelectedLayer &&
        layer.layer_id === this.oSelectedLayer.layer_id) {
        return true;
      } else {
        return false;
      }
    });
    if (!layerExists) {
      this.oSelectedLayer = null;
      this.aActiveFilter = [new LayerFilterOperator()];
    }
  }

  /**
   * Appelé au changement de couche
   * @param oLayer
   */
  public onLayerChange(oLayer: VmapLayer) {
    this.aLayerColumns = this.getAvailableColumns_(this.oSelectedLayer);
    this.aActiveFilter = [new LayerFilterOperator()];
  }

  /**
   * Appelé à la sélection d'une colonne
   * @param sColumn
   */
  public onSelectColumn(index: number, sColumn: string) {
    if (this.vitisCheck.isArray(this.aLayerColumns) && this.aLayerColumns.length > 0) {
      const oSelectedColumn = this.aLayerColumns.find(col => {
        return col.column === sColumn;
      });
      this.aActiveFilter[index].oSelectedColumn = oSelectedColumn;
    }
  }

  /**
   * Affiche le bouton permettant d'ajouter une ligne
   * @param index
   */
  public openAddExpression(index: number) {
    if (this.aActiveFilter.length <= 1) {
      document.getElementById(`vmap_tool_filter_dropdown_${index}`).classList.toggle('show');
      this.isOpenDropDown = !this.isOpenDropDown;
    } else {
      this.addExpression(this.aActiveFilter[0].sql_operator, index);
    }
  }

  /**
   * Ajoute une ligne au filtre
   * @param sql_operator
   * @param index
   */
  public addExpression(sql_operator: string, index: number) {
    this.aActiveFilter[index].sql_operator = sql_operator;
    this.aActiveFilter = [
      ...this.aActiveFilter,
      new LayerFilterOperator()
    ];
    const dropElem = document.getElementById(`vmap_tool_filter_dropdown_${index}`);
    if (dropElem) {
      dropElem.classList.remove('show');
    }
  }

  /**
   * Supprime une ligne
   * @param index
   */
  public deleteExpression(index: number) {
    if (this.aActiveFilter.length === 1) {
      this.aActiveFilter = [new LayerFilterOperator()];
    } else {
      if (index === this.aActiveFilter.length - 1) {
        this.aActiveFilter[index - 1].sql_operator = null;
      }
      this.aActiveFilter = [...this.aActiveFilter.slice(0, index), ...this.aActiveFilter.slice(Number(index) + 1)];
    }
  }

  /**
   * Retourne les opérateurs en fonction du type de colonne
   * @param index
   * @returns
   */
  public getAvailableOperators(index: number): Array<string> {
    const aReturn = [];
    if (this.aActiveFilter &&
      this.aActiveFilter[index] &&
      this.aActiveFilter[index].oSelectedColumn &&
      (this.aActiveFilter[index].oSelectedColumn.type ||
        this.aActiveFilter[index].oSelectedColumn.filter_type)) {

      // Suite à une évolution le type est désormais contenu dans filter_type, pour des raisons de retrocompatibilité on continue de faire fontionner type
      const type = this.aActiveFilter[index].oSelectedColumn.filter_type ?
        this.aActiveFilter[index].oSelectedColumn.filter_type : this.aActiveFilter[index].oSelectedColumn.type;

      switch (type) {
        case 'string':
          return ['=', '~', '!=', 'IS NOT NULL', 'IS NULL'];
          break;
        case 'number':
          return ['=', '!=', '<', '<=', '>', '>=', 'IS NOT NULL', 'IS NULL'];
          break;
        case 'bool':
          return ['IS TRUE', 'IS FALSE', 'IS NOT NULL', 'IS NULL'];
          break;
        case 'date':
          return ['=', '!=', '<', '<=', '>', '>='];
          break;
        default:
          return ['=', '!=', 'IS NOT NULL', 'IS NULL'];
          break;
      }
    }

    return aReturn;
  }

  /**
   * Lance la requête de filtre
   */
  public sendFilter() {
    const oAPIFilter = this.getAPIFilter_(this.aActiveFilter);
    this.vmapToolsService.refreshMap(this.olMap, false, null, oAPIFilter);

    // Ajoute le deuxième icône sur le bouton "Filtrer une couche".
    if (this.vitisCheck.isArray(oAPIFilter) && oAPIFilter.length > 0) {
      this.setGreenCircle_();
    } else {
      this.unsetGreenCircle_();
    }
  }

  /**
   * Retourne les couches disponibles pour filtrage
   * @returns
   */
  private getAvailableLayers_(olMap: Map): Array<VmapLayer> {
    const aAvailableLayers = [];
    const aLayers = this.mapManagerService.getLayersArray(olMap);

    for (let i = 0; i < aLayers.length; i++) {
      const oLayer = aLayers[i];
      if (oLayer.type === 'postgres' &&
        oLayer.is_filter === true) {
        const columns = this.getAvailableColumns_(oLayer);
        if (this.vitisCheck.isArray(columns) && columns.length > 0) {
          aAvailableLayers.push(oLayer);
        }
      }
    }

    return aAvailableLayers.sort((a, b) => a.name > b.name && 1 || -1);
  }

  /**
   * Retourne la définition des colonnes de la couche en se basant sur json_columns
   * @param oLayer
   * @returns
   */
  private getAvailableColumns_(oLayer: VmapLayer): Array<LayerColumn> | null {
    if (this.vitisCheck.isString(oLayer.json_columns)) {
      const columns = JSON.parse(oLayer.json_columns);
      if (this.vitisCheck.isArray(columns) && columns.length > 0) {
        return columns.filter((column: any) => {
          return column.filter === true;
        });
      }
    } else {
      return null;
    }
  }

  /**
   * Transforme le filtre en filtre compréhensible par l'API
   * @param aFilter
   * @returns
   */
  private getAPIFilter_(aFilter: Array<LayerFilterOperator>) {
    if (this.vitisCheck.isArray(aFilter) && aFilter.length > 0) {

      const aAPIFilter = [];
      const oFilter = {
        relation: aFilter[0].sql_operator || 'AND',
        operators: [],
      };

      for (let i = 0; i < aFilter.length; i++) {
        const operator = aFilter[i];

        if (operator.column &&
          operator.column.length > 0 &&
          operator.compare_operator &&
          operator.compare_operator.length > 0) {

          if (operator.value === null) {
            operator.value = '';
          }

          // Filtre de type "date" => UNIQUEMENT SI OPERATEUR = "=" ?
          if (this.vitisCheck.isDefAndNotNull(operator.oSelectedColumn) &&
          operator.oSelectedColumn.filter_type == 'date' &&
          operator.compare_operator == '=') {
            // Date de début (00:00:00).
            const iTimeStamp: number = moment(new Date(parseInt(operator.value + '000', 10))).hour(0).minute(0).second(0).utc().unix();
            oFilter.operators.push({
              column: operator.column,
              compare_operator: '>=',
              value: iTimeStamp
            });
            // Date de fin (23:59:00)
            oFilter.operators.push({
              column: operator.column,
              compare_operator: '<=',
              value: iTimeStamp + 86399
            });
          } else if (operator.compare_operator === '~') {
            oFilter.operators.push({
              column: operator.column,
              compare_operator: 'LIKE',
              value: `%${operator.value}%`,
              compare_operator_options: {
                column_type: operator.oSelectedColumn.type,
                case_insensitive: true,
              }
            });
          } else if (operator.compare_operator === 'IS TRUE' || operator.compare_operator === 'IS FALSE') {
            oFilter.operators.push({
              column: operator.column,
              compare_operator: '=',
              value: operator.compare_operator === 'IS TRUE' ? true : false,
              compare_operator_options: {
                column_type: operator.oSelectedColumn.type,
              }
            });
          } else {
            oFilter.operators.push({
              column: operator.column,
              compare_operator: operator.compare_operator,
              value: operator.value,
              compare_operator_options: {
                column_type: operator.oSelectedColumn.type,
              }
            });
          }
        }
      }

      if (oFilter.operators.length > 0) {
        aAPIFilter.push({
          layer_id: this.oSelectedLayer.layer_id,
          filter: oFilter,
        });
      }

      return aAPIFilter;
    }
  }

  /**
   * Ajoute un rond vert à côté de l'icône du bouton "Filtrer une couche".
   */
  private setGreenCircle_() {
    this.oTool.secondIcon = 'green-circle';
  }

  /**
   * Supprime le rond vert à côté de l'icône du bouton "Filtrer une couche".
   */
  private unsetGreenCircle_() {
    delete this.oTool.secondIcon;
  }

  /**
   * Retourne le format de date à afficher
   * @returns string
   */
  public getDateFormat(): string {
    // Valeur par défaut
    let sDateFormat = 'DD/MM/YYYY';

    // Valeur de l'utilisateur en cours
    const oUser = this.authService.getUserInfos();
    if (this.vitisCheck.isDefAndNotNull(oUser)) {
      if (this.vitisCheck.isDefAndNotNull(oUser.formatdate_id)) {
        sDateFormat = oUser.formatdate_id;
      }
    }
    return sDateFormat;
  }
}
