import { Component, OnInit, Input, OnDestroy, ViewChild } from '@angular/core';
import { environment } from 'src/environments/environment';
import { VitisRequest, VitisClone, VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { MapToolDef } from '../../map-tools.model';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import { Feature } from 'ol';
import Geometry from 'ol/geom/Geometry';
import Point from 'ol/geom/Point';
import { transform } from 'ol/proj';
import { MapCorePopupService, Popup } from 'src/app/vitis/map-core/services/map-core-popup/map-core-popup.service';
import { ActivatedRoute } from '@angular/router';
import { MapCoreManagerService } from 'src/app/vitis/map-core/services/map-core-manager/map-core-manager.service';
import { VmapToolsService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools.service';
import { Subscription } from 'rxjs';
import { EventsKey } from 'ol/events';
import { unByKey } from 'ol/Observable';
import { VmapToolsFormComponent } from 'src/app/modules/vmap/carto/components/vmap-tools/components/vmap-tools-search/vmap-tools-form/vmap-tools-form.component';
import { VmapFormService } from 'src/modules/vmap/components/carto/services/vmap-form/vmap-form.service';
import { IsSmallScreenService } from 'src/app/vitis/core/services/utils/is-small-screen/is-small-screen.service';

@Component({
  selector: 'app-vmap-tools-search',
  templateUrl: './vmap-tools-search.component.html',
  styleUrls: ['./vmap-tools-search.component.scss']
})
export class VmapToolsSearchComponent implements OnInit, OnDestroy {

  @Input() olMap;
  @Input() oTool: MapToolDef;

  /**
   * Enfant de type VmapToolsFormComponent
   */
  @ViewChild(VmapToolsFormComponent, { static: false, read: VmapToolsFormComponent }) set oVmapToolsForm(oVmapToolsForm: VmapToolsFormComponent) {
    if (this.vitisCheck.isDefAndNotNull(oVmapToolsForm)) {
      this.oFormReader = oVmapToolsForm.oFormReader;
    }
  }

  public isMapLayersSearch: boolean;
  public mapLayers: Array<any>; // Permet uniquement de savoir si la carte a changé
  public mapLayersLocalisable: Array<any>; // Liste réellement utilisée
  public layer_id: any;
  private map_id: any;
  private selectedLayer: any;
  private popup: Popup;

  public sDefinition: any;
  public sFormName = 'search';
  public sDatasForm: any;

  private sCurrDatasForm: any;

  public results: any;
  public resultsEmpty = false;
  public resultsDisplayed = 6;

  private currentLimit: number;

  public optionsApi = [{
    url: 'https://api-adresse.data.gouv.fr/search/?q=[search]&limit=[limit]',
    value: 'BAN',
    label: 'COMPONENT_MAP_SEARCH_BAN',
    enabled: environment.properties.vmap_search_ban_enabled
  },
  {
    url: 'https://nominatim.openstreetmap.org/search?format=json&addressdetails=1&limit=[limit]&extratags=1&namedetails=1&polygon_geojson=1&countrycodes=fr&q=[search]&vitis_version=29192',
    value: 'OSM',
    label: 'COMPONENT_MAP_SEARCH_OSM',
    enabled: environment.properties.vmap_search_osm_enabled
  }];
  public inputSearch: any;

  private feature: Feature<Geometry>;

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];
  private mapSubscriptions: EventsKey[] = [];

  public isSmallScreen;

  public oFormReader: any;

  /**
   * Constructeur
   */
  constructor(
    private mapUtils: MapCoreUtilsService,
    private vitisRequest: VitisRequest,
    private vitisClone: VitisClone,
    public vitisCheck: VitisCheck,
    private mapPopup: MapCorePopupService,
    private route: ActivatedRoute,
    private mapCoreManagerService: MapCoreManagerService,
    private vmapToolsService: VmapToolsService,
    private isSmallScreenService: IsSmallScreenService,
    private mapToolsService: VmapToolsService,
    private mapFormService: VmapFormService,
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit(): void {

    this.isSmallScreen = this.isSmallScreenService.getIsSmallScreen();
    this.subscriptions.push(this.isSmallScreenService.isSmallScreenObservable.subscribe(isSmallScreen => {
      this.isSmallScreen = isSmallScreen;
    }));

    if (this.isSmallScreen) {
      this.initLayers();
      this.mapSubscriptions.push(this.olMap.on('change', () => {
        this.resetLocalisation();
        this.initLayers();
      }));
    }

    this.subscriptions.push(this.route.data.subscribe(mode => {
      this.subscriptions.push(this.route.queryParams.subscribe(oParams => {
        this.map_id = oParams.map_id;
      }));
    }));

    if (this.oTool) {
      // Ouverture de l'outil
      this.subscriptions.push(this.oTool.onOpen.subscribe(() => {
        this.initLayers();
      }));

      // Fermeture de l'outil
      this.subscriptions.push(this.oTool.onClose.subscribe(() => {
        this.resetLocalisation();
      }));
    }
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
    this.mapSubscriptions.forEach(mapSubscription => unByKey(mapSubscription));
    this.resetLocalisation();
  }

  /**
   * Initalise les couches disponible à la localisation
   */
  private initLayers(): void {

    const oMapDef = this.olMap.get('map_definition');
    const mapLayers = [];
    const aLayers = this.mapCoreManagerService.getLayersList(this.olMap);
    for (let i = 0; i < aLayers.length; i++) {
      const oLayer = aLayers[i];
      mapLayers.push({
        layer_id: oLayer.layer_id,
        visibility: oLayer.visibility,
        is_location: oLayer.is_location,
      });
    }

    // Couche par défaut
    let mapDefaultLayer: any = 'OSM';
    if (oMapDef.options &&
      oMapDef.options.default_layer_location_id) {
      mapDefaultLayer = oMapDef.options.default_layer_location_id;
    }

    // Liste des couches est différente ?
    let bLayersChanged = false;
    try {
      bLayersChanged = JSON.stringify(mapLayers) !== JSON.stringify(this.mapLayers);
    } catch (error) { }

    if (bLayersChanged) {
      this.mapLayers = mapLayers;
      this.mapLayersLocalisable = aLayers.filter(layer => layer.is_location === true);
      this.layer_id = mapDefaultLayer;
      this.onChangeLayer(this.layer_id);
    }
  }

  /**
   * Fonction appelé quand on sélectionne une couche
   * @param layer_id
   */
  public onChangeLayer(layer_id: any): void {
    this.results = [];
    this.resultsEmpty = false;
    this.resetLocalisation();
    if (isNaN(Number(layer_id))) {
      this.sDefinition = {};
      this.sDatasForm = {};
      this.isMapLayersSearch = false;
      this.selectedLayer = null;
    } else {
      this.isMapLayersSearch = true;
      const aLayers = this.mapCoreManagerService.getLayersList(this.olMap);
      this.selectedLayer = aLayers.find(oLayer => {
        if (oLayer.layer_id && Number(oLayer.layer_id) === Number(layer_id)) {
          return true;
        }
      });

      if (this.selectedLayer && this.selectedLayer.published_form_json_def) {
        this.sDefinition = this.vitisClone.clone(this.selectedLayer.published_form_json_def);
        this.sDefinition.search.eventContainer = this;
        this.sDefinition.search.event = 'sendSearch()';

        // Le bouton est caché mais obligatoire pour que l'appui sur la touche entrée fonctionne
        this.sDefinition.search.rows.push({
          fields: [{
            type: 'button',
            class: 'hidden btn-ungroup btn-group-sm',
            nb_cols: 12,
            buttons: [{
              type: 'submit',
              name: 'form_submit',
              label: 'submit',
              class: 'hidden btn-primary'
            }]
          }]
        });
        this.sFormName = this.vitisClone.clone('search');
        this.sCurrDatasForm = {};
        this.sDatasForm = this.vitisClone.clone({});

        // Contenu CSS
        if (this.vitisCheck.isDefAndNotNull(this.selectedLayer.published_form_css_def)) {
          setTimeout(() => { // Attend que l'HTML soit chargé
            this.mapFormService.loadCss(this.selectedLayer.published_form_css_def);
          });
        }
      }
    }
  }

  /**
   * Event retourné quand le formulaire est en mode édition
   * @param updadedForm
   */
  public getFormChanges(updadedForm): void {
    this.sCurrDatasForm = updadedForm;
  }

  /**
   * Envoie la requête pour chercher l'object avec ce qui est entré en paramètre
   */
  public sendSearch(limit: number = 10): void {
    let resultPromise;
    this.resultsEmpty = false;
    this.resetLocalisation();
    if (isNaN(this.layer_id)) {
      if (this.inputSearch) {
        const optionApi = this.optionsApi.find(option => (option.value === this.layer_id && option.enabled !== false));
        if (optionApi) {
          const url = optionApi.url.replace('[search]', this.inputSearch).replace('[limit]', String(limit));
          resultPromise = this.vitisRequest.ajaxRequestPromise({
            method: 'GET',
            url
          });
        } else {
          console.error('No search API selected');
        }
      }
    } else {
      const operators = [];
      Object.keys(this.sCurrDatasForm).forEach(keys => {
        if (this.sCurrDatasForm[keys]) {
          const correctSyntax = this.likeSQLSyntax(this.sCurrDatasForm[keys], this.oFormReader.getFormField(keys).type);
          operators.push({
            'column': keys,
            'compare_operator': correctSyntax.compare_operator,
            'value': correctSyntax.value,
            'compare_operator_options': { 'case_insensitive': correctSyntax.case_insensitive }
          });
        }
      });

      const sOrderColumn = this.selectedLayer.select_field || this.selectedLayer.idfield;
      resultPromise = this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        httpMethodOverride: true,
        url: `${environment.properties.app_api_url}/vmap/layers/${this.layer_id}/query`,
        data: {
          'filter': {
            'relation': 'AND',
            operators
          },
          'order_by': sOrderColumn, // Select_field de la couche
          'limit': this.vitisCheck.isNumber(limit) ? limit : 10,
          'result_srid': this.olMap.getView().getProjection().getCode().substring(5),
        }
      });
    }
    resultPromise.then(resp => {
      if (resp && resp.data) {
        if (resp.data.data) {
          this.results = resp.data.data;
        } else if (resp.data.features) {
          this.results = resp.data.features;
        } else {
          this.results = resp.data;
        }
        if (!this.results.length) {
          this.resultsEmpty = true;
        }
        if (this.results.length === 1) {
          this.showLocalisation(this.results[0]);
        }
        this.resultsDisplayed = limit;
        this.currentLimit = limit;
      }
    });
  }

  /**
   * Retourne le nom à afficher pour la popup et lors du retour des résultats de recherche
   * @param data
   * @returns
   */
  public getCurrentName(data): string {
    if (!isNaN(this.layer_id)) {
      return data[this.selectedLayer.select_field || this.selectedLayer.idfield];
    } else if (this.layer_id === 'OSM') {
      return data.display_name;
    } else if (this.layer_id === 'BAN') {
      return data.properties.label;
    }
  }

  /**
   * Afffiche la popup et zoom sur le résultat cliqué
   * @param data
   */
  public showLocalisation(data): void {
    let geometry;
    const displayedName = this.getCurrentName(data);
    if (this.selectedLayer && this.selectedLayer.geomfield) {
      geometry = this.mapUtils.getGeomFromEWKT(data[this.selectedLayer.geomfield], this.olMap);
    } else {
      if (this.layer_id === 'OSM') {
        if (data.geojson) {
          const EWKT = this.mapUtils.getEWKTFromGeojson(data.geojson, 'EPSG:4326');
          geometry = this.mapUtils.getGeomFromEWKT(EWKT, this.olMap);
        } else {
          geometry = new Point([+data.lon, +data.lat]);
        }
      } else if (this.layer_id === 'BAN') {
        geometry = new Point(transform(data.geometry.coordinates, 'EPSG:4326', this.olMap.getView().getProjection().getCode()));
      }
    }
    this.resetLocalisation();
    this.feature = new Feature({ geometry });
    this.popup = this.mapPopup.displayPopup(displayedName, new Feature({ geometry: geometry }), this.olMap, { hideCloseBtn: true });

    // Si on est en version mobile
    if (this.isSmallScreen) {
      this.mapUtils.centerMapOnGeom(this.olMap, geometry, [200, 0, 200, 0]);
      this.vmapToolsService.getRightPanelEmiter().next(0);
      this.vmapToolsService.getActiveRightPanelEmiter().next('');
    } else {
      this.mapUtils.centerMapOnGeom(this.olMap, geometry, this.vmapToolsService.getCenterPadding());
    }
  }

  /**
   * Augmente le nombre de résultat de recherche si demandé
   */
  public displayMoreResults(): void {
    this.sendSearch(this.currentLimit + 10);
  }

  /**
   * Supprime la popup et le dessin de la localisation précédemment cliquée
   * @returns
   */
  public resetLocalisation(): boolean {
    if (this.popup) {
      this.popup.close();
      this.popup = null;
      this.feature = null;
      return true;
    }
    return false;
  }

  /**
   * Réinitialise la localisation
   */
  public clearLocalisation() {
    this.resetLocalisation();
    this.results.length = 0;
  }

  /**
   * Retourne la bonne syntaxe pour la recherche sql
   * @param value
   * @param sType
   * @returns
   */
  private likeSQLSyntax(value: any, sType: string = 'text'): any {
    const e: any = {
      value
    };
    switch (sType) {
      // Colonne de type texte
      case 'text':
      case 'textarea':
        if (this.selectedLayer.location_search_mode_id) {
          e.compare_operator = 'LIKE';
          switch (Number(this.selectedLayer.location_search_mode_id)) {
            case 2: // FORM_SEARCH_START
              e.value = value + '%';
              e.case_insensitive = true;
              break;
            case 3: // FORM_SEARCH_END
              e.value = '%' + value;
              e.case_insensitive = true;
              break;
            case 4: // FORM_SEARCH_CENTER
              e.value = '%' + value + '%';
              e.case_insensitive = true;
              break;
            case 5: // FORM_SEARCH_PLAIN_TEXT
              e.value = '%' + value + '%';
              e.case_insensitive = true;
              break;
            case 1: // FORM_SEARCH_STRICT
              e.compare_operator = '=';
              break;
          }
        }
        break;
      default:
        e.compare_operator = '=';
        break;

    }
    return e;
  }

  /**
   * Permet de faire le filtre parmi les Api actives
   * @param optionsApi
   * @returns
   */
  public filterApi(optionsApi: any[]) {
    return optionsApi.filter(e => e.enabled !== false);
  }
}
