import { Component, ViewChild, AfterViewInit, Input, OnChanges, Output, EventEmitter, OnDestroy } from '@angular/core';
import { FormreaderComponent } from 'src/app/vitis/formreader/formreader.component';
import { VitisClone } from 'src/app/vitis/core/services/utils/vitis-utils';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import { transform } from 'node_modules/ol/proj.js';
import GeoJSON from 'node_modules/ol/format/GeoJSON';
import { Subscription } from 'rxjs';

@Component({
  selector: 'app-vmap-tools-form',
  templateUrl: './vmap-tools-form.component.html',
  styleUrls: ['./vmap-tools-form.component.scss']
})
export class VmapToolsFormComponent implements AfterViewInit, OnChanges, OnDestroy {

  @ViewChild(FormreaderComponent) oFormReader: FormreaderComponent;

  @Input() map;
  @Input() oLayer;
  @Input() sDefinition;
  @Input() sFormName;
  @Input() sDatas;
  @Input() indexSelection;

  @Input() hiddenBtns;
  @Input() validForm;
  @Input() cancelForm;

  @Output() sendUpdateForm = new EventEmitter();
  @Output() sendCloseForm = new EventEmitter();
  @Output() getFormChanges = new EventEmitter();
  @Output() getForm: EventEmitter<FormreaderComponent> = new EventEmitter();

  private datas: any;

  public sLayerId;

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];

  /**
   * Constructeur
   */
  constructor(private vitisClone: VitisClone,
    private mapUtils: MapCoreUtilsService,
  ) { }


  /**
   * Appelé lorsqu'il y a des changements sur les input
   */
  ngOnChanges(changes): void {
    if (changes.oLayer && changes.oLayer.currentValue.layer_id) {
      this.sLayerId = changes.oLayer.currentValue.layer_id;
    }
    if (this.oFormReader) {
      if (changes) {
        if (changes.sFormName && changes.sFormName.currentValue) {
          this.oFormReader.setFormDefinitionName(changes.sFormName.currentValue);
        }
        if (changes.sDefinition && changes.sDefinition.currentValue && changes.sDatas && changes.sDatas.currentValue) {
          this.sDefinition = this.initFormatForm(changes.sDefinition.currentValue);
          this.oFormReader.initForm(changes.sDefinition.currentValue, this.sFormName, this.vitisClone.clone(changes.sDatas.currentValue));
        }
        if (changes.sDatas && changes.sDatas.currentValue) {
          this.datas = this.vitisClone.clone(changes.sDatas.currentValue);
          this.oFormReader.setFormValues(this.vitisClone.clone(changes.sDatas.currentValue));
        }
      }
      this.getForm.emit(this.oFormReader);
    }
  }

  /**
   * Appelé après le chargement du DOM
   */
  ngAfterViewInit(): void {
    setTimeout(() => {
      this.datas = this.vitisClone.clone(this.sDatas);
      this.sDefinition = this.initFormatForm(this.sDefinition);
      this.oFormReader.initForm(this.sDefinition, this.sFormName, this.vitisClone.clone(this.sDatas));

      if (this.getFormChanges) {
        this.subscriptions.push(this.oFormReader.oFormValues.valueChanges.subscribe(form => {
          this.getFormChanges.next(form);
        }));
      }
      this.getForm.emit(this.oFormReader);
    });
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
    // Supprime les css des formulaires lors de leur fermeture
    const oldDiv = document.getElementsByClassName('vmap-form-css');
    for (let i = 0; i < oldDiv.length; i++) {
      oldDiv[i].remove();
    }
  }

  /**
   * Retourne les valeurs du formulaire
   * @returns
   */
  public getFormValues(): any {
    return this.oFormReader.getFormValues();
  }

  /**
   * Retourne les valeurs du formulaire
   * @returns
   */
  public getIsFormValid(): any {
    return this.oFormReader.isFormValid();
  }

  /**
   * Appel la fonction update du parent
   */
  public updateForm(): void {
    this.sendUpdateForm.next(this.oFormReader.getFormValues());
  }

  /**
   * Appel la fonction pour fermer le formulaire 'update' depuis le parent
   */
  public closeForm(): void {
    this.sendCloseForm.next(undefined);
  }

  /**
   * Initialisation appelé lors de la modifications de la géométrie de la feature manuellement
   */
  public initEventModifCord(): void {

    /**
     * Met à jour le format content si une modification a été effectué depuis le format de saisie ou la projection
     * @param EWKT
     * @param format
     */
    const updateFormatContent = (EWKT, format, projection) => {
      if (format === 'Geojson') {
        const geojson = this.mapUtils.getGeojsonFromEWKT(EWKT);
        this.oFormReader.oFormValues.controls['geojson'].setValue(geojson);
      } else if (format === 'WKT') {
        const wkt = this.mapUtils.getWKTFromEWKT(EWKT);
        this.oFormReader.oFormValues.controls['wkt'].setValue(wkt.geom);
      } else if (format === 'XY') {
        const geom = this.mapUtils.getGeomFromEWKTAndProj(EWKT);
        const coordinates = geom.getCoordinates();
        if (coordinates.length > 0) {
          this.oFormReader.oFormValues.controls['longitude'].setValue(coordinates[0]);
          this.oFormReader.oFormValues.controls['latitude'].setValue(coordinates[1]);
        }
      }
    };

    this.subscriptions.push(this.oFormReader.oFormValues.get('projection').valueChanges.subscribe(projection => {
      const value = this.oFormReader.oFormValues.value;
      if (value.projection && projection && value.projection !== projection) {
        let EWKT;
        const oGeoJson = new GeoJSON({});
        if (value.format === 'WKT' && value.projection && value.wkt) {
          const EWKTtmp = this.mapUtils.getEWKTFromWKT({ proj: value.projection, geom: value.wkt });
          const geom = this.mapUtils.getGeomFromEWKT(EWKTtmp, this.map);
          EWKT = this.mapUtils.getEWKTFromGeojson({
            type: value.type,
            coordinates: geom.transform(this.map.getView().getProjection(), projection).getCoordinates()
          }, projection);
        } else if (value.format === 'Geojson' && value.geojson && value.projection) {
          const geojson = JSON.parse(value.geojson);
          const geom: any = oGeoJson.readGeometry(geojson);
          EWKT = this.mapUtils.getEWKTFromGeojson({
            type: value.type,
            coordinates: geom.transform(value.projection, projection).getCoordinates()
          }, projection);
        } else if (value.format === 'XY' && value.longitude && value.latitude) {
          const coordinates = transform([value.longitude, value.latitude], this.map.getView().getProjection(), projection);
          EWKT = this.mapUtils.getEWKTFromGeojson({
            type: value.type,
            coordinates
          }, projection);
        }
        if (EWKT) {
          updateFormatContent(EWKT, value.format, projection);
        }
      }
    }));
    this.subscriptions.push(this.oFormReader.oFormValues.get('format').valueChanges.subscribe(format => {
      const value = this.oFormReader.oFormValues.value;
      if (value.format && format && value.format !== format) {
        let EWKT;
        if (value.format === 'WKT' && value.wkt) {
          EWKT = this.mapUtils.getEWKTFromWKT({ proj: value.projection, geom: value.wkt });
          this.oFormReader.oFormValues.controls['wkt'].setValue(null);
        } else if (value.format === 'Geojson' && value.geojson && value.projection) {
          EWKT = this.mapUtils.getEWKTFromGeojson(JSON.parse(value.geojson), value.projection);
          this.oFormReader.oFormValues.controls['geojson'].setValue(null);
        } else if (value.format === 'XY' && value.longitude && value.latitude && value.projection) {
          EWKT = this.mapUtils.getEWKTFromGeojson({
            type: value.type,
            coordinates: [value.longitude, value.latitude]
          }, value.projection);
          this.oFormReader.oFormValues.controls['latitude'].setValue(null);
          this.oFormReader.oFormValues.controls['longitude'].setValue(null);
        }
        if (EWKT) {
          updateFormatContent(EWKT, format, value.projection);
        }
      }
    }));
  }

  /**
   * Initialise le style de la definition du formulaire + ajoute les boutons evenements
   * @param sDefinition
   * @returns
   */
  public initFormatForm(sDefinition): any {
    const validButton = this.validForm || 'FORM_UPDATE';
    const cancelButton = this.cancelForm || 'FORM_CLOSE';
    if (this.sendUpdateForm && this.sendCloseForm) {
      const isFormButtonsSet = sDefinition.update.rows.find(field => field.fields[0].type === 'button');
      if (!isFormButtonsSet) {
        sDefinition.update.event = 'updateForm()';
        sDefinition.update.rows.push({
          'fields': [{
            'type': 'button',
            'class': 'btn-ungroup btn-group-sm',
            'nb_cols': 12,
            'buttons': [{
              'id': 'form_submit_' + this.indexSelection,
              'type': 'submit',
              'name': 'form_submit',
              'label': validButton,
              'class': this.hiddenBtns === 'true' ? ' hidden' : 'btn-primary',
            },
            {
              'id': 'form_close_' + this.indexSelection,
              'type': 'button',
              'name': 'form_close',
              'label': cancelButton,
              'class': this.hiddenBtns === 'true' ? ' hidden' : 'btn-primary',
              'event': 'closeForm()',
            }
            ]
          }]
        });

        // Gestion des onglets
        if (sDefinition[this.sFormName].tabs) {
          if (sDefinition.update.tabs.type === 'container') {
            sDefinition.update.tabs.list.push({
              'elements': [
                'form_submit',
                'form_close'
              ],
              'class': 'tabs-container-transparent'
            });
          } else {
            for (let i = 0; i < sDefinition.update.tabs.list.length; i++) {
              const oTab = sDefinition.update.tabs.list[i];
              if (oTab.elements) {
                oTab.elements.push('form_submit');
                oTab.elements.push('form_close');
              }
            }
          }
        }
      }
    }
    sDefinition.update.eventContainer = this;
    return sDefinition;
  }
}
