import { Component, Input, OnInit, OnChanges, ViewChild, OnDestroy } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import { Subscription } from 'rxjs';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import { VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { DataGridComponent } from 'src/app/vitis/data-grid/data-grid.component';
import { DataGridOptions } from 'src/app/vitis/data-grid/data-grid.model';
import { WorkspaceListOptions } from 'src/app/vitis/workspacelist/workspacelist.model';
import { environment } from 'src/environments/environment';
import { VmapSelectService } from 'src/modules/vmap/components/carto/services/vmap-select/vmap-select.service';

declare let $: any;

@Component({
  selector: 'app-vmap-tools-requestor-modal',
  templateUrl: './vmap-tools-requestor-modal.component.html',
  styleUrls: ['./vmap-tools-requestor-modal.component.scss']
})
export class VmapToolsRequestorModalComponent implements OnInit, OnChanges, OnDestroy {

  @ViewChild(DataGridComponent, { static: true }) oDataGrid: DataGridComponent;

  @Input() sDatas;


  private aLayers: any;
  private layer: any;

  public currentLayer: string; // NAME DE LA LAYER SELECTIONNÉE

  // Largeur de page
  public pageLength: Number = environment.properties.rows_per_page;

  // Options de la grille de données.
  public oDataGridOptions: DataGridOptions = {
    datatableOptions: {},
    datagridOptions: WorkspaceListOptions
  };

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];

  /**
   * Constructeur
   */
  constructor(private mapSelectService: VmapSelectService,
    private vitisRequest: VitisRequest,
    private vitisCheck: VitisCheck,
    private auth: AuthService,
    private translate: TranslateService) { }

  /**
   * Chargement du composant
   */
  ngOnInit(): void {
    // Nombre d'éléments à afficher sur la grille
    this.pageLength = Number(environment.properties.rows_per_page);

    this.subscriptions.push(this.mapSelectService.getObservableSelectLayers().subscribe(aLayers => {
      this.aLayers = aLayers;
      this.oDataGrid.oDataGridOptions = this.oDataGridOptions;
    }));
  }

  /**
   * Appelé lorsqu'il y a des changements sur les input
   */
  ngOnChanges(changes) {
    if (changes && changes.sDatas && changes.sDatas.currentValue) {
      this.currentLayer = Object.keys(changes.sDatas.currentValue)[0];
      this.initDataGridColumns();
    }
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
  }

  /**
   * Charge les données des couches
   * @param layer_name
   */
  public loadLayerDatas(layer_name) {
    this.currentLayer = this.currentLayer === layer_name ? null : layer_name;
    this.initDataGridColumns();
  }

  /**
   * Initialise les colonnes
   */
  private initDataGridColumns(): void {
    if (this.aLayers && this.aLayers.length) {
      this.layer = this.aLayers.find(layer => layer.name === this.currentLayer);
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: `${environment.properties.app_api_url}/vmap/pglayers/${this.layer.layer_id}/columns`,
      }).then((resp: any) => {
        let columns = [];
        if (resp && resp.data && resp.data.data) {
          columns = resp.data.data.filter(e => e && e.name).map(column => {
            return {
              align: 'left',
              field: column.name,
              title: column.label,
              translateTitle: column.label,
              resizeable: true,
              searchable: true,
              sortable: true
            };
          });
        }
        this.oDataGridOptions.datagridOptions = {
          module: 'vmap',
          ressource: 'vmap/layers/:id/query',
          id_field: this.layer.layer_id,
          display: true,
          columns
        };
        this.initDataGrid();
      });
    }
  }

  /**
   * Initialise la datagrid
   */
  public initDataGrid(): void {
    // Ordre
    const order = [0, 'desc'];

    const rowId = 'nom';

    // Paramètres de la grille de données.
    this.oDataGridOptions.datatableOptions = {
      select: true,
      selectMode: 'multi',
      order: order, // Ordre
      displayStart: 0,  // Offset
      rowId: rowId,
      ajax: (data: any, callback: any, settings: any) => {

        this.getResourceData_(data).then((oResponse: any) => {

          // Remplace les données
          callback(oResponse);

        }).catch(() => {
          callback({
            recordsTotal: 0,
            recordsFiltered: 0,
            data: []
          });
        });
      }
    };

    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.columns)) {

        let oDatagridColumns = [];
        if (this.vitisCheck.isArray(this.oDataGridOptions.datagridOptions.columns)) {
          oDatagridColumns = this.oDataGridOptions.datagridOptions.columns.filter(column => {
            // Privilège(s) obligatoire(s) pour afficher la colonne.
            if (this.vitisCheck.isDefAndNotNull(column.privilege)) {
              if (this.vitisCheck.isString(column.privilege)) {
                column.privilege = [column.privilege];
              }
              if (this.vitisCheck.isArray(column.privilege)) {
                const aUserPrivileges = this.auth.getPrivileges();
                for (let i = 0; i < column.privilege.length; i++) {
                  const sPrivilege = column.privilege[i];
                  if (this.vitisCheck.isString(sPrivilege) &&
                    aUserPrivileges.indexOf(sPrivilege) !== -1) {
                    return true;
                  }
                }
              } else {
                console.error('Column privilege not string or array', column.privilege);
              }
              return false;
            }
            return true;
          });
        }

        // Traitement des colonnes
        for (let i = 0; i < oDatagridColumns.length; i++) {

          const column = oDatagridColumns[i];

          // Index par défaut de la colonne
          column.index = i;

          // Traduction
          column.translateTitle = this.translate.instant(column.title);

          column.data = column.field;

          // Visibilité de la colonne
          column.visible = column.visible === false ? false : true;

          // Colonne type
          column.type = column.template ? 'html' : 'string';

          // Alignement du contenu de la colonne.
          column.className = 'table-align-' + (column.align ? column.align : 'left');

          // Service web pour filtres
          if (column.searchWebService === undefined &&
            column.searchSuggestions !== false &&
            !this.vitisCheck.isArray(column.searchSuggestions)) {
            column.searchWebService = {
              resource: this.getRessourceUrl(),
              filterOnComplete: true,
              params: {
                attributs: column.field,
                order_by: column.field,
                distinct: true,
                sort_order: 'ASC',
                limit: 10,
              },
              label: column.field,
              value: column.field,
            };
          }
        }

        this.oDataGridOptions.datagridOptions.columns = oDatagridColumns;
      }
    }

    // Initialise les boutons icons
    this.initIconButtons();

    setTimeout(() => {

      // Création de la liste.
      this.oDataGrid.initTable(this.oDataGridOptions);
    });
  }

  /**
   * Initialise les boutons présent dans chaque ligne du tableau
   */
  public initIconButtons(): void {

    const aButtons: any = [];

    if (this.oDataGridOptions.datagridOptions.display === true &&
      (environment.properties.data_grid_update_btn_only === false ||
        this.oDataGridOptions.datagridOptions.update !== true)) {
      aButtons.push({
        type: 'display',
        icon: 'icon-vitis-search',
        event: (data: any) => { }
      });
    }
    if (this.oDataGridOptions.datagridOptions.update === true) {
      aButtons.push({
        type: 'update',
        icon: 'icon-vitis-draw',
        event: (data: any) => { }
      });
    }

    this.oDataGridOptions.iconButtons = aButtons;
  }

  /**
   * Génére l'url pour récupérer les données
   * @returns {string}
   */
  protected getRessourceUrl(): string {
    const sRessourceUrl = this.oDataGridOptions.datagridOptions.ressource;
    return environment.properties.app_api_url + '/' + sRessourceUrl.replace(':id', this.layer.layer_id);
  }

  /**
   * Récupère les donéées au format DataTables
   * @param  {any} params
   * @returns Promise
   */
  public getResourceData_(params: any): Promise<any> {
    return new Promise((resolve, reject) => {

      this.oDataGridOptions.loadingData_ = true;

      // Lance la requête
      this.vitisRequest.ajaxRequest({
        method: 'GET',
        url: this.getRessourceUrl(),
        ajaxLoader: false,
        // params: this.getRequestParams_(params),
        success: (resp: any) => {

          this.oDataGridOptions.loadingData_ = false;

          const oReturn = {
            recordsTotal: 0,
            recordsFiltered: 0,
            data: []
          };

          if (this.vitisCheck.isDefAndNotNull(resp.data)) {
            if (this.vitisCheck.isDefAndNotNull(resp.data.status === 1)) {
              if (this.vitisCheck.isArray(resp.data.data)) {
                oReturn.data = resp.data.data;
              }
              if (this.vitisCheck.isDefAndNotNull(resp.data.total_row_number)) {
                oReturn.recordsTotal = resp.data.total_row_number;
                oReturn.recordsFiltered = resp.data.list_count;
              }
            }
          }

          resolve(oReturn);
        },
        error: (err: any) => {
          console.error('error while loading values :', err);

          this.oDataGridOptions.loadingData_ = false;

          const oReturn = {
            recordsTotal: 0,
            recordsFiltered: 0,
            data: []
          };

          resolve(oReturn);
        }
      });
    });
  }

  /**
   * Ferme le requêteur
   */
  public closeRequestor(): void {
    $('#modal-requestor-layers').modal('hide');
  }

}
