import { Component, Input, OnDestroy, OnInit } from '@angular/core';
import { Feature, Map } from 'ol';
import { MapToolDef } from '../../map-tools.model';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { VmapLayer } from 'src/app/vitis/map-core/services/map-models.model';
import { MapCoreManagerService } from 'src/app/vitis/map-core/services/map-core-manager/map-core-manager.service';
import { VmapSelectService } from 'src/modules/vmap/components/carto/services/vmap-select/vmap-select.service';
import { VmapToolsService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools.service';
import { ActivatedRoute, Router } from '@angular/router';
import { Subscription } from 'rxjs';
import { EventsKey } from 'ol/events';
import { unByKey } from 'ol/Observable';
import { MapCoreDrawService } from 'src/app/vitis/map-core/services/map-core-draw/map-core-draw.service';
import { fromCircle } from 'ol/geom/Polygon';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import { Geometry, MultiPoint } from 'ol/geom';
import { environment } from 'src/environments/environment';
import { VmapToolsSetCircleRadiusService } from '../../../../services/vmap-tools/vmap-tools-set-circle-radius/vmap-tools-set-circle-radius.service';
import { IsSmallScreenService } from 'src/app/vitis/core/services/utils/is-small-screen/is-small-screen.service';

@Component({
  selector: 'app-vmap-tools-select',
  templateUrl: './vmap-tools-select.component.html',
  styleUrls: ['./vmap-tools-select.component.scss']
})
export class VmapToolsSelectComponent implements OnInit, OnDestroy {

  @Input() olMap: Map;
  @Input() oTool: MapToolDef;
  @Input() insert: boolean = true;

  public aSelectableLayers: Array<VmapLayer> = [{ name: 'Premier plan' }];
  public isRequestor: boolean = false;
  public isSmallScreen: boolean = false;
  private oSelectLayer: VmapLayer;
  private bToolOpen = false;
  private subscribeLayer;
  private subscribeLayerVisibility;

  public selectByGeomActive: string; // type de géométrie de sélection ou null si désactivé
  private olDrawinteraction; // Interraction de dessin
  private onMapDrawEndKey: EventsKey; // Souscription à drawend

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];
  private mapSubscriptions: EventsKey[] = [];

  /**
   * Constructeur
   */
  constructor(
    private mapManagerService: MapCoreManagerService,
    private mapSelectService: VmapSelectService,
    private mapToolsService: VmapToolsService,
    private route: ActivatedRoute,
    private router: Router,
    private vitisCheck: VitisCheck,
    private mapCoreDrawService: MapCoreDrawService,
    private mapCoreUtilsService: MapCoreUtilsService,
    private isSmallScreenService: IsSmallScreenService,
    private setCircleRadiusService: VmapToolsSetCircleRadiusService,
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit(): void {

    this.isSmallScreen = this.isSmallScreenService.getIsSmallScreen();
    this.subscriptions.push(this.isSmallScreenService.isSmallScreenObservable.subscribe(isSmallScreen => {
      this.isSmallScreen = isSmallScreen;
    }));

    this.initSelectLayers();
    this.olMap.on('change', () => {
      if (this.subscribeLayer) {
        this.subscribeLayer.unsubscribe();
      }
      setTimeout(() => {
        this.initSelectLayers();
      });
    });
    // @ts-ignore
    this.olMap.on('vectortiles_loaded', () => {
      if (this.subscribeLayer) {
        this.subscribeLayer.unsubscribe();
      }
      setTimeout(() => {
        this.initSelectLayers();
      });
    });

    this.subscribeLayerVisibility = this.mapManagerService.mapVisibilityChange.subscribe((change: any) => {
      if (this.subscribeLayer) {
        this.subscribeLayer.unsubscribe();
      }
      this.initSelectLayers();
    });

    this.subscriptions.push(this.mapSelectService.getObservableSelectLayers().subscribe((oSelectedLayers) => {
      this.aSelectableLayers = oSelectedLayers;
    }));

    if (this.oTool) {
      // Ouverture de l'outil
      this.subscriptions.push(this.oTool.onOpen.subscribe(() => {
        this.mapToolsService.setRightPanel(0, '');
        this.bToolOpen = true;
      }));

      // Fermeture de l'outil
      this.subscriptions.push(this.oTool.onClose.subscribe(() => {
        this.bToolOpen = false;

        // Desactive l'interraction de dessin
        this.stopGeomSelection();
      }));
    }

    // Event lors du click sur la carte
    this.mapSubscriptions.push(this.olMap.on('singleclick', (e: any) => {
      if (!this.vitisCheck.isDefAndNotNull(this.selectByGeomActive)) {
        this.onMapClick(this.olMap, e);
      }
    }));

    // Initialise la couche de dessin pour les sélections
    this.initDrawLayer();
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
    this.mapSubscriptions.forEach(mapSubscription => unByKey(mapSubscription));
    if (this.subscribeLayerVisibility) {
      this.subscribeLayerVisibility.unsubscribe();
    }
    if (this.subscribeLayer) {
      this.subscribeLayer.unsubscribe();
    }
    this.setCircleRadiusService.eventInnerDrawstartUnbind();
    this.setCircleRadiusService.eventDrawstartUnbind();
  }

  /**
   * Initialise les layers de la carte
   */
  private initSelectLayers() {
    const oFirstLayer = { name: 'Premier plan' };
    const selectLayers = this.getSelectableLayers(this.olMap);
    this.aSelectableLayers = selectLayers.length ? [
      oFirstLayer,
      ...selectLayers
    ] : [];

    // Couche de sélection
    this.subscribeLayer = this.mapSelectService.getSelectLayerObs().subscribe((oSelectedLayer) => {

      // Selectionne la couche que si elle n'est pas déjà sélectionné pour éviter les route.navigate en boucle
      if (!this.vitisCheck.isSameObject(this.oSelectLayer, oSelectedLayer)) {
        this.oSelectLayer = oSelectedLayer;

        // Sélection premier plan si la couche d'interrogation n'est pas dans la liste
        if (this.aSelectableLayers.length > 0 && this.aSelectableLayers.indexOf(this.oSelectLayer) === -1) {
          this.selectLayer(this.aSelectableLayers[0]);
        }
      }
    });

    // Outil requêteur
    this.isRequestor = this.mapToolsService.isToolActive(this.olMap, 'requestor');

    setTimeout(() => {
      if (this.oSelectLayer &&
          this.oSelectLayer.name &&
          this.aSelectableLayers &&
          this.aSelectableLayers[0] &&
          this.aSelectableLayers[0].name &&
          this.oSelectLayer.name === this.aSelectableLayers[0].name) {
        // ne rien faire car c'est la même couche
      } else {
        this.selectLayer(this.aSelectableLayers[0]);
      }
      this.mapSelectService.setSelectLayers(this.aSelectableLayers);
    });
  }

  /**
   * Choix de la couche à interroger
   * @param oLayer
   */
  public selectLayer(oLayer: VmapLayer) {

    // Ferme la sélection actuelle
    this.mapSelectService.closeSelections();
    this.mapToolsService.setRightPanel(0, '');

    this.mapSelectService.setSelectLayer(oLayer);
  }

  /**
   * Sélectionne le mode (modify/insert) avec la sélection qui va avec
   * @param mode
   * @param selection
   */
  public selectMode(mode: string, selection: any) {
    this.mapSelectService.closeSelections();
    this.mapSelectService.setSelectMode(mode);
    this.mapSelectService.setSelection(selection);
    if (this.oTool) {
      this.oTool.close();
    }
  }

  /**
   * Lance une sélection par géométrie
   * @param sGeomtype
   */
  public startGeomSelection(sGeomtype: string) {

    // Utilisé pour toggle
    const bIsActive = this.vitisCheck.isString(this.selectByGeomActive) && this.selectByGeomActive === sGeomtype;

    // Desactive l'interraction de dessin
    this.stopGeomSelection();

    // Si déjà actif alors on s'arrête là
    if (bIsActive) {
      return;
    }

    // Type de sélection en cours
    this.selectByGeomActive = sGeomtype;

    // Set ou update l'interaction d'ajout si elle n'existe pas ou si on change de geomtype
    this.olDrawinteraction = this.mapCoreDrawService.addDrawInteraction(sGeomtype.toLowerCase(), this.olMap, this.olMap.get('select_by_geom_layer'));

    this.setCircleRadiusService.setCircleListener(this.olDrawinteraction, this.olMap, 'finishCircleDrawing', this);

    this.onMapDrawEndKey = this.olDrawinteraction.on('drawend', (evt) => {
      // Récupération de la géométrie de l'event
      let oGeometry = evt.feature.getGeometry();
      let bValidGeom = true;

      // Si la géométrie est vide, elle n'est pas valide
      if(!this.vitisCheck.isDefAndNotNull(oGeometry)) {
        bValidGeom = false;
      }

      if (oGeometry.getType() === 'MultiPoint') {
        oGeometry = this.getMultipointGeom(this.olMap.get('select_by_geom_layer'), oGeometry);
      }

      // Si il s'agit d'un cercle, le convertir en polygone
      if (oGeometry.getType() === 'Circle') {
        oGeometry = fromCircle(oGeometry);
      }

      // Si il s'agit d'une ligne ou d'un LineString mais sans longueur, on ne le traite pas
      if (oGeometry.getType() === 'LineString') {
        if (oGeometry.getCoordinates().length <= 1) {
          bValidGeom = false;
        }
      }

      // Si la géométrie est de type Polygon, on vérifie le nombre de points qu'il contient, si il n'y a qu'un seul point, on ne le traite pas
      if (oGeometry.getType() === 'Polygon') {
        if (oGeometry.getCoordinates()[0].length <= 3) {
          bValidGeom = false;
        }
      }

      // Si il s'agit d'une sélection valide
      if (bValidGeom) {
        // Récupération de la géométrie au format EWKT
        const sWKTGeom = this.mapCoreUtilsService.getEWKTFromGeom(oGeometry, this.olMap);

        // Intersection avec la couche de sélection
        this.onSelectGeom(this.olMap, sWKTGeom, oGeometry);

        // Stoppe la sélection par géométrie
        setTimeout(() => {
          if (oGeometry.getType() !== 'MultiPoint') {
            this.stopGeomSelection();
          }
        });
      }
    });
    this.mapSubscriptions.push(this.onMapDrawEndKey);
  }

  /**
   * Liste des couches interrogeables
   * @param olMap
   * @returns
   */
  private getSelectableLayers(olMap: Map): Array<VmapLayer> {
    const aSelectableLayers = [];
    const aLayers = this.mapManagerService.getLayersArray(olMap);

    for (let i = 0; i < aLayers.length; i++) {
      const oLayer = aLayers[i];
      if (oLayer.is_map_selection === true && oLayer.have_select_right === true && oLayer.is_geometry_layer) {
        if (oLayer.is_map_hidden === true) {
          if (oLayer.map_parent_hidden) {
            const layer_parent_hidden = aLayers.find(layer => layer.layer_id == oLayer.map_parent_hidden);
            if (layer_parent_hidden && layer_parent_hidden.olLayer && layer_parent_hidden.olLayer.getVisible() === true) {
              aSelectableLayers.push(oLayer);
            }
          } else {
            aSelectableLayers.push(oLayer);
          }
        } else if ((oLayer.olLayer && oLayer.olLayer.getVisible() === true)) {
          aSelectableLayers.push(oLayer);
        }
      }
    }

    return aSelectableLayers.sort((a, b) => a.name > b.name && 1 || -1);
  }

  /**
   * Event appelé au click sur la carte
   * @param olMap
   * @param e
   */
  private onMapClick(olMap: Map, e: any) {
    if (this.oSelectLayer && !this.isDrawActive_(olMap)) {
      if (e.coordinate) {
        const oLayersFilters = this.olMap.get('layers_filters');
        this.mapSelectService.selectByCoordinates(e.coordinate, this.oSelectLayer, olMap, {
          get_geom: true,
          layers_filters: oLayersFilters ? oLayersFilters : {},
        }).then((aData) => {
          const datas = aData.oLayer ? aData.data : aData;

          // retrouve aData.oLayer dans this.aSelectableLayers (qui contient plus d'infos sur la couche)
          if (aData.oLayer) {
            const layer = this.aSelectableLayers.find((selectableLayer) => {
              return selectableLayer.layer_id === aData.oLayer.layer_id;
            });
            if (layer) {
              aData.oLayer = layer;
            }
          }

          // Set les nouvelles sélections
          const aSelections = [];
          if (datas && datas[0]) {
            for (let i = 0; i < datas.length; i++) {
              aSelections.push({
                data: datas[i],
                clickCoords: e.coordinate,
                oLayer: aData.oLayer || this.oSelectLayer,
              });
            }
          }

          // Affiche les popups
          this.mapSelectService.displayPopups(olMap, aSelections);

          // Sélectionne le mode (modify/insert) avec la sélection qui va avec
          this.selectMode('modify', aSelections);

          // Ferme l'outil
          if (aSelections.length > 0 && this.oTool) {
            this.oTool.close();
          }
        });
      }
    }
  }

  /**
   * Sélection par géométrie
   * @param olMap
   * @param sWKTGeom
   * @param oGeometry
   */
  private onSelectGeom(olMap: Map, sWKTGeom: string, oGeometry: Geometry) {
    this.mapSelectService.setRequestorSelection({
      layer: this.oSelectLayer,
      geom: sWKTGeom,
      olGeometry: oGeometry,
    });
  }

  /**
   * Retourne true si une interraction de dessin est en cours
   * @param olMap
   * @returns
   */
  private isDrawActive_(olMap: Map): boolean {
    const aInteractions = olMap.getInteractions().getArray();
    for (let i = 0; i < aInteractions.length; i++) {
      const olInteraction = aInteractions[i];
      if (olInteraction.get('vitis_draw') === true && olInteraction.getActive() === true) {
        return true;
      }
    }
    return false;
  }

  /**
   * Initialise le vector Layer pour ajouter les sélections à la carte
   */
  private initDrawLayer(): void {
    if (!this.vitisCheck.isDefAndNotNull(this.olMap.get('select_by_geom_layer'))) {
      this.olMap.set('select_by_geom_layer', this.mapCoreDrawService.initDrawLayer(this.olMap, null));
    }
  }

  /**
   * Stoppe toutes les interractions liées à la sélection par dessin
   */
  private stopGeomSelection() {

    // Interraction de dessin
    if (this.olDrawinteraction) {
      this.mapCoreDrawService.removeInteractions(this.olMap);
    }

    // Souscription à drawend
    if (this.onMapDrawEndKey) {
      unByKey(this.onMapDrawEndKey);
    }

    this.setCircleRadiusService.eventInnerDrawstartUnbind();
    this.setCircleRadiusService.eventDrawstartUnbind();

    this.olDrawinteraction = null;
    this.selectByGeomActive = null;

    // Supprime les dessins
    if (this.vitisCheck.isDefAndNotNull(this.olMap.get('select_by_geom_layer'))) {
      this.olMap.get('select_by_geom_layer').getSource().clear();
    }
  }

  /**
   * Concatène les anciens points de la drawlayer avec la nouvelle géométrie
   * @param oDrawLayer
   * @param oGeometry
   * @returns
   */
  private getMultipointGeom(oDrawLayer: any, oGeometry: MultiPoint) {
    const aFeatures = oDrawLayer.getSource().getFeatures();
    let aCoordinates = [];

    // Coords des features de la couche
    for (let i = 0; i < aFeatures.length; i++) {
      const oLayerGeom = aFeatures[i].getGeometry();
      if (oLayerGeom.getType() === 'MultiPoint') {
        aCoordinates = aCoordinates.concat(oLayerGeom.getCoordinates());
      }
    }

    if (oGeometry.getType() === 'MultiPoint') {
      aCoordinates = aCoordinates.concat(oGeometry.getCoordinates());
    }

    const olConcatgeom = new MultiPoint(aCoordinates);
    return olConcatgeom;
  }

  /**
   * Lance le dessin du cercle quand on lui a spécifié une valeur dans la popup
   * @param feature
   */
  public finishCircleDrawing(feature: Feature) {
    this.olDrawinteraction.finishDrawing();
  }
}
