import { Injectable } from '@angular/core';
import { VitisCheck, vitisAlert, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { Feature } from 'ol';
import { Circle, Geometry, LineString, MultiLineString, MultiPoint, MultiPolygon, Point, Polygon } from 'ol/geom';
import { VmapLayer } from 'src/app/vitis/map-core/services/map-models.model';
import { FormreaderService } from 'src/app/vitis/formreader/services/formreader.service';
import { environment } from 'src/environments/environment';
import { VmapToolsService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools.service';
import { TranslateService } from '@ngx-translate/core';

@Injectable({
  providedIn: 'root'
})
export class LayerToolsService {

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private formreaderSrvc: FormreaderService,
    private vitisRequest: VitisRequest,
    private mapToolsService: VmapToolsService,
    private translate: TranslateService
  ) { }

  /**
   * Retourne le type de géométrie de la couche
   * @param oLayer
   * @returns
   */
  public getLayerGeomType(oLayer: VmapLayer): string {
    let geomType = '';
    if (oLayer) {
      if (this.vitisCheck.isDefAndNotNull(oLayer['ms_geomtype']) &&
        this.vitisCheck.isDefAndNotNull(oLayer['ms_geomtype']['label'])) {
        if (oLayer.is_multi_geom === true) {
          geomType += 'MULTI';
        }
        if (oLayer['ms_geomtype']['label'] === 'GEOMETRY') { // La géométrie de la couche est de type GEOMETRY
          geomType += 'POINT';
        } else { // La géométrie de la couche est de type POINT, LIGNE OU POLYGONE
          geomType += oLayer['ms_geomtype']['label'];
        }
        return geomType;
      }
    }
    // La couche ne possède pas de géométrie
    return geomType;
  }

  /**
   * Initialise la feature en fonction du type de géométrie
   * @param geomType
   * @returns
   */
  public getInitFeature(geomType: string): Feature<Geometry> {
    let geometry;
    switch (geomType) {
      case 'POLYGON':
        geometry = new Polygon([]);
        break;
      case 'MULTIPOLYGON':
        geometry = new MultiPolygon([]);
        break;
      case 'POINT':
        geometry = new Point([]);
        break;
      case 'CIRCLE':
        geometry = new Circle([]);
        break;
      case 'MULTIPOINT':
        geometry = new MultiPoint([]);
        break;
      case 'LINE':
      case 'LINESTRING':
        geometry = new LineString([]);
        break;
      case 'MULTILINESTRING':
        geometry = new MultiLineString([]);
        break;
      default:
        geometry = new Point([]);
    }
    return new Feature({ geometry });
  }

  /**
   * Requête d'insertion d'une feature en base de donnée
   * @param formFeature
   * @param layer_id
   * @param OlMap
   * @returns
   */
  public insertFeatureLayer(formFeature, layer_id: number, OlMap): Promise<void> {
    return new Promise((resolve, reject) => {
      this.formreaderSrvc.manageLargesFiles(formFeature).then((managedData: any) => {
        this.vitisRequest.ajaxRequestPromise({
          method: 'POST',
          url: `${environment.properties.app_api_url}/vmap/layers/${layer_id}/query`,
          data: this.formreaderSrvc.transformToDataForm(managedData),
        }).then((resp: any) => {
          if (resp && resp.data && resp.data.data) {
            this.mapToolsService.refreshLayer(OlMap, layer_id, true);
            vitisAlert.fireToast({
              icon: 'success',
              title: this.translate.instant('COMPONENT_MAP_ENABLE_UTILS_ADD_SUCCESS_TOAST')
            });
            resolve(resp);
          }
        }).catch(e => {
          console.error(e);
          this.translate.get(['ERROR_INSERT_LAYER_OBJECT', 'ERROR_INSERT_LAYER_OBJECT_DETAIL']).subscribe(async (translation) => {
            vitisAlert.alert({
              type: 'error',
              title: translation['ERROR_INSERT_LAYER_OBJECT'],
              text: translation['ERROR_INSERT_LAYER_OBJECT_DETAIL'],
            });
          });
          reject(e);
        });
      });
    });
  }
}
