import { Injectable } from '@angular/core';
import { Geolocation } from 'ol';
import { Geometry } from 'ol/geom';
import { unByKey } from 'ol/Observable';
import { BehaviorSubject } from 'rxjs';
import { LocalStorageService } from 'src/app/vitis/core/services/utils/localstorage/local-storage.service';

@Injectable({
  providedIn: 'root'
})
export class VmapToolsGeolocationService {

  private geolocation: Geolocation;
  private isTrackingGps = null;

  private isAccuraryGps = null;
  private subscribeAccuraryGps = new BehaviorSubject<boolean>(false);

  private localisation: number[] = [];
  private trackingLocalisation = new BehaviorSubject<number[]>([]);

  private accuracyGeometry: Geometry;
  private trackingAccuracyGeometry = new BehaviorSubject<number[]>([]);

  private listenerKeyAccuracyChange;
  private listenerKeyPositionChange;


  /**
   * Constructeur
   */
  constructor(
    private localStorageService: LocalStorageService,
  ) { }

  /**
   * Initialise la géolocaliation
   * @param mapProjection
   */
  public initGeolocation(mapProjection) {
    this.geolocation = new Geolocation({
      trackingOptions: {
        enableHighAccuracy: true,
      },
      tracking: true,
      projection: mapProjection
    });

    if (this.listenerKeyAccuracyChange) {
      unByKey(this.listenerKeyAccuracyChange);
    }
    if (this.listenerKeyPositionChange) {
      unByKey(this.listenerKeyPositionChange);
    }

    this.listenerKeyPositionChange = this.geolocation.on('change:position', (event: any) => {
      this.localisation = event.target.getPosition();
      this.trackingLocalisation.next(event.target.getPosition());
    });

    this.listenerKeyAccuracyChange = this.geolocation.on('change:accuracyGeometry', (event: any) => {
      this.accuracyGeometry = event.target.getAccuracyGeometry();
      this.trackingAccuracyGeometry.next(event.target.getAccuracyGeometry());
    });
  }

  /**
   * Annule et reset la géolocalisation gps
   */
  public disableGeolocation() {
    if (this.listenerKeyAccuracyChange) {
      unByKey(this.listenerKeyAccuracyChange);
    }
    if (this.listenerKeyPositionChange) {
      unByKey(this.listenerKeyPositionChange);
    }
    this.geolocation = null;
    this.localisation = [];
    this.accuracyGeometry = null;
  }

  /**
   * Retourne l'état du traceur de position
   * @returns
   */
  public isTrackingEnable(): boolean {
    if (this.isTrackingGps === null) {
      const localStgVal = this.localStorageService.get('vmap_gps_tracking_enabled');
      this.isTrackingGps = localStgVal ?? false;
    }
    return this.isTrackingGps;
  }

  /**
   * Met à jour l'état du traceur de position
   * @param isEnable
   */
  public setTrackingGps(isEnable: boolean) {
    this.isTrackingGps = isEnable;
    this.localStorageService.set('vmap_gps_tracking_enabled', isEnable);
  }

  /**
   * Retourne l'état de la précision de la position
   * @returns
   */
  public isAccuracyEnable() {
    if (this.isAccuraryGps === null) {
      const localStgVal = this.localStorageService.get('vmap_gps_accuracy_enabled');
      this.isAccuraryGps = localStgVal ?? false;
    }
    return this.isAccuraryGps;
  }

  /**
   * Met à jour l'état de la précision de la position
   * @param isEnable
   */
  public setAccuracyGps(isEnable: boolean) {
    this.isAccuraryGps = isEnable;
    this.subscribeAccuraryGps.next(isEnable);
    this.localStorageService.set('vmap_gps_accuracy_enabled', isEnable);
  }

  /**
   * Retourne les events lier à l'état de la précision de la position gps
   * @returns
   */
  public getSubscribeAccuraryGps(): BehaviorSubject<boolean> {
    return this.subscribeAccuraryGps;
  }

  /**
   * Retourne la position de l'appareil
   * @returns
   */
  public getPosition(): number[] {
    return this.localisation;
  }

  /**
   * Ecoute et retourne à chaque déplacement la position de l'appareil
   * @returns
   */
  public getTrackingPosition(): BehaviorSubject<number[]> {
    return this.trackingLocalisation;
  }

  /**
   * Retourne la précision de la position de l'appareil
   * @returns
   */
  public getAccuracyGeometry(): Geometry {
    return this.accuracyGeometry;
  }

  /**
   * Ecoute et retourne à chaque déplacement la précision de la position de l'appareil
   * @returns
   */
  public getTrackingAccuracyGeometry(): BehaviorSubject<number[]> {
    return this.trackingAccuracyGeometry;
  }
}
