import { Injectable } from '@angular/core';
import { Circle, Geometry, LineString, Point } from 'ol/geom';
import { transform } from 'ol/proj';
import { getArea, getLength } from 'ol/sphere';
import { Fill, Stroke, Style } from 'ol/style';
import CircleStyle from 'ol/style/Circle';
import { MapCorePopupService, Popup } from 'src/app/vitis/map-core/services/map-core-popup/map-core-popup.service';

class Infos {
  geometryUid?: number;
  geometry: any;
  sPopUpData: any;
  popup: Popup;
  length?: any;
  area?: any;
}

@Injectable({
  providedIn: 'root'
})
export class VmapToolsMeasureService {

  public measuresTypes = ['POINT', 'LINE', 'POLYGON', 'CIRCLE'];

  public infosMeasures = [];

  public optionsProjection = [
    { value: 'EPSG:4326', text: '[EPSG:4326]-WGS84.LL' },
    { value: 'EPSG:2154', text: '[EPSG:2154]-RGF93.Lambert-93' },
    { value: 'EPSG:3857', text: '[EPSG:3857]-WGS84.Spherical Mercator' }
  ];

  public olMeasureStyle = new Style({
    fill: new Fill({
      color: 'rgba(255, 255, 255, 0.2)',
    }),
    stroke: new Stroke({
      color: '#ffcc33',
      width: 2,
    }),
    image: new CircleStyle({
      radius: 7,
      fill: new Fill({
        color: '#ffcc33',
      }),
    })
  });

  /**
   * Constructeur
   */
  constructor(
    private mapPopup: MapCorePopupService,
  ) { }

  /**
   * Renvoie le format correct en km ou m pour une distance
   * @param value
   * @param decimal
   * @param power
   * @returns
   */
  public formatMeasures(value, decimal = 2, power: number = 1): string {
    const km = value / Math.pow(1000, power);
    if ((value < 0 && (value * -1) > 1) || km > 1) {
      return `${km.toFixed(decimal)} km`;
    }
    return `${value.toFixed(decimal)} m`;
  }

  /**
   * Renvoie le nombre de décimal après la virgule
   * @param proj
   * @returns
   */
  public formatDecimal(proj): number {
    return proj === 'EPSG:4326' ? 7 : 2;
  }

  /**
   * Renvoie les infos de la mesure
   * @param feature
   * @param mapProjection
   */
  public infosMeasure(feature, olMap): Infos {
    const geometry: any = feature.getGeometry();
    const geometryUid = geometry.ol_uid; // Sauvegarde de l'uid de la géométrie pour les interactions
    let sPopupData: any; // Contenu dans la popup
    const infosMeasure: any = {
      geometryUid: geometry.ol_uid,
      geometry: { type: geometry.getType() },
      sPopupData,
      projection: 'EPSG:4326' // Projection de défaut de la carte
    };
    const mapProjection = olMap.getView().getProjection();
    switch (geometry.getType()) {
      case 'Point':
        const coordinates = new Point(transform(geometry.getCoordinates(), mapProjection, infosMeasure.projection)).getCoordinates();
        infosMeasure.geometry.coordinates = coordinates;
        sPopupData = {
          select: {
            value: infosMeasure.projection,
            options: this.optionsProjection,
            event: (evtSelect) => {
              const projectionValue = evtSelect.target.value;
              const infosMeasureUpdate = this.infosMeasures.find(infos => infos.geometryUid == geometry.ol_uid);
              const transformCoordinates = new Point(transform(infosMeasureUpdate.geometry.coordinates, infosMeasureUpdate.projection, projectionValue)).getCoordinates();
              infosMeasureUpdate.geometry.coordinates = transformCoordinates;
              infosMeasureUpdate.projection = projectionValue;
              infosMeasureUpdate.sPopupData.X = transformCoordinates[0].toFixed(this.formatDecimal(projectionValue));
              infosMeasureUpdate.sPopupData.Y = transformCoordinates[1].toFixed(this.formatDecimal(projectionValue));
              infosMeasureUpdate.popup.setPopupMessage({ X: infosMeasureUpdate.sPopupData.X, Y: infosMeasureUpdate.sPopupData.Y });
            }
          },
          X: coordinates[0].toFixed(7),
          Y: coordinates[1].toFixed(7)
        };
        break;
      case 'Line':
      case 'LineString':
        infosMeasure.length = this.calcLength(geometry, mapProjection);
        infosMeasure.geometry.coordinates = geometry.getCoordinates();
        sPopupData = this.formatMeasures(infosMeasure.length);
        break;
      case 'Polygon':
        infosMeasure.area = this.calcArea(geometry, mapProjection);
        infosMeasure.geometry.coordinates = geometry.getCoordinates();
        sPopupData = `${this.formatMeasures(infosMeasure.area, 2, 2)}²`;
        break;
      case 'Circle':
        const radius = this.calcRadius(geometry, mapProjection);
        infosMeasure.radius = radius;
        infosMeasure.geometry.center = geometry.getCenter();
        sPopupData = `Périmètre : ${this.formatMeasures(2 * Math.PI * radius)}`;
        sPopupData += `<br>Rayon : ${this.formatMeasures(radius)}`;
        break;
    }

    infosMeasure.popup = this.mapPopup.displayPopup(sPopupData, feature, olMap, { hideCloseBtn: true, displayFeature: false });
    infosMeasure.sPopupData = sPopupData;
    infosMeasure.geometryUid = geometryUid;
    this.infosMeasures.push(infosMeasure);
    return infosMeasure;
  }

  /**
   * Vide le tableau d'informations des mesures
   * Ferme toute les popup appartenant aux mesures
   */
  public emptiedInfosMeasures() {
    this.infosMeasures.forEach(infos => infos.popup.close());
    this.infosMeasures = [];
  }

  /**
   * Calcule l'aire d'un polygone en prenant en compte le système de projection
   * @param geometry
   * @returns
   */
  public calcArea(geometry: Geometry, mapProjection) {
    return getArea(geometry, { projection: mapProjection });
  }

  /**
   * Calcule la longueur d'une ligne en prenant en compte le système de projection
   * @param geometry
   * @returns
   */
  public calcLength(geometry: Geometry, mapProjection) {
    return getLength(geometry, { projection: mapProjection });
  }

  /**
   * Calcule le rayon d'un cercle en prenant en compte le système de projection
   * @param geometry
   * @returns
   */
  public calcRadius(circle: Circle, mapProjection) {
    const lineString = new LineString([circle.getFirstCoordinate(), circle.getLastCoordinate()], circle.getLayout());
    return getLength(lineString, { projection: mapProjection });
  }
}
