import { Injectable } from '@angular/core';
import { Feature, Map } from 'ol';
import moment from 'moment-timezone';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import { MapCoreJSONService } from 'src/app/vitis/map-core/services/map-core-json/map-core-json.service';
import { VmapToolsService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools.service';
import { Geometry } from 'ol/geom';

interface PrintParamsOptions {
  olMap: Map;
  templateDef: any; // Définition du template à imprimer
  mapScale?: number; // Échelle à afficher dans l'impression
  mapDef?: any; // Définition de la carte à imprimer
  mapFeature?: Feature<Geometry>; // Feature à afficher durant l'impression
  mapMeasures?: any; // Mesures à afficher durant l'impression
  mapExtent?: Array<number>; // Étendue de la carte
  selectResolutionVal?: number; // Facteur de résolution
  scopeVals?: any; // Valeurs à ajouter dans le scope
}

interface PrintParamsReturn {
  aMaps: Array<any>;
  aScope: Array<any>;
  aIncludes: Array<any>;
}

@Injectable({
  providedIn: 'root'
})
export class VmapToolsPrintService {

  /**
   * Constructeur
   */
  constructor(
    private auth: AuthService,
    private mapUtils: MapCoreUtilsService,
    private mapCoreJSONService: MapCoreJSONService,
    private vmapToolsService: VmapToolsService,
  ) { }

  /**
   * Forme et retourne les paramètres d'impression
   * @param opt_options
   * @returns
   */
  public getPrintParams(opt_options: PrintParamsOptions): PrintParamsReturn {

    const mapAttributions = document.getElementById('map-attributions');
    const oUserInfos = this.auth.getUserInfos();
    const aIncludes = [];
    const olMap = opt_options.olMap;
    const templateDef = opt_options.templateDef;
    const mapScale = opt_options.mapScale ? opt_options.mapScale : Math.round(this.mapUtils.getScale(olMap, { pretty: false }));
    const mapDef = opt_options.mapDef ? opt_options.mapDef : this.mapCoreJSONService.getMapDef(olMap);
    const mapMeasures = opt_options.mapMeasures ? opt_options.mapMeasures : null;
    const mapFeature = opt_options.mapFeature ? opt_options.mapFeature : null;
    const mapExtent = opt_options.mapExtent ? opt_options.mapExtent : olMap.getView().calculateExtent();
    const selectResolutionVal = opt_options.selectResolutionVal ? opt_options.selectResolutionVal : 1;
    const scopeVals = opt_options.scopeVals ? opt_options.scopeVals : {};

    const aScope: any = {
      map_scale: `1:${mapScale}`,
      date: moment.tz(oUserInfos.timezone_id).format(oUserInfos.formatdate_id),
      date_hour: moment.tz(oUserInfos.timezone_id).format(oUserInfos.formatdate_id + ' HH:mm:ss'),
      layer_sources: Array.from(mapAttributions.getElementsByTagName('li')).map(el => el.innerText),
      map_legend: this.vmapToolsService.getLayersLegends(olMap),
      user_login: oUserInfos.login,
      ...scopeVals,
    };

    const aMaps = this.getMapsDef_(olMap, aScope, templateDef, mapDef, mapExtent, mapMeasures, mapFeature, selectResolutionVal);

    return {
      aMaps,
      aScope,
      aIncludes,
    };
  }

  /**
   * Retourne la définition des carte à imprimer
   * @param olMap
   * @param aScope
   * @param templateDef
   * @param mapDef
   * @param mapMeasures
   * @param mapExtent
   * @param selectResolutionVal
   * @returns
   */
  private getMapsDef_(olMap: Map, aScope: any, templateDef: string, mapDef: any, mapExtent?: any, mapMeasures?: any, mapFeature?: Feature<Geometry>, selectResolutionVal?: number): Array<any> {
    const aMaps: any = [];
    const olMapCompare = olMap.get('map_compare');
    if (olMapCompare) { // Mode comparaison
      // Mode de comparaison activé dans le scope
      aScope.compare_mode = true;
      // Légende de la carte de comparaison
      const aLayersCompareLegends = this.vmapToolsService.getLayersLegends(olMapCompare);
      aScope.map_legend_compare = aLayersCompareLegends;
      const aMapImageCompare1Size = this.getTemplateMapSize_(olMap, templateDef, 'map_image_compare_1');
      const aMapImageCompare2Size = this.getTemplateMapSize_(olMap, templateDef, 'map_image_compare_2');
      const iResizeFactor = this.calcResizeFactor_(aMapImageCompare1Size, olMap.getSize());

      // Def carte 1
      const oMap1: any = {
        target: '#map_image_compare_1',
        map_definition: {
          map_json: mapDef,
          image_size: Math.round(aMapImageCompare1Size[0] * selectResolutionVal) + '|' + Math.round(aMapImageCompare1Size[1] * selectResolutionVal),
          resolution_coeff: iResizeFactor,
          extent: mapExtent.join('|'),
        },
      };
      // Mesures
      if (mapMeasures) {
        oMap1.map_definition.measure_features = mapMeasures;
      }

      // Features
      if (mapFeature) {
        const sEWKTFeature = this.mapUtils.getEWKTFromGeom(mapFeature.getGeometry(), olMap);
        if (sEWKTFeature) {
          oMap1.map_definition.features = sEWKTFeature;
          oMap1.map_definition.features_zoom = 100;
        }
      }

      // Def carte 2
      const oMap2 = {
        target: '#map_image_compare_2',
        map_definition: {
          map_json: this.mapCoreJSONService.getMapDef(olMapCompare),
          image_size: Math.round(aMapImageCompare2Size[0] * selectResolutionVal) + '|' + Math.round(aMapImageCompare2Size[1] * selectResolutionVal),
          resolution_coeff: iResizeFactor,
          extent: mapExtent.join('|')
        }
      };
      aMaps.push(oMap1, oMap2);
    } else { // Mode normal
      const aMapImageSize = this.getTemplateMapSize_(olMap, templateDef, 'map_image');
      const iResizeFactor = this.calcResizeFactor_(aMapImageSize, olMap.getSize(), selectResolutionVal);

      // Options carte
      const oMap: any = {
        target: '#map_image',
        map_definition: {
          map_json: mapDef,
          image_size: Math.round(aMapImageSize[0] * selectResolutionVal) + '|' + Math.round(aMapImageSize[1] * selectResolutionVal),
          resolution_coeff: iResizeFactor,
          extent: mapExtent.join('|'),
        },
      };

      // Mesures
      if (mapMeasures) {
        oMap.map_definition.measure_features = mapMeasures;
      }

      // Features
      if (mapFeature) {
        const sEWKTFeature = this.mapUtils.getEWKTFromGeom(mapFeature.getGeometry(), olMap);
        if (sEWKTFeature) {
          oMap.map_definition.features = sEWKTFeature;
          oMap.map_definition.features_zoom = 100;
        }
      }

      aMaps.push(oMap);
    }

    const aMapOverviewImageSize = this.getTemplateMapSize_(olMap, templateDef, 'map_overview');
    const iOverviewResizeFactor = this.calcResizeFactor_(aMapOverviewImageSize, olMap.getSize());
    const aOverviewExtent = [
      mapExtent[0] - (mapExtent[2] - mapExtent[0]),
      mapExtent[1] - (mapExtent[3] - mapExtent[1]),
      mapExtent[2] + (mapExtent[2] - mapExtent[0]),
      mapExtent[3] + (mapExtent[3] - mapExtent[1])
    ];
    if (aMapOverviewImageSize && aMapOverviewImageSize[0] > 0 && aMapOverviewImageSize[1] > 0) {
      aMaps.push({
        target: '#map_overview',
        map_definition: {
          map_json: mapDef,
          image_size: Math.round(aMapOverviewImageSize[0] * selectResolutionVal) + '|' + Math.round(aMapOverviewImageSize[1] * selectResolutionVal),
          resolution_coeff: iOverviewResizeFactor,
          extent: aOverviewExtent.join('|'),
          features: '',
          features_zoom: 400
        }
      });
    }
    return aMaps;
  }

  /**
   * Fonction permettant de récupérer la taille de la balise "map_image"
   * servant à définir la taille de la boîte d'impression
   * @returns
   */
  private getTemplateMapSize_(olMap: Map, templateDef: string, sTarget = 'map_image'): Array<number> {
    let aTemplateMapSize = [0, 0];

    const tempDiv = document.createElement('div');
    tempDiv.style.cssText = 'position: fixed; top: 9999px;';
    tempDiv.innerHTML = templateDef;
    document.body.appendChild(tempDiv);

    const mapElement = document.getElementById(sTarget);

    if (mapElement) {
      const style = window.getComputedStyle(mapElement, null);
      aTemplateMapSize = [
        parseFloat(style.width.replace('px', '')),
        parseFloat(style.height.replace('px', ''))
      ];
    }

    document.body.removeChild(tempDiv);

    // Taille de la carte dans le navigateur
    const aOlMapSize = olMap.getSize();

    // Calcul du facteur de resize de la zone d'impression
    const iResizeFactor = this.calcResizeFactor_(aTemplateMapSize, aOlMapSize);
    aTemplateMapSize[0] = aTemplateMapSize[0] / iResizeFactor;
    aTemplateMapSize[1] = aTemplateMapSize[1] / iResizeFactor;

    return aTemplateMapSize;
  }

  /**
   * Fonction récursive qui calcule le facteur de resize de la zone d'impression
   * @param aTemplateMapSize
   * @param aOlMapSize
   * @param iResizeFactor
   * @returns
   */
  private calcResizeFactor_(aTemplateMapSize: Array<number>, aOlMapSize: Array<number>, iResizeFactor = 1): number {
    if (aOlMapSize[0] < aTemplateMapSize[0] / iResizeFactor ||
      aOlMapSize[1] < aTemplateMapSize[1] / iResizeFactor) {
      return this.calcResizeFactor_(aTemplateMapSize, aOlMapSize, iResizeFactor + 1);
    } else {
      return iResizeFactor;
    }
  }
}
