import { Injectable } from '@angular/core';
import { Feature, Map } from 'ol';
import moment from 'moment-timezone';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import { MapCoreJSONService } from 'src/app/vitis/map-core/services/map-core-json/map-core-json.service';
import { MapCoreManagerService } from 'src/app/vitis/map-core/services/map-core-manager/map-core-manager.service';
import { VmapToolsService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools.service';
import { Geometry } from 'ol/geom';
import { environment } from 'src/environments/environment';
import { VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';

interface ReportParamsOptions {
  olMap: Map;
  reportDef: any; // Définition du rapport à imprimer
  mapDef?: any; // Définition de la carte à imprimer
  mapExtent?: Array<number>; // Étendue de la carte
  mapFeatures?: Array<Feature<Geometry>>; // Features à afficher durant l'impression
  jsonDefs?: Array<any>; // Définition des objets json
  userIncludes?: Array<any>; // Images à inclure
  userScopes?: Array<any>; // Scopes utilisateur à inclure
  userMaps?: Array<any>; // Cartes d'utilistateur
  isMultiObject?: boolean; // Le rapport est multiobjet ?
}

interface ReportParamsReturn {
  aMaps: Array<any>;
  aScopes: Array<any>;
  aIncludes: Array<any>;
  aJsonDefs: Array<any>;
}

@Injectable({
  providedIn: 'root'
})
export class VmapToolsReportService {

  /**
   * Constructeur
   */
  constructor(
    private auth: AuthService,
    private mapUtils: MapCoreUtilsService,
    private mapCoreJSONService: MapCoreJSONService,
    private mapCoreManagerService: MapCoreManagerService,
    private vmapToolsService: VmapToolsService,
    private vitisRequest: VitisRequest,
    private vitisCheck: VitisCheck,
  ) { }

  /**
   * Effectue l'impression d'un rapport
   * @param olMap
   * @param iPrintReportId
   * @param aBOs
   * @returns
   */
  public printReportById(olMap: Map, iPrintReportId: number | string, aBOs: any[]): Promise<any> {
    return new Promise((resolve, reject) => {

      if (!this.vitisCheck.isArray(aBOs) &&
        this.vitisCheck.isObject(aBOs)) {
        aBOs = [aBOs];
      }

      // Récupère le formulaire
      const url = `${environment.properties.app_api_url}/vmap/printreport/${iPrintReportId}`;
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url,
      }).then((resp: any) => {
        if (resp &&
          resp.data &&
          resp.data.data) {
          const oPrintReport = resp.data.data;

          // Lance l'impression
          this.printReport(olMap, oPrintReport, aBOs).then(() => {
            resolve(true);
          });
        } else {
          console.error('Error while finding print report ' + iPrintReportId, resp);
          reject();
        }
      }).catch((err: any) => {
        console.error('Error while finding print report ' + iPrintReportId, err);
        reject();
      });
    });
  }

  /**
   * Effectue l'impression d'un rapport
   * @param olMap
   * @param oPrintReport
   * @param aBOs
   * @returns
   */
  public printReport(olMap: Map, oPrintReport: any, aBOs: Array<any>): Promise<any> {
    return new Promise(async (resolve, reject) => {
      // Paramètres
      const sJsonObjects = oPrintReport.jsonobjects;
      const bIsMultiObject = oPrintReport.multiobject;
      const aJsonDefs = [];
      const aUserIncludes = [];
      const aUserScopes = [];
      const aUserMaps = [];
      let aFeatures = [];

      try {
        if (bIsMultiObject) {
          // Définition JSON des objets
          const oJsonDef = this.compileJsonObjects(sJsonObjects, aBOs);
          aJsonDefs.push(oJsonDef);

          // Récupération des objets JSON
          aUserIncludes.push(this.getUserIncludes(oJsonDef));
          aUserScopes.push(this.getUserScope(oJsonDef, aBOs, bIsMultiObject));
          aUserMaps.push(await this.getUserMaps(oJsonDef));

          // Features à imprimer
          aFeatures = aFeatures.concat(this.getFeaturesFromBusinessObjects(olMap, aBOs));

        } else {
          for (let i = 0; i < aBOs.length; i++) {
            // Définition JSON des objets
            const oJsonDef = this.compileJsonObjects(sJsonObjects, aBOs, i);
            aJsonDefs.push(oJsonDef);

            // Récupération des objets JSON
            aUserIncludes.push(this.getUserIncludes(oJsonDef));
            aUserScopes.push(this.getUserScope(oJsonDef, aBOs, bIsMultiObject, i));
            aUserMaps.push(await this.getUserMaps(oJsonDef));

            // Feature à imprimer par objet
            aFeatures = aFeatures.concat(this.getFeaturesFromBusinessObjects(olMap, [aBOs[i]]));
          }
        }
      } catch (e) {
        console.error(e);
        reject(e);
      }

      // Récupération des paramètres d'impression
      const oPrintParams = this.getPrintParams({
        olMap: olMap,
        reportDef: oPrintReport.htmldefinition,
        jsonDefs: aJsonDefs,
        userIncludes: aUserIncludes,
        userScopes: aUserScopes,
        userMaps: aUserMaps,
        mapDef: this.mapCoreJSONService.getMapDef(olMap),
        mapFeatures: aFeatures,
        isMultiObject: bIsMultiObject
      });

      if (oPrintParams &&
        oPrintParams.aScopes &&
        oPrintParams.aIncludes &&
        oPrintParams.aMaps &&
        oPrintParams.aJsonDefs) {

        const sScope = JSON.stringify(oPrintParams.aScopes);
        const sIncludes = JSON.stringify(oPrintParams.aIncludes);
        const sMaps = JSON.stringify(oPrintParams.aMaps);
        const sJsonDefinition = JSON.stringify(oPrintParams.aJsonDefs);

        const data = {
          scope_json: sScope,
          includes_json: sIncludes,
          maps_json: sMaps,
          json_definition: sJsonDefinition
        };

        const printreport_id = oPrintReport.printreport_id;
        const url = `${environment.properties.app_api_url}/vmap/printreportservice/${printreport_id}/print`;

        // Récupère le formulaire
        this.vitisRequest.ajaxRequestPromise({
          method: 'POST',
          url,
          data
        }).then((resp) => {
          if (resp.data && resp.data.status === 1 && this.vitisCheck.isString(resp.data.data)) {
            const sDownloadUrl = resp.data.data;
            const oUserInfos = this.auth.getUserInfos();
            const date = moment.tz(oUserInfos.timezone_id).format(oUserInfos.formatdate_id);

            this.vitisRequest.ajaxDownloadFile(sDownloadUrl, 'rapport_' + date + '.pdf');
          }
          resolve(true);

        }).catch((error) => {
          console.error('Error occured while printing...');
          reject(error);
        });
      }
    });
  }

  /**
   * Renvoie la liste des variables de type objets spécifiés dans le scope par l'utilisateur
   * @param aJsonDef
   * @param aBOs
   * @param bIsMultiObject
   * @param iIndex
   * @returns
   */
  public getUserScope(aJsonDef: any, aBOs: Array<any>, bIsMultiObject: boolean, iIndex?: number): Array<any> {
    const aUserScope = [];

    for (const oJsonObject of aJsonDef) {
      if (oJsonObject.type === 'object') {
        aUserScope[oJsonObject.target] = oJsonObject.content;
      }
    }

    aUserScope['BO'] = bIsMultiObject === true ? aBOs : aBOs[iIndex];

    return aUserScope;
  }

  /**
   * Renvoie la liste des includes d'images spécifiés par l'utilisateur dans les objets JSON
   * @param aJsonDef
   * @returns
   */
  public getUserIncludes(aJsonDef: any): Array<any> {
    const aUserIncludes = [];

    for (const oJsonObject of aJsonDef) {
      if (oJsonObject.type === 'image') {
        aUserIncludes.push({
          imageUrl: oJsonObject.imageUrl,
          target: oJsonObject.target
        });
      }
    }

    return aUserIncludes;
  }

  /**
   * Renvoie la liste des cartes spécifiées par l'utilisateur dans les objets JSON
   * @param aJsonDef
   * @returns
   */
  public getUserMaps(aJsonDef: any): Promise<Array<any>> {
    return new Promise(async (resolve, reject) => {
      const aUserMaps = [];

      for (const oJsonObject of aJsonDef) {
        if (oJsonObject.type === 'map') {
          try {
            const oMapDef = await this.mapCoreManagerService.ajaxGetMap(oJsonObject.map_id);
            aUserMaps.push({
              target: oJsonObject.target,
              map_definition: {
                map_id: oJsonObject.map_id,
                map_json: oMapDef,
                features_zoom: oJsonObject.features_zoom
              }
            });

          } catch (e) {
            reject(e);
          }
        }
      }

      resolve(aUserMaps);
    });
  }

  /**
   * Permet de précompiler les objets JSON afin de rendre utilisables le ou les objets métier sélectionnés
   * @param sJsonDef
   * @param aBOs
   * @returns
   */
  public compileJsonObjects(sJsonDef: string, aBOs: any, iIndex: number = 0): any {

    if (!sJsonDef) {
      return [];
    }

    const regexElement = new RegExp(/\{\{(.*)\}\}/, 'g');
    const aResults = sJsonDef.match(regexElement);
    let sCompiledJsonDef = sJsonDef;

    if (aResults) {
      for (const oResult of aResults) {
        const args = oResult.replace('{{', '').replace('}}', '').trim().split('.');

        // Un objet BO est trouvé
        if (args.length > 1 && args[0].startsWith('BO')) {
          let index = iIndex;
          const key = args[1];

          // Si le premier element est 'BO' avec index, le chercher
          if (args[0] !== 'BO') {
            const regexIndex = new RegExp(/\[(.*)\]/, 'g');
            index = args[0].match(regexIndex).map((element) => parseInt(element, 10))[0] ?? iIndex;
          }

          sCompiledJsonDef = sCompiledJsonDef.replace(oResult, aBOs[index][key]);
        }
      }
    }

    return JSON.parse(sCompiledJsonDef);
  }

  /**
   * Récupère et renvoie un tableau de features à partir d'un array de business objects
   * @param olMap
   * @param aBOs
   */
  public getFeaturesFromBusinessObjects(olMap: Map, aBOs: Array<any>): Array<Feature<Geometry>> {
    let aFeatures: Array<Feature<Geometry>> = [];

    for (const oBO of aBOs) {
      if (oBO.geom) {
        aFeatures = aFeatures.concat(this.mapUtils.getFeaturesFromGeom(oBO.geom, olMap));
      }
    }

    return aFeatures;
  }

  /**
   * Forme et retourne les paramètres d'impression
   * @param opt_options
   * @returns
   */
  public getPrintParams(opt_options: ReportParamsOptions): ReportParamsReturn {
    const mapAttributions = document.getElementById('map-attributions');
    const oUserInfos = this.auth.getUserInfos();
    const olMap = opt_options.olMap;
    const reportDef = opt_options.reportDef;
    const mapDef = opt_options.mapDef ? opt_options.mapDef : this.mapCoreJSONService.getMapDef(olMap);
    const mapExtent = opt_options.mapExtent ?? olMap.getView().calculateExtent();
    const aMapFeatures = opt_options.mapFeatures ?? [];
    const aJsonDefs = opt_options.jsonDefs ?? [];
    const aUserIncludes = opt_options.userIncludes ?? [];
    const aUserScopes = opt_options.userScopes ?? [];
    const aUserMaps = opt_options.userMaps ?? [];
    const bIsMultiObject = opt_options.isMultiObject ?? false;

    const oCommonScope = {
      date: moment.tz(oUserInfos.timezone_id).format(oUserInfos.formatdate_id),
      date_hour: moment.tz(oUserInfos.timezone_id).format(oUserInfos.formatdate_id + ' HH:mm:ss'),
      layer_sources: Array.from(mapAttributions.getElementsByTagName('li')).map(el => el.innerText),
      map_legend: this.vmapToolsService.getLayersLegends(olMap),
      token: this.auth.getToken()
    };

    const aMaps: Array<any> = this.getMapsDef_(olMap, reportDef, mapDef, aUserMaps, aMapFeatures, bIsMultiObject, mapExtent);
    const aScopes: Array<any> = aUserScopes.map((scope) => {
      return { ...oCommonScope, ...scope };
    });
    const aIncludes: Array<any> = aUserIncludes;

    return {
      aMaps,
      aScopes,
      aIncludes,
      aJsonDefs
    };
  }

  /**
   * Retourne la définition des cartes à imprimer
   * @param olMap
   * @param aUserMaps
   * @param reportDef
   * @param mapDef
   * @param mapMeasures
   * @param mapExtent
   * @param selectResolutionVal
   * @returns
   */
  private getMapsDef_(olMap: Map, reportDef: string, mapDef: any, aUserMaps: any, mapFeatures: Array<Feature<Geometry>>, isMultiObject: boolean, mapExtent?: any,): Array<any> {
    const aMaps: Array<any> = [];

    // Options carte principale
    const aMapImageSize = this.getReportMapSize_(olMap, reportDef, 'map_image');
    const iResizeFactor = this.calcResizeFactor_(aMapImageSize, olMap.getSize());
    const iLength = !isMultiObject ? mapFeatures.length : 1;

    // Options overview
    const aOverviewImageSize = this.getReportMapSize_(olMap, reportDef, 'map_overview');
    const iOverviewResizeFactor = this.calcResizeFactor_(aOverviewImageSize, olMap.getSize());
    const aOverviewExtent = [
      mapExtent[0] - (mapExtent[2] - mapExtent[0]),
      mapExtent[1] - (mapExtent[3] - mapExtent[1]),
      mapExtent[2] + (mapExtent[2] - mapExtent[0]),
      mapExtent[3] + (mapExtent[3] - mapExtent[1])
    ];

    for (let i = 0; i < iLength; i++) {
      const aReportMaps: Array<any> = [];
      let features: string, featuresZoom: number;

      // Features
      if (mapFeatures.length > 0) {
        const aEWKTFeatures: Array<string> = [];

        if (isMultiObject) {
          for (const feature of mapFeatures) {
            aEWKTFeatures.push(this.mapUtils.getEWKTFromGeom(feature.getGeometry(), olMap));
          }
        } else {
          aEWKTFeatures.push(this.mapUtils.getEWKTFromGeom(mapFeatures[i].getGeometry(), olMap));
        }

        features = aEWKTFeatures.join('|');
        featuresZoom = 100;
      }

      // Si des cartes supplémentaires sont fournies
      if (aUserMaps && aUserMaps.length > 0) {
        for (const oUserMap of aUserMaps[i]) {
          const oUserMapDef = oUserMap.map_definition;

          // Réassignation de la définition et des features
          oUserMapDef.image_size = aMapImageSize.join('|');
          oUserMapDef.resolution_coeff = iResizeFactor;
          oUserMapDef.extent = mapExtent.join('|');
          oUserMapDef.features = features;
          oUserMapDef.features_zoom = featuresZoom;
          oUserMapDef.token = this.auth.getToken();
          oUserMap.map_definition = oUserMapDef;

          // Ajout aux cartes
          aReportMaps.push(oUserMap);
        }
      }

      // Gestion overview
      if (aOverviewImageSize && aOverviewImageSize[0] > 0 && aOverviewImageSize[1] > 0) {
        aReportMaps.push({
          target: '#map_overview',
          map_definition: {
            map_json: mapDef,
            image_size: aOverviewImageSize.join('|'),
            resolution_coeff: iOverviewResizeFactor,
            extent: aOverviewExtent.join('|'),
            features,
            features_zoom: 400,
            token: this.auth.getToken()
          }
        });
      }

      aMaps.push(aReportMaps);
    }
    return aMaps;
  }

  /**
   * Fonction permettant de récupérer la taille de la balise "map_image"
   * servant à définir la taille de la boîte d'impression
   * @returns
   */
  private getReportMapSize_(olMap: Map, reportDef: string, sTarget = 'map_image'): Array<number> {
    let aReportMapSize = [0, 0];

    const tempDiv = document.createElement('div');
    tempDiv.style.cssText = 'position: fixed; top: 9999px;';
    tempDiv.innerHTML = reportDef;
    document.body.appendChild(tempDiv);

    const mapElement = document.getElementById(sTarget);

    if (mapElement) {
      // Passage de la position de l'élément image en absolute
      mapElement.style.position = 'absolute';

      const style = window.getComputedStyle(mapElement, null);
      aReportMapSize = [
        parseFloat(style.width.replace('px', '')),
        parseFloat(style.height.replace('px', ''))
      ];
    }

    document.body.removeChild(tempDiv);

    // Taille de la carte dans le navigateur
    const aOlMapSize = olMap.getSize();

    // Calcul du facteur de resize de la zone d'impression
    const iResizeFactor = this.calcResizeFactor_(aReportMapSize, aOlMapSize);
    aReportMapSize[0] = Math.round(aReportMapSize[0] / iResizeFactor);
    aReportMapSize[1] = Math.round(aReportMapSize[1] / iResizeFactor);

    return aReportMapSize;
  }

  /**
   * Fonction récursive qui calcule le facteur de resize de la zone d'impression
   * @param aReportMapSize
   * @param aOlMapSize
   * @param iResizeFactor
   * @returns
   */
  private calcResizeFactor_(aReportMapSize: Array<number>, aOlMapSize: Array<number>, iResizeFactor = 1): number {
    if (aOlMapSize[0] < aReportMapSize[0] / iResizeFactor ||
      aOlMapSize[1] < aReportMapSize[1] / iResizeFactor) {
      return this.calcResizeFactor_(aReportMapSize, aOlMapSize, iResizeFactor + 1);
    } else {
      return iResizeFactor;
    }
  }
}
