import { Injectable } from '@angular/core';
import { Circle, Geometry } from 'ol/geom';
import { Feature, Map } from 'ol';
import { MapCorePopupService, Popup } from 'src/app/vitis/map-core/services/map-core-popup/map-core-popup.service';
import { EventsKey } from 'ol/events';
import { unByKey } from 'ol/Observable';
import { VmapToolsMeasureService } from 'src/modules/vmap/components/carto/services/vmap-tools/vmap-tools-measure/vmap-tools-measure.service';
import moment from 'moment-timezone';
import { FunctionService } from 'src/app/vitis/core/services/utils/function/function.service';

export class InfosCircle {
  geometry: any;
  sPopUpData: string;
  popup: Popup;
}

@Injectable({
  providedIn: 'root'
})
export class VmapToolsSetCircleRadiusService {


  private infosCircle: InfosCircle;
  private bCursorInPopup: boolean;

  private listenerKeyInteractionDrawStart: EventsKey;
  private listenerKeyAddingUpdateFeature: EventsKey;
  private listenerKeyMouseEnter: any;
  private listenerKeyMouseLeave: any;

  /**
   * Constructeur
   */
  constructor(
    private mapPopup: MapCorePopupService,
    private vmapToolsMeasureService: VmapToolsMeasureService,
    private functionService: FunctionService
  ) { }

  /**
   * Demarre le dessin d'un cercle avec possibilité de saisir le rayon
   * @param interaction
   * @param map
   * @param sFunction
   * @param oFunctionContainer
   */
  public setCircleListener(interaction: any, map: Map, sFunction: string, oFunctionContainer: any) {
    // Pour les cercles : permet l'affichage de la popup
    this.listenerKeyInteractionDrawStart = interaction.on('drawstart', (evt) => {
      // Unlisten le précédent change si il existe
      this.eventInnerDrawstartUnbind();

      if (evt.feature.getGeometry().getType() == 'Circle') {
        // Ajout de la popup sur la carte
        this.infosCircle = this.getInfosCircle(evt.feature, map, sFunction, oFunctionContainer);

        this.listenerKeyMouseEnter = document.getElementsByClassName('ol-popup')[0].addEventListener('mouseenter', (cursorEvt) => {
          this.bCursorInPopup = true;
        });
        this.listenerKeyMouseLeave = document.getElementsByClassName('ol-popup')[0].addEventListener('mouseleave', (cursorEvt) => {
          this.bCursorInPopup = false;
        });

        // Modifie la valeur de la popup au changement de la geométrie
        let startModify = moment();
        this.listenerKeyAddingUpdateFeature = evt.feature.getGeometry().on('change', (event) => {
          const currentModify = moment();
          if (currentModify.diff(startModify, 'milliseconds') > 50 && !this.bCursorInPopup) {
            startModify = currentModify;
            this.setGeometryPopup(map, event.target, this.infosCircle);
          }
        });
      }
    });
  }

  /**
   * Supprime les eventlistener en lien avec le drawstart
   */
  public eventInnerDrawstartUnbind() {
    if (this.listenerKeyAddingUpdateFeature) {
      unByKey(this.listenerKeyAddingUpdateFeature);
    }
    if (this.listenerKeyMouseEnter) {
      unByKey(this.listenerKeyMouseEnter);
    }
    if (this.listenerKeyMouseLeave) {
      unByKey(this.listenerKeyMouseLeave);
    }
    if (this.infosCircle && this.infosCircle.popup) {
      this.infosCircle.popup.close();
    }
  }

  /**
   * Supprime l'event drawstart
   */
  public eventDrawstartUnbind() {
    if (this.listenerKeyInteractionDrawStart) {
      unByKey(this.listenerKeyInteractionDrawStart);
    }
  }

  /**
   * Renvoi les infos de la mesure
   * @param feature
   */
  private getInfosCircle(feature: Feature, map: Map, sFunction: string, oFunctionContainer: any): InfosCircle {
    const geometry: Geometry = feature.getGeometry();
    const sPopUpData: string = ''; // Contenu dans la popup
    const infosCircle: any = {
      geometry: { type: geometry.getType() },
      sPopUpData: sPopUpData,
      projection: 'EPSG:4326' // Projection de défaut de la carte
    };
    infosCircle.popup = this.mapPopup.displayPopup(sPopUpData, feature, map, {
      hideCloseBtn: true, displayFeature: false, buttons: [
        {
          type: 'btn',
          position: 'right',
          text: 'Ok',
          class: 'btn-primary',
          bo: this,
          event:
            () => {
              const input: HTMLInputElement = document.getElementById('set-radius-circle') as HTMLInputElement;
              const circle: Geometry = feature.getGeometry();
              if (!(circle instanceof Circle)) {
                console.error('La géométrie n\'est pas un cercle');
              } else {
                circle.setRadius(parseInt(input.value, 10));
                feature.setGeometry(circle);
                this.functionService.callFunction(sFunction, feature, oFunctionContainer);
              }
            }
        }
      ]
    });
    infosCircle.sPopUpData = sPopUpData;
    return infosCircle;
  }

  /**
   * Met à jour le contenu et la position d'une popup
   * @param olMap
   * @param geometry
   * @param infosCircle
   */
  private setGeometryPopup(olMap: Map, geometry, infosCircle: InfosCircle) {
    let popupMessage;
    const popupPosition = geometry;
    switch (geometry.getType()) {
      case 'Circle':
        const radius = this.vmapToolsMeasureService.calcRadius(geometry, olMap.getView().getProjection());
        infosCircle.geometry.center = geometry.getCenter();
        popupMessage = `<div><div class="mb-2">Rayon (en m) : </div><input type="text" class="form-control form-control-sm" id="set-radius-circle" value="${radius.toFixed(2)} "></input></div>`;
        infosCircle.sPopUpData = popupMessage;
        break;
    }

    infosCircle.popup.setPopupMessage(popupMessage);
    infosCircle.popup.setPopupPosition(popupPosition);
  }
}
