import { Injectable, EventEmitter } from '@angular/core';
import { Map } from 'ol';
import { Layer } from 'ol/layer';
import { Source } from 'ol/source';
import ImageWMS from 'ol/source/ImageWMS';
import TileWMS from 'ol/source/TileWMS';
import OSM from 'ol/source/OSM';
import XYZ from 'ol/source/XYZ';
import WMTS from 'ol/source/WMTS';
import VectorTile from 'ol/source/VectorTile';
import { vitisAlert, VitisCheck, VitisClone, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { MapCoreManagerService } from 'src/app/vitis/map-core/services/map-core-manager/map-core-manager.service';
import { MapCoreJSONService } from 'src/app/vitis/map-core/services/map-core-json/map-core-json.service';
import { environment } from 'src/environments/environment';
import { TranslateService } from '@ngx-translate/core';
import { VmapLayer } from 'src/app/vitis/map-core/services/map-models.model';
import { LegendObject } from 'src/modules/vmap/components/carto/components/vmap-layers-panel/components/vmap-layers-legend/vmap-layers-legend.model';
import { nativeMapToolsDef } from 'src/app/modules/vmap/carto/components/vmap-tools/map-tools-buttons';
import { MapToolDef } from 'src/app/modules/vmap/carto/components/vmap-tools/map-tools.model';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import { NgbModal } from '@ng-bootstrap/ng-bootstrap';

interface CloseToolsOptions {
  bExceptBottom?: boolean; // Ne ferme pas les outils de type bottom
  bOnlyTypeMenu?: boolean; // ne ferme que les outils de type menu
}

export interface GtfWorkspace {
  name: string;
  workspace_id: number;
  layer_id?: number;
  layer_workspace_id?: number;
  allow_display_card?: boolean;
  allow_requestor?: boolean;
  allow_select_all_data?: boolean;
  allow_select_current_filter?: boolean;
  selection_max_entities?: number;
}

@Injectable({
  providedIn: 'root'
})
export class VmapToolsService {

  // Gère les tailles
  private oLeftPanelEmiter: EventEmitter<number> = new EventEmitter();
  private oRightPanelEmiter: EventEmitter<number> = new EventEmitter();
  private oBottomPanelEmiter: EventEmitter<number> = new EventEmitter();

  // État actuel des panneaux en cours
  private iLeftPanel: number;
  private iRightPanel: number;
  private iBottomPanel: number;

  // Gère l'outil en cours
  private oActiveToolEmiter: EventEmitter<string> = new EventEmitter();

  // Gère les panneaux en cours
  private oActiveLeftPanelEmiter: EventEmitter<string> = new EventEmitter();
  private oActiveRightPanelEmiter: EventEmitter<string> = new EventEmitter();
  private oActiveBottomPanelEmiter: EventEmitter<string> = new EventEmitter();

  // État actuel des panneaux en cours
  private sActiveLeftPanel: string = '';
  private sActiveRightPanel: string = '';
  private sActiveBottomPanel: string = '';

  // Renvoie les panneaux qui sont ouverts
  private oCurrentOpenedLeftPanel: EventEmitter<string> = new EventEmitter();
  private oCurrentOpenedRightPanel: EventEmitter<string> = new EventEmitter();
  private oCurrentOpenedBottomPanel: EventEmitter<string> = new EventEmitter();

  // Boutons panneaux venant d'être cliqués
  private oPanelBtnClicked: EventEmitter<string> = new EventEmitter();

  // Renvoie l'outil qui est ouvert
  private currentOpenedTool: EventEmitter<string> = new EventEmitter();

  // Event de rafraîchissement de la carte
  private oRefreshMapEmitter: EventEmitter<number> = new EventEmitter();

  // Outils disponible
  private oToolsDef: any;

  /**
   * Constructeur
   */
  constructor(
    private mapCoreManagerService: MapCoreManagerService,
    private mapCoreJSONService: MapCoreJSONService,
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private translate: TranslateService,
    private auth: AuthService,
    private modalService: NgbModal,
  ) {
    this.getActiveLeftPanelEmiter().subscribe((next: string) => {
      this.sActiveLeftPanel = next;
      this.oCurrentOpenedLeftPanel.next(next);
    });

    this.getActiveRightPanelEmiter().subscribe((next: string) => {
      this.sActiveRightPanel = next;
      this.oCurrentOpenedRightPanel.next(next);
    });

    this.getActiveBottomPanelEmiter().subscribe((next: string) => {
      this.sActiveBottomPanel = next;
      this.oCurrentOpenedBottomPanel.next(next);
    });

    this.getLeftPanelEmiter().subscribe((next: number) => {
      this.iLeftPanel = next;
    });

    this.getRightPanelEmiter().subscribe((next: number) => {
      this.iRightPanel = next;
    });

    this.getBottomPanelEmiter().subscribe((next: number) => {
      this.iBottomPanel = next;
    });

    // Charge la liste des outils disponibles
    this.loadToolsDef();
  }

  /**
   * Ferme tous les panneaux ouverts
   * @returns
   */
  public closeAllPanels(): void {
    // Fermeture du panneau de gauche
    this.setLeftPanel(0, '');

    // Fermeture du panneau de droite
    this.setRightPanel(0, '');

    // Fermeture du panneau du bas
    this.setBottomPanel(0, '');
  }

  /**
   * Définit un outil à une position et une taille donnée
   * @param position
   * @param cols
   * @param tool
   */
  public setPanel(position: string, cols: number, tool: string) {
    if (position === 'left') {
      this.setLeftPanel(cols, tool);
    } else if (position === 'right') {
      this.setRightPanel(cols, tool);
    } else if (position === 'bottom') {
      this.setBottomPanel(cols, tool);
    }
  }

  /**
   * Change l'état du panneau de gauche
   */
  public setLeftPanel(cols: number, tool: string): void {
    this.getLeftPanelEmiter().next(cols);

    // Permet de ne pas fermer / ouvrir l'outil si la taille change
    if (this.sActiveLeftPanel !== tool) {
      this.getActiveLeftPanelEmiter().next(tool);
    }
  }

  /**
   * Change l'état du panneau de droite
   */
  public setRightPanel(cols: number, tool: string): void {
    this.getRightPanelEmiter().next(cols);

    // Permet de ne pas fermer / ouvrir l'outil si la taille change
    if (this.sActiveRightPanel !== tool) {
      this.getActiveRightPanelEmiter().next(tool);
    }
  }

  /**
   * Change l'état du panneau du bas
   */
  public setBottomPanel(cols: number, tool: string): void {
    this.getBottomPanelEmiter().next(cols);

    // Permet de ne pas fermer / ouvrir l'outil si la taille change
    if (this.sActiveBottomPanel !== tool) {
      this.getActiveBottomPanelEmiter().next(tool);
    }
  }

  /**
   * Emiter appelé lors de l'ouverture d'un des menus de gauche
   * la valeur vaut 0, 1, 2, 3 en fonction de la largeur de la fenêtre
   * @returns
   */
  public getLeftPanelEmiter(): EventEmitter<number> {
    return this.oLeftPanelEmiter;
  }

  /**
   * Emiter appelé lors de l'ouverture d'un des menus de droite
   * la valeur vaut 0, 1, 2, 3 en fonction de la largeur de la fenêtre
   * @returns
   */
  public getRightPanelEmiter(): EventEmitter<number> {
    return this.oRightPanelEmiter;
  }

  /**
   * Emiter appelé lors de l'ouverture d'un des menus du bas
   * la valeur vaut 0, 1, 2, 3 en fonction de la largeur de la fenêtre
   * @returns
   */
  public getBottomPanelEmiter(): EventEmitter<number> {
    return this.oBottomPanelEmiter;
  }

  /**
   * Emiter appelé lors de l'ouverture d'un des menus
   * la valeur vaut le nom de l'outil
   * @returns
   */
  public getActiveToolEmiter(): EventEmitter<string> {
    return this.oActiveToolEmiter;
  }

  /**
   * Emiter appelé lors de l'ouverture d'un des panneaux de gauche
   * la valeur vaut le nom de l'outil
   * @returns
   */
  public getActiveLeftPanelEmiter(): EventEmitter<string> {
    return this.oActiveLeftPanelEmiter;
  }

  /**
   * Emiter appelé lors de l'ouverture d'un des panneaux de droite
   * la valeur vaut le nom de l'outil
   * @returns
   */
  public getActiveRightPanelEmiter(): EventEmitter<string> {
    return this.oActiveRightPanelEmiter;
  }

  /**
   * Emiter appelé lors de l'ouverture d'un des panneaux du bas
   * la valeur vaut le nom de l'outil
   * @returns
   */
  public getActiveBottomPanelEmiter(): EventEmitter<string> {
    return this.oActiveBottomPanelEmiter;
  }

  /**
   * Emiter appelé lors du changement de l'outil qui est ouvert
   * sur le panneau de gauche
   * @returns
   */
  public getCurrentOpenedLeftPanel(): EventEmitter<string> {
    return this.oCurrentOpenedLeftPanel;
  }

  /**
   * Emiter appelé lors du changement de l'outil qui est ouvert
   * sur le panneau de droite
   * @returns
   */
  public getCurrentOpenedRightPanel(): EventEmitter<string> {
    return this.oCurrentOpenedRightPanel;
  }

  /**
   * Emiter appelé lors du changement de l'outil qui est ouvert
   * sur le panneau du bas
   * @returns
   */
  public getCurrentOpenedBottomPanel(): EventEmitter<string> {
    return this.oCurrentOpenedBottomPanel;
  }

  /**
   * Emiter appelé lors du click sur un des boutons pannel
   * @returns
   */
  public getPanelBtnClickedEmiter(): EventEmitter<string> {
    return this.oPanelBtnClicked;
  }

  /**
  * Sécifie le bouton pannel ayant été clické
  */
  public setPanelBtnClicked(tool: string): void {
    this.getPanelBtnClickedEmiter().next(tool);
  }

  /**
   * Emiter appelé lors de l'ouverture d'un des menus
   * la valeur vaut le nom de l'outil ouvert
   * @returns
   */
  public getCurrentOpenedTool(): EventEmitter<string> {
    return this.currentOpenedTool;
  }

  /**
   * Retourne un emitter appelé au rafraichissement de la carte
   * @returns
   */
  public getRefreshMapEmitter(): EventEmitter<number> {
    return this.oRefreshMapEmitter;
  }

  /**
   * Rafraîchit la map en ne conservant rien (comme si on faisait F5)
   * @param olMap
   */
  public reloadMap(olMap: Map): void {
    const reloadMapTrigger = olMap.get('reloadMapTrigger');
    if (reloadMapTrigger) {
      reloadMapTrigger.next();
    }
  }

  /**
   * Rafraîchit la carte en cours
   *
   * @param olMap
   * @param keepLoadedTiles si true alors garde les tuiles tant qu'elles n'ont pas été chargées
   * @param aLayersIds permet de définir les couches à recharger
   * @param oAPIFilter permet d'effectuer un filtre sur les couches postgres
   */
  public refreshMap(olMap: Map, keepLoadedTiles = false, aLayersIds = null, oAPIFilter = null): void {
    const iMapId = olMap.get('map_id');
    const checkLayerId = this.vitisCheck.isArray(aLayersIds) && aLayersIds.length > 0;

    if (iMapId) {
      // Permet de rafraichir le mapfile
      this.mapCoreManagerService.ajaxGetMap(iMapId, oAPIFilter).then((oMapDef: any) => {
        const aLayers = olMap.getLayers().getArray();
        for (let i = 0; i < aLayers.length; i++) {
          const olLayer: any = aLayers[i];
          const olSource = olLayer.getSource();

          // Test aLayersIds
          if (checkLayerId && !this.isLayerInIds_(olLayer, aLayersIds)) {
            continue;
          }

          if (
            olSource instanceof ImageWMS ||
            olSource instanceof TileWMS
          ) {
            // Modification du nom de la couche (WMS) qui a changée au moment du ajaxGetMap
            // Mis à jour des paramètres de la couche
            const newLayerDef = this.findLayerInMapDef(olLayer.get('layer_id'), oMapDef);

            if (newLayerDef &&
              newLayerDef.params &&
              newLayerDef.params.LAYERS) {
              olSource.updateParams({ 'LAYERS': newLayerDef.params.LAYERS, 'TIMESTAMP': Date.now() });

              const aNewLayers = newLayerDef.params.LAYERS.split(',');
              if (this.vitisCheck.isArray(aNewLayers)) {

                // sublayers
                olLayer.set('sublayers', aNewLayers);

                // activeSublayers
                olLayer.set('activeSublayers', aNewLayers);

                // queryableSublayers
                if (this.vitisCheck.isDefAndNotNull(newLayerDef.is_queryable_getfeatureinfo)) {
                  olLayer.set('queryableSublayers', aNewLayers);
                }
              }
            }
          }

          if (
            olSource instanceof ImageWMS ||
            olSource instanceof TileWMS ||
            olSource instanceof OSM ||
            olSource instanceof XYZ ||
            olSource instanceof WMTS ||
            olSource instanceof VectorTile
          ) {
            if (keepLoadedTiles !== true) {
              olSource.refresh();
            }
          }

          if (oAPIFilter) {
            olMap.set('layers_filters', oAPIFilter);
          } else {
            olMap.set('layers_filters', null);
          }
        }

        this.oRefreshMapEmitter.next(Date.now());
      });
    } else {
      console.error('cannot found map_id', olMap);
    }
  }

  /**
   * Rafraîchit une couche de la carte
   * @param olMap
   * @param iLayerId
   * @param keepLoadedTiles
   */
  public refreshLayer(olMap: Map, iLayerId: number, keepLoadedTiles = false): void {
    // Conservation des filtres sur les couches.
    const aLayersFilters = olMap.get('layers_filters');
    this.refreshMap(olMap, keepLoadedTiles, [iLayerId], aLayersFilters);
  }

  /**
   * Centre sur la position initiale de la carte
   * @param olMap
   */
  public centerHome(olMap: Map): void {

    // Vue d'après la définition
    const view = this.mapCoreJSONService.getViewFromDef(olMap.get('map_definition'), {
      size: olMap.getSize()
    });
    const mapCenter = view.getCenter();
    const mapResolution = view.getResolution();

    if (mapCenter && mapResolution) {
      olMap.getView().setCenter(mapCenter);
      olMap.getView().setResolution(mapResolution);
    } else {
      console.error('Cannot get map original center and resolution');
    }
  }

  /**
   * Retourne true si la couche fait partie des aIds
   * @param olLayer
   * @param aIds
   * @returns
   */
  private isLayerInIds_(olLayer: Layer<Source>, aIds: Array<number>): boolean {
    if (this.vitisCheck.isArray(aIds) && aIds.length > 0) {
      if (this.vitisCheck.isArray(olLayer.get('fusion_layers'))) {
        for (let i = 0; i < olLayer.get('fusion_layers').length; i++) {
          const olChildLayer = olLayer.get('fusion_layers')[i];
          const isInIds = this.isLayerInIds_(olChildLayer, aIds);
          if (isInIds === true) {
            return true;
          }
        }
      } else {
        if (olLayer.get('layer_id') && aIds.indexOf(olLayer.get('layer_id')) !== -1) {
          return true;
        }
      }
    }
    return false;
  }

  /**
   * Retourne le tableau de légendes
   * @param olMap
   * @param scale
   * @returns
   */
  public getLayersLegends(olMap: Map, scale?: number) {
    const aLayersLegends = [];
    const aLayers = this.mapCoreManagerService.getLayersArray(olMap);
    let scaleParam = '';

    for (let i = 0; i < aLayers.length; i++) {
      const olLayer = aLayers[i].olLayer;
      if (olLayer.get('visible') === true) {
        if (olLayer.get('legend')) {
          if (olLayer.get('layerType') === 'postgres') {
            scaleParam = '';

            if (olLayer.get('is_dynamic_legend') && this.vitisCheck.isDefAndNotNull(scale)) {
              scaleParam = '&SCALE=' + scale;
            }

            if (olLayer.get('type') === 'tilewms' || olLayer.get('type') === 'imagewms') {
              const aWMSLegendObjects = this.getWMSLegendObjects_(aLayers[i], scaleParam);
              if (aWMSLegendObjects) {
                aLayersLegends.push({
                  collapsed: false,
                  layerName: olLayer.get('name'),
                  legendObjects: aWMSLegendObjects,
                  pictogram: aLayers[i].pictogram,
                  legend: aLayers[i].legend
                });
              }
            }
          } else if (this.vitisCheck.isDefAndNotNull(aLayers[i].legend)) {
            aLayersLegends.push({
              collapsed: false,
              layerName: olLayer.get('name'),
              legend: aLayers[i].legend,
              pictogram: aLayers[i].pictogram,
              legendObjects: [{
                layerName: olLayer.get('name'),
                legendName: null,
                legendType: 'image',
                legendURL: aLayers[i].legend,
                service_login: null,
                service_password: null,
                url: null,
              }]
            });
          }
        } else if (olLayer.get('layerType') === 'wms') {
          aLayersLegends.push({
            collapsed: false,
            layerName: olLayer.get('name'),
            legend: aLayers[i].legend,
            pictogram: aLayers[i].pictogram,
            legendObjects: [{
              layerName: olLayer.get('name'),
              legendName: null,
              legendType: 'image',
              legendURL: this.getWmsLegendUrl(aLayers[i]),
              service_login: null,
              service_password: null,
              url: null,
            }]
          });
        }
      }
    }

    // Tri par ordre alpha
    aLayersLegends.sort((a, b) => {
      if (a.layerName < b.layerName) {
        return -1;
      }
      if (a.layerName > b.layerName) {
        return 1;
      }
      return 0;
    });

    return aLayersLegends;
  }

  /**
   * Parse la couche et retourne le tableau de LegendObject
   * @param oLayer
   * @param scaleParam
   * @returns
   */
  private getWMSLegendObjects_(oLayer: VmapLayer, scaleParam: any): Array<LegendObject> {
    const olLayer = oLayer.olLayer;
    const sLayers = olLayer.getSource().getParams()['LAYERS'];
    let aLayerParam = [];

    if (this.vitisCheck.isDefAndNotNull(sLayers)) {
      aLayerParam = olLayer.getSource().getParams()['LAYERS'].split(',');
    }

    // Permet de ne pas visualiser les couches non visibles
    const aActiveSublayers = olLayer.get('activeSublayers');
    if (this.vitisCheck.isDefAndNotNull(aActiveSublayers)) {
      aLayerParam = aActiveSublayers;
    }

    const sSourceUrl = olLayer.get('type') === 'imagewms' ? olLayer.getSource().getUrl() : olLayer.get('type') === 'tilewms' ? olLayer.getSource().getUrls()[0] : null;
    if (!sSourceUrl) {
      return;
    }

    const aLegendsArray = [];
    for (let iii = 0; iii < aLayerParam.length; iii++) {
      const separator = sSourceUrl.indexOf('?') !== -1 ? '&' : '?';
      let legendURL = sSourceUrl + separator + 'request=GetLegendGraphic&LAYER=' + aLayerParam[iii] + '&FORMAT=image/png&VERSION=1.1.0&service=WMS' + scaleParam;
      let legendName: any;
      let legendType = 'image';

      if (this.vitisCheck.isDefAndNotNull(oLayer.definition_sym_json)) {
        const oSymDef = this.vitisCheck.isString(oLayer.definition_sym_json) ? JSON.parse(oLayer.definition_sym_json) : oLayer.definition_sym_json;

        // Type singleclass
        if (oSymDef &&
          oSymDef.type === 'singleclass') {
          legendName = oSymDef.legendName ? oSymDef.legendName : oLayer.name;
          legendURL += `&RULE=${aLayerParam[iii]}`;
          legendType = 'vmap';
        }
      }

      if (this.vitisCheck.isDefAndNotNull(oLayer.legend)) {
        legendURL = oLayer.legend;
      }

      aLegendsArray.push({
        url: sSourceUrl,
        layerName: olLayer.get('name'),
        service_login: this.vitisCheck.isDefAndNotNull(olLayer.get('service_login')) ? olLayer.get('service_login') : null,
        service_password: this.vitisCheck.isDefAndNotNull(olLayer.get('service_password')) ? olLayer.get('service_password') : null,
        legendURL,
        legendName,
        legendType,
      });
    }

    return aLegendsArray;
  }
  /**
   * Génère l'url pour accéder à la légende d'une couche qui provient d'un flux wms
   * @returns
   */
  private getWmsLegendUrl(oLayer: VmapLayer): String {
    const olLayer = oLayer.olLayer;
    const sUrl = olLayer.getSource().getUrl();
    const separator = sUrl.indexOf('?') !== -1 ? '&' : '?';
    const sLayer = olLayer.getSource().getParams()['LAYERS'];
    const sFormat = olLayer.getSource().getParams()['FORMAT'];
    const sversion = olLayer.getSource().getParams()['VERSION'];
    const sService = 'WMS';
    const sUrlfinal = sUrl + separator + 'SERVICE=' + sService + '&VERSION=' + sversion + '&REQUEST=getlegendgraphic&FORMAT=' + sFormat + '&layer=' + sLayer;

    return sUrlfinal;
  }
  /**
   * Calcul du padding à donner lors de l'appel à centerMapOnGeom en fonction des menus ouverts
   * @returns
   */
  public getCenterPadding(): Array<number> {
    return [
      200, // Haut
      this.getRightPadding_(), // Droite
      this.getBottomPadding_(), // Bas
      this.getLeftPadding_(), // Gauche
    ];
  }

  /**
   * Calcul du padding du bas
   * @returns
   */
  private getBottomPadding_() {
    switch (this.iBottomPanel) {
      case 1:
        return 400;
        break;
      default:
        return 200;
        break;
    }
  }

  /**
   * Calcul du padding de droite
   * @returns
   */
  private getRightPadding_() {
    switch (this.iRightPanel) {
      case 1:
        return 400;
        break;
      case 2:
        return 600;
        break;
      default:
        return 200;
        break;
    }
  }

  /**
   * Calcul du padding de gauche
   * @returns
   */
  private getLeftPadding_() {
    switch (this.iLeftPanel) {
      case 1:
        return 400;
        break;
      case 2:
        return 600;
        break;
      default:
        return 200;
        break;
    }
  }

  /**
   * Charge les outils depuis la table toolmap
   * @returns
   */
  public ajaxGetToolsmap(): Promise<any> {
    return new Promise((resolve, reject) => {
      // Charge les modules complémentaires depuis la base
      this.vitisRequest.ajaxRequestPromise({
        url: `${environment.properties.app_api_url}/vmap/toolmap`,
        method: 'GET',
        displayErrorMessage: false,
      }).then((resp: any) => {
        const oCustomTool = {};
        if (resp.data && this.vitisCheck.isArray(resp.data.data)) {
          for (let i = 0; i < resp.data.data.length; i++) {
            const oTool = resp.data.data[i];
            if (oTool &&
              oTool.toolmap_id &&
              oTool.name &&
              (oTool.icon || oTool.icon_svg)) {
              oCustomTool[oTool.name] = {
                name: oTool.name,
                icon: oTool.icon,
                icon_svg: oTool.icon_svg,
                title: oTool.title ?? '',
                type: oTool.type ?? 'panel',
                component: oTool.component ?? 'app-vmap-tools-simple-html',
                panelPosition: oTool.panel_position ?? 'right',
                panelSize: oTool.panel_size ?? 1,
                htmlContent: oTool.html_content ?? '',
                jsContent: oTool.js_content ?? '',
                cssContent: oTool.css_content ?? '',
              };
            } else {
              console.error('tool not valid', oTool);
            }
          }
        }
        resolve(oCustomTool);
      });
    });
  }

  /**
   * Charge les outils disponibles en concaténant ceux du code avec les modules complémentaires
   * @returns
   */
  public loadToolsDef() {
    return new Promise((resolve, reject) => {
      if (this.vitisCheck.isDefAndNotNull(this.oToolsDef)) {
        resolve(this.oToolsDef);
      } else {
        this.ajaxGetToolsmap().then((oCustomTool) => {
          this.oToolsDef = { ...nativeMapToolsDef, ...oCustomTool };
          resolve(this.oToolsDef);
        }).catch(() => {
          console.error('error while loading toolmap');
          this.oToolsDef = { ...nativeMapToolsDef };
          resolve(this.oToolsDef);
        });
      }
    });
  }

  /**
   * Retourne les outils disponibles
   * @param olMap
   * @returns
   */
  public getMapToolsDef(olMap: Map) {
    return this.oToolsDef;
  }

  /**
   * Ferme l'ensemble des outils sauf exceptions (voir options)
   */
  public closeTools(olMap: Map, options?: CloseToolsOptions) {
    const oMapToolsDef = this.getMapToolsDef(olMap);
    options = this.vitisCheck.isObject(options) ? options : {};
    for (const key in oMapToolsDef) {
      if (Object.prototype.hasOwnProperty.call(oMapToolsDef, key)) {
        const oTool = oMapToolsDef[key];

        let close = true;
        if (options.bOnlyTypeMenu && oTool.type !== 'submenu') {
          close = false;
        }
        if (options.bExceptBottom && oTool.panelPosition === 'bottom') {
          close = false;
        }

        if (close) {
          this.openTool_(oTool, false);
        }
      }
    }
  }

  /**
   * Ouvre un outil
   * @param olMap
   * @param sToolName
   * @param bHandleMenus si true alors gère l'affichage en fonction des autres outils
   */
  public openTool(olMap: Map, sToolName: string, bHandleMenus = true) {
    const oTool = this.getToolByName(olMap, sToolName);
    if (bHandleMenus) {
      this.openToolWithMenus_(olMap, oTool, true);
    } else {
      this.openTool_(oTool, true);
    }
  }

  /**
   * Ferme un outil
   * @param olMap
   * @param sToolName
   * @param bHandleMenus si true alors gère l'affichage en fonction des autres outils
   */
  public closeTool(olMap: Map, sToolName: string, bHandleMenus = true) {
    const oTool = this.getToolByName(olMap, sToolName);
    if (bHandleMenus) {
      this.openToolWithMenus_(olMap, oTool, false);
    } else {
      this.openTool_(oTool, false);
    }
  }

  /**
   * Retourne un outil par son nom
   * @param olMap
   * @param sToolName
   * @returns
   */
  public getToolByName(olMap: Map, sToolName: string) {
    const oMapToolsDef = this.getMapToolsDef(olMap);

    let oTool: any;
    for (const key in oMapToolsDef) {
      if (Object.prototype.hasOwnProperty.call(oMapToolsDef, key)) {
        if (oMapToolsDef[key].name === sToolName) {
          oTool = oMapToolsDef[key];
        }
      }
    }

    if (this.vitisCheck.isDefAndNotNull(oTool)) {
      return oTool;
    } else {
      console.error('tool not defined', sToolName, oMapToolsDef);
    }
  }

  /**
   * Ferme un outil de type panneau à la position donnée
   */
  public closePanelAt(olMap: Map, position: string) {
    const oMapToolsDef = this.getMapToolsDef(olMap);
    for (const key in oMapToolsDef) {
      if (Object.prototype.hasOwnProperty.call(oMapToolsDef, key)) {
        const oTool: MapToolDef = oMapToolsDef[key];

        if (oTool.type === 'panel' && oTool.panelPosition === position) {
          if (oTool.onClose && oTool.collapsed_ === false) {
            oTool.onClose.next(true);
          }
          oTool.collapsed_ = true;
        }
      }
    }
  }

  /**
   * Ouvre ou ferme l'outil
   * @param bOpen
   */
  private openTool_(oTool: MapToolDef, bOpen: boolean) {
    if (oTool.onOpen) {
      if (bOpen === true && oTool.collapsed_ !== false) {
        oTool.onOpen.next(true);
      }
    }
    if (oTool.onClose && oTool.collapsed_ === false) {
      if (bOpen === false) {
        oTool.onClose.next(true);
      }
    }
    oTool.collapsed_ = !bOpen;
  }

  /**
   * Ouvre l'outil et gère les différents menus
   * @param olMap
   * @param oTool
   * @param bOpen
   */
  private openToolWithMenus_(olMap: Map, oTool: MapToolDef, bOpen: boolean) {

    if (oTool.type === 'submenu') {

      // Ferme les panneaux de gauche et droite
      this.setLeftPanel(0, '');
      this.setRightPanel(0, '');

      if (bOpen) {
        // Ferme les outils de gauche et de droite
        this.closeTools(olMap, {
          bExceptBottom: true,
        });
        this.openTool_(oTool, true);
      } else {
        this.openTool_(oTool, false);
      }
    } else if (oTool.type === 'panel') {

      if (oTool.collapsed_ === false) {
        if (oTool.panelPosition === 'left') {
          this.setLeftPanel(0, '');

        } else if (oTool.panelPosition === 'right') {
          this.setRightPanel(0, '');

        } else if (oTool.panelPosition === 'bottom') {
          this.setBottomPanel(0, '');
        }
        this.openTool_(oTool, false);

      } else {
        if (oTool.panelPosition === 'left') {
          this.setLeftPanel(oTool.panelSize, oTool.name);

        } else if (oTool.panelPosition === 'right') {
          this.setRightPanel(oTool.panelSize, oTool.name);

        } else if (oTool.panelPosition === 'bottom') {
          this.setBottomPanel(oTool.panelSize, oTool.name);
        }

        this.closeTools(olMap, {
          bOnlyTypeMenu: true,
        });
        this.closePanelAt(olMap, oTool.panelPosition);
        this.openTool_(oTool, true);
      }
    }
  }

  /**
   * Outils de la carte avec les positions
   * @param olMap
   * @returns
   */
  public getActiveTools(olMap: Map) {
    const oMapDef = olMap.get('map_definition');
    if (oMapDef && oMapDef.maptools && oMapDef.maptools.tools) {
      return oMapDef.maptools.tools;
    }
  }

  /**
   * Outils de la carte au format d'un tableau
   * @param olMap
   * @returns
   */
  public getActiveToolsAsArray(olMap: Map) {
    const aMapTools = [];
    const oMapTools = this.getActiveTools(olMap);

    for (const sPosition in oMapTools) {
      if (Object.prototype.hasOwnProperty.call(oMapTools, sPosition)) {
        if (this.vitisCheck.isArray(oMapTools[sPosition])) {
          for (let i = 0; i < oMapTools[sPosition].length; i++) {
            const sTool = oMapTools[sPosition][i];
            if (aMapTools.indexOf(sTool) === -1) {
              aMapTools.push(sTool);
            }
          }
        }
      }
    }

    return aMapTools;
  }

  /**
   * Retourne true/false si un outil est disponible pour cette carte
   * @param olMap
   * @param sTool
   * @returns
   */
  public isToolActive(olMap: Map, sTool: string) {
    const aMapTools = this.getActiveToolsAsArray(olMap);
    if (this.vitisCheck.isArray(aMapTools) && aMapTools.indexOf(sTool) !== -1) {
      return true;
    } else {
      return false;
    }
  }

  /**
   * Récupère les traitements GTF associée à la couche et à l'utilisateur
   */
  public getLayerFMEWorkspaces(layer_id: number): Promise<any> {
    return new Promise((resolve, reject) => {
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: `${environment.properties.app_api_url}/vmap/layers/${layer_id}/fme-userworkspaces`,
      }).then((resp: any) => {
        if (resp && resp.data && resp.data.data) {
          resolve(resp.data.data);
        }
      }).catch(() => {
        reject();
        vitisAlert.alert({
          type: 'error',
          text: this.translate.instant('ERROR_CALLING_GTF_API'),
        });
      });
    });
  }

  /**
   * récupère les formulaires d'un traitement GTF (JSON, JS, CSS) à partir de l'API GTF
   */
  public getGtfWorkspaceForm(workspace_id: number, layer_id: number) {
    if (this.auth.getPrivileges().includes('gtf_user')) {
      const url = `${environment.properties.app_api_url}/vmap/layers/${layer_id}/fme-userworkspaces/${workspace_id}/form/published`;
      return this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: url,
      });
    } else {
      vitisAlert.alert({
        type: 'error',
        text: this.translate.instant('REQUESTOR_WORKSPACES_MISSING_GTF_RIGHTS'),
      });
      return;
    }
  }

  /**
   * Affiche le formulaire permettant de lancer une demande GTF
   * @param component VmapGtfWorkspaceModalComponent
   * @param oWorkspace
   * @param oLayer
   * @param oEntity
   * @param oDataTableInstance
   * @param oDataGridOptions
   */
  public displayGTFWorkspaceModal(component, oWorkspace: GtfWorkspace, oLayer: any, oEntity?: any, oDataTableInstance?: any, oDataGridOptions?: any) {
    return new Promise((resolve, reject) => {
      this.getGtfWorkspaceForm(oWorkspace.workspace_id, oLayer.layer_id).then((resp: any) => {
        const openModal = this.modalService.open(component);
        openModal.componentInstance.JSONform = resp.data.data['json_form'];
        openModal.componentInstance.workspace = oWorkspace;
        openModal.componentInstance.layer = oLayer;
        openModal.componentInstance.entity = oEntity;
        if (oDataTableInstance) {
          openModal.componentInstance.dt = oDataTableInstance;
        }
        if (oDataGridOptions) {
          openModal.componentInstance.parentDataGridOptions = oDataGridOptions;
        }
        openModal.componentInstance.closeModal = () => {
          openModal.close();
        };
        setTimeout(() => {
          openModal.componentInstance.initForm('insert');
        });
        resolve(openModal);
      });
    });
  }

  /**
   * Retourne la taille (en pixel) du formulaire, en fonction de son identifiant de taille
   * @param  {number} formSizeId
   * @param  {string} unit
   */
  public getDecimalFormSize(formSizeId: number, unit: string = 'px'): number {
    let size: number = 0;
    switch (formSizeId) {
      case 0:
        size = 0;
        break;
      case 1:
        size = 350;
        break;
      case 2:
        size = 700;
        break;
      case 3:
        size = 1050;
        break;
      default:
        size = 0;
    }
    return size;
  }

  /**
   * Fonction récursive qui retrouve une couche dans l'arbre via son id
   * @param layerId
   * @param oNode oMapDef ou noeud dans l'arbre
   * @returns
   */
  private findLayerInMapDef(layerId: number, oNode: any): any {
    if (oNode &&
      oNode.layer_id) {
      if (oNode.layer_id === layerId) {
        return oNode;
      }
    } else if (oNode &&
      oNode.children &&
      this.vitisCheck.isArray(oNode.children)) {
      for (const children of oNode.children) {
        const layer = this.findLayerInMapDef(layerId, children);
        if (layer) {
          return layer;
        }
      }
    }
  }
}
