import { Component, OnInit, Input } from '@angular/core';
import { ActivatedRoute, Params, Router } from '@angular/router';
import { vitisAlert, VitisCheck, VitisClone, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';
import { TranslateService } from '@ngx-translate/core';
import { VmapLayer, } from 'src/app/vitis/map-core/services/map-models.model';
import { FormreaderService } from 'src/app/vitis/formreader/services/formreader.service';
import { Map } from 'ol';
import { Subject, Observable } from 'rxjs';

@Component({
  selector: 'app-layer-filter-request-form',
  templateUrl: './layer-filter-request-form.component.html',
  styleUrls: ['./layer-filter-request-form.component.scss']
})
export class LayerFilterRequestFormComponent implements OnInit {

  @Input() public oLayer: VmapLayer;
  @Input() public olMap: Map;

  public obsRunTest = new Subject();
  public oDragAndDropGridRows: Array<any> = [];
  public oDragAndDropGridRowsNotModified: Array<any> = [];
  public aLayersList: any;
  public aColumns: any;
  public placeholder: HTMLDivElement;
  public sDragElementID: string;
  public bRequestorLayer: boolean;
  public sFormType: string;

  private iDraggedElementIndex: number;
  private iOnDragElementIndex: number;

  // Liste des type disponibles pour filtre
  public aFilterColumnTypeList = [{
    key: 'string',
    label: 'STUDIO_COMPONENT_WEBSERVICE_GRID_TYPE_TEXT_LABEL'
  }, {
    key: 'number',
    label: 'STUDIO_COMPONENT_WEBSERVICE_GRID_TYPE_NUMBER_LABEL'
  }, {
    key: 'bool',
    label: 'STUDIO_COMPONENT_WEBSERVICE_GRID_TYPE_BOOLEAN_LABEL'
  }, {
    key: 'date',
    label: 'STUDIO_COMPONENT_WEBSERVICE_GRID_TYPE_DATE_LABEL'
  }];

  // Liste des type disponibles pour requêteur
  public aRequestorColumnTypeList = [{
    key: 'string',
    label: 'STUDIO_COMPONENT_WEBSERVICE_GRID_TYPE_TEXT_LABEL'
  }, {
    key: 'number',
    label: 'STUDIO_COMPONENT_WEBSERVICE_GRID_TYPE_NUMBER_LABEL'
  }, {
    key: 'date',
    label: 'STUDIO_COMPONENT_WEBSERVICE_GRID_TYPE_DATE_LABEL'
  }];

  public bddDatas = {
    layer_id: null,
    schema: null,
    database: null,
    table: null,
    datasource_id: null
  };

  /**
   * Constructeur
   */
  constructor(
    private route: ActivatedRoute,
    private formreaderSrvc: FormreaderService,
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private translate: TranslateService,
    private router: Router,
    private vitisClone: VitisClone,
  ) {

  }

  /**
   * Chargement du composant
   */
  ngOnInit() {
    this.route.data.subscribe(oMode => {
      if (oMode && oMode.options && oMode.options.type) {
        // Type de formulaire (requêteur ou filtre)
        this.sFormType = oMode.options.type;
        if (this.sFormType !== 'requestor' && this.sFormType !== 'filter') {
          console.error('type not valid, must be requestor or filter', this.sFormType);
          return;
        }

        this.route.params.subscribe(oParams => {
          if (this.vitisCheck.isDefAndNotNull(oParams) && this.vitisCheck.isDefAndNotNull(oParams.id)) {
            this.getParams(oParams.id).then((bValid: boolean) => {
              if (this.vitisCheck.isDefAndNotNull(bValid) && bValid == true) {
                this.getColumns();
              } else {
                console.error('Cannot get Parameter');
                vitisAlert.alert({
                  text: 'Impossible de récupérer les données de paramètre',
                  type: 'error',
                });
              }
            });
          } else {
            vitisAlert.alert({
              type: 'error',
              title: 'Error'
            });
          }
        });
      }
    });
  }

  /**
   * Vérifie si le formulaire a été modifié
   */
  canDeactivate(): Observable<boolean> | Promise<boolean> | boolean {
    const bIsModified: Boolean = this.bReturnModified(this.oDragAndDropGridRows, this.oDragAndDropGridRowsNotModified, this.bRequestorLayer);
    if (bIsModified) {
      return new Promise((resolve) => {
        this.translate.get(['FORM_MODIFIED_WARNING_TITLE',
          'FORM_MODIFIED_WARNING',
          'CONFIRM_CANCEL',
          'CONFIRM_CONFIRM']
        ).subscribe((translations) => {
          vitisAlert.alert({
            title: translations['FORM_MODIFIED_WARNING_TITLE'],
            text: translations['FORM_MODIFIED_WARNING'],
            type: 'warning',
            showCancelButton: true,
            confirmButtonText: translations['CONFIRM_CONFIRM'],
            cancelButtonText: translations['CONFIRM_CANCEL'],
          }).then((result: any) => {
            if (result.isConfirmed) {
              resolve(result.value);
            }
          });
        });
      });
    }
    return true;
  }

  /**
   * Fonction déclenchée lors d'un début de drag and drop sur un item
   * @param event
   */
  public onDragStart(event, draggedElement) {
    // Tout dans le setTimeout (sinon bug après scroll sur chrome).
    setTimeout(() => {
      this.iDraggedElementIndex = this.oDragAndDropGridRows.indexOf(draggedElement);

      // On récupère la div element-item en cours de drag
      const draggingElement = event.target;

      // On lui applique quelques styles pour qu'il ne soit plus visible pendant le drag
      draggingElement.style.position = 'absolute';
      draggingElement.style.zIndex = -10;

      // On récupère tout le groupe de l'élément
      const element = draggingElement.parentNode;
      this.sDragElementID = element.id;

      // On crée la div de placeholder


      this.placeholder = document.createElement('div');
      this.placeholder.classList.add('placeholder');

      // Insertion du placeholder
      element.parentNode.insertBefore(
        this.placeholder,
        element.nextSibling
      );
    });
  }

  /**
   * Fonction déclenchée lors du survol d'un élément en cours de drag and drop
   * @param event
   */
  public onDragOver(event, onDragElement) {

    let focusedElement: any;
    const element = event.target.closest('.element');
    // Drag au dessus d'un element
    if (this.vitisCheck.isDefAndNotNull(element)) {
      focusedElement = element;
    }

    const { bottom, top } = focusedElement.getBoundingClientRect();
    const elHeigth = bottom - top;
    const offsetY = bottom - event.clientY;

    // Par en bas
    if (offsetY < elHeigth / 2) {
      focusedElement.parentNode.insertBefore(
        this.placeholder,
        focusedElement
      );
      this.iOnDragElementIndex = this.oDragAndDropGridRows.indexOf(onDragElement);
    } else if (offsetY > elHeigth / 2) {
      // Par en haut
      focusedElement.parentNode.insertBefore(
        this.placeholder,
        focusedElement.nextSibling
      );
      this.iOnDragElementIndex = this.oDragAndDropGridRows.indexOf(onDragElement);
    }
    if (this.vitisCheck.isDefAndNotNull(this.iOnDragElementIndex) && this.vitisCheck.isDefAndNotNull(this.iDraggedElementIndex)) {
      this.changeArrayIndex(this.oDragAndDropGridRows, this.iDraggedElementIndex, this.iOnDragElementIndex);
      this.iDraggedElementIndex = this.iOnDragElementIndex;
    }
  }

  /**
   * Fonction déclenchée lors du drop d'un élément
   * @param event
   */
  public onDragEnd(event, id) {
    const draggingElement = event.target;
    draggingElement.style.removeProperty('position');
    draggingElement.style.removeProperty('z-index');
    // On met l'élément que l'on déplacait à la place du placeholder
    if (this.placeholder.parentNode) {
      this.placeholder.parentNode.replaceChild(draggingElement.parentNode, this.placeholder);
    }
  }

  /**
   * Fonction pour récupèrer les paramètres pour la requête getColumns() si json_columns  n'est pas nul alors on initialise notre objet oDragAndDropGridRows
   */
  public getParams(layer_id): Promise<boolean> {
    return new Promise((resolve, reject) => {
      this.bddDatas.layer_id = layer_id;
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: `${environment.properties.app_api_url}/vmap/layers/${this.bddDatas.layer_id}`,
      }).then((resp) => {
        if (resp.data) {

          if (resp.data.status === 1) {
            this.bddDatas.database = resp.data.data.database;
            this.bddDatas.schema = resp.data.data.schema;
            this.bddDatas.table = resp.data.data.pg_table;
            this.bddDatas.datasource_id = resp.data.data.datasource_id;
            this.bRequestorLayer = resp.data.data.is_requestor_layer;

            // Utilisation de table spéciale pour le requeteur
            if (this.sFormType === 'requestor' && resp.data.data.is_specific_table_requestor == true) {
              this.bddDatas.table = resp.data.data.table_requestor;
              this.bddDatas.schema = resp.data.data.schema_requestor;
            }

            if (this.bRequestorLayer === true) {
              if (resp.data.data.json_columns && resp.data.data.json_columns.length > 0) {
                this.oDragAndDropGridRows = [...this.parseJSONColumns(resp.data.data.json_columns)];
                this.oDragAndDropGridRowsNotModified = [...this.parseJSONColumns(resp.data.data.json_columns)];
              }
            } else {
              this.translate.get('COMPONENT_MAP_FILTER_ERROR_REQUESTOR').subscribe((sTranslate: string) => {
                vitisAlert.alert({
                  type: 'error',
                  text: sTranslate,
                });
              });
              reject(false);
            }
            resolve(true);
          } else {
            reject(false);
          }
        }
      });
    });
  }

  /**
   * Fonction pour récupérer les noms des colonnes via une requête si l'objet contenant json_columns est null alors on l'initialise
   * On n'initialise pas json_columns, il sera initialisé que si l'utilisateur sauvegarde
   */
  public getColumns(): Promise<boolean> {
    return new Promise((resolve, reject) => {
      const { layer_id, ...params } = this.bddDatas;
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: `${environment.properties.app_api_url}/vmap/pglayers/columns`,
        params: params,
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status === 1) {
            if (this.vitisCheck.isArray(resp.data.data) && resp.data.data.length > 0) {
              this.aColumns = resp.data.data;
              // On verifie que notre objet oDragAndDropGridRows contient bien les élements de json_columns si non on l'initialise
              if (this.vitisCheck.isDefAndNotNull(this.oDragAndDropGridRows) && this.oDragAndDropGridRows.length > 0) {

                // Si le nombre de champs dans la table est différent on réinitialise la liste en fusionnant les données.
                // Utile dans le cas où une table spécifique a été choisi pour le requeteur et dans le cas d'une modification des colonnes de la table
                if (this.oDragAndDropGridRows.length !== this.aColumns.length) {
                  if (this.bRequestorLayer === true) {
                    const preConfirm: any = [];
                    let iFindIndex: number;
                    let iCurrentColumnsIndex: number = 0;
                    const aOtherColumns: Array<any> = this.vitisClone.clone(this.aColumns);
                    // Ajoute en premier les colonnes qui étaient présente (on vérifie d'abord si elle existe bien dans le nouveau aColumns)
                    for (const key in this.oDragAndDropGridRows) {
                      if (Object.prototype.hasOwnProperty.call(this.oDragAndDropGridRows, key)) {
                        // this.oDragAndDropGridRows[key];
                        iFindIndex = this.aColumns.findIndex((column: any) => column.column_name == this.oDragAndDropGridRows[key].column);

                        if (iFindIndex !== -1) {
                          delete aOtherColumns[iFindIndex];
                          preConfirm.push({
                            requestor: this.oDragAndDropGridRows[key].requestor,
                            filter: this.oDragAndDropGridRows[key].filter,
                            column: this.oDragAndDropGridRows[key].column,
                            name: this.oDragAndDropGridRows[key].name,
                            index: iCurrentColumnsIndex,
                            requestor_type: this.oDragAndDropGridRows[key].requestor_type,
                            filter_type: this.oDragAndDropGridRows[key].requestor_type,
                            request: false,
                          });
                          iCurrentColumnsIndex++;

                        }
                      }
                    }

                    // Ajoute les colonnes restantes, celles qui n'étaient pas présente avant.
                    for (const key in aOtherColumns) {
                      if (Object.prototype.hasOwnProperty.call(aOtherColumns, key)) {
                        preConfirm.push({
                          requestor: false,
                          filter: false,
                          column: aOtherColumns[key].column_name,
                          name: aOtherColumns[key].column_name,
                          index: iCurrentColumnsIndex,
                          requestor_type: aOtherColumns[key].requestor_type,
                          filter_type: aOtherColumns[key].requestor_type,
                          request: false,
                        });
                        iCurrentColumnsIndex++;

                      }
                    }
                    this.oDragAndDropGridRows = preConfirm;
                  }
                }
              } else {
                if (this.bRequestorLayer === true) {
                  const preConfirm: any = [];
                  for (let i = 0; i < this.aColumns.length; i++) {
                    preConfirm.push({
                      requestor: false,
                      filter: false,
                      column: this.aColumns[i].column_name,
                      name: this.aColumns[i].column_name,
                      index: i,
                      requestor_type: this.aColumns[i].requestor_type,
                      filter_type: this.aColumns[i].requestor_type,
                      request: false,
                    });
                  }
                  this.oDragAndDropGridRows = preConfirm;

                }
              }
              resolve(true);
            }
          }
        }
        reject(false);
      });
    });
  }

  /**
   * Modifie notre tableau pour mettre à jour le déplacement des éléments via le drag and drop
   */
  private changeArrayIndex(array, fromIndex, toIndex) {
    const element = array[fromIndex];
    array.splice(fromIndex, 1);
    array.splice(toIndex, 0, element);
  }

  /**
   * Envoie la requete pour update
   */
  public updateModification(): void {
    // Veriffie que les nouvelles données soient valides
    this.verificationLabel().then((bValid: boolean) => {
      if (this.vitisCheck.isDefAndNotNull(bValid) && bValid == true && this.bRequestorLayer === true) {
        return new Promise((resolve, reject) => {
          this.vitisRequest.ajaxRequestPromise({
            method: 'PUT',
            url: `${environment.properties.app_api_url}/vmap/layers/${this.bddDatas.layer_id}`,
            data: this.formreaderSrvc.transformToDataForm({ json_columns: JSON.stringify(this.oDragAndDropGridRows)})
          }).then((resp) => {
            if (resp.data) {
              if (resp.data.status === 1) {
                if (this.vitisCheck.isDefAndNotNull(resp.data.data)) {
                  this.oDragAndDropGridRowsNotModified = this.oDragAndDropGridRows;
                  this.translate.get('FORM_VALIDATION_OK').subscribe((sTranslate: string) => {
                    vitisAlert.fireToast({
                      icon: 'success',
                      title: sTranslate
                    });
                  });
                  resolve(true);
                }
              }
            } else {
              reject(false);
            }
          });
        });
      } else {
        return false;
      }
    });
  }

  /**
   * retourne à la liste des couches
   */
  public returnToLayer() {
    this.router.navigate(['../../../'], {
      relativeTo: this.route.parent
    });
  }

  /**
   * Verrifie si les deux objets sont égaux
   */
  public bReturnModified(oObjecta, oObjectb, brequestor): boolean {
    let bIsModified: boolean = false;
    if (brequestor === true) {
      if (oObjectb.length > 0) {
        for (let i = 0; i < oObjecta.length; i++) {
          if ((oObjecta[i].name !== oObjectb[i].name) || (oObjecta[i].filter_type !== oObjectb[i].filter_type) || (oObjecta[i].requestor_type !== oObjectb[i].requestor_type) || (oObjecta[i].filter !== oObjectb[i].filter) || (oObjecta[i].request !== oObjectb[i].request)) {
            bIsModified = true;
            break;
          }
        }
      } else {
        bIsModified = true;
      }
    }
    return bIsModified;
  }

  /**
   * Verifie que le label n'est pas vide ainsi que le type etc
   */
  public verificationLabel(): Promise<boolean> {
    return new Promise((resolve, reject) => {
      for (let i = 0; i <= (this.oDragAndDropGridRows.length) - 1; i++) {
        if (this.vitisCheck.isDefAndNotNull(this.oDragAndDropGridRows[i].name) &&
        this.oDragAndDropGridRows[i].name != '' &&
        this.vitisCheck.isDefAndNotNull(this.oDragAndDropGridRows[i].filter) &&
        this.vitisCheck.isDefAndNotNull(this.oDragAndDropGridRows[i].requestor) &&
        (this.oDragAndDropGridRows[i].filter === false || this.vitisCheck.isDefAndNotNull(this.oDragAndDropGridRows[i].filter_type)) &&
        (this.oDragAndDropGridRows[i].requestor === false || this.vitisCheck.isDefAndNotNull(this.oDragAndDropGridRows[i].requestor_type))
        ) {
        } else if (this.oDragAndDropGridRows[i].name === '') {
          this.translate.get('COMPONENT_MAP_FILTER_ERROR_LABEL').subscribe((translation) => {
            vitisAlert.alert({
              type: 'error',
              text: translation,
            });
          });
          reject(false);
        } else {
          this.translate.get('COMPONENT_MAP_FILTER_ERROR_VERIFICATION').subscribe((translation) => {
            vitisAlert.alert({
              type: 'error',
              text: translation,
            });
          });
          reject(false);
        }
      }
      resolve(true);
    });
  }

  /**
   * Traite la colonne json_columns
   * @param sJsonColumns
   */
  private parseJSONColumns(sJsonColumns: string): Array<any> {
    const aJsonColumns = JSON.parse(sJsonColumns);
    const aMandatoryOptions = ['column', 'name'];
    const oDefaultColumn = {
      column: '',
      name: '',
      index: 0,
      filter: false,
      requestor: false,
      filter_type: 'string',
      requestor_type: 'string',
    };

    for (let i = 0; i < aJsonColumns.length; i++) {
      const oColumn = aJsonColumns[i];

      // Vérifie les valeurs obligatoires
      aMandatoryOptions.forEach(option => {
        if (!this.vitisCheck.isDefAndNotNull(oColumn[option])) {
          console.error(`column not valid : missing ${option}`, oColumn);
        }
      });

      // Vérifie les valeurs non obligatoires
      for (const key in oDefaultColumn) {
        if (Object.prototype.hasOwnProperty.call(oDefaultColumn, key)) {
          if (!this.vitisCheck.isDefAndNotNull(oColumn[key])) {
            oColumn[key] = oDefaultColumn[key];
          }
        }
      }
    }

    return aJsonColumns;
  }
}
