import { Injectable, EventEmitter } from '@angular/core';
import { VitisRequest, VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';
import { BehaviorSubject, Observable, Subject } from 'rxjs';
import { StyleSubFields } from 'src/modules/vmap/components/models/json-symbology/json-symbology-style.model';
import { LayerSymbologieSubType } from '../../../models/layer-symbology.model';

@Injectable({
  providedIn: 'root'
})
export class LayerSymbologyService {

  public columnsAvancedInput = new BehaviorSubject([]);
  public attributName: string;

  public symbols = new BehaviorSubject([]);
  public fonts = new BehaviorSubject([]);
  public positions = new BehaviorSubject({});
  public aligns = new BehaviorSubject({});
  public types = new BehaviorSubject({});
  public forces = new BehaviorSubject({});
  public lineCap = new BehaviorSubject({});
  public linejoin = new BehaviorSubject({});
  public obsRunTest = new Subject(); // Appelé pour lancer le test sur la couche
  public regions = new BehaviorSubject({});
  public sizeunits = new BehaviorSubject({});
  public units = new BehaviorSubject({});

  public bddDatas = {
    layer_id: null,
    schema: null,
    database: null,
    table: null,
    datasource_id: null
  };

  /**
   * Constructeur
   */
  constructor(
    private vitisRequest: VitisRequest,
    private vitisCheck: VitisCheck,
  ) {
    this.getSymbols();
    this.getFonts();
    this.getPositions();
    this.getAligns();
    this.getTypes();
    this.getForces();
    this.getLineCap();
    this.getLinejoin();
    this.getRegions();
    this.getSizeUnits();
    this.getUnits();
  }

  /**
   * Requête pour répupérer les colonnes de la table
   * @returns
   */
  public getColumns(isWithCluster: boolean = false): Promise<boolean> {
    const { layer_id, ...params } = this.bddDatas;
    return new Promise((resolve, reject) => {
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: `${environment.properties.app_api_url}/vmap/pglayers/columns`,
        params,
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status === 1) {
            if (this.vitisCheck.isArray(resp.data.data) && resp.data.data.length > 0) {
              let value = resp.data.data;
              if (isWithCluster) {
                value = [
                  ...resp.data.data,
                  ...[{ column_name: 'Cluster_FeatureCount', data_type: 'integer' }]
                ];
              }
              this.columnsAvancedInput.next(value);
              resolve(true);
            }
          }
        }
        reject(false);
      });
    });
  }

  /**
   * Requête pour récupérer les symboles dusponibles
   * @returns
   */
  public getSymbols(): Promise<boolean> {
    return new Promise((resolve, reject) => {
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: environment.properties.app_api_url + '/vmap/symbols',
        params: {
          attributs: 'name|label'
        }
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status === 1) {
            if (resp.data.data) {
              const arr = resp.data.data;
              const formatOptions = arr.reduce((acc, curr) => {
                return { ...acc, [curr.name]: curr.label || curr.name };
              }, {});
              this.symbols.next(formatOptions);
              resolve(true);
            }
          }
        }
        reject(false);
      });
    });
  }

  /**
   * Requête pour récupérer les polices dusponibles
   * @returns
   */
  public getFonts(): Promise<boolean> {
    return new Promise((resolve, reject) => {
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: environment.properties.app_api_url + '/vmap/fonts',
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status === 1) {
            if (resp.data.data) {
              const arr = resp.data.data;
              const formatOptions = arr.reduce((acc, curr) => {
                return { ...acc, [curr.name]: curr.name };
              }, {});
              this.fonts.next(formatOptions);
              resolve(true);
            }
          }
        }
        reject(false);
      });
    });
  }

  /**
   * Retourne les positions disponibles
   * @returns
   */
  public getPositions(): Promise<boolean> {
    return new Promise((resolve, reject) => {

      this.positions.next({
        ul: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_UL',
        uc: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_UC',
        ur: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_UR',
        cl: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_CL',
        cc: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_CC',
        cr: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_CR',
        ll: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_LL',
        lc: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_LC',
        lr: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_LR',
        auto: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_POSITION_AUTO'
      });
      resolve(true);
    });
  }

  /**
   * Retourne les alignements disponibles
   * @returns
   */
  public getAligns(): Promise<boolean> {
    return new Promise((resolve, reject) => {

      this.aligns.next({
        left: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_ALIGN_LEFT',
        center: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_ALIGN_CENTER',
        right: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_ALIGN_RIGHT'
      });
      resolve(true);
    });
  }

  /**
   * Retourne les types disponibles
   * @returns
   */
  public getTypes(): Promise<boolean> {
    return new Promise((resolve, reject) => {
      this.types.next({
        bitmap: 'bitmap',
        truetype: 'truetype',
      });
      resolve(true);
    });
  }

  /**
   * Retourne les forces disponibles
   * @returns
   */
  public getForces(): Promise<boolean> {
    return new Promise((resolve, reject) => {
      this.forces.next({
        true: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_FORCE_TRUE',
        false: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_FORCE_FALSE',
        group: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABEL_FORCE_GROUP'
      });
      resolve(true);
    });
  }

  /** Retourne les types de cap disponible */
  public getLineCap(): Promise<boolean> {
    return new Promise((resolve, reject) => {
      this.lineCap.next({
        BUTT: 'COMPONENT_MAP_EDIT_SYMBOLOGY_BUTT',
        ROUND: 'COMPONENT_MAP_EDIT_SYMBOLOGY_ROUND',
        SQUARE: 'COMPONENT_MAP_EDIT_SYMBOLOGY_SQUARE'
      });
      resolve(true);
    });
  }

  /** Retourne les types de linejoin disponible */
  public getLinejoin(): Promise<boolean> {
    return new Promise((resolve, reject) => {
      this.linejoin.next({
        ROUND: 'COMPONENT_MAP_EDIT_SYMBOLOGY_ROUND',
        MITER: 'COMPONENT_MAP_EDIT_SYMBOLOGY_MITER',
        BEVEL: 'COMPONENT_MAP_EDIT_SYMBOLOGY_BEVEL',
        NONE: 'COMPONENT_MAP_EDIT_SYMBOLOGY_NONE',
      });
      resolve(true);
    });
  }

  /**
   * Retourne les regions disponibles pour les cluster
   * @returns
   */
  public getRegions(): Promise<boolean> {
    return new Promise((resolve, reject) => {
      this.regions.next({
        ellipse: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLUSTER_ELLIPSE',
        rectangle: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLUSTER_RECTANGLE'
      });
      resolve(true);
    });
  }

  /**
   * Retourne les unités disponible pour la couche
   * @returns
   */
  public getSizeUnits(): Promise<boolean> {
    return new Promise((resolve, reject) => {
      this.sizeunits.next({
        feet: 'COMPONENT_MAP_EDIT_SYMBOLOGY_SIZEUNITS_FEET',
        inches: 'COMPONENT_MAP_EDIT_SYMBOLOGY_SIZEUNITS_INCHES',
        kilometers: 'COMPONENT_MAP_EDIT_SYMBOLOGY_SIZEUNITS_KILOMETERS',
        meters: 'COMPONENT_MAP_EDIT_SYMBOLOGY_SIZEUNITS_METERS',
        miles: 'COMPONENT_MAP_EDIT_SYMBOLOGY_SIZEUNITS_MILES',
        nauticalmiles: 'COMPONENT_MAP_EDIT_SYMBOLOGY_SIZEUNITS_NAUTICALMILES',
        pixels: 'COMPONENT_MAP_EDIT_SYMBOLOGY_SIZEUNITS_PIXELS'
      });
      resolve(true);
    });
  }

  /**
   * Retourne les unités disponible pour la couche
   * @returns
   */
  public getUnits(): Promise<boolean> {
    return new Promise((resolve, reject) => {
      this.units.next({
        dd: 'COMPONENT_MAP_EDIT_SYMBOLOGY_UNITS_DD',
        feet: 'COMPONENT_MAP_EDIT_SYMBOLOGY_UNITS_FEET',
        inches: 'COMPONENT_MAP_EDIT_SYMBOLOGY_UNITS_INCHES',
        kilometers: 'COMPONENT_MAP_EDIT_SYMBOLOGY_UNITS_KILOMETERS',
        meters: 'COMPONENT_MAP_EDIT_SYMBOLOGY_UNITS_METERS',
        miles: 'COMPONENT_MAP_EDIT_SYMBOLOGY_UNITS_MILES',
        nauticalmiles: 'COMPONENT_MAP_EDIT_SYMBOLOGY_UNITS_NAUTICALMILES',
        percentages: 'COMPONENT_MAP_EDIT_SYMBOLOGY_UNITS_PERCENTAGES',
        pixels: 'COMPONENT_MAP_EDIT_SYMBOLOGY_UNITS_PIXELS'
      });
      resolve(true);
    });
  }

  /**
   * Traduit une expression en objet
   * @param expression
   * @returns
   */
  public expressionTaductionArrToString(expression): string {
    return (typeof expression == 'object' && !this.vitisCheck.isEmptyObject(expression)) ? expression.reduce((acc, curr) =>
      `${acc} [${curr.table}] ${curr.operator} ${curr.value} ${curr.sql_operator || ''}`, '') : '';
  }

  /**
   * Retourne un pourcentage arrondi
   * @param iVal
   * @returns
   */
  public getPrettyPercent(iVal: number): string {
    return `${iVal ? Math.round(iVal * 100) : 100}%`;
  }

  /**
   * Retourne l'observable obsRunTest
   * @returns
   */
  public getObsRunTest() {
    return this.obsRunTest;
  }

  /**
   * Lance le test de couche en envoyant une nouvelle valeur sur l'observable obsRunTest
   * @param oJsonSymbologyExpression
   */
  public runTest(oJsonSymbologyExpression: any) {
    this.obsRunTest.next(oJsonSymbologyExpression);
  }

  /**
   * Récupère la liste des types qui peuvent être ajouté
   */
  public getAvailableSubTypeList(
    currGeomType: string,
    isLabel: boolean = false,
    isGlobal: boolean = false,
    currType?: string,
    isFile: boolean = false
  ): Array<string> {

    if (isGlobal) { // Paramètres de la couche
      const aAvailableSubType: Array<string> = [
        'visibility'
      ];
      if (isFile === false) {
        aAvailableSubType.push('opacity', 'units');
        if (currGeomType === 'point' && this.vitisCheck.isDefAndNotNull(currType) && currType == 'cluster') {
          aAvailableSubType.push('cluster');
        }
      }
      return aAvailableSubType;
    } else {
      if (!isLabel) {
        const aAvailableSubType: Array<string> = [
          'visibility',
          'opacity',
          'symbolOffset',
          'symbolSizeVariation',
        ];

        if (currGeomType === 'point') {
          aAvailableSubType.push('symbolOutline');
        }

        if (currGeomType === 'polygon') {
          aAvailableSubType.push('polygonSymbolRepeat');
          aAvailableSubType.push('polygonSymbolCenter');
        }

        if (currGeomType === 'line') {
          aAvailableSubType.push('lineDash');
          aAvailableSubType.push('lineSymbol');
        }

        return aAvailableSubType;
      } else {
        return [
          'visibility',
          'shadow',
          'shadowBackground',
          'wrap',
          'sizeAndPosition',
          'outline',
          'other',
        ];
      }
    }
  }

  /** Récupère la liste des tout les sous type disponible */
  public getSubType(): LayerSymbologieSubType {
    return {
      visibility: {
        label: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_VISIBILITY',
        icon: '',
        defaultDeletable: false
      },
      opacity: {
        label: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_OPACITY',
        icon: '',
        defaultDeletable: true
      },
      units: {
        label: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_UNITS',
        icon: '',
        defaultDeletable: true
      },
      cluster: {
        label: 'COMPONENT_MAP_EDIT_SYMBOLOGY_PARAMS',
        icon: '',
        defaultDeletable: false
      },
      symbolOffset: {
        label: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_SYMBOL_OFFSET',
        icon: '',
        defaultDeletable: true
      },
      symbolSizeVariation: {
        label: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_SYMBOL_SIZE_VARIATION',
        icon: '',
        defaultDeletable: true
      },
      symbolOutline: {
        label: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_SYMBOL_OUTLINE',
        icon: '',
        defaultDeletable: true
      },
      polygonSymbolRepeat: {
        label: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_POLYGON_SYMBOL_REPEAT',
        icon: '',
        defaultDeletable: true
      },
      polygonSymbolCenter: {
        label: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_POLYGON_SYMBOL_CENTER',
        icon: '',
        defaultDeletable: true
      },
      lineDash: {
        label: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_LINE_DASH',
        icon: '',
        defaultDeletable: true
      },
      lineSymbol: {
        label: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_PARAMS_LINE_SYMBOL',
        icon: '',
        defaultDeletable: true
      },
      shadow: {
        label: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABELS_SHADOW',
        icon: '',
        defaultDeletable: true
      },
      shadowBackground: {
        label: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABELS_SHADOW_BACKGROUND',
        icon: '',
        defaultDeletable: true
      },
      wrap: {
        label: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABELS_WRAP_WORD',
        icon: '',
        defaultDeletable: true
      },
      sizeAndPosition: {
        label: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABELS_SAS',
        icon: '',
        defaultDeletable: true
      },
      outline: {
        label: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABELS_OUTLINE',
        icon: '',
        defaultDeletable: true
      },
      other: {
        label: 'COMPONENT_MAP_EDIT_SYMBOLOGY_CLASS_LABELS_OTHER',
        icon: '',
        defaultDeletable: true
      }
    };
  }

  /**
   * Set la valeur dans otherFields en fonction du chemin dans event
   */
  public setOtherFieldValue(event: StyleSubFields, object: any) {
    const sEventFirstKey = Object.keys(event)[0];
    const sEventSecondKey = Object.keys(event[sEventFirstKey])[0];
    const sEventValue = event[sEventFirstKey][sEventSecondKey];
    for (const key in object.otherFields) {
      if (Object.prototype.hasOwnProperty.call(object.otherFields, key)) {
        if (key === sEventFirstKey) {
          object.otherFields[key][sEventSecondKey] = sEventValue;
        }
      }
    }
    return object;
  }

  /**
   * Créér un géométrie par défaut
   * @param currGeomType
   */
  public setDefaultValue(currGeomType: string) {
    if (currGeomType == 'point') {
      return {
        symbol: 'circle',
        fillcolor: '#000000'
      };
    }
    if (currGeomType == 'line') {
      return {
        width: '10',
        fillcolor: '#000000'
      };
    }
    if (currGeomType == 'polygon') {
      return {
        width: '2',
        fillcolor: '#FF0000',
        outlineColor: '#000000'
      };
    }
    return {};
  }

  /**
   * Vérifie si les conditions sont remplis
   * @param key
   * @param oConditions
   */
  public isReadOnly(key: string, oConditions: any) {
    for (const k in oConditions) {
      if (Object.prototype.hasOwnProperty.call(oConditions, k)) {
        const condition = oConditions[k];
        if (k === key && condition) {
          return true;
        }
      }
    }
    return false;
  }
}
