<?php

namespace App\Module\Vmap\WebService\Controller;

use App\Module\Vitis\WebService\Controller\GenericController;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller pour générer le fichier epsg pour mapserver.
 */
class CoordsysController extends GenericController {
  /**
   * Met à jour une entité pour une route de type PUT Module/Objet/{{id}}.
   * @param mixed $id id de l'entité à trouver
   * @return Response réponse à retourner par l'API
   */
  public function updateData($id): Response {
    $sData = $this->requestToJson();
    // Sauvegarde en base de donnée
    $oEntity = $this->updateDataInDb($id, $sData);
    if ($oEntity['status'] == 0) {
      return $this->formatGenericResponseForErrorOnUpdate();
    }
    // Création du fichier EPSG pour mapserver
    $this->generateEpsgFile();
    //
    return $this->formatResponse($oEntity);
  }

  /**
   * Créé une entité pour une route de type POST Module/Objet.
   * @return Response réponse à retourner par l'API
   */
  public function postData(): Response {
    $sData = $this->requestToJson();
    // Sauvegarde en base de donnée
    $oEntity = $this->insertDataInDb($sData);

    if ($oEntity == null) {
      return $this->formatGenericResponseForErrorOnInsert();
    }

    if ($oEntity['status'] == 0) {
      return $this->formatGenericResponseForErrorOnInsert();
    }
    // Création du fichier EPSG pour mapserver
    $this->generateEpsgFile();

    return $this->formatResponse($oEntity);
  }

  /**
   * Supprime une entité pour une route de type POST Module/Objet/{{id}}.
   * @param mixed $id id de l'entité à supprimer
   * @return Response réponse à retourner par l'API
   */
  public function deleteData($id): Response {
    // Sauvegarde en base de donnée
    $oEntity = $this->deleteDataInDb($id);
    if ($oEntity['status'] == 0) {
      return $this->formatGenericResponseForErrorOnDelete();
    }
    // Création du fichier EPSG pour mapserver
    $this->generateEpsgFile();

    return $this->formatResponse($oEntity);
  }

  /**
   * Supprime les entités d'une liste pour une route de type POST Module/Objet.
   * @return Response réponse à retourner par l'API
   */
  public function deleteDatas(): Response {
    $aData = $this->massiveRequestGetIdList();
    if (empty($aData)) {
      return $this->formatGenericResponseForMissingIdList();
    }
    // Sauvegarde en base de donnée
    $aEntities = $this->deleteDataInDb($aData['idList']);
    if ($aEntities['status'] == 0) {
      return $this->formatGenericResponseForErrorOnDelete();
    }
    // Création du fichier EPSG pour mapserver
    $this->generateEpsgFile();

    return $this->formatResponse($aEntities);
  }

  /**
   * Génération du fichier EPSG pour mapserver.
   * @return void
   */
  private function generateEpsgFile(): void {
    // Récupère tous les systèmes de coordonnées.
    $qb = $this->oDatabaseConnection->getQueryAccessor()->createQueryBuilder();
    $qb->select('coordsys_id, label, epsg_definition')
      ->from('[SCHEMA].coordsys', 'c')
      ->orderBy('coordsys_id', 'ASC');
    $aParams = [
      'SCHEMA' => ['value' => $this->oProperties->get('schema_vmap'), 'type' => 'schema_name'],
    ];
    try {
      $aAllCoordsys = $qb->executeWithParams($aParams)->fetchAll();
    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('Error getting coordinate systems: ' . $e->getMessage());
    }

    // Génération du fichier epsg.
    try {
      if (!empty($aAllCoordsys)) {
        $sEpsgFile = '';
        // Template pour les systèmes de coordonnées.
        $sCoordsysTemplate = '# {{label}}' . chr(10) . '<{{coordsys_id}}> {{epsg_definition}} <>' . chr(10);
        foreach($aAllCoordsys as $aCoordsys) {
          $sEpsgFile .= $sCoordsysTemplate;
          foreach($aCoordsys as $key => $value) {
            $sEpsgFile = str_replace('{{' . $key . '}}', $value, $sEpsgFile);
          }
        }
      }
      // Suppression du fichier epsg.
      $oLocalFileAccessor = $this->getLocalFileAccessor();
      $sEpsgFilePath = $this->oProperties->get('vmap_ms_dir') . '/proj/epsg';
      if ($oLocalFileAccessor->file_exists($sEpsgFilePath)) {
        $oLocalFileAccessor->unlink($sEpsgFilePath);
      }
      // Ecriture du fichier epsg.
      $oLocalFileAccessor->file_put_contents($sEpsgFilePath, $sEpsgFile);
    } catch (\Exception $e) {
      $this->oLoggerManager->get('API_ERROR')->error('Error during epsg file generation : ' . $e->getMessage());
    }
  }
}
