<?php

namespace App\Module\Vmap\WebService\Controller;

use App\Module\Vitis\WebService\Controller\GenericController;
use App\Module\Vitis\WebService\Repository\GenericRepository;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\Normalizer\VitisNormalizer;
use App\Module\Vitis\WebService\Service\Serializer\VitisSerializer;
use App\Module\Vitis\WebService\Service\VitisRequest;
use App\Module\Vmap\WebService\Entity\AbstractLayer;
use App\Module\Vmap\WebService\Entity\LayerFile;
use App\Module\Vmap\WebService\Entity\LayerOSM;
use App\Module\Vmap\WebService\Entity\LayerPostgres;
use App\Module\Vmap\WebService\Entity\LayerVectorTiles;
use App\Module\Vmap\WebService\Entity\LayerWMS;
use App\Module\Vmap\WebService\Entity\LayerWMTS;
use App\Module\Vmap\WebService\Entity\LayerXYZ;
use App\Module\Vmap\WebService\Repository\LayerRepository;
use App\Module\Vmap\WebService\Service\LayerService;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Serializer\Normalizer\AbstractNormalizer;

/**
 * Controller pour le datasource treeview de vMap.
 *
 * @author Ludivine Premont <ludivine.premont@veremes.com>
 */
class LayerController extends GenericController {

  /**
   * Créé plusieurs entités pour une route de type POST Module/Objet.
   * @return Response réponse à retourner par l'API
   */
  public function postDatas(): Response {
    $aData = $this->requestToArray();
    if (empty($aData)) {
      $aReturn = ['status' => 0];
      $aReturn['message'] = 'data parameter is missing';
      $aReturn['errorMessage'] = 'ERROR_MISSING_DATA_PARAMETER';
      return $this->formatResponse($aReturn, Response::HTTP_NOT_ACCEPTABLE);
    }

    // chargement de la connexion à la base de donnée
    $oDataAccessor = $this->getDataAccessor();
    // Sauvegarde en base de donnée
    $oEntity = $oDataAccessor->insertEntities($aData['data'], true); // Uniquement en ORM
    if ($oEntity['status'] == 0) {
      $oEntity['message'] = 'Error during update';
      $oEntity['errorMessage'] = 'ERROR_GENERIC_CONTROLLER_INSERT';
      return $this->formatResponse($oEntity, Response::HTTP_UNAUTHORIZED);
    }
    $this->formatResponse($oEntity);

    return $this->oResponse;
  }

  /**
   * Récupère une entité AbstractLayer
   * @return Response réponse à retourner par l'API
   */
  public function getData($id): Response {
    $aData = $this->requestToArray();

    // Attribution de l'entité oLayer grâce à son layertype
    $sClassName = $this->getLayerClassForId($id);

    if (empty($sClassName)) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot find layer with id : ' . $id);
      $oResponse = [
        'status' => 0,
        'message' => 'Error during get',
        'errorMessage' => 'ERROR_GENERIC_CONTROLLER_GET',
      ];
      return $this->formatResponse($oResponse, Response::HTTP_UNAUTHORIZED);
    }

    // surcharge de la classe initiale pour récupérer les information de l'entité
    $parameters = $this->getRequestParameters();
    $oDataAccessor = $this->getDataAccessor();
    $parameters['options']['entity'] = $sClassName;
    $oDataAccessor->setOptions($parameters['options']);

    $oEntity = $oDataAccessor->getEntity($id, $aData);

    if ($oEntity['status'] == 0) {
      return $this->formatGenericResponseForErrorOnSelect();
    }
    $aSerializerContext = $this->getSerializerContext();

    return $this->formatResponse($oEntity, Response::HTTP_OK, $aSerializerContext);
  }

  /**
   * Met à jour une entité AbstractLayer
   * @return Response réponse à retourner par l'API
   */
  public function updateData($id): Response {
    $sData = $this->requestToJson();

    // Récupération du repository parent AbstractLayer
    $oLayerRepository = $this->oDatabaseConnection->getAdminEntityManager()->getRepository(AbstractLayer::class);
    if (!$oLayerRepository instanceof LayerRepository) {
      throw new \Exception("Repository is not LayerRepository", 1);
    }

    $layertype = $layertype ?? $oLayerRepository->findLayerTypeWithId($id);

    $parameters = match ($layertype) {
      LayerPostgres::TYPE => $this->oRouteConfig->getRequestParameters('vmap_put_datasourcetreeview_pglayers_v2'),
      LayerFile::TYPE => $this->oRouteConfig->getRequestParameters('vmap_put_datasourcetreeview_filelayers_v2'),
      LayerXYZ::TYPE => $this->oRouteConfig->getRequestParameters('vmap_put_datasourcetreeview_xyzlayers_v2'),
      LayerOSM::TYPE => $this->oRouteConfig->getRequestParameters('vmap_put_datasourcetreeview_osmlayers_v2'),
      LayerWMS::TYPE => $this->oRouteConfig->getRequestParameters('vmap_put_datasourcetreeview_wmslayers_v2'),
      LayerWMTS::TYPE => $this->oRouteConfig->getRequestParameters('vmap_put_datasourcetreeview_wmtslayers_v2'),
      LayerVectorTiles::TYPE => $this->oRouteConfig->getRequestParameters('vmap_put_datasourcetreeview_vectortileslayers_v2'),
      default => $this->oRouteConfig->getRequestParameters()
    };

    $oDataAccessor = $this->getDataAccessor();
    $oDataAccessor->setOptions($parameters['options']);
    $aReturn = $this->updateDataInDb($id, $sData, $oDataAccessor);

    if ($aReturn['status'] == 0) {
      return $this->formatGenericResponseForErrorOnUpdate();
    }

    $aReturn['data']->setDatemaj(new \DateTime());

    // Sauvegarde en base de donnée
    $this->oDatabaseConnection->getUserEntityManager()->persist($aReturn['data']);
    $this->oDatabaseConnection->getUserEntityManager()->flush();

    // rafraichissement en cas de modifications interne à la base
    $this->oDatabaseConnection->getUserEntityManager()->refresh($aReturn['data']);

    return $this->formatResponse($aReturn, Response::HTTP_OK);
  }

  /**
   * Ajoute une entité AbstractLayer
   * @return Response réponse à retourner par l'API
   */
  public function postData(): Response {
    $sData = $this->requestToJson();
    $layertype = $this->requestToArray()['layertype'];
    if (empty($layertype)) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot find layertype');
      $oResponse = [
        'status' => 0,
        'message' => 'Error during post',
        'errorMessage' => 'ERROR_GENERIC_CONTROLLER_POST',
      ];
      return $this->formatResponse($oResponse, Response::HTTP_UNAUTHORIZED);
    }

    $parameters = match ($layertype) {
      LayerPostgres::TYPE => $this->oRouteConfig->getRequestParameters('vmap_post_datasourcetreeview_pglayers_v2'),
      LayerFile::TYPE => $this->oRouteConfig->getRequestParameters('vmap_post_datasourcetreeview_filelayers_v2'),
      LayerXYZ::TYPE => $this->oRouteConfig->getRequestParameters('vmap_post_datasourcetreeview_xyzlayers_v2'),
      LayerOSM::TYPE => $this->oRouteConfig->getRequestParameters('vmap_post_datasourcetreeview_osmlayers_v2'),
      LayerWMS::TYPE => $this->oRouteConfig->getRequestParameters('vmap_post_datasourcetreeview_wmslayers_v2'),
      LayerWMTS::TYPE => $this->oRouteConfig->getRequestParameters('vmap_post_datasourcetreeview_wmtslayers_v2'),
      LayerVectorTiles::TYPE => $this->oRouteConfig->getRequestParameters('vmap_post_datasourcetreeview_vectortileslayers_v2'),
      default => $this->oRouteConfig->getRequestParameters()
    };

    $oDataAccessor = $this->getDataAccessor();
    $oDataAccessor->setOptions($parameters['options']);
    $oEntity = $this->insertDataInDb($sData, $oDataAccessor);

    if ($oEntity == null) {
      return $this->formatGenericResponseForErrorOnInsert();
    }

    if ($oEntity['status'] == 0) {
      return $this->formatGenericResponseForErrorOnInsert();
    }

    return $this->formatResponse($oEntity);
  }

  /**
   * Retourne la classe pour un layer type donné
   *
   * @param  string      $layertype
   * @return string|null
   */
  private function getLayerClassForType(string $layertype) : ?string {
    return AbstractLayer::getLayerChildrenLayerClassName($layertype);
  }

  /**
   * Retourne la classe pour un identifiant de couche
   * 
   * @param  mixed       $id
   * @return string|null
   */
  private function getLayerClassForId(mixed $id) : ?string {
    // Récupération du repository parent AbstractLayer
    $oLayerRepository = $this->oDatabaseConnection->getAdminEntityManager()->getRepository(AbstractLayer::class);
    if (!$oLayerRepository instanceof LayerRepository) {
      throw new \Exception("Repository is not LayerRepository", 1);
    }

    $layertype = $layertype ?? $oLayerRepository->findLayerTypeWithId($id);
    return  $this->getLayerClassForType($layertype);
  }
}
