<?php

namespace App\Module\Vmap\WebService\Controller;

use App\Module\Vitis\WebService\Controller\GenericController;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\Token\JweTokenManager;
use App\Module\Vitis\WebService\Service\VitisHttpClient;
use App\Module\Vitis\WebService\Service\VitisRequest;
use App\Module\Vmap\WebService\Entity\LayerFmeWorkbench;
use Symfony\Component\HttpClient\Exception\ServerException;
use Symfony\Component\HttpClient\Exception\TransportException;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller pour la récupération des informations relatives aux workbench FME liés aux couches
 *
 * @author Kyllian Beasse <kyllian.beasse@veremes.com>
 */
class LayerFmeWorkbenchController extends GenericController {
  /**
   * @var array structure de filtre pour l'jaout de demande
   */
  private const MAIN_FILTER = [
    'relation' => 'AND',
    'operators' => [],
  ];
  private const LAYER_FILTER = [
    'column' => 'layer_id',
    'compare_operator' => '=',
  ];
  private const USER_FILTER = [
    'column' => 'user_id',
    'compare_operator' => '=',
  ];

  /**
   * @var array
   */
  protected const HTTP_CLIENT_OPTIONS = [
    'timeout' => 900, // Temps max file_get_contents (900 = 15min)
    'headers' => [],
    'verify_host' => false,
    'verify_peer' => false,
  ];

  /**
   * @var JweTokenManager
   */
  protected $oJweTokenManager;

  /**
   * Initialise le controler et instancie les services requis
   * @param Properties $oPropertiesSrvc Service d'accès aux properties
   * @param VitisRequest $oVitisRequest Service d'abstraction de la requête Symfony
   * @param Route $oRouteConfig Service d'accès à la configuration de la route
   * @param LoggerManager $oLoggerManager Service d'utilsation des logs
   * @param DatabaseConnection $oDatabaseConnection Service d'accès à la base de données
   * @param FileManager $oFileManager Service de gestion des fichiers
   * @param JweTokenManager $oJweTokenManager Service de manipulation des token JWE
   */
  public function __construct(Properties $oPropertiesSrvc, VitisRequest $oVitisRequest, Route $oRouteConfig, LoggerManager $oLoggerManager, DatabaseConnection $oDatabaseConnection, FileManager $oFileManager, JweTokenManager $oJweTokenManager) {
    parent::__construct($oPropertiesSrvc, $oVitisRequest, $oRouteConfig, $oLoggerManager, $oDatabaseConnection, $oFileManager);
    $this->oJweTokenManager = $oJweTokenManager;
  }

  /**
   * Récupère une liste d'entités LayerFmeWorkbench (nécessite gtf_amdin)
   * @return Response réponse à retourner par l'API
   */
  public function getDatasById($id): Response {
    return $this->getLayerWorkspaces($id, false);
  }

  /**
   * Récupère une liste d'entités LayerFmeWorkbench
   * @param id id de la couche
   * @param bUserOnly si true alors utilise gtf/userworkspaces, autrement utilise gtf/workspaces
   * @return Response réponse à retourner par l'API
   */
  public function getLayerWorkspaces($id, $bUserOnly = false): Response {
    $sGtfToken = $this->convertvMapTokenIntoGtfToken();

    if (empty($sGtfToken)) {
      $sMessage = 'Error retriving token from GTF API, check vMap configuration';
      return $this->formatGenericError('ERROR_GENERIC_CONTROLLER_GET', '',  $sMessage, Response::HTTP_SERVICE_UNAVAILABLE);
    }

    $curUserId = $this->getCurrentUserEntity()->getUserId();
    $sUrl = $this->getGtfUrl() . ($bUserOnly ? 'gtf/userworkspaces' : 'gtf/workspaces');

    if (empty($sUrl)) {
      return $this->formatGenericError('ERROR_GENERIC_CONTROLLER_GET', '', 'Missing rights to list GTF workspace', Response::HTTP_FORBIDDEN);
    }

    $aData = $this->requestToArray();
    $keepIndex = ['workspace_id', 'name'];
    $aReturn = ['data' => [], 'status' => 1];

    // Rajout des paramètres et des headers à la requête
    $oOption = self::HTTP_CLIENT_OPTIONS;
    $oOption['headers']['Authorization'] = $sGtfToken;
    $oOption['query'] = $aData;

    // Ajoute le filtre pour ne retourner que les workbenchs associés à la couche, dont l'utilisateur est celui actuellement connecté
    $aDefaultFilter = self::MAIN_FILTER;
    if (!empty($id)) {
      $aLayerFilter = self::LAYER_FILTER;
      $aLayerFilter['value'] = $id;
      $aDefaultFilter['operators'][] = $aLayerFilter;
    }
    if (!empty($curUserId)) {
      $aUserFilter = self::USER_FILTER;
      $aUserFilter['value'] = $curUserId;
      $aDefaultFilter['operators'][] = $aUserFilter;
    }

    $aData["filter"] = $aDefaultFilter;

    // chargement de la connexion à la base de donnée
    $aEntities = $this->getDataWithRequestParams($aData);
    if ($aEntities['status'] == 0) {
      return $this->formatGenericResponseForErrorOnSelect();
    }

    // utilisation du client pimpé vitis pour utilisation, proxy, ...
    $oHttpClient = new VitisHttpClient($this->oProperties);
    $client = $oHttpClient->createClient();
    $oResponse = $client->request('GET', $sUrl, $oOption);
    $oContent = null;

    try {
      $oContent = \json_decode($oResponse->getContent(false), true);
      if (isset($oContent['data'])) {
        $oContent = $oContent['data'];
      } else {
        $oContent = null;
      }
    } catch (TransportException | ServerException $e) {
      $sMessage = 'Error retriving data from GTF API, check vMap configuration';
      return $this->formatGenericError('ERROR_GENERIC_CONTROLLER_GET', '', $sMessage, Response::HTTP_SERVICE_UNAVAILABLE);

    }

    if ($oContent === null || !isset($oContent) || $oContent == []) {
      $aEntities['data'] = [];
      return $this->formatResponse($aEntities, Response::HTTP_OK);
    }

    foreach ($oContent as $key => $value) {
      $filteredArray = array_intersect_key($value, array_flip($keepIndex));
      foreach ($aEntities['data'] as $key => $bValue) {
        if ($filteredArray['workspace_id'] == $bValue['workspace_id']) {
          $filteredArray = array_merge($filteredArray, $bValue);
          array_push($aReturn['data'], $filteredArray);
        }
      }
    }

    // transforme le tableau de réponse dans le résultat souhaité
    $aSerializerContext = $this->getSerializerContext();
    return $this->formatResponse($aReturn, Response::HTTP_OK, $aSerializerContext);
  }

  /**
   * Retourne l'url principal pour accéder aux webservice de gtf
   *
   * @return string
   */
  private function getGtfUrl(): string {
    // récupération de l'url principal
    $gtfUrl = $this->oProperties->get('gtf_api_url');
    // gestion absence / de fin
    if (substr($gtfUrl, -1) != '/') {
      $gtfUrl .= '/';
    }
    // ajout des constante jusqu'au module gtf
    $gtfUrl .= 'v2/';

    return $gtfUrl;
  }

  /**
   * contact l'API de GTF pour récupérer un token à partir des infos dans le token qui contact l'application
   *
   * @return string|null token de GTF ou null en cas d'erreur
   */
  private function convertvMapTokenIntoGtfToken(): ?string{
    // Token décrypté
    $token = $this->oVitisRequest->getToken();
    if (empty($token)) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot contact GTF api without token');
      return null;
    }
    $aTokenJWEDecrypt = $this->oJweTokenManager->decryptToken($token);

    $sLogin = $aTokenJWEDecrypt['data']['user']['login'] ?? '';
    $sPassword = $aTokenJWEDecrypt['data']['password'] ?? '';

    if (empty($sLogin) || empty($sPassword)) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot find user info in this Token');
      return null;
    }

    $sGtfTokenUrl = $this->getGtfUrl() . 'vitis/privatetoken';

    // utilisation du client pimpé vitis pour utilisation, proxy, ...
    $oHttpClient = new VitisHttpClient($this->oProperties);
    $client = $oHttpClient->createClient();
    $oOption = self::HTTP_CLIENT_OPTIONS;
    $oOption['json'] = [
      'user' => $sLogin,
      'password' => $sPassword,
    ];

    $oResponse = $client->request('POST', $sGtfTokenUrl, $oOption);

    try {
      $aContent = $oResponse->toArray();
      if ($aContent['status'] != 1) {
        $this->oLoggerManager->get('API_ERROR')->error('an error occured during GTF token request');
        return null;
      }
      return $aContent['data']['token'];
    } catch (TransportException | ServerException $e) {
      $this->oLoggerManager->get('API_ERROR')->error('an error occured during GTF token request : ' . $e->getMessage());
      return null;
    }
  }

  /**
   * Sert de proxy et envoie la requête à GTF
   * @return Response réponse à retourner par l'API
   */
  private function proxyGTFrequest($sUrl, $sMethod='GET'): Response {
    $aData = $this->requestToArray();
    $sGtfToken = $this->convertvMapTokenIntoGtfToken();
    $aUserRights = $this->getCurrentUserEntity()?->getRoles();

    if (empty($aUserRights)) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot read user role');
      return null;
    }

    // si aucun des trois pivilège est dans la liste de sprivilège de l'utilsiateur il faut qu'il retouche ses droits
    $aGtfRole = ['gtf_admin', 'gtf_author', 'gtf_user'];
    if (count(array_diff($aGtfRole,$aUserRights)) === count($aGtfRole)) {
      $this->oLoggerManager->get('API_ERROR')->error('GTF roles are missing for this user');
      return null;
    }

    if (empty($sGtfToken)) {
      $sMessage = 'Error retriving token from GTF API, check vMap configuration';
      return $this->formatGenericError('ERROR_GENERIC_CONTROLLER_GET', '', $sMessage, Response::HTTP_SERVICE_UNAVAILABLE);
    }

    // preset de la structure de réponse
    $aReturn = ['status' => 0];

    // Rajout des paramètres et des headers à la requête
    $oOption = self::HTTP_CLIENT_OPTIONS;
    $oOption['headers']['Authorization'] = $sGtfToken;

    switch ($sMethod) {
      case 'POST':
        $oOption['body'] = $aData;
        break;
      default:
        $oOption['query'] = $aData;
        break;
    }

    // utilisation du client pimpé vitis pour utilisation, proxy, ...
    $oHttpClient = new VitisHttpClient($this->oProperties);
    $client = $oHttpClient->createClient();

    $oResponse = $client->request($sMethod, $sUrl, $oOption);
    $aContent = [];

    try {
      $aContent = $oResponse->toArray();
    } catch (TransportException | ServerException $e) {
      $sMessage = 'Error retriving data from GTF API, check your configuration';
      return $this->formatGenericError('ERROR_GENERIC_CONTROLLER_GET', '', $sMessage, Response::HTTP_SERVICE_UNAVAILABLE);
    }

    // chargement de la connexion à la base de donnée
    $aEntities = $this->getDataWithRequestParams($aData);
    if ($aEntities['status'] == 0) {
      return $this->formatGenericResponseForErrorOnSelect();
    }

    return $this->formatResponse($aContent, $oResponse->getStatusCode());
  }

  /**
   * Récupère la liste des traitements GTF (nécessite gtf_amin)
   * @return Response réponse à retourner par l'API
   */
  public function getGtfWorkspaces(): Response {
    $aUserRights = $this->getCurrentUserEntity()?->getRoles();

    if (empty($aUserRights)) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot read user role');
      return null;
    }

    $aGtfRole = ['gtf_admin', 'gtf_author'];
    if (count(array_diff($aGtfRole, $aUserRights)) === count($aGtfRole)) {
      $this->oLoggerManager->get('API_ERROR')->error('GTF roles are missing for this user');
      $sMessage = 'User must be gtf_admin';
      return $this->formatGenericError('REQUESTOR_WORKSPACES_MISSING_GTF_RIGHTS_ADMIN', '', $sMessage, Response::HTTP_UNAUTHORIZED);
    }
    $sUrl = $this->getGtfUrl() . 'gtf/workspaces';
    return $this->proxyGTFrequest($sUrl);
  }

  /**
   * Récupère la liste des traitements GTF associés à la couche et à l'utilisateur actif
   * @return Response réponse à retourner par l'API
   */
  public function getGtfUserWorkspaces($id): Response {
    return $this->getLayerWorkspaces($id, true);
  }

  /**
   * Ajoute une liste de workbenchID dans la table layer_workspace
   * @return Response réponse à retourner par l'API
   */
  public function postDatas() {
    $routeParameters = $this->getRequestParameters();
    $aData = $this->requestToArray();
    $curUserId = $this->getCurrentUserEntity()->getUserId();

    try {
      $aEntities = [];
      for ($i = 0; $i < count($aData['data']); $i++) {
        $sData = $aData['data'][$i];
        $layerFmeWorkbench = new LayerFmeWorkbench();
        $layerFmeWorkbench->setWorkspaceId($sData['workspace_id']);
        $layerFmeWorkbench->setLayerId($sData['layer_id']);
        $layerFmeWorkbench->setUserId($curUserId);

        if (array_key_exists('layer_index', $sData)) {
          $layerFmeWorkbench->setLayerIndex($sData['layer_index']);
        }
        $serialisation = $layerFmeWorkbench->jsonSerialize($routeParameters);
        $this->oDatabaseConnection->getAdminEntityManager()->persist($layerFmeWorkbench);

        $serialisation['layer_workspace_id'] = $layerFmeWorkbench->getLayerWorkspaceId();
        $aEntities[] = $serialisation;
      }
      $this->oDatabaseConnection->getAdminEntityManager()->flush();

      $aReturn = [
        'data' => $aEntities,
        'status' => 1,
      ];
      $this->formatResponse($aReturn, Response::HTTP_OK);
    } catch (\Exception $e) {
      $this->formatResponse(['status' => 0], Response::HTTP_UNAUTHORIZED);
    }
    return $this->oResponse;
  }

  /**
   * Modifie des entité LayerFmeWorkbench
   * @return Response réponse à retourner par l'API
   */
  public function updateDatas(): Response {
    $aData = $this->requestToArray();
    $oDataAccessor = $this->getDataAccessor();

    try {
      $aEntities = [];
      foreach ($aData['data'] as $data) {
        $sData = json_encode($data);
        $oEntity = $oDataAccessor->updateEntity($data['layer_workspace_id'], $sData);
        $aEntities[] = $oEntity;
      }

      $aReturn = [
        'data' => $aEntities,
        'status' => 1,
      ];
      $this->formatResponse($aReturn, Response::HTTP_OK);
    } catch (\Exception $e) {
      $this->formatResponse(['status' => 0], Response::HTTP_UNAUTHORIZED);
    }

    return $this->oResponse;
  }

  /**
   * Récupère le formulaire d'un workspace GTF
   * @return Response réponse à retourner par l'API
   */
  public function getWorkspaceForm($id, $workspace_id, $form): Response {
    $sUrl = $this->getGtfUrl() . 'gtf/userworkspaces/' . $workspace_id . '/form/' . $form;
    return $this->proxyGTFrequest($sUrl);
  }

  /**
   * Permet d'envoyer une demande GTF
   * @return Response réponse à retourner par l'API
   */
  public function postOrder($id, $workspace_id): Response {
    
    // Test l'existance et les droits de l'utilisateur sur le workspace
    try {
      $aWorkspaceResp = $this->proxyGTFrequest($this->getGtfUrl() . 'gtf/userworkspaces/' . $workspace_id);
    } catch (\Throwable $th) {
      $sMessage = 'GTF workspace unavailable';
      return $this->formatGenericError('GTF_WORKSPACE_UNAVAILABLE', '', $sMessage, Response::HTTP_SERVICE_UNAVAILABLE);
    }

    // Lancement de la demande
    $sUrl = $this->getGtfUrl() . 'gtf/userorders';
    return $this->proxyGTFrequest($sUrl, 'POST');
  }


}
