<?php

namespace App\Module\Vmap\WebService\Controller;

use App\Module\Vitis\WebService\Controller\GenericController;
use App\Module\Vitis\WebService\Repository\GenericRepository;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\VitisRequest;
use App\Module\Vmap\WebService\Entity\AbstractLayer;
use App\Module\Vmap\WebService\Entity\LayerFile;
use App\Module\Vmap\WebService\Entity\LayerOSM;
use App\Module\Vmap\WebService\Entity\LayerPostgres;
use App\Module\Vmap\WebService\Entity\LayerVectorTiles;
use App\Module\Vmap\WebService\Entity\LayerWMS;
use App\Module\Vmap\WebService\Entity\LayerWMTS;
use App\Module\Vmap\WebService\Entity\LayerXYZ;
use App\Module\Vmap\WebService\Entity\Map;
use App\Module\Vmap\WebService\Service\MapServerService;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller pour le MapJSON de vMap.
 *
 */
class MapJSONController extends GenericController {
  // Exemple de JSON
  //
  // {
  //   type: string;
  //   tree?: any; // arbre de définition de la carte
  //   map_id?: number;
  //   center?: any;
  //   extent?: any;
  //   source?: any; // utilisé pour cartes Bing
  //   circle_radius?: number;
  //   contour_color?: string;
  //   contour_size?: number;
  //   controls: {
  //     map_projection?: boolean,
  //     mobile_gps_zoom?: boolean,
  //     mouse_position?: boolean,
  //     scale_line?: boolean,
  //     zoom_controls?: boolean,
  //     location_controls?: boolean,
  //     attribution?: boolean, // True par défaut
  //   };
  //   coord_accuracy?: number;
  //   draw_color?: string;
  //   interactions: {
  //     business_objects?: boolean,
  //     draw_point?: boolean,
  //     draw_line?: boolean,
  //     draw_polygon?: boolean,
  //     full_screen?: boolean,
  //     modify?: boolean,
  //     multi_geometry?: boolean,
  //     remove?: boolean,
  //     remove_all?: boolean,
  //   };
  //   output_format?: string;
  //   proj?: string;
  //   map_geocoders?: any;
  //   map_default_geocoders?: string;
  // }

  /**
   * Clé utilisée pour l'id dans les chemins
   * @var string
   */
  private const ID_PATH_KEY = '@id';
  /**
   * Clé utilisée pour la valeur du champ dans les chemins
   * @var string
   */
  private const VALUE_PATH_KEY = '@value';
  /**
   * Clé utilisée pour la valeur du champ dans les chemins
   * @var string
   */
  private const COLUMN_PATH_KEY = '@column';

  /**
   * @var MapServerService
   */
  private $oMapServerService;

  /**
   * Initialise le controler et instancie les services requis
   * @param Properties $oPropertiesSrvc Service d'accès aux properties
   * @param VitisRequest $oVitisRequest Service d'abstraction de la requête Symfony
   * @param Route $oRouteConfig Service d'accès à la configuration de la route
   * @param LoggerManager $oLoggerManager Service d'utilsation des logs
   * @param DatabaseConnection $oDatabaseConnection Service d'accès à la base de données
   * @param FileManager $oFileManager Service de gestion des fichiers
   * @param MapServerService $oMapServerService
   */
  public function __construct(Properties $oPropertiesSrvc, VitisRequest $oVitisRequest, Route $oRouteConfig, LoggerManager $oLoggerManager, DatabaseConnection $oDatabaseConnection, FileManager $oFileManager, MapServerService $oMapServerService) {
    parent::__construct($oPropertiesSrvc, $oVitisRequest, $oRouteConfig, $oLoggerManager, $oDatabaseConnection, $oFileManager);
    $this->oMapServerService = $oMapServerService;
  }

  /**
   * Récupère une carte au format MapJSON
   * @param mixed $id id de la carte à trouver
   * @return Response réponse à retourner par l'API
   */
  public function getData($id): Response {
    $aData = $this->requestToArray();

    // chargement de la connexion à la base de donnée
    $oDataAccessor = $this->getDataAccessor();
    $oEntity = $oDataAccessor->getEntity($id, $aData);

    if ($oEntity['status'] === 0) {
      return $this->oResponse;
    }

    // Filtre à appliquer sur les couches
    $aFilter = null;
    if (array_key_exists('filter', $aData)) {
      $tmpFilter = json_decode($aData['filter'], true);
      if (is_array($tmpFilter) && count($tmpFilter) > 0) {
        $aFilter = $tmpFilter;
      }
    }
    $oMapTree = $this->getMapTree($oEntity['data'], $aFilter);

    if ($oMapTree === null) {
      return $this->formatResponse([
        'status' => 0,
        'message' => 'Error during get',
        'errorMessage' => 'ERROR_GENERIC_CONTROLLER_GET',
      ], Response::HTTP_UNAUTHORIZED);
    }

    $oMap = [
      'name' => $oEntity['data']['name'],
      'children' => $oMapTree,
      'maptools' => $this->getMapTools($oEntity['data']),
      'options' => [
        'optimized' => true,
        'default_layer_location_id' => $oEntity['data']['default_layer_location_id'],
      ],
    ];

    // Commentaire de la carte.
    if (!empty($oEntity['data']['comment_activation']) && $oEntity['data']['comment_activation'] === true) {
      $oMap['comment'] = [
        'text' => $oEntity['data']['comment'],
        'display' => $oEntity['data']['comment_display'],
        'last_modification' => $oEntity['data']['comment_last_modification']
      ];
    }

    $oEntity['data'] = $oMap;

    $this->formatResponse($oEntity, Response::HTTP_OK);
    return $this->oResponse;
  }

  /**
   * Retourne maptools parsé
   * @param array $aMap
   * @return array
   */
  private function getMapTools(array $aMap): array{
    return json_decode($aMap['maptools'], true) ?? [];
  }

  /**
   * Retourne l'arbre définition de la carte
   * @param array $aMap
   * @param array $aFilter filtre à appliquer sur les couches
   * @return array|null
   */
  private function getMapTree(array $aMap, ?array $aFilter = null): ?array{
    $aTree = [[
      'view' => [
        'extent' => explode('|', $aMap['extent']),
        'projection' => $aMap['coordsys_epsg'],
      ],
    ]];

    $mPublishedLayers = $this->publishLayers($aMap, $aFilter);

    if ($mPublishedLayers === null) {
      return null;
    }

    $aLayers = $this->orderLayersByTheme($mPublishedLayers);

    // Sort les valeurs par le champ 'name'
    usort($aLayers, function (mixed $a, mixed $b) {
      return strcmp($a['name'], $b['name']);
    });

    foreach ($aLayers as $aLayer) {
      $aTree[] = $aLayer;
    }

    return $aTree;
  }

  /**
   * Retourne la définition des couches (avec héritage) d'une carte
   * @param array $aMap
   * @return array|null
   */
  private function getMapLayers(array $aMap): ?array{
    // Récupération du contenu de layer_map
    try {
      $oMapRepository = $this->oDatabaseConnection->getAdminEntityManager()->getRepository(Map::class);

      if (!$oMapRepository instanceof GenericRepository) {
        $this->oLoggerManager->get('API_ERROR')->error('Bad type repository');
        return null;
      }

      $oMap = $oMapRepository->findById($aMap['map_id'])[0];
    } catch (\Exception $e) {
      $this->oLoggerManager->get('API_ERROR')->error('Cannot find map with id : ' . $aMap['map_id']);
      return null;
    }

    $aLayers = $oMap->getLayersWithEntity();
    $oLayerRepository = $this->oDatabaseConnection->getAdminEntityManager()->getRepository(AbstractLayer::class);

    if (!$oLayerRepository instanceof GenericRepository) {
      $this->oLoggerManager->get('API_ERROR')->error('Bad type repository');
      return null;
    }

    // Remplit le reste des infos de la couche
    foreach ($aLayers as $key => $aLayerMapDef) {
      if (isset($aLayerMapDef['layer_id'])) {
        $oLayer = $oLayerRepository->findById($aLayerMapDef['layer']->getLayerId())[0];
        unset($aLayerMapDef['layer']);

        // Attributs de AbstractLayer
        $oLayerRoute = $this->oRouteConfig->getRequestParameters('vmap_layer');
        $aLayerDef = $oLayer->jsonSerialize($oLayerRoute);

        // is_location : si dans layer vaut true alors dépend de layer_map (true=true, false=false)
        if (array_key_exists('is_location', $aLayerDef) && $aLayerDef['is_location'] === true) {
          if (array_key_exists('is_location', $aLayerMapDef) && $aLayerMapDef['is_location'] === true) {
            $aLayerDef['is_location'] = true;
          } else {
            $aLayerDef['is_location'] = false;
          }
        } else {
          $aLayerDef['is_location'] = false;
        }

        // is_map_selection : si dans layer vaut true alors dépend de layer_map (true=true, false=false)
        if (array_key_exists('is_map_selection', $aLayerDef) && $aLayerDef['is_map_selection'] === true) {
          if (array_key_exists('is_map_selection', $aLayerMapDef) && $aLayerMapDef['is_map_selection'] === true) {
            $aLayerDef['is_map_selection'] = true;
          } else {
            $aLayerDef['is_map_selection'] = false;
          }
        } else {
          $aLayerDef['is_map_selection'] = false;
        }

        // is_data_grid : si dans layer vaut true alors dépend de layer_map (true=true, false=false)
        if (array_key_exists('is_data_grid', $aLayerDef) && $aLayerDef['is_data_grid'] === true) {
          if (array_key_exists('is_data_grid', $aLayerMapDef) && $aLayerMapDef['is_data_grid'] === true) {
            $aLayerDef['is_data_grid'] = true;
          } else {
            $aLayerDef['is_data_grid'] = false;
          }
        } else {
          $aLayerDef['is_data_grid'] = false;
        }
        // is_filter : si dans layer vaut true alors dépend de layer_map (true=true, false=false)
        $aLayerDef['is_filter'] = (array_key_exists('is_filter', $aLayerMapDef) && $aLayerMapDef['is_filter'] === true && $aLayerDef['is_filter'] === true);

        // gestion definition_sym_json
        if (array_key_exists('definition_sym_json', $aLayerDef) && array_key_exists('definition_sym_json_map', $aLayerMapDef)) {
          if (!empty($aLayerMapDef['definition_sym_json_map'])) {
            $aLayerDef['definition_sym_json'] = $aLayerMapDef['definition_sym_json_map'];
          }
        }
        
        // Transforme le nom du pictogramme et le nom de la legende en son chemin public depuis le serveur
        $aLayerDef = $this->transformFileToUrl($oLayerRoute, $aLayerDef, 'pictogram');
        $aLayerDef = $this->transformFileToUrl($oLayerRoute, $aLayerDef, 'legend');

        // Ajoute les droits d'utilisateur sur les couches postgres
        if ($aLayerDef['layertype'] === LayerPostgres::TYPE) {
          $oEm = $this->oDatabaseConnection->getUserEntityManager();

          if (array_key_exists('datasource', $aLayerDef)) {
            // création de l'entity manager si datasource non privée
            if ($aLayerDef['datasource']['pg_public'] == true) {
              $oEm = $this->oDatabaseConnection->createExternalEntityManager($aLayerDef['database'], $aLayerDef['datasource']['pg_login'], $aLayerDef['datasource']['pg_password'], DatabaseConnection::SGBD, $aLayerDef['datasource']['pg_server'], $aLayerDef['datasource']['pg_port']);
            } else if ($aLayerDef['database'] !== $this->oProperties->get('db_name')) {
              $oEm = $this->oDatabaseConnection->createExternalEntityManager($aLayerDef['database']);
            }
          }

          $sSchema = $aLayerDef['schema'];
          $sTable = $aLayerDef['pg_table'];
          if ($aLayerDef['is_specific_table_form'] && $aLayerDef['is_specific_table_form'] == true
            && $aLayerDef['schema_form'] != '' && $aLayerDef['table_form'] != '') {
            $sSchema = $aLayerDef['schema_form'];
            $sTable = $aLayerDef['table_form'];
          }

          $aPrivilege = $this->oDatabaseConnection->getQueryAccessor($oEm)->getTableRights($sSchema, $sTable);
          if (is_array($aPrivilege) && !empty($aPrivilege)) {
            $aLayerDef = array_merge($aLayerDef, [
              'have_select_right' => $aPrivilege['SELECT'],
              'have_insert_right' => $aPrivilege['INSERT'],
              'have_update_right' => $aPrivilege['UPDATE'],
              'have_delete_right' => $aPrivilege['DELETE'],
            ]);
          } else {
            $aLayerDef = array_merge($aLayerDef, [
              'have_select_right' => false,
              'have_insert_right' => false,
              'have_update_right' => false,
              'have_delete_right' => false,
            ]);
          }

          $bHaveRequestorSelectRight = false;
          $sSchema = $aLayerDef['schema'];
          $sTable = $aLayerDef['pg_table'];
          if ($aLayerDef['is_specific_table_requestor'] && $aLayerDef['is_specific_table_requestor'] == true
            && $aLayerDef['schema_requestor'] != '' && $aLayerDef['table_requestor'] != '') {
            $sSchema = $aLayerDef['schema_requestor'];
            $sTable = $aLayerDef['table_requestor'];
          }

          $aPrivilege = $this->oDatabaseConnection->getQueryAccessor($oEm)->getTableRights($sSchema, $sTable);
          if (is_array($aPrivilege) && !empty($aPrivilege)) {
            $bHaveRequestorSelectRight = $aPrivilege['SELECT'];
          }
          $aLayerDef = array_merge($aLayerDef, [
            'have_requestor_select_right' => $bHaveRequestorSelectRight,
          ]);
        }

        $aLayers[$key] = array_merge($aLayerMapDef, $aLayerDef);
      }
    }
    return $aLayers;
  }

  /**
   * Transforme les couches en calques
   * @param array $aMap
   * @param array $aFilter filtre à appliquer sur les couches
   * @return array|null
   */
  private function publishLayers(array $aMap, ?array $aFilter = null): ?array{
    $aLayers = $this->getMapLayers($aMap);
    if ($aLayers === null) {
      return null;
    }

    // Publication du flux
    $bHasLayerPostgres = false;
    $bHasLayerFile = false;
    foreach ($aLayers as $key => $aLayer) {
      if ($aLayer['layertype'] === LayerPostgres::TYPE) {
        $bHasLayerPostgres = true;
      } else if ($aLayer['layertype'] === LayerFile::TYPE) {
        $bHasLayerFile = true;
      }
      // Identifiant sur le nom qui servira à éviter les doublons dans le mapfile
      $aLayers[$key]['uniquename'] = $aLayer['name'] .= '_' . uniqid();
    }

    // Générer le fichier fonts.list avec les données en bdd
    $this->oMapServerService->generateFontList();

    $sMapfileName = '';
    if ($bHasLayerPostgres || $bHasLayerFile) {
      $sMapfileName = $this->oMapServerService->createMapfile($aMap, $aLayers, $aFilter, false);
    }

    $aLayers = array_filter($aLayers, function (mixed $k) {
      return !is_string($k);
    }, ARRAY_FILTER_USE_KEY);

    // Couches au format OpenLayers
    foreach ($aLayers as $key => $aLayer) {

      $aLayers[$key] = match($aLayer['layertype']) {
        LayerWMS::TYPE => LayerWMS::getLayerParams($aLayer),
        LayerWMTS::TYPE => LayerWMTS::getLayerParams($aLayer),
        LayerOSM::TYPE => LayerOSM::getLayerParams($aLayer),
        LayerXYZ::TYPE => LayerXYZ::getLayerParams($aLayer),
        LayerVectorTiles::TYPE => LayerVectorTiles::getLayerParams($aLayer),
        LayerFile::TYPE => $this->getFileLayerParams($aLayer, $sMapfileName),
        LayerPostgres::TYPE => $this->getPostgresLayerParams($aLayer, $sMapfileName),
      };
    }

    return $aLayers;
  }

  /**
   * Renvoie les paramètres spécifiques a OSM
   * @param array $aLayer
   * @param string $sMapfileName
   * @return array
   */
  private function getFileLayerParams(array $aLayer, string $sMapfileName): array{
    $sOnlineResourceUrl = str_replace('@file', $sMapfileName, MapServerService::PRIVATE_WMS_MAIN_URL);
    $sOnlineResourceUrl = $this->oProperties->replacePropertiesInString($sOnlineResourceUrl);
    $aLayer['url'] = $sOnlineResourceUrl;
    return LayerFile::getLayerParams($aLayer);
  }

  /**
   * Renvoie les paramètres spécifiques a OSM
   * @param array $aLayer
   * @param string $sMapfileName
   * @return array
   */
  private function getPostgresLayerParams(array $aLayer, string $sMapfileName): array{
    $sOnlineResourceUrl = str_replace('@file', $sMapfileName, MapServerService::PRIVATE_WMS_MAIN_URL);
    $sOnlineResourceUrl = $this->oProperties->replacePropertiesInString($sOnlineResourceUrl);
    $aLayer['url'] = $sOnlineResourceUrl;
    $aFinalLayer = LayerPostgres::getLayerParams($aLayer);

    // Couches non géométriques
    if (empty($aFinalLayer['definition_sym_json']) || $aFinalLayer['is_geometry_layer'] === false) {
      $aFinalLayer['is_map_hidden'] = true;
      $aFinalLayer['is_geometry_layer'] = false;
    }

    // Supprime les données confidentielles ou inutiles & optimise le JSON du formulaire
    $aFinalLayer = LayerPostgres::simplifyLayer($aFinalLayer);

    return $aFinalLayer;
  }

  /**
   * Ordonne les couches sous forme d'arbre en fonction de leur thème
   * @param array $aLayers
   * @return array
   */
  private function orderLayersByTheme(array $aLayers): array{
    $aReturn = [];
    $aThemes = $this->getLayersThemes($aLayers);

    foreach ($aThemes as $aTheme) {
      if (count($aTheme) > 1) {
        $sParentTheme = $aTheme[0];
        array_shift($aTheme);

        $aTree = [
          'name' => $sParentTheme,
          'children' => $this->findLayersByTheme([$sParentTheme], $aLayers),
        ];

        foreach ($aTheme as $sTheme) {
          $aTree['children'][] = $this->getThemeLayers($sTheme, $sParentTheme, $aLayers);
        }

        $aReturn[] = $aTree;
      } else {
        $sTheme = $aTheme[0];
        if ($sTheme !== 'null') {
          $groupedLayers = $this->findLayersByTheme($aTheme, $aLayers);
          // Sort les valeurs par le champ 'name'
          usort($groupedLayers, function (mixed $a, mixed $b) {
            return strcmp($a['name'], $b['name']);
          });

          $aReturn[] = [
            'name' => $sTheme,
            'children' => $groupedLayers,
          ];
        } else {
          $aWithoutThemeLayers = $this->findLayersByTheme($aTheme, $aLayers);
          foreach ($aWithoutThemeLayers as $aWithoutThemeLayer) {
            $aReturn[] = $aWithoutThemeLayer;
          }
        }
      }
    }

    return $aReturn;
  }

  /**
   * Trouve les layers en fonction du thème passé en paramètre
   * @param array $aTheme
   * @param array $aLayers
   * @return array
   */
  private function findLayersByTheme(array $aTheme, array $aLayers): array{
    $aResult = [];

    foreach ($aLayers as $aLayer) {
      $sLayerTheme = 'null';
      if (!empty($aLayer['theme']) && !empty($aLayer['theme']['label'])) {
        $sLayerTheme = $aLayer['theme']['label'];
      }

      if (join('/', $aTheme) === $sLayerTheme) {
        $aResult[] = $aLayer;
      }
    }

    return $aResult;
  }

  /**
   * Regarde de façon stricte (aucune différence entre les deux tableaux) si un tableau est déjà présent dans un tableau à deux dimmensions
   * @param array $arr1 tableau à deux dimmensions cible
   * @param array $arr2 tableau à rechercher
   * @return bool
   */
  private function array_diff_multidimensionnal(array $arr1, array $arr2): bool{
    foreach ($arr1 as $subArray) {
        if(count(array_diff($subArray, $arr2)) == 0 && count(array_diff($arr2, $subArray)) == 0) {
          return true;
        }
    }
    return false;
}

  /**
   * Retourne le tableau de thèmes
   */
  private function getLayersThemes(array $aLayers): array{
    $aThemes = [];
    $aOrderedThemes = [];
    foreach ($aLayers as $aLayer) {
      $sTheme = 'null';
      if (!empty($aLayer['theme']) && !empty($aLayer['theme']['label'])) {
        $sTheme = $aLayer['theme']['label'];
      }
      $aTheme = explode('/', $sTheme);
      if (count($aTheme) <= 2) {
        if(!$this->array_diff_multidimensionnal($aThemes, $aTheme)) $aThemes[] = $aTheme;
      } else {
        $aTheme = [$sTheme];
        if(!$this->array_diff_multidimensionnal($aThemes, $aTheme)) $aThemes[] = $aTheme;
      }
    }

    $aOrderedThemes = $this->orderThemes($aThemes, $aOrderedThemes);
    return $aOrderedThemes;
  }

  /**
   * Ordonne les thèmes sous deux niveaux max
   */
  private function orderThemes(array $aThemes, array $aOrderedThemes): array{
    foreach ($aThemes as $aTheme) {
      $arrColumns = array_column($aOrderedThemes, 0);
      $arrSearch = array_search($aTheme[0], $arrColumns, true); // Recherche si le parent niveau 0 existe déjà

      if ($arrSearch === false) { // Thème non trouvé
        $aOrderedThemes[] = $aTheme;
      } else { // Thème en commun trouvé
        if ($aOrderedThemes[$arrSearch] != $aTheme) {
          // Recherche sur le niveau 2
          if (count($aTheme) > 1) {
            $arrColumns2 = array_column($aOrderedThemes, 1);
            $arrSearch2 = array_search($aTheme[1], $arrColumns2, true); // Recherche si le parent niveau 1 existe déjà
            if ($arrSearch2 === false) { // Thème non trouvé
              $aOrderedThemes[$arrSearch][] = $aTheme[1];
            }
          }
        }
      }
    }

    return $aOrderedThemes;
  }

  /**
   * Permet la récupération des couches dans des sous-themes
   * @param string $sTheme
   * @param string $sParentTheme
   * @param array $aLayers
   * @return array
   */
  private function getThemeLayers(string $sTheme, string $sParentTheme, array $aLayers): array{
    $aTree = [
      'name' => $sTheme,
      'children' => $this->findLayersByTheme([$sParentTheme, $sTheme], $aLayers),
    ];
    return $aTree;
  }

  /**
   * Transforme un champ en son chemin public depuis le serveur
   *
   * @param array $aLayerDef
   * @param string $sField
   * @param [type] $oLayerRoute
   * @return array
   */
  private function transformFileToUrl($oLayerRoute, array $aLayerDef, string $sField) {
    if (array_key_exists($sField, $aLayerDef)) {
      $sFilePath = $aLayerDef[$sField];
      $aFieldOptions = [];
      foreach ($oLayerRoute['options']['fields'] as $field) {
        if ($field['name'] === $sField) {
          $aFieldOptions = $field;
        }
      }
      $sPath = $this->oProperties->replacePropertiesInString($aFieldOptions['path']);
      // remplacement de l'id
      $sPath = str_replace(self::ID_PATH_KEY, $aLayerDef[$oLayerRoute['options']['field_id']], $sPath);
      // remplacement des champs rensigné
      $sPath = str_replace(self::VALUE_PATH_KEY, $sFilePath, $sPath);
      // remplacement du nom du champs
      $sPath = str_replace(self::COLUMN_PATH_KEY, $aFieldOptions['name'], $sPath);
      // Renvoie l'url public de l'image sur le serveur
      $sUrl = $this->oFileManager->getAccessorForPath($sPath)->getProxyPassUrl($sPath);

      if (!empty($sUrl)) {
        $aLayerDef[$sField] = $sUrl;
      }
    }

    return $aLayerDef;
  }
}
