<?php

namespace App\Module\Vmap\WebService\Controller;

use App\Module\Vitis\WebService\Controller\StudioFormController;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller qui permet de récupérer et d'intéragir avec les formulaires studio.
 *
 * La route devra contenir id et form
 * id : identifiant de l'enregistrement
 * form : formulaire à retourner (default, published, custom)
 *
 * En put form peut prendre la valeur init pour sauvegarder tous les formulaires (si ils sont vide en bdd)
 *
 * Exemple :
 * "route": "/v2/vmap/datasourcetreeview/pglayers/{id}/form/{form}",
 *
 * Dans les options il y aura
 * schema: Schema contenant les données
 * table: Table contenant les données
 * field_id: Champ identifiant
 * En plus il faudra utiliser l'attribut fields et y placer un tableau d'entités
 * Chaque entité est définie par name, form_field
 * name : nom de l'attribut en sortie
 * form_field : définit le nom du champ pour chaque type de formulaire
 *
 * Exemple :
 * "fields": {
 * "json": {
 *   "name": "json_form",
 *   "form_field": {
 *     "default": "default_form_json_def",
 *     "published": "published_form_json_def",
 *     "custom": "custom_form_json_def"
 *   }
 * },
 * "js": {
 *   "name": "js_form",
 *   "form_field": {
 *     "default": "default_form_js_def",
 *     "published": "published_form_js_def",
 *     "custom": "custom_form_js_def"
 *   }
 *  },
 * "css": {
 *    "name": "css_form",
 *    "form_field": {
 *      "default": "default_form_css_def",
 *      "published": "published_form_css_def",
 *      "custom": "custom_form_css_def"
 *    }
 *  }
 * }
 */
class StudioFormObjectController extends StudioFormController {



  /**
   * Récupère le contenu du formulaire
   * @param mixed $id
   * @param mixed $form
   * @return Response réponse à retourner par l'API
   */
  public function getForm($id, $form): Response {
    $aRequestOptions = $this->getRequestParameters();
    $aReturn = [];
    $haveResponse = false;

    if (!empty($id) && !empty($form) && is_array($aRequestOptions['options']['fields'])) {
      $aReturn['data'] = [];
      foreach ($aRequestOptions['options']['fields'] as $i => $oField) {
        $sFieldContent = $this->getFieldContent($aRequestOptions['options'], $oField, $id, $form);
        if ($sFieldContent !== null) {
          $haveResponse = true;
          $aReturn['data'][$oField['name']] = $sFieldContent;
        }
      }
    }

    if ($haveResponse === false) {
      $aReturn['message'] = 'can\'t get form content';
      return $this->formatResponse($aReturn, Response::HTTP_NOT_ACCEPTABLE);
    }

    $aReturn['status'] = 1;
    $this->formatResponse($aReturn);
    return $this->oResponse;
  }

  /**
   * Retourne le contenu d'un champ de la base
   * @param mixed $oOptions
   * @param mixed $oField
   * @param mixed $id
   * @param mixed $form
   * @return string|null
   */
  private function getFieldContent($oOptions, $oField, $id, $form): ?string {
    if (!empty($oField['name']) && !empty($oField['form_field'][$form])) {
      // Requête sql
      $qb = $this->oDatabaseConnection->getQueryAccessor()->createQueryBuilder();

      try {
        $qb->select('a.[FIELD]')
          ->from('[SCHEMA].[TABLE]', 'a')
          ->where($qb->expr()->eq('a.[FIELD_ID]', '[ID_VALUE]'));

        $aParams = [
          'FIELD' => ['value' => $oField['form_field'][$form], 'type' => 'column_name'],
          'SCHEMA' => ['value' => $this->oProperties->replacePropertiesInString($oOptions['schema']), 'type' => 'schema_name'],
          'TABLE' => ['value' => $oOptions['table'], 'type' => 'table_name'],
          'FIELD_ID' => ['value' => $oOptions['field_id'], 'type' => 'column_name'],
          'ID_VALUE' => ['value' => $id, 'type' => 'mixed'],
        ];

        $oContent = $qb->executeWithParams($aParams)->fetchAllAssociative();
      } catch (\Exception $e) {
        $this->oLoggerManager->get('SQL_ERROR')->error('Studio : can\'t get form content: ' . $e->getMessage());
        return null;
      }

      return $oContent[0][$oField['form_field'][$form]] ?? '';
    }
    return null;
  }

  /**
   * Enregistre le contenu du formulaire
   * @param mixed $id
   * @param mixed $form
   * @return Response
   */
  public function updateFields($id, $form): Response {
    $aRequestOptions = $this->getRequestParameters();
    $aReturn = [
      'status' => 0,
    ];
    if (!empty($id) && !empty($form) && is_array($aRequestOptions['options']['fields'])) {
      if ($form === 'init') {
        // Si c'est une demande init on sauvegarde le formulaire dans les différents types (Default, Published, Perso)
        $result = $this->saveInitialForms_($id);
      } else if ($form === 'initautogenerated') {
        // Si c'est une demande initautogenerated on sauvegarde le formulaire dans les différents types (Default, Published, Perso)
        $result = $this->saveInitialForms_($id, true);
      } else {
        $result = $this->updateFields_($id, $form);
      }
      if ($result == null) {
        $aReturn['message'] = 'Error during update form content';
        $aReturn['errorMessage'] = 'ERROR_GENERIC_CONTROLLER_UPDATE';
        return $this->formatResponse($aReturn, Response::HTTP_UNAUTHORIZED);
      }
      return $this->getForm($id, $result);
    }

    $this->formatResponse($aReturn);
    return $this->oResponse;
  }

  /**
   * Met à jour les champs d'un formulaire
   * @param mixed $id
   * @param string $form
   * @return string|null
   */
  private function updateFields_($id, string $form): ?string{
    $aRequestOptions = $this->getRequestParameters();
    $aData = $this->requestToArray();

    // Requête sql
    $qb = $this->oDatabaseConnection->getQueryAccessor()->createQueryBuilder();

    $qb->update('[SCHEMA].[TABLE]', 'a');

    $aParams = [
      'SCHEMA' => ['value' => $this->oProperties->replacePropertiesInString($aRequestOptions['options']['schema']), 'type' => 'schema_name'],
      'TABLE' => ['value' => $aRequestOptions['options']['table'], 'type' => 'table_name'],
      'FIELD_ID' => ['value' => $aRequestOptions['options']['field_id'], 'type' => 'column_name'],
      'ID_VALUE' => ['value' => $id, 'type' => 'mixed'],
    ];
    foreach ($aRequestOptions['options']['fields'] as $key => $oField) {
      $name = $oField['form_field'][$form];
      if (isset($aData[$key])) {
        $sFieldParamsName = 'FIELD_' . $name;
        $sFieldParamsValue = 'VALUE_' . $name;
        $aParams[$sFieldParamsName] = ['value' => $name, 'type' => 'column_name'];
        $aParams[$sFieldParamsValue] = ['value' => $aData[$key], 'type' => 'mixed'];
        $qb->set('[' . $sFieldParamsName . ']', '[' . $sFieldParamsValue . ']');
      }
    }
    $qb->where($qb->expr()->eq('a.[FIELD_ID]', '[ID_VALUE]'));

    try {
      $qb->executeWithParams($aParams);
    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('Studio : can\'t put form content: ' . $e->getMessage());
      return null;
    }

    return $form;
  }

  /**
   * Met à jour les différents type de formulaire (Default, Published, Perso) si ils sont vide
   * @param mixed $id
   * @param bool $isAutogenerated Si true, force la publication des formulaires
   * @return string|null
   */
  private function saveInitialForms_($id, bool $isAutogenerated = false): ?string{
    $aRequestOptions = $this->getRequestParameters();

    $qb = $this->oDatabaseConnection->getQueryAccessor()->createQueryBuilder();

    if (isset($aRequestOptions['options']['fields']['json'])) {
      $sLastForm = '';
      foreach ($aRequestOptions['options']['fields']['json']['form_field'] as $sForm => $sField) {
        try {
          $qb->select('a.[FIELD]')
            ->from('[SCHEMA].[TABLE]', 'a')
            ->where($qb->expr()->eq('a.[FIELD_ID]', '[ID_VALUE]'));

          $aParams = [
            'FIELD' => ['value' => $sField, 'type' => 'column_name'],
            'SCHEMA' => ['value' => $this->oProperties->replacePropertiesInString($aRequestOptions['options']['schema']), 'type' => 'schema_name'],
            'TABLE' => ['value' => $aRequestOptions['options']['table'], 'type' => 'table_name'],
            'FIELD_ID' => ['value' =>  $aRequestOptions['options']['field_id'], 'type' => 'column_name'],
            'ID_VALUE' => ['value' => $id, 'type' => 'mixed'],
          ];

          $oContent = $qb->executeWithParams($aParams)->fetchAll();
        } catch (\Exception $e) {
          $this->oLoggerManager->get('SQL_ERROR')->error('Studio check empty: can\'t get form content: ' . $e->getMessage());
          return null;
        }

        // Si le chamsp Json est vide on enregistre le formulaire
        if (($isAutogenerated === true) || ($isAutogenerated !== true && $oContent[0][$sField] == '')) {
          $result = $this->updateFields_($id, $sForm);
          if ($result == null) {
            return null;
          }
        }
        $sLastForm = $sForm;
      }
      return $sLastForm;
    }

    return null;
  }
}
