<?php

namespace App\Module\Vmap\WebService\Entity;

use App\Module\Vitis\WebService\Entity\AbstractEntity;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use Doctrine\Common\Collections\ArrayCollection;
use Doctrine\ORM\EntityManager;
use Doctrine\ORM\Mapping\Column;
use Doctrine\ORM\Mapping\DiscriminatorColumn;
use Doctrine\ORM\Mapping\DiscriminatorMap;
use Doctrine\ORM\Mapping\Entity;
use Doctrine\ORM\Mapping\GeneratedValue;
use Doctrine\ORM\Mapping\Id;
use Doctrine\ORM\Mapping\InheritanceType;
use Doctrine\ORM\Mapping\JoinColumn;
use Doctrine\ORM\Mapping\JoinTable;
use Doctrine\ORM\Mapping\ManyToMany;
use Doctrine\ORM\Mapping\ManyToOne;
use Doctrine\ORM\Mapping\SequenceGenerator;
use Doctrine\ORM\Mapping\Table;
use Doctrine\Persistence\Mapping\ClassMetadata;
use Doctrine\Persistence\ObjectManager;

/**
 * Getter et Setter pour les élements venant de la table layer
 *
 * @author Ludivine Premont <ludivine.premont@veremes.com>
 *
 * @Entity(repositoryClass="App\Module\Vmap\WebService\Repository\LayerRepository")
 * @Table(schema="@prop(schema_vmap)",name="layer")
 * @InheritanceType("JOINED")
 * @DiscriminatorColumn(name="layertype", type="string")
 * @DiscriminatorMap({ "postgres" = "LayerPostgres", "wms" = "LayerWMS", "wmts" = "LayerWMTS", "xyz" = "LayerXYZ", "osm" = "LayerOSM", "file" = "LayerFile", "vectortiles" = "LayerVectorTiles", "layerfme" = "LayerFmeWorkbench" })
 */
abstract class AbstractLayer extends AbstractEntity {
  /**
   * Tableau permettant de mapper les classes enfants
   */
  public const LAYER_TYPE_MAP = [
    LayerWMS::TYPE => LayerWMS::class,
    LayerWMTS::TYPE => LayerWMTS::class,
    LayerOSM::TYPE => LayerOSM::class,
    LayerXYZ::TYPE => LayerXYZ::class,
    LayerFile::TYPE => LayerFile::class,
    LayerVectorTiles::TYPE => LayerVectorTiles::class,
    LayerPostgres::TYPE => LayerPostgres::class,
  ];

  /**
   * @var EntityManager|null
   */
  public $oEntityManager;

  /**
   * @var LoggerManager|null
   */
  protected $oLoggerManager;

  /**
   * @var int
   * @Id
   * @Column(type="integer", name="layer_id", nullable=false)
   * @GeneratedValue(strategy="SEQUENCE")
   * @SequenceGenerator(sequenceName="@prop(schema_vmap).layer_layer_id_seq")
   */
  protected $layer_id;

  /**
   * @var string
   * @Column(type="string", length=100, nullable=true)
   */
  protected $name;

  /**
   * @var string
   * @Column(type="string", name="pictogram")
   */
  protected $pictogram;

  /**
   * @var string
   * @Column(type="string", name="legend")
   */
  protected $legend;

  /**
   * @var Datasource|null
   * @ManyToOne(targetEntity="App\Module\Vmap\WebService\Entity\Datasource", inversedBy="layers")
   * @JoinColumn(name="datasource_id", referencedColumnName="datasource_id")
   */
  protected $datasource;

  /**
   * @var int
   * @Column(type="integer")
   */
  protected $datasource_id;

  /**
   * @var \DateTime
   * @Column(type="datetime", name="datemaj", nullable=true)
   */
  protected $datemaj;

  /**
   * @var int
   * @Column(type="integer")
   */
  protected $source_id;

  // /**
  //  * @var string
  //  * @Column(type="string")
  //  */
  // protected $layertype;

  /**
   * @var Source|null
   * @ManyToOne(targetEntity="App\Module\Vmap\WebService\Entity\Source", inversedBy="layers")
   * @JoinColumn(name="source_id", referencedColumnName="source_id")
   */
  protected $source;

  /**
   * @var string
   * @Column(type="string", nullable=true)
   */
  protected $extent;

  /**
   * @var int|null
   * @Column(type="integer", nullable=true)
   */
  protected $min_layer_scale;

  /**
   * @var int|null
   * @Column(type="integer", nullable=true)
   */
  protected $max_layer_scale;

  /**
   * @var bool|null
   * @Column(type="boolean", nullable=true)
   */
  protected $active;

  /**
   * @var bool|null
   * @Column(type="boolean", nullable=true)
   */
  protected $is_background_layer;

  /**
   * @var ThemeLayer|null
   * @ManyToOne(targetEntity="App\Module\Vmap\WebService\Entity\ThemeLayer", inversedBy="layers")
   * @JoinColumn(name="theme_id", referencedColumnName="themelayer_id")
   */
  protected $theme;

  /**
   * @var ArrayCollection<Map>
   * @ManyToMany(targetEntity="Map")
   * @JoinTable(
   *   name="@prop(schema_vmap).layer_map",
   *   joinColumns={@JoinColumn(name="layer_id", referencedColumnName="layer_id")},
   *   inverseJoinColumns={@JoinColumn(name="map_id", referencedColumnName="map_id")}
   * )
   */
  protected $aMaps;

  /**
   * @var int
   * @Column(type="integer")
   */
  protected $theme_id;

  public function __construct() {
    $this->aMaps = new ArrayCollection();
  }

  /**
   * @return int identifiant de la couche
   */
  public function getLayerId() {
    // pas sensé arriver
    // if (gettype($this->layer_id) == 'object') {
    //   return $this->layer_id->getLayerId();
    // }
    return $this->layer_id;
  }

  /**
   * @return string|null nom de la couche
   */
  public function getName(): ?string {
    return $this->name;
  }

  /**
   * @param string $name nom de la couche
   * @return self
   */
  public function setName(?string $name): self {
    $this->name = $name;
    return $this;
  }

  /**
   * @return string|null pictogram
   */
  public function getPictogram(): ?string {
    return $this->pictogram;
  }

  /**
   * @param string $pictogram
   * @return self
   */
  public function setPictogram(?string $pictogram): self {
    $this->pictogram = $pictogram;
    return $this;
  }

  /**
   * @return string|null legend
   */
  public function getLegend(): ?string {
    return $this->legend;
  }

  /**
   * @param string $legend
   * @return self
   */
  public function setLegend(?string $legend): self {
    $this->legend = $legend;
    return $this;
  }
  /**
   * @return array Datasource de la couche
   */
  public function getDatasource(): ?array{
    if (!isset($this->datasource)) {
      return null;
    }
    // return context
    return [
      'datasource_id' => $this->getDatasourceId(),
      'name' => $this->datasource->getName(),
      'description' => $this->datasource->getDescription(),
      'datasourcetype_id' => $this->datasource->getDatasourceTypeId(),
      'datasourcetype' => $this->datasource->getDatasourceType(),
      'pg_database' => $this->datasource->getPgDatabase(),
      'pg_server' => $this->datasource->getPgServer(),
      'pg_port' => $this->datasource->getPgPort(),
      'pg_login' => $this->datasource->getPgLogin(),
      'pg_password' => $this->datasource->getPgPassword(),
      'pg_public' => $this->datasource->getPgPublic(),
      'directory_path' => $this->datasource->getDirectoryPath(),
      'flux_url' => $this->datasource->getFluxUrl(),
      'headers' => $this->datasource->getHeaders(),
      'flux_options' => $this->datasource->getFluxOptions(),
      'flux_login' => $this->datasource->getFluxLogin(),
      'flux_password' => $this->datasource->getFluxPassword(),
      'flux_version' => $this->datasource->getFluxVersion(),
      'flux_type' => $this->datasource->getFluxType(),
    ];
  }

  /**
   * @param Datasource $datasource Datasource de la couche
   * @return self
   */
  public function setDatasource(Datasource $datasource): self {
    $this->datasource = $datasource;
    return $this;
  }

  /**
   * @param int|array $datasource_id Datasource de la couche
   * @param int $datasource_id Datasource de la couche
   * @return self
   */
  public function setDatasourceId(int $datasource_id): self {
    $this->datasource_id = $datasource_id;
    if (isset($this->oEntityManager)) {
      $datasource = $this->oEntityManager->getReference(Datasource::class, $datasource_id);
      $this->setDatasource($datasource);
    }
    return $this;
  }

  /**
   * @return int
   */
  public function getDatasourceId(): ?int {
    return $this->datasource_id;
  }

  /**
   * @return \DateTime|null Date de la dernière mise à jour de la couche
   */
  public function getDatemaj(): ?\DateTime {
    return $this->datemaj;
  }

  /**
   * @param \DateTime $datemaj
   * @return self
   */
  public function setDatemaj(\DateTime $datemaj): self {
    $this->datemaj = $datemaj;
    return $this;
  }

  /**
   * @return array|null Source de la couche
   */
  public function getSource(): ?array{
    if (!isset($this->source)) {
      return null;
    }
    // return context
    return [
      'source_id' => $this->source->getSourceId(),
      'name' => $this->source->getName(),
      'url' => $this->source->getUrl(),
    ];
  }

  /**
   * @param Source $source Source de la couche
   * @return self
   */
  public function setSource(?Source $source): self {
    $this->source = $source;
    return $this;
  }

  /**
   * @param int $source_id Source de la couche
   * @return self
   */
  public function setSourceId(?int $source_id): self {
    $this->source_id = $source_id;
    if (isset($this->oEntityManager)) {
      if (empty($source_id)) {
        $this->setSource(null);
      } else {
        $source = $this->oEntityManager->getReference(Source::class, $source_id);
        $this->setSource($source);
      }
    }
    return $this;
  }

  /**
   * @return int
   */
  public function getSourceId(): ?int {
    return $this->source_id;
  }

  /**
   * @return string|null étendue de la couche
   */
  public function getExtent(): ?string {
    return $this->extent;
  }

  /**
   * @param string $extent étendue de la couche
   * @return self
   */
  public function setExtent(?string $extent): self {
    $this->extent = $extent;
    return $this;
  }

  /**
   * @return int|null Echelle minimum de la couche
   */
  public function getMinLayerScale(): ?int {
    return $this->min_layer_scale;
  }

  /**
   * @param int $min_layer_scale  Echelle minimum de la couche
   * @return self
   */
  public function setMinLayerScale(?int $min_layer_scale): self {
    $this->min_layer_scale = $min_layer_scale;
    return $this;
  }

  /**
   * @return int|null Echelle maximum de la couche
   */
  public function getMaxLayerScale(): ?int {
    return $this->max_layer_scale;
  }

  /**
   * @param int $max_layer_scale  Echelle maximum de la couche
   * @return self
   */
  public function setMaxLayerScale(?int $max_layer_scale): self {
    $this->max_layer_scale = $max_layer_scale;
    return $this;
  }

  /**
   * @return bool|null La couche est active
   */
  public function getActive(): ?bool {
    return $this->active;
  }

  /**
   * @param bool $active La couche est active
   * @return self
   */
  public function setActive($active): self {
    $this->active = $active;
    return $this;
  }

  /**
   * @return bool|null La couche est un fond de plan
   */
  public function getIsBackgroundLayer(): ?bool {
    return $this->is_background_layer;
  }

  /**
   * @param bool $is_background_layer La couche est un fond de plan
   * @return self
   */
  public function setIsBackgroundLayer($is_background_layer): self {
    $this->is_background_layer = $is_background_layer;
    return $this;
  }

  /**
   * @return array|null Theme de la couche
   */
  public function getTheme(): ?array{
    return [
      'theme_id' => !isset($this->theme) ? null : $this->theme->getThemeId(),
      'label' => !isset($this->theme) ? '' : $this->theme->getLabel(),
    ];
  }

  /**
   * @param ThemeLayer $theme Theme de la couche
   * @return self
   */
  public function setTheme(?ThemeLayer $theme): self {
    $this->theme = $theme;
    return $this;
  }

  /**
   * @param int $theme_id Theme de la couche
   * @return self
   */
  public function setThemeId(?int $theme_id): self {
    $this->theme_id = $theme_id;
    if (isset($this->oEntityManager)) {
      if (empty($theme_id)) {
        $this->setTheme(null);
      } else {
        $theme = $this->oEntityManager->getReference(ThemeLayer::class, $theme_id);
        $this->setTheme($theme);
      }
    }
    return $this;
  }

  /**
   * @return int
   */
  public function getThemeId(): ?int {
    return $this->theme_id;
  }

  /**
   * @return string
   */
  public function getLayertype(): ?string {
    foreach (self::LAYER_TYPE_MAP as $sLayerType => $sClassName) {
      if (get_class($this) === $sClassName) {
        return $sLayerType;
      }
    }
    return null;
  }

  /**
   * @return array liste des map_id
   */
  public function getMaps(): array{
    $aMaps = [];

    if (!$this->aMaps->isEmpty()) {
      foreach ($this->aMaps as $oMap) {
        $aMaps[] = [
          'map_id' => $oMap->getMapId(),
          'name' => $oMap->getName(),
        ];
      }
    }
    return $aMaps;
  }

  /**
   * @param string $sType nom du type à chercher
   * @return string|null namespace de la classe enfant ou null si le type n'existe pas
   */
  public static function getLayerChildrenLayerClassName(string $sType): ?string {
    return self::LAYER_TYPE_MAP[$sType] ?? null;
  }

  /**
   * récupère la config de la couche pour MapJsonController
   * @param array $aLayer
   * @return array
   */
  public static function getLayerParams(array $aLayer): array{
    return [
      'layer_id' => $aLayer['layer_id'],
      'type' => $aLayer['layertype'],
      'name' => $aLayer['name'],
      'theme' => $aLayer['theme'],
      'opacity' => 1,
      'visible' => $aLayer['is_visible'] === false ? false : true,
      'source' => $aLayer['source'],
      'extent' => $aLayer['extent'],
      'min_layer_scale' => $aLayer['min_layer_scale'],
      'max_layer_scale' => $aLayer['max_layer_scale'],
      'active' => $aLayer['active'],
      'is_background_layer' => $aLayer['is_background_layer'],
      'index' => $aLayer['layer_index'],
      'is_location' => $aLayer['is_location'],
      'is_map_selection' => $aLayer['is_map_selection'],
      'is_data_grid' => $aLayer['is_data_grid'],
      'is_filter'=>$aLayer['is_filter'],
      'is_map_hidden' => $aLayer['is_map_hidden'],
      'map_parent_hidden' => $aLayer['map_parent_hidden'],
      'legend' => $aLayer['legend'],
    ];
  }

  /* ---------------------------------------------------------------- */
  /* -------------------------- Interfaces -------------------------- */
  /* ---------------------------------------------------------------- */

  /**
   * Permet la récupération d'un gestionnaire d'entité pour interagir avec la base
   * @return void
   */
  public function injectObjectManager(ObjectManager $objectManager, ClassMetadata $classMetadata): void {
    if ($objectManager instanceof EntityManager) {
      $this->oEntityManager = $objectManager;
    }
  }

  /**
   * Permet au gestionnaire d'event de doctrine de charger le service des logger lors du chargement
   * @return void
   */
  public function setLoggerManagerService(LoggerManager $oService): void {
    $this->oLoggerManager = $oService;
  }
}
