<?php

namespace App\Module\Vmap\WebService\Entity;

use App\Module\Vitis\WebService\Entity\AbstractEntity;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use Doctrine\Common\Collections\ArrayCollection;
use Doctrine\ORM\EntityManager;
use Doctrine\ORM\Mapping\Column;
use Doctrine\ORM\Mapping\DiscriminatorColumn;
use Doctrine\ORM\Mapping\DiscriminatorMap;
use Doctrine\ORM\Mapping\Entity;
use Doctrine\ORM\Mapping\GeneratedValue;
use Doctrine\ORM\Mapping\Id;
use Doctrine\ORM\Mapping\InheritanceType;
use Doctrine\ORM\Mapping\JoinColumn;
use Doctrine\ORM\Mapping\JoinTable;
use Doctrine\ORM\Mapping\ManyToMany;
use Doctrine\ORM\Mapping\ManyToOne;
use Doctrine\ORM\Mapping\SequenceGenerator;
use Doctrine\ORM\Mapping\Table;
use Doctrine\Persistence\Mapping\ClassMetadata;
use Doctrine\Persistence\ObjectManager;

/**
 * Getter et Setter pour les élements venant de la table publication
 *
 * @author Laurie Ambroise <laurie.ambroise@veremes.com>
 *
 * @Entity(repositoryClass="App\Module\Vmap\WebService\Repository\PublicationRepository")
 * @Table(schema="@prop(schema_vmap)",name="publication")
 * @InheritanceType("JOINED")
 * @DiscriminatorColumn(name="publicationtype", type="string")
 * @DiscriminatorMap({ "postgres" = "PublicationPostgres", "file" = "PublicationFile" })
 */
abstract class AbstractPublication extends AbstractEntity {
  /**
   * Tableau permettant de mapper les classes enfants
   */
  public const PUBLICATION_TYPE_MAP = [
    PublicationFile::TYPE => PublicationFile::class,
    PublicationPostgres::TYPE => PublicationPostgres::class,
  ];

  /**
   * @var EntityManager|null
   */
  public $oEntityManager;

  /**
   * @var LoggerManager|null
   */
  protected $oLoggerManager;

  /**
   * @var int
   * @id
   * @Column(type="integer", name="publication_id", nullable=false)
   * @GeneratedValue(strategy="SEQUENCE")
   * @SequenceGenerator(sequenceName="@prop(schema_vmap).publication_publication_id_seq")
   */
  protected $publication_id;

  /**
   * @var string
   * @Column(type="string", length=100, nullable=true)
   */
  protected $title;

  /**
   * @var string
   * @Column(type="string", length=100, nullable=true)
   */
  protected $name;

  /**
   * @var Datasource|null
   * @ManyToOne(targetEntity="App\Module\Vmap\WebService\Entity\Datasource", inversedBy="publications")
   * @JoinColumn(name="datasource_id", referencedColumnName="datasource_id")
   */
  protected $datasource;

  /**
   * @var int
   * @Column(type="integer")
   */
  protected $datasource_id;

  /**
   * @var \DateTime
   * @Column(type="datetime", name="datemaj", nullable=true)
   */
  protected $datemaj;

  /**
   * @var string
   * @Column(type="string", nullable=true)
   */
  protected $extent;

  /**
   * @var int|null
   * @Column(type="integer", nullable=true)
   */
  protected $min_layer_scale;

  /**
   * @var int|null
   * @Column(type="integer", nullable=true)
   */
  protected $max_layer_scale;

  /**
   * @var bool|null
   * @Column(type="boolean", nullable=true)
   */
  protected $active;

  /**
   * @var ArrayCollection<Flux>
   * @ManyToMany(targetEntity="Flux")
   * @JoinTable(
   *   name="@prop(schema_vmap).publication_flux",
   *   joinColumns={@JoinColumn(name="publication_id", referencedColumnName="publication_id")},
   *   inverseJoinColumns={@JoinColumn(name="flux_id", referencedColumnName="flux_id")}
   * )
   */
  protected $aFlux;

  public function __construct() {
    $this->aFlux = new ArrayCollection();
  }

  /**
   * @return int identifiant de la publication
   */
  public function getPublicationId() {
    // c'est pas sensé arriver ça
    // if (gettype($this->publication_id) == 'object') {
    //   return $this->publication_id->getPublicationId();
    // }
    return $this->publication_id;
  }

  /**
   * @return string|null titre de la publication
   */
  public function getTitle(): ?string {
    return $this->title;
  }

  /**
   * @param string $title titre de la publication
   * @return self
   */
  public function setTitle(?string $title): self {
    $this->title = $title;
    return $this;
  }

  /**
   * @return string|null nom de la publication
   */
  public function getName(): ?string {
    return $this->name;
  }

  /**
   * @param string $name nom de la publication
   * @return self
   */
  public function setName(?string $name): self {
    $this->name = $name;
    return $this;
  }

  /**
   * @return array Datasource de la publication
   */
  public function getDatasource(): ?array{
    if (!isset($this->datasource)) {
      return null;
    }
    // return context
    return [
      'datasource_id' => $this->getDatasourceId(),
      'name' => $this->datasource->getName(),
      'description' => $this->datasource->getDescription(),
      'datasourcetype_id' => $this->datasource->getDatasourceTypeId(),
      'datasourcetype' => $this->datasource->getDatasourceType(),
      'pg_database' => $this->datasource->getPgDatabase(),
      'pg_server' => $this->datasource->getPgServer(),
      'pg_port' => $this->datasource->getPgPort(),
      'pg_login' => $this->datasource->getPgLogin(),
      'pg_password' => $this->datasource->getPgPassword(),
      'pg_public' => $this->datasource->getPgPublic(),
      'directory_path' => $this->datasource->getDirectoryPath(),
      'flux_url' => $this->datasource->getFluxUrl(),
      'flux_options' => $this->datasource->getFluxOptions(),
      'flux_login' => $this->datasource->getFluxLogin(),
      'flux_password' => $this->datasource->getFluxPassword(),
      'flux_version' => $this->datasource->getFluxVersion(),
      'flux_type' => $this->datasource->getFluxType(),
    ];
  }

  /**
   * @param Datasource $datasource Datasource de la publication
   * @return self
   */
  public function setDatasource(Datasource $datasource): self {
    $this->datasource = $datasource;
    return $this;
  }

  /**
   * @param int|array $datasource_id Datasource de la publication
   * @param int $datasource_id Datasource de la publication
   * @return self
   */
  public function setDatasourceId(int $datasource_id): self {
    $this->datasource_id = $datasource_id;
    if (isset($this->oEntityManager)) {
      $datasource = $this->oEntityManager->getReference(Datasource::class, $datasource_id);
      $this->setDatasource($datasource);
    }
    return $this;
  }

  /**
   * @return int
   */
  public function getDatasourceId(): ?int {
    return $this->datasource_id;
  }

  /**
   * @return \DateTime|null Date de la dernière mise à jour de la publication
   */
  public function getDatemaj(): ?\DateTime {
    return $this->datemaj;
  }

  /**
   * @param \DateTime $datemaj
   * @return self
   */
  public function setDatemaj(\DateTime $datemaj): self {
    $this->datemaj = $datemaj;
    return $this;
  }

  /**
   * @return string|null étendue de la publication
   */
  public function getExtent(): ?string {
    return $this->extent;
  }

  /**
   * @param string $extent étendue de la publication
   * @return self
   */
  public function setExtent(?string $extent): self {
    $this->extent = $extent;
    return $this;
  }

  /**
   * @return int|null Echelle minimum de la publication
   */
  public function getMinLayerScale(): ?int {
    return $this->min_layer_scale;
  }

  /**
   * @param int $min_layer_scale  Echelle minimum de la publication
   * @return self
   */
  public function setMinLayerScale(?int $min_layer_scale): self {
    $this->min_layer_scale = $min_layer_scale;
    return $this;
  }

  /**
   * @return int|null Echelle maximum de la publication
   */
  public function getMaxLayerScale(): ?int {
    return $this->max_layer_scale;
  }

  /**
   * @param int $max_layer_scale  Echelle maximum de la publication
   * @return self
   */
  public function setMaxLayerScale(?int $max_layer_scale): self {
    $this->max_layer_scale = $max_layer_scale;
    return $this;
  }

  /**
   * @return bool|null La publication est active
   */
  public function getActive(): ?bool {
    return $this->active;
  }

  /**
   * @param bool $active La publication est active
   * @return self
   */
  public function setActive($active): self {
    $this->active = $active;
    return $this;
  }

  /**
   * @return string
   */
  public function getPublicationtype(): ?string {
    foreach (self::PUBLICATION_TYPE_MAP as $sPublicationType => $sClassName) {
      if (get_class($this) === $sClassName) {
        return $sPublicationType;
      }
    }
    return null;
  }

  /**
   * @return array liste des flux
   */
  public function getFlux(): array{
    $aFlux = [];

    if (!$this->aFlux->isEmpty()) {
      foreach ($this->aFlux as $oFlux) {
        $aFlux[] = [
          'flux_id' => $oFlux->getFluxId(),
          'name' => $oFlux->getName(),
        ];
      }
    }
    return $aFlux;
  }

  /**
   * @param string $sType nom du type à chercher
   * @return string|null namespace de la classe enfant ou null si le type n'existe pas
   */
  public static function getPublicationChildrenPublicationClassName(string $sType): ?string {
    return self::PUBLICATION_TYPE_MAP[$sType] ?? null;
  }

  /**
   * récupère la config de la publication pour MapJsonController
   * @param array $aPublication
   * @return array
   */
  public static function getPublicationParams(array $aPublication): array{
    return [
      'publication_id' => $aPublication['publication_id'],
      'type' => $aPublication['publicationtype'],
      'name' => $aPublication['name'],
      'title' => $aPublication['title'],
      'datasource_id' => $aPublication['datasource_id'],
      'datemaj' => $aPublication['datemaj'],
      'extent' => $aPublication['extent'],
      'min_layer_scale' => $aPublication['min_layer_scale'],
      'max_layer_scale' => $aPublication['max_layer_scale'],
      'active' => $aPublication['active'],
    ];
  }

  /* ---------------------------------------------------------------- */
  /* -------------------------- Interfaces -------------------------- */
  /* ---------------------------------------------------------------- */

  /**
   * Permet la récupération d'un gestionnaire d'entité pour interagir avec la base
   * @return void
   */
  public function injectObjectManager(ObjectManager $objectManager, ClassMetadata $classMetadata): void {
    if ($objectManager instanceof EntityManager) {
      $this->oEntityManager = $objectManager;
    }
  }

  /**
   * Permet au gestionnaire d'event de doctrine de charger le service des logger lors du chargement
   * @return void
   */
  public function setLoggerManagerService(LoggerManager $oService): void {
    $this->oLoggerManager = $oService;
  }
}
