<?php

namespace App\Module\Vmap\WebService\Repository;

use App\Module\Vitis\WebService\Query\SortableNullValueWalker;
use App\Module\Vitis\WebService\Repository\GenericRepository;
use App\Module\Vitis\WebService\Service\DataAccessor\CommonAccessorFunctions;
use App\Module\Vmap\WebService\Entity\PublicationFile;
use App\Module\Vmap\WebService\Entity\PublicationPostgres;
use Doctrine\DBAL\Connection;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\ORM\Mapping\ClassMetadata;
use Doctrine\ORM\Tools\Pagination\Paginator;

/**
 * Repository des couches de vmap.
 *
 * @author Laurie Ambroise <laurie.ambroise@veremes.com>
 */

class PublicationRepository extends GenericRepository {
  /**
   * {@inheritdoc}
   */
  public function __construct(EntityManagerInterface $em, ClassMetadata $class) {
    parent::__construct($em, $class);
  }

  /**
   * Permet de filtrer et paginer les entités Doctrine.
   * @param array $aParams Objet permettant de filtrer et paginer les entités
   *                          - $aParams.order_by     (string)        : colonne pour trier les entités
   *                          - $aParams.sort_order   (string)        : sens de tri (ASC|DESC)
   *                          - $aParams.limit        (integer)       : nombre d'entité à retourner
   *                          - $aParams.offset       (integer)       : décallage des entité par rapport à la première
   *                          - $aParams.attributs    (string)        : Liste des champs à afficher
   *                          - $aParams.filter       (string)        : Json inline permettant de générer les filtres
   *                          - $aParams.distinct     (bool)          : Pas de doublon sur les entités
   * @param CommonAccessorFunctions $oAccessorFunc service de manipulation des attributs
   * @param string|null $sSchema
   * @param string|null $sTable
   * @return array Liste des entités
   */
  public function search(array $aParams, CommonAccessorFunctions $oAccessorFunc, ?string $sSchema = null, ?string $sTable = null): array
  {
    // initialisation query builder
    $qb = $this->createQueryBuilder(self::ENTITY_ALIAS);
    $sOrderByColumn = null;

    // order_by
    if (isset($aParams['order_by'])) {
      $sSortOrder = (isset($aParams['sort_order'])) ? $aParams['sort_order'] : 'asc';
      $sColumnOrderBy = $oAccessorFunc->getFieldColumn($aParams['order_by']);
      if (isset($sColumnOrderBy)) {
        $sOrderByColumn = self::ENTITY_ALIAS . '.' . $sColumnOrderBy;
        $qb->orderBy(self::ENTITY_ALIAS . '.' . $sColumnOrderBy, $sSortOrder);
      }
    }

    // filter
    $aParams['filter'] = $aParams['filter'] ?? null;

    $filter = (array) json_decode($aParams['filter'], true);
    if (isset($filter['operators'])) {
      $instanceOf = '';
      $filter['operators'] = array_values(array_filter($filter['operators'], function (mixed $operator) use (&$instanceOf) {
        if ($operator['column'] == 'publicationtype') {
          $instanceOf = $operator['value'];
          return false;
        }
        return true;
      }));
      // Instance of sur la bonne classe si filter publicationtype existe
      if (!empty($instanceOf)) {
        switch ($instanceOf) {
          case 'postgres':
            $qb->where($qb->expr()->isInstanceOf(self::ENTITY_ALIAS, PublicationPostgres::class));
            break;
          case 'file':
            $qb->where($qb->expr()->isInstanceOf(self::ENTITY_ALIAS, PublicationFile::class));
            break;
        }
      }
    }

    $filter = json_encode($filter);
    if ($oAccessorFunc->needFilter($filter)) {
      $aFilterReturn = $oAccessorFunc->prepareAndDecodeFilter($filter);

      if (!empty($aFilterReturn['request'])) {
        $qb->andWhere($aFilterReturn['request']);

        foreach ($aFilterReturn['params'] as $sKey => $aValue) {
          if (isset($aValue['type'])) {
            $qb->setParameter($sKey, $aValue['value'], $aValue['type']);
          } else {
            $qb->setParameter($sKey, $aValue['value']);
          }
        }
      }
    }

    $oFullPaginator = new Paginator($qb, true);
    $oFullPaginator->setUseOutputWalkers(false);

    // offset
    if (isset($aParams['offset']) && $aParams['offset'] > 0) {
      $qb->setFirstResult($aParams['offset']);
    }

    // limit
    if (isset($aParams['limit']) && $aParams['limit'] > 0) {
      $qb->setMaxResults($aParams['limit']);
    }
    $query = $qb->getQuery();
    if ($sOrderByColumn !== null) {
      // ajout de NULLS LAST
      $query->setHint(SortableNullValueWalker::HINT, [
        $sOrderByColumn => SortableNullValueWalker::NULLS_LAST,
      ]);
    }

    $aEntities = $query->getResult();

    // Formatage de la réponse
    $aEntities = [
      'data' => $aEntities,
      'total_row_selected' => $oFullPaginator->count(),
      'list_count' => count($aEntities),
      'status' => 1,
    ];

    return $aEntities;
  }
}
