import { ChangeDetectorRef, Component, Inject, Injector, OnInit, Optional } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { TranslateService } from '@ngx-translate/core';
import { environment } from 'src/environments/environment';
import { AppModeConfig } from '../app.model';
import { DashboardComponent } from '../core/components/dashboard/dashboard.component';
import { AuthService } from '../core/services/auth/auth.service';
import { FunctionService } from '../core/services/utils/function/function.service';
import { VitisCheck, VitisClone, VitisRequest } from '../core/services/utils/vitis-utils';
import { Subject, Subscription } from 'rxjs';
import { RoutingService } from '../core/services/utils/routing/routing.service';


@Component({
  selector: 'app-additional-route',
  templateUrl: './additional-route.component.html',
  styleUrls: ['./additional-route.component.scss']
})
export class AdditionalRouteComponent implements OnInit {

  public aSections: Array<any> = [];
  public showRightComponent: boolean = true;
  public sSelectedSectionId;
  public bAvailableSections = true;

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];
  public sidenavOpen: boolean;

  /**
   * Constructeur
   */
  constructor(
    public translate: TranslateService,
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private route: ActivatedRoute,
    private router: Router,
    private auth: AuthService,
    @Optional() @Inject(DashboardComponent) private dashboardComponent: DashboardComponent,
    private routingService: RoutingService,
  ) { }

  /**
   *
   */
  ngOnInit(): void {
    // Traductions (obligatoire sinon ne fonctionne pas dans le module en lazy loading)
    this.translate.addLangs(['en', 'fr']);
    this.translate.setDefaultLang('en');
    this.translate.use(environment.properties.language ? environment.properties.language : 'en');

    if (environment.properties.enable_module_configuration) {
      this.addModuleConfigurationSections();
    }

    this.router.navigate(['.'], { relativeTo: this.route, queryParamsHandling: 'merge' });

    // Check la section en cours
    if (this.route.firstChild) {
      this.subscriptions.push(this.route.firstChild.paramMap.subscribe(params => {
        this.sSelectedSectionId = params.get('id');
      }));
    }

    // Vérifie si le menu des modes est ouvert pour positionner les sections
    if (this.vitisCheck.isDefAndNotNull(this.dashboardComponent)) {
      this.sidenavOpen = this.dashboardComponent.sidenavOpen;
      this.subscriptions.push(this.dashboardComponent.sidenavOpenChange.subscribe((sidenavOpen: any) => {
        this.sidenavOpen = sidenavOpen;
      }));
    }
  }

  /**
   * À la destruction
   */
  ngOnDestroy() {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
  }

  /**
   * Ajout les configurations de module
   */
  private addModuleConfigurationSections(): Promise<Array<any>> {
    return new Promise((resolve, reject) => {
      const sUrl = `${environment.properties.app_api_url}/vitis/mods_config`;
      const aSections: Array<any> = [];
      // Récupère en base les configurations de modules
      this.vitisRequest.ajaxRequestPromise({
        url: sUrl,
        method: 'GET',
        params: { order_by: 'name' }
      }).then((resp: any) => {
        if (resp.data.status === 1 && this.vitisCheck.isDefAndNotNull(resp.data.data)) {
          const modsConfigs = resp.data.data;
          for (let i = 0; i < modsConfigs.length; i++) {
            const modConfig = modsConfigs[i];
            modConfig.role_min = this.vitisCheck.isString(modConfig.role_min) ? [modConfig.role_min] : modConfig.role_min;
            if (this.routingService.canAccessRoute(modConfig.role_min)) {
              // Construit les sections
              aSections.push({
                id: modConfig.mod_configuration_id,
                label: modConfig.name
              });
            }
          }

          if (aSections.length > 0) {
            this.aSections = aSections;
            this.openModuleConf(this.aSections[0]);

            // Check la section en cours
            setTimeout(() => {
              if (this.route.firstChild) {
                this.subscriptions.push(this.route.firstChild.paramMap.subscribe(params => {
                  this.sSelectedSectionId = params.get('id');
                }));
              }
            });
          } else {
            this.bAvailableSections = false;
          }
        }
      }).catch(() => {
        this.bAvailableSections = false;
      });
    });
  }

  /**
   * Ouvre la route
   * @param object
   * @param sFormDefinitionName
   */
  public openRoute(object: any) {
    if (object.route) {
      this.router.navigate(Object.keys(object).map(key => object[key]), { relativeTo: this.route, queryParamsHandling: 'merge' });
    } else {
      this.router.navigate(['.'], { relativeTo: this.route, queryParamsHandling: 'merge' });
    }
  }

  /**
   * Formate l'objet pour lancer la route
   * @param object
   */
  public openModuleConf(object: any) {
    this.openRoute({
      route: 'module-conf',
      id: object.id,
      form: 'update'
    });
  }
}
