import { Component, OnDestroy, OnInit, ViewChild } from '@angular/core';
import { ActivatedRoute } from '@angular/router';
import { vitisAlert, VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { FormreaderComponent } from 'src/app/vitis/formreader/formreader.component';
import { environment } from 'src/environments/environment';
import { FormreaderService } from 'src/app/vitis/formreader/services/formreader.service';
import { TranslateService } from '@ngx-translate/core';

@Component({
  selector: 'app-module-conf-form',
  templateUrl: './module-conf-form.component.html',
  styleUrls: ['./module-conf-form.component.scss']
})
export class ModuleConfFormComponent implements OnInit, OnDestroy {

  @ViewChild(FormreaderComponent, { static: false }) oFormReader: FormreaderComponent;

  private routeParamsSubscription: any;
  public oParams: any;
  public oFormOptions: any;
  public formIsLoad: boolean;
  private routeDataSubscription: any;
  public oOptions: any;

  /**
   *
   */
  constructor(
    private route: ActivatedRoute,
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private formreaderSrvc: FormreaderService,
    public translate: TranslateService,
  ) { }

  /**
   *
   */
  ngOnInit(): void {
    this.routeParamsSubscription = this.route.params.subscribe(oParams => {
      // Les parametres de la route change c'est que l'on change de formulaire donc on considère le formulaire comme non chargé
      this.formIsLoad = false;

      this.oParams = oParams;

      this.getModuleConfiguration().then((oOptions) => {
        this.oOptions = oOptions;

        // Pour que initform soit lancé il faut l'id et les data
        if (this.oParams &&
          oParams.id && this.oOptions) {
          this.initForm();
        }
      });

    });
  }


  /**
   * Destruction du composant
   */
  ngOnDestroy() {
    this.oParams = null;
    if (this.routeParamsSubscription) {
      this.routeParamsSubscription.unsubscribe();
    }
    if (this.routeDataSubscription) {
      this.routeDataSubscription.unsubscribe();
    }
  }

  /**
   * Ajout les configurations de module
   */
  private getModuleConfiguration(): Promise<any> {

    return new Promise((resolve, reject) => {
      const sUrl = `${environment.properties.app_api_url}/vitis/mods_config/${this.oParams.id}`;
      let oOptions;
      // Récupère en base les configurations de modules
      this.vitisRequest.ajaxRequestPromise({
        url: sUrl,
        method: 'GET'
      }).then((resp: any) => {
        const aPrivileges: Array<string> = [];
        if (resp.data.status === 1 && this.vitisCheck.isDefAndNotNull(resp.data.data)) {
          const modConfig = resp.data.data;
          oOptions = {
            json_def: this.vitisCheck.isString(modConfig.json_def) ? JSON.parse(modConfig.json_def) : modConfig.json_def,
            js_def: modConfig.js_def,
            ressource: modConfig.ressource,
            json_properties: modConfig.json_properties
          };
        }
        resolve(oOptions);
      });
    });
  }

  /**
   * Initialise le formulaire
   */
  public initForm() {
    // Charge le JS
    this.setEventContainer().then(() => {
      // Charge le formulaire
      this.oFormOptions = {
        form: this.oOptions.json_def,
        formName: this.oParams.form,
        formValues: {},
      };
      // Le formulaire est chargé
      this.formIsLoad = true;
    });
  }

  /**
   * Prépare le event container et les fonction commune
   */
  private setEventContainer(): Promise<any> {
    return new Promise((resolve, reject) => {
      // Sélectionne l'event container en fonction de l'option formobject de la route
      const eventContainer = this;
      if (this.vitisCheck.isString(this.oOptions.js_def)) {
        this.formreaderSrvc.loadJs(this.oFormReader, this.oOptions.js_def);
      }
      // Ajout l'event container
      this.oOptions.json_def[this.oParams.form].eventContainer = eventContainer;

      resolve(true);
    });
  }


  /**
   * Ajoute les méthodes communes aux formulaires de properties (loadProperties, updateProperties et getProperties)
   * Remplit le formulaire
   */
  public loadProperties() {
    const sRessourceURL = environment.properties.app_api_url + '/vitis/properties';
    let oFormValues = {};
    const jsonProperties = this.oOptions.json_properties ?? [];

    this.vitisRequest.ajaxRequestPromise({
      method: 'GET',
      url: sRessourceURL,
    }).then((resp) => {
      if (resp.data) {
        if (resp.data.status === 1) {
          // Applique au formulaire
          setTimeout(() => {
            oFormValues = resp.data.data;
            for (const key in oFormValues) {
              if (Object.hasOwnProperty.call(oFormValues, key)) {
                if (jsonProperties.indexOf(key) !== -1) {
                  try {
                    oFormValues[key] = JSON.stringify(oFormValues[key], null, 2);
                  } catch (error) { }
                }
              }
            }

            this.oFormReader.setFormValues(oFormValues);
          }, 500);
        }
      }
    });
  }

  /**
   * Ajoute les méthodes communes aux formulaires de properties (loadProperties, updateProperties et getProperties)
   * Fonction de mise à jour des properties
   */
  public updateProperties(oFormValues: any): Promise<boolean> {
    return new Promise((resolve, reject) => {
      const sRessourceURL = environment.properties.app_api_url + '/' + this.oOptions.ressource;
      const jsonProperties = this.oOptions.json_properties ?? [];

      const oData = {};
      for (const key in oFormValues) {
        if (this.vitisCheck.isDefAndNotNull(oFormValues[key])) {
          if (jsonProperties.indexOf(key) !== -1) {
            oData[key] = JSON.parse(oFormValues[key]);
          } else {
            oData[key] = oFormValues[key];
          }
        }
      }

      this.vitisRequest.ajaxRequestPromise({
        method: 'PUT',
        url: sRessourceURL,
        data: oData
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status === 1) {
            this.getProperties().then((oProperties) => {

              // MAJ variable properties
              environment.properties = {
                ...environment.properties,
                ...oProperties
              };

              // toast
              this.translate.get('FORM_PROPERTIES_EDIT_OK').subscribe((translation) => {
                vitisAlert.fireToast({
                  type: 'success',
                  title: translation
                });
                resolve(true);
              });
            }).catch(() => {
              console.error('Could not refresh properties');
            });
          }
        }
      });
    });
  }

  /**
   * Ajoute les méthodes communes aux formulaires de properties (loadProperties, updateProperties et getProperties)
   * Get Properties
   */
  public getProperties(): Promise<any> {
    return new Promise((resolve, reject) => {

      const sRessourceURL = environment.properties.app_api_url + '/vitis/properties';
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: sRessourceURL,
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status === 1) {
            resolve(resp.data.data);
          } else {
            reject();
          }
        } else {
          reject();
        }
      });
    });
  }
}
