import { Component, OnInit } from '@angular/core';
import { ActivatedRoute } from '@angular/router';
import { environment } from 'src/environments/environment';
import { VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import { TranslateService } from '@ngx-translate/core';
import SwaggerUI from 'swagger-ui';

declare let $: any;

@Component({
  selector: 'app-api-doc',
  templateUrl: './api-doc.component.html',
  styleUrls: ['./api-doc.component.scss']
})
export class ApiDocComponent implements OnInit {

  /**
   * Constructeur
   */
  constructor(
    private route: ActivatedRoute,
    private vitisRequest: VitisRequest,
    private auth: AuthService,
    private translate: TranslateService,
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit() {
    this.route.data.subscribe(oParams => {
      // Affiche la première doc
      this.initApiDoc();
    });
  }

  /**
   * Affichage de la doc. Swagger de l'API.
   * @returns string
   */
  private initApiDoc() {
    // Chargement du fichier de définition de l'API.
    this.loadApiDefinition().then((oApiDefinition: any) => {
      oApiDefinition = this.translateApiDefintion(oApiDefinition);
      // Conversion d'une collection Postman vers Swagger.
      this.initApiDefinition(oApiDefinition);
      // Affichage de la doc. Swagger.
      this.initSwaggerUi(oApiDefinition);
    });
  }

  /**
   * Charge le fichier json de définition de l'API au format Swagger.
   * @returns string
   */
  private loadApiDefinition() {
    return new Promise((resolve, reject) => {
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: environment.properties.app_api_url + '/vitis/api_doc'
      }).then((resp: any) => {
        if (resp.status === 200) {
          resolve(resp.data.data);
        } else {
          console.error('Could not get the API definition file from request');
          reject();
        }
      }).catch((err) => {
        reject();
      });
    });
  }

  /**
   * Affichage de la doc. Swagger de l'API.
   * @param  {object} oApiDefinition
   * @returns string
   */
  public initSwaggerUi(oApiDefinition: any) {
    const self = this;
    // Test d'utilisation.
    const ui = SwaggerUI({
      dom_id: '#swagger-ui',
      layout: 'BaseLayout',
      presets: [
        SwaggerUI.presets.apis,
        SwaggerUI.SwaggerUIStandalonePreset
      ],
      spec: oApiDefinition,
      docExpansion: 'none',
      operationsSorter: 'alpha',
      tagsSorter: 'alpha',
      onComplete: function () {
        setTimeout(() => {
          self.translateSwaggerComponent();
        }, 100);
      }
    });
  }

  /**
   * Initialisation des paramètres de la doc.
   * @param  {object} oApiDefinition
   * @returns void
   */
  public initApiDefinition(oApiDefinition: any) {
    // Host de l'API Rest.
    const iSlashLastIndex = environment.properties.app_api_url.lastIndexOf('/');
    oApiDefinition.host = environment.properties.app_api_url.substring(0, iSlashLastIndex).replace('https://', '');
    // Version de l'API Rest.
    oApiDefinition.info.version = environment.properties.version;
    oApiDefinition.info.title = `${environment.properties.application}`.toUpperCase();
    // Définition du paramètre 'Authorization' avec le token de l'utilisateur.
    this.setAllParameters(oApiDefinition, 'Authorization', this.auth.getToken());
  }

  /**
   * Définition d'un paramètre dans tous les web services.
   * @param  {object} oApiDefinition
   * @param  {string} sParameterName
   * @param  {any} value
   * @returns void
   */
  public setAllParameters(oApiDefinition: any, sParameterName: string, value: any) {
    let oParameter: any, aMethods: any, aParameters: Array<any>;
    const aWebServices: any = Object.keys(oApiDefinition.paths);
    for (let i = 0; i < aWebServices.length; i++) {
      aMethods = Object.keys(oApiDefinition.paths[aWebServices[i]]);
      for (let j = 0; j < aMethods.length; j++) {
        aParameters = oApiDefinition.paths[aWebServices[i]][aMethods[j]].parameters;
        for (let k = 0; k < aParameters.length; k++) {
          oParameter = aParameters[k];
          if (oParameter.name == sParameterName) {
            // oParameter.default = value;
            oParameter.value = value;
            break;
          }
        }
      }
    }
  }

  /**
   * traduit les clé dans la défintion venant de l'API
   * @param oDef
   * @returns oDef
   */
  private translateApiDefintion(oDef: any): any {

    for (const index of Object.keys(oDef.tags)) {
      if (oDef.tags[index].name) {
        oDef.tags[index].name = this.translate.instant(oDef.tags[index].name);
      }
      if (oDef.tags[index].description) {
        oDef.tags[index].description = this.translate.instant(oDef.tags[index].description);
      }
    }

    if (!oDef.paths) {
      oDef.paths = [];
    }
    for (const path of Object.keys(oDef.paths)) {
      for (const method of Object.keys(oDef.paths[path])) {
        const itemDef = oDef.paths[path][method];
        if (itemDef.summary) {
          itemDef.summary = this.translate.instant(itemDef.summary);
        }

        if (itemDef.description) {
          itemDef.description = this.translate.instant(itemDef.description);
        } else {
          itemDef.description = '';
        }

        if (itemDef.public == true) {
          itemDef.description += '<br>';
          itemDef.description += this.translate.instant('SWAGGER_DESCRIPTION_ADDON_PUBLIC');
        }

        if (itemDef.roles) {
          itemDef.description += '<br>';
          itemDef.description += this.translate.instant('SWAGGER_DESCRIPTION_ADDON_ROLES_LIST');
          itemDef.description += '<ul>';
          for (const sRole of itemDef.roles) {
            itemDef.description += '<li>' + sRole + '</li>';
          }
          itemDef.description += '</ul>';
        }

        if (itemDef.dynamic) {
          itemDef.description += '<br>';
          itemDef.description += this.translate.instant('SWAGGER_DESCRIPTION_ADDON_DYNAMIC_RETURN');
        }


        for (const code in itemDef.responses) {
          if (itemDef.responses[code].description) {
            itemDef.responses[code].description = this.translate.instant(itemDef.responses[code].description);
          }
        }
        for (const index in itemDef.tags) {
          if (itemDef.tags[index]) {
            itemDef.tags[index] = this.translate.instant(itemDef.tags[index]);
          }
        }

        for (const index in itemDef.parameters) {
          if (itemDef.parameters[index].description) {
            itemDef.parameters[index].description = this.translate.instant(itemDef.parameters[index].description);
          }
        }

        oDef.paths[path][method] = itemDef;
      }
    }


    return oDef;
  }

  /**
   * ajoute une classe correpondant à la langue pour traduirte les constante de swagger
   */
  private translateSwaggerComponent(): void {
    $('.swagger-ui').addClass(environment.properties.language);
  }
}
