import { NgModule } from '@angular/core';
import { Router, Route, Routes, RouterModule } from '@angular/router';
import { environment } from 'src/environments/environment';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import { AuthGuardService as AuthGuard } from 'src/app/vitis/core/services/auth/auth-guard.service';
import { RoleGuardService as RoleGuard } from 'src/app/vitis/core/services/auth/role-guard.service';
import { CanDeactivateGuard } from 'src/app/vitis/core/services/deactivate/can-deactivate.guard';
import { DashboardComponent } from 'src/app/vitis/core/components/dashboard/dashboard.component';
import { GenericComponent } from 'src/app/vitis/core/components/generic/generic.component';
import { ConfigurationSupervisionComponent } from 'src/app/vitis/core/components/supervision/supervision.component';
import { ErrorPageComponent } from 'src/app/vitis/core/components/error-page/error-page.component';
import { WidgetComponent } from './core/components/widget/widget.component';
import { TranslateService, LangChangeEvent } from '@ngx-translate/core';

import { oModules, oModulesComponents } from '../modules/modules-components';
import { AppModeConfig } from './app.model';
import { VitisCheck, VitisClone } from './core/services/utils/vitis-utils';
import { CustomTranslationService } from 'src/app/vitis/customtranslation/services/customtranslation.service';
import { RoutingService } from './core/services/utils/routing/routing.service';

const routes: Routes = [
  {
    path: 'login',
    loadChildren: () => import('./login/login.module').then(m => m.LoginModule),
  },
  {
    path: 'oauth2',
    loadChildren: () => import('./oauth2/oauth2.module').then(m => m.OAuth2Module),
  },
  {
    path: 'dashboard',
    component: DashboardComponent,
    canActivate: [AuthGuard]
  },
  {
    path: 'widget',
    component: WidgetComponent,
    canActivate: [AuthGuard]
  },
  // page pour les erreur de tokens expirés sur les widgets
  { path: 'error/:key', component: ErrorPageComponent },
  { path: '', redirectTo: 'dashboard', pathMatch: 'full' }
];

@NgModule({
  imports: [RouterModule.forRoot(routes, { relativeLinkResolution: 'legacy' })],
  exports: [RouterModule]
})
export class AppRoutingModule {

  private aWidgetRoutes_: Array<any>;
  private aModeRoutes_: Array<any>;

  /**
   * Constructeur
   */
  constructor(
    private router: Router,
    private auth: AuthService,
    private vitisCheck: VitisCheck,
    private vitisClone: VitisClone,
    private translate: TranslateService,
    private customTranslationService: CustomTranslationService,
    private routingService: RoutingService,
  ) {
    // Charge les routes
    this.loadDashboardRoutes();

    // Recharge les routes lors de la connexion de l'utilisateur (permet d'utiliser les privileges)
    this.auth.onLogin.subscribe((bLogged: boolean) => {
      if (bLogged) {
        this.loadDashboardRoutes();
      }
    });

    // Recharge les routes lors d'un changement de privilèges
    this.auth.onPrivilegesChange.subscribe((aPrivilege: Array<string>) => {
      this.loadDashboardRoutes();
    });
  }

  /**
   * Charge les routes
   * @returns void
   */
  public loadDashboardRoutes(): void {
    // Permet de ne pas charger deux fois les routes et causer des problèmes pour les sections
    if (!this.aModeRoutes_) {

      // Routes des modes définis dans app.json
      this.aModeRoutes_ = this.getModesRoutes_(environment.appStruct.modes);

      // Routes des widgets définis dans app.json
      this.aWidgetRoutes_ = this.getModesRoutes_(environment.appStruct.widgets);

      // Routes additionnelles
      if (environment.appStruct.routes &&
        environment.appStruct.routes[0]) {
        this.aModeRoutes_ = this.addAdditionalRoutes(environment.appStruct.routes, this.aModeRoutes_);
      }
    }

    if (this.aModeRoutes_) {
      if (this.aModeRoutes_[0]) {
        if (this.aModeRoutes_[0].path) {

          // Premier des modes
          const defaultMode = this.getDefaultRoute_(this.aModeRoutes_);

          let defaultRouteExists = false;

          // Ajoute les routes de dashboard
          this.router.config.forEach(root => {
            if (root.path === 'dashboard') {
              root.children = [];
              for (let i = 0; i < this.aModeRoutes_.length; i++) {
                const element = this.aModeRoutes_[i];
                root.children.push(element);
              }
              root.children.push({ path: '**', redirectTo: defaultMode });
            }
            if (root.path === 'widget') {
              root.children = [];
              for (let i = 0; i < this.aWidgetRoutes_.length; i++) {
                const element = this.aWidgetRoutes_[i];
                root.children.push(element);
              }
              root.children.push({ path: '**', redirectTo: defaultMode });
            }
            if (root.path === '**') {
              defaultRouteExists = true;
              root.redirectTo = 'dashboard/' + defaultMode;
            }
          });

          // Route par défaut
          if (!defaultRouteExists) {
            this.router.config.push({ path: '**', redirectTo: 'dashboard/' + defaultMode });
          }

          // Chargement des traductions personnalisées.
          if (this.vitisCheck.isDefAndNotNull(this.translate.translations) &&
            Object.keys(this.translate.translations).length > 0) {
            this.customTranslationService.loadCustomTranslations();
          } else {
            this.translate.onLangChange.subscribe((event: LangChangeEvent) => {
              this.customTranslationService.loadCustomTranslations();
            });
          }
        } else {
          console.error('this.aModeRoutes_[0].path not defined');
        }
      } else {
        console.error('this.aModeRoutes_[0] not defined');
      }
    } else {
      console.error('this.aModeRoutes_ not defined');
    }
  }

  /**
   * Ajoute les routes additionnelles
   * @param aAdditionalRoutes
   * @param aModeRoutes
   */
  public addAdditionalRoutes(aAdditionalRoutes: Array<AppModeConfig>, aModeRoutes: Array<any>): any {

    for (let i = 0; i < aAdditionalRoutes.length; i++) {
      const aRoutes: Routes = this.getModeRoute_(aAdditionalRoutes[i]);

      // Ajoute les routes additionnelles
      for (let ii = 0; ii < aRoutes.length; ii++) {
        const additionalRoute = aRoutes[ii];

        // Supprime si la route est déjà présente
        for (let iii = aModeRoutes.length - 1; iii >= 0; iii--) {
          const oModeRoute = aModeRoutes[iii];
          if (oModeRoute.path === additionalRoute.path) {
            aModeRoutes.splice(iii, 1);
          }
        }

        // Ajoute la nouvelle route
        aModeRoutes.push(additionalRoute);
      }
    }

    return aModeRoutes;
  }

  /**
   * Retourne la première route sur laquelle l'utilisateur a des droits
   * @param  {Array<any>} aModesRoutes
   * @returns any
   */
  private getDefaultRoute_(aModesRoutes: Array<any>): any {
    const aUserPrivileges = this.auth.getPrivileges();
    if (aUserPrivileges) {
      for (let i = 0; i < aModesRoutes.length; i++) {
        if (aModesRoutes[i].data) {
          if (aModesRoutes[i].data.privilege) {
            if (this.routingService.canAccessRoute(aModesRoutes[i].data.privilege, aModesRoutes[i].data.privilege_array_mode) &&
              !aModesRoutes[i].path.includes(':')) {
              return aModesRoutes[i].path;
            }
          }
        }
      }
      console.error('No route with this user privileges', aUserPrivileges);
    }

    return '';
  }

  /**
   * Récupère les routes définies dans les modes
   * @returns Routes
   */
  private getModesRoutes_(aModes: Array<any>): Routes {

    let aModesRoutes: Routes = [];

    // Remplit aModesRoutes
    if (aModes) {
      if (aModes.length && aModes[0]) {

        // Ajoute les différents modes
        for (let i = 0; i < aModes.length; i++) {

          const aModeRoutes = this.getModeRoutes_(aModes[i]);

          if (aModeRoutes) {
            if (aModeRoutes.length && aModeRoutes[0]) {
              aModesRoutes = [...aModesRoutes, ...aModeRoutes];
            }
          }
        }
      }
    }

    return aModesRoutes;
  }

  /**
   * Récupère la/les routes définies dans le mode
   * @returns Routes
   */
  private getModeRoutes_(oMode: any): Routes {

    let aModeRoutes: Routes = [];

    // Objets dans le mode
    if (oMode.objects) {
      if (oMode.objects.length && oMode.objects[0]) {

        // Ajoute les routes des objets
        for (let i = 0; i < oMode.objects.length; i++) {
          const oObject = oMode.objects[i];

          // Ajoute le parent à la route de l'objet
          if (oObject.route) {
            // Uniquement si la route du parent n'a pas déjà été ajoutée
            if (!oObject.defaultRoute) {
              oObject.defaultRoute = oObject.route;
              oObject.route = oMode.route + '/' + oObject.route;
            }
          }

          // Récupère la route de l'objet
          const aObjectRoutes = this.getModeRoutes_(oObject);

          // Incorpore aux routes du parent
          if (aObjectRoutes) {
            if (aObjectRoutes.length && aObjectRoutes[0]) {
              aModeRoutes = [...aModeRoutes, ...aObjectRoutes];
            }
          }
        }

        // Crée une redirection vers le premier des objets
        if (aModeRoutes[0]) {
          if (aModeRoutes[0].path) {
            aModeRoutes.push({
              path: oMode.route,
              redirectTo: aModeRoutes[0].path
            });
          }
        }
      }
    } else {
      aModeRoutes = this.getModeRoute_(oMode);
    }

    return aModeRoutes;
  }

  /**
   * Récupère la route du mode / objet
   * @returns Routes
   */
  private getModeRoute_(oMode: any): Routes {

    const aModeRoutes: Routes = [];

    // Type de composant à utiliser
    let modeComponent: any;
    switch (oMode.type) {
      case 'directory':
        return [{
          path: oMode.route,
          loadChildren: () => import('./directory/directory.module').then(m => m.DirectoryModule),
          canActivate: [RoleGuard],
          data: oMode
        }];
        break;
      case 'sectionSimpleform':
        // Ajoute les routes d'édition des objets
        const aSFListRoutes = this.getListRoutes_(oMode);
        return [{
          path: oMode.route,
          loadChildren: () => import('./simpleform/simpleform.module.routing').then(m => m.SimpleformWithRoutingModule),
          canActivate: [RoleGuard],
          data: oMode
        }, ...aSFListRoutes];
        break;
      case 'simpleform':
        return [{
          path: oMode.route,
          loadChildren: () => import('./simpleform/simpleform.module.routing').then(m => m.SimpleformWithRoutingModule),
          canActivate: [RoleGuard],
          data: oMode
        }];
        break;
      case 'displayform':
        return [{
          path: oMode.route,
          loadChildren: () => import('./displayform/displayform.module').then(m => m.DisplayformModule),
          canActivate: [RoleGuard],
          data: oMode
        }];
        break;
      case 'studio':
        return [{
          path: oMode.route,
          loadChildren: () => import('./studio/studio.module').then(m => m.StudioModule),
          canActivate: [RoleGuard],
          data: oMode
        }];
        break;
      case 'workspacelist':
        // Ajoute les routes d'édition des objets
        const aWKListRoutes = this.getListRoutes_(oMode);
        return [{
          path: oMode.route,
          loadChildren: () => import('./workspacelist/workspacelist.module.routing')
            .then(m => m.WorkspacelistWithRoutingModule),
          canActivate: [RoleGuard],
          data: oMode
        }, ...aWKListRoutes];
        break;
      case 'workspacerelationlist':
        const aWKRListRoutes = this.getListRoutes_(oMode);
        return [{
          path: oMode.route,
          loadChildren: () => import('./workspacerelationlist/workspacerelationlist.module.routing').then(m => m.WorkspaceRelationlistWithRoutingModule),
          canActivate: [RoleGuard],
          data: oMode
        }, ...aWKRListRoutes];
        break;
      case 'domainlist':
        return [{
          path: oMode.route,
          loadChildren: () => import('./domainlist/domainlist.module').then(m => m.DomainlistModule),
          canActivate: [RoleGuard],
          data: oMode
        }];
        break;
      case 'logs':
        return [{
          path: oMode.route,
          loadChildren: () => import('./logs/logs.module').then(m => m.LogsModule),
          canActivate: [RoleGuard],
          data: oMode
        }];
        break;
      case 'separator':
        return [];
        break;
      case 'config_supervision':
        modeComponent = ConfigurationSupervisionComponent;
        break;
      case 'api_doc':
        return [{
          path: oMode.route,
          loadChildren: () => import('./api-doc/api-doc.module').then(m => m.ApiDocModule),
          canActivate: [RoleGuard],
          data: oMode
        }];
        break;
      case 'module':
        if (oMode.options &&
          oMode.options.module_name &&
          oModules &&
          oModules[oMode.options.module_name]) {
          const moduleRoute = {
            path: oMode.route,
            loadChildren: () => oModules[oMode.options.module_name],
            data: oMode,
            canActivate: [RoleGuard],
          };
          if (oModules[oMode.options.module_name].loadChildren) {
            moduleRoute.loadChildren = oModules[oMode.options.module_name].loadChildren;
          }
          return [moduleRoute];
        }
        break;
      case 'module_component':
        if (oMode.options &&
          oMode.options.component_name &&
          oModulesComponents &&
          oModulesComponents[oMode.options.component_name]) {
          modeComponent = oModulesComponents[oMode.options.component_name];
        }
        break;
      case 'draganddrop-grid':
        const aDDGListRoutes = this.getListRoutes_(oMode);
        return [{
          path: oMode.route,
          loadChildren: () => import('./draganddrop-grid/draganddrop-grid.module.routing').then(m => m.DragAndDropGridWithRoutingModule),
          canActivate: [RoleGuard],
          data: oMode
        }, ...aDDGListRoutes];
        break;
      case 'customtranslation':
        const aCTRoutes = this.getListRoutes_(oMode);
        return [{
          path: oMode.route,
          loadChildren: () => import('./customtranslation/customtranslation.module.routing')
            .then(m => m.CustomTranslationWithRoutingModule),
          canActivate: [RoleGuard],
          data: oMode
        }, ...aCTRoutes];
        break;
      case 'additionalRoute':
        const aARListRoutes = this.getListRoutes_(oMode);
        return [{
          path: oMode.route,
          loadChildren: () => import('./additional-route/additional-route.module.routing').then(m => m.AdditionalRouteWithRoutingModule),
          canActivate: [RoleGuard],
          data: oMode
        }, ...aARListRoutes];
        break;
      default:
        modeComponent = GenericComponent;
        break;
    }

    // Ajout du chemin vers le mode
    aModeRoutes.unshift({
      path: oMode.route,
      component: modeComponent,
      canActivate: [RoleGuard],
      canDeactivate: [CanDeactivateGuard],
      data: oMode
    });

    return aModeRoutes;
  }

  /**
   * Retourne les routes d'édition des objets
   * @param  {any} oMode
   * @returns Routes
   */
  private getListRoutes_(oMode: any): Routes {

    let aModeRoutes: Routes = [];

    // Ajoute les routes vers les formulaires
    if (oMode.options) {

      if (oMode.sections) {
        /**
         * Si sections existe avec map_field et form_path
         * Sections devient un object qu'on modifie en tableau d'object
         */
        const getObjectsKeys: Array<string> = Object.keys(oMode.sections);
        let sections = [];
        if (getObjectsKeys[0] != '0') {
          sections = getObjectsKeys.reduce((acc, curr) => {
            return [...new Set([...acc, ...oMode.sections[curr]])];
          }, []);
        }
        sections = sections.length ? sections : oMode.sections;
        for (let i = 0; i < sections.length; i++) {

          const oSection = sections[i];
          const sSectionRoute = oSection.route;

          if (oMode.options.display === true
            || oMode.options.update === true
            || oMode.options.insert === true
            || typeof oMode.options.display == 'string'
            || typeof oMode.options.update == 'string'
            || typeof oMode.options.insert == 'string') {

            // Renseigne le mode parent
            oSection.section_parent_mode = oMode;

            oSection.route = oMode.route + '/:form' + '/' + sSectionRoute;
            aModeRoutes = [...aModeRoutes, ...this.getModeRoute_(oSection)];

            if (oMode.options.id_field) {
              oSection.route = oMode.route + '/:id/:form' + '/' + sSectionRoute;
              aModeRoutes = [...aModeRoutes, ...this.getModeRoute_(oSection)];
            }
          } else {
            // Renseigne le mode parent
            oSection.section_parent_mode = oMode;

            oSection.route = oMode.route + '/' + sSectionRoute;
            aModeRoutes = [...aModeRoutes, ...this.getModeRoute_(oSection)];
          }
        }
      } else if (oMode.options.form_path ||
        oMode.options.edition_component ||
        oMode.options.edition_module) {
        if (oMode.options.display === true
          || oMode.options.update === true
          || oMode.options.insert === true
          || typeof oMode.options.display == 'string'
          || typeof oMode.options.update == 'string'
          || typeof oMode.options.insert == 'string') {

          // Composant custom
          let oCustomComponent: any;
          if (oMode.options.edition_component &&
            oModulesComponents[oMode.options.edition_component]) {
            oCustomComponent = oModulesComponents[oMode.options.edition_component];
          }

          // Module custom
          let oCustomModule: any;
          if (oMode.options.edition_module &&
            oModules[oMode.options.edition_module]) {
            oCustomModule = oModules[oMode.options.edition_module];
          }

          // Options custom
          const oEditionMode = this.vitisClone.clone(oMode);
          if (oMode.options.edition_options) {
            oEditionMode.options = oMode.options.edition_options;
          }

          // Sous formulaires
          let parentRoute = oMode.route;
          if (parentRoute.indexOf('/:form/') !== -1) {
            parentRoute = parentRoute.replace('/:form/', '/:form_parent/');
            parentRoute = parentRoute.replace('/:id/', '/:id_parent/');
          }

          if (oCustomComponent) {
            aModeRoutes.push({
              path: parentRoute + '/:form',
              component: oCustomComponent,
              canActivate: [RoleGuard],
              data: oEditionMode
            });
            aModeRoutes.push({
              path: parentRoute + '/:id/:form',
              component: oCustomComponent,
              canActivate: [RoleGuard],
              canDeactivate: [CanDeactivateGuard],
              data: oEditionMode
            });
          } else if (oCustomModule) {
            aModeRoutes.push({
              path: parentRoute + '/:form',
              loadChildren: () => oCustomModule,
              canActivate: [RoleGuard],
              data: oEditionMode
            });
            aModeRoutes.push({
              path: parentRoute + '/:id/:form',
              loadChildren: () => oCustomModule,
              canActivate: [RoleGuard],
              data: oEditionMode
            });
          } else {
            // Utilisation SimpleFormModule
            aModeRoutes.push({
              path: parentRoute + '/:form',
              loadChildren: () => import('./simpleform/simpleform.module.routing').then(m => m.SimpleformWithRoutingModule),
              canActivate: [RoleGuard],
              data: oMode
            });
            aModeRoutes.push({
              path: parentRoute + '/:id/:form',
              loadChildren: () => import('./simpleform/simpleform.module.routing').then(m => m.SimpleformWithRoutingModule),
              canActivate: [RoleGuard],
              data: oMode
            });
          }
        }
      }

      // Objets standalone
      if (oMode.standaloneObjects) {
        for (let i = 0; i < oMode.standaloneObjects.length; i++) {
          const oObject = oMode.standaloneObjects[i];

          oObject.route = oMode.route + '/' + oObject.route;
          aModeRoutes = [...aModeRoutes, ...this.getModeRoute_(oObject)];
        }
      }
    }

    return aModeRoutes;
  }
}
