import '@angular/compiler';
import { Component } from '@angular/core';
import { environment } from 'src/environments/environment';
import { TranslateService } from '@ngx-translate/core';
import { VitisCheck, VitisRequest, VitisClone } from './core/services/utils/vitis-utils';
import { InterfaceService } from './core/services/utils/interface/interface.service';
import { LocalStorageService } from './core/services/utils/localstorage/local-storage.service';
import { NavigationEnd, Router } from '@angular/router';
import { first, filter } from 'rxjs/operators';
import { vitisAlert } from './core/services/utils/alert/vitisalert';

@Component({
  selector: 'app-root',
  templateUrl: './app.component.html',
  styleUrls: ['./app.component.scss']
})
export class AppComponent {

  public isAcceptedCookie: boolean = false;
  private unactivatedCookiesRoutes = ['/widget'];

  /**
   * Constructeur
   */
  constructor(
    private translate: TranslateService,
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private interfaceService: InterfaceService,
    private localStorageService: LocalStorageService,
    private router: Router
  ) {

    // Change le titre de l'application
    if (environment.appStruct.name) {
      document.title = environment.appStruct.name;
    }

    // Change le favicon
    if (environment.appStruct.favicon) {
      const link: any = document.querySelector('link[rel*=\'icon\']') || document.createElement('link');
      link.type = 'image/x-icon';
      link.rel = 'shortcut icon';
      link.href = environment.appStruct.favicon;
      document.getElementsByTagName('head')[0].appendChild(link);
    }
    // Traductions
    this.translate.addLangs(['en', 'fr']);
    this.translate.setDefaultLang('en');

    // Properties
    const serverProperties = this.getPropertiesFromServerSync_();
    environment.properties = { ...serverProperties, ...environment.properties };
    // Sélection de la langue d'après les properties
    this.translate.use(environment.properties.language ? environment.properties.language : 'en');

    // Initialise le thème de l'application
    this.interfaceService.toggleTheme();

    // Ajout le css personnalisé
    this.interfaceService.setDynamicStyle();

    // Cookies acceptés
    this.isAcceptedCookie = this.localStorageService.get('vitis_accepted_cookie');

    // Routes pour les quelles le bandeau de validation de cookies est désactivé
    this.router.events.pipe(filter((event: any) => event instanceof NavigationEnd)).pipe(first()).subscribe(event => {
      if (event && event.url && this.unactivatedCookiesRoutes) {
        for (let i = 0; i < this.unactivatedCookiesRoutes.length; i++) {
          const route = this.unactivatedCookiesRoutes[i];
          if (event.url.startsWith(route)) {
            this.isAcceptedCookie = true;
          }
        }
      }
    });
  }

  /**
   * Effectue une requête synchrone et récupère les properties depuis le serveur
   * @returns Object
   */
  private getPropertiesFromServerSync_(): Object {

    let properties = {};
    this.vitisRequest.ajaxRequest({
      async: false,
      responseType: '',
      method: 'GET',
      url: `${environment.properties.app_api_url}/vitis/properties`,
      success: (resp: any) => {
        if (this.vitisCheck.isString(resp.data)) {
          const oResult = JSON.parse(resp.data);
          if (oResult.data.VM_STATUS) {
            properties = oResult.data;
          }
        }
      },
      error: (resp: any) => {
        if (this.vitisCheck.isString(resp.data)) {
          const oResult = JSON.parse(resp.data);
          this.translate.get(oResult.errorMessage).subscribe((translation) => {
            const options: any = {
              type: 'error',
              title: translation
            };
            vitisAlert.alert(options);
          });
        }
      }
    });

    return properties;
  }

  /**
   * Sauvagarde l'acceptation des cookies dans le localstorage
   */
  public acceptCookie() {
    this.localStorageService.set('vitis_accepted_cookie', true);
    this.isAcceptedCookie = this.localStorageService.get('vitis_accepted_cookie');
  }
}
