import { Injectable } from '@angular/core';
import { Subject, Observable } from 'rxjs';
import { environment } from 'src/environments/environment';

@Injectable({
  providedIn: 'root'
})
export class CheckService {
  isClassRegex: RegExp = /^class\s|^function\s+[A-Z]/;
  isConventionalClassRegex: RegExp = /^function\s+[A-Z]/;
  isNativeClassRegex: RegExp = /^class\s/;

  /**
   * Constructeur
   */
  constructor() { }

  /**
   * Get the object type string
   * @param {*} value
   * @returns {string}
   */
  public getObjectType(value: any): string {
    return Object.prototype.toString.call(value);
  }

  /**
   * Checks to see if a value is an object
   * @param {*} value
   * @returns {boolean}
   */
  public isObject(value: any): boolean {
    // null is object, hence the extra check
    return value !== null && typeof value === 'object';
  }

  /**
   * Checks to see if a value is an object and only an object
   * @param {*} value
   * @returns {boolean}
   */
  public isPlainObject(value: any): boolean {
    return this.isObject(value) && value.__proto__ === Object.prototype;
  }

  /**
   * Checks to see if a value is empty
   * @param {*} value
   * @returns {boolean}
   */
  public isEmpty(value: any): boolean {
    return value == null;
  }

  /**
   * Is empty object
   * @param {*} value
   * @returns {boolean}
   */
  public isEmptyObject(value: any): boolean {
    // We could use Object.keys, but this is more effecient
    for (const key in value) {
      if (value.hasOwnProperty(key)) {
        return false;
      }
    }
    return true;
  }

  /**
   * Is ES6+ class
   * @param {*} value
   * @returns {boolean}
   */
  public isNativeClass(value: any): boolean {
    // NOTE TO DEVELOPER: If any of this changes, isClass must also be updated
    return (
      typeof value === 'function' && this.isNativeClassRegex.test(value.toString())
    );
  }

  /**
   * Is Conventional Class
   * Looks for function with capital first letter MyClass
   * First letter is the 9th character
   * If changed, isClass must also be updated
   * @param {*} value
   * @returns {boolean}
   */
  public isConventionalClass(value: any): boolean {
    return (
      typeof value === 'function' &&
      this.isConventionalClassRegex.test(value.toString())
    );
  }

  /**
   * Is Class
   * @param {*} value
   * @returns {boolean}
   */
  public isClass(value: any): boolean {
    return typeof value === 'function' && this.isClassRegex.test(value.toString());
  }

  /**
   * Checks to see if a value is an error
   * @param {*} value
   * @returns {boolean}
   */
  public isError(value: any): boolean {
    return value instanceof Error;
  }

  /**
   * Checks to see if a value is a date
   * @param {*} value
   * @returns {boolean}
   */
  public isDate(value: any): boolean {
    return this.getObjectType(value) === '[object Date]';
  }

  /**
   * Checks to see if a value is an arguments object
   * @param {*} value
   * @returns {boolean}
   */
  public isArguments(value: any): boolean {
    return this.getObjectType(value) === '[object Arguments]';
  }

  /**
   * Checks to see if a value is a function but not an asynchronous function
   * @param {*} value
   * @returns {boolean}
   */
  public isSyncFunction(value: any): boolean {
    return this.getObjectType(value) === '[object Function]';
  }

  /**
   * Checks to see if a value is an asynchronous function
   * @param {*} value
   * @returns {boolean}
   */
  public isAsyncFunction(value: any): boolean {
    return this.getObjectType(value) === '[object AsyncFunction]';
  }

  /**
   * Checks to see if a value is a function
   * @param {*} value
   * @returns {boolean}
   */
  public isFunction(value: any): boolean {
    return this.isSyncFunction(value) || this.isAsyncFunction(value);
  }

  /**
   * Checks to see if a value is an observable
   * @param {*} value
   * @returns {boolean}
   */
  public isObservabe(value: any): boolean {
    if (value instanceof Subject ||
      value instanceof Observable) {
      return true;
    } else {
      return false;
    }
  }

  /**
   * Checks to see if a value is an regex
   * @param {*} value
   * @returns {boolean}
   */
  public isRegExp(value: any): boolean {
    return this.getObjectType(value) === '[object RegExp]';
  }

  /**
   * Checks to see if a value is an array
   * @param {*} value
   * @returns {boolean}
   */
  public isArray(value: any): boolean {
    return (
      (typeof Array.isArray === 'function' && Array.isArray(value)) ||
      this.getObjectType(value) === '[object Array]'
    );
  }

  /**
   * Checks to see if a valule is a number
   * @param {*} value
   * @returns {boolean}
   */
  public isNumber(value: any): boolean {
    return (typeof value === 'number' || this.getObjectType(value) === '[object Number]') &&
      !isNaN(value);
  }

  /**
   * Checks to see if a value is a string
   * @param {*} value
   * @returns {boolean}
   */
  public isString(value: any): boolean {
    return typeof value === 'string' || this.getObjectType(value) === '[object String]';
  }

  /**
   * Checks to see if a valule is a boolean
   * @param {*} value
   * @returns {boolean}
   */
  public isBoolean(value: any): boolean {
    return (
      value === true ||
      value === false ||
      this.getObjectType(value) === '[object Boolean]'
    );
  }

  /**
   * Checks to see if a value is null
   * @param {*} value
   * @returns {boolean}
   */
  public isNull(value: any): boolean {
    return value === null;
  }

  /**
   * Checks to see if a value is undefined
   * @param {*} value
   * @returns {boolean}
   */
  public isUndefined(value: any): boolean {
    return typeof value === 'undefined';
  }

  /**
   * Checks to see if a value is a Map
   * @param {*} value
   * @returns {boolean}
   */
  public isMap(value: any): boolean {
    return this.getObjectType(value) === '[object Map]';
  }

  /**
   * Checks to see if a value is a WeakMap
   * @param {*} value
   * @returns {boolean}
   */
  public isWeakMap(value: any): boolean {
    return this.getObjectType(value) === '[object WeakMap]';
  }

  /**
   * Checks if value is defined and not null
   * @param  {*} value
   * @returns boolean
   */
  public isDefAndNotNull(value: any): boolean {
    return (this.isUndefined(value) || this.isNull(value) ? false : true);
  }

  /**
   * Compare deux objets et retourne true si les attributs sont identiques
   * @param x
   * @param y
   */
  public isSameObject(x: any, y: any) {
    if (x === y) { return true; }
    // if both x and y are null or undefined and exactly the same

    // if they are not strictly equal, they both need to be Objects
    if (!(x instanceof Object) || !(y instanceof Object)) {
      return false;
    }

    // they must have the exact same prototype chain, the closest we can do is
    // test there constructor.
    if (x.constructor !== y.constructor) {
      return false;
    }

    for (const p in x) {
      if (!x.hasOwnProperty(p)) { continue; }
      // other properties were tested using x.constructor === y.constructor

      // allows to compare x[ p ] and y[ p ] when set to undefined
      if (!y.hasOwnProperty(p)) {
        return false;
      }

      if (x[p] === y[p]) { continue; }
      // if they have the same strict value or identity then they are equal

      // Numbers, Strings, Functions, Booleans must be strictly equal
      if (typeof (x[p]) !== 'object') {
        return false;
      }

      // Objects and Arrays must be tested recursively
      if (!this.isSameObject(x[p], y[p])) {
        return false;
      }
    }

    for (const p in y) {
      // allows x[ p ] to be set to undefined
      if (y.hasOwnProperty(p) && !x.hasOwnProperty(p)) {
        return false;
      }
    }
    return true;
  }

  /**
   * check si l'objet passé est un Blob
   * @param {*} value
   * @returns {boolean}
   */
  public isBlob(value: any): boolean {
    return value instanceof Blob;
  }
  /**
   * check si l'objet passé est un File (blob spécialisé)
   * @param {*} value
   * @returns {boolean}
   */
  public isFile(value: any): boolean {
    return value instanceof File;
  }
  /**
   * check si l'objet passé est un File (blob spécialisé)
   * @param {*} value
   * @returns {boolean}
   */
  public isArrayOfFileNotempty(value: any): boolean {

    if (!this.isArray(value)) {
      return false;
    }

    if (value.length === 0) {
      return false;
    }

    return value[0] instanceof File;
  }

  /**
   * Permet de savoir si une valeur est un fichier uploadable (en dessous de la taille max)
   * @param {any} value valeur à tester
   * @param {number} iLimitation permet de fixer une taille en octet max arbitraire complémentaire
   * @returns
   */
  public isUploadable(value: any, iLimitation: number = -1): boolean {

    if (!this.isBlob(value)) {
      return false;
    }

    if (value.size > iLimitation && iLimitation > -1) {
      return false;
    }

    if (value.size > environment.properties.max_upload_file_size_restricted) {
      return false;
    }

    return true;
  }
}
