import { Injectable } from '@angular/core';
import { VitisCheck } from '../vitis-utils';

@Injectable({
  providedIn: 'root'
})
export class FunctionService {

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
  ) { }

  /**
   * Appelle une fonction appartenant à oFunctionContainer
   * @param  {any} sFunction
   * @param  {any} oValues
   * @param  {any} oFunctionContainer? objet ou tableau d'objets dans lequel est possiblement contenu la fonction
   */
  public callFunction(sFunction: any, oValues: any, oFunctionContainer?: any) {

    oFunctionContainer = this.vitisCheck.isDefAndNotNull(oFunctionContainer) ? oFunctionContainer : this;

    if (this.vitisCheck.isFunction(sFunction)) {
      return sFunction.call(oFunctionContainer, oValues);
    } else {

      const oFunction = this.parseFunction_(sFunction, oValues, oFunctionContainer);

      // Check et lance la fonction
      if (this.vitisCheck.isFunction(oFunction.function)) {
        return oFunction.function.call(
          oFunction.container,
          ...oFunction.params
        );
      } else {
        console.error('function ' + oFunction.name + ' not found in ', oFunctionContainer);
        return false;
      }
    }
  }

  /**
   * Vérifie si la chaine de caractère passée est un appel de fonction
   * @param {string} sString Chaine de caractère.
   * @param {boolean} bCheckFunction? Vérification de l'existence de la fonction (vrai par défaut).
   * @param {any} oFunctionContainer? objet ou tableau d'objets dans lequel est possiblement contenu la fonction
   * @return {boolean}
   */
  public isFunctionCall(sString: string, bCheckFunction?: boolean, oFunctionContainer?: any): boolean {

    bCheckFunction = this.vitisCheck.isDefAndNotNull(bCheckFunction) ? bCheckFunction : true;
    oFunctionContainer = this.vitisCheck.isDefAndNotNull(oFunctionContainer) ? oFunctionContainer : this;

    let bReturn = false;
    if (this.vitisCheck.isString(sString)) {
      if (sString.substring(0, 1) === '=') {
        return false;
      }
      const aMatchResult = sString.match(/(.+)\((.*)\)/);
      if (aMatchResult !== null) {
        if (bCheckFunction) {
          const sFunctionName = sString.substring(0, sString.indexOf('('));

          if (this.vitisCheck.isFunction(sFunctionName)) {
            bReturn = true;
          } else if (this.vitisCheck.isDefAndNotNull(oFunctionContainer)) {
            if (this.vitisCheck.isFunction(oFunctionContainer[sFunctionName])) {
              bReturn = true;
            }
            if (this.vitisCheck.isArray(oFunctionContainer)) {
              for (let i = 0; i < oFunctionContainer.length; i++) {
                const container = oFunctionContainer[i];
                if (this.vitisCheck.isFunction(container[sFunctionName])) {
                  bReturn = true;
                }
              }
            }
          } else if (this.vitisCheck.isFunction(this[sFunctionName])) {
            bReturn = true;
          }
          if (!bReturn) {
            console.error('Function not found : ' + sString);
          }
        } else {
          bReturn = true;
        }
      }
    }

    return bReturn;
  }

  /**
   * Parse la chaine passée et retourne un objet avec les paramètres suivants
   * function: Fonction à lancer
   * container: Container de la fonction
   * name: Nom de la fonction sans les paramètres
   * params: Paramètres de la fonction avec valeurs du formulaire
   *
   * @param  {any} sFunction
   * @param  {any} oValues
   * @param  {any} oFunctionContainer? objet ou tableau d'objets dans lequel est possiblement contenu la fonction
   * @returns any
   */
  private parseFunction_(sFunction: any, oValues: any, oFunctionContainer?: any): any {

    oFunctionContainer = this.vitisCheck.isDefAndNotNull(oFunctionContainer) ? oFunctionContainer : this;

    const oReturn = {
      function: sFunction,
      container: oFunctionContainer,
      name: sFunction,
      params: []
    };

    // Cas où oFunctionContainer est un tableau
    if (this.vitisCheck.isArray(oFunctionContainer)) {
      for (let i = 0; i < oFunctionContainer.length; i++) {
        const oTmpReturn = this.parseFunction_(sFunction, oValues, oFunctionContainer[i]);
        if (this.vitisCheck.isFunction(oTmpReturn.function)) {
          return oTmpReturn;
        }
      }
    }

    // Parse sFunction et récupère les paramètres
    oReturn.params = this.parseFunctionParams_(sFunction);
    oReturn.params.push(oValues);

    // Rretrouve le nom de la fonction
    if (sFunction.indexOf('(') !== -1) {
      oReturn.name = sFunction.substring(0, sFunction.indexOf('('));
    }

    // Retrouve le container de la fonction
    if (this.vitisCheck.isFunction(oFunctionContainer[oReturn.name])) {
      oReturn.function = oFunctionContainer[oReturn.name];
    }

    return oReturn;
  }

  /**
   * Parse les paramètres de la fonction si elle est de type string
   * @param  {any} sFunction
   * @returns Array
   */
  private parseFunctionParams_(sFunction: any): Array<any> {
    let aParams = [];

    if (this.vitisCheck.isString(sFunction)) {

      const iOpenParanthesis = (sFunction.match(/\(/g) || []).length;
      const iCloseParanthesis = (sFunction.match(/\)/g) || []).length;

      if (iOpenParanthesis === 1 && iCloseParanthesis === 1) {

        let sParams = '';
        sParams = sFunction.substring(sFunction.indexOf('('));
        sParams = sParams.substring(1, sParams.length - 1);
        if (sParams.length > 0) {
          aParams = sParams.split(',');
        }
        for (let i = 0; i < aParams.length; i++) {
          if (this.vitisCheck.isString(aParams[i])) {
            aParams[i] = aParams[i].trim();

            // Supprime les quotes
            if ((aParams[i][0] === '\'' && aParams[i][aParams[i].length - 1] === '\'')
              || (aParams[i][0] === '"' && aParams[i][aParams[i].length - 1] === '"')) {
              aParams[i] = aParams[i].substring(1, aParams[i].length - 1);
            }
          }
        }

      } else {
        if (iOpenParanthesis > 0) {
          console.error('Parsing event params failed');
        }
      }
    }

    return aParams;
  }
}
