import { Injectable } from '@angular/core';
import { VitisCheck } from '../vitis-utils';
import { environment } from 'src/environments/environment';

@Injectable({
  providedIn: 'root'
})
export class InterfaceService {

  /**
   * List des champs de la base de données qui ne sont pas pour le style css
   */
  private ignoredFields = ['theme_id', 'name'];

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
  ) { }

  /**
   * Change le theme de l'application en fonction du contenu des properties ou du paramètre passé
   * @param {string} currentTheme
   * @param {Array<any>} themes
   */
  public toggleTheme(currentTheme?: string, themes?: Array<any>) {
    // Si le thème en cours n'est pas passé en paramètre (cas test thème), on récupère celui des properties
    if (!this.vitisCheck.isDefAndNotNull(currentTheme)) {
      currentTheme = environment.properties.app_theme;
    }

    // Si les themes ne sont pas passé en paramètre (cas test thème personnalisé), on récupère ceux des properties
    if (!this.vitisCheck.isDefAndNotNull(themes)) {
      themes = environment.properties.themes;
    }

    if (this.vitisCheck.isDefAndNotNull(currentTheme) && currentTheme != '') {
      for (let i = 0; i < themes.length; i++) {
        const theme = themes[i];
        // Si c'est le theme sélectionné
        if (theme.theme_id == currentTheme) {
          // Parcours les propriétés du theme
          for (const stylePropertie in theme) {
            if (Object.prototype.hasOwnProperty.call(theme, stylePropertie)) {
              document.documentElement.style.removeProperty('--application-' + stylePropertie.replace(/_/g, '-'));

              // On ignore theme_id et name car ce ne sont pas des variables pour le css, et on ignore les valeurs vide
              if (this.ignoredFields.indexOf(stylePropertie) === -1 && theme[stylePropertie] != '') {
                // Applique le style
                document.documentElement.style.setProperty('--application-' + stylePropertie.replace(/_/g, '-'), theme[stylePropertie]);
              }
            }
          }
        }

      }
    }
  }

  /**
   * Construit et renvoie un theme personalisé en fonction de oData
   * @param {Array<any>} oData
   */
  public getPersonalizeTheme(oData: Array<any>) {
    // Sauvegarde du thème personnalisé
    const personalizeTheme = {
      theme_id: oData['app_theme']
    };
    for (const stylePropertie in environment.properties.themes[0]) {
      if (Object.prototype.hasOwnProperty.call(environment.properties.themes[0], stylePropertie)) {
        if (this.ignoredFields.indexOf(stylePropertie) === -1) {
          personalizeTheme[stylePropertie] = oData[stylePropertie];
        }
      }
    }

    return personalizeTheme;
  }


  /**
   * Applique le css personnalisé au site
   * @param {string} customCss
   */
  public setDynamicStyle(customCss?: string) {

    if (!this.vitisCheck.isDefAndNotNull(customCss)) {
      customCss = environment.properties.custom_css;
    }

    if (this.vitisCheck.isDefAndNotNull(customCss) && customCss != '') {
      // Ajoute le css custom
      const oScript = document.createElement('style');
      const oCSS = document.createTextNode(customCss);
      oScript.appendChild(oCSS);
      document.getElementsByTagName('body')[0].appendChild(oScript);
    }
  }
}
