import { Injectable } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import sha256 from 'crypto-js/sha256';
import Base64 from 'crypto-js/enc-base64';

import { RequestsService } from './requests.service';
import { vitisAlert } from '../alert/vitisalert';
import { environment } from 'src/environments/environment';
import { VitisChunk, VitisFile } from './file-requests.model';

declare let $: any;


/**
 * le but de ce service et de permettre l'envoi de gros fichiers sur l'API de vitis en le découpant en paquets
 * chaque paquet est envoyé sur le serveur, puis une requête de finalisation est envoyé au serveur pour controler et refusionner le fichier
 * le principe est sensiblement le même que l'upload de gros fichiers sur AWS S3
 *
 * @author Anthony Borghi
 */


@Injectable({
  providedIn: 'root'
})
export class FileRequestsService {

  /**
   * permet de jouer sur le nombre de paquet pour la boucle récursive
   * si on envoi trop de paquet simultanément ça peut saturer le serveur (surtout en dev)
   * je préconise de ne pas dépasser les 20
   */
  private limit_batch: number = 1;

  /**
   * permet d'incorporer une latence entre chaque lancement dans un même paquet
   * le but est le même que limit_batch, si limit_batch vaut 1 cette variable ne fait rien ou presque
   * par défaut avec 200ms on devrait être bon pour une connexion type fibre/vdsl
   * Apache est aussi moins sujet à ce genre de problème que le serveur symfony
   */
  private timeout_batch: number = environment.properties.transactional_upload_delay;

  /**
   * Id des trasactions en cours
   */
  private currentTransactionsIds: Array<string> = [];

  /**
   * variable de mémorisation de la modale
   */
  private alertOptions: any = {};

  /**
   * Constructeur
   */
  constructor(
    private vitisRequests: RequestsService,
    private translateService: TranslateService,
  ) { }

  /**
   * Envoi un tableau de fichier via les transactions
   * @param {File[]} aFiles
   * @returns {Promise<string[]>}
   */
  public uploadFiles(aFiles: File[]): Promise<string[]> {
    this.currentTransactionsIds.length = 0;
    return new Promise<string[]>((resolve, reject) => {
      const aTransactionId: string[] = [];
      let iProcessed: number = 0;
      for (let i = 0; i < aFiles.length; i++) {
        setTimeout(() => {
          const oFile = aFiles[i];
          this.uploadFile_(oFile).then((sTransactionId: string) => {
            aTransactionId.push(sTransactionId);
            iProcessed++;
          }).catch(() => {
            iProcessed++;
          }).finally(() => {
            if (iProcessed >= aFiles.length) {
              resolve(aTransactionId);
            }
          });
        }, i * this.timeout_batch);
      }
    });
  }

  /**
   * Envoi un fichier en utilisant les transactions
   * @param {File} oFile
   * @returns {Promise<string>}
   */
  public uploadFile(oFile: File): Promise<string> {
    this.currentTransactionsIds.length = 0;
    return this.uploadFile_(oFile);
  }

  /**
   * Envoi un fichier en utilisant les transactions
   * @param {File} oFile
   * @returns {Promise<string>}
   */
  private uploadFile_(oFile: File): Promise<string> {
    return new Promise<string>((resolve, reject) => {

      const aChunkedFile: Blob[] = this.chunkBlob(oFile);

      this.hashTransaction(aChunkedFile, oFile.name).then((oTransaction: VitisFile) => {
        this.beginTransaction(oTransaction).then((sTransactionId) => {

          // Mémorise la transaction id
          this.currentTransactionsIds.push(sTransactionId);

          this.sendChunks(sTransactionId, oTransaction).then(() => {
            this.endTransaction(sTransactionId).then(() => {
              resolve(sTransactionId);
            }).catch(() => {
              console.error('cannot complete the transaction');
              reject();
            });
          }).catch(() => {
            console.error('cannot send all chunks');
            reject();
          });
        }).catch(() => {
          console.error('cannot start transaction');
          reject();
        });
      });
    });
  }

  /**
   * Détermine si le fichier à besoin d'être envoyé en multipart ou non
   * @param {Blob} oFile fichier à tester
   * @returns {boolean}
   */
  public needUploadbyTransaction(oFile: Blob): boolean {
    // si la taille du fichier est sous le seuil
    if (oFile.size < environment.properties.transactional_upload_threshold) {
      return false;
    }
    // ou au dessus du nombre de tronçon autorisé on ne passe pas en transactionnel
    if (oFile.size > environment.properties.max_upload_file_size) {
      return false;
    }
    return true;
  }

  /**
   * Découpe le blob en plusieurs sous-blob
   * @param {Blob} oFile
   * @returns {Blob[]}
   */
  private chunkBlob(oFile: Blob): Blob[] {
    const aBlob: Blob[] = [];
    const iChunkSize: number = environment.properties.chunk_size;
    let iIndex: number = 0;
    let eof: boolean = false;

    while (!eof) {
      // calcul de l'octet d'arret
      const iEndByte: number = (iIndex + 1) * iChunkSize;
      let oTmpBlob: Blob;
      // si on doit s'arréter aprés la fin du fichier on ne précise pas la fin et on coupe la boucle
      if (iEndByte > oFile.size) {
        oTmpBlob = oFile.slice(iIndex * iChunkSize);
        eof = true;
      } else {
        oTmpBlob = oFile.slice(iIndex * iChunkSize, iEndByte);
      }
      aBlob.push(oTmpBlob);
      iIndex++;
    }

    return aBlob;
  }

  /**
   * hash chaque blob et génére un hash pour le fichier entier
   *
   * Pour le hash du fichier, on prend le hash du premier blob tel quel, que l'on concatène avec celui du blob suivant
   * la concaténation des deux hash est ensuite elle-même hashé en sha256 et encodé en base64, puis on recommence
   * de façon à traiter tous les blobs successivement dans l'ordre croissant des index
   *
   *
   * @param {Blob[]} aBlob
   * @returns {Promise<VitisFile>}
   */
  private hashTransaction(aBlob: Blob[], sFileName: string): Promise<VitisFile> {
    return new Promise<VitisFile>((resolve, reject) => {
      const aVitisChunk: VitisChunk[] = [];
      let iBlobProcessed: number = 0;
      let sFileHash: string = '';
      for (let i = 0; i < aBlob.length; i++) {
        const oChunk: Blob = aBlob[i];
        this.hashBlob(oChunk).then((str: string) => {
          const tmpVitisChunck: VitisChunk = {
            index: i,
            file: oChunk,
            hash: str
          };
          aVitisChunk.push(tmpVitisChunck);


        }).finally(() => {
          iBlobProcessed++;
          if (iBlobProcessed >= aBlob.length) {
            // génére le chunk pour le fichier
            for (const oFileChunk of aVitisChunk.sort((a, b) => (a.index < b.index) ? -1 : 1)) {
              if (sFileHash == '') {
                sFileHash = oFileChunk.hash;
              } else {
                sFileHash = Base64.stringify(sha256(sFileHash + oFileChunk.hash));
              }
            }

            resolve({
              name: sFileName,
              hash: sFileHash,
              blobs: aVitisChunk
            });
          }
        });
      }
    });
  }

  /**
   * hash un blob
   *
   * Le hashage se fait en récupérant les 256 premiers bytes du blob que l'on passe ne hexa
   * (dans le cas du binaire c'était complqiué de récupérer la totalité de la chaine)
   * La chaine hexa est ensuite hashé en sha256 puis encodé en base64
   *
   * @param {Blob} oFile
   * @returns {Promise<string>}
   */
  private hashBlob(oFile: Blob): Promise<string> {

    return new Promise<string>((resolve, reject) => {
      const hashedBlob: Blob = oFile.slice(0, 256);
      hashedBlob['arrayBuffer']().then((buf) => {
        // Passage en chaine hexa car le binaire ne me permet pas de cécupérer le bon nombre de bytes (UTF8)
        const str = this.buf2hex(buf);
        resolve(Base64.stringify(sha256(str)/* .toString() */));
      }).catch(() => {
        reject();
      });
    });
  }

  /**
   * Transforme un ArrayBuffer en string hexa
   * @param {ArrayBuffer} buffer
   * @returns {string}
   */
  private buf2hex(buffer: ArrayBuffer): string {
    return [...new Uint8Array(buffer)]
      .map(x => x.toString(16).padStart(2, '0'))
      .join('');
  }

  /**
   * Débute une transaction
   * @param {VitisFile} oFile
   * @returns {Promise<string>}
   */
  private beginTransaction(oFile: VitisFile): Promise<string> {
    return new Promise<string>((resolve, reject) => {
      this.showLoadingModal();
      // envoi de la requête
      this.vitisRequests.ajaxRequestPromise({
        method: 'POST',
        ajaxLoader: true,
        url: environment.properties.app_api_url + '/vitis/lfs',
        data: {
          md5_hash_content: oFile.hash,
          nb_chunk: oFile.blobs.length,
          file_name: oFile.name,
        }
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status === 1) {
            // récupération du transaction_id
            this.addProgressBar(oFile, resp.data.data.transaction_id);
            resolve(resp.data.data.transaction_id);
          } else {
            reject();
          }
        }
      });
    });
  }

  /**
   * traite tout le lot de tronçon en récursif
   * @param {string} transaction_id
   * @param {VitisFile} oFile
   * @param {number} offset default 0
   * @returns {Promise<void>}
   */
  private sendChunks(transaction_id: string, oFile: VitisFile, offset: number = 0): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      // s'il n'y a plus de donnée à traité c'est fini
      if (offset >= oFile.blobs.length) {
        resolve();
        return;
      }
      let iProcessed: number = 0;
      // prend la limitation la plus basse entre ce qu'il reste dans la pile ou le limit
      const iLimit = (this.limit_batch > (oFile.blobs.length - offset)) ? oFile.blobs.length - offset : this.limit_batch;
      // traite un lot de tronçons
      for (let i = offset; i < (offset + iLimit); i++) {
        // on ajoute un delay progressif dans la boucle pour ne pas saturer le serveur
        setTimeout(() => {
          const oBlob = oFile.blobs[i];
          this.addChunk(oBlob, transaction_id).then(() => {
            iProcessed++;
            this.updateProgressBar(oFile, transaction_id, i);
          }).catch(() => {
            reject();
          }).finally(() => {
            if (iProcessed >= iLimit) {
              // si le lot est traité on passe au suivant
              this.sendChunks(transaction_id, oFile, offset + iLimit).then(() => {
                resolve();
              });
            }
          });
        }, (i - offset) * this.timeout_batch);
      }
    });
  }

  /**
   *  Envoi un tronçon sur le serveur
   * @param {VitisChunk} oBlob
   * @param {string} transaction_id
   * @returns {Promise<void> }
   */
  private addChunk(oBlob: VitisChunk, transaction_id: string): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      const formData = new FormData();
      formData.append('chunk', oBlob.file);
      formData.append('md5_hash_content', oBlob.hash);


      this.vitisRequests.ajaxRequestPromise({
        method: 'POST',
        ajaxLoader: false,
        displayErrorMessage: false,
        url: environment.properties.app_api_url + '/vitis/lfs/' + transaction_id + '/add/' + oBlob.index,
        data: formData
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status == 1) {
            resolve();
            return;
          }
        }
        this.abortTransaction(transaction_id).finally(() => {
          reject();
        });
      }).catch(() => {
        this.abortTransaction(transaction_id).finally(() => {
          reject();
        });
      });
    });
  }

  /**
   * Finalise la transaction
   * @param {string} transaction_id
   * @returns {Promise<void>}
   */
  private endTransaction(transaction_id: string): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      this.updateToVerifState(transaction_id);
      this.cleanTransactionId(transaction_id);
      // envoi de la requête
      this.vitisRequests.ajaxRequestPromise({
        method: 'PUT',
        ajaxLoader: false,
        displayErrorMessage: false,
        url: environment.properties.app_api_url + '/vitis/lfs/' + transaction_id + '/complete',
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status === 1) {
            // récupération du transaction_id
            resolve();
            this.updateToCompleteState(transaction_id);
          } else {
            this.abortTransaction(transaction_id).finally(() => {
              reject();
            });
          }
        }
      }).catch(() => {
        this.abortTransaction(transaction_id).finally(() => {
          reject();
        });
      });
    });
  }

  /**
   * Finalise la transaction
   * @param {string} transaction_id
   * @returns {Promise<void>}
   */
  private abortTransaction(transaction_id: string): Promise<void> {
    return new Promise<void>((resolve, reject) => {
      // envoi de la requête
      if (this.currentTransactionsIds.indexOf(transaction_id) !== -1) {
        this.vitisRequests.ajaxRequestPromise({
          method: 'DELETE',
          ajaxLoader: false,
          url: environment.properties.app_api_url + '/vitis/lfs/' + transaction_id + '/abort',
        }).then((resp) => {
          if (resp.data) {
            if (resp.data.status === 1) {
              // récupération du transaction_id
              resolve();
              this.cleanTransactionId(transaction_id);
              this.updateToAbortState(transaction_id);
            } else {
              reject();
            }
          }
        });
      }
    });
  }

  /**
   * Inistialise la modale de supervision
   */
  private showLoadingModal(): void {

    if (vitisAlert.isVisible()) {
      return;
    }

    this.translateService.get([
      'TRANSACTIONAL_LOADING_MODAL_TITLE',
      'TRANSACTIONAL_LOADING_MODAL_IS_PROCESSING',
      'TRANSACTIONAL_LOADING_MODAL_ABORD',
    ]).subscribe((translation) => {
      this.alertOptions = {
        title: translation['TRANSACTIONAL_LOADING_MODAL_TITLE'],
        width: 650,
        html: '<div style="color:#ff8c00;">' + translation['TRANSACTIONAL_LOADING_MODAL_IS_PROCESSING'] + '</div><div id="transactional-modal-status" style="padding-top: 10px;"></div>',
        showCloseButton: false,
        showCancelButton: false,
        showDenyButton: true,
        showConfirmButton: false,
        allowOutsideClick: false,
        denyButtonText: translation['TRANSACTIONAL_LOADING_MODAL_ABORD'],
      };
      vitisAlert.alert(this.alertOptions).then((result) => {
        if (result.isDenied === true) {
          for (let i = 0; i < this.currentTransactionsIds.length; i++) {
            const transaction_id = this.currentTransactionsIds[i];
            this.abortTransaction(transaction_id);
          }
        }
      });
    });
  }

  /**
   * Ajoute une bar de progression pour un fichier donné
   * @param oFile
   * @param transaction_id
   */
  private addProgressBar(oFile: VitisFile, transaction_id: string): void {
    const self = this;
    const sHtml: string = '<div id="tr-progress-' + transaction_id + '" class="transaction-progress-group">' +
      '<div class="row" style="margin-bottom:5px;">' +
      '<div class="col-9" style="text-align:left; white-space: nowrap;overflow: hidden;text-overflow: ellipsis;">' + oFile.name +
      '</div>' +
      '<div id="tr-percent-' + transaction_id + '" class="col" style="text-align:right;">0 %</div>' +
      '</div>' +
      '<div class="progress">' +
      '<div id="tr-progress-bar-' + transaction_id + '" class="progress-bar bg-default" role="progressbar" style="width: 0%"></div>' +
      '</div>' +
      '</div>';
    $('#transactional-modal-status').append(sHtml);
    $('#tr-abort-' + transaction_id).on('click', () => {
      self.abortTransaction(transaction_id);
    });
  }

  /**
   * Met à jour la bar de prohression pour un fichier donné
   * @param oFile
   * @param transaction_id
   * @param processed
   */
  private updateProgressBar(oFile: VitisFile, transaction_id: string, processed: number): void {
    const ratio: number = Math.trunc((processed / (oFile.blobs.length)) * 10000) / 100;
    $('#tr-percent-' + transaction_id).html(ratio + ' %');
    $('#tr-progress-bar-' + transaction_id).css('width', ratio + '%');
  }

  /**
   * modifie la ligne de la barre de chargement
   * et ferme la modale si nécessaire
   */
  private updateToVerifState(transaction_id: string): void {
    this.translateService.get('TRANSACTIONAL_LOADING_MODAL_CHECK').subscribe((translation) => {

      const sHtml: string = '<div style="color:orange;">' + translation + '</div>';
      $('#tr-percent-' + transaction_id).html(sHtml);
      $('#tr-progress-bar-' + transaction_id).parent().remove();
    });
  }

  /**
   * modifie la ligne de la barre de chargement
   * et ferme la modale si nécessaire
   * @param {string} transaction_id
   */
  private updateToCompleteState(transaction_id: string): void {
    this.translateService.get('TRANSACTIONAL_LOADING_MODAL_COMPLETE').subscribe((translation) => {
      const sHtml: string = '<div style="color:green;">' + translation + '</div>';
      $('#tr-percent-' + transaction_id).html(sHtml);

      // Check si on a encore une transaction en cours
      if ($('.transaction-progress-group').length == 0) {
        vitisAlert.close();
      } else {
        vitisAlert.update({
          ...this.alertOptions,
          html: vitisAlert.getHtmlContainer(),
          showCloseButton: true
        });
      }
    });
  }

  /**
   * modifie la ligne de la barre de chargement
   * et ferme la modale si nécessaire
   * @param {string} transaction_id
   */
  private updateToAbortState(transaction_id: string): void {
    this.translateService.get('TRANSACTIONAL_LOADING_MODAL_ABORT').subscribe((translation) => {
      const sHtml: string = '<div style="color:red;">' + translation + '</div>';
      $('#tr-percent-' + transaction_id).html(sHtml);
    });

    // Check si on a encore une transaction en cours
    if ($('#tr-progress-bar-' + transaction_id).length > 0) {
      $('#tr-progress-bar-' + transaction_id).removeClass('bg-default').addClass('bg-danger');
    }

    $('#tr-abort-' + transaction_id).remove();

    // Check si on a encore une transaction en cours
    if ($('.transaction-progress-group').length == 0) {
      vitisAlert.close();
    } else {
      vitisAlert.update({
        ...this.alertOptions,
        html: vitisAlert.getHtmlContainer(),
        showCloseButton: true
      });
    }
  }

  /**
   * Nettoie currentTransactionsIds et enlève la transaction id
   * @param transaction_id
   */
  private cleanTransactionId(transaction_id: string): void {
    const index = this.currentTransactionsIds.indexOf(transaction_id);
    if (index > -1) {
      this.currentTransactionsIds.splice(index, 1);
    }
  }
}
