import { Component, OnInit, ViewChild } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { TranslateService } from '@ngx-translate/core';
import { VitisCheck, vitisAlert, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { DataGridComponent } from '../data-grid/data-grid.component';
import { DataGridOptions, DataTableOptions } from 'src/app/vitis/data-grid/data-grid.model';
import { WorkspaceListOptions } from '../workspacelist/workspacelist.model';
import { environment } from 'src/environments/environment';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import { CustomTranslationService } from './services/customtranslation.service';
import ResizeObserver from 'resize-observer-polyfill';

declare let $: any;

@Component({
  selector: 'app-customtranslation',
  templateUrl: './customtranslation.component.html',
  styleUrls: ['./customtranslation.component.scss']
})
export class CustomTranslationComponent implements OnInit {

  /**
   * Enfant de type DataGridComponent
   */
  @ViewChild(DataGridComponent, { static: true }) oDataGrid: DataGridComponent;

  // Composant de la grille de données.
  public oDataGrid_: any;

  // Options de la grille de données.
  public oDataGridOptions: DataGridOptions = {
    datatableOptions: {},
    datagridOptions: WorkspaceListOptions
  };

  // Mode en cours
  public oMode: any;

  // Types de filtre disponible
  public oFilterTypes: any = {
    custom: 'VITIS_WORKSPACELIST_FILTER_CUSTOM',
    form: 'VITIS_WORKSPACELIST_FILTER_FORM'
  };

  // Filtre en cours
  public oFilter: any = {
    type: 'custom',
    nodes: []
  };

  /**
   * Constructeur
   */
  constructor(
    private route: ActivatedRoute,
    private router: Router,
    private vitisCheck: VitisCheck,
    private translate: TranslateService,
    private auth: AuthService,
    private vitisRequest: VitisRequest,
    private customTranslationService: CustomTranslationService
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit() {
    // Mode en cours d'utilisation
    this.route.data.subscribe(mode => {
      this.oMode = mode;
      // options app.json
      if (mode.options) {
        this.oDataGridOptions.datagridOptions = mode.options;
      }
      this.initDataGrid();
    });
  }

  /**
  * Initialise la grille de données.
  * @param {DataGridComponent} oDataGrid
  * @returns void
  */
  private initDataGrid() {

    // Options de la grille de données.
    const datatableOptions: DataTableOptions = {
      select: 'true',
      selectMode: 'multi',
      order: [0, 'asc'], // Ordre
      displayStart: 0,  // Offset
      rowId: '',
      clientPagination: true,
      ajax: (data: any, callback: any, settings: any) => {

        const oResponse = {
          recordsTotal: 0,
          recordsFiltered: 0,
          data: []
        };

        const oResults = this.setDataGridContent();
        if (oResults.length > 0) {
          oResponse.data = oResults;
          oResponse.recordsTotal = oResults.length;
          oResponse.recordsFiltered = oResults.length;
        }

        callback(oResponse);
      }
    };

    // Paramètres du plugin DataTable.
    if (this.vitisCheck.isDefAndNotNull(this.oMode.options.columns)) {
      this.oDataGridOptions.datagridOptions.columns = this.oMode.options.columns;
    }

    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.id_field)) {
      datatableOptions.rowId = this.oDataGridOptions.datagridOptions.id_field;
    }

    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.columns)) {

        let oDatagridColumns = [];
        if (this.vitisCheck.isArray(this.oDataGridOptions.datagridOptions.columns)) {
          oDatagridColumns = this.oDataGridOptions.datagridOptions.columns.filter(column => {
            // Privilège(s) obligatoire(s) pour afficher la colonne.
            if (this.vitisCheck.isDefAndNotNull(column.privilege)) {
              if (this.vitisCheck.isString(column.privilege)) {
                column.privilege = [column.privilege];
              }
              if (this.vitisCheck.isArray(column.privilege)) {
                const aUserPrivileges = this.auth.getPrivileges();
                for (let i = 0; i < column.privilege.length; i++) {
                  const sPrivilege = column.privilege[i];
                  if (this.vitisCheck.isString(sPrivilege) &&
                    aUserPrivileges.indexOf(sPrivilege) !== -1) {
                    return true;
                  }
                }
              } else {
                console.error('Column privilege not string or array', column.privilege);
              }
              return false;
            }
            return true;
          });
        }

        // Traitement des colonnes
        for (let i = 0; i < oDatagridColumns.length; i++) {

          const column = oDatagridColumns[i];

          // Index par défaut de la colonne
          column.index = i;

          // Traduction
          column.translateTitle = this.translate.instant(column.title);

          column.data = column.field;

          // Visibilité de la colonne
          column.visible = column.visible === false ? false : true;

          // Colonne type
          column.type = column.template ? 'html' : 'string';

          // Alignement du contenu de la colonne.
          column.className = 'table-align-' + (column.align ? column.align : 'left');
        }

        this.oDataGridOptions.datagridOptions.columns = oDatagridColumns;
      }
    }

    this.oDataGridOptions.datatableOptions = datatableOptions;

    // Bouton delete
    if (this.oDataGridOptions.datagridOptions.delete === true) {
      this.oDataGridOptions.buttons = [{
        type: 'button',
        class: 'btn-default',
        label: 'LIST_CONFIGURATION_CUSTOM_TRANSLATIONS_DELETE_BTN',
        icon: 'fa fa-trash',
        event: () => {
          this.deleteSelectedRows();
        }
      }];
    }

    // Initialise les boutons icons
    this.initIconButtons();

    setTimeout(() => {

      this.oDataGridOptions.datatableOptions.tableHeight = this.calcDataTableHeight();
      // Création de la grille.
      this.oDataGrid.initTable(this.oDataGridOptions);
      this.customTranslationService.dataTableInstance = this.oDataGrid;
    });
  }

  /**
   * Initialise les boutons présent dans chaque ligne du tableau
   */
  public initIconButtons(): void {
    const aButtons: any = [];
    let buttonUpdateExist = false;
    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.update) &&
      this.oDataGridOptions.datagridOptions.update === true) {
      aButtons.push({
        type: 'update',
        icon: 'icon-vitis-draw',
        event: (data: any) => {
          this.displayRowForm_(data, 'update');
        }
      });
      buttonUpdateExist = true;
    }

    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.display) &&
      this.oDataGridOptions.datagridOptions.display === true) {
      // Vérifie qu'il y a un bouton update
      if (environment.properties.data_grid_update_btn_only === true && buttonUpdateExist) {
        this.oDataGridOptions.datagridOptions.display = false;
      } else {
        aButtons.push({
          type: 'display',
          icon: 'icon-vitis-search',
          event: (data: any) => {
            this.displayRowForm_(data, 'display');
          }
        });
      }
    }

    this.oDataGridOptions.iconButtons = aButtons;
  }

  /**
   * Calcule la hauteur de la table
   */
  private calcDataTableHeight = () => {
    if (this.vitisCheck.isNumber(this.oMode.options.height)) {
      return this.oMode.options.height;
    } else {
      const dashboardHeight = $('#dashboard-content').height();
      const titleHeight = $('.mode-title').height();
      const filterHeight = $('.grid-filter').height();
      const menusHeight = 200;
      const tableHeight = dashboardHeight - (menusHeight + filterHeight + titleHeight);
      return tableHeight;
    }
  };

  /**
   * Affichage des traductions dans la grille.
   */
  private setDataGridContent = (): any[] => {
    // Traductions de la langue sélectionnée.
    const oTranslations: any = this.translate.translations[environment.properties.language];
    const oOriginalTranslations: any = this.customTranslationService.oTranslations[environment.properties.language];
    const aRows: any = [];
    // Formatage des données de traduction pour la grille.
    for (const sKey in oTranslations) {
      if (sKey != '') {
        aRows.push({
          translation_id: sKey,
          original_text: oOriginalTranslations[sKey] ? oOriginalTranslations[sKey] : '',
          text: (this.customTranslationService.aCustomTranslations.indexOf(sKey) !== -1) ? (oTranslations[sKey] ? oTranslations[sKey] : '') : '',
          custom_translation: (this.customTranslationService.aCustomTranslations.indexOf(sKey) !== -1)
        });
      }
    }

    // Application des filtres de la grille.
    let aFilteredRows: any = [];
    if (this.oFilter.nodes.length > 0) {
      let bAddRow: boolean;
      for (let i = 0; i < aRows.length; i++) {
        bAddRow = true;
        for (let j = 0; j < this.oFilter.nodes.length; j++) {
          // Si le filtre n'est pas vérifié : pas d'ajout de la ligne.
          if (String(aRows[i][this.oFilter.nodes[j].column.field]).toLowerCase().indexOf(
            String(this.oFilter.nodes[j].value).toLowerCase()) === -1) {
            bAddRow = false;
            // Si l'opérateur de comparaison est un "AND" : ligne suivante.
            if (this.oFilter.relationOperator == 'AND') {
              break;
            }
          } else {
            bAddRow = true;
            // Si l'opérateur de comparaison est un "OR" : ajoute la ligne et ligne suivante.
            if (this.oFilter.relationOperator == 'OR') {
              break;
            }
          }
        }
        if (bAddRow) {
          aFilteredRows.push(aRows[i]);
        }
      }
    } else {
      aFilteredRows = aRows;
    }

    return aFilteredRows;
  };

  /**
   * Affiche le formulaire correspondant
   * @param  {any} oRow
   * @param  {String} sFormName
   */
  private displayRowForm_(oRow: any, sFormName: String) {
    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.id_field)) {
        if (this.vitisCheck.isDefAndNotNull(oRow)) {
          if (this.vitisCheck.isDefAndNotNull(oRow[this.oDataGridOptions.datagridOptions.id_field])) {
            if (this.vitisCheck.isDefAndNotNull(sFormName)) {
              // Redirection vers le formulaire d'édition.
              const sId = String(oRow[this.oDataGridOptions.datagridOptions.id_field]).toLowerCase();
              const aModeRoute = this.oMode.route.split('/');
              this.router.navigate([sId, sFormName], { relativeTo: this.route.parent });
            }
          }
        }
      }
    }
  }

  /**
   * Suppression des traductions personnalisées sélectionnées.
   */
  private deleteSelectedRows(): void {
    const aTranslationId = this.oDataGrid_.getSelectedIds_();
    // Au moins un enregistrement doit être sélectionné.
    if (aTranslationId.length > 0) {
      // Au moins une traduction personnalisée doit être sélectionné.
      const aCustomTranslation: any = [];
      for (let i = 0; i < aTranslationId.length; i++) {
        if (this.customTranslationService.aCustomTranslations.indexOf(aTranslationId[i]) !== -1) {
          aCustomTranslation.push(aTranslationId[i]);
        }
      }
      if (aCustomTranslation.length > 0) {
        if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
          if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.ressource)) {
            // Chemin de la ressource REST.
            let sRessourceURL = environment.properties.app_api_url;
            sRessourceURL += '/' + this.oDataGridOptions.datagridOptions.ressource;
            vitisAlert.alert({
              type: 'warning',
              title: this.translate.instant('LIST_DELETE_CONFIRM'),
              showCancelButton: true
            }).then((result: any) => {
              if (result.isConfirmed === true) {
                // Suppression des traductions personnalisées en base.
                this.vitisRequest.ajaxRequestPromise({
                  method: 'DELETE',
                  url: sRessourceURL,
                  data: {
                    idList: aCustomTranslation
                  }
                }).then((resp) => {
                  if (resp.data) {
                    if (resp.data.status === 1) {
                      // Message de succès.
                      vitisAlert.fireToast({
                        icon: 'success',
                        title: this.translate.instant('LIST_DELETE_OK')
                      });
                      // Recharge les traductions personnalisées.
                      this.customTranslationService.loadCustomTranslations()
                        .then(() => {
                          // Recharge les données de la grille.
                          // this.setDataGridContent();
                        });
                    }
                  }
                });
              }
            });
          }
        }
      } else {
        vitisAlert.alert({
          type: 'error',
          title: this.translate.instant('LIST_CONFIGURATION_CUSTOM_TRANSLATIONS_DELETE_NO_TRANSLATION'),
        });
      }
    } else {
      vitisAlert.alert({
        type: 'error',
        title: this.translate.instant('LIST_DELETE_NO_SELECTION'),
      });
    }
  }

  /**
   * Recharge la data en appliquant les filtres
   */
  public reloadData() {
    this.oDataGrid.reloadTable();
  }
}
