import { Component, ViewContainerRef, OnInit, ComponentRef, Injector, EventEmitter, ViewChildren, QueryList, AfterViewInit, ViewChild, OnDestroy } from '@angular/core';
import { LocalStorage, VitisCheck, VitisClone, VitisRequest, vitisAlert, SessionStorage } from 'src/app/vitis/core/services/utils/vitis-utils';
import { WorkspaceListButton } from '../workspacelist/workspacelist.model';
import { DataGridOptions } from 'src/app/vitis/data-grid/data-grid.model';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import { UntypedFormBuilder, UntypedFormGroup, Validators } from '@angular/forms';
import exportFromJSON from 'export-from-json';
import { Subject, Subscription } from 'rxjs';
import { DragAndDropListComponent } from '../drag-and-drop-list/drag-and-drop-list.component';
import { environment } from 'src/environments/environment';
import { TranslateService } from '@ngx-translate/core';

declare let $: any;

@Component({
  selector: 'app-data-grid',
  templateUrl: './data-grid.component.html',
  styleUrls: ['./data-grid.component.scss']
})
export class DataGridComponent implements OnInit, AfterViewInit, OnDestroy {

  @ViewChild(DragAndDropListComponent, { static: false }) oDragAndDropList: DragAndDropListComponent;

  // Options de la grille de données.
  public oDataGridOptions: DataGridOptions;
  public oDataGridOptionsColumns;

  private records: Array<any>;
  public oDataGridOptionsRecords: Array<any> = [];
  public datasIsLoading: boolean;
  public oDataGridOptionsRecordsFiltered: number = 0;
  public oDataGridOptionsRecordsTotal: number = 0;

  // Sélection des lignes
  public oDataGridOptionsSelectedRecords: Array<any> = [];
  public isAllRowsSelected: boolean = false;

  // Sort
  public indexColumnSort: number = 0;
  public directionSort: string = 'desc';

  // Paginations
  public page: number = 1;
  public pageLength: number = environment.properties.rows_per_page;
  public lengthMenu: number[] = [5, 10, 20, 50, 100];
  public offsetPage = 0;

  // Paramètres requêtes ajax
  public ajaxParams: any = {};

  public uniqueId: string = this.getUniqueId_();
  public containerTableId: string = 'containerTable' + this.uniqueId;
  public tableId: string = 'table' + this.uniqueId;

  // Formats d'exportation de la liste.
  public exportFormats: any = {
    'csv': {
      replace_label: true,
      no_data_value: ''
    },
    'xls': {
      replace_label: true,
      no_data_value: ''
    },
    'json': {
      replace_label: false,
      no_data_value: null
    },
    'xml': {
      replace_label: false,
      no_data_value: ''
    }
  };
  public form: UntypedFormGroup;

  // Observable d'initialisation complète (grille affichée et données chargées).
  public onInitComplete: EventEmitter<string> = new EventEmitter();
  public selectRowEvent: Subject<any> = new Subject();
  public deselectRowEvent: Subject<any> = new Subject();

  /**
   * Données pour la personnalisation des colonnes
   */
  public list_settings: Array<any> = []; // Paramètres de la liste avec les affichages non modifié
  public curr_list_settings: Array<any> = []; // Paramètres de la liste avec les affichages en cours
  public selectedDisplay: number; // Id de l'affichage en cours
  public defaultDisplay: number; // Id de l'affichage de défaut
  public newDisplay: boolean = false; // Booléen si l'affichage actuel est nouveau
  public displayHasChanged: boolean = false; // Booléen si l'affichage actuel à été modifié
  public displayNewLabel: string = ''; // Label de l'affichage courant
  public displayPagingElement: boolean = true;

  private subscribeOnChangeColumnOrder: Subscription; // Subscription du changement des colonnes

  public modalCreateDisplay = 'modal_id_create_display_' + Date.now();
  public modalCustomColumns = 'modal_id_custom_columns_' + Date.now();

  // Body datagrid
  @ViewChildren('bodyRecords') bodyRecords: QueryList<any>;

  /**
   *
   */
  constructor(
    private vitisCheck: VitisCheck,
    private vitisClone: VitisClone,
    public viewContainer: ViewContainerRef,
    private fb: UntypedFormBuilder,
    private auth: AuthService,
    private vitisRequest: VitisRequest,
    private localStorage: LocalStorage,
    private translate: TranslateService,
    private sessionStorage: SessionStorage,
  ) {

  }

  /**
   * Retourne les affichages public
   */
  get publicDisplays() {
    return this.vitisCheck.isDefAndNotNull(this.curr_list_settings) ?
      this.curr_list_settings.filter(display => display.ispublic === true) : [];
  }

  /**
   * Retourne les affichages personnalisé
   */
  get persoDisplays() {
    return this.vitisCheck.isDefAndNotNull(this.curr_list_settings) ?
      this.curr_list_settings.filter(display => display.ispublic === false && display.list_settings_id > 0) : [];
  }

  /**
   * Calcul dynamiquement la longueur du colspan de la cellule qui pas d'enregistrements
   */
  get colspanEmptyRecords() {
    let colspan = 0;
    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions)) {
      if (this.oDataGridOptions.datatableOptions && this.oDataGridOptions.datatableOptions.select !== false) {
        colspan += 1;
      }
      if (this.oDataGridOptions.iconButtons && this.oDataGridOptions.iconButtons.length > 0) {
        colspan += this.oDataGridOptions.iconButtons.length;
      }
      if (this.oDataGridOptionsColumns && this.oDataGridOptionsColumns.length > 0) {
        colspan += this.oDataGridOptionsColumns.length;
      }
    }
    return colspan;
  }

  /**
   * Retourne les informations de pagination de data grid
   */
  get infosTable() {
    return {
      itemStart: ((this.page - 1) * this.pageLength) > 0 ? ((this.page - 1) * this.pageLength) : 1,
      itemEnd: ((this.page * this.pageLength) <
        this.oDataGridOptionsRecordsTotal) ? (this.page * this.pageLength) : this.oDataGridOptionsRecordsTotal,
      totalItems: this.oDataGridOptionsRecordsTotal,
      selectedItems: this.oDataGridOptionsSelectedRecords.length
    };
  }

  /**
   *
   */
  ngOnInit(): void {
    // formulaire d'exportation de la liste.
    this.form = this.fb.group({
      export_file_name: ['export', Validators.required],
      export_file_format: ['csv', Validators.required],
      export_rows_number: ['20']
    });
  }

  /**
   *
   */
  ngAfterViewInit(): void {
    // Event quand le tableau est initialisé et rempli
    this.bodyRecords.changes.subscribe(changes => {
      this.onInitComplete.emit();
      setTimeout(() => {
        if (this.vitisCheck.isDefAndNotNull(this.selectedDisplay)) {
          this.setColumnsOrder(this.getDisplayById().definition.column);
        }
        this.formatFields(this.oDataGridOptions);
        this.ajustTableHeight();
      });
    });
  }

  /**
   *
   */
  ngOnDestroy(): void {
    if (this.subscribeOnChangeColumnOrder) {
      this.subscribeOnChangeColumnOrder.unsubscribe();
    }
  }

  /**
   * Initialise la table de datagrid
   */
  public initTable(oDataGridOptions: DataGridOptions) {
    if (this.vitisCheck.isDefAndNotNull(oDataGridOptions.datatableOptions)) {

      if (this.vitisCheck.isDefAndNotNull(oDataGridOptions.datatableOptions.displayStart)) {
        this.offsetPage = oDataGridOptions.datatableOptions.displayStart;
      }

      if (this.vitisCheck.isDefAndNotNull(oDataGridOptions.datatableOptions.paging)) {
        this.displayPagingElement = oDataGridOptions.datatableOptions.paging !== false;
      }

      if (this.vitisCheck.isDefAndNotNull(oDataGridOptions.datatableOptions.pageLength)) {
        this.pageLength = oDataGridOptions.datatableOptions.pageLength;
      } else {
        this.pageLength = this.getPageLength_();
      }

      if (this.vitisCheck.isDefAndNotNull(oDataGridOptions.datatableOptions.lengthMenu)) {
        this.lengthMenu = oDataGridOptions.datatableOptions.lengthMenu;
      } else {
        this.lengthMenu = this.getLengthMenu_();
      }

      if (this.vitisCheck.isDefAndNotNull(oDataGridOptions.datatableOptions.order)) {
        if (this.vitisCheck.isObject(oDataGridOptions.datatableOptions.order[0])) {
          this.indexColumnSort = oDataGridOptions.datatableOptions.order[0].column;
          this.directionSort = oDataGridOptions.datatableOptions.order[0].dir;
        } else {
          this.indexColumnSort = oDataGridOptions.datatableOptions.order[0];
          this.directionSort = oDataGridOptions.datatableOptions.order[1];
        }
      }

      if (oDataGridOptions.datatableOptions.colReorder === true) {
        /**
         * Initialise le drag and drop pour la personnalisation des colonnes
         */
        this.oDragAndDropList.initOptions({
          columns: this.vitisClone.clone(oDataGridOptions.datagridOptions.columns),
          label_field: 'translateTitle',
          id_field: 'field'
        });

        this.defaultDisplay = oDataGridOptions.datatableOptions.listSettings.isdefault;
        this.list_settings = [
          this.vitisClone.clone(this.getStandardDisplay(this.defaultDisplay)),
          ...this.vitisClone.clone(oDataGridOptions.datatableOptions.listSettings.userDisplays)
        ];
        this.curr_list_settings = this.vitisClone.clone(this.list_settings);
        this.selectedDisplay = this.defaultDisplay;

        /**
         * Event appelé lors de changement dans la personnalisation des colonnes
         */
        this.subscribeOnChangeColumnOrder = this.oDragAndDropList.onChange.subscribe((order: Array<any>) => {
          if (this.getDisplayById().ispublic === true) { // Si affichage public on génère un nouvel affichage
            const newDisplay = this.createNewLocalDisplay(order);
            this.curr_list_settings.push(newDisplay);
            this.newDisplay = true;
          } else {
            this.getDisplayById().definition.column = this.vitisClone.clone(order);
          }
          this.setColumnsOrder(order);
          this.displayHasChanged = true;
        });
      }

      const _this = this;
      $(`#${this.tableId}`).on('click', 'tr', function () {
        _this.onSelectRow(this.oRow);
      });

      $(`#${this.tableId}`).on('click', 'tr .container-icon-event', function (e) {
        e.stopPropagation();
      });
    }

    this.oDataGridOptionsColumns = this.vitisClone.clone(oDataGridOptions.datagridOptions.columns);

    this.oDataGridOptions = oDataGridOptions;

    if (this.vitisCheck.isDefAndNotNull(oDataGridOptions.datatableOptions.ajax)) {
      /**
       * Charge les données de la table
       */
      this.reloadTable(false, this.getParamsOrder());
    } else if (this.vitisCheck.isDefAndNotNull(oDataGridOptions.datatableOptions.oResults)) {
      const oResults = oDataGridOptions.datatableOptions.oResults;
      this.records = this.vitisClone.clone(oResults.data);
      this.oDataGridOptionsRecords = this.vitisClone.clone(oResults.data);
      this.oDataGridOptionsRecordsFiltered = oResults.recordsFiltered;
      this.oDataGridOptionsRecordsTotal = oResults.recordsTotal;
    }
  }

  /**
   * Redimensionne datagrid
   */
  public ajustTableHeight() {
    if (this.vitisCheck.isDefAndNotNull(document.getElementById(this.containerTableId)) &&
      this.vitisCheck.isDefAndNotNull(this.oDataGridOptions)) {
      document.getElementById(this.containerTableId).style.height = this.oDataGridOptions.datatableOptions.tableHeight + 'px';
    }
  }

  /**
   * Retourne le tableau contenant le nombre de lignes affichables
   */
  private getLengthMenu_(): number[] {
    const lengthMenu = [5, 10, 20, 50, 100];

    if (this.oDataGridOptions &&
      this.oDataGridOptions.datagridOptions &&
      this.oDataGridOptions.datagridOptions.max_rows) {

      const iMaxRows = this.oDataGridOptions.datagridOptions.max_rows;

      for (let i = lengthMenu.length - 1; i >= 0; i--) {
        if (lengthMenu[i] > iMaxRows) {
          lengthMenu.splice(i, 1);
        }
      }
    }

    return lengthMenu;
  }

  /**
   * Retourne le nombre d'éléments à afficher
   */
  private getPageLength_(): number {
    let pageLength: number = this.pageLength;

    // Concordance avec le tableau de choix
    this.lengthMenu = this.getLengthMenu_();
    if (this.lengthMenu.indexOf(pageLength) === -1) {
      if (pageLength > this.lengthMenu[this.lengthMenu.length - 1]) {
        pageLength = this.lengthMenu[this.lengthMenu.length - 1];
      } else {
        pageLength = 20;
      }
    }

    return pageLength;
  }

  /**
   * Recharge les données de datagrid
   */
  public reloadTable(resetFilter = false, params: any = {}) {

    // Recharge l'affichage courant sur le dernier enregistrer dans la sessionStorage
    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions) &&
      this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datatableOptions.listSettings)) {
      const currSessionStorage = this.sessionStorage.get(this.oDataGridOptions.datatableOptions.listSettings.route);
      if (currSessionStorage && this.vitisCheck.isDefAndNotNull(currSessionStorage.lastSelectedDisplay) &&
        currSessionStorage.lastSelectedDisplay !== 0 && currSessionStorage.lastSelectedDisplay !== this.selectedDisplay) {
        this.changeSelectedDisplay(this.getDisplayById(currSessionStorage.lastSelectedDisplay), false);
        return;
      }
    }

    // Reinitialise le filtre si on change d'affichage
    if (resetFilter === true) {
      params.filter = this.getDisplayById().definition.filter;
      this.displayHasChanged = false;
    }

    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions) &&
      this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datatableOptions.ajax)) {

      params.start = this.offsetPage;
      params.length = this.pageLength;

      if (!this.vitisCheck.isDefAndNotNull(params.order)) {
        const paramsOrder = this.getParamsOrder();
        params.order = paramsOrder.order;
        params.columns = paramsOrder.columns;
      }

      this.ajaxParams = params;
      this.datasIsLoading = true;
      this.oDataGridOptions.datatableOptions.ajax(params, (oResults: any) => {

        // Pagination front end
        if (this.oDataGridOptions.datatableOptions.clientPagination === true) {
          oResults = this.paginateResultsFrontEnd(oResults, params);
        }

        // Exploitation des résultats
        this.oDataGridOptionsRecords = [];
        if (this.vitisCheck.isDefAndNotNull(oResults)) {
          this.records = this.vitisClone.clone(oResults.data);
          this.oDataGridOptionsRecords = this.vitisClone.clone(oResults.data);
          this.oDataGridOptionsRecordsFiltered = oResults.recordsFiltered;
          this.oDataGridOptionsRecordsTotal = oResults.recordsTotal;
          if (this.oDataGridOptions.datatableOptions.colReorder === true) {
            if (this.newDisplay === true) {
              this.getDisplayById().definition.filter = this.vitisClone.clone(oResults.filter);
            } else {
              const oldFilter = this.list_settings.find(oldDisplay =>
                oldDisplay.list_settings_id === this.selectedDisplay).definition.filter;
              if ((oResults.filter.value || oldFilter.value) &&
                !this.vitisCheck.isSameObject(oResults.filter, oldFilter)) {
                if (this.getDisplayById().ispublic === true) {
                  const newDisplay = this.createNewLocalDisplay(this.getDisplayById().definition.column);
                  this.curr_list_settings.push(newDisplay);
                  this.newDisplay = true;
                }
                this.displayHasChanged = true;
              }
            }
          }
        }
        this.datasIsLoading = false;
      }, null);
    }
  }

  /**
   * Attribut les colonnes dans l'odre de leur indexation
   * @param oDataGridColumns
   */
  private setColumnsOrder(columnsOrder) {
    if (this.oDataGridOptionsColumns && this.oDataGridOptionsColumns.length > 0 && columnsOrder) {
      const oDataGridOptionsColumns = [];
      for (let indexOrder = 0; indexOrder < columnsOrder.length; indexOrder++) {
        const column = columnsOrder.find(columnOrder => columnOrder.index === indexOrder);
        const currColumn = this.vitisClone.clone(this.oDataGridOptionsColumns.find(optionColumn => optionColumn.field === (column.field ?? column.id)));
        currColumn.index = indexOrder;
        currColumn.visible = column.visible;
        oDataGridOptionsColumns[indexOrder] = currColumn;
      }
      this.oDataGridOptionsColumns = oDataGridOptionsColumns;
      setTimeout(() => {
        // Met à jour les formats des cellules après update des colonnes
        // Utile pour afficher les field render
        this.formatFields(this.oDataGridOptions);
      });
      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions) &&
        this.oDataGridOptions.datatableOptions.colReorder === true) {
        this.oDragAndDropList.setDraggedRows(this.oDataGridOptionsColumns);
      }
    }
  }

  /**
   * Retourne l'ordre de l'affichage standard
   * Appelé au début de l'initialisation du tableau
   * @param defaultId
   * @returns
   */
  private getStandardDisplay(defaultId): object {
    return this.list_settings.length > 0 ?
      {
        ...this.list_settings.find(display => display.list_settings_id === 0),
        isdefault: defaultId === 0,
      } : {
        list_settings_id: 0,
        name: 'Affichage Standard',
        definition: {
          filter: {
            type: 'custom',
            nodes: []
          },
          column:
            this.oDataGridOptions.datagridOptions.columns.map(column =>
              ({ index: column.index, visible: column.visible, field: column.field }))
        },
        isdefault: defaultId === 0,
        ispublic: true
      };
  }

  /**
   * Retourne l'affichage par son Id
   * @param display_id
   * @returns
   */
  public getDisplayById(display_id = this.selectedDisplay) {
    return this.curr_list_settings.find(display => display.list_settings_id === display_id);
  }

  /**
   * Créé un nouvel affichage perso en local
   * @param column
   * @returns
   */
  private createNewLocalDisplay(column, name?) {
    const currDisplay = this.getDisplayById();
    if (this.vitisCheck.isDefAndNotNull(name)) {
      this.displayNewLabel = name;
    } else {
      this.displayNewLabel = 'Affichage personnalisé';
      if (currDisplay.ispublic === false) {
        this.displayNewLabel = currDisplay.name + ' ' + this.curr_list_settings.length;
      }
    }
    this.selectedDisplay = this.curr_list_settings.length * - 1;
    return {
      list_settings_id: this.selectedDisplay,
      name: this.displayNewLabel,
      definition: {
        filter: this.vitisClone.clone(currDisplay.definition.filter),
        column
      },
      route: this.oDataGridOptions.datatableOptions.listSettings.route,
      isdefault: false,
      ispublic: false
    };
  }

  /**
   * Fonction appelée lors d'un changement sur un autre affichage existant
   * @param display nouvel affichage à charger
   */
  public updateSelectedDisplay(display) {
    if (this.newDisplay === true) {
      // Gère le clique sur un changement d'affichage qui n'a pas été créé
      if (display.list_settings_id === this.selectedDisplay) {
        return;
      }
      vitisAlert.alert({
        title: 'Abandonner le nouvel affichage',
        type: 'warning',
        showCancelButton: true,
        confirmButtonText: this.translate.instant('CONFIRM_CONFIRM'),
        cancelButtonText: this.translate.instant('CONFIRM_CANCEL'),
      }).then((result) => {
        if (result.value) {
          this.curr_list_settings = this.curr_list_settings.filter(oldDisplay => oldDisplay.list_settings_id !== this.selectedDisplay);
          this.changeSelectedDisplay(display);
        }
      });
    } else {
      this.changeSelectedDisplay(display);
    }
  }

  /**
   * Change l'affichage en cours | reinitialise les variables
   * @param display
   */
  private changeSelectedDisplay(display: any, resetFilter = true) {
    this.selectedDisplay = display.list_settings_id;
    this.sessionStorage.add(this.oDataGridOptions.datatableOptions.listSettings.route, {
      lastSelectedDisplay: this.selectedDisplay
    });
    this.displayNewLabel = display.name;
    this.reloadTable(resetFilter);
    this.newDisplay = false;
    this.displayHasChanged = false;
  }

  /**
   * Fonction appelée quand l'utilisateur créé un nouvel affichage
   */
  public createUserDisplay(showModal: boolean = false): Promise<any> {
    return new Promise((resolve, reject) => {
      if (showModal === true) {
        $('#' + this.modalCreateDisplay).modal('show');
        return;
      }
      const { list_settings_id, ...newDisplay } = this.getDisplayById();
      this.vitisRequest.ajaxRequest({
        url: environment.properties.app_api_url + '/vitis/list_settings',
        method: 'POST',
        data: {
          ...newDisplay,
          definition: JSON.stringify(newDisplay.definition),
          name: this.displayNewLabel
        },
        success: (resp: any) => {
          this.reloadUserDisplays().then(() => {
            this.changeSelectedDisplay(resp.data.data);
            $('#' + this.modalCreateDisplay).modal('hide');
            resolve(resp.data.data);
          });
        },
        catch: (e: any) => {

        }
      });
    });
  }

  /**
   * Supprime l'affichage en cours et charge l'affichage par défaut
   */
  public deleteUserDisplay() {
    if (this.defaultDisplay != this.selectedDisplay) {
      this.vitisRequest.ajaxRequest({
        url: environment.properties.app_api_url + '/vitis/list_settings/' + this.selectedDisplay,
        method: 'DELETE',
        success: (resp: any) => {
          // popup de success
          this.reloadUserDisplays().then(() => {
            this.updateSelectedDisplay(this.getDisplayById(this.defaultDisplay));
          });
        },
        catch: (e: any) => {

        }
      });
    } else {
      vitisAlert.alert({
        title: 'Vous ne pouvez pas supprimer un affichage par défaut',
        type: 'error',
        showCancelButton: true
      });
    }
  }

  /**
   * Envoie une requête de mise à jour des colonnes personnalisées
   */
  public saveDisplayColumns() {
    this.saveUserDisplay({
      definition: JSON.stringify({
        filter: this.getDisplayById().definition.filter,
        column: this.oDragAndDropList.oDragAndDropGridRows
      })
    }, this.selectedDisplay, true);
  }

  /**
   * Envoie une requête de mise à jour du nom de l'affichage
   */
  public saveDisplayName(exec) {
    if (exec === true) {
      this.saveUserDisplay({
        name: this.displayNewLabel
      }, this.selectedDisplay, true);
      $('#' + this.modalCreateDisplay).modal('hide');
    } else {
      this.displayNewLabel = this.getDisplayById().name;
      $('#' + this.modalCreateDisplay).modal('show');
    }
  }

  /**
   * Requête de mise à jour d'un affichage
   * @param data data à mettre à jour
   * @param displayId id de l'affichage à mettre à jour
   * @param reload booleen de reload les affichages suite au modifications
   */
  public saveUserDisplay(data, displayId, reload): Promise<void> {
    return new Promise((resolve, reject) => {
      this.vitisRequest.ajaxRequest({
        url: environment.properties.app_api_url + '/vitis/list_settings/' + displayId,
        method: 'PUT',
        data,
        success: (resp: any) => {
          if (reload === true) {
            this.reloadUserDisplays().then(() => {
              this.displayHasChanged = false;
              this.updateSelectedDisplay(this.getDisplayById(this.selectedDisplay));
              resolve();
            });
          } else {
            resolve();
          }
        },
        catch: (e: any) => {
          reject(e);
        }
      });
    });
  }

  /**
   * Annule les modifications de l'affichage courant
   */
  public resetCurrentDisplay() {
    this.curr_list_settings = this.curr_list_settings.map(display => {
      if (display.list_settings_id === this.selectedDisplay) {
        return this.vitisClone.clone(this.list_settings.find(oldDisplay => oldDisplay.list_settings_id === this.selectedDisplay));
      }
      return display;
    });
    this.reloadTable(true);
  }

  /**
   * Ajoute l'affichage en cours en affichage par défaut
   */
  public async setDisplayToDefault() {
    if (this.selectedDisplay !== this.defaultDisplay) {
      if (this.defaultDisplay != 0) {
        await this.saveUserDisplay({
          isdefault: false
        }, this.defaultDisplay, false);
      }

      const currDisplay = this.getDisplayById();
      // Si l'affichage est public alors on le dupplique et on set l'affichage comme affichage par défaut
      if (currDisplay.ispublic && this.selectedDisplay != 0) {
        const newDisplay = this.createNewLocalDisplay(currDisplay.definition.column, currDisplay.name);
        this.curr_list_settings.push(newDisplay);
        this.createUserDisplay(false).then(() => {
          this.setDisplayToDefault();
        });
      } else {
        if (this.selectedDisplay != 0) {
          await this.saveUserDisplay({
            isdefault: true
          }, this.selectedDisplay, false);
        }
        this.defaultDisplay = this.selectedDisplay;
        this.reloadUserDisplays().then(() => {
          this.displayHasChanged = false;
          this.updateSelectedDisplay(this.getDisplayById(this.selectedDisplay));
        });
      }
    }
  }

  /**
   * Recharge les affichages de l'utilisateur
   */
  private reloadUserDisplays(): Promise<void> {
    return new Promise((resolve) => {
      this.vitisRequest.ajaxRequest({
        method: 'GET',
        url: environment.properties.app_api_url + '/vitis/list_settings',
        success: (resp: any) => {
          const userDisplays = resp.data.data;
          this.localStorage.set('userDisplays', userDisplays);
          const currUserDisplays = userDisplays.filter(display =>
            display.route === this.oDataGridOptions.datatableOptions.listSettings.route).map(display =>
            ({ ...this.vitisClone.clone(display), definition: JSON.parse(display.definition) })
          );
          if (this.oDataGridOptions.datatableOptions.colReorder === true) {
            this.list_settings = [
              this.vitisClone.clone(this.getStandardDisplay(this.defaultDisplay)),
              ...this.vitisClone.clone(currUserDisplays)
            ];
            this.curr_list_settings = this.vitisClone.clone(this.list_settings);
            resolve();
          }
        },
        catch: (e: any) => {

        }
      });
    });
  }

  /**
   * Formate les champs d'un tableau selon le template d'une colonne
   * @param {any} oDataGridOptions
   */
  public formatFields(oDataGridOptions: any) {
    if (oDataGridOptions &&
      oDataGridOptions.datagridOptions) {

      const aFieldRefs: any = document.getElementsByTagName('field-render');
      for (let i = aFieldRefs.length - 1; i >= 0; i--) {

        const oFieldRef = aFieldRefs[i];
        const oRow = oFieldRef.row;
        const oColumn = oDataGridOptions.datagridOptions.columns.find(column =>
          column.field === oFieldRef.column_field || column.data === oFieldRef.data || column.data === oFieldRef.column_field);
        const oParentNode = oFieldRef.parentNode;

        if (this.vitisCheck.isFunction(oColumn.template)) { // Si c'est une classe
          const oClass = oColumn.template;
          const oOptions = this.vitisCheck.isObject(oColumn.template_options) ? oColumn.template_options : {};
          const componentRef: ComponentRef<any> = this.viewContainer.createComponent(oClass, {
            injector: Injector.create({
              providers: [
                {
                  provide: 'DatagridProvider',
                  useValue: {
                    oRow,
                    oColumn,
                    oFieldRef,
                    ...oOptions,
                  }
                }
              ]
            })
          });

          // Supprime la chaine de base
          oFieldRef.parentNode.removeChild(oFieldRef);

          // Ajoute à l'élément la nouvelle chaine
          oParentNode.appendChild(componentRef.location.nativeElement);
        }
      }
    }
  }

  /**
   *
   */
  public getSelectedIds_(): Array<any> {
    return this.oDataGridOptionsSelectedRecords;
  }

  /**
   * Vide le tableau des sélections
   */
  public clearSelection() {
    this.oDataGridOptionsSelectedRecords = [];
  }

  /**
   * Récupère la liste des ligne sélectionné
   */
  public getSelectedRows_(): Array<any> {
    const sKey = this.oDataGridOptions.datagridOptions.id_field || this.oDataGridOptions.datatableOptions.rowId;

    if (!this.vitisCheck.isDefAndNotNull(sKey)) {
      return [];
    }

    const aReturn = [];

    for (const record_id of this.oDataGridOptionsSelectedRecords) {
      for (const record of this.oDataGridOptionsRecords) {

        if (!this.vitisCheck.isDefAndNotNull(record)) {
          continue;
        }

        if (record[sKey] === record_id) {
          aReturn.push(record);
        }
      }
    }

    return aReturn;
  }

  /**
   * Ajoute une ligne dans le data grid
   * @param oRow
   */
  public addRow_(oRow): any {
    this.records = [...this.records, this.vitisClone.clone(oRow)];
    const datagridRow = this.vitisClone.clone(oRow);
    this.oDataGridOptionsRecords = [...this.oDataGridOptionsRecords, datagridRow];
    this.oDataGridOptionsRecordsFiltered += 1;
    this.oDataGridOptionsRecordsTotal += 1;
    return datagridRow;
  }

  /**
   * Nettoye le tableau
   */
  public clear() {
    this.records = [];
    this.oDataGridOptionsRecords = [];
    this.oDataGridOptionsRecordsFiltered = 0;
    this.oDataGridOptionsRecordsTotal = 0;
  }

  /**
   * Trouve un identifiant html unique
   * @returns string
   */
  private getUniqueId_(): string {
    let id = '_' + Math.random().toString(36).substring(2, 11);
    if (document.getElementById(id)) {
      id = this.getUniqueId_();
    }
    return id;
  }

  /**
   * Sélectionne/Désélectionne toute les lignes
   */
  public onSelectAllRows(isSelect?) {
    if (this.vitisCheck.isDefAndNotNull(isSelect)) {
      this.isAllRowsSelected = isSelect;
    } else {
      this.isAllRowsSelected = !this.isAllRowsSelected;
    }
    if (this.isAllRowsSelected) {
      this.oDataGridOptionsSelectedRecords = this.oDataGridOptionsRecords.map(row => row[this.oDataGridOptions.datatableOptions.rowId]);
    } else {
      this.oDataGridOptionsSelectedRecords = [];
    }
  }

  /**
   * Sélectionne/Déselectionne la ligne cliquée
   * @param selectedRow ligne sélectionnée
   */
  public onSelectRow(selectedRow) {
    const id_field = selectedRow[this.oDataGridOptions.datatableOptions.rowId];
    if (this.oDataGridOptionsSelectedRecords.includes(id_field)) {
      this.oDataGridOptionsSelectedRecords = this.oDataGridOptionsSelectedRecords.filter(field => field != id_field);
      this.deselectRowEvent.next(selectedRow);
    } else {
      if (this.oDataGridOptions.datatableOptions.selectMode === 'single' &&
        this.oDataGridOptionsSelectedRecords.length > 0) {
        return;
      }
      this.oDataGridOptionsSelectedRecords.push(id_field);
      this.selectRowEvent.next(selectedRow);
    }
  }

  /**
   * Retourne si la ligne est sélectionnée
   */
  public isRowSelected(row) {
    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions)) {
      return this.oDataGridOptionsSelectedRecords.includes(row[this.oDataGridOptions.datatableOptions.rowId]);
    }
  }

  /**
   *
   */
  public onSort(indexColumnSort) {
    if (this.indexColumnSort === indexColumnSort) {
      this.directionSort = this.directionSort ? (this.directionSort === 'asc' ? 'desc' : '') : 'asc';
    } else {
      this.indexColumnSort = indexColumnSort;
      this.directionSort = 'asc';
    }
    this.reloadTable(false, this.getParamsOrder());
  }

  /**
   *
   */
  private getParamsOrder() {
    const paramsDatatable: any = {};
    if (this.vitisCheck.isDefAndNotNull(this.directionSort) && this.directionSort.length > 0) {
      paramsDatatable.order = [{
        dir: this.directionSort,
        column: this.indexColumnSort
      }];
      if (this.oDataGridOptions.datatableOptions.colReorder === true) {
        paramsDatatable.columns = this.getDisplayById().definition.column.map(column =>
          ({ ...column, data: column.field }));
      } else {
        paramsDatatable.columns = this.vitisClone.clone(this.oDataGridOptions.datagridOptions.columns);
      }
    }
    return paramsDatatable;
  }

  /**
   * Initialise l'événement d'un bouton
   * @param {WorkspaceListButton} button
   * @param {oWorkspaceList} any container de l'objet workspacelist
   */
  public initButtonEvent_(button: WorkspaceListButton, oWorkspaceList?: any) {
    // Transforme la chaine de caractères en vrai fonction
    if (this.vitisCheck.isString(button.event)) {
      button.sEvent = button.event;
    }
    if (this.vitisCheck.isString(button.eventOnClose)) {
      button.sEvent = button.eventOnClose;
    }
    if (this.vitisCheck.isString(button.sEvent)) {
      button.event = () => {

        // Nom de la fonction sans les paranthèses
        let sFunction = button.sEvent;
        if (sFunction.indexOf('(') !== -1) {
          sFunction = sFunction.substring(0, sFunction.indexOf('('));
        }

        // Lance la fonction depuis workspacelist
        if (this.vitisCheck.isDefAndNotNull(oWorkspaceList)) {
          if (this.vitisCheck.isFunction(oWorkspaceList[sFunction])) {
            return oWorkspaceList[sFunction].call(oWorkspaceList);
          }
        }

        // Lance la fonction depuis datagrid
        if (this.vitisCheck.isFunction(this[sFunction])) {
          return this[sFunction].call(this);
        }
      };
    }

    // Recursivité
    if (this.vitisCheck.isArray(button.buttons)) {
      for (let i = 0; i < button.buttons.length; i++) {
        const buttonChild = button.buttons[i];
        this.initButtonEvent_(buttonChild, oWorkspaceList);
      }
    }
  }

  /**
   * Initialise l'affichage d'un bouton
   * @param {WorkspaceListButton} button
   * @param {oWorkspaceList} any container de l'objet workspacelist
   */
  public initButtonDisplay_(button: WorkspaceListButton, oWorkspaceList?: any) {
    // Privilège(s) obligatoire(s) pour afficher le bouton.
    if (this.vitisCheck.isDefAndNotNull(button.privilege)) {
      const aUserPrivileges = this.auth.getPrivileges();
      if (this.vitisCheck.isString(button.privilege)) {
        if (aUserPrivileges.indexOf(<string>button.privilege) !== -1) {
          button.display = true;
        } else {
          button.display = false;
        }
      } else if (this.vitisCheck.isArray(button.privilege)) {
        button.display = false;
        for (let ii = 0; ii < button.privilege.length; ii++) {
          const sPrivilege = button.privilege[ii];
          if (aUserPrivileges.indexOf(sPrivilege) !== -1) {
            button.display = true;
            break;
          }
        }
      }
    }

    // Recursivité
    if (this.vitisCheck.isArray(button.buttons)) {
      for (let i = 0; i < button.buttons.length; i++) {
        const buttonChild = button.buttons[i];
        this.initButtonDisplay_(buttonChild, oWorkspaceList);
      }
    }
  }

  /**
   * Recherche dans les colonnes du mode celle dont le field correspond à sColumnName
   * @param {String} sColumnName
   * @returns any
   */
  public getColumnByName_(sColumnName: String): any {

    if (this.oDataGridOptions &&
      this.oDataGridOptions.datagridOptions &&
      this.oDataGridOptions.datagridOptions.columns) {

      // Pour chaque colonne
      for (let i = 0; i < this.oDataGridOptions.datagridOptions.columns.length; i++) {
        const column = this.oDataGridOptions.datagridOptions.columns[i];

        if (column.field.trim() === sColumnName.trim()) {
          return column;
        }
      }
    }
  }

  /**
   * Affichage de la fenêtre modale d'exportation de la liste.
   */
  public displayExportModal() {
    $('#datagrid_export_modal').modal('show');
  }

  /**
   * Affichage de la modale de custom des colonnes
   */
  public displayCustomColumnsModal() {
    $('#' + this.modalCustomColumns).modal('show');
  }

  /**
   * Téléchargement du fichier d'exportation de la liste.
   */
  public exportDataGrid() {
    // récupération des valeurs du formulaire ou valeur par défaut
    const iMaxRow: number = (this.form.value.export_rows_number) ? this.form.value.export_rows_number : 0;
    const sFormat: string = (this.form.value.export_file_format) ? this.form.value.export_file_format : 'csv';
    const sFileName: string = (this.form.value.export_file_name) ? this.form.value.export_file_name : 'export';

    this.getDataGridDataForExport(iMaxRow, this.exportFormats[sFormat]).then((aData: any[]) => {
      switch (sFormat) {
        case 'csv':
        case 'json':
        case 'xml':
        case 'xls':
          const aConfig: any = {
            data: aData,
            fileName: sFileName,
            exportType: sFormat,
          };
          // transforme le json et télécharge le résultat
          exportFromJSON(aConfig);
          break;
        default:
          console.error('format not available : ' + sFormat);
          break;
      }
    }).catch(() => {
      console.error('error occured while exporting data');
    }).finally(() => {
      $('#datagrid_export_modal').modal('hide');
    });
  }

  /**
   * récupère les lignes visible dans data-grid, remplace les clé et transforme le tout en array type json
   * @param {number} iNbRows nombre de ligne à récupérer si 0 on récupère tout
   * @param {any} aConfig
   * @returns {any} json
   */
  public getDataGridDataForExport(iNbRows: number = 0, aConfig: any = {}): Promise<Array<any>> {
    const aFinalData: any[] = [];
    const aColumns: any[] = this.oDataGridOptions.datagridOptions.columns;
    let isSelected = false;

    return new Promise((resolve, reject) => {
      for (let i = 0; i < this.oDataGridOptionsRecords.length; i++) {

        // décide si la signe fait partie de celles sélectionnées
        if (this.vitisCheck.isDefAndNotNull(aConfig.aSelectedRows)) {
          isSelected = this.isSelectedRow(this.oDataGridOptionsRecords[i], aConfig.aSelectedRows);
        }

        const aTmpItem: any = {};
        for (const oColumn of aColumns) {
          if (('translateTitle' in oColumn || 'title' in oColumn) && ('field' in oColumn || 'data' in oColumn)) {
            // si la colonne à une correspondance dans les données
            // on récupère la donnée en remplaçant la clé par le label de la colonne
            const sKey: string = (aConfig.replace_label) ? (oColumn.translateTitle || oColumn.title) : (oColumn.field || oColumn.data);
            if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptionsRecords[i][oColumn.field || oColumn.data])) {
              aTmpItem[sKey] = this.oDataGridOptionsRecords[i][oColumn.field || oColumn.data];
            } else {
              aTmpItem[sKey] = aConfig.no_data_value;
            }
          }
        }
        if (this.vitisCheck.isDefAndNotNull(aConfig.aSelectedRows)) {
          if (isSelected) {
            aFinalData.push(aTmpItem);
          }
        } else {
          aFinalData.push(aTmpItem);
        }
        if (iNbRows > 0 && iNbRows === i + 1) {
          break;
        }
      }
      resolve(aFinalData);
    });
  }

  /**
   * Retourne true/false si la ligne fait partie des sélectionnées
   * @param oRow
   * @param aSelectedRows
   * @returns
   */
  private isSelectedRow(oRow: any, aSelectedRows: any[]): boolean {
    for (let i = 0; i < aSelectedRows.length; i++) {
      const oSelectedRow = aSelectedRows[i];

      let isSelected = true;
      for (const key in oRow) {
        if (Object.prototype.hasOwnProperty.call(oRow, key)) {
          if (oSelectedRow[key] !== oRow[key]) {
            isSelected = false;
          }
        }
      }

      if (isSelected) {
        return true;
      }
    }

    return false;
  }

  /**
   * Change de page ou du nombres d'affichage par page
   * @param limit
   * @param pageNumber
   */
  public onChangeParamsTable(limit?, pageNumber?) {

    if (this.newDisplay === true) {
      return;
    }

    if (this.vitisCheck.isDefAndNotNull(limit)) {
      this.pageLength = Number(limit);
      // Permet de mettre à jour la page et l'offset quand la page est > 1
      if (this.page > 1) {
        pageNumber = 1;
        while ((pageNumber * this.pageLength - 1) < this.offsetPage) {
          pageNumber = pageNumber + 1;
        }
      }
    }

    if (this.vitisCheck.isDefAndNotNull(pageNumber)) {
      this.page = pageNumber;
      this.offsetPage = (this.page - 1) * this.pageLength;

    }
    this.reloadTable(false, this.ajaxParams);
  }

  /**
   * Effectue la pagination côté frontend
   * @param oResults
   * @param params
   * @returns
   */
  private paginateResultsFrontEnd(oResults: any, params: any): any {
    if (oResults &&
      oResults.data) {

      // Ordre
      if (params &&
        params.order &&
        params.order[0] &&
        this.vitisCheck.isDefAndNotNull(params.order[0].column)) {
        if (params.columns &&
          params.columns[params.order[0].column] &&
          params.columns[params.order[0].column].field) {
          const sOrderField = params.columns[params.order[0].column].field;
          oResults.data = oResults.data.sort((a, b) => {
            if (this.vitisCheck.isNumber(a[sOrderField]) &&
              this.vitisCheck.isNumber(b[sOrderField])) {
              if (params.order[0].dir === 'desc') {
                return b[sOrderField] - a[sOrderField];
              } else {
                return a[sOrderField] - b[sOrderField];
              }
            } else if (this.vitisCheck.isString(a[sOrderField]) &&
              this.vitisCheck.isString(b[sOrderField])) {
              const nameA = a[sOrderField].toUpperCase();
              const nameB = b[sOrderField].toUpperCase();
              if (params.order[0].dir === 'desc') {
                if (nameA < nameB) {
                  return 1;
                } else if (nameA > nameB) {
                  return -1;
                } else {
                  return 0;
                }
              } else {
                if (nameA < nameB) {
                  return -1;
                } else if (nameA > nameB) {
                  return 1;
                } else {
                  return 0;
                }
              }
            }
          });
        }
      }

      // Pagination
      if (this.vitisCheck.isNumber(params.length) &&
        this.vitisCheck.isNumber(params.start)) {
        oResults.data = oResults.data.slice(params.start, params.start + params.length);
      }
    }

    return oResults;
  }

}
