import { Component, OnInit, Input, ElementRef, ViewChild } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import { FormreaderField } from '../../formreader-field';
import { VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { DataGridComponent } from 'src/app/vitis/data-grid/data-grid.component';
import { DataGridOptions } from 'src/app/vitis/data-grid/data-grid.model';
import { DataGridStudioService } from 'src/app/vitis/data-grid/services/data-grid-studio.service';
import { NgbModal, NgbModalRef } from '@ng-bootstrap/ng-bootstrap';
import { FormReaderDataGridModalComponent } from './components/modal/data-grid-modal.component';
import { WklDateComponent } from 'src/app/vitis/workspacelist/template/wkl-date/wkl-date.component';
import { WklDownloadBtnComponent } from 'src/app/vitis/workspacelist/template/wkl-download-btn/wkl-download-btn.component';
import { WklExternalLinkComponent } from 'src/app/vitis/workspacelist/template/wkl-external-link/wkl-external-link.component';
import { WorkspaceListOptions } from 'src/app/vitis/workspacelist/workspacelist.model';
import { environment } from 'src/environments/environment';

declare let $: any;

/**
 * Type date-grid
 *
 * Arguments:
 * type : type de champ (date, datetime)
 * name : attribut utilisé par le champ
 * label : libellé utilisé par le champ
 * options : options de la grille de données
 * nb_cols : nombre de colonnes à utiliser (de 1 à 12)
 * id? : id du champ à utiliser
 * required? : champ requis
 *
 * Exemple:
 * {
 *   "type": "data-grid",
 *   "name": "data-grid",
 *   "label": "Grille de données",
 *   "nb_cols": 12,
 *   "options": {
 *     "rowId": 'id',
 *     "searching": false,
 *     "paging": false,
 *     "info": false,
 *     "select": true,
 *     "selectMode": 'multi',
 *     insert: true,
 *     update: true,
 *     display: true,
 *     delete: true,
 *     "columns": [{
 *         "field": "date",
 *         "title": "FORM_HISTORY_DATE_PUBLICATION_FME_PROJECTS_METADATA",
 *         "width": "15%"
 *       },
 *       {
 *         "field": "author",
 *         "title": "FORM_HISTORY_AUTHOR_PUBLICATION_FME_PROJECTS_METADATA",
 *         "width": "20%"
 *       },
 *       {
 *         "field": "description",
 *         "title": "FORM_HISTORY_DESCRIPTION_PUBLICATION_FME_PROJECTS_METADATA",
 *         "width": "65%"
 *       }
 *     ],
 *     "data": [
 *         [ "Admin-Export", "exportv3.fmw", true, "CTLPASPAD" ],
 *         [ "Admin-Form-Checker", "formChecker.fmw", false, "2EW0ET1BDHLWC" ]
 *     ],
 *     "iconButtons": [{
 *        "type": "display",
          "icon": "icon-vitis-search",
          "event": (data: any) => {
            this.displayModalForm(data, "display");
          }
 *       }
 *     ]
 *     "buttons": [{
 *         "type": "button",
 *         "class": "btn-primary",
 *         "label": "FORM_BTN_DELETE_HISTORY_METADATA_PUBLICATION_FME_PROJECTS",
 *         "icon": "fa fa-trash",
 *         "event": "deleteHistory()"
 *       },
 *       {
 *         "type": "button",
 *         "class": "btn-primary",
 *         "label": "FORM_BTN_ADD_HISTORY_METADATA_PUBLICATION_FME_PROJECTS",
 *         "icon": "fa fa-plus",
 *         "event": "addHistory()"
 *       }
 *     ]
 *   }
 * }
 */

@Component({
  selector: 'app-formreader-data-grid',
  templateUrl: './data-grid.component.html',
  styleUrls: ['./data-grid.component.scss']
})
export class FormReaderDataGridComponent extends FormreaderField implements OnInit {

  @Input() sSize: string;
  @Input() oField: any;
  @Input() oFormValues: any;
  @Input() oFormReader: any;

  /**
   * Enfant de type DataGridComponent
   */
  @ViewChild(DataGridComponent, { static: false }) oDataGrid: DataGridComponent;

  // Instance de la grille de données.
  public oModalFormOptions: any;

  // Fenêtre modale.
  public modalRef: NgbModalRef;
  // Affichage de la grille de données.
  public showDataGrid: boolean = true;
  // Options de la grille de données.
  public oDataGridOptions: DataGridOptions = {
    datatableOptions: {},
    datagridOptions: WorkspaceListOptions
  };

  /**
   * Constructeur
   */
  constructor(
    public translate: TranslateService,
    public vitisCheck: VitisCheck,
    public element: ElementRef,
    private vitisRequest: VitisRequest,
    private dataGridStudioService: DataGridStudioService,
    private modalService: NgbModal,
  ) {
    super(translate, vitisCheck);
  }

  /**
   * Chargement du composant
   */
  ngOnInit() {
    // Fonction getValue
    this.oField.getValue = () => {
      return this.getValue(this.oFormValues, this.oField);
    };

    // Fonction setValue
    this.oField.setValue = (oValue: any) => {
      if (this.vitisCheck.isString(oValue) && oValue != '') {
        oValue = JSON.parse(oValue);
      }
      const oNewValue = this.setValue(this.oFormValues, this.oField, oValue);
      this.refreshGrid();
      return oNewValue;
    };

    this.oField.refreshTooltip = () => {
      return this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    };

    // Tooltip
    this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);

    // Set contenu manuellement
    this.oField.setContent = (aRows: Array<any>) => {
      return this.setContent(aRows);
    };

    this.oField.displayModalForm = (sFormUrl: string, sFormMode: string = 'insert', oFormValues: any = {}, oOptions: any = {}, oFormDefinition: any = {}, sJsForm: string = null) => {
      return this.displayModalForm(sFormUrl, sFormMode, oFormValues, oOptions, oFormDefinition, sJsForm);
    };

    // Ajout fonctions
    this.oField.refreshGrid = () => {
      this.refreshGrid();
    };


    setTimeout(() => {
      this.initDataGrid();
    }, 100);
  }

  /**
   * Remplit la grille
   * @param aRows
   */
  public setContent(aRows: Array<any>) {
    this.oField.oDataGrid.clear();
    for (let i = 0; i < aRows.length; i++) {
      this.oField.oDataGrid.addRow_(aRows[i]);
      if (aRows[i].selected === true) {
      }
    }
  }

  /**
   * Retourne les valeurs de la grille
   */
  protected getValues(params: any = {}) {
    return new Promise((resolve, reject) => {
      const oResults = {
        recordsTotal: 0,
        recordsFiltered: 0,
        data: []
      };
      if (this.vitisCheck.isDefAndNotNull(this.oFormValues.value[this.oField.name]) && this.oFormValues.value[this.oField.name] !== '') {
        let aDatas = [];
        if (this.vitisCheck.isDefAndNotNull(params.start) && this.vitisCheck.isDefAndNotNull(params.length)) {
          aDatas = this.oFormValues.value[this.oField.name].slice(params.start, params.start + params.length);
        } else {
          aDatas = this.oFormValues.value[this.oField.name];
        }
        oResults.data = aDatas;
        oResults.recordsTotal = this.oFormValues.value[this.oField.name].length;
        oResults.recordsFiltered = aDatas.length;
        resolve(oResults);
      } else {
        resolve(oResults);
      }
    });
  }

  /**
   * Initialise la grille de données.
   * @param {DataGridComponent} oDataGrid
   * @returns void
   */
  protected initDataGrid() {
    if (this.vitisCheck.isDefAndNotNull(this.oField.options)) {
      // Trie des colonnes en fonction de l'index
      if (this.vitisCheck.isDefAndNotNull(this.oField.options.columns) && this.oField.options.columns.length > 0) {
        this.oField.options.columns = this.oField.options.columns.sort((a: any, b: any) => {
          // Les colonnes sans index sont mis au début
          const indexA = this.vitisCheck.isDefAndNotNull(a.index) ? a.index : -1;
          const indexB = this.vitisCheck.isDefAndNotNull(b.index) ? b.index : -1;
          if (indexA == indexB) {
            // Trie par ordre alphabetique les colonnes avec index
            if (indexA !== -1) {
              if (a.title < b.title) { return -1; }
              if (a.title > b.title) { return 1; }
            }
            return 0;
          }
          return indexA - indexB;
        });
      }

      // Paramètres du plugin DataTable.
      const oDataGridOptions: DataGridOptions = {
        datatableOptions: {
          rowId: this.oField.options.rowId || 'id',
          info: this.oField.options.info,
          select: this.oField.options.select || false,
          selectMode: this.oField.options.selectMode || false,
          lengthMenu: this.oField.options.lengthMenu ?? [5, 10, 20, 50, 100],
          pageLength: this.oField.options.pageLength,
          order: this.oField.options.bOrder || [0, 'asc'],
          update: this.oField.options.update === true,
          display: this.oField.options.display === true,
          delete: this.oField.options.delete === true,
          insert: this.oField.options.insert === true,
          paging: this.oField.options.paging === true,
          ajax: (data: any, callback: any, settings: any) => {
            this.getValues(data).then((oResponse: any) => {
              callback(oResponse); // Remplace les données
            }).catch(() => {
              callback({
                recordsTotal: 0,
                recordsFiltered: 0,
                data: []
              });
            });
          }
        },
        datagridOptions: {
          refresh: this.oField.options.refresh || false,
          export: this.oField.options.export || false,
          columns: this.formatColumnsByType(this.oField.options.columns)
        },
        buttons: this.vitisCheck.isArray(this.oField.options.buttons) ? this.oField.options.buttons : [],
        iconButtons: this.vitisCheck.isArray(this.oField.options.iconButtons) ? this.oField.options.iconButtons : [],
      };

      // Paramètres de la requête ajax.
      if (this.vitisCheck.isDefAndNotNull(this.oField.options.ajax)) {
        oDataGridOptions.datatableOptions.ajax = this.oField.options.ajax;
      }

      if (this.vitisCheck.isDefAndNotNull(this.oField.options.bOrder)) {
        oDataGridOptions.datatableOptions.order = this.oField.options.bOrder;
      }

      if (this.vitisCheck.isDefAndNotNull(oDataGridOptions.datagridOptions)) {
        if (this.vitisCheck.isDefAndNotNull(oDataGridOptions.datagridOptions.columns)) {

          let oDatagridColumns = [];
          if (this.vitisCheck.isArray(oDataGridOptions.datagridOptions.columns)) {
            oDatagridColumns = oDataGridOptions.datagridOptions.columns;
          }

          // Traitement des colonnes
          for (let i = 0; i < oDatagridColumns.length; i++) {

            const column = oDatagridColumns[i];

            // Index par défaut de la colonne
            column.index = i;

            column.field = column.data;

            // Traduction
            column.translateTitle = this.translate.instant(column.title);

            // Visibilité de la colonne
            column.visible = column.visible === false ? false : true;

            // Colonne type
            column.type = column.template ? 'html' : 'string';

            // Alignement du contenu de la colonne.
            column.className = 'table-align-' + (column.align ? column.align : 'left');
          }

          this.oDataGridOptions.datagridOptions.columns = oDatagridColumns;
        }
      }


      if (this.oField.options.refresh === true) {
        oDataGridOptions.datagridOptions.refresh = true;
      }

      if (this.oField.options.export === true) {
        oDataGridOptions.datagridOptions.export = true;
      }

      // Evènements sur les boutons si l'on n'est pas dans le studio.
      if (!this.vitisCheck.isDefAndNotNull(this.oField.studio_id_)) {

        if (this.vitisCheck.isDefAndNotNull(oDataGridOptions.buttons)
          && oDataGridOptions.buttons.length > 0) {
          for (let i = 0; i < oDataGridOptions.buttons.length; i++) {
            if (this.vitisCheck.isDefAndNotNull(oDataGridOptions.buttons[i].event) && !this.vitisCheck.isFunction(oDataGridOptions.buttons[i].event)) {
              const sButtonEvent = oDataGridOptions.buttons[i].event;
              oDataGridOptions.buttons[i].event = () => {
                try {
                  this.oFormReader.fieldsService.callFunction(sButtonEvent, this.oFormReader.getFormValues(), [this.oFormReader, this.oFormReader.getFormContainer()]);
                } catch (e) {
                  console.error(e);
                }
              };
            }
          }
        }

        if (this.vitisCheck.isDefAndNotNull(oDataGridOptions.iconButtons)
          && oDataGridOptions.iconButtons.length > 0) {
          for (let i = 0; i < oDataGridOptions.iconButtons.length; i++) {
            if (this.vitisCheck.isDefAndNotNull(oDataGridOptions.iconButtons[i].event) && !this.vitisCheck.isFunction(oDataGridOptions.iconButtons[i].event)) {
              const sButtonEvent = oDataGridOptions.iconButtons[i].event;
              oDataGridOptions.iconButtons[i].event = () => {
                try {
                  this.oFormReader.fieldsService.callFunction(sButtonEvent, this.oFormReader.getFormValues(), [this.oFormReader, this.oFormReader.getFormContainer()]);
                } catch (e) {
                  console.error(e);
                }
              };
            }
          }
        }

        // Ajoute les boutons (Ajouter, supprimer...).
        this.initButtons(oDataGridOptions);

        // Ajoute les boutons (Ajouter, supprimer...).
        this.initIconButtons(oDataGridOptions);
      }

      this.oDataGridOptions = oDataGridOptions;

      setTimeout(() => {
        // Création de la grille.
        this.oDataGrid.initTable(oDataGridOptions);

        this.oField.oDataGrid = this.oDataGrid;
      }, 500);

      // Référence du composant DataGridComponent;
      this.oField.oFormReaderDataGridComponent = this;
      // Sauve les options de la grille de données.
      this.oDataGridOptions = oDataGridOptions;
    }
  }

  /**
   * Initialise les boutons
   * @param  {Array} oDataGridOptions
   * @returns void
   */
  public initButtons(oDataGridOptions: DataGridOptions): void {
    const aButtons: any = [];
    // Bouton insert
    if (this.oField.options.insert === true) {
      aButtons.unshift({
        type: 'button',
        name: 'insert',
        class: 'btn-primary',
        label: 'LIST_BUTTON_ADD',
        icon: 'fa fa-plus',
        event: (sFormUrl) => {
          this.displayModalForm(sFormUrl, 'insert');
        }
      });
    }

    // Bouton delete
    if (this.oField.options.delete === true) {
      aButtons.unshift({
        type: 'button',
        name: 'delete',
        class: 'btn-primary',
        label: 'LIST_BUTTON_DELETE',
        icon: 'fa fa-trash',
        event: () => {
        }
      });
    }

    // Ajout des boutons d'action.
    if (aButtons.length > 0) {
      if (this.vitisCheck.isArray(oDataGridOptions.buttons)) {
        oDataGridOptions.buttons = oDataGridOptions.buttons.concat(aButtons);
      } else {
        oDataGridOptions.buttons = aButtons;
      }
    }
  }

  /**
   * Initialise les boutons présent dans chaque ligne du tableau
   */
  public initIconButtons(oDataGridOptions): void {
    const aButtons: any = [];
    let buttonUpdateExist = false;
    if (this.vitisCheck.isDefAndNotNull(oDataGridOptions.datatableOptions.update) &&
      oDataGridOptions.datatableOptions.update === true) {
      aButtons.push({
        type: 'update',
        icon: 'icon-vitis-draw',
        event: (sFormUrl: any) => {
          this.displayModalForm(sFormUrl, 'update');
        }
      });
      buttonUpdateExist = true;
    }

    if (this.vitisCheck.isDefAndNotNull(oDataGridOptions.datatableOptions.display) &&
      oDataGridOptions.datatableOptions.display === true) {
      // Vérifie qu'il y a un bouton update
      if (environment.properties.data_grid_update_btn_only === true && buttonUpdateExist) {
        oDataGridOptions.datatableOptions.display = false;
      } else {
        aButtons.push({
          type: 'display',
          icon: 'icon-vitis-search',
          event: (sFormUrl: any) => {
            this.displayModalForm(sFormUrl, 'display');
          }
        });
      }
    }

    oDataGridOptions.iconButtons = aButtons;
  }

  /**
   * Affiche un formulaire dans une fenêtre modale
   * @param  {string} sFormUrl
   * @param  {string} sFormMode
   * @param  {any} oFormValues
   * @param  {any} oModalOptions
   * @param  {any} oFormDefinition
   * @param  {string|null} sJsForm
   */
  public async displayModalForm(sFormUrl: string, sFormMode: string = 'insert', oFormValues: any = {}, oModalOptions: any = {}, oFormDefinition: any = {}, sJsForm = null) {
    // Chargement du formulaire ou récupération de la définition passée.
    if (sFormUrl != '') {
      oFormDefinition = await this.loadForm(sFormUrl);
    } else if (Object.keys(oFormDefinition).length == 0) {
      console.warn('No form definition to display in modal window.');
      return false;
    }

    // Définition et valeurs par défaut du formulaire pour le FormReader.
    const oModalFormOptions: any = {
      form: oFormDefinition,
      formName: sFormMode,
      formValues: oFormValues
    };
    // JS à charger
    if (sJsForm) {
      oModalFormOptions.oFormJS = {
        sJS: sJsForm,
        oAdditionalScope: {},
        oAdditionalServices: {
          oParentFormreader: this.oFormReader,
        }
      };
    }
    // Exécution du FormReader.
    this.oModalFormOptions = oModalFormOptions;
    // Titre de la modale.
    if (!this.vitisCheck.isDefAndNotNull(oModalOptions.title)) {
      oModalOptions.title = this.oModalFormOptions.form[sFormMode].title;
    }

    oModalOptions.this_ = this;
    // Affichage de la fenêtre modale du formulaire.
    this.modalRef = this.modalService.open(FormReaderDataGridModalComponent);
    // Options de la fenêtre modale.
    this.modalRef.componentInstance.mergeModalOptions(oModalOptions);
    // Définition du formulaire.
    this.modalRef.componentInstance.oModalFormOptions = this.oModalFormOptions;
  }

  /**
   * Charge un formulaire.
   * @param  {string} sFormUrl
   * @returns Promise
   */
  public loadForm(sFormUrl: string): Promise<any> {
    return new Promise((resolve, reject) => {
      // Chargement du formulaire.
      this.vitisRequest.ajaxRequestObservable({
        method: 'GET',
        url: sFormUrl,
      }).subscribe(resp => {
        if (resp.status === 200 && resp.data) {
          resolve(resp.data);
        }
      });
    });
  }

  /**
   * Lance le adjust de la grille pour réaligner les colonnes
   */
  public refreshGrid() {
    setTimeout(() => {
      if (this.vitisCheck.isDefAndNotNull(this.oField) &&
        this.vitisCheck.isDefAndNotNull(this.oDataGrid)) {
        this.oDataGrid.reloadTable();
      } else {
        console.error('Datagrid n\'est pas chargé, impossible d\'ajuster les colonnes (refreshGrid)');
      }
    }, 500);
  }

  /**
   * Ferme la modale du formulaire.
   * @returns void
   */
  public closeModalForm(): void {
    this.modalRef.componentInstance.closeModalForm();
  }

  /**
   * Formatage des colonnes suivant leur type.
   * @param  {any} aColumns
   * @returns any
   */
  public formatColumnsByType(aColumns: any): any {
    if (this.vitisCheck.isArray(aColumns) && aColumns.length > 0) {
      for (let i = 0; i < aColumns.length; i++) {
        switch (aColumns[i].type) {
          case 'image':
          case 'document':
            aColumns[i].align = 'center';
            aColumns[i].template = WklDownloadBtnComponent;
            break;
          case 'link':
            aColumns[i].align = 'left';
            aColumns[i].template = WklExternalLinkComponent;
            break;
          case 'date':
            aColumns[i].type = 'date';
            aColumns[i].searchType = 'date';
            aColumns[i].template = WklDateComponent;
            break;
          default:
        }
      }
    }
    return aColumns;
  }
}
