import { Component, OnInit, Input, ElementRef } from '@angular/core';
import { FormreaderField } from '../../formreader-field';
import { TranslateService } from '@ngx-translate/core';
import { VitisCheck, vitisAlert } from 'src/app/vitis/core/services/utils/vitis-utils';
import { UntypedFormGroup, UntypedFormControl } from '@angular/forms';

/**
 * Type liste éditable
 *
 * Arguments:
 * type : type de champ
 * name : attribut utilisé par le champ
 * label : libellé utilisé par le champ
 * nb_cols : nombre de colonnes à utiliser (de 1 à 12)
 * size : nombre de lignes à utiliser
 * prompt_title : titre à afficher dans la modale d'ajout
 * prompt_type : type de champ à ajouter dans la modale d'ajout (text, number, email, url)
 * id? : id du champ à utiliser
 * required? : champ requis
 * default_value? : valeur par défaut
 *
 * Exemple:
 * {
 *   "type": "editable_list",
 *   "name": "scales",
 *   "label": "Échelles",
 *   "prompt_title": "Nouvelle échelle",
 *   "prompt_type": "number",
 *   "required": false,
 *   "nb_cols": 6,
 *   "size": 10
 * }
 */
@Component({
  selector: 'app-editable-list',
  templateUrl: './editable-list.component.html',
  styleUrls: ['./editable-list.component.scss']
})
export class EditableListComponent extends FormreaderField implements OnInit {

  @Input() sSize: string;
  @Input() oField: any;
  @Input() oFormValues: any;
  public oTemporaryForm: any;

  /**
   * Constructeur
   */
  constructor(
    translate: TranslateService,
    vitisCheck: VitisCheck,
    public element: ElementRef
  ) {
    super(translate, vitisCheck);
  }

  /**
   * Chargement du composant
   */
  ngOnInit() {

    // Fonction getValue
    this.oField.getValue = () => {
      return this.getValue(this.oFormValues, this.oField);
    };

    // Fonction setValue
    this.oField.setValue = (oValue: any) => {
      return this.setValue(this.oFormValues, this.oField, oValue);
    };

    this.oField.refreshTooltip = () => {
      return this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    };

    // Tooltip
    this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);

    // Création FormGroup spécifique
    this.createSpecialFormGroup_();
  }

  /**
   * Création d'un FormGroup pour connaître la valeur selectionnée afin d'utiliser les différents boutons
   * Le FormGroup oFormValues n'est utilisé que pour la boucle
   */
  private createSpecialFormGroup_() {
    const formControl = {};
    formControl[this.oField.name] = new UntypedFormControl({ value: '', disabled: this.oField.disabled });
    this.oTemporaryForm = new UntypedFormGroup(formControl);
  }

  /**
   * Affiche un prompt et ajoute un élement à la liste éditable
   */
  public addElementInEditableList() {
    const this_ = this;
    if (!this.vitisCheck.isDefAndNotNull(this_.oField.prompt_title)) {
      this_.oField.prompt_title = 'FORM_MODAL_ADD_ELEMENT_LIST';
    }
    this.translate.get(this_.oField.prompt_title).subscribe(function (translations) {

      vitisAlert.alert({
        text: translations,
        input: this_.oField.prompt_type
      }).then((result) => {
        if (this_.vitisCheck.isDefAndNotNull(result.value) && result.value !== '') {
          if (!this_.vitisCheck.isDefAndNotNull(this_.oFormValues.value[this_.oField.name])) {
            this_.oFormValues.value[this_.oField.name] = [];
          }
          const aOptions = this_.oFormValues.value[this_.oField.name];
          const isInArray = aOptions.find(function (a) {
            return String(a) === result.value;
          });
          if (!isInArray) {
            const value = this_.oFormValues.value[this_.oField.name];
            value.push(result.value);
            this_.oFormValues.controls[this_.oField.name].setValue(value);
          }
        }
      });
    });
  }

  /**
   * Supprime un element à une liste editable après confirmation
   */
  public removeElementInEditableList() {
    const this_ = this;
    const value = this.oTemporaryForm.value[this_.oField.name];
    this.translate
      .get('FORM_MODAL_DELETE_VALUE_TOOLS_CONFIGURATION_CONFIGURATION_VMAP_CONFIG', { value: value })
      .subscribe(function (translations) {
        if (this_.vitisCheck.isDefAndNotNull(this_.oFormValues.value[this_.oField.name])) {
          const aOptions = this_.oFormValues.value[this_.oField.name];

          vitisAlert.alert({
            text: translations,
            showCancelButton: true
          }).then(result => {
            if (result.value) {
              const newValue = aOptions.filter((a) => String(a) !== String(value));
              this_.oFormValues.controls[this_.oField.name].setValue(newValue);
              this_.oFormValues.controls[this_.oField.name].touched = true;
            }
          });
        }
      });
  }

  /**
   * Déplace un element dans une liste editable (direction -1 pour up et 1 pour down)
   * @param {number} direction
   */
  public moveElementInEditableList(direction: number) {
    const selectedValue = this.oTemporaryForm.value[this.oField.name];
    const aOptions = this.oFormValues.value[this.oField.name];
    if (this.vitisCheck.isDefAndNotNull(selectedValue) && this.vitisCheck.isDefAndNotNull(aOptions)) {
      const iIndex = aOptions.findIndex((item: any) => String(item) === String(selectedValue));
      if (iIndex !== -1) {
        const iIndex2 = iIndex + direction;
        if (this.vitisCheck.isDefAndNotNull(aOptions[iIndex2])) {
          const tmp = aOptions[iIndex];
          aOptions[iIndex] = aOptions[iIndex2];
          aOptions[iIndex2] = tmp;
        }
      }
    }
  }

  /**
   * Retourne les valeurs sélectionnées de la liste au format texte séparées par des "|"
   * @param oFormValues
   * @param oField
   *
   * @returns {String}
   */
  public getValue(oFormValues: any, oField: any): string {
    let sReturn: any = '';
    if (oFormValues) {
      if (oFormValues.value) {
        if (oFormValues.value[oField.name]) {
          sReturn = oFormValues.value[oField.name];

          // Retour au format string séparé par des "|"
          if (this.vitisCheck.isArray(sReturn)) {
            sReturn = sReturn.join('|');
          }
        }
      }
    }
    return sReturn;
  }

  /**
     * Setter de valeur du champ
     * @param  {any} oFormValues
     * @param  {any} oField
     * @param  {any} oValue
     */
  public setValue(oFormValues: any, oField: any, oValue: any) {

    if (this.vitisCheck.isString(oValue)) {
      oValue = oValue.split('|');
    }

    const oPatchValues = {};
    if (this.vitisCheck.isDefAndNotNull(oField)) {
      if (this.vitisCheck.isDefAndNotNull(oField.name)) {
        oPatchValues[oField.name] = oValue;
      }
    }

    oFormValues.patchValue(oPatchValues);
  }
}
