import { Component, OnInit, Input, ElementRef } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import { FormreaderField } from '../../formreader-field';
import { VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';
import { UntypedFormControl, UntypedFormGroup } from '@angular/forms';

/**
 * Type selecteur d'icône
 *
 * Arguments:
 * type : type de champ
 * name : attribut utilisé par le champ
 * label : libellé utilisé par le champ
 * nb_cols : nombre de colonnes à utiliser (de 1 à 12)
 * id? : id du champ à utiliser
 * required? : champ requis
 * default_value? : valeur par défaut
 *
 * Exemple:
 * {
 *   "type": "icon",
 *   "name": "icon",
 *   "label": "Icône",
 *   "required": true,
 *   "nb_cols": 12,
 * }
 */
@Component({
  selector: 'app-icon',
  templateUrl: './icon.component.html',
  styleUrls: ['./icon.component.scss']
})
export class IconComponent extends FormreaderField implements OnInit {

  @Input() sSize: string;
  @Input() oField: any;
  @Input() oFormValues: any;

  public iconList: Array<any>;
  public availableIcons: Array<any>;
  public emptyIcon: any = {
    name: '',
    css_class: '',
    tags: ['vide', 'empty'],
  };
  public selectedIcon = this.emptyIcon;
  public filterGroup = new UntypedFormGroup({
    sFilter: new UntypedFormControl()
  });

  /**
   * Constructeur
   */
  constructor(
    translate: TranslateService,
    vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    public element: ElementRef
  ) {
    super(translate, vitisCheck);
  }

  /**
   * Chargement du composant
   */
  ngOnInit() {

    // Fonction getValue
    this.oField.getValue = () => {
      return this.getValue(this.oFormValues, this.oField);
    };

    // Fonction setValue
    this.oField.setValue = (oValue: any) => {
      this.selectedIcon = {
        css_class: oValue
      };
      return this.setValue(this.oFormValues, this.oField, String(oValue));
    };

    this.oField.refreshTooltip = () => {
      return this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    };

    // Tooltip
    this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);

    // Liste des icones disponibles
    this.getAvailableIcons_().then((availableIcons) => {
      this.availableIcons = availableIcons;
      this.iconList = this.getIconList_(this.availableIcons, '');
    });

    this.filterGroup.get('sFilter').valueChanges.subscribe(value => {
      this.iconList = this.getIconList_(this.availableIcons, value);
    });
  }

  /**
   * Sélection d'une icone
   * @param oIcon
   */
  public selectIcon(oIcon: any) {
    this.selectedIcon = oIcon;
    this.oField.setValue(oIcon.css_class);
  }

  /**
   * Télécharge la liste des icones disponibles
   */
  private getAvailableIcons_(): Promise<any> {
    return new Promise((resolve, reject) => {
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: environment.properties.app_api_url + '/vitis/icons',
      }).then((resp: any) => {
        if (resp.data &&
          resp.data.data &&
          resp.data.data) {

          const aIcons = [this.emptyIcon, ...resp.data.data];
          for (let i = 0; i < aIcons.length; i++) {
            const icon = aIcons[i];
            try {
              icon.tags = JSON.parse(icon.tags);
            } catch (error) {
              icon.tags = [];
            }
          }

          resolve(aIcons);
        } else {
          console.error('could not get userorders', resp);
          reject();
        }
      }).catch((err) => {
        console.error('could not get userorders', err);
        reject();
      });
    });
  }

  /**
   * Retourne la liste filtrée
   * @param availableIcons
   * @param sFilter
   */
  public getIconList_(availableIcons: Array<any>, sFilter: string): Array<any> {
    if (!availableIcons || !sFilter || sFilter === '') {
      return availableIcons;
    }
    return availableIcons.filter(icon =>
      (icon.name.indexOf(sFilter.toLowerCase().normalize('NFD').replace(/[\u0300-\u036f]/g, '')) !== -1) ||
      (icon.tags.find((tag: string) => tag.indexOf(sFilter.toLowerCase().normalize('NFD').replace(/[\u0300-\u036f]/g, '')) !== -1)));
  }
}
