import { Component, OnInit, Input, HostListener, ElementRef } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';

import { FormreaderField } from '../../formreader-field';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';

/**
 * Type monaco-editor
 *
 * Arguments:
 * type : type de champ
 * name : attribut utilisé par le champ
 * label : libellé utilisé par le champ
 * nb_cols : nombre de colonnes à utiliser (de 1 à 12)
 * id? : id du champ à utiliser
 * height?: hauteur du champ de saisi monaco
 * auto_complete_custom?: tableau de completionItems pour ajouter de l'auto complétion à monaco editor
 * auto_complete_tooltip?: si un tableau d'auto complétion est défini, permet de générer automatiquement un tooltip contenant le détail de l'autocomplétion
 *
 * Exemple:
 * {
 *     "type": "monaco-editor",
 *     "name": "code",
 *     "label": "Code",
 *     "nb_cols": 12
 * }
 *
 * Documentation pour les completionItems:
 * https://microsoft.github.io/monaco-editor/api/interfaces/monaco.languages.completionitem.html
 */
@Component({
  selector: 'app-monaco-editor',
  templateUrl: './monaco-editor.component.html',
  styleUrls: ['./monaco-editor.component.scss']
})
export class MonacoEditorComponent extends FormreaderField implements OnInit {

  @Input() sSize: string;
  @Input() oField: any;
  @Input() oFormValues: any;

  public height: string;

  /**
   * Constructeur
   */
  constructor(
    translate: TranslateService,
    vitisCheck: VitisCheck,
    public element: ElementRef
  ) {
    super(translate, vitisCheck);
  }

  /**
   * Permet d'écouter les changements de taille
   * @param event
   */
  @HostListener('window:resize', ['$event'])

  /**
   * Retourne la hauteur du champ
   */
  getScreenSize(event?) {
    if (this.oField.height) {
      this.height = this.oField.height;
    } else {
      // Hauteur par défaut = la moitié de la hauteur de l'écran.
      this.height = (window.innerHeight / 2) + 'px';
    }
  }

  /**
   * Chargement du composant
   */
  ngOnInit() {
    // Fonction getValue
    this.oField.getValue = () => {
      let oValue = this.getValue(this.oFormValues, this.oField);
      if (this.oField.options && this.oField.options.language === 'php') {
        if (oValue.indexOf('<?php\n') != -1) {
          oValue = oValue.replace('<?php\n', '');
        } else if (oValue.indexOf('<?php') != -1) {
          oValue = oValue.replace('<?php', '');
        }
      }
      return oValue;
    };

    // Fonction setValue
    this.oField.setValue = (oValue: any) => {
      if (this.oField.options && this.oField.options.language === 'php') {
        if (oValue.indexOf('<?php') == -1) {
          oValue = '<?php\n' + oValue;
        }
      }
      return this.setValue(this.oFormValues, this.oField, oValue);
    };

    this.oField.refreshTooltip = () => {
      return this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    };

    this.getScreenSize();

    // Créer le tooltip des variables autorisé à partir de l'auto complete
    if (this.vitisCheck.isDefAndNotNull(this.oField.auto_complete_custom) &&
      this.vitisCheck.isDefAndNotNull(this.oField.auto_complete_tooltip)) {

      this.oField.tooltip = {
        title: 'TOOLTIP_MONACOEDITOR_AUTOCOMPLETE_TITLE',
        content: '',
        container: 'body',
        width: '950px',
        sanitize: false,
        trigger: 'manual',
        html: true
      };

      const translates = ['TOOLTIP_MONACOEDITOR_AUTOCOMPLETE_VARIABLE', 'TOOLTIP_MONACOEDITOR_AUTOCOMPLETE_DESCRIPTION', 'TOOLTIP_MONACOEDITOR_AUTOCOMPLETE_EXAMPLE'];

      for (let i = 0; i < this.oField.auto_complete_custom.length; i++) {
        const element = this.oField.auto_complete_custom[i];

        if (this.vitisCheck.isDefAndNotNull(element.documentation)) {
          translates.push(this.oField.auto_complete_custom[i].documentation);
        }

        if (!this.vitisCheck.isDefAndNotNull(element.kind)) {
          element.kind = 'monaco.languages.CompletionItemKind.Variable';
        }
      }

      this.translate.get(translates).subscribe((translations) => {
        let content = '<table border="1" style="width: 100%;">';
        content += '<tr style="width: 100%;">';
        content += '<th scope="col" style="">' + translations['TOOLTIP_MONACOEDITOR_AUTOCOMPLETE_VARIABLE'] + '</th>';
        content += '<th scope="col" style="">' + translations['TOOLTIP_MONACOEDITOR_AUTOCOMPLETE_DESCRIPTION'] + '</th>';
        content += '<th scope="col" style="">' + translations['TOOLTIP_MONACOEDITOR_AUTOCOMPLETE_EXAMPLE'] + '</th>';
        content += '</tr>';

        for (let i = 0; i < this.oField.auto_complete_custom.length; i++) {
          const autocompleteCustom = this.oField.auto_complete_custom[i];

          if (this.vitisCheck.isDefAndNotNull(autocompleteCustom.documentation)) {
            content += '<tr style="width: 100%;">';
            content += '<td style="text-align: left;">' + autocompleteCustom.label + '</td>';
            content += '<td style="text-align: left;">' + translations[autocompleteCustom.documentation] + '</td>';
            content += '<td style="text-align: left;">' + autocompleteCustom.example + '</td>';
            content += '</tr>';
          }
        }

        content += '</table>';
        this.oField.tooltip.content = content;

        // Tooltip
        this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
      });
    } else {
      // si pas de tooltip pour l'auto_complétion on passe ne mode classique
      this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    }

  }

  /**
   * Evenement quand monaco editor est initialisé
   * Si il y en a dans la configuration de oField, ajoute les auto completions custom
   */
  public onMonacoEditorInit() {
    // Ajout provideCompletionItems si il y a en a.
    if (this.vitisCheck.isDefAndNotNull(this.oField.auto_complete_custom)) {
      const this_ = this;
      const monaco = window['monaco'];
      monaco.languages.registerCompletionItemProvider(this_.oField.options.language, {
        provideCompletionItems: function (model, position, context, token) {
          // Un seul retour de la liste des suggestions (sinon doublons).
          if (model['_versionId'] != window['iSuggestionVersionId']) {
            window['iSuggestionVersionId'] = model['_versionId'];
            return {
              suggestions: this_.oField.auto_complete_custom
            };
          }
        }
      });
    }

  }

}
