import { Component, OnInit, Input, ElementRef } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import { FormreaderField } from '../../formreader-field';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { Validators } from '@angular/forms';

/**
 * Type nombre
 *
 * Arguments:
 * type : type de champ (integer/number/float) float=number
 * name : attribut utilisé par le champ
 * label : libellé utilisé par le champ
 * nb_cols : nombre de colonnes à utiliser (de 1 à 12)
 * id? : id du champ à utiliser
 * required? : champ requis
 * default_value? : valeur par défaut
 * pattern? : regex à utiliser pour contraintre la saisie
 *
 * Exemple:
 * {
 *     "type": "number",
 *     "name": "name",
 *     "label": "Nom",
 *     "required": false,
 *     "nb_cols": 12
 * }
 */
@Component({
  selector: 'app-number',
  templateUrl: './number.component.html',
  styleUrls: ['./number.component.scss']
})
export class NumberComponent extends FormreaderField implements OnInit {

  @Input() sSize: string;
  @Input() oField: any;
  @Input() oFormValues: any;

  /**
   * Constructeur
   */
  constructor(
    translate: TranslateService,
    vitisCheck: VitisCheck,
    public element: ElementRef
  ) {
    super(translate, vitisCheck);
  }

  /**
   * Chargement du composant
   */
  ngOnInit() {

    // Fonction getValue
    this.oField.getValue = () => {
      let val = this.getValue(this.oFormValues, this.oField);

      // Cas où l'on veuille vider la valeur
      if (this.vitisCheck.isNull(val)) {
        val = '';
      }

      return val;
    };

    // Fonction setValue
    this.oField.setValue = (oValue: any) => {
      return this.setValue(this.oFormValues, this.oField, oValue);
    };

    this.oField.refreshTooltip = () => {
      return this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    };

    // Tooltip
    this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);

    // validateurs min/max
    if (this.oField.options &&
      (this.oField.options.min || this.oField.options.max)) {
      this.updateFieldValidator_(this.oField);
    }
  }

  /**
   * Permet de contraindre la saisie pour le type integer
   * @param  {any} event
   * @returns boolean
   */
  public numberOnly(event: any): boolean {
    if (this.oField.type === 'integer') {
      const charCode = (event.which) ? event.which : event.keyCode;
      if (charCode > 31 && (charCode < 48 || charCode > 57)) {
        return false;
      }
    }
    return true;
  }

  /**
  * Création d'un validateur spécifique
  * @param oField
  */
  private updateFieldValidator_(oField: any) {

    let aValidators = [];

    // Récupère les validators inclus par FormReader
    if (this.vitisCheck.isArray(oField.validators_list)) {
      aValidators = oField.validators_list;
    }

    // Validateur min
    if (oField.options && oField.options.min && !isNaN(oField.options.min)) {
      aValidators.push(Validators.min(oField.options.min));
    }

    // Validateur max
    if (oField.options && oField.options.max && !isNaN(oField.options.max)) {
      aValidators.push(Validators.max(oField.options.max));
    }

    this.oFormValues.controls[oField.name].setValidators(aValidators);

    if (!oField.validation_messages) {
      oField.validation_messages = [];
    }

    // Texte validateur min
    if (oField.options && oField.options.min && !isNaN(oField.options.min)) {
      oField.validation_messages.push({
        type: 'min',
        message: this.translate.instant('FORM_READER_ERROR_NUMBER_MIN')
      });
    }

    // Texte validateur max
    if (oField.options && oField.options.max && !isNaN(oField.options.max)) {
      oField.validation_messages.push({
        type: 'max',
        message: this.translate.instant('FORM_READER_ERROR_NUMBER_MAX')
      });
    }
  }
}
