import { Component, OnInit, Input, ElementRef, AfterViewInit } from '@angular/core';
import { FormreaderField } from '../../formreader-field';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { TranslateService } from '@ngx-translate/core';
import { environment } from 'src/environments/environment';

declare let tinyMCE: any;

/**
 * Type tinymce
 *
 * Arguments:
 * type : type de champ
 * name : attribut utilisé par le champ
 * label : libellé utilisé par le champ
 * nb_cols : nombre de colonnes à utiliser (de 1 à 12)
 * id? : id du champ à utiliser
 * required? : champ requis
 * readonly? : Désactivation du composant.
 * options : options qui permettent l'initialisation du tinymce
 *
 * Exemple:
 * {
 *     "type": "tinymce",
 *     "name": "body",
 *     "label": "Message",
 *     "required": false,
 *     "nb_cols": 12
 *     "options": {
 *       plugins: 'autoresize',
 *       min_height: 200
 *     }
 * }
 */
@Component({
  selector: 'app-tinymce',
  templateUrl: './tinymce.component.html',
  styleUrls: ['./tinymce.component.scss']
})
export class TinymceComponent extends FormreaderField implements OnInit, AfterViewInit {

  @Input() sSize: string;
  @Input() oField: any;
  @Input() oFormValues: any;

  /**
   * Constructeur
   */
  constructor(
    translate: TranslateService,
    vitisCheck: VitisCheck,
    public element: ElementRef
  ) {
    super(translate, vitisCheck);
  }

  /**
   * Chargement du composant
   */
  ngOnInit() {

    // Fonction getValue
    this.oField.getValue = () => {
      return this.getValue(this.oFormValues, this.oField);
    };

    // Fonction setValue
    this.oField.setValue = (oValue: any) => {
      return this.setValue(this.oFormValues, this.oField, oValue);
    };

    this.oField.refreshTooltip = () => {
      return this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    };

    this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);

    // Définit la langue de TinyMCE en fonction de la
    if (this.vitisCheck.isDefAndNotNull(this.oField.options)) {
      switch (environment.properties.language) {
        case 'fr':
          this.oField.options.language = 'fr_FR';
          break;

        default:
          this.oField.options.language = 'en';
          break;
      }
    }
  }

  /**
   * Appelé après le chargement du DOM
   */
  ngAfterViewInit() {
    // Mode "readonly".
    if (this.oField.readonly === true) {
      if (this.vitisCheck.isDefAndNotNull(this.oField.id)) {
        tinyMCE.get(this.oField.id).setMode('readonly');
      }
    }
  }

  /**
   * Retourne true si des options ont été fournies
   * @returns
   */
  public haveOptions() {
    return !this.vitisCheck.isEmptyObject(this.oField.options);
  }
}
