import { Component, OnInit, Input, ElementRef, ViewChild, AfterViewInit } from '@angular/core';
import { TranslateService } from '@ngx-translate/core';
import { Subject } from 'rxjs';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { TreeviewComponent } from 'src/app/vitis/shared/components/treeview/treeview.component';
import { FormreaderField } from '../../formreader-field';

@Component({
  selector: 'app-treeview-formreader',
  templateUrl: './treeview-formreader.component.html',
  styleUrls: ['./treeview-formreader.component.scss']
})
export class TreeviewFormreaderComponent extends FormreaderField implements OnInit, AfterViewInit {

  @Input() sSize: string;
  @Input() oField: any;
  @Input() oFormValues: any;
  @Input() oFormReader: any;

  @ViewChild(TreeviewComponent, { static: false }) oTreeview: TreeviewComponent;

  private defaultNode: any = {
    text: '',
    value: '',
    collapsed: true,
    selected: false,
    selectable: false,
    visible: true,
  };

  private nodeAttribute = 'nodes';

  private memorize_last_value: boolean = true;
  private last_value: any[] = [];

  public checkAll: boolean = false;

  /**
   * Constructeur
   */
  constructor(
    translate: TranslateService,
    vitisCheck: VitisCheck,
    public element: ElementRef
  ) {
    super(translate, vitisCheck);
  }

  /**
   * Chargement du composant
   */
  ngOnInit() {

    // Fonction getValue
    this.oField.getValue = () => {
      return this.getValue(this.oFormValues, this.oField);
    };

    // Fonction setValue
    this.oField.setValue = (oValue: any) => {
      return this.setValue(this.oFormValues, this.oField, oValue);
    };

    this.oField.refreshTooltip = () => {
      return this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);
    };

    this.oField.toggleSelectAll = () => {
      return this.toggleSelectAll();
    };

    // Tooltip
    this.tooltip(this.oFormValues, this.oField, this.element.nativeElement);

    // Observable des options
    this.oField.subscribeOptions = new Subject();

    // Gère l'attribut des enfants
    this.nodeAttribute = this.oField.options.nodeAttribute;

    if (this.oField.datasource) {
      // Observe l'arrivée des choix (pour utilisation avec datasource)
      if (this.vitisCheck.isDefAndNotNull(this.oField.options)) {
        if (this.vitisCheck.isDefAndNotNull(this.oField.options.observableChoises)) {
          this.oField.options.observableChoises.subscribe((aChoises: any) => {

            this.generateData(aChoises);
            this.memorize_last_value = false;

            // Set valeur "" si aucune valeur
            setTimeout(() => {
              if (!this.vitisCheck.isDefAndNotNull(this.oField.getValue()) && !this.oField.disabled) {
                this.oField.setValue([]);
              }

              if (this.last_value.length != 0) {
                this.oField.setValue(this.last_value);
              }

              this.checkAll = this.isAllSelected();
            });
          });
        }
      }
    }
  }

  /**
   * Appelé après le chargement du DOM
   */
  ngAfterViewInit(): void {
    this.oField.getTreeview = () => this.oTreeview;
  }

  /**
   * généère les données à partir de la datasources
   * @param aChoises
   */
  private generateData(aChoises: any[]): void {

    let aTree: any[] = [];
    const sDelimiter: string = this.oField.options.node_delimiter;

    for (const choice of aChoises) {
      if (!choice.label || choice.label == '') {
        continue;
      }
      if (!choice.value || choice.value == '') {
        continue;
      }

      aTree = this.processNode(choice.label.split(sDelimiter), choice.value, aTree);
    }

    this.oField.options.data = aTree;
  }

  /**
   * généère un chemin de noeud
   * @param aPath chemin
   * @param val valeur du noeud final
   * @param Tree tableau de noeud enfants
   * @returns tableau d enoeud enfant complété
   */
  private processNode(aPath: string[], val: any, Tree: any[]): any[] {

    if (aPath.length === 0) {
      return [];
    }

    const sItem: string = aPath.shift();
    let found = false;

    for (const index in Tree) {

      if (!Tree[index]) {
        continue;
      }
      const oNode: any = Tree[index];
      if (oNode.text == sItem) {
        found = true;
        let children: any[] = (oNode[this.nodeAttribute]) ? oNode[this.nodeAttribute] : [];
        children = this.processNode(aPath, val, children);
        if (children.length > 0) {
          Tree[index][this.nodeAttribute] = children;
        }
      }
    }

    if (!found) {
      const children = this.processNode(aPath, val, []);
      const oNode = {
        ...this.defaultNode,
        text: sItem,
      };

      if (children.length > 0) {
        oNode[this.nodeAttribute] = children;
      } else {
        oNode.value = val;
        oNode.selectable = (this.oField.disabled != true);
      }

      Tree.push(oNode);
    }

    return Tree;
  }

  /**
   * set value pour le formreader
   * @param oFormValues
   * @param oField
   * @param oValue
   * @returns
   */
  public setValue(oFormValues: any, oField: any, oValue: any): void {
    if (!Array.isArray(oValue)) {
      oValue = [oValue];
    }

    if (this.memorize_last_value) {
      this.last_value = oValue;
      return;
    }

    for (const oVal of oValue) {
      this.oField.options.data = this.setValue_(oVal, this.oField.options.data);
    }
    if (!this.vitisCheck.isDefAndNotNull(oValue) || oValue.length === 0) {
      if (this.checkAll) {
        this.checkAll = false;
      }
    }
    this.oField.subscribeOptions.next(oValue);
  }
  /**
   * Set réellement une valeur sur le composant
   * @param {any} oValue
   * @param {any[]} aTree
   * @returns {any[]}
   */
  private setValue_(oValue: any, aTree: any[]): any[] {

    for (const index in aTree) {
      if (!aTree[index]) {
        continue;
      }
      if (aTree[index].value === oValue) {
        aTree[index].selected = true;
      }

      if (aTree[index][this.nodeAttribute]) {
        aTree[index][this.nodeAttribute] = this.setValue_(oValue, aTree[index][this.nodeAttribute]);
      }
    }

    return aTree;
  }

  /**
   * getValue pour le formreader
   * @param oFormValues
   * @param oField
   * @returns
   */
  public getValue(oFormValues: any, oField: any): any[] {
    const aItems: any[] = this.oTreeview.getSelectedItems();
    const aVal: any[] = [];
    for (const Item of aItems) {
      if (this.vitisCheck.isDefAndNotNull(Item.value) && Item.value !== '') {
        aVal.push(Item.value);
      } else {
        aVal.push(Item[this.oField.options.pathAttribute]);
      }
    }
    return aVal;
  }

  /**
   * défini si tout est check dans l'arbre
   * @returns {boolean}
   */
  public isAllSelected(): boolean {
    return this.isAllSelected_(this.oField.options.data);
  }

  /**
   * parcours l'arbre pour savoir si tout est sélectionné
   * @param aTree
   * @returns {boolean}
   */
  private isAllSelected_(aTree: any[]): boolean {
    for (const oNode of aTree) {
      if (oNode.value !== '' && oNode.selected == false) {
        return false;
      }

      if (Array.isArray(oNode[this.nodeAttribute])) {
        if (!this.isAllSelected_(oNode[this.nodeAttribute])) {
          return false;
        }
      }
    }

    return true;
  }

  /**
   * Change le status de tous les noeud sélectionnable
   */
  public toggleSelectAll(): void {
    this.checkAll = !this.checkAll;
    this.changeAllSelectstatus(this.oField.options.data, this.checkAll);
    this.memorize_last_value = false;
    this.setValue(this.oFormValues, this.oField, this.oTreeview.getSelectedValues());
  }

  /**
   * change le status de tous les élement qui ont une valeur
   * @param aTree
   * @param bStatus
   * @returns
   */
  private changeAllSelectstatus(aTree: any, bStatus: boolean): any[] {
    for (const index in aTree) {
      if (!aTree[index]) {
        continue;
      }
      if (aTree[index].selectable == true) {
        aTree[index].selected = bStatus;
      }

      if (Array.isArray(aTree[index][this.nodeAttribute])) {
        aTree[index][this.nodeAttribute] = this.changeAllSelectstatus(aTree[index][this.nodeAttribute], bStatus);
      }
    }

    return aTree;
  }

  /**
   * Event selection change dans le treeview
   * @param event
   */
  public selectedChange(event) {
    this.memorize_last_value = false;
    this.setValue(this.oFormValues, this.oField, event);
    if (this.checkAll) {
      this.checkAll = false;
    }
  }
}
