import { Injectable } from '@angular/core';
import { VitisCheck, vitisAlert, VitisRequest, VitisFileRequest, VitisClone } from 'src/app/vitis/core/services/utils/vitis-utils';
import { Tabs } from '../formreader.model';
import { environment } from 'src/environments/environment';
import { TranslateService } from '@ngx-translate/core';
import { Subject } from 'rxjs';
import { ActivatedRoute, Router } from '@angular/router';
import { AuthService } from '../../core/services/auth/auth.service';
import { InterfaceService } from '../../core/services/utils/interface/interface.service';
import { SessionStorageService } from '../../core/services/utils/sessionstorage/session-storage.service';
import { LocalStorageService } from '../../core/services/utils/localstorage/local-storage.service';
import { FormJS } from '../formreader.component';
import { FormControl, Validators } from '@angular/forms';

/**
 * Service FormReader
 *
 * Injectable et contient des fonctions de base sur le formulaire
 *
 */
@Injectable({
  providedIn: 'root'
})
export class FormreaderService {

  private mapChangeEvent = new Subject<any>();
  private mapChangeEventObservable = this.mapChangeEvent.asObservable();

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private vitisFileRequest: VitisFileRequest,
    private translate: TranslateService,
    private route: ActivatedRoute,
    private router: Router,
    private auth: AuthService,
    private interfaceService: InterfaceService,
    private sessionStorageService: SessionStorageService,
    private localStorageService: LocalStorageService,
    private vitisClone: VitisClone,
  ) { }

  /**
   * Retourne l'élément du formulaire
   * @param  {string} sFieldName
   * @param  {any} oFormDefinition
   * @param  {any} sFormDefinitionName
   * @returns Any
   */
  public getFormField(sFieldName: string, oFormDefinition: any, sFormDefinitionName: any): any {
    const aFields = this.getFormFields(oFormDefinition, sFormDefinitionName);
    for (let i = 0; i < aFields.length; i++) {
      const oField = aFields[i];
      if (oField.name === sFieldName) {
        return oField;
      }

      if (oField.type === 'button') {
        for (let j = 0; j < oField.buttons.length; j++) {
          const oFieldBtn = oField.buttons[j];
          if (oFieldBtn.name === sFieldName) {
            return oFieldBtn;
          }
        }
      }
    }
  }

  /**
   * Retourne un tableau contenant les elements du formulaire
   * @param  {any} oFormDefinition
   * @param  {any} sFormDefinitionName
   * @returns Array
   */
  public getFormFields(oFormDefinition: any, sFormDefinitionName: any): Array<any> {

    // Linges
    const aRows = this.getFormRows(oFormDefinition, sFormDefinitionName);

    // Default
    const aFields = [];

    // Vérifie que tout soit bien définit et ajoute la ligne
    if (this.vitisCheck.isArray(aRows)) {
      for (let i = 0; i < aRows.length; i++) {
        if (this.vitisCheck.isArray(aRows[i].fields)) {
          for (let ii = 0; ii < aRows[i].fields.length; ii++) {
            aFields.push(aRows[i].fields[ii]);
          }
        }
      }
    }

    return aFields;
  }

  /**
   * Retourne un tableau contenant les noms des champs
   * @param oFormDefinition
   * @param sFormDefinitionName
   */
  public getFormFieldsNames(oFormDefinition: any, sFormDefinitionName: string): Array<string> {
    const aFieldsNames = [];
    const aFields = this.getFormFields(oFormDefinition, sFormDefinitionName);

    for (let i = 0; i < aFields.length; i++) {
      const oField = aFields[i];
      if (oField.name) {
        aFieldsNames.push(oField.name);
      } else if (this.vitisCheck.isDefAndNotNull(oField.buttons)) {
        for (let ii = 0; ii < oField.buttons.length; ii++) {
          if (this.vitisCheck.isDefAndNotNull(oField.buttons[ii].name)) {
            aFieldsNames.push(oField.buttons[ii].name);
          }
        }
      }
    }

    return aFieldsNames;
  }

  /**
   * Retourne un tableau contenant les lignes du formulaire
   * @param  {any} oFormDefinition
   * @param  {any} sFormDefinitionName
   * @returns Array
   */
  public getFormRows(oFormDefinition: any, sFormDefinitionName: any): Array<any> {

    // Vérifie que tout soit bien définit et retourne les colones
    if (this.vitisCheck.isDefAndNotNull(oFormDefinition)) {
      if (this.vitisCheck.isDefAndNotNull(sFormDefinitionName)) {
        if (this.vitisCheck.isDefAndNotNull(oFormDefinition[sFormDefinitionName])) {
          if (this.vitisCheck.isArray(oFormDefinition[sFormDefinitionName].rows)) {
            return oFormDefinition[sFormDefinitionName].rows;
          }
        }
      }
    }

    // Default
    return [];
  }

  /**
   * Retourne le nombre de colones utilisées pour le formulaire
   * @param  {any} oFormDefinition
   * @param  {any} sFormDefinitionName
   * @returns string
   */
  public getFormSize(oFormDefinition: any, sFormDefinitionName: any): string {

    // Récupère la taille du formulaire
    if (this.vitisCheck.isDefAndNotNull(oFormDefinition)) {
      if (this.vitisCheck.isDefAndNotNull(sFormDefinitionName)) {
        if (this.vitisCheck.isDefAndNotNull(oFormDefinition[sFormDefinitionName])) {
          if (this.vitisCheck.isDefAndNotNull(oFormDefinition[sFormDefinitionName].nb_cols)) {
            return oFormDefinition[sFormDefinitionName].nb_cols;
          }
        }
      }
    }

    // Default
    return 'standard';
  }

  /**
   * Retourne la taille des champs utiliés par le formulaire
   * @param  {any} oFormDefinition
   * @param  {any} sFormDefinitionName
   * @returns string
   */
  public getFormInputsSize(oFormDefinition: any, sFormDefinitionName: any): string {

    if (this.vitisCheck.isDefAndNotNull(oFormDefinition)) {
      if (this.vitisCheck.isDefAndNotNull(sFormDefinitionName)) {
        if (this.vitisCheck.isDefAndNotNull(oFormDefinition[sFormDefinitionName])) {
          if (this.vitisCheck.isDefAndNotNull(oFormDefinition[sFormDefinitionName].nb_cols)) {
            return oFormDefinition[sFormDefinitionName].input_size;
          }
        }
      }
    }

    // Default
    return 'xs';
  }

  /**
   * Ajoute un filtre au datasource
   * Si bAutoRemove vaut true alors supprime tous les filtres
   * déjà exsistants utilisant la même colonne que oFilter
   * @param {any} oDatasource
   * @param {any} oFilter
   * @param {boolean} bAutoRemove=false
   * @returns any nouveau datasource
   */
  public addDatasourceFilter(oDatasource: any, oFilter: any, bAutoRemove: boolean = false): any {

    // Crée les paramètres si ils n'existent pas
    if (!oDatasource.parameters) {
      oDatasource.parameters = {};
    }

    // Crée le filtre si il n'existe pas encore
    if (!oDatasource.parameters.filter) {
      oDatasource.parameters.filter = this.createDatasourceFilter();
    }

    // Supprime l'encien filtre potentiel
    if (bAutoRemove) {
      oDatasource.parameters.filter = this.removeDatasourceOperator(oDatasource.parameters.filter, oFilter.column);
    }


    // Ajoute le nouveau filtre
    oDatasource.parameters.filter = this.addDatasourceOperator(oDatasource.parameters.filter, oFilter);

    return oDatasource;
  }

  /**
   * Supprime un filtre au datasource
   * @param {any} oDatasource
   * @param {any} oFilter
   * @returns any nouveau datasource
   */
  public removeDatasourceFilter(oDatasource: any, oFilter: any): any {

    // Supprime l'encien filtre potentiel
    if (oDatasource.parameters &&
      oDatasource.parameters.filter) {
      oDatasource.parameters.filter = this.removeDatasourceOperator(oDatasource.parameters.filter, oFilter.column);
    }

    return oDatasource;
  }

  /**
   * Crée un filtre vide pour datasource
   * @returns any nouveau filtre
   */
  public createDatasourceFilter(): any {
    const oDatasourceFilter = {
      relation: 'AND',
      operators: []
    };
    return oDatasourceFilter;
  }

  /**
   * Ajoute un opérateur à un filtre datasource
   * @param  {any} oFilter
   * @param  {any} oOperator
   * @returns any nouveau filtre
   */
  public addDatasourceOperator(oFilter: any, oOperator: any): any {
    if (oFilter &&
      this.vitisCheck.isDefAndNotNull(oFilter.operators)) {
      oFilter.operators.push(oOperator);
    }
    return oFilter;
  }

  /**
   * Supprime le filtre donné les opérateurs mentionnant la colonne donnée
   * @param  {any} oFilter
   * @param  {string} sColumn
   * @returns any nouveau filtre
   */
  public removeDatasourceOperator(oFilter: any, sColumn: string): any {
    if (oFilter) {
      if (oFilter[sColumn]) {
        delete oFilter[sColumn];
      }
      if (this.vitisCheck.isArray(oFilter.operators)) {
        for (let i = oFilter.operators.length - 1; i >= 0; i--) {
          const operator = oFilter.operators[i];
          if (operator.column === sColumn) {
            oFilter.operators.splice(i, 1);
          }
          if (operator.operators &&
            operator.operators[0] &&
            operator.operators[0].column === sColumn) {
            oFilter.operators.splice(i, 1);
          }
        }
      }
    }
    return oFilter;
  }

  /**
   * Rend sous les champs du formulaire disabled
   * @param oFormDefinition
   * @param sFormDefinitionName
   */
  public makeFormDisabled(oFormDefinition: any, sFormDefinitionName: any) {
    const fields = this.getFormFields(oFormDefinition, sFormDefinitionName);
    for (let i = 0; i < fields.length; i++) {
      const field = fields[i];
      field.disabled = true;
    }
    return oFormDefinition;
  }

  /**
   * Test si la structure de tabs est bonne
   * @param oTabs
   */
  public isTabsValid(oTabs: Tabs): boolean {

    let bTabsStructOk = false;
    if (oTabs &&
      oTabs.list &&
      oTabs.list[0] &&
      oTabs.list[0].elements) {
      bTabsStructOk = true;
    }

    return bTabsStructOk;
  }

  /**
   * Envoi les fichier en avance si nécessaire
   * @param oValues
   * @returns
   */
  public manageLargesFiles(oValues: any): Promise<any> {
    return new Promise<any>((resolve, reject) => {
      let fileProcessed: number = 0;
      let fileTotal: number = 0;
      const fileStruct: any = {};
      // ajout d'une structure pour simplifier le travail de l'API
      for (const oFieldName in oValues) {
        if (oValues.hasOwnProperty(oFieldName)) {
          const val = oValues[oFieldName];
          // cherche les fichiers
          if (this.vitisCheck.isFile(val)) {
            // check si la taille implique un upload par transaction
            if (this.vitisFileRequest.needUploadbyTransaction(val)) {
              fileTotal++;
              this.vitisFileRequest.uploadFile(val).then((sTransactionId: string) => {
                if (!fileStruct[oFieldName]) {
                  fileStruct[oFieldName] = [];
                }
                fileStruct[oFieldName].push(sTransactionId);
                delete oValues[oFieldName];
              }).catch(() => {
                console.error('Cannot send the file');
              }).finally(() => {
                fileProcessed++;
                if (fileProcessed >= fileTotal) {
                  oValues['vitis_transaction_files'] = JSON.stringify(fileStruct);
                  resolve(oValues);
                }
              });
            }
          } else if (this.vitisCheck.isArrayOfFileNotempty(val)) {
            // gestion multi fichier
            let bNeeded: boolean = false;
            // si un fichier dans la liste en a besoin toute la liste passe en transactionnel
            for (const oFile of val) {
              if (this.vitisFileRequest.needUploadbyTransaction(oFile)) {
                bNeeded = true;
              }
            }

            if (bNeeded) {
              fileTotal++;
              this.vitisFileRequest.uploadFiles(val).then((aTransactionId: string[]) => {
                if (!fileStruct[oFieldName]) {
                  fileStruct[oFieldName] = [];
                }
                fileStruct[oFieldName] = fileStruct[oFieldName].concat(aTransactionId);
                delete oValues[oFieldName];
              }).catch(() => {
                console.error('Cannot send the files');
              }).finally(() => {
                fileProcessed++;
                if (fileProcessed >= fileTotal) {
                  oValues['vitis_transaction_files'] = JSON.stringify(fileStruct);
                  resolve(oValues);
                }
              });
            }
          }
        }
      }
      // s'il n'y a pas de fichier dans le formdata on skip
      if (fileTotal === 0) {
        resolve(oValues);
      }
    });
  }

  /**
   * Transforme l'objet oValues en FormData
   * @param {any} oValues
   * @param {boolean} isStringifyObject
   * @returns {FormData}
   */
  public transformToDataForm(oValues: any, isStringifyObject: boolean = false): FormData {
    const oReturnValues = new FormData();

    for (const oFieldName in oValues) {
      if (oValues.hasOwnProperty(oFieldName)) {
        const val = oValues[oFieldName];

        if (this.vitisCheck.isArray(val)) {
          if (val.length === 0) {
            oReturnValues.append(oFieldName + '[]', val);
          }
          for (let j = 0; j < val.length; j++) {
            if (val[j]) {
              if (this.vitisCheck.isDefAndNotNull(val[j].name)) {
                oReturnValues.append(oFieldName + '[]', val[j], val[j].name);
              } else {
                oReturnValues.append(oFieldName + '[]', val[j]);
              }
            }
          }
        } else if (isStringifyObject === true && this.vitisCheck.isObject(val)) {
          oReturnValues.append(oFieldName, JSON.stringify(val));
        } else {
          if (val !== null) { oReturnValues.append(oFieldName, val); }
        }
      }
    }

    return oReturnValues;
  }

  /**
   * Charge le javascript pour le formulaire
   * @param {any} oFormReader
   * @param {string} sJsForm
   * @param {FormJS} oFormJS
   */
  public loadJs(oFormReader: any, sJsForm: string, oFormJS?: FormJS) {
    try {
      // Supprime les js précédement chargé
      const oldDiv = document.getElementsByClassName('formreader-js');
      for (let i = 0; i < oldDiv.length; i++) {
        oldDiv[i].remove();
      }

      // Injecte le JS
      if (sJsForm) {
        // Ajoute le nouveau javascript
        const newDiv = document.createElement('div');
        newDiv.setAttribute('class', 'formreader-js');

        const oScript = document.createElement('script');
        oScript.setAttribute('type', 'text/javascript');

        const js = document.createTextNode(sJsForm);
        oScript.appendChild(js);
        newDiv.appendChild(oScript);

        document.getElementsByTagName('body')[0].appendChild(newDiv);
      }

      let oScope = oFormReader;
      if (oFormJS && oFormJS.oAdditionalScope) {
        oScope = { oFormReader, ...oFormJS.oAdditionalScope };
      }

      let services = {
        vitisRequest: this.vitisRequest,
        vitisCheck: this.vitisCheck,
        vitisAlert: vitisAlert,
        vitisClone: this.vitisClone,
        translate: this.translate,
        auth: this.auth,
        formreaderSrvc: this,
        route: this.route,
        router: this.router,
        interfaceService: this.interfaceService,
        sessionStorageService: this.sessionStorageService,
        localStorageService: this.localStorageService,
        oFormReader: oFormReader,
      };
      if (oFormJS && oFormJS.oAdditionalServices) {
        services = { ...services, ...oFormJS.oAdditionalServices };
      }

      // Execute le js
      if ((<any>window).constructor_form) {
        (<any>window).constructor_form(oScope, environment, services);
      }

    } catch (error) {
      console.error('could not load js :', error);
    }
  }

  /**
   * Update la valeur de olMap à chaque changement de celle-ci
   */
  public emitMapEvent(olMap: any) {
    this.mapChangeEvent.next(olMap);
  }

  /**
   * Permet de rajouter des champs dans le formulaire dynamiquement
   * @returns
   */
  public createDynamicFormControl(default_value?: string, options?: any): FormControl {
    const dynamicFormControl = new FormControl(default_value ?? '');
    if (options?.validators.length > 0) {
      const validators = [];
      if (options.validators.includes('required')) {
        validators.push(Validators.required);
      }
      dynamicFormControl.setValidators(validators);
    }
    return dynamicFormControl;
  }
}
