
import { Component, OnInit, ViewChild } from '@angular/core';
import { UntypedFormBuilder, UntypedFormGroup, Validators } from '@angular/forms';
import { Router } from '@angular/router';
import { environment } from '../../../environments/environment';
import { AuthService } from '../core/services/auth/auth.service';
import { VitisRequest, VitisCheck, vitisAlert, vitisStringSimplifier, SessionStorage } from 'src/app/vitis/core/services/utils/vitis-utils';
import { LocalStorageService } from 'src/app/vitis/core/services/utils/localstorage/local-storage.service';
import { FormreaderComponent } from '../formreader/formreader.component';
import { FormreaderService } from '../formreader/services/formreader.service';
import { TranslateService } from '@ngx-translate/core';
import { LoginConfig } from '../app.model';
import { IsSmallScreenService } from '../core/services/utils/is-small-screen/is-small-screen.service';

declare let $: any;

@Component({
  selector: 'app-login',
  templateUrl: './login.component.html',
  styleUrls: ['./login.component.scss']
})
export class LoginComponent implements OnInit {

  public form: UntypedFormGroup;
  public appDescription: any;
  public rememberMeEnabled: boolean = true;
  public rememberMeDefault: boolean = true;
  public passwordForgottenEnabled: boolean = false;
  public signUpEnabled: boolean = false;
  public domains: any = [];
  public suffix: String;

  public sModalTitle: string = '';
  public initFormValues: any = null;

  private config: LoginConfig = {
    signup_form: 'assets/forms/signup/signup.json',
    signup_route: 'vitis/accounts/signup',
  };

  public isSmallScreen;

  @ViewChild(FormreaderComponent) oFormReader: FormreaderComponent;

  /**
   * Constructeur
   */
  constructor(
    private fb: UntypedFormBuilder,
    private authService: AuthService,
    private vitisRequest: VitisRequest,
    private vitisCheck: VitisCheck,
    private localStorageService: LocalStorageService,
    private router: Router,
    private formReaderService: FormreaderService,
    private translate: TranslateService,
    private sessionStorage: SessionStorage,
    private isSmallScreenService: IsSmallScreenService,
  ) {
    // Initialise la taille de l'écran actuel
    this.sessionStorage.set('smallScreen', this.isSmallScreenService.getIsSmallScreen());
  }

  /**
   * Chargement du composant
   */
  ngOnInit() {

    if (environment.appStruct.login) {
      this.config = {
        ...this.config,
        ...environment.appStruct.login
      };
    }

    this.isSmallScreen = this.isSmallScreenService.getIsSmallScreen();
    this.isSmallScreenService.isSmallScreenObservable.subscribe(isSmallScreen => {
      this.isSmallScreen = isSmallScreen;
    });

    // Liste des domaines
    this.vitisRequest.ajaxRequestPromise({
      method: 'GET',
      url: environment.properties.app_api_url + '/vitis/domains/suffix'
    }).then((resp: any) => {
      if (resp.data && resp.data.status === 1) {
        this.domains = resp.data.data;
      }
    });

    if (environment.properties) {

      // Description de l'application
      if (environment.properties.app_description) {
        this.appDescription = environment.properties.app_description;
      }

      // Affichage de "Se souvenir de moi" d'apprès les properties
      if (this.vitisCheck.isDefAndNotNull(environment.properties.login_remember_me)) {
        this.rememberMeEnabled = environment.properties.login_remember_me;
      }

      // Valeur de "Se souvenir de moi" d'apprès les properties
      if (this.vitisCheck.isDefAndNotNull(environment.properties.login_remember_me_default)) {
        this.rememberMeDefault = environment.properties.login_remember_me_default;
      }

      // Valeur de "Se souvenir de moi" d'apprès le localstorage
      if (this.vitisCheck.isDefAndNotNull(this.localStorageService.get('login_remember_me'))) {
        this.rememberMeDefault = this.localStorageService.get('login_remember_me');
      }

      // Affichage de "Mot de passe oublié" d'apprès les properties
      if (this.vitisCheck.isDefAndNotNull(environment.properties.password_forgotten)) {
        this.passwordForgottenEnabled = environment.properties.password_forgotten;
      }

      // Affichage de "Création de compte" d'apprès les properties
      if (this.vitisCheck.isDefAndNotNull(environment.properties.sign_up)) {
        this.signUpEnabled = environment.properties.sign_up;
      }
    }

    // formulaire
    this.form = this.fb.group({
      user: ['', Validators.required],
      password: ['', Validators.required],
      remember: ['']
    });

    // Valeurs par défaut
    this.form.patchValue({
      remember: this.rememberMeDefault
    });
  }

  /**
   * Login function
   * @returns void
   */
  public login(): void {
    const val = this.form.value;

    // Mémorise la valeur de remember me
    this.localStorageService.set('login_remember_me', val.remember);

    if (val.user && val.password) {
      let login = val.user;
      if (this.suffix != null && this.suffix.length > 0) {
        login = val.user + this.suffix;
      }
      this.authService.login(login, val.password, val.remember)
        .subscribe(
          (result) => {
            if (result === true) {
              if (this.authService.getAfterLoginRoute()) {
                this.router.navigateByUrl(this.authService.getAfterLoginRoute());
              } else {
                this.router.navigate(['**'], { queryParamsHandling: 'merge' });
              }
            } else {
              // Message d'erreur à afficher.
              let sErrorTranslationId = 'ERROR_LOGIN';
              if (this.vitisCheck.isDefAndNotNull(result.data) &&
                this.vitisCheck.isDefAndNotNull(result.data.errorMessage)) {
                sErrorTranslationId = result.data.errorMessage;
              }
              // Affichage du message d'erreur.
              vitisAlert.alert({
                text: this.translate.instant(sErrorTranslationId),
                type: 'error'
              });
            }
          }
        );
    }
  }

  /**
   * Lance la modal avec le formulaire forgotten_password_email ou forgotten_password_login
   * @param sForm sForm email|login
   */
  public passwordForgotten(sForm: string) {
    const oFields = [];
    if (sForm === 'email') {
      this.sModalTitle = 'FORM_LOGIN_FORGET_PASSWORD';
      oFields.push({
        fields: [{
          type: 'email',
          name: 'fpwd_email',
          label: 'SIGNUPMAIL',
          nb_cols: 12,
          required: true,
          id: 'domain_1_1',
          pattern: '^(([a-zA-Z]|[0-9])|([-]|[_]|[.]))+[@](([a-zA-Z0-9])|([-])){2,63}[.](([a-zA-Z0-9]){2,63})+$'
        }]
      });
    }

    if (sForm === 'login') {
      this.sModalTitle = 'FORM_LOGIN_FORGET_PASSWORD';
      oFields.push({
        fields: [{
          type: 'subtitle',
          name: 'email_several_exist',
          label: 'SEVERAL_MAIL_FIND',
          nb_cols: 12,
        }, {
          type: 'text',
          name: 'fpwd_login',
          id: 'login_form_user_login',
          label: 'FORM_LOGIN_USER',
          autofocus: true,
          required: true,
          nb_cols: 12
        }]
      });
    }

    oFields.push({
      fields: [{
        type: 'button',
        class: 'btn-ungroup btn-group-sm',
        nb_cols: 12,
        buttons: [{
          type: 'submit',
          name: 'btn_send',
          label: 'FORM_FORGOTTEN_PASSWORD_SEND',
          class: 'btn-primary btn-block'
        }]
      }]
    });

    const oForm: any = {
      update: {
        name: 'vitis_forgotten_password_update_form',
        eventContainer: this,
        event: 'sendFormForgottenPassword()',
        input_size: 'xs',
        nb_cols: 12,
        rows: oFields
      }
    };

    // Lance le formulaire dans FormReader
    this.oFormReader.initForm(oForm, 'update', this.initFormValues);

    $('#login-formreader-modal').modal('show');
  }

  /**
   * Envoi le formulaire pour demander changement de mot de passe
   */
  public sendFormForgottenPassword() {
    if (this.oFormReader.isFormValid()) {
      const values = this.oFormReader.getFormValues();

      // Récupération des données du form
      const formData = this.formReaderService.transformToDataForm(values);

      this.vitisRequest.ajaxRequest({
        method: 'POST',
        url: environment.properties.app_api_url + '/vitis/accounts/fpwd',
        data: formData,
        displayErrorMessage: false,
        success: (resp: any) => {
          if (resp.data && resp.data.status == 1) {
            // Le mail a été envoyé
            this.translate.get('CHECKMAIL').subscribe((translate) => {
              vitisAlert.fireToast({
                type: 'success',
                title: translate
              });

              $('#login-formreader-modal').modal('hide');
            });
          }
        },
        error: (resp: any) => {
          if (resp.data) {
            if (resp.data.status == 0 && resp.data.errorMessage === 'SEVERAL_USER_FOUNDED') {
              this.passwordForgotten('login');
            } else {
              this.translate.get(resp.data.errorMessage).subscribe((translate) => {
                vitisAlert.alert({
                  type: 'error',
                  text: translate
                });
              });
            }
          } else {
            this.translate.get('ERROR_SERVER_REQUEST_500').subscribe((translate) => {
              vitisAlert.alert({
                type: 'error',
                text: translate
              });
            });
          }
        }
      });
    }
  }
  /**
   * charge le formulaire d'inscription et son js si nécessaire
   */
  public signUp(): void {

    // charge le js s'il est présent
    if (this.config.signup_js) {
      this.vitisRequest.loadExternalJs(this.config.signup_js).then(() => {
        // Lance le JS associé
        if ((<any>window).constructor_form) {
          (<any>window).constructor_form(this, environment, {
            vitisRequest: this.vitisRequest,
            vitisCheck: this.vitisCheck,
            vitisAlert: vitisAlert,
            translate: this.translate,
            vitisStringSimplifier: vitisStringSimplifier,
          });
        }

        this.loadSignUpForm();
      });
    } else {
      this.loadSignUpForm();
    }
  }

  /**
   * charge le formulaire d'inscription
   */
  private loadSignUpForm(): void {
    this.sModalTitle = 'MODAL_SIGN_UP_TITLE';

    // Chargement et initialisation du formulaire des stats.
    this.loadForm().then((oForm: any) => {

      // Lance le formulaire dans FormReader
      this.oFormReader.initForm(oForm, 'update', this.initFormValues);

      // charge le js s'il est présent
      if (this.config.signup_js) {
        this.vitisRequest.loadExternalJs(this.config.signup_js).then(() => {
          // Lance le JS associé
          if ((<any>window).constructor_form) {
            (<any>window).constructor_form(this, environment, {
              vitisRequest: this.vitisRequest,
              vitisCheck: this.vitisCheck,
              vitisAlert: vitisAlert,
              translate: this.translate,
              vitisStringSimplifier: vitisStringSimplifier,
            });
          }
        });
      }

      $('#login-formreader-modal').modal('show');
    });
  }

  /**
   * Envoi le formulaire pour demander créer un compte
   */
  public sendFormSignUp() {
    if (this.oFormReader.isFormValid()) {
      const values: any = this.oFormReader.getFormValues();

      if (values.password !== values.password_confirm) {
        console.error('password_confirm not valid');
        vitisAlert.alert({
          type: 'error',
          text: this.translate.instant('CONFIRMPASSWORD'),
        });
        return;
      }

      // Récupération des données du form
      const formData = this.formReaderService.transformToDataForm(values);

      this.sendSignUpRequest(formData);
    }
  }

  /**
   * envoi de la requête avec un formdata permet en cas de js spécifique ça permet de juste gérer le formdata
   * @param formData
   */
  public sendSignUpRequest(formData: FormData) {
    this.vitisRequest.ajaxRequestPromise({
      method: 'POST',
      url: environment.properties.app_api_url + '/' + this.config.signup_route,
      data: formData
    }).then((resp) => {
      if (resp.data) {
        if (resp.data.status === 1) {
          this.translate.get('CHECKMAIL').subscribe(($translate) => {
            // Ferme la popup.
            $('#login-formreader-modal').modal('hide');
            //
            vitisAlert.fireToast({
              type: 'success',
              text: $translate
            });
          });
        }
      }
    }).catch((e) => {
      // On vide le captcha
      const captchaField = this.oFormReader.getFormField('captcha');
      captchaField.setValue('');
    });
  }

  /**
   * Charge le formulaire json d'inscription
   */
  private loadForm() {
    return new Promise((resolve, reject) => {
      // Téléchargement du fichier.
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: this.config.signup_form
      }).then((resp: any) => {
        if (resp.status === 200 && resp.data) {
          const oForm = resp.data;
          if (this.vitisCheck.isDefAndNotNull(oForm) && this.vitisCheck.isDefAndNotNull(oForm.update)) {
            oForm.update.eventContainer = this;
            resolve(oForm);
          } else {
            console.error('Could not get signup form from request');
            reject();
          }
        } else {
          reject();
        }
      }).catch((err) => {
        reject();
      });
    });
  }
}
