import { Component, Input, OnInit, ViewContainerRef, ViewChild, AfterViewInit } from '@angular/core';
import { VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { Map } from 'ol';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import { LocalStorageService } from 'src/app/vitis/core/services/utils/localstorage/local-storage.service';
declare let $: any;

@Component({
  selector: 'app-core-controls-map-comment',
  templateUrl: './core-controls-map-comment.component.html',
  styleUrls: ['./core-controls-map-comment.component.scss']
})
export class CoreControlsMapCommentComponent implements OnInit, AfterViewInit {

  @Input() olMap: Map;

  // Paramètres du commentaire.
  public oMapComment: any;
  // Bouton "Ne plus afficher".
  public bCommentNoDisplay: boolean;
  // Infos de l'utilisateur.
  public oUserInfos: any;

  @ViewChild('container', { read: ViewContainerRef }) container: ViewContainerRef;

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private authService: AuthService,
    private localStorageService: LocalStorageService,
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit(): void {
    // Mise à jour du commentaire de la carte.
    this.oMapComment = this.refreshMapComment();
    this.oUserInfos = this.authService.getUserInfos();
    // Changement de carte.
    this.olMap.on('change', () => {
      // Mise à jour du commentaire de la carte.
      this.oMapComment = this.refreshMapComment();
      // Affichage de la fenêtre de commentaire de la carte.
      if (this.vitisCheck.isDefAndNotNull(this.oMapComment) &&
        this.hasToDisplayComment()) {
        this.displayModal();
      }
    });
  }

  /**
   * Appelé après le chargement du DOM
   */
  ngAfterViewInit(): void {
    // Affichage de la fenêtre de commentaire de la carte.
    if (this.vitisCheck.isDefAndNotNull(this.oMapComment) &&
      this.hasToDisplayComment()) {
      this.displayModal();
    }
  }

  /**
   * Retourne la clé posée sur le sessionStorage concernant le choix de l'utilisateur s'il a cliqué sur "Ne plus afficher"
   * @param map_id
   * @returns
   */
  private getSessionKey(map_id: number | string): string {
    return `vmap_map_comment_not_read_anymore_${map_id}`;
  }

  /**
   * Retourne false si l'utilisateur a cliqué sur "Ne plus afficher" et que le commentaire n'a pas été modifié depuis
   * @returns
   */
  private hasToDisplayComment(): boolean {
    if (this.oMapComment.display === false) {
      return false;
    }

    // Retrouve le choix de l'utilisateur s'il a cliqué sur "Ne plus afficher"
    const memo = this.localStorageService.get(this.getSessionKey(this.olMap.get('map_id')));
    if (memo && memo > this.oMapComment.last_modification * 1000) {
      return false;
    }

    return true;
  }

  /**
   * Retourne le comentaire de la carte.
   * @returns {any}
   */
  private refreshMapComment(): any {
    const oMapDef = this.olMap.get('map_definition');
    if (this.vitisCheck.isDefAndNotNull(oMapDef.comment)) {
      this.bCommentNoDisplay = !oMapDef.comment.display;
      return oMapDef.comment;
    }
  }

  /**
   * Affichage de la fenêtre modale du commentaire de la carte.
   * @param {number} iTimeout
   * @returns {void}
   */
  private displayModal(iTimeout: number = 1000): void {
    // Affichage si commentaire récent ou autorisé par l'utilisateur.
    setTimeout(() => {
      if (this.vitisCheck.isDefAndNotNull(this.container)) {
        // Contenu du commentaire dans la fenêtre.
        this.container.element.nativeElement.innerHTML = this.oMapComment.text;
        // Affichage de la fenêtre modale.
        $('#map_comment_modal').modal('show');
      }
    }, iTimeout);
  }

  /**
   * Evènement de fermeture de la fenêtre de commentaire.
   * @returns {void}
   */
  public onModalClose(): void {
    if (this.bCommentNoDisplay === true) {
      this.localStorageService.set(this.getSessionKey(this.olMap.get('map_id')), Date.now());
    }
  }
}
