import { Component, Input, OnInit } from '@angular/core';
import Map from 'ol/Map';
import {
  GPX,
  GeoJSON,
  IGC,
  KML,
  TopoJSON
} from 'ol/format';
import { MapCoreManagerService } from '../../services/map-core-manager/map-core-manager.service';
import { TranslateService } from '@ngx-translate/core';
import { vitisAlert } from 'src/app/vitis/core/services/utils/vitis-utils';
import shp from 'shpjs';
import Feature from 'ol/Feature';
import { Geometry } from 'ol/geom';

@Component({
  selector: 'app-core-draganddrop',
  templateUrl: './core-draganddrop.component.html',
  styleUrls: ['./core-draganddrop.component.scss']
})
export class CoreDraganddropComponent implements OnInit {

  @Input() olMap: Map;

  public dropinActive = false;

  /**
   * Constructeur
   */
  constructor(
    private mapCoreManagerService: MapCoreManagerService,
    private translate: TranslateService,
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit(): void {
    this.setInteraction();
  }

  /**
   * Ajoute l'interaction Drag & drop à la carte
   */
  private setInteraction() {

    const target = this.olMap.getTargetElement().parentElement;
    target.ondragenter = (e) => {
      this.dropinActive = true;
      return false;
    };
    target.ondragleave = (e: any) => {
      if (!target.contains(e.relatedTarget)) {
        this.dropinActive = false;
      }
      return false;
    };
    target.ondragover = (e) => {
      e.stopPropagation();
      e.preventDefault();
      e.dataTransfer.dropEffect = 'copy';
      return false;
    };
    target.ondrop = (e) => {
      e.stopPropagation();
      e.preventDefault();
      this.dropinActive = false;
      if (e.dataTransfer &&
        e.dataTransfer.files &&
        e.dataTransfer.files[0]) {
        this.addVectorFile(e.dataTransfer.files[0]);
      }
      return false;
    };
  }

  /**
   * Lit un fichier et crée la couche vectorielle correspondante
   * @param oFile
   * @returns
   */
  private addVectorFile(oFile: File) {
    const reader = new FileReader();
    const availableExtentions = ['shp', 'gpx', 'json', 'geojson', 'kml', 'topojson'];
    const posLastDot = oFile.name.lastIndexOf('.');
    const extention = oFile.name.substring(posLastDot + 1, oFile.name.length);

    if (!availableExtentions.includes(extention)) {
      return this.translate.get('COMPONENT_MAP_VECTOR_FILE_FORMAT_ERROR').subscribe((sTranslate: string) => {
        vitisAlert.fireToast({
          type: 'warning',
          title: sTranslate
        });
      });
    }

    if (extention === 'shp') { // Format shapefile

      vitisAlert.alert({
        title: this.translate.instant('COMPONENT_MAP_VECTOR_FILE_PROJECTION'),
        input: 'select',
        inputOptions: {
          'EPSG:2154': '[EPSG:2154]-RGF93/Lambert-93',
          'EPSG:4326': '[EPSG:4326]-WGS84/Long-Lat',
          'EPSG:3857': '[EPSG:3857]-WGS84/Spherical Mercator',
        },
        showCancelButton: true,
      }).then((result) => {
        if (result && result.value) {
          reader.readAsArrayBuffer(oFile);
          reader.onload = () => {
            const sFeatures = shp.parseShp(reader.result);
            const geojsonFeatures = sFeatures.map(sFeature => ({
              geometry: sFeature,
              type: 'Feature'
            }));
            const format = new GeoJSON();
            try {
              const features = format.readFeatures({
                type: 'FeatureCollection',
                features: geojsonFeatures,
              }, {
                dataProjection: result.value,
                featureProjection: this.olMap.getView().getProjection().getCode(),
              });
              if (features && features.length > 0) {
                this.addFeaturesOnMap(format, features, oFile.name);
              }
            } catch (err) {
            }
          };
        }
      });
    } else { // autre formats
      reader.readAsText(oFile);
      const this_ = this;
      reader.onloadend = (evt: any) => {
        if (evt.target.readyState === FileReader.DONE) {
          let sFeatures;
          const formatConstructors = [GPX, GeoJSON, KML, TopoJSON];
          try {
            sFeatures = JSON.parse(evt.target.result);
          } catch (e) {
            sFeatures = evt.target.result;
          }

          // Essaye de lire la source selon les formats suivants
          for (let i = 0; i < formatConstructors.length; i++) {
            try {
              const format = new formatConstructors[i]();
              const features = format.readFeatures(sFeatures, {
                featureProjection: this.olMap.getView().getProjection().getCode(),
              });
              if (features && features.length > 0) {
                this_.addFeaturesOnMap(format, features, oFile.name);
              }
            } catch (err) {

            }
          }
        }
      };
    }
  }

  /**
   * Ajoute la couche vectorielle à la carte
   * @param format
   * @param features
   * @param layerName
   */
  public addFeaturesOnMap(format: GPX | GeoJSON | IGC | KML | TopoJSON, features: Array<Feature<Geometry>>, layerName: string) {
    this.mapCoreManagerService.addLayerVector(this.olMap, features, layerName, true);
  }
}
