import { Component, Input, OnInit, Output, EventEmitter, ElementRef, Inject, OnDestroy } from '@angular/core';
import { DOCUMENT } from '@angular/common';
import { MapJSON } from 'src/app/vitis/map-core/services/map-models.model';
import { Map } from 'ol';
import { Vector as VectorLayer } from 'ol/layer.js';
import Point from 'ol/geom/Point.js';
import Feature from 'ol/Feature';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { MapCoreDrawService } from 'src/app/vitis/map-core/services/map-core-draw/map-core-draw.service';
import { MapCoreUtilsService } from 'src/app/vitis/map-core/services/map-core-utils/map-core-utils.service';
import { MapCoreManagerService } from '../../services/map-core-manager/map-core-manager.service';
import VectorSource from 'ol/source/Vector';
import { Geometry } from 'ol/geom';
import { Subscription } from 'rxjs';

@Component({
  selector: 'app-core-interactions',
  templateUrl: './core-interactions.component.html',
  styleUrls: ['./core-interactions.component.scss']
})
export class CoreInteractionsComponent implements OnInit, OnDestroy {

  @Input() olMap: Map;
  @Output() public setValueEmitter = new EventEmitter();

  // Interactions autorisées dans le MapJSON
  public enabledInteractions: any = {
    fullScreen: false,
    removeAll: false,
    remove: false,
    modify: false,
    point: false,
    line: false,
    polygon: false,
    multipolygon: false,
    select: false,
  };

  // Interactions actives
  public activeInteractions: any = {
    fullScreen: false,
    removeAll: false,
    remove: false,
    modify: false,
    point: false,
    line: false,
    polygon: false,
    multipolygon: false,
    select: false,
  };

  // Interaction en cours
  public currentInteraction: any;

  // Dessin en cours
  public sEWKTDraw: string;

  // Géométries multiples
  public multiGeom: boolean = false;

  // Définition de la carte
  public mapJSON: MapJSON;

  // Couche de dessin
  public olDrawLayer: VectorLayer<VectorSource<Geometry>>;

  // Conteneur de la carte OL3
  public mapContainer: any;

  // Variables privées
  private ondrawendEvt_: Function;
  private sLastType_: string;

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private mapCoreDrawService: MapCoreDrawService,
    private mapCoreUtilsService: MapCoreUtilsService,
    private mapCoreManagerService: MapCoreManagerService,
    private el: ElementRef,
    @Inject(DOCUMENT) private document
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit(): void {
    this.mapContainer = this.el.nativeElement.parentNode;

    this.mapJSON = this.olMap.get('map_json');

    this.olDrawLayer = this.olMap.get('draw_layer');

    if (this.vitisCheck.isDefAndNotNull(this.mapJSON)) {
      this.initMapInteractions_(this.mapJSON.interactions);
    }

    // Écoute les changements d'interractions
    this.subscriptions.push(this.mapCoreDrawService.getActiveInteractionObs().subscribe(oActiveInteraction => {
      this.currentInteraction = oActiveInteraction;
    }));
  }

  /**
   * Fonction de destruction
   */
  ngOnDestroy(): void {
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
  }

  /**
   * Active/desactive une interraction de la carte
   * @param  {string} sType
   */
  public toggleInteraction(sType: string) {

    // État actuel
    let isActive = false;

    // Outils desktop
    if (this.vitisCheck.isDefAndNotNull(this.currentInteraction) && sType.toLowerCase() == this.sLastType_.toLowerCase()) {
      isActive = this.currentInteraction.getActive();
    }

    // Supprime les interactions
    this.removeInteractions_();

    if (!isActive) {
      // Ajoute l'interaction à la carte
      this.addInteraction_(sType);
    }
  }

  /**
   * Initialise les interractions de carte
   * @param  {any} oInteractions
   */
  private initMapInteractions_(oInteractions: any) {

    // Full Screen
    if (oInteractions.full_screen) {
      // Si il est possible d'afficher en fullScreen
      if (
        this.document.fullscreenEnabled ||
        this.document.webkitFullscreenEnabled ||
        this.document.mozFullScreenEnabled ||
        this.document.msFullscreenEnabled
      ) {
        this.enabledInteractions.fullScreen = true;
      }
    }
    // Remove all
    if (oInteractions.remove_all) {
      this.enabledInteractions.removeAll = true;
    }
    // Remove
    if (oInteractions.remove) {
      this.enabledInteractions.remove = true;
    }
    // Modify
    if (oInteractions.modify) {
      this.enabledInteractions.modify = true;
    }
    // Point
    if (oInteractions.draw_point) {
      this.enabledInteractions.point = true;
    }
    // Line
    if (oInteractions.draw_line) {
      this.enabledInteractions.line = true;
    }
    // Polygon
    if (oInteractions.draw_polygon) {
      this.enabledInteractions.polygon = true;
    }
    // Multi Polygon
    if (oInteractions.draw_multipolygon) {
      this.enabledInteractions.multipolygon = true;
    }
    // select
    if (oInteractions.select) {
      this.enabledInteractions.select = true;

      // Sélection auto après un dessin
      if (oInteractions.autoselect) {
        this.subscriptions.push(this.mapCoreDrawService.getActiveInteractionObs().subscribe(oActiveInteraction => {
          if (oActiveInteraction && oActiveInteraction.get('vitis_draw') === true) {
            oActiveInteraction.on('change:geometry', (evt: any) => {
              setTimeout(() => {
                const olFeature = evt.feature;
                this.mapCoreDrawService.selectFeature(olFeature, this.olDrawLayer);
              });
            });
          }
        }));
      }
    }

    // Edition multiple
    if (oInteractions.multi_geometry) {
      this.multiGeom = true;
    }

    // Interaction par défaut
    if (oInteractions.active_interaction) {
      this.addInteraction_(oInteractions.active_interaction);
    }

    // Interaction par défaut
    if (this.vitisCheck.isFunction(oInteractions.ondrawend)) {
      this.ondrawendEvt_ = oInteractions.ondrawend;
    }
  }

  /**
   * Ajoute une interraction
   * @param  {string} sType
   */
  private addInteraction_(sType: string) {
    this.sLastType_ = sType;
    this.addDesktopInteraction_(sType);
  }

  /**
   * Supprime les interractions précédament ajoutées
   */
  private removeInteractions_() {

    // Enlève les interactions de la carte
    this.mapCoreDrawService.removeInteractions(this.olMap);

    // Enlève l'interaction en cours de ce composant
    delete this.currentInteraction;

    // Spécifie que plus aucune interaction n'est active
    for (const key in this.activeInteractions) {
      if (this.activeInteractions.hasOwnProperty(key)) {
        this.activeInteractions[key] = false;
      }
    }
  }

  /**
  * Ajoute une interraction (pour desktop)
  * @param  {string} sType
  */
  private addDesktopInteraction_(sType: string) {

    if (sType != 'fullScreen') {
      this.mapCoreDrawService.addDrawInteraction(sType, this.olMap, this.olDrawLayer, {
        multiple: this.multiGeom
      });
    }

    // Ajoute l'interaciton à la carte
    switch (sType) {
      case 'fullScreen':
        // Si on est déjà en mode plein écran
        if (
          this.document.fullscreenElement ||
          this.document.webkitFullscreenElement ||
          this.document.mozFullScreenElement ||
          this.document.msFullscreenElement
        ) {
          if (this.document.exitFullscreen) {
            this.document.exitFullscreen();
          } else if (this.document.webkitExitFullscreen) {
            this.document.webkitExitFullscreen();
          } else if (this.document.mozCancelFullScreen) {
            this.document.mozCancelFullScreen();
          } else if (this.document.msExitFullscreen) {
            this.document.msExitFullscreen();
          }
        } else {
          const elem = this.mapContainer;
          if (elem.requestFullscreen) {
            elem.requestFullscreen();
          } else if (elem.mozRequestFullScreen) {
            elem.mozRequestFullScreen();
          } else if (elem.webkitRequestFullscreen) {
            elem.webkitRequestFullscreen();
          }
        }
        this.activeInteractions.fullScreen = true;
        break;
      case 'removeAll':
        this.activeInteractions.removeAll = true;
        break;
      case 'remove':
        this.activeInteractions.remove = true;
        break;
      case 'modify':
        this.activeInteractions.modify = true;
        break;
      case 'point':
        this.activeInteractions.point = true;
        break;
      case 'linestring':
        this.activeInteractions.line = true;
        break;
      case 'polygon':
        this.activeInteractions.polygon = true;
        break;
      case 'multipolygon':
        this.activeInteractions.multipolygon = true;
        break;
      case 'select':
        this.activeInteractions.select = true;
        break;
      default:
        console.error('Interaction type not valid', sType);
    }

    // Callback après avoir dessiné
    if (this.vitisCheck.isDefAndNotNull(this.currentInteraction)) {
      this.currentInteraction.on('drawend', (evt: any) => {

        // Dessin
        const olFeature = evt.feature;
        this.setDrawFeature_(olFeature);
      });
    }

    this.olDrawLayer.on('change', () => {
      const olFeatures = this.olDrawLayer.getSource().getFeatures();
      this.setValueEmitter.emit(this.mapCoreUtilsService.getEWKTFromFeatures(olFeatures, this.olMap));
    });
  }

  /**
   * Set la feature de dessin
   * @param  {Feature} olFeature
   */
  private setDrawFeature_(olFeature: Feature<Geometry>) {

    // Arrete l'interraction après avoir dessiné
    this.removeInteractions_();

    // Dessin resultant
    delete this.sEWKTDraw;
    if (this.vitisCheck.isDefAndNotNull(olFeature)) {
      if (this.vitisCheck.isDefAndNotNull(olFeature.getGeometry())) {
        this.sEWKTDraw = this.mapCoreUtilsService.getEWKTFromGeom(olFeature.getGeometry(), this.olMap);
      }
    }

    if (this.vitisCheck.isFunction(this.ondrawendEvt_)) {
      this.ondrawendEvt_.call(this, olFeature.getGeometry());
    }
  }

}
