import { Component, Input, OnInit, Output, EventEmitter, Inject, Optional, ChangeDetectorRef, SimpleChanges, OnChanges, OnDestroy } from '@angular/core';
import Map from 'ol/Map';
import { Vector as VectorLayer } from 'ol/layer.js';
import { transform, transformExtent, get as getProjection } from 'ol/proj.js';
import * as olExtent from 'ol/extent';
import Feature from 'ol/Feature';
import { DashboardComponent } from '../core/components/dashboard/dashboard.component';
import { VitisCheck } from '../core/services/utils/vitis-utils';
import { MapJSON } from 'src/app/vitis/map-core/services/map-models.model';
import { MapCoreManagerService } from './services/map-core-manager/map-core-manager.service';
import { MapCoreFormreaderService } from './services/map-core-fromreader/map-core-formreader.service';
import { MapCoreStudioService } from './services/map-core-studio/map-core-studio.service';
import { MapCoreDrawService } from './services/map-core-draw/map-core-draw.service';
import { MapCoreUtilsService } from './services/map-core-utils/map-core-utils.service';
import { MapCorePopupOverlayService } from './services/map-core-popup-overlay/map-core-popup-overlay.service';
import { FormreaderService } from '../formreader/services/formreader.service';
import { Geometry } from 'ol/geom';
import VectorSource from 'ol/source/Vector';
import { Observable, Subscription } from 'rxjs';

@Component({
  selector: 'app-map-core',
  templateUrl: './map-core.component.html',
  styleUrls: ['./map-core.component.scss']
})
export class MapCoreComponent implements OnInit, OnChanges, OnDestroy {

  @Input() public mapJSON: MapJSON; // Définition pour utilisation FormReader
  @Input() public style: any;
  @Input() public templateType: string;
  @Input() public idStudio: string;
  @Input() public formreaderId: string;
  @Input() public sGeom: string;
  @Input() public initMapObs: Observable<Map>;

  @Output() public setValueEmitter = new EventEmitter();
  @Output() public setFeatureEmitter = new EventEmitter();
  @Output() public selectFeatureEmitter = new EventEmitter<any>();
  @Output() public changeExtentEmitter = new EventEmitter<any>();

  public targetId: string; // Id de la div portant la carte
  public targetCompareId: string; // Id de la div portant la carte
  public olMap: Map; // Composant Map OpenLayers

  public refreshMapOptionSubscription: any; // Utilisé lorqu'il y a une modification depuis le studio
  public refreshMapFormreaderSubscription: any; // Utilisé lorqu'il y a une modification depuis FormReader

  public haveControls: Boolean = false; // Contrôles de la carte
  public haveInteractions: Boolean = false; // Panneau d'outils de dessin sur la droite
  public haveDragNDrop: Boolean = false; // Drag & drop de fichiers vectoriels

  public tileSize: Array<number> = [512, 512];

  // Mémorise toutes les souscriptions de manière à les desouscrire à la destruction du composant
  private subscriptions: Subscription[] = [];

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private formreaderService: FormreaderService,
    private mapCoreManagerService: MapCoreManagerService,
    private mapCoreFormreaderService: MapCoreFormreaderService,
    private mapCoreStudioService: MapCoreStudioService,
    private changeDetectorRef: ChangeDetectorRef,
    private mapCoreDrawService: MapCoreDrawService,
    private mapCoreUtilsService: MapCoreUtilsService,
    private mapCorePopupOverlayService: MapCorePopupOverlayService,
    @Optional() @Inject(DashboardComponent) private dashboardComponent: DashboardComponent,
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit(): void {

    // Définit la valeur par défaut de targetId
    this.targetId = this.mapCoreManagerService.getUniqueId();
    this.targetCompareId = this.mapCoreManagerService.getUniqueId();

    // Rétro compatibilité
    if (this.vitisCheck.isDefAndNotNull(this.mapJSON)) {
      this.checkRetrocompatibility(this.mapJSON);
    }

    // Ajoute les projections
    this.mapCoreManagerService.addCustomProjections();

    // Écoute les changements de sidenavOpen et resize la carte pour éviter les applatissements.
    // -> Pas en mode "widget".
    if (this.vitisCheck.isDefAndNotNull(this.dashboardComponent)) {
      this.subscriptions.push(this.dashboardComponent.sidenavOpenChange.subscribe((sidenavOpen: any) => {
        if (this.olMap) {
          this.mapCoreManagerService.resizeMap(this.olMap);
        }
      }));
    }

    if (this.vitisCheck.isDefAndNotNull(this.mapJSON)) {
      // Utilisation FormReader avec MapJSON
      this.initFromJSON_();
      // Timeout le temps que la carte soit initialisée
      setTimeout(() => {
        this.haveDragNDrop = true;
      });
    }

    // Observe les features (vectorielles) sélectionnées et déclenche selectFeatureEmitter
    this.subscriptions.push(this.mapCoreDrawService.getSelectFeatureObs().subscribe((olSelectedFeature) => {
      this.selectFeatureEmitter.emit(olSelectedFeature);
    }));

    // Nécessaire pour fonctionnement asynchrone avec vMap
    if (this.initMapObs) {
      this.subscriptions.push(this.initMapObs.subscribe(olMap => {
        this.olMap = olMap;
        this.haveDragNDrop = true;
      }));
    }
  }

  /**
   * Appelé à la destruction
   */
  ngOnDestroy() {
    if (this.vitisCheck.isDefAndNotNull(this.refreshMapOptionSubscription)) {
      this.refreshMapOptionSubscription.unsubscribe();
    }

    if (this.vitisCheck.isDefAndNotNull(this.refreshMapFormreaderSubscription)) {
      this.refreshMapFormreaderSubscription.unsubscribe();
    }
    this.subscriptions.forEach(subscription => subscription.unsubscribe());
  }

  /**
   * Detecte les changements venant du setValue du composant map-formreader
   */
  ngOnChanges(changements: SimpleChanges) {
    const aFeatures = this.mapCoreFormreaderService.setValues(changements, this.olMap);
    if (this.vitisCheck.isDefAndNotNull(aFeatures) && this.vitisCheck.isArray(aFeatures)) {
      // Zoom sur les features
      this.zoomOnFeatures(aFeatures);

      // Transmet les features aux autres composants
      this.setFeatureEmitter.emit(aFeatures);
    }
  }

  /**
   * Rétrocompatibilité de la définition d'une carte FormReader.
   * @param {any} oMapJson Définition d'une carte FormReader
   */
  public checkRetrocompatibility(oMapJson: any) {
    // Anciens paramètres à remplacer.
    const oCompatibilityParams = {
      interactions: {
        RA: 'remove_all',
        RO: 'remove',
        ED: 'modify',
        DP: 'draw_point',
        DL: 'draw_line',
        DPol: 'draw_polygon',
        DMPol: 'draw_multipolygon'
      }
    };
    const aInteractionsKeys = Object.keys(oMapJson.interactions);
    let sInteraction: string;
    for (let i = 0; i < aInteractionsKeys.length; i++) {
      sInteraction = oCompatibilityParams.interactions[aInteractionsKeys[i]];
      if (this.vitisCheck.isDefAndNotNull(sInteraction)) {
        oMapJson.interactions[sInteraction] = oMapJson.interactions[aInteractionsKeys[i]];
        delete oMapJson.interactions[aInteractionsKeys[i]];
      }
    }
  }

  /**
   * Initialise les attributs Vitis de l'objet Map
   * @param olMap
   */
  public setMapAttributes(olMap: Map) {

    // Couche de dessin
    this.olMap.set('draw_layer', this.mapCoreDrawService.initDrawLayer(this.olMap));

    // Couche de popup
    this.olMap.set('popup_layer', this.mapCorePopupOverlayService.initPopupOverlay(this.olMap));

    // Type de template (fromreader / fullscreen)
    this.olMap.set('template_type', this.templateType);
  }

  /**
   * Retourne les features dessinées
   */
  public getDrawFeatures(): Array<Feature<Geometry>> {
    return this.olMap.get('draw_layer').getSource().getFeatures();
  }

  /**
   * Sélectionne une feature dessinée
   * @param olSelectedFeature
   */
  public selectDrawFeature(olSelectedFeature: Feature<Geometry>) {
    this.mapCoreDrawService.selectFeature(olSelectedFeature, this.olMap.get('draw_layer'));
  }

  /**
   * Supprime une feature
   * @param oFeature
   * @returns
   */
  public removeFeature(oFeature: Feature<Geometry>) {
    return this.olMap.get('draw_layer').getSource().removeFeature(oFeature);
  }

  /**
   * Zoom la carte sur une feature EWKT
   * @param sFeature
   */
  public zoomOnFeatureByString(sFeature: string) {
    const olFeature = new Feature({ geometry: this.mapCoreUtilsService.getGeomFromEWKT(sFeature, this.olMap) });
    this.zoomOnFeature(olFeature);
  }

  /**
   * Zoom la carte sur une feature OL
   * @param sFeature
   */
  public zoomOnFeature(olFeature: Feature<Geometry>) {
    this.zoomOnFeatures([olFeature]);
  }

  /**
   * Zoom la carte sur une feature OL
   * @param sFeature
   * @param bFit Si true alors zoom au maximum, sinon retire un niveau de zoom
   */
  public zoomOnFeatures(aFeatures: Array<Feature<Geometry>>, bFit = false) {

    const aExtent = aFeatures[0].getGeometry().getExtent().slice(0);
    aFeatures.forEach(function (feature) {
      olExtent.extend(aExtent, feature.getGeometry().getExtent());
    });

    // Zoom sur les étendues
    this.olMap.getView().fit(aExtent);

    // Zoom en cas de point
    if (aFeatures.length === 1 && aFeatures[0].getGeometry().getType() === 'Point') {
      this.olMap.getView().setZoom(14);
    } else {
      if (bFit === false) {
        this.olMap.getView().setZoom(this.olMap.getView().getZoom() - 1);
      }
    }

  }

  /**
   * Transmet la valeur au composant parent
   * @param value
   */
  public setValue(value: any) {
    this.setValueEmitter.emit(value);
  }

  /**
   * Initialisation pour formreader
   */
  private initFromJSON_() {

    // Initialisation de la carte Formreader
    this.initMapFromJSON_();

    // Operations pour studio
    if (this.vitisCheck.isDefAndNotNull(this.idStudio)) {
      this.studioInit_();
    }

    // Opération pour le formreader
    if (this.vitisCheck.isDefAndNotNull(this.formreaderId)) {
      this.formReaderInit_();
    }
  }

  /**
   * Initialise les opérations pour studio
   */
  private studioInit_() {
    this.refreshMapOptionSubscription = this.mapCoreStudioService.mapOptionsEventListner(this.idStudio).subscribe((data) => {
      if (this.vitisCheck.isDefAndNotNull(this.olMap)) {
        setTimeout(() => {
          if (data === 'refresh') {
            this.refreshMapFromJSON_();
          }
          if (data === 'resize') {
            // Redimensionne la carte pour éviter les applatissements
            this.mapCoreManagerService.resizeMap(this.olMap);
          }
          // Recharge la carte
          if (data === 'reloadProj' || data === 'reload') {

            if (data === 'reloadProj') {

              // Reprojette la carte en cas de changement de système de projection
              const aExtent = this.olMap.getView().calculateExtent(this.olMap.getSize()); // Étendue en cours
              const center = this.olMap.getView().getCenter(); // Centre en cours
              const proj = this.olMap.getView().getProjection().getCode(); // Code de la projection en cours
              const oNewProj = getProjection(this.mapJSON.proj); // Nouvelle projection
              const aNewExtent = transformExtent(aExtent, proj, this.mapJSON.proj); // Nouvelle étendue
              const aNewProjExtent = oNewProj.getExtent(); // Étendue de la nouvelle projection

              // Si la nouvelle étendue est inclue dans l'étendue du système de projection alors on applique la reprojection
              // Autrement on applique l'étendue du nouveau système de projection
              if (olExtent.containsExtent(aNewProjExtent, aNewExtent)) {
                this.mapJSON.center.extent = aNewExtent;
                this.mapJSON.center.coord = transform(center, proj, this.mapJSON.proj);
              } else {
                this.mapJSON.center.extent = aNewProjExtent;
                this.mapJSON.center.coord = olExtent.getCenter(aNewProjExtent);
              }
            }

            this.reloadMapJSON_();
          }
        }, 200);
      }
    });
  }

  /**
   * Initialise les opérations pour FormReader
   */
  private formReaderInit_() {
    this.refreshMapFormreaderSubscription = this.mapCoreFormreaderService.mapFormreaderEventListner(this.formreaderId).subscribe((data) => {
      if (this.vitisCheck.isDefAndNotNull(this.olMap)) {
        setTimeout(() => {
          const olLayer = this.selectLayer_(data);

          // Update visibility
          if (data.action === 'setVisibility' && this.vitisCheck.isDefAndNotNull(data.visible)) {
            if (this.vitisCheck.isDefAndNotNull(olLayer)) {
              olLayer.setVisible(data.visible);
            }
          }
          // Update style
          if (data.action === 'setStyle' && this.vitisCheck.isDefAndNotNull(data.style)) {
            if (this.vitisCheck.isDefAndNotNull(olLayer)) {
              const oParams = olLayer.getSource().getParams();
              oParams.STYLES = data.style;
              olLayer.getSource().updateParams(oParams);
            }
          }

          // Zoom on Extent
          if (data.action === 'zoomOnExtent' && this.vitisCheck.isDefAndNotNull(data.extent)) {
            if (this.vitisCheck.isDefAndNotNull(this.olMap)) {
              this.olMap.getView().fit(data.extent);
            }
          }

          // Reload map
          if (data.action === 'reload') {
            this.reloadMapJSON_();
          }

          if (data.action === 'activeDrawInteraction') {
            if (this.vitisCheck.isDefAndNotNull(this.olMap)) {
              const olDrawLayer: VectorLayer<VectorSource<Geometry>> = this.olMap.get('draw_layer');
              const olInteraction = this.mapCoreDrawService.addDrawInteraction(data.geomType, this.olMap, olDrawLayer);

              olInteraction.on('drawend', (evt: any) => {
                // Dessin
                const olFeature = evt.feature;
                olFeature.set('layer_grid_id', this.mapCoreManagerService.getUniqueId());
                this.setFeatureEmitter.emit(olFeature);

                // Remove interactions quand le dessin est fini
                if (data.removeInteractionAfterDraw) {
                  this.mapCoreDrawService.removeInteractions(this.olMap);
                }
              });
            }
          }

          if (data.action === 'setFeatureStyle') {
            this.mapCoreUtilsService.setStyleByJSON(data.feature, data.style);
            // Rafraichit la feature
            setTimeout(() => {
              data.feature.changed();
            });
          }
        }, 200);
      }
    });
  }

  /**
   * Initialise la carte depuis mapJSON
   */
  private initMapFromJSON_() {

    // Utilisation FormReader
    this.subscriptions.push(this.mapCoreFormreaderService.initMapFormReader(this.mapJSON, this.targetId, this.tileSize).subscribe((olMap) => {

      // Enregistre l'objet Map
      this.olMap = olMap;

      // Enregistre l'identifiant de carte utilisé
      this.olMap.set('map_json', this.mapJSON);

      // Attribue les autres paramètres (draw_layer...)
      this.setMapAttributes(olMap);

      // Affiche les composants de la carte
      this.displaySubComponents_();

      // Track extent for Studio
      if (this.templateType == 'embedded') {
        this.trackExtent_();
      }
    }));
  }

  /**
   * Rafraîchit la carte
   */
  private refreshMapFromJSON_() {
    if (this.vitisCheck.isDefAndNotNull(this.olMap)) {
      // Cache les menu & controle pour les reinitialisers
      this.haveInteractions = false;
      this.haveControls = false;
      if (this.vitisCheck.isDefAndNotNull(this.olMap.get('scale_line'))) {
        this.olMap.removeControl(this.olMap.get('scale_line'));
      }
      if (this.vitisCheck.isDefAndNotNull(this.olMap.get('mouse_position'))) {
        this.olMap.removeControl(this.olMap.get('mouse_position'));
      }
      this.changeDetectorRef.detectChanges();


      // Enregistre l'identifiant de carte utilisé
      this.olMap.set('map_json', this.mapJSON);

      // Attribue les autres paramètres (draw_layer...)
      this.setMapAttributes(this.olMap);

      // Affiche les composants de la carte
      this.displaySubComponents_();


      // Préviens la carte des changements
      this.olMap.changed();
    }
  }

  /**
   * Recharge la carte à partir du mapJSON
   */
  private reloadMapJSON_() {
    this.olMap.setTarget(null);
    delete this.olMap;
    this.haveInteractions = false;
    this.haveControls = false;
    this.changeDetectorRef.detectChanges();
    this.initMapFromJSON_();
  }

  /**
   * Sélectionne une couche en fonction de data.layer_id
   * @param data
   */
  private selectLayer_(data: any) {
    let oResponseLayer: any;
    if (this.vitisCheck.isDefAndNotNull(this.olMap)) {
      const oLayers = this.olMap.getLayers().getArray();
      for (let i = 0; i < oLayers.length; i++) {
        const oLayer: any = oLayers[i];
        if (this.vitisCheck.isDefAndNotNull(oLayer.get('layer_grid_id')) && oLayer.get('layer_grid_id') === data.layerGridId) {
          oResponseLayer = oLayer;
        }
      }
    }
    return oResponseLayer;
  }

  /**
   * Track l'étendu pour l'envoyé au formreader puis au studio si idStudio exist
   */
  private trackExtent_() {
    if (this.olMap) {
      const olView = this.olMap.getView();
      if (olView) {
        // Étendue
        this.olMap.on('moveend', () => {
          this.changeExtentEmitter.next(this.olMap);
          if (this.vitisCheck.isDefAndNotNull(this.idStudio)) {
            this.mapCoreStudioService.emitMapEvent(this.idStudio, this.olMap);
          } else if (this.vitisCheck.isDefAndNotNull(this.formreaderId)) {
            this.formreaderService.emitMapEvent(this.olMap);
          }
        });
      }
    }
  }

  /**
   * On ne charge les sous composants que quand la carte est chargé.
   * @param {any} mapJSON
   */
  private displaySubComponents_() {

    if (this.mapJSON) {
      // Interractions
      this.haveInteractions = this.vitisCheck.isDefAndNotNull(this.mapJSON.interactions);

      // Controles
      this.haveControls = this.vitisCheck.isDefAndNotNull(this.mapJSON.controls);
    }
  }
}
