import { Injectable } from '@angular/core';

// Vitis
import { VitisCheck, VitisClone } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';

// OpenLayers
import Map from 'ol/Map.js';
import { Draw, Modify, Snap, Select } from 'ol/interaction.js';
import { createBox, createRegularPolygon } from 'ol/interaction/Draw';
import { Vector as VectorLayer } from 'node_modules/ol/layer.js';
import { Vector as VectorSource } from 'ol/source.js';
import { Circle as CircleStyle, Fill, Stroke, Style } from 'ol/style.js';
import { click, pointerMove } from 'ol/events/condition';
import Feature from 'ol/Feature';
import Polygon from 'ol/geom/Polygon';
import { Subject } from 'rxjs';
import Geometry from 'ol/geom/Geometry';
import { LineString, MultiLineString, MultiPolygon, Point } from 'ol/geom';
import { Coordinate } from 'ol/coordinate';

@Injectable({
  providedIn: 'root'
})
export class MapCoreDrawService {

  private selectedFeatureObs = new Subject<Feature<Geometry> | null>();
  private activeInteractionObs = new Subject<Feature<Geometry> | null>();

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private vitisClone: VitisClone,
  ) { }

  /**
   * Retourne l'observable des features sélectionnées
   * @returns
   */
  public getSelectFeatureObs(): Subject<Feature<Geometry> | null> {
    return this.selectedFeatureObs;
  }

  /**
   * Retourne l'observable de l'interaction active
   * @returns
   */
  public getActiveInteractionObs(): Subject<Feature<Geometry> | null> {
    return this.activeInteractionObs;
  }

  /**
   * Initialise la couche de dessin
   * @param  {Map} olMap
   * @returns VectorLayer
   */
  public initDrawLayer(olMap: Map, vectorSource?: VectorSource<Geometry>, olStyle?: Style): VectorLayer<VectorSource<Geometry>> {

    const vectorLayer = this.createDrawLayer(olMap, vectorSource, olStyle);
    olMap.addLayer(vectorLayer);

    // Termine les dessins par le clic droit
    olMap.getViewport().addEventListener('contextmenu', function (evt) {
      evt.preventDefault();

      const aInteractions = olMap.getInteractions().getArray();
      for (let i = 0; i < aInteractions.length; i++) {
        const olInteraction: any = aInteractions[i];
        if (olInteraction.get('vitis_draw') === true) {
          try {
            // Termine le dessin en cours
            olInteraction.finishDrawing();
            // Desactive la fonctionnalié pour ne pas commencer un autre dessin
            olInteraction.setActive(false);
            // Réactive la fonctionnalité
            setTimeout(() => {
              olInteraction.setActive(true);
            }, 500);
          } catch (e) { }
        }
      }
    });

    return vectorLayer;
  }

  /**
   * Crée une couche vecteur correspondant au seddin
   * @returns VectorLayer
   */
  public createDrawLayer(olMap: Map, source?: VectorSource<Geometry>, olStyle?: Style): VectorLayer<VectorSource<Geometry>> {

    let oDrawStyle: any;
    // Lecture properties
    if (this.vitisCheck.isDefAndNotNull(environment.properties)) {
      if (this.vitisCheck.isDefAndNotNull(environment.properties.map)) {
        if (this.vitisCheck.isDefAndNotNull(environment.properties.map.draw_style)) {
          oDrawStyle = environment.properties.map.draw_style;
        }
      }
    }

    // Lecture map_json
    if (this.vitisCheck.isDefAndNotNull(olMap.get('map_json'))) {
      const oMapJson = olMap.get('map_json');
      if (this.vitisCheck.isDefAndNotNull(oMapJson.draw_color)) {
        oDrawStyle.draw_color = oMapJson.draw_color;
      }
      if (this.vitisCheck.isDefAndNotNull(oMapJson.contour_color)) {
        oDrawStyle.contour_color = oMapJson.contour_color;
      }
      if (this.vitisCheck.isDefAndNotNull(oMapJson.contour_size)) {
        oDrawStyle.contour_size = oMapJson.contour_size;
      }
      if (this.vitisCheck.isDefAndNotNull(oMapJson.circle_radius)) {
        oDrawStyle.circle_radius = oMapJson.circle_radius;
      }
    }

    const drawColor = this.vitisCheck.isDefAndNotNull(oDrawStyle['draw_color']) ? oDrawStyle['draw_color'] : 'rgba(54, 184, 255, 0.6)';
    const contourColor = this.vitisCheck.isDefAndNotNull(oDrawStyle['contour_color']) ? oDrawStyle['contour_color'] : 'rgba(0, 0, 0, 0.4)';
    const contourSize = this.vitisCheck.isDefAndNotNull(oDrawStyle['contour_size']) ? oDrawStyle['contour_size'] : 2;
    const circleRadius = this.vitisCheck.isDefAndNotNull(oDrawStyle['circle_radius']) ? oDrawStyle['circle_radius'] : 7;

    const vectorStyle = olStyle ? olStyle : new Style({
      fill: new Fill({
        color: drawColor
      }),
      stroke: new Stroke({
        color: contourColor,
        width: contourSize
      }),
      image: new CircleStyle({
        radius: circleRadius,
        fill: new Fill({
          color: drawColor
        }),
        stroke: new Stroke({
          color: contourColor,
          width: contourSize
        })
      })
    });

    const vectorSource = source || new VectorSource();
    const vectorLayer = new VectorLayer({
      source: vectorSource,
      style: vectorStyle,
      zIndex: 1000,
    });

    return vectorLayer;
  }

  /**
   * Ajoute l'interraction dessin
   * @param  {string} sType
   * @param  {Map} olMap
   * @param  {VectorLayer} olDrawLayer
   * @param  {any} oOptions?
   */
  public addDrawInteraction(sType: string, olMap: Map, olDrawLayer: VectorLayer<VectorSource<Geometry>>, oOptions?: any) {

    oOptions = this.vitisCheck.isDefAndNotNull(oOptions) ? oOptions : {};

    let olInteraction: any;
    let olInteractionSelect: any;

    // Snaping
    const olInteractiontSnap: any = this.vitisCheck.isDefAndNotNull(oOptions.snap) ? oOptions.snap : new Snap({
      source: olDrawLayer.getSource()
    });
    // Désactive l'accrochage car il doit être ajouté après les interactions de dessin
    olMap.removeInteraction(olInteractiontSnap);

    const drawOptions: any = {
      source: olDrawLayer.getSource()
    };

    if (this.vitisCheck.isDefAndNotNull(oOptions.drawStyle)) {
      drawOptions.style = oOptions.drawStyle;
    }

    switch (sType) {
      // Type Remove all
      case 'removeAll':
        olDrawLayer.getSource().clear();
        break;

      // Type Remove
      case 'remove':
        olInteraction = new Select({
          condition: click
        });

        olInteraction.on('select', (evt: any) => {
          for (let i = 0; i < evt.selected.length; i++) {
            const feature = evt.selected[i];
            olDrawLayer.getSource().removeFeature(feature);
            olInteraction.getFeatures().remove(feature);
          }
        });
        break;

      // Type modify
      case 'modify':
        olInteraction = new Modify({
          source: olDrawLayer.getSource(),
          snapToPointer: true
        });
        break;

      // Type point
      case 'point':
        olInteraction = new Draw({
          ...drawOptions,
          type: 'Point'
        });
        break;
      case 'multipoint':
        olInteraction = new Draw({
          ...drawOptions,
          type: 'MultiPoint'
        });
        break;

      // Type linestring
      case 'line':
      case 'multiline':
      case 'linestring':
        olInteraction = new Draw({
          ...drawOptions,
          type: 'LineString'
        });
        break;

      // Type polygon
      case 'polygon':
        olInteraction = new Draw({
          ...drawOptions,
          type: 'Polygon'
        });
        break;

      // Type multipolygon
      case 'multipolygon':
        olInteraction = new Draw({
          ...drawOptions,
          type: 'MultiPolygon'
        });
        break;

      // Type circle
      case 'circle':
        olInteraction = new Draw({
          ...drawOptions,
          type: 'Circle'
        });
        break;

      // Type box
      case 'box':
        olInteraction = new Draw({
          ...drawOptions,
          type: 'Circle',
          geometryFunction: createBox()
        });
        break;

      // Type square
      case 'square':
        olInteraction = new Draw({
          ...drawOptions,
          type: 'Circle',
          geometryFunction: createRegularPolygon(4)
        });
        break;

      // Type select
      case 'select':
        olInteraction = this.setSelectInteraction(olMap, olDrawLayer);
        break;

      // Type removeVertex
      case 'removeVertex':

        if (oOptions &&
          oOptions.feature &&
          oOptions.pointsLayer) {

          // Tableau de points correspondant aux sommets de la feature
          const aVertexPoints: Array<Feature> = this.getFeaturesVertex(oOptions.feature);
          const pointsSource = oOptions.pointsLayer.getSource();
          pointsSource.clear();
          pointsSource.addFeatures(aVertexPoints);

          let initialCoods: number[];
          olInteraction = new Modify({
            source: pointsSource,
            snapToPointer: true,
            deleteCondition: (event: any) => {
              if (event.type == 'pointerdown') {
                // mémorise les coordonnées lors du clickdown
                const featurePoint = pointsSource.getClosestFeatureToCoordinate(event.coordinate);
                initialCoods = this.vitisClone.clone(featurePoint.getGeometry().getCoordinates());
              }
              if (event.type == 'pointerup') {
                const featurePoint = pointsSource.getClosestFeatureToCoordinate(event.coordinate);

                // Vérifie que le point n'ai pas bougé
                if (featurePoint.getGeometry().getCoordinates()[0] === initialCoods[0] &&
                  featurePoint.getGeometry().getCoordinates()[1] === initialCoods[1]) {

                  if (this.removeFeatureVertex(initialCoods, oOptions.feature)) {

                    // suppression dans la couche de points
                    pointsSource.removeFeature(featurePoint);

                    // Event de changement de la feature
                    if (oOptions.onRemoveVertex && oOptions.onRemoveVertex.next) {
                      oOptions.onRemoveVertex.next(oOptions.feature);
                    }
                  } else {
                    console.error('cannot remove feature vertex');
                  }
                } else {
                  console.error('please do not move the vertex');
                  featurePoint.getGeometry().setCoordinates(initialCoods);
                  return false;
                }
              }
              return true;
            }
          });
        }
        break;
      default:
        console.error('Cannot draw type', sType);
        break;
    }

    if (this.vitisCheck.isDefAndNotNull(olInteraction)) {

      // Permet d'identifier les interactions ajoutées par cette fonction
      olInteraction.set('vitis_draw', true);

      // Événements drawend
      olInteraction.on('drawend', (evt: any) => {
        // Multiple features
        if (oOptions.multiple === false) {
          olDrawLayer.getSource().clear();
        }
      });

      // Ajoute l'interraction
      olMap.addInteraction(olInteraction);
    }

    // Re-ajoute l'accrochage car il doit être ajouté après les interactions de dessin
    olMap.addInteraction(olInteractiontSnap);

    // Renseigne l'interaction active
    this.activeInteractionObs.next(olInteraction);

    if (oOptions && oOptions.getInteractions) {
      return {
        'interaction': olInteraction,
        'select': olInteractionSelect,
        'snap': olInteractiontSnap
      };
    }
    return olInteraction;
  }

  /**
   * Supprime toutes les interractions de type dessin
   * @param  {Map} olMap
   */
  public removeInteractions(olMap: Map) {
    // On fait une copie du tableau (pas par référence) sinon quand on fait le removeInteraction la boucle s'arrête.
    const aInteractions = [...olMap.getInteractions().getArray()];
    for (let i = 0; i < aInteractions.length; i++) {
      const oInteraction = aInteractions[i];
      if (oInteraction.get('vitis_draw')) {
        olMap.removeInteraction(oInteraction);
      }
    }

    // Renseigne l'interaction active
    this.activeInteractionObs.next(null);
  }

  /**
   * Active l'interraction select
   */
  public setSelectInteraction(olMap: Map, olDrawLayer: VectorLayer<VectorSource<Geometry>>) {

    const select = new Select({
      layers: [olDrawLayer],
      style: null,
      multi: true,
    });
    select.set('vitis_draw', true);
    olMap.addInteraction(select);

    select.on('select', (e: any) => {
      if (e && e.selected && e.selected[0]) {
        const olSelectedFeature = e.selected[0];
        this.selectFeature(olSelectedFeature, olDrawLayer);
      } else {
        this.selectFeature(null, olDrawLayer);
      }
    });

    return select;
  }

  /**
   * Sélectionne une feature
   *
   * @param olSelectedFeature
   * @param olDrawLayer
   */
  public selectFeature(olSelectedFeature: Feature<Geometry>, olDrawLayer: VectorLayer<VectorSource<Geometry>>) {

    // Cache le potentiel précédent contour de sélection
    this.hideSelectedFeatureContour(olDrawLayer);

    if (olSelectedFeature) {

      // Valeur de l'observable
      this.selectedFeatureObs.next(olSelectedFeature);

      // Affiche le contour de sélection
      this.displaySelectedFeatureContour(olSelectedFeature, olDrawLayer);

    } else {

      // Valeur de l'observable
      this.selectedFeatureObs.next(null);
    }
  }

  /**
   * Ajoute la feature de contour de sélection
   * @param olFeature
   * @param select
   * @param olDrawLayer
   */
  private displaySelectedFeatureContour(olFeature: Feature<Geometry>, olDrawLayer: VectorLayer<VectorSource<Geometry>>) {
    const fill = new Fill({
      color: 'rgba(255,255,255,0)'
    });
    const stroke = new Stroke({
      color: 'rgba(0,0,0,0.4)',
      width: 2,
      lineDash: [10]
    });
    const selectStyle = new Style({
      image: new CircleStyle({
        fill: fill,
        stroke: stroke,
        radius: 50
      }),
      fill: fill,
      stroke: stroke
    });

    if (olFeature) {
      const featureExtentArray = olFeature.getGeometry().getExtent();
      let selectionFeature;
      if (olFeature.getGeometry().getType() === 'Point') {
        selectionFeature = new Feature({
          geometry: olFeature.getGeometry()
        });
      } else {
        const xmin = featureExtentArray[0];
        const ymin = featureExtentArray[1];
        const xmax = featureExtentArray[2];
        const ymax = featureExtentArray[3];
        const featureExtentGeometry = new Polygon([
          [
            [xmin, ymin],
            [xmin, ymax],
            [xmax, ymax],
            [xmax, ymin]
          ]
        ]);
        selectionFeature = new Feature({
          geometry: featureExtentGeometry
        });
      }
      selectionFeature.setStyle(selectStyle);
      selectionFeature.set('isSelectFeatureContour', true);
      olDrawLayer.getSource().addFeature(selectionFeature);
    }
  }

  /**
   * Supprime la feature de contour de sélection
   * @param olDrawLayer
   */
  private hideSelectedFeatureContour(olDrawLayer: VectorLayer<VectorSource<Geometry>>) {
    const aFeatures = olDrawLayer.getSource().getFeatures();
    for (let i = aFeatures.length - 1; i >= 0; i--) {
      const olFeature = aFeatures[i];
      if (olFeature.get('isSelectFeatureContour') === true) {
        olDrawLayer.getSource().removeFeature(olFeature);
      }
    }
  }

  /**
   * Calcul la distance entre deux points
   * @param p1
   * @param p2
   */
  public sdistance(p1: Array<number>, p2: Array<number>) {
    const dx = p1[0] - p2[0];
    const dy = p1[1] - p2[1];
    return dx * dx + dy * dy;
  }

  /**
   * Retourne une liste de points qui correspondent aux vertex d'une géométrie
   * @param feature
   * @returns
   */
  public getFeaturesVertex(feature: Feature): Array<Feature> {
    let aFeaturesVertex: Array<Feature> = [];
    const sGeomType: string = feature.getGeometry().getType();
    let oGeom: MultiPolygon | Polygon | MultiLineString | LineString;
    if (sGeomType === 'MultiPolygon') {
      oGeom = feature.getGeometry() as MultiPolygon;
      const aGeomCoordinates: Array<any> = oGeom.getCoordinates();
      for (let i = 0; i < aGeomCoordinates.length; i++) {
        aFeaturesVertex = [...aFeaturesVertex, ...aGeomCoordinates[i][0].filter((geom, index) => (index != 0) ? geom : null).map(geom => new Feature({ geometry: new Point(geom) }))];
      }
    }
    if (sGeomType === 'Polygon') {
      oGeom = feature.getGeometry() as Polygon;
      aFeaturesVertex = oGeom.getCoordinates()[0].filter((geom, index) => (index != 0) ? geom : null).map(geom => new Feature({ geometry: new Point(geom) }));
    }
    if (sGeomType === 'MultiLineString') {
      oGeom = feature.getGeometry() as MultiLineString;
      const aGeomCoordinates: Array<any> = oGeom.getCoordinates();
      for (let i = 0; i < aGeomCoordinates.length; i++) {
        aFeaturesVertex = [...aFeaturesVertex, ...aGeomCoordinates[i].map(geom => new Feature({ geometry: new Point(geom) }))];
      }
    }
    if (sGeomType === 'LineString') {
      oGeom = feature.getGeometry() as LineString;
      aFeaturesVertex = oGeom.getCoordinates().map(geom => new Feature({ geometry: new Point(geom) }));
    }
    return aFeaturesVertex;
  }

  /**
   * Suppression d'un sommet dans une feature
   * @param coordinate
   * @param olFeature
   * @returns
   */
  public removeFeatureVertex(coordinate: Coordinate, olFeature: Feature): boolean {

    // Vérifie si il y a assez de vertex pour en supprimer 1.
    if (!this.checkRemoveVertex(coordinate, olFeature)) {
      return false;
    }

    const geomType = olFeature.getGeometry().getType();
    let geom: Polygon | LineString | MultiPolygon | MultiLineString;
    if (geomType == 'Polygon') {
      geom = olFeature.getGeometry() as Polygon;
    }
    if (geomType == 'LineString') {
      geom = olFeature.getGeometry() as LineString;
    }
    if (geomType == 'MultiPolygon') {
      geom = olFeature.getGeometry() as MultiPolygon;
    }
    if (geomType == 'MultiLineString') {
      geom = olFeature.getGeometry() as MultiLineString;
    }

    // Récupère les coordonnées de la géométrie et le point le plus proche des coordonnées supprimé
    const aCoord: any = geom.getCoordinates();
    const getClosestPoint = geom.getClosestPoint(coordinate);

    if (geomType == 'Polygon') {
      const index = aCoord[0].findIndex((coord) => (coord[0] === getClosestPoint[0] && coord[1] === getClosestPoint[1]));
      if (index != -1) {
        if (index == 0) {
          aCoord[0][aCoord[0].length - 1] = aCoord[0][index + 1];
        }
        aCoord[0].splice(index, 1);
      }
    }
    if (geomType == 'LineString') {
      const index = aCoord.findIndex((coord) => (coord[0] === getClosestPoint[0] && coord[1] === getClosestPoint[1]));
      if (index != -1) {
        aCoord.splice(index, 1);
      }
    }
    if (geomType == 'MultiPolygon') {
      for (let i = 0; i < aCoord.length; i++) {
        const index = aCoord[i][0].findIndex((coord) => (coord[0] === getClosestPoint[0] && coord[1] === getClosestPoint[1]));
        if (index != -1) {
          if (index == 0) {
            aCoord[i][0][aCoord[i][0].length - 1] = aCoord[i][0][index + 1];
          }
          aCoord[i][0].splice(index, 1);
        }
      }
    }
    if (geomType == 'MultiLineString') {
      for (let i = 0; i < aCoord.length; i++) {
        const index = aCoord[i].findIndex((coord) => (coord[0] === getClosestPoint[0] && coord[1] === getClosestPoint[1]));
        if (index != -1) {
          aCoord[i].splice(index, 1);
        }
      }
    }

    // Attribut les nouvelles coordonnées à la feature
    geom.setCoordinates(aCoord);

    return true;
  }

  /**
   * Vérifie si il y a assez de vertex pour en supprimer 1.
   * @param event
   * @param olDrawLayer
   */
  public checkRemoveVertex(coordinate: Coordinate, olFeature: Feature) {
    let geom: any, aGeomCoordinates: any[];
    const sGeomType = olFeature.getGeometry().getType();
    switch (sGeomType) {
      case 'Polygon':
        geom = olFeature.getGeometry() as Polygon;
        // Vérifie si il y a au moins 3 sommets (donc 4 coordonnées)
        if (geom.getCoordinates()[0].length < 5) {
          return false;
        }
        break;
      case 'MultiPolygon':
        geom = olFeature.getGeometry() as MultiPolygon;
        aGeomCoordinates = geom.getCoordinates();
        for (let i = 0; i < aGeomCoordinates.length; i++) {
          // Vérifie si il y a au moins 3 sommets (donc 4 coordonnées)
          if (aGeomCoordinates[i][0].findIndex((coord) => (coord[0] === coordinate[0] && coord[1] === coordinate[1])) != -1 && aGeomCoordinates[i][0].length < 5) {
            return false;
          }
        }
        break;
      case 'LineString':
        geom = olFeature.getGeometry() as LineString;
        // Vérifie si il y a au moins 2 sommets
        if (geom.getCoordinates().length < 3) {
          return false;
        }
        break;
      case 'MultiLineString':
        geom = olFeature.getGeometry() as MultiLineString;
        aGeomCoordinates = geom.getCoordinates();
        for (let i = 0; i < aGeomCoordinates.length; i++) {
          // Vérifie si il y a au moins 2 sommets
          if (aGeomCoordinates[i].findIndex((coord) => (coord[0] === coordinate[0] && coord[1] === coordinate[1])) != -1 && aGeomCoordinates[i].length < 3) {
            return false;
          }
        }
        break;
      default:
        console.error(`Type ${sGeomType} not valid for vertex remove`);
        return false;
        break;
    }
    return true;
  }
}
