import { Injectable } from '@angular/core';
import { VitisCheck, VitisRequest, vitisAlert } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';
import { Subject, Observable, of } from 'rxjs';
import { MapCoreJSONService } from '../map-core-json/map-core-json.service';
import { MapDefinition, MapCatalog, MapJSON, VmapLayer } from '../map-models.model';
import { TranslateService } from '@ngx-translate/core';

// OpenLayers
import Map from 'node_modules/ol/Map';
import LayerGroup from 'ol/layer/Group';
import View from 'ol/View';
import Layer from 'ol/layer/Layer';
import ImageWMS from 'ol/source/ImageWMS';
import TileWMS from 'ol/source/TileWMS';
import ImageLayer from 'ol/layer/Image';
import { Projection, addProjection, addCoordinateTransforms, transform } from 'ol/proj';
import { GPX, GeoJSON, IGC, KML, TopoJSON } from 'ol/format';
import Collection from 'ol/Collection';
import { Source } from 'ol/source';
import VectorLayer from 'ol/layer/Vector';
import { Vector as VectorSource } from 'ol/source';

@Injectable({
  providedIn: 'root'
})
export class MapCoreManagerService {

  private oMapCatalog_: MapCatalog;
  private oProjections_: any;

  public mapVisibilityChange = new Subject();

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private mapCoreJSONService: MapCoreJSONService,
    private translate: TranslateService
  ) { }

  /**
   * Retourne le MapCatalog sous la forme d'un observable
   * @returns Observable<MapCatalog>
   */
  public getMapCatalog(reloadCatalog?: boolean): Observable<MapCatalog> {
    const oReturn = new Subject<MapCatalog>();

    if (this.oMapCatalog_ && !reloadCatalog) {
      setTimeout(() => {
        oReturn.next(this.oMapCatalog_);
      });
    } else {
      this.ajaxGetMapCatalog().then((oMapCatalog) => {
        this.oMapCatalog_ = oMapCatalog;
        oReturn.next(oMapCatalog);
      }).catch(() => {
        oReturn.next(null);
      });
    }

    return oReturn;
  }

  /**
   * Retourne l'identifiant de la carte par défaut
   * @param  {MapCatalog} oMapCatalog
   * @returns number
   */
  public getDefaultMapFromCatalog(oMapCatalog: MapCatalog): number {

    // Si usedMap
    if (this.vitisCheck.isDefAndNotNull(oMapCatalog.usedMap) &&
      this.vitisCheck.isDefAndNotNull(oMapCatalog.maps[oMapCatalog.usedMap])) {
      const oUsedMap = oMapCatalog.maps[oMapCatalog.usedMap];
      if (oUsedMap &&
        oUsedMap.map_id) {
        return oUsedMap.map_id;
      }
    } else if (this.vitisCheck.isDefAndNotNull(oMapCatalog.defaultMap) &&
      this.vitisCheck.isNumber(oMapCatalog.defaultMap)) {
      return oMapCatalog.defaultMap;
    } else {
      let iMapId: number;
      let iMinIndex: number;

      // Première carte
      if (oMapCatalog.maps &&
        oMapCatalog.maps[0] &&
        oMapCatalog.maps[0].map_id) {
        iMapId = oMapCatalog.maps[0].map_id;
      }

      // Recherche avec catalog_index
      for (let i = 0; i < oMapCatalog.maps.length; i++) {
        const oMap = oMapCatalog.maps[i];
        if (oMap.map_id &&
          oMap.catalog_index) {
          if (!iMinIndex ||
            iMinIndex > oMap.catalog_index) {
            iMapId = oMapCatalog.maps[i].map_id;
            iMinIndex = oMapCatalog.maps[i].catalog_index;
          }
        }
      }
      return iMapId;
    }
  }

  /**
   * Charge une carte par son map_id
   * @param iMapId
   * @param targetId
   * @param tileSize
   */
  public initOlMap(iMapId: number, targetId: string, tileSize: Array<number>, olExistingMap?: Map): Observable<any> {
    const oReturn = new Subject<any>();
    this.ajaxGetMap(iMapId).then((oMapDef: any) => {

      // Initialise la carte
      const olMap = this.initMapFromDef(oMapDef, targetId, tileSize, olExistingMap);
      if (olMap) {
        oReturn.next(olMap);
      }
    }).catch(() => {
      this.translate.get(['COMPONENT_MAP_ERROR_GET_MAP_CATALOG_NO_MAP', 'COMPONENT_MAP_ERROR_GET_MAP_CATALOG_UNAUTHORIZED']).subscribe((translations: string) => {
        vitisAlert.alert({
          type: 'error',
          title: translations['COMPONENT_MAP_ERROR_GET_MAP_CATALOG_NO_MAP'],
          html: translations['COMPONENT_MAP_ERROR_GET_MAP_CATALOG_UNAUTHORIZED']
        });
      });
      oReturn.next({ error: 'unauthorized' });
    });
    return oReturn;
  }

  /**
   * Génère le mapfile privé et retourne une promesse
   * @returns Promise
   */
  public generatePrivateWMSService(): Promise<void> {
    return new Promise((resolve, reject) => {
      this.vitisRequest.ajaxRequestPromise({
        method: 'POST',
        url: environment.properties.app_api_url + '/vm4ms/WmsServices/MapFile',
        data: {
          wmsservice_id: 'private',
          type: 'prod'
        }
      }).then((resp: any) => {
        resolve();
      }).catch((error) => {
        reject();
      });
    });
  }

  /**
   * Retourne la définition d'une carte par son map_id
   * @param  {number} iMapId
   * @returns Promise
   */
  public ajaxGetMap(iMapId: number, aFilter: any = null): Promise<any> {
    return new Promise((resolve, reject) => {

      const params: any = {};
      if (aFilter && aFilter.length > 0) {
        params.filter = aFilter;
      }

      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: environment.properties.app_api_url + '/vmap/mapjsons/' + iMapId,
        params,
      }).then((resp: any) => {
        if (resp.data) {
          if (resp.data.status === 1) {
            if (this.vitisCheck.isDefAndNotNull(resp.data.data)) {
              resolve(resp.data.data);
            } else {
              console.error('Cannot get map ' + iMapId);
              reject();
            }
          } else {
            console.error('Cannot get map ' + iMapId);
            reject();
          }
        } else {
          console.error('Cannot get map ' + iMapId);
          reject();
        }
      });
    });
  }

  /**
   * Charge le mapcatalog
   * @returns Promise
   */
  public ajaxGetMapCatalog(): Promise<MapCatalog> {
    return new Promise((resolve, reject) => {
      this.vitisRequest.ajaxRequest({
        method: 'GET',
        url: environment.properties.app_api_url + '/vmap/mapcatalog',
        success: (response: any) => {
          if (response.data &&
            response.data.status === 1 &&
            this.vitisCheck.isDefAndNotNull(response.data.data) &&
            this.vitisCheck.isDefAndNotNull(response.data.data.maps)) {
            resolve(response.data.data);
          } else {
            const emptyCatalog: MapCatalog = {
              'maps': [],
              'services': {}
            };
            resolve(emptyCatalog);
          }
        },
        error: (error: any) => {
          reject();
        }
      });
    });
  }

  /**
   * Initialise une carte depuis la définition
   * @param oMapDef
   * @param targetId
   * @param tileSize
   */
  public initMapFromDef(oMapDef: MapDefinition, targetId: string, tileSize: Array<number>, olExistingMap?: Map, measureFeatures?: any) {

    const olMap: Map = olExistingMap ? olExistingMap : new Map({
      target: targetId,
      controls: []
    });

    const olView: View = this.mapCoreJSONService.getViewFromDef(oMapDef, {
      size: olMap.getSize(),
      tileSize: tileSize
    });

    const vectortilesLoaded = new Subject();
    const olLayers: Array<Layer<Source>> = this.mapCoreJSONService.getLayersFromDef(oMapDef, {
      size: olMap.getSize(),
      tileSize: tileSize
    }, olMap, vectortilesLoaded);

    if (this.vitisCheck.isDefAndNotNull(olView)) {
      olView.set('home_position', {
        center: olView.getCenter(),
        resolution: olView.getResolution()
      });
      olMap.setView(olView);
    } else {
      console.error('View not defined');
    }

    if (this.vitisCheck.isArray(olLayers)) {

      olMap.set('original_layers', olLayers);

      let aOptimizedLayers;
      if (oMapDef.options && oMapDef.options.optimized) {
        aOptimizedLayers = this.optimiseLayers([...olLayers]);
        olMap.setLayerGroup(new LayerGroup({
          layers: aOptimizedLayers
        }));
      } else {
        olMap.setLayerGroup(new LayerGroup({
          layers: olLayers
        }));
      }

      olMap.set('map_layers', olLayers);
      olMap.set('map_optimized_layers', aOptimizedLayers);
    } else {
      console.error('Layers not defined');
    }

    // Enregistre la définition de la carte
    olMap.set('map_definition', oMapDef);

    olMap.set('measure_features', measureFeatures);

    // Lance l'événement "changed"
    olMap.changed();

    // Chargement vectortiles terminé (car elles se chargent en asynchrone)
    const vectortilesLoadedSubscriber = vectortilesLoaded.subscribe((aLayersDef: Array<any>) => {

      // Recharge l'ordre des couches
      const aLayersOrder = [];
      if (this.vitisCheck.isArray(aLayersDef)) {
        for (let i = 0; i < aLayersDef.length; i++) {
          if (aLayersDef[i].layer_id) {
            aLayersOrder.push(aLayersDef[i].layer_id);
          }
        }
      }
      this.setLayersOrder(aLayersOrder, olMap);
      vectortilesLoadedSubscriber.unsubscribe();
      olMap.dispatchEvent('vectortiles_loaded');
    });

    return olMap;
  }

  /**
   * Refraichit les couches de la carte sans garder de cache
   * @param olMap
   */
  public refreshMapLayers(olMap: Map, keepLoadedTiles = false) {

    const aLayers = olMap.getLayers().getArray();
    for (let i = 0; i < aLayers.length; i++) {
      const olLayer: any = aLayers[i];
      const olSource = olLayer.getSource();

      if (
        olSource instanceof ImageWMS ||
        olSource instanceof TileWMS
      ) {
        const params = olSource.getParams();
        params.TIMESTAMP = Date.now();
        olSource.updateParams(params);

        if (keepLoadedTiles !== true) {
          olSource.refresh();
        }
      }
    }
  }

  /**
   * Rafraichi les couches de la cartes en fonctions de la définition donnée
   * @param olMap
   * @param oMapDef
   */
  private refreshLayersFromDef_(oMapDef: MapDefinition, olMap: Map): Map {
    const olLayers: Array<Layer<Source>> = this.mapCoreJSONService.getLayersFromDef(oMapDef, {
      size: olMap.getSize(),
      tileSize: [512, 512]
    }, olMap);

    if (this.vitisCheck.isArray(olLayers)) {
      olMap.setLayerGroup(new LayerGroup({
        layers: olLayers
      }));
    } else {
      console.error('Layers not defined');
    }

    // Enregistre la définition de la carte
    olMap.set('map_definition', oMapDef);

    // Lance l'événement "changed"
    olMap.changed();

    return olMap;
  }

  /**
   * Rafraichit la taille de la carte et évite les applatissements
   */
  public resizeMap(olMap: Map) {
    for (let i = 0; i < 50; i++) {
      setTimeout(() => {
        if (this.vitisCheck.isDefAndNotNull(olMap)) {
          olMap.updateSize();
        }
      }, i * 20);
    }
  }

  /**
   * Retourne un unique identifiant HTML
   * @returns string
   */
  public getUniqueId(): string {
    return 'map_' + Date.now() + Math.trunc(Math.random() * 10000);
  }

  /**
   * Charge les projections disponibles
   * @returns Promise
   */
  public ajaxGetProjections(): Promise<any> {
    return new Promise((resolve, reject) => {
      const url = 'assets/data/proj.json';

      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data)) {
          resolve(resp.data);
        } else {
          console.error('Error getting JSON form');
          reject(false);
        }
      });
    });
  }

  /**
   * Retourne les projections sous la forme d'un observable
   * @returns Observable<any>
   */
  public getProjections(): Observable<any> {
    const oReturn = new Subject<any>();

    if (this.oProjections_) {
      setTimeout(() => {
        oReturn.next(this.oProjections_);
      });
    } else {
      this.ajaxGetProjections().then((oProjections) => {
        oReturn.next(oProjections);
      });
    }

    return oReturn;
  }

  /**
   * Créer une nouvelle carte à partir d'une existante
   * @param {Map} olMap
   * @param {any} newMapTreeDef
   */
  public setNewMap(olMap: Map, newMapTreeDef: any) {
    olMap.set('name', 'Nouvelle carte');
    olMap.set('new_map', true);
    olMap.set('map_id', null);
    olMap = this.initMapFromDef(newMapTreeDef, this.getUniqueId(), [512, 512], olMap);
  }

  /**
   * Ajoute une couche à la carte
   * @param {Map} olMap
   * @param {Layer} oLayer
   */
  public addLayer(olMap: Map, oLayer: any) {

    const newLayer = {};
    let doesServiceExist: number = 0;
    let oTree = { ... this.getLayersTree_(olMap) };
    const aLayers = olMap.getLayers().getArray();
    const aCenter = olMap.getView().getCenter();
    const sResolution = olMap.getView().getResolution();
    const sZoom = olMap.getView().getZoom();

    // Test si le service est déjà inclus, si il l'est, la variable
    // doesServiceExist prend la valeur de l'index du service
    for (let i = 0; i < oTree.children.length; i++) {
      if (oTree.children[i].name === oLayer.serviceName) {
        doesServiceExist = i;
      }
    }

    // Test si la couche est interrogeable (non checkée par défaut)
    oLayer.select = null;
    if (oLayer.queryable === 'true') {
      oLayer.select = false;
    }

    // Recrée la vue (dont on ne se sert pas)
    oTree.children[0].view.zoom = sZoom;
    oTree.children[0].view.center = aCenter;
    oTree.children[0].view.resolution = sResolution;

    // Recrée un arbre de couches
    if ((oLayer.layerType === 'tilewms') || (oLayer.layerType === 'imagewms')) {
      oTree = this.prepareLayerTilewmsImagewms(oLayer, oTree, doesServiceExist, newLayer, aLayers);
    } else if (oLayer.layerType === 'osm') {
      oTree = this.prepareLayerOSM(oLayer, oTree, doesServiceExist, newLayer);
    } else if (oLayer.layerType === 'xyz') {
      oTree = this.prepareLayerXYZ(oLayer, oTree, doesServiceExist, newLayer);
    } else if (oLayer.layerType === 'bing') {
      oTree = this.prepareLayerBing(oLayer, oTree, doesServiceExist, newLayer);
    } else if (oLayer.layerType === 'wmts') {
      oTree = this.prepareLayerWMTS(oLayer, oTree, doesServiceExist, newLayer);
    } else if (oLayer.layerType === 'imagevector') {
      oTree = this.prepareLayerImageVector(oLayer, oTree, doesServiceExist, newLayer);
    } else {
      console.error('type :"' + oLayer.layerType + '" non pris en compte');
    }

    if (oTree === 0 || oTree === false || oTree === null) {
      return oTree;
    }

    olMap = this.refreshLayersFromDef_(oTree, olMap);
  }

  /**
   * Retourne l'arbre
   * @param olMap
   * @returns
   */
  public getLayersTree(olMap: Map) {
    const aTree = [];
    const oDef = olMap.get('map_definition');

    if (oDef && oDef.children) {
      // Commence à 1 car 0 correspond à la vue
      for (let i = 1; i < oDef.children.length; i++) {
        aTree.push(oDef.children[i]);
      }
    }

    return aTree;
  }

  /**
   * Retourne les couches actuelles sous forme d'arbre
   * @param olMap
   * @returns
   */
  public getLayersList(olMap: Map) {
    const aList = [];

    let aLayers = this.deoptimiseLayers(olMap);
    aLayers = this.joinVectorTileslayers_(aLayers);
    const oMapDef = olMap.get('map_definition');

    for (const oLayer of aLayers) {
      const iLayerID = oLayer.get('layer_id');
      const oLayerDef = this.findLayerInTree(iLayerID, oMapDef);

      if (oLayerDef) {
        if (!this.vitisCheck.isDefAndNotNull(oLayerDef.olLayer)) {
          oLayerDef.olLayer = oLayer;
        }

        if (this.vitisCheck.isString(oLayerDef.definition_sym_json)) {
          try {
            oLayerDef.definition_sym_json = JSON.parse(oLayerDef.definition_sym_json);
          } catch (error) { }
        }

        if (this.vitisCheck.isString(oLayerDef.published_form_json_def)) {
          try {
            oLayerDef.published_form_json_def = JSON.parse(oLayerDef.published_form_json_def);
          } catch (error) { }
        }

        const oTreeLayer = {
          ...oLayerDef,
          ...{
            id: this.getUniqueId(),
            opacity: oLayer.get('opacity'),
            visibility: oLayer.get('visible') ? 1 : 0,
          }
        };
        aList.push(oTreeLayer);
      }
    }

    return aList.sort((a, b) => {
      if (a.olLayer && b.olLayer) {
        if (a.olLayer.getZIndex() < b.olLayer.getZIndex()) {
          return -1;
        } else if (a.olLayer.getZIndex() > b.olLayer.getZIndex()) {
          return 1;
        } else {
          return 0;
        }
      } else {
        return 0;
      }
    });
  }

  /**
   * Retourne un nouveau tableau de couches optimisé avec des couches fusionnées de manière à faire le moins de requêtes possibles
   * @param aLayers
   * @returns
   */
  public optimiseLayers(aLayers: Array<Layer<Source>>): Array<Layer<Source>> {
    const optiLayers = [];

    for (let i = 0; i < aLayers.length; i++) {
      const olLayer = aLayers[i];

      if (olLayer) {
        if (aLayers[i + 1]) {
          const olNextLayer = aLayers[i + 1];
          if (this.areLayersFusionable_(olLayer, olNextLayer)) {
            const oNewLayer = this.fusionLayersWMS_(olLayer, olNextLayer);
            aLayers[i + 1] = oNewLayer;
          } else {
            optiLayers.push(olLayer);
          }
        } else {
          optiLayers.push(olLayer);
        }
      }
    }

    return optiLayers;
  }

  /**
   * Retourne la liste des couches défusionnées
   * @param olMap
   * @returns
   */
  public deoptimiseLayers(olMap: Map): Array<Layer<Source>> {
    const aLayers = [...olMap.getLayers().getArray()];
    const aDeoptimizedLayers = [];

    for (let i = 0; i < aLayers.length; i++) {
      const aFusionLayers = aLayers[i].get('fusion_layers');
      if (this.vitisCheck.isDefAndNotNull(aFusionLayers)) {
        for (const oFusionLayer of aFusionLayers) {
          aDeoptimizedLayers.push(oFusionLayer);
        }
      } else {
        aDeoptimizedLayers.push(aLayers[i]);
      }
    }

    return aDeoptimizedLayers;
  }

  /**
   * Change e DPI des couches Mapserver
   * @param aLayers
   * @param iDpi
   * @returns
   */
  public changeLayersDPI(aLayers: Array<Layer<ImageWMS | TileWMS>>, iDpi: number): Array<Layer<ImageWMS | TileWMS>> {
    for (let i = 0; i < aLayers.length; i++) {
      const olLayer = aLayers[i];
      if (olLayer.get('layerType') === 'postgres') {
        const oSourceParams = olLayer.getSource().getParams();
        oSourceParams['MAP.RESOLUTION'] = 72;
        oSourceParams['MAP.DEFRESOLUTION'] = iDpi;
      }
    }
    return aLayers;
  }

  /**
   * Retourne les couches sous forme de tableau
   * @param olMap
   * @returns
   */
  public getLayersArray(olMap: Map): any {
    const aTree = this.getLayersTree(olMap);
    const aLayers = this.recursiveGetLayerChildren_({
      children: aTree
    });

    return aLayers;
  }

  /**
   * LayersTree_ getter (also set the visibility of the layers with the new ones)
   *
   * @param olMap
   * @returns {any} Layers tree
   */
  private getLayersTree_(olMap: Map): any {

    const aLayers = olMap.getLayers().getArray();
    const oMapDef = olMap.get('map_definition');

    /**
     * Recursive function that update parameters recursively
     * @param {object} node
     */
    const updateOnTree = (node) => {
      // Recursive
      if (this.vitisCheck.isDefAndNotNull(node.children)) {
        for (let i = 0; i < node.children.length; i++) {
          updateOnTree(node.children[i]);
        }
      }
      // Update
      if (this.vitisCheck.isDefAndNotNull(node.olLayer)) {
        node.visible = node.olLayer.getVisible();
        node.index = aLayers.indexOf(node.olLayer);
        node.opacity = node.olLayer.getOpacity();
      }
    };

    for (let i = 0; i < oMapDef.children.length; i++) {
      updateOnTree(oMapDef.children[i]);
    }

    return oMapDef;
  }

  /**
 * Add the proxy to the wms/wmts... url
 * @param {string} url
 * @returns {String}
 */
  private parseProxyUrl_(url: string): string {

    if (!this.vitisCheck.isDefAndNotNull(url)) {
      return undefined;
    }

    const proxyUrl = environment.properties.proxy_url;

    url = url.replace(/\?/g, '&');

    const parsedProxyUrl = proxyUrl + '?url=' + url;

    return parsedProxyUrl;
  }

  /**
   * Créer un toast d'erreur traduit
   * @param {string} message
   */
  public notifyBasicError(message: string) {
    this.translate.get(message).subscribe((translation) => {
      vitisAlert.fireToast({
        icon: 'error',
        title: translation
      });
    });
  }

  /**
   * Set l'opacité d'une couche et gère les couches fusionnées
   * @param oLayer
   * @param olMap
   * @param fOpacity
   */
  public setLayerOpacity(oLayer: any, olMap: Map, fOpacity: number) {
    const iLayerId = oLayer.layer_id;
    const aMapLayers = olMap.getLayers().getArray();

    for (let i = 0; i < aMapLayers.length; i++) {
      // Si le layer id correspond à une couche normale
      if (aMapLayers[i].get('layer_id') === iLayerId) {
        aMapLayers[i].setOpacity(fOpacity);
        return;

        // Sinon, on vérifie si il y a une couche fusionnée, puis on regarde à l'intérieur
      } else {
        const aFusionLayers = aMapLayers[i].get('fusion_layers');

        if (this.vitisCheck.isDefAndNotNull(aFusionLayers)) {
          // Si la couche fusionnée contient la couche souhaitée
          if (aFusionLayers.some(oFL => oFL.get('layer_id') === iLayerId)) {
            // On récupère puis supprime la couche fusionnée actuelle
            const olFusionLayer: any = olMap.removeLayer(aMapLayers[i]);
            // On sépare la couche modifiée de la couche fusionnée, puis on renvoit les deux distinctement
            const aRefreshedLayers = this.refreshFusionLayersOpacity_(olFusionLayer, iLayerId, fOpacity).reverse();

            // On restaure la couche séparée et la couche re-fusionnée
            for (const refreshedLayer of aRefreshedLayers) {
              olMap.getLayers().insertAt(i, refreshedLayer);
            }
          }
        }
      }
    }
  }

  /**
   * Prend la liste ordonnée des ID des couches puis les réordonne sur l'objet Map
   * @param aNewLayersOrder
   * @param olMap
   */
  public setLayersOrder(aNewLayersOrder: Array<number>, olMap: Map) {
    const aDeoptimizedLayers = this.deoptimiseLayers(olMap);
    const aOrdenedLayers = [];

    for (const layerId of aNewLayersOrder) {
      for (const oLayer of aDeoptimizedLayers) {
        if (oLayer.get('layer_id') === layerId) {
          aOrdenedLayers.push(oLayer);
        }
      }
    }

    for (const oLayer of aDeoptimizedLayers) {
      if (aOrdenedLayers.indexOf(oLayer) === -1) {
        aOrdenedLayers.push(oLayer);
      }
    }

    const aOptimizedLayers = this.optimiseLayers(aOrdenedLayers);
    for (let i = 0; i < aOptimizedLayers.length; i++) {
      const olLayer = aOptimizedLayers[i];
      if (!(olLayer instanceof VectorLayer) || this.vitisCheck.isNumber(olLayer.get('layer_id'))) {
        olLayer.setZIndex(i + 1);
      }
    }

    olMap.getLayerGroup().setLayers(new Collection(aOptimizedLayers));
  }

  /**
   * Set une couche visible / non visible et gère les couches fusionnées
   * @param oLayer
   * @param olMap
   * @param bVisible
   * @returns
   */
  public setLayerVisible(oLayer: any, olMap: Map, bVisible: boolean) {
    const olLayer = oLayer.olLayer;
    const aMapLayers: any = olMap.getLayers().getArray();
    let visibilityChange = false;

    for (let i = 0; i < aMapLayers.length; i++) {

      const targetLayerId = olLayer.get('layer_id');
      const loopLayerId = aMapLayers[i].get('layer_id');
      const aFusionLayers = aMapLayers[i].get('fusion_layers');

      if (olLayer === aMapLayers[i] || (
        this.vitisCheck.isDefAndNotNull(targetLayerId) &&
        this.vitisCheck.isDefAndNotNull(loopLayerId) &&
        targetLayerId === loopLayerId
      )) {
        aMapLayers[i].setVisible(bVisible);
        visibilityChange = true;
      } else if (this.vitisCheck.isDefAndNotNull(aFusionLayers)) {
        for (let ii = 0; ii < aFusionLayers.length; ii++) {
          if (olLayer === aFusionLayers[ii]) {
            aFusionLayers[ii].setVisible(bVisible);
            this.mapVisibilityChange.next({ oLayer, bVisible });
            this.refreshFusionLayersVisible_(aMapLayers[i]);
          }
        }
      }
    }

    if (visibilityChange) {
      this.mapVisibilityChange.next({ oLayer, bVisible });
    }
  }

  /**
   * Add a Layer from a vector content
   * @param {string|object} content
   * @param {string} serviceName
   * @param {string} layerTitle
   * @returns {boolean}
   */
  public addLayerFromVectorContent(olMap: Map, content: any, serviceName: string, layerTitle: string): boolean {

    let sFeatures: string;
    let oFeaturesReaded: any;

    try {
      sFeatures = JSON.parse(content);
    } catch (e) {
      sFeatures = content;
    }

    // Récupère l'objet features depuis la chaine de caractères
    const formatConstructors = [
      GPX,
      GeoJSON,
      IGC,
      KML,
      TopoJSON
    ];

    // Essaye de lire la source celon les formats suivants
    for (let i = 0; i < formatConstructors.length; i++) {
      const format = new formatConstructors[i]();
      try {
        const features = format.readFeatures(sFeatures, {
          featureProjection: olMap.getView().getProjection()
        });
        // Si la lecture a réussi, alors oFeaturesReaded prend la valeur de la géométrie
        if (features.length > 0) {
          oFeaturesReaded = features;
          i = formatConstructors.length;
        }
      } catch (err) {
      }
    }

    if (this.vitisCheck.isDefAndNotNull(oFeaturesReaded)) {

      // Ajoute la couche
      const oLayer = {
        layerType: 'imagevector',
        serviceName: serviceName,
        layerTitle: layerTitle,
        features: oFeaturesReaded
      };
      this.addLayer(olMap, oLayer);
      return true;
    } else {
      this.notifyBasicError('VMAP_UNABLE_READ_GEOMETRY');
      console.error('Impossible de lire les géométries: ', content);
      return false;
    }
  }

  /**
   * Ajoute les projections non supportées par OL
   */
  public addCustomProjections() {

    const lambert93 = new Projection({
      code: 'EPSG:2154',
      // Cette extent est celle du niveau de zoom 0, on peut les trouver sur http://epsg.io/
      extent: [-378305.81, 6093283.21, 1212610.74, 7186901.68],
      units: 'm'
    });
    addProjection(lambert93);

    // Ajoute les fonctions de transformation de EPSG:4326 vers EPSG:2154
    addCoordinateTransforms('EPSG:4326', 'EPSG:2154', this.WGSToLambert, this.lambertToWGS);
    addCoordinateTransforms('EPSG:3857', 'EPSG:2154',
      function (coordinate3857: any) {
        const coordinate4326 = transform(coordinate3857, 'EPSG:3857', 'EPSG:4326');
        const coordinate2154 = transform(coordinate4326, 'EPSG:4326', 'EPSG:2154');
        return coordinate2154;
      },
      function (coordinate2154: any) {
        const coordinate4326 = transform(coordinate2154, 'EPSG:2154', 'EPSG:4326');
        const coordinate3857 = transform(coordinate4326, 'EPSG:4326', 'EPSG:3857');
        return coordinate3857;
      });
  }

  /**
   * Retourne une layer d'après son layer_id
   * @param iLayerId
   * @param oTree
   * @returns
   */
  public findLayerInTree(iLayerId: number, oTree: any): VmapLayer {
    const aLayers = this.recursiveGetLayerChildren_(oTree);
    for (let i = 0; i < aLayers.length; i++) {
      const oLayer: any = aLayers[i];
      if (oLayer.layer_id === iLayerId) {
        return oLayer;
      }
    }
  }

  /**
   * Ajoute une couche vecteur (qui ne sera pas gardée au rechargement de la carte)
   *
   * @param olMap
   * @param features
   * @param layerName
   */
  public addLayerVector(olMap: Map, features: any, layerName: string, zoomOnLayer: boolean): any {

    // Faux layer_id car la donnée est temporaire
    const iFakeId = Math.floor(Math.random() * (9999 - 1000 + 1)) + 1000;

    // ZIndex au dessus des autres
    const zIndex = olMap.getLayers().getLength();

    const vectorSource = new VectorSource({
      features: features,
    });
    const vectorLayer = new VectorLayer({
      source: vectorSource,
    });

    const oLayer: any = {
      olLayer: vectorLayer,
      visible: true,
      name: layerName,
      layerType: 'vector',
      layer_id: iFakeId,
    };

    // Utile pour faire apparaître la couche dans la liste des couches pour modifier l'emplacement
    vectorLayer.set('layer_id', iFakeId);
    vectorLayer.set('name', layerName);

    // Place le vector en top position
    vectorLayer.setZIndex(zIndex);
    olMap.addLayer(vectorLayer);
    const oDef = olMap.get('map_definition');
    oDef.children = [...oDef.children.slice(0, 1), oLayer, ...oDef.children.slice(1)];
    olMap.set('map_definition', oDef);
    olMap.changed();

    if (zoomOnLayer === true) {
      olMap.getView().fit(vectorSource.getExtent());
    }
  }

  /**
   * Tansform WGS83 coordinates to Lambert93
   * @param {ol.coordinates} coordinates WGS coordinates to transform
   * @returns {ol.coordinates} Lambert93 coordinates
   */
  private WGSToLambert(coordinates: Array<any>): Array<any> {

    const longitude = coordinates[0];
    const latitude = coordinates[1];
    const deg2rad = function (angle: any) {
      return (angle / 180) * Math.PI;
    };
    // variables:
    const a = 6378137; // demi grand axe de l'ellipsoide (m)
    const e = 0.08181919106; // première excentricité de l'ellipsoide
    const l0 = deg2rad(3);
    const lc = deg2rad(3);
    const phi0 = deg2rad(46.5); // latitude d'origine en radian
    const phi1 = deg2rad(44); // 1er parallele automécoïque
    const phi2 = deg2rad(49); // 2eme parallele automécoïque

    const x0 = 700000; // coordonnées à l'origine
    const y0 = 6600000; // coordonnées à l'origine

    const phi = deg2rad(latitude);
    const l = deg2rad(longitude);
    // calcul des grandes normales
    const gN1 = a / Math.sqrt(1 - e * e * Math.sin(phi1) * Math.sin(phi1));
    const gN2 = a / Math.sqrt(1 - e * e * Math.sin(phi2) * Math.sin(phi2));
    // calculs des latitudes isométriques
    const gl1 = Math.log(Math.tan(Math.PI / 4 + phi1 / 2) * Math.pow((1 - e * Math.sin(phi1)) / (1 + e * Math.sin(phi1)), e / 2));
    const gl2 = Math.log(Math.tan(Math.PI / 4 + phi2 / 2) * Math.pow((1 - e * Math.sin(phi2)) / (1 + e * Math.sin(phi2)), e / 2));
    const gl0 = Math.log(Math.tan(Math.PI / 4 + phi0 / 2) * Math.pow((1 - e * Math.sin(phi0)) / (1 + e * Math.sin(phi0)), e / 2));
    const gl = Math.log(Math.tan(Math.PI / 4 + phi / 2) * Math.pow((1 - e * Math.sin(phi)) / (1 + e * Math.sin(phi)), e / 2));
    // calcul de l'exposant de la projection
    const n = (Math.log((gN2 * Math.cos(phi2)) / (gN1 * Math.cos(phi1)))) / (gl1 - gl2);
    // calcul de la constante de projection
    const c = ((gN1 * Math.cos(phi1)) / n) * Math.exp(n * gl1);
    // calcul des coordonnées
    const ys = y0 + c * Math.exp(-1 * n * gl0);
    const x93 = x0 + c * Math.exp(-1 * n * gl) * Math.sin(n * (l - lc));
    const y93 = ys - c * Math.exp(-1 * n * gl) * Math.cos(n * (l - lc));
    return [x93, y93];
  }

  /**
   * Tansform Lambert93 coordinates to WGS83
   * @param {ol.coordinates} coordinates Lambert93 coordinates to transform
   * @returns {ol.coordinates} WGS84 coordinates
   */
  private lambertToWGS(coordinates: Array<any>): Array<any> {

    Math.atanh = Math.atanh || function (x_) {
      return Math.log((1 + x_) / (1 - x_)) / 2;
    };

    Math.tanh = Math.tanh || function (x_) {
      if (x_ === Infinity) {
        return 1;
      } else if (x_ === -Infinity) {
        return -1;
      } else {
        return (Math.exp(x_) - Math.exp(-x_)) / (Math.exp(x_) + Math.exp(-x_));
      }
    };

    const x = parseFloat(coordinates[0]).toFixed(10);
    const y = parseFloat(coordinates[1]).toFixed(10);
    const b6 = 6378137.0000;
    const b7 = 298.257222101;
    const b8 = 1 / b7;
    const b9 = 2 * b8 - b8 * b8;
    const b10 = Math.sqrt(b9);
    const b13 = 3.000000000;
    const b14 = 700000.0000;
    const b15 = 12655612.0499;
    const b16 = 0.7256077650532670;
    const b17 = 11754255.426096;
    const delx = parseFloat(x) - b14;
    const dely = parseFloat(y) - b15;
    const gamma = Math.atan(-(delx) / dely);
    const r = Math.sqrt((delx * delx) + (dely * dely));
    const latiso = Math.log(b17 / r) / b16;
    const sinphiit0 = Math.tanh(latiso + b10 * Math.atanh(b10 * Math.sin(1)));
    const sinphiit1 = Math.tanh(latiso + b10 * Math.atanh(b10 * sinphiit0));
    const sinphiit2 = Math.tanh(latiso + b10 * Math.atanh(b10 * sinphiit1));
    const sinphiit3 = Math.tanh(latiso + b10 * Math.atanh(b10 * sinphiit2));
    const sinphiit4 = Math.tanh(latiso + b10 * Math.atanh(b10 * sinphiit3));
    const sinphiit5 = Math.tanh(latiso + b10 * Math.atanh(b10 * sinphiit4));
    const sinphiit6 = Math.tanh(latiso + b10 * Math.atanh(b10 * sinphiit5));
    const longrad = gamma / b16 + b13 / 180 * Math.PI;
    const latrad = Math.asin(sinphiit6);
    const lon = (longrad / Math.PI * 180);
    const lat = (latrad / Math.PI * 180);
    return ([lon, lat]);
  }

  /**
   * Prépare la couche Tilewms ou Imagewms pour l'ajout
   * @param {any} oLayer
   * @param {any} oTree
   * @param {number} doesServiceExist
   * @param {any} newLayer
   * @param {Array<any>} aLayers
   * @returns {any}
   */
  private prepareLayerTilewmsImagewms(oLayer: any, oTree: any, doesServiceExist: number, newLayer: any, aLayers: Array<any>): any {
    // Test si la couche existe
    const aWMSServicesParams = [];

    // Mémorise les couches présentes
    for (let i = 0; i < aLayers.length; i++) {
      if (aLayers[i].get('type') === 'tilewms') {
        // TileWMS peut supporter plusieurs URL, si ce cas se présente on vérifie chacune des url
        const aUrls = aLayers[i].getSource().getUrls();
        for (let ii = 0; ii < aUrls.length; ii++) {
          aWMSServicesParams.push({
            url: aUrls[ii],
            layers: aLayers[i].getSource().getParams().LAYERS
          });
        }
      }
      if (aLayers[i].get('type') === 'imagewms') {

        aWMSServicesParams.push({
          url: aLayers[i].getSource().getUrl(),
          layers: aLayers[i].getSource().getParams().LAYERS
        });
      }
    }
    // Les compare à la couche ajoutée
    for (let i = 0; i < aWMSServicesParams.length; i++) {
      if (aWMSServicesParams[i].url === environment.properties.use_proxy_for_tiles ?
        this.parseProxyUrl_(oLayer.url) : oLayer.url && aWMSServicesParams[i].layers === oLayer.layerName) {
        this.notifyBasicError('VMAP_LAYER_ALREADY_PRESENT');
        return null;
      }
    }

    // Login/Password
    oLayer.service_options = '';
    if (this.vitisCheck.isString(oLayer.login) && this.vitisCheck.isString(oLayer.password)) {
      if (oLayer.login.trim().length > 0 && oLayer.password.trim().length > 0) {
        oLayer.service_options = JSON.stringify({
          'login': oLayer.login,
          'password': oLayer.password
        });
      }
    }

    // Ajout de la couche dans oTree
    if (doesServiceExist === 0) {
      newLayer = {
        'name': oLayer.serviceName,
        'children': [{
          'name': oLayer.layerTitle,
          'index': 1000000,
          'layerType': 'tilewms',
          'visible': 'true',
          'legend': 'true',
          'select': oLayer.select,
          'url': oLayer.url,
          'service_options': oLayer.service_options,
          'params': {
            'LAYERS': oLayer.layerName,
            'VERSION': oLayer.version,
            'STYLES': this.vitisCheck.isDefAndNotNull(oLayer.style) ? oLayer.style : ''
          }
        }]
      };
      oTree.children.push(newLayer);
    } else {
      newLayer = {
        'name': oLayer.layerTitle,
        'index': 1000000,
        'layerType': 'tilewms',
        'visible': 'true',
        'legend': 'true',
        'select': oLayer.select,
        'url': oLayer.url,
        'service_options': oLayer.service_options,
        'params': {
          'LAYERS': oLayer.layerName,
          'VERSION': oLayer.version,
          'STYLES': this.vitisCheck.isDefAndNotNull(oLayer.style) ? oLayer.style : ''
        }
      };

      oTree.children[doesServiceExist].children.push(newLayer);
    }
    return oTree;
  }

  /**
   * Prépare la couche OSM pour l'ajout
   * @param {any} oLayer
   * @param {any} oTree
   * @param {number} doesServiceExist
   * @param {any} newLayer
   * @returns {any}
   */
  private prepareLayerOSM(oLayer: any, oTree: any, doesServiceExist: number, newLayer: any): any {
    // Test si la couche est déjà incluse
    for (let i = 0; i < oTree.children.length; i++) {
      if (oTree.children[i].children) {
        for (let ii = 0; ii < oTree.children[i].children.length; ii++) {
          if (oTree.children[i].children[ii].url === oLayer.url) {
            this.notifyBasicError('VMAP_LAYER_ALREADY_PRESENT');
            return 0;
          }
          if (oLayer.url == '' && oTree.children[i].children[ii].url == 'https://{a-c}.tile.openstreetmap.org/{z}/{x}/{y}.png') {
            this.notifyBasicError('VMAP_LAYER_ALREADY_PRESENT');
            return 0;
          }
        }
      }
    }
    // Ajout de la couche dans oTree
    if (doesServiceExist === 0) {
      newLayer = {
        'name': oLayer.serviceName,
        'children': [{
          'index': 1000000,
          'name': oLayer.layerTitle,
          'visible': 'true',
          'layerType': 'osm',
          'url': oLayer.url
        }]
      };
      oTree.children.push(newLayer);
    } else {
      newLayer = {
        'name': oLayer.layerTitle,
        'index': 1000000,
        'visible': 'true',
        'layerType': 'osm',
        'url': oLayer.url
      };

      oTree.children[doesServiceExist].children.push(newLayer);
    }
    return oTree;
  }

  /**
   * Prépare la couche XYZ pour l'ajout
   * @param {any} oLayer
   * @param {any} oTree
   * @param {number} doesServiceExist
   * @param {any} newLayer
   * @returns {any}
   */
  private prepareLayerXYZ(oLayer: any, oTree: any, doesServiceExist: number, newLayer: any): any {
    // Test si la couche est déjà incluse
    for (let i = 0; i < oTree.children.length; i++) {
      if (oTree.children[i].children) {
        for (let ii = 0; ii < oTree.children[i].children.length; ii++) {
          if (oTree.children[i].children[ii].url === oLayer.url) {
            this.notifyBasicError('VMAP_LAYER_ALREADY_PRESENT');
            return 0;
          }
        }
      }
    }
    // Ajout de la couche dans oTree
    if (doesServiceExist === 0) {
      newLayer = {
        'name': oLayer.serviceName,
        'children': [{
          'index': 1000000,
          'name': oLayer.layerTitle,
          'visible': 'true',
          'layerType': 'xyz',
          'url': oLayer.url
        }]
      };
      oTree.children.push(newLayer);
    } else {
      newLayer = {
        'name': oLayer.layerTitle,
        'index': 1000000,
        'visible': 'true',
        'layerType': 'xyz',
        'url': oLayer.url
      };

      oTree.children[doesServiceExist].children.push(newLayer);
    }
    return oTree;
  }

  /**
   * Prépare la couche Bing pour l'ajout
   * @param {any} oLayer
   * @param {any} oTree
   * @param {number} doesServiceExist
   * @param {any} newLayer
   * @returns {any}
   */
  private prepareLayerBing(oLayer: any, oTree: any, doesServiceExist: number, newLayer: any): any {
    // Test si la couche est déjà incluse
    for (let i = 0; i < oTree.children.length; i++) {
      if (oTree.children[i].children) {
        for (let ii = 0; ii < oTree.children[i].children.length; ii++) {
          if (oTree.children[i].children[ii].imagerySet === oLayer.layerName) {
            this.notifyBasicError('VMAP_LAYER_ALREADY_PRESENT');
            return 0;
          }
        }
      }
    }
    // Ajout de la couche dans oTree
    if (doesServiceExist === 0) {
      newLayer = {
        'name': oLayer.serviceName,
        'children': [{
          'name': oLayer.layerName,
          'index': 1000000,
          'imagerySet': oLayer.layerName,
          'visible': 'true',
          'layerType': 'bing',
          'key': oLayer.key,
          'culture': oLayer.culture
        }]
      };
      oTree.children.push(newLayer);
    } else {
      newLayer = {
        'name': oLayer.layerName,
        'index': 1000000,
        'imagerySet': oLayer.layerName,
        'visible': 'true',
        'layerType': 'bing',
        'key': oLayer.key,
        'culture': oLayer.culture
      };

      oTree.children[doesServiceExist].children.push(newLayer);
    }
    return oTree;
  }

  /**
   * Prépare la couche WMTS pour l'ajout
   * @param {any} oLayer
   * @param {any} oTree
   * @param {number} doesServiceExist
   * @param {any} newLayer
   * @returns {any}
   */
  private prepareLayerWMTS(oLayer: any, oTree: any, doesServiceExist: number, newLayer: any): any {
    // Test si la couche est déjà incluse
    for (let i = 0; i < oTree.children.length; i++) {
      if (oTree.children[i].children) {
        for (let ii = 0; ii < oTree.children[i].children.length; ii++) {
          if (oTree.children[i].children[ii].layer === oLayer.layerName) {
            this.notifyBasicError('VMAP_LAYER_ALREADY_PRESENT');
            return 0;
          }
        }
      }
    }
    // Ajout de la couche dans oTree
    if (doesServiceExist === 0) {
      newLayer = {
        'name': oLayer.serviceName,
        'children': [{
          'name': oLayer.layerTitle,
          'index': 1000000,
          'layer': oLayer.layerName,
          'visible': 'true',
          'layerType': 'wmts',
          'matrixSet': oLayer.matrixSet,
          'style': oLayer.style,
          'format': oLayer.format,
          'requestEncoding': oLayer.requestEncoding,
          'service_options': oLayer.service_options,
          'version': oLayer.version,
          'url': oLayer.url
        }]
      };
      oTree.children.push(newLayer);
    } else {
      newLayer = {
        'name': oLayer.layerTitle,
        'index': 1000000,
        'layer': oLayer.layerName,
        'visible': 'true',
        'layerType': 'wmts',
        'matrixSet': oLayer.matrixSet,
        'style': oLayer.style,
        'format': oLayer.format,
        'requestEncoding': oLayer.requestEncoding,
        'service_options': oLayer.service_options,
        'version': oLayer.version,
        'url': oLayer.url
      };

      oTree.children[doesServiceExist].children.push(newLayer);
    }
    return oTree;
  }

  /**
   * Prépare la couche ImageVector pour l'ajout
   * @param {any} oLayer
   * @param {any} oTree
   * @param {number} doesServiceExist
   * @param {any} newLayer
   * @returns {any}
   */
  private prepareLayerImageVector(oLayer: any, oTree: any, doesServiceExist: number, newLayer: any): any {
    // Test si la couche est déjà incluse
    for (let i = 0; i < oTree.children.length; i++) {
      if (oTree.children[i].children) {
        for (let ii = 0; ii < oTree.children[i].children.length; ii++) {
          if (oTree.children[i].children[ii].layerType === 'imagevector') {
            if (oTree.children[i].children[ii].url !== undefined) {
              if (oTree.children[i].children[ii].url === environment.properties.use_proxy_for_tiles ?
                this.parseProxyUrl_(oLayer.url) : oLayer.url) {
                this.notifyBasicError('VMAP_LAYER_ALREADY_PRESENT');

                return 0;
              }
            }
          }
        }
      }
    }
    // Ajout de la couche dans oTree
    if (doesServiceExist === 0) {
      newLayer = {
        'name': oLayer.serviceName,
        'children': [{
          'name': oLayer.layerTitle,
          'index': 1000000,
          'visible': 'true',
          'layerType': 'imagevector',
          'url': oLayer.url,
          'features': oLayer.features,
          'style': oLayer.style
        }]
      };
      oTree.children.push(newLayer);
    } else {
      newLayer = {
        'name': oLayer.layerTitle,
        'index': 1000000,
        'visible': 'true',
        'layerType': 'imagevector',
        'url': oLayer.url,
        'features': oLayer.features,
        'style': oLayer.style
      };
      oTree.children[doesServiceExist].children.push(newLayer);
    }
    return oTree;
  }

  /**
   * Définit si deux couches sont fusionnables (car elles utilisent le même flux)
   * @param olLayer1
   * @param olLayer2
   * @returns
   */
  private areLayersFusionable_(olLayer1: any, olLayer2: any): boolean {

    const olSource1 = olLayer1.getSource();
    const olSource2 = olLayer2.getSource();

    // Vérification type sources
    if (!(olSource1 instanceof ImageWMS && olSource2 instanceof ImageWMS)) {
      return false;
    }

    // URL
    if (!(this.vitisCheck.isFunction(olSource1.getUrl) && this.vitisCheck.isFunction(olSource2.getUrl))) {
      return false;
    }
    if (olSource1.getUrl() !== olSource2.getUrl()) {
      return false;
    }

    // Params
    const oSourceParams1 = olSource1.getParams();
    const oSourceParams2 = olSource2.getParams();
    for (const key in oSourceParams2) {
      if (key !== 'LAYERS') {
        if (Object.prototype.hasOwnProperty.call(oSourceParams1, key) &&
          Object.prototype.hasOwnProperty.call(oSourceParams2, key)) {
          if (oSourceParams1[key] !== oSourceParams2[key]) {
            return false;
          }
        } else {
          return false;
        }
      }
    }

    // Visible
    if (olLayer1.get('visible') !== olLayer2.get('visible')) {
      return false;
    }

    // Opacity
    if (olLayer1.get('opacity') !== olLayer2.get('opacity')) {
      return false;
    }

    // Vérification values
    const oValues1 = olLayer1.values_;
    const oValues2 = olLayer2.values_;
    const acceptableDifKeys = ['source', 'name', 'layer_id', 'sublayers', 'activeSublayers', 'fusion_layers'];

    // Vérifie que les valeurs de 1 soient dans 2
    for (const key in oValues1) {
      if (acceptableDifKeys.indexOf(key) === -1) {
        if (Object.prototype.hasOwnProperty.call(oValues1, key) &&
          !this.vitisCheck.isUndefined(oValues1[key])) {
          if (!Object.prototype.hasOwnProperty.call(oValues2, key) ||
            this.vitisCheck.isUndefined(oValues2[key])) {
            return false;
          }
          if (oValues1[key] !== oValues2[key]) {
            return false;
          }
        }
      }
    }

    // Vérifie que les valeurs de 2 soient dans 1
    for (const key in oValues2) {
      if (acceptableDifKeys.indexOf(key) === -1) {
        if (Object.prototype.hasOwnProperty.call(oValues2, key) &&
          !this.vitisCheck.isUndefined(oValues2[key])) {
          if (!Object.prototype.hasOwnProperty.call(oValues1, key) ||
            this.vitisCheck.isUndefined(oValues1[key])) {
            return false;
          }
          if (oValues2[key] !== oValues1[key]) {
            return false;
          }
        }
      }
    }

    return true;
  }

  /**
   * Fusionne olLayer2 dans olLayer1
   * @param olLayer1
   * @param olLayer2
   */
  private fusionLayersWMS_(olLayer1: any, olLayer2: any): Layer<ImageWMS | TileWMS> {

    const olSource1 = olLayer1.getSource();
    const olSource2 = olLayer2.getSource();

    let oNewSource: any;
    const oNewSourceOptions = {
      params: {
        ...olSource1.getParams(),
        ...{
          LAYERS: `${olSource1.getParams().LAYERS},${olSource2.getParams().LAYERS}`
        }
      },
      projection: olLayer1.getSource().getProjection(),
      url: olLayer1.getSource().getUrl(),
    };

    if (olSource1 instanceof ImageWMS) {
      oNewSource = new ImageWMS(oNewSourceOptions);
    }
    if (olSource1 instanceof TileWMS) {
      oNewSource = new TileWMS(oNewSourceOptions);
    }


    if (oNewSource) {
      const olNewLayer: any = new ImageLayer({
        source: oNewSource,
        opacity: olLayer1.getOpacity(),
        visible: olLayer1.getVisible(),
        extent: olLayer1.getExtent(),
        zIndex: olLayer1.getZIndex(),
        minResolution: olLayer1.getMinResolution(),
        maxResolution: olLayer1.getMaxResolution(),
        // minZoom: olLayer1.getMinZoom(),
        // maxZoom: olLayer1.getMaxZoom(),
      });

      // Set variable fusion_layers
      let aFusionLayers = [];
      if (olLayer1.get('fusion_layers')) {
        aFusionLayers = [...aFusionLayers, ...olLayer1.get('fusion_layers')];
      } else {
        aFusionLayers.push(olLayer1);
      }
      if (olLayer2.get('fusion_layers')) {
        aFusionLayers = [...aFusionLayers, ...olLayer2.get('fusion_layers')];
      } else {
        aFusionLayers.push(olLayer2);
      }
      olNewLayer.set('fusion_layers', aFusionLayers);


      // Set autres variables
      const oValues1 = olLayer1.values_;
      const oValues2 = olLayer2.values_;
      const acceptableDifKeys = ['source', 'name', 'layer_id', 'sublayers', 'activeSublayers', 'fusion_layers'];
      for (const key in oValues1) {
        if (acceptableDifKeys.indexOf(key) === -1) {
          if (Object.prototype.hasOwnProperty.call(oValues1, key)) {
            const val = oValues1[key];
            olNewLayer.set(key, oValues1[key]);
          }
        }
      }
      for (const key in oValues2) {
        if (acceptableDifKeys.indexOf(key) === -1) {
          if (Object.prototype.hasOwnProperty.call(oValues2, key)) {
            const val = oValues2[key];
            olNewLayer.set(key, oValues2[key]);
          }
        }
      }

      olNewLayer.values = olNewLayer.values_;
      return olNewLayer;
    }
  }

  /**
   * Rafraichit les couches dans une fusion de couches ayant changé d'opacité
   * @param olLayer
   * @param iTargetLayerId
   * @param fOpacity
   */
  private refreshFusionLayersOpacity_(olLayer: Layer<Source>, iTargetLayerId: number, fOpacity: number) {
    let aRefreshedLayers = [];
    const aFusionLayers = olLayer.get('fusion_layers');

    if (this.vitisCheck.isDefAndNotNull(aFusionLayers)) {
      // On parcourt les couches fusionnées jusqu'à trouver la couche que l'on souhaite modifier
      for (let i = 0; i < aFusionLayers.length; i++) {
        if (aFusionLayers[i].get('layer_id') === iTargetLayerId) {
          aFusionLayers[i].setOpacity(fOpacity);
        }
      }

      // On réoptimise ces couches, ce qui va séparer la couche modifiée des autres
      aRefreshedLayers = this.optimiseLayers(aFusionLayers);
    }

    return aRefreshedLayers;
  }

  /**
   * Rafraichit les couches visibles dans une fusion de couches
   * @param olLayer
   */
  private refreshFusionLayersVisible_(olLayer: Layer<any>) {
    let aRefreshedLayers = [];
    const aFusionLayers = olLayer.get('fusion_layers');
    if (this.vitisCheck.isDefAndNotNull(aFusionLayers)) {
      for (let i = 0; i < aFusionLayers.length; i++) {
        const aChildLayer = aFusionLayers[i];
        if (aChildLayer.getVisible()) {
          const aChildLayerLayers = aFusionLayers[i].getSource().getParams().LAYERS.split(',');
          aRefreshedLayers = [...aRefreshedLayers, ...aChildLayerLayers];
        }

      }
    }

    if (aRefreshedLayers.length > 0) {
      const oSourceParams = olLayer.getSource().getParams();
      oSourceParams.LAYERS = aRefreshedLayers.join(',');
      olLayer.getSource().updateParams(oSourceParams);
      olLayer.setVisible(true);
    } else {
      olLayer.setVisible(false);
    }
  }

  /**
   * Fonction récursive permettant de sortir les couches de l'arbre
   * @param oNode
   * @returns
   */
  private recursiveGetLayerChildren_(oNode: any): any {
    let aLayers = [];

    if (oNode.layer_id) {
      aLayers.push(oNode);
    } else if (this.vitisCheck.isArray(oNode.children)) {
      for (let i = 0; i < oNode.children.length; i++) {
        const childLayers = this.recursiveGetLayerChildren_(oNode.children[i]);
        if (this.vitisCheck.isArray(childLayers)) {
          aLayers = [...aLayers, ...childLayers];
        }
      }
    }

    return aLayers;
  }

  /**
   * Parfois les couches vectortiles sont composées de deux couches OL3
   * Cette fonction regroupe les doublons vectortiles
   *
   * @param aLayers
   * @returns
   */
  private joinVectorTileslayers_(aLayers: Array<Layer<Source>>): Array<Layer<Source>> {
    const aReturn = [];

    for (let i = 0; i < aLayers.length; i++) {
      const oLayer = aLayers[i];

      const existingLayer = aReturn.find((l) => {
        return l.get('layer_id') === oLayer.get('layer_id');
      });

      if (existingLayer) {
        const vectorFustionLayers = existingLayer.get('vector_fusion_layers') || [];
        vectorFustionLayers.push(oLayer);
        existingLayer.set('vector_fusion_layers', vectorFustionLayers);
      } else {
        aReturn.push(oLayer);
      }
    }

    return aReturn;
  }
}
