import { Injectable } from '@angular/core';

// Vitis
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';
import { MapCoreUtilsService } from '../map-core-utils/map-core-utils.service';

import Map from 'ol/Map.js';
import Overlay from 'ol/Overlay.js';
import Feature from 'node_modules/ol/Feature';
import { Geometry } from 'ol/geom';
import { environment } from 'src/environments/environment';
import { IsSmallScreenService } from 'src/app/vitis/core/services/utils/is-small-screen/is-small-screen.service';

declare let $: any;

class PopupOptions {
  bo?: any;
  buttons?: Array<PopupOptionsBtn>;
  hideCloseBtn?: boolean;
  clickCoords?: Array<number>;
  displayFeature?: boolean; // Si false n'affiche pas la feature
  resolutionCoeff?: number;
}

class PopupOptionsBtn {
  type: string;
  src: string;
  event: any;
  position: string;
  text: string;
  class: string;
}

class OptionSelect {
  value: any;
  text: string;
}


export class Popup {

  private vitisCheck: VitisCheck;
  private mapUtils: MapCoreUtilsService;
  private popupOverlay: Overlay;
  private olMap: Map;
  private oInfos: any;
  private olFeature: Feature<Geometry>;
  private oBo: any;
  private oOptions: PopupOptions = {};

  private domHTML_: any;
  private domHTMLContainer_: any;
  private domHTMLButtonsContainer_: any;
  private domHTMLCloser_: any;
  private domHTMLContent_: any;

  private isSmallScreen;

  /**
   * Constructeur
   */
  constructor(
    vitisCheck: VitisCheck,
    mapUtils: MapCoreUtilsService,
    oInfos: any,
    olFeature: Feature<Geometry>,
    olMap: Map,
    options?,
    isSmallScreen?
  ) {

    this.vitisCheck = vitisCheck;
    this.mapUtils = mapUtils;
    this.olMap = olMap;
    this.oInfos = oInfos;
    this.olFeature = olFeature;
    this.isSmallScreen = isSmallScreen;

    // Ajout de la feature à la carte
    const olLocationPopupLayer = this.olMap.get('popup_layer');
    if (options && options.displayFeature !== false &&
      this.vitisCheck.isDefAndNotNull(olLocationPopupLayer) &&
      !olLocationPopupLayer.getSource().getFeatures().includes(this.olFeature)) {
      olLocationPopupLayer.getSource().addFeature(olFeature);
    }

    if (this.vitisCheck.isDefAndNotNull(options)) {
      this.oOptions = options;
      if (this.vitisCheck.isDefAndNotNull(options.bo)) {
        this.oBo = options.bo;
      }
    }

    // Étiquette
    if (this.vitisCheck.isDefAndNotNull(this.oInfos)) {

      // Crée les éléments dom HTML
      this.createPopupElement_(this.vitisCheck.isObject(this.oInfos));

      // Crée une overlay qui contiendra la popup
      this.createPopupOverlay_();

      if (this.vitisCheck.isObject(this.oInfos)) {
        // Ajoute les infos sous forme de table
        this.displayMessageTable(this.oInfos);
      } else {
        // Ajoute les infos sous forme de message
        this.displayMessage(this.oInfos);
      }

      // Ajout des boutons spécifiques
      if (this.vitisCheck.isDefAndNotNull(options)) {
        if (this.vitisCheck.isArray(options.buttons)) {
          for (let i = 0; i < options.buttons.length; i++) {
            this.addButton(options.buttons[i]);
          }
        }
      }

      // Événement de fermeture
      if (this.domHTMLCloser_) {
        this.domHTMLCloser_.addEventListener('click', () => {
          this.close();
        });
      }
    }
  }

  /**
   * Affiche un message de type string ou dom dans la popup
   * @param  {any} message
   */
  public displayMessage(message: any) {
    this.removeMessages();
    this.addMessage(message);
  }

  /**
   * Affiche le contenu de l'objet data sous forme de tableau dans la popup
   * @param  {any} data
   */
  public displayMessageTable(data: any) {
    this.removeMessages();
    this.addMessageTable(data);
  }

  /**
   * Supprime les messages de la popup
   */
  public removeMessages() {
    this.domHTMLContent_.innerHTML = '';
  }

  /**
   * Ajoute un message dans la popup
   * @param  {any} message
   */
  public addMessage(message: any) {
    this.domHTMLContent_.innerHTML = message;
  }

  /**
   * Ajoute un message de type table dans la popup
   * @param  {any} data
   */
  public addMessageTable(data: any) {

    if (!this.vitisCheck.isDefAndNotNull(data)) {
      console.error('MapPopup: data not defined');
      return 0;
    }

    const HTMLMessageContainer = document.createElement('table');

    for (const item in data) {

      // car feature est le mot clé utilisé pour stocker la feature dans infocontainer
      if (item === 'feature') {
        continue;
      }

      const row = document.createElement('tr');

      if (item === 'select') { // Si l'on souhaite ajouter un select
        const columnSelect = document.createElement('td');
        columnSelect.className = 'map-popup-cell';
        const selectOptions = data[item];
        // Taille du select dans le tableau par défaut taille 2 car il y a 2 colonnes value et text
        columnSelect.colSpan = selectOptions.colspan || 2;

        // Création du select
        const select = document.createElement('select');
        // Assignement des options avec text et value
        selectOptions.options.forEach((option: OptionSelect) => {
          const optionElm = document.createElement('option');
          optionElm.value = option.value;
          optionElm.text = option.text;
          select.add(optionElm, null);
        });
        // Assignement de l'event appelé lors du changement d'options
        if (this.vitisCheck.isDefAndNotNull(selectOptions.event)) {
          select.onchange = selectOptions.event;
        }

        // Ajoute une valeur au select
        if (this.vitisCheck.isDefAndNotNull(selectOptions.value)) {
          select.value = selectOptions.value;
        }

        // Ajout du select au tableau
        columnSelect.appendChild(select);
        row.appendChild(columnSelect);
      } else {
        const column1 = document.createElement('td');
        column1.className = 'map-popup-cell';
        row.appendChild(column1);

        const column2 = document.createElement('td');
        column2.className = 'map-popup-cell';
        row.appendChild(column2);

        column1.innerHTML = item;
        column2.innerHTML = data[item];
      }

      HTMLMessageContainer.appendChild(row);
    }

    this.domHTMLContent_.appendChild(HTMLMessageContainer);
    return true;
  }

  /**
   * Ajout de boutons spéciaux
   * @param  {any} oButton
   */
  public addButton(oButton: any) {

    // Défintion du bouton
    const domBtn = document.createElement('div');

    // Élément sur lequel se declenche l'action
    let triggerElem: HTMLDivElement | HTMLButtonElement = domBtn;

    // Type image
    if (oButton.type === 'image') {
      const domBtnImg = document.createElement('img');
      domBtnImg.classList.add('ol-popup-img-btn');
      if (this.vitisCheck.isString(oButton.src)) {
        domBtnImg.src = oButton.src;
      }
      if (this.vitisCheck.isFunction(oButton.src)) {
        domBtnImg.src = oButton.src.call(this, this.oBo);
      }

      // Curseur de type pointer si un événement est associé
      if (oButton.event) {
        domBtnImg.style.cursor = 'pointer';
      }

      // Ajoute l'image
      domBtn.appendChild(domBtnImg);

      // C'est l'image qui déclenchera l'événement
      triggerElem = domBtnImg;
    }

    // Type Button
    if (oButton.type === 'btn') {
      const domBtnBtn = document.createElement('button');
      this.domHTMLContainer_.classList.add('ol-popup-large');
      domBtnBtn.classList.add('btn');
      domBtnBtn.classList.add(oButton.class);
      domBtnBtn.innerText = oButton.text;
      // Curseur de type pointer si un événement est associé
      if (oButton.event) {
        domBtnBtn.style.cursor = 'pointer';
      }

      // Ajoute le bouton
      domBtn.appendChild(domBtnBtn);

      // C'est le bouton qui déclenchera l'événement
      triggerElem = domBtnBtn;
    }

    // Container du bouton
    if (oButton.position === 'right') {
      const domBtnsContainer = document.createElement('div');
      domBtnsContainer.classList.add('col-4');
      domBtnsContainer.classList.add('ol-btn-right-vertical-align');
      this.domHTMLContainer_.classList.add('row');
      this.domHTMLContent_.classList.add('col-8');
      domBtnsContainer.appendChild(domBtn);
      this.domHTMLContainer_.appendChild(domBtnsContainer);
    }

    // Event
    if (oButton.event) {
      triggerElem.addEventListener('click', () => {
        oButton.event.call(this, this.oBo);
      });
    }
  }

  /**
   * Ferme la popup
   */
  public close() {
    // Suppression de la feature de la carte
    const olLocationPopupLayer = this.olMap.get('popup_layer');
    if (this.vitisCheck.isDefAndNotNull(olLocationPopupLayer) &&
      olLocationPopupLayer.getSource().getFeatures().includes(this.olFeature)) {
      olLocationPopupLayer.getSource().removeFeature(this.olFeature);
    }

    if (this.vitisCheck.isDefAndNotNull(this.oInfos)) {
      if (this.domHTMLCloser_) {
        this.domHTMLCloser_.blur();
      }
      this.popupOverlay.setPosition(undefined);
    }
  }

  /**
   * Crée les éléments dom HTML
   */
  private createPopupElement_(bTableType = true) {

    // Popup
    this.domHTML_ = document.createElement('div');
    this.domHTML_.className = 'ol-popup';

    if (bTableType !== true) {
      this.domHTML_.className += ' one-line';
    }

    // Container
    this.domHTMLContainer_ = document.createElement('div');
    this.domHTML_.appendChild(this.domHTMLContainer_);

    // Boutons
    this.domHTMLButtonsContainer_ = document.createElement('div');
    this.domHTMLButtonsContainer_.className = 'ol-popup-buttons-container';
    this.domHTMLContainer_.appendChild(this.domHTMLButtonsContainer_);

    // Bouton close
    if (this.oOptions.hideCloseBtn !== true) {
      this.domHTMLCloser_ = document.createElement('a');
      this.domHTMLCloser_.className = 'ol-popup-closer icon-vitis-cross2';
      this.domHTMLButtonsContainer_.appendChild(this.domHTMLCloser_);
    }

    // Contenu
    this.domHTMLContent_ = document.createElement('div');
    this.domHTMLContent_.classList.add('popup-content');
    this.domHTMLContainer_.appendChild(this.domHTMLContent_);
  }

  /**
   * Crée une overlay qui contiendra la popup
   */
  private createPopupOverlay_() {
    const position = this.mapUtils.getGeometryCenter(this.olFeature.getGeometry());

    // Overlay contenant la popup
    this.popupOverlay = new Overlay(({
      element: this.domHTML_,
      autoPan: true,
      autoPanAnimation: {
        duration: 250
      },
      positioning: 'bottom-center'
    }));

    const resolutionCoeff = this.oOptions.resolutionCoeff ?? 1;

    // Met à jour la taille/position de la popup en fonction de la résolution
    this.domHTML_.style.transform = `scale(${resolutionCoeff})`;
    const marginBottom = 2 * (resolutionCoeff - 1);
    if (this.olFeature.getGeometry().getType() === 'Point') {
      this.domHTML_.style.marginBottom = `calc(${marginBottom + (marginBottom / 2) + (0.5 * resolutionCoeff)}rem)`;
    } else {
      this.domHTML_.style.marginBottom = `calc(${marginBottom}rem)`;
    }
    this.olMap.addOverlay(this.popupOverlay);
    this.popupOverlay.setOffset([1 * (resolutionCoeff - 1), 0]);
    this.popupOverlay.setPosition(position);
  }

  /**
   * Met à jour le contenu de la popup
   * @param oInfos
   */
  public setPopupMessage(oInfos: any) {
    if (this.vitisCheck.isObject(oInfos)) {
      // Set les valeurs dont les key apparaissent dans oInfos
      this.setMessageTable(oInfos);
    } else {
      // Reset le message
      this.displayMessage(oInfos);
    }
  }

  /**
   * Met à jour le message de la popup qui est sous forme d'un tableau
   * Remarque : Seul les key dans le data se mettront à jour
   * @param data
   * @returns
   */
  public setMessageTable(data: any) {
    if (!this.vitisCheck.isDefAndNotNull(data)) {
      console.error('MapPopup: data not defined');
      return 0;
    }

    const tdElements = this.domHTMLContent_.getElementsByTagName('td');
    const keysToUpdate = Object.keys(data);
    for (let i = 0; i < tdElements.length; i++) {
      if (keysToUpdate.includes(tdElements[i].innerHTML)) {
        tdElements[i + 1].innerHTML = data[tdElements[i].innerHTML];
      }
    }
  }

  /**
   * Modifier la position de la popup
   * @param position
   */
  public setPopupPosition(geometry: Geometry) {
    const position = this.mapUtils.getGeometryCenter(geometry);
    this.popupOverlay.setPosition(position);
  }
}

@Injectable({
  providedIn: 'root'
})
export class MapCorePopupService {

  private aCurrentPopups: Array<Popup> = [];
  private isSmallScreen;

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private mapUtils: MapCoreUtilsService,
    private isSmallScreenService: IsSmallScreenService
  ) {
    this.isSmallScreen = this.isSmallScreenService.getIsSmallScreen();
    this.isSmallScreenService.isSmallScreenObservable.subscribe(isSmallScreen => {
      this.isSmallScreen = isSmallScreen;
    });
  }

  /**
   * Instancie et renvoie une popup
   * @param  {any} oInfos
   * @param  {Feature} olFeature
   * @param  {Map} olMap
   * @param  {any} options?
   */
  public displayPopup(oInfos: any, olFeature: Feature<Geometry>, olMap: Map, options?: any) {
    if (olFeature && olFeature.getGeometry()) {
      const oPopup = new Popup(this.vitisCheck, this.mapUtils, oInfos, olFeature, olMap, options, this.isSmallScreen);
      this.aCurrentPopups.push(oPopup);
      return oPopup;
    }
  }

  /**
   * Ferme les popups en cours
   */
  public closePopups() {
    for (let i = 0; i < this.aCurrentPopups.length; i++) {
      const oPopup = this.aCurrentPopups[i];
      oPopup.close();
    }
  }
}
