import { Injectable } from '@angular/core';

// Vitis
import { VitisCheck, VitisRequest, VitisClone } from 'src/app/vitis/core/services/utils/vitis-utils';
import { environment } from 'src/environments/environment';

// OpenLayers
import Map from 'ol/Map.js';
import Geometry from 'ol/geom/Geometry';
import Point from 'node_modules/ol/geom/Point.js';
import WKT from 'ol/format/WKT';
import GeoJSON from 'node_modules/ol/format/GeoJSON';
import { getDistance } from 'ol/sphere.js';
import { transform } from 'node_modules/ol/proj.js';
import Feature from 'node_modules/ol/Feature';
import { Circle as CircleStyle, Fill, Stroke, Style, Text, Icon } from 'ol/style.js';
import { StyleJSON } from '../map-models.model';
import LinearRing from 'ol/geom/LinearRing';
import LineString from 'ol/geom/LineString';
import MultiLineString from 'ol/geom/MultiLineString';
import MultiPoint from 'ol/geom/MultiPoint';
import MultiPolygon from 'ol/geom/MultiPolygon';
import Polygon from 'ol/geom/Polygon';
import { GeometryCollection } from 'ol/geom';
import { Coordinate } from 'ol/coordinate';

@Injectable({
  providedIn: 'root'
})
export class MapCoreUtilsService {

  /**
   * Constructeur
   */
  constructor(
    private vitisClone: VitisClone,
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest
  ) { }

  /**
  * Retourne la géométrie au format EWKT
  * @param  {Point} geom
  * @param  {Map} olMap
  * @returns string
  */
  public getEWKTFromGeom(geom: Geometry, olMap: Map): string {

    const proj: string = olMap.getView().getProjection().getCode();

    if (!this.vitisCheck.isDefAndNotNull(geom)) {
      console.error('geom is not defined');
      return null;
    }
    if (!this.vitisCheck.isString(proj)) {
      console.error('proj is not a string');
      return null;
    }
    if (!(proj.indexOf('EPSG:') === 0)) {
      console.error('proj is not an EPSG code');
      return null;
    }

    const oWKT = new WKT();
    const SRID = proj.substring(5);

    let oWKTGeom = oWKT.writeGeometry(geom, {
      dataProjection: proj,
      featureProjection: proj
    });

    // Corrections pour la 3D car en EWKT il ne faut pas de "Z"
    oWKTGeom = oWKTGeom.substring(0, oWKTGeom.indexOf('(')).replace(' Z', '') + oWKTGeom.substring(oWKTGeom.indexOf('('));

    const sEWKTGeom = 'SRID=' + SRID + ';' + oWKTGeom;
    return sEWKTGeom;
  }

  /**
  * Retourne la géométrie au format EWKT en changeant sa projection
  * @param  {Point} geom
  * @param  {Map} olMap
  * @param proj
  * @returns string
  */
  public getEWKTFromGeomAndProj(geom: any, olMap: Map, transformProj): string {

    const proj: string = 'EPSG:' + transformProj;

    const oWKT = new WKT();
    const SRID = proj.substring(5);

    let oWKTGeom = oWKT.writeGeometry(geom.transform(olMap.getView().getProjection(), proj), {
      dataProjection: proj,
      featureProjection: proj
    });

    // Corrections pour la 3D car en EWKT il ne faut pas de "Z"
    oWKTGeom = oWKTGeom.substring(0, oWKTGeom.indexOf('(')).replace(' Z', '') + oWKTGeom.substring(oWKTGeom.indexOf('('));

    const sEWKTGeom = 'SRID=' + SRID + ';' + oWKTGeom;

    return sEWKTGeom;
  }

  /**
   * Transforme une géométrie EWKT en géométrie OpenLayers
   * @param  {string} sEWKTGeom
   * @param  {Map} olMap
   * @returns Geometry
   */
  public getGeomFromEWKT(sEWKTGeom: string, olMap: Map): any {

    if (!this.vitisCheck.isString(sEWKTGeom)) {
      console.error('sEWKTGeom is not a string: ', sEWKTGeom);
      return null;
    }

    if (this.isEWKTGeom(sEWKTGeom)) {

      const oWKTFormat = new WKT();
      const oWKT = this.getWKTFromEWKT(sEWKTGeom);
      const olWKTGeom = oWKT.geom;
      const sEWKTGeomProj = oWKT.proj;
      const sMapProj = olMap.getView().getProjection().getCode();

      const oGeom: any = oWKTFormat.readGeometry(olWKTGeom, {
        dataProjection: sEWKTGeomProj,
        featureProjection: sMapProj
      });

      if (oGeom &&
        oGeom.getCoordinates &&
        oGeom.getCoordinates() &&
        this.isValidCoordinates(oGeom.getCoordinates())) {
        return oGeom;
      } else {
        console.error('Geom is not an EWKT Geometry: ', sEWKTGeom);
        return null;
      }
    } else {
      console.error('Geom is not an EWKT Geometry: ', sEWKTGeom);
      return null;
    }
  }

  /**
   * Transforme une géométrie EWKT en géométrie OpenLayers
   * @param  {string} sEWKTGeom
   * @param  {Map} olMap
   * @returns Geometry
   */
  public getGeomFromEWKTAndProj(sEWKTGeom: string, projection?): any {

    if (!this.vitisCheck.isString(sEWKTGeom)) {
      console.error('sEWKTGeom is not a string: ', sEWKTGeom);
      return null;
    }

    if (this.isEWKTGeom(sEWKTGeom)) {

      const oWKTFormat = new WKT();
      const oWKT = this.getWKTFromEWKT(sEWKTGeom);
      const olWKTGeom = oWKT.geom;
      const sEWKTGeomProj = oWKT.proj;

      const oGeom = oWKTFormat.readGeometry(olWKTGeom, {
        dataProjection: sEWKTGeomProj,
        featureProjection: projection || sEWKTGeomProj
      });

      return oGeom;

    } else {
      console.error('Geom is not an EWKT Geometry: ', sEWKTGeom);
      return null;
    }
  }

  /**
   * Parse une geom EWKT et retourne un objet composé de proj et wkt
   * @param {string} sEWKTGeom
   * @returns {object}
   * @export
   */
  public getWKTFromEWKT(sEWKTGeom: string): any {

    if (!this.vitisCheck.isString(sEWKTGeom)) {
      console.error('sEWKTGeom is not a string');
      return null;
    }

    let sWKTGeom: string = sEWKTGeom.substring(sEWKTGeom.indexOf(';') + 1);
    const sProj: string = 'EPSG:' + sEWKTGeom.slice(5, sEWKTGeom.indexOf(';'));

    /**
     * Le WKT et EWKT n'ont pas la même syntaxe pour les géométries 3d
     * WKT: "MULTILINESTRING Z ((710630.741851595 6153582.35273146 0,710624.342249182 6153579.04143213 0 (...)"
     * EWKT: "SRID=2154;MULTILINESTRING((710630.741851595 6153582.35273146 0,710624.342249182 6153579.04143213 0 (...)"
     */

    // Vérifications pour le Z (3d)
    let bIs3DGeom = false;
    const sLastParenthesis = sEWKTGeom.substring(
      sEWKTGeom.lastIndexOf('(') + 1,
      sEWKTGeom.indexOf(')', sEWKTGeom.lastIndexOf('('))
    );
    if (sLastParenthesis.split(',')[0].split(' ').length === 3) {
      bIs3DGeom = true;
    }

    // Corrections pour la 3D
    if (bIs3DGeom) {
      sWKTGeom = sWKTGeom.substring(0, sWKTGeom.indexOf('(')) + ' Z' + sWKTGeom.substring(sWKTGeom.indexOf('('));
    }

    return {
      'proj': sProj,
      'geom': sWKTGeom
    };
  }

  /**
   * Parse un WKT et retourne une geom EWKT
   * @param sWKT
   * @returns
   */
  public getEWKTFromWKT(sWKT: any): string {

    if (!this.vitisCheck.isDefAndNotNull(sWKT.geom)) {
      console.error('geom is not defined');
      return null;
    }

    if (!this.vitisCheck.isDefAndNotNull(sWKT.proj)) {
      console.error('proj is not defined');
      return null;
    }

    if (!(sWKT.proj.indexOf('EPSG:') === 0)) {
      console.error('proj is not an EPSG code');
      return null;
    }

    const SRID = sWKT.proj.substring(5);
    let oWKTGeom = sWKT.geom;

    // Corrections pour la 3D car en EWKT il ne faut pas de "Z"
    oWKTGeom = oWKTGeom.substring(0, oWKTGeom.indexOf('(')).replace(' Z ', '') + oWKTGeom.substring(oWKTGeom.indexOf('('));

    const sEWKTGeom = 'SRID=' + SRID + ';' + oWKTGeom;
    return sEWKTGeom;
  }

  /**
   * Retourne true si sEWKTGeom est de type EWKT
   * @param {string} sEWKTGeom
   * @returns {boolean}
   * @export
   */
  public isEWKTGeom(sEWKTGeom: string): boolean {

    if (sEWKTGeom.substring(0, 5).toUpperCase() !== 'SRID=') {
      return false;
    }

    const oWKTFormat = new WKT();
    const oWKTGeom = this.getWKTFromEWKT(sEWKTGeom)['geom'];

    try {
      oWKTFormat.readGeometry(oWKTGeom);
    } catch (e) {
      console.error('readGeometry failed: ', sEWKTGeom);
      return false;
    }

    return true;
  }

  /**
   * Retourne true si sWKTGeom est de type WKT
   * @param {string} sWKTGeom
   * @returns {boolean}
   * @export
   */
  public isWKTGeom(sWKTGeom: string): boolean {

    const oWKTFormat = new WKT();
    try {
      oWKTFormat.readGeometry(sWKTGeom);
    } catch (e) {
      return false;
    }

    return true;
  }

  /**
   * Retourne true si sGeoJsonGeom est de type GeoJson
   * @param {string} sGeoJsonGeom
   * @returns {boolean}
   * @export
   */
  public isGeoJsonGeom(sGeoJsonGeom: string): boolean {

    const oGeoJsonFormat = new GeoJSON();
    try {
      oGeoJsonFormat.readGeometry(sGeoJsonGeom);
    } catch (e) {
      try {
        oGeoJsonFormat.readFeatures(sGeoJsonGeom);
      } catch (err) {
        return false;
      }
    }

    return true;
  }

  /**
   * Retourne l'échelle de la carte
   * si opt_options.pretty alors le retour est de tyle string ex : 1:25,000
   * sinon renvoie un nombre
   * @param  {Map} olMap
   * @param  {any} opt_options?
   * @returns any
   */
  public getScale(olMap: Map, opt_options?: any): any {

    opt_options = this.vitisCheck.isDefAndNotNull(opt_options) ? opt_options : {};
    opt_options.pretty = this.vitisCheck.isDefAndNotNull(opt_options.pretty) ? opt_options.pretty : false;

    const projection = olMap.getView().getProjection();
    const line = olMap.getView().calculateExtent([37.795275591, 0]);
    const c1 = transform([line[0], line[1]], projection, 'EPSG:4326');
    const c2 = transform([line[2], line[3]], projection, 'EPSG:4326');
    const length = getDistance(c1, c2);

    // donc 1m sur la carte correspond à (length mètres dans la réalité x 100)
    let scale: any = length * 100;

    // Rend l'échelle sous format (1:25,000)
    if (opt_options.pretty === true) {
      scale = this.getPrettyScale(scale);
    }

    return scale;
  }

  /**
   * Setter échelle
   * @param olMap
   * @param iScale
   */
  public setScale(olMap: Map, iScale: number): void {

    // calcule et va à l'échelle celon une règle de 3
    const currentScale = this.getScale(olMap);
    const currentResolution = olMap.getView().getResolution();

    const scaleResolution = iScale * currentResolution / currentScale;

    olMap.getView().setResolution(scaleResolution);
  }

  /**
   * Retourne l'échelle de la carte en mode pretty ex : 1:25,000
   * @param {number} scale
   * @returns {String} scale (pretty)
   */
  public getPrettyScale(scale: number): string {

    // Rend l'échelle sous format (1:25,000)
    let sScale = String(Math.round(scale));
    let j = 1;
    for (let i = sScale.length - 1; i > 0; i--) {
      if (j % 3 === 0) {
        sScale = sScale.slice(0, i) + ',' + sScale.slice(i + Math.abs(0));
      }
      j++;
    }
    sScale = '1:' + sScale;

    return sScale;
  }

  /**
   * Centre la carte sur la géométrie passée
   * @param  {Map} olMap
   * @param  {Geometry} olGeom
   * @param {number[]} padding
   * @param {boolean} keepMapZoom
   */
  public centerMapOnGeom(olMap: Map, olGeom: any, padding = [200, 200, 200, 200], keepMapZoom: boolean = false) {

    // Tests existence geom
    if (!this.vitisCheck.isDefAndNotNull(olGeom)) {
      return;
    }
    const coords = olGeom.getCoordinates();
    if (!coords || !this.vitisCheck.isDefAndNotNull(coords[0])) {
      return;
    }

    const zoomPoint = environment.properties.map_zoom_threshold_for_points;
    if (olGeom.getType() === 'Point' && !zoomPoint) {
      olMap.getView().fit(olGeom, {
        maxZoom: olMap.getView().getZoom(),
        padding,
      });
    } else {
      olMap.getView().fit(olGeom, {
        maxZoom: (olGeom.getType() === 'Point' ? Number(zoomPoint) : keepMapZoom) ? olMap.getView().getZoom() : undefined,
        padding,
      });
    }
  }

  /**
   * Centre la carte sur la localisation gps
   * @param  {Map} olMap
   * @param  {Geometry} olGeom
   */
  public zoomLocationGPS(olMap: Map, olGeom: any, padding = [200, 200, 200, 200], bForceNoZoom = false) {
    const zoomGps = environment.properties.map_zoom_threshold_for_gps_location;
    if (zoomGps && bForceNoZoom !== true) {
      olMap.getView().fit(olGeom, {
        maxZoom: olGeom.getType() === 'Point' ? Number(zoomGps) : undefined,
        padding,
      });
    } else {
      olMap.getView().setCenter(this.getGeometryCenter(olGeom));
    }
  }

  /**
   * Centre la carte sur la géométrie passée
   * @param  {Map} olMap
   * @param   aGeoms
   */
  public centerMapOnGeoms(olMap: Map, aGeoms: Array<Geometry>, padding = [200, 200, 200, 200]) {
    if (aGeoms.length > 1) {
      let maxZoom = 15;
      if (this.vitisCheck.isDefAndNotNull(environment.properties.map.default_feature_zoom)) {
        maxZoom = environment.properties.map.default_feature_zoom;
      }

      const oGeomCollection = new GeometryCollection(aGeoms);

      olMap.getView().fit(oGeomCollection.getExtent(), {
        maxZoom: undefined,
        padding,
      });
    } else {
      this.centerMapOnGeom(olMap, aGeoms[0], padding);
    }
  }

  /**
   * Récupère le centre de la géométrie
   * @param  {Geometry} olGeom
   * @returns any
   */
  public getGeometryCenter(olGeom: any): any {

    let position: any;

    if (olGeom.getType() === 'Polygon') {
      position = olGeom.getInteriorPoint().getLastCoordinate();
    } else if (olGeom.getType() === 'MultiPolygon') {

      const aPolygons = olGeom.getPolygons();
      let maxArea = 0;
      let bigestPolygon = 0;

      // Retrouve le plus gros polygone
      for (let i = 0; i < aPolygons.length; i++) {
        if (maxArea < aPolygons[i].getArea()) {
          maxArea = aPolygons[i].getArea();
          bigestPolygon = i;
        }
      }
      position = aPolygons[bigestPolygon].getInteriorPoint().getLastCoordinate();

    } else if (olGeom.getType() === 'LineString') {
      const centerPoint = Math.round(olGeom.getCoordinates().length / 2);
      position = olGeom.getCoordinates()[centerPoint];
    } else if (olGeom.getType() === 'Circle') {
      position = olGeom.getCenter();
    } else if (olGeom.getType() === 'Point' || olGeom.getType() === 'MultiPoint') {
      position = olGeom.getLastCoordinate();
    } else if (olGeom.getType() === 'GeometryCollection') {
      return 0;
    } else {
      return olGeom.getLastCoordinate();
    }

    return position;
  }

  /**
    * Retourne la géométrie au format EWKT
    * @param  {[Feature]} features
    * @param  {Map} olMap
    * @returns string
    */
  public getEWKTFromFeatures(features: Array<Feature<Geometry>>, olMap: Map): string {

    const proj: string = olMap.getView().getProjection().getCode();

    if (!this.vitisCheck.isDefAndNotNull(features)) {
      console.error('features is not defined');
      return null;
    }
    if (!this.vitisCheck.isString(proj)) {
      console.error('proj is not a string');
      return null;
    }
    if (!(proj.indexOf('EPSG:') === 0)) {
      console.error('proj is not an EPSG code');
      return null;
    }

    const oWKT = new WKT();
    const SRID = proj.substring(5);

    let oWKTGeom = oWKT.writeFeatures(features, {
      dataProjection: proj,
      featureProjection: proj
    });

    // Corrections pour la 3D car en EWKT il ne faut pas de "Z"
    oWKTGeom = oWKTGeom.substring(0, oWKTGeom.indexOf('(')).replace(' Z', '') + oWKTGeom.substring(oWKTGeom.indexOf('('));

    const sEWKTGeom = 'SRID=' + SRID + ';' + oWKTGeom;
    return sEWKTGeom;
  }

  /**
   * Parse une geom EWKT
   * @param {string} sEWKTGeom
   * @returns {object}
   * @export
   */
  public getGeojsonFromEWKT(sEWKTGeom: string): any {

    try {
      const oWKT = new WKT();
      const geom = oWKT.readGeometry(this.getWKTFromEWKT(sEWKTGeom).geom);

      const oGeoJson = new GeoJSON({});
      return oGeoJson.writeGeometry(geom);
    } catch (error) {
      return null;
    }
  }

  /**
   * Parse un GEOJSON et retourne une geom EWKT
   * @param sGeoJson
   * @param proj
   * @returns
   */
  public getEWKTFromGeojson(sGeoJson: any, proj): string {

    if (!this.vitisCheck.isDefAndNotNull(sGeoJson.coordinates)) {
      console.error('coordinates is not defined');
      return null;
    }

    const oGeoJson = new GeoJSON({});
    const geom = oGeoJson.readGeometry(sGeoJson);

    const oWKT = new WKT({});
    const SRID = proj.substring(5);

    const oWKTGeom = oWKT.writeGeometry(geom, {
      dataProjection: proj,
      featureProjection: proj
    });

    const sEWKTGeom = 'SRID=' + SRID + ';' + oWKTGeom;
    return sEWKTGeom;
  }

  /**
   * Parse une geom EWKT
   * @param {string} sEWKTGeom
   * @returns {object}
   * @export
   */
  public getGeojsonFromGeom(olGeom: Geometry): string {

    try {
      const oGeoJson = new GeoJSON({});
      return oGeoJson.writeGeometry(olGeom);
    } catch (error) {
      return null;
    }
  }

  /**
   * Parse une geom EWKT
   * @param {string} sEWKTGeom
   * @returns {object}
   * @export
   */
  public getGeojsonFromFeatures(features: [Feature<Geometry>], olMap: Map): string {

    const proj: string = olMap.getView().getProjection().getCode();

    if (!this.vitisCheck.isDefAndNotNull(features)) {
      console.error('features is not defined');
      return null;
    }
    if (!this.vitisCheck.isString(proj)) {
      console.error('proj is not a string');
      return null;
    }
    if (!(proj.indexOf('EPSG:') === 0)) {
      console.error('proj is not an EPSG code');
      return null;
    }

    const oGeoJson = new GeoJSON({});

    const sGeoJSONGeom = oGeoJson.writeFeatures(features, {
      dataProjection: 'EPSG:4326',
      featureProjection: proj
    });

    return sGeoJSONGeom;
  }

  /**
   * Transforme une géométrie EWKT ou WKT ou geojson en features OpenLayers
   * @param  {string} sGeom
   * @param  {Map} olMap
   * @returns Features
   */
  public getFeaturesFromGeom(sGeom: string | any, olMap: Map): any {

    if (this.vitisCheck.isObject(sGeom)) {
      sGeom = JSON.stringify(sGeom);
    }

    if (!this.vitisCheck.isString(sGeom)) {
      console.error('sGeom is not a string: ', sGeom);
      return null;
    }

    const isGeoJsonGeom = this.isGeoJsonGeom(sGeom);
    const isEWKTGeom = this.isEWKTGeom(sGeom);
    if (!isGeoJsonGeom && !this.isWKTGeom(sGeom) && !isEWKTGeom) {
      console.error('Geom is not an EWKT or WKT or GeoJSON Geometry: ', sGeom);
      return null;
    }

    // Valeurs par défaut (wkt)
    const sMapProj = olMap.getView().getProjection().getCode();
    let oFormat: GeoJSON | WKT = new WKT();
    const options: any = {
      featureProjection: sMapProj,
      dataProjection: sMapProj
    };

    // GeoJSON
    if (isGeoJsonGeom) {
      oFormat = new GeoJSON();
    }

    // EWKT
    if (isEWKTGeom) {
      const oWKT = this.getWKTFromEWKT(sGeom);
      sGeom = oWKT.geom;
      options.dataProjection = oWKT.proj;
    }

    let oFeature: any;
    try {
      oFeature = oFormat.readFeature(sGeom, options);
    } catch (e) {
      try {
        oFeature = oFormat.readFeatures(sGeom, options);
      } catch (err) {
        console.error('Cannor read features from geom ', sGeom);
      }
    }

    return this.splitGeometryCollection(oFeature);
  }

  /**
   * Découpe la feature si elle est de type geometry collection pour obtenir plusieurs feature
   * @param {Feature} oFeature
   */
  public splitGeometryCollection(oFeature: any): Array<Feature<Geometry>> {
    if (this.vitisCheck.isArray(oFeature)) {
      return oFeature;
    }
    if (oFeature.getGeometry().getType() === 'GeometryCollection') {
      const geometries = oFeature.getGeometry().getGeometries();
      const aFeatures: Array<Feature<Geometry>> = [];
      for (let i = 0; i < geometries.length; i++) {
        const geometry = geometries[i];
        aFeatures.push(new Feature(geometry));
      }

      return aFeatures;
    }

    return [oFeature];
  }

  /**
   * Get the feature JSON style definition
   * @param olFeature
   */
  public getStyleAsJSON(olFeature: any): any {
    switch (olFeature.getGeometry().getType()) {
      case 'Point':
      case 'MultiPoint':
        return this.getPointJSONStyle_(olFeature.getStyle());
        break;
      case 'LineString':
      case 'MultiLineString':
        return this.getLineJSONStyle_(olFeature.getStyle());
        break;
      case 'Polygon':
      case 'MultiPolygon':
        return this.getPolygonJSONStyle_(olFeature.getStyle());
        break;
      default:
        console.error('undefined type: ', olFeature.getGeometry().getType());
        break;
    }
  }

  /**
   * Get the point JSON style definition
   * @param {ol.style.Style} olStyle
   * @returns {object} JSON style definition
   */
  private getPointJSONStyle_(olStyle: any): any {

    const oStyle: any = {
      draw: {
        color: 'rgba(54,184,255,0.6)',
        outline_color: 'rgba(0,0,0,0.4)',
        size: 7
      },
      text: null
    };

    if (this.vitisCheck.isDefAndNotNull(olStyle)) {
      if (this.vitisCheck.isDefAndNotNull(olStyle.getImage())) {
        if (this.vitisCheck.isDefAndNotNull(olStyle.getImage().getGlyph)) {
          if (this.vitisCheck.isDefAndNotNull(olStyle.getImage().getGlyph())) {
            if (this.vitisCheck.isDefAndNotNull(olStyle.getImage().getGlyph().name)) {
              oStyle.draw.symbol = olStyle.getImage().getGlyph().name;
            }
          }
        }
        if (this.vitisCheck.isDefAndNotNull(olStyle.getImage().radius2_)) {
          oStyle.draw.size = olStyle.getImage().radius2_;
        }
        if (this.vitisCheck.isDefAndNotNull(olStyle.getImage().getFill)) {
          if (this.vitisCheck.isDefAndNotNull(olStyle.getImage().getFill())) {
            oStyle.draw.color = olStyle.getImage().getFill().color_;
          }
        }
        if (this.vitisCheck.isDefAndNotNull(olStyle.getImage().getStroke)) {
          if (this.vitisCheck.isDefAndNotNull(olStyle.getImage().getStroke())) {
            oStyle.draw.outline_color = olStyle.getImage().getStroke().color_;
          }
        }
        if (this.vitisCheck.isDefAndNotNull(olStyle.getImage().getSrc)) {
          if (this.vitisCheck.isDefAndNotNull(olStyle.getImage().getSrc())) {
            oStyle.draw.image = olStyle.getImage().getSrc();
          }
        }
        if (this.vitisCheck.isDefAndNotNull(olStyle.getImage().getRotation)) {
          if (this.vitisCheck.isDefAndNotNull(olStyle.getImage().getRotation())) {
            oStyle.draw.rotation = olStyle.getImage().getRotation();
          }
        }
      }
      oStyle.text = this.getTextJSONStyle_(olStyle);
    }

    return oStyle;
  }

  /**
  * Get the line JSON style definition
  * @param {ol.style.Style} olStyle
  * @returns {object} JSON style definition
  */
  private getLineJSONStyle_(olStyle: Style): any {

    const oStyle: any = {
      draw: {
        color: 'rgba(54,184,255,0.6)',
        outline_color: 'rgba(0,0,0,0.4)',
        size: 7
      },
      text: null
    };

    if (this.vitisCheck.isDefAndNotNull(olStyle)) {
      if (this.vitisCheck.isDefAndNotNull(olStyle.getFill())) {
        if (this.vitisCheck.isDefAndNotNull(olStyle.getFill().getColor())) {
          oStyle.draw.color = olStyle.getFill().getColor();
        }
      }
      if (this.vitisCheck.isDefAndNotNull(olStyle.getStroke())) {
        if (this.vitisCheck.isArray(olStyle.getStroke().getLineDash())) {
          oStyle.draw.dash = olStyle.getStroke().getLineDash()[0];
        }
        if (this.vitisCheck.isDefAndNotNull(olStyle.getStroke().getWidth())) {
          oStyle.draw.size = olStyle.getStroke().getWidth();
        }
      }
      oStyle.text = this.getTextJSONStyle_(olStyle);
    }
    return oStyle;
  }

  /**
  * Get the polygon JSON style definition
  * @param {ol.style.Style} olStyle
  * @returns {object} JSON style definition
  */
  private getPolygonJSONStyle_(olStyle: Style): any {

    const oStyle = {
      draw: {
        color: 'rgba(54,184,255,0.6)',
        outline_color: 'rgba(0,0,0,0.4)',
        size: 7
      },
      text: null
    };

    if (this.vitisCheck.isDefAndNotNull(olStyle)) {
      if (this.vitisCheck.isDefAndNotNull(olStyle.getFill())) {
        if (this.vitisCheck.isDefAndNotNull(olStyle.getFill().getColor())) {
          oStyle.draw.color = olStyle.getFill().getColor().toString();
        }
      }
      if (this.vitisCheck.isDefAndNotNull(olStyle.getStroke())) {
        if (this.vitisCheck.isDefAndNotNull(olStyle.getStroke().getColor())) {
          oStyle.draw.outline_color = olStyle.getStroke().getColor().toString();
        }
        if (this.vitisCheck.isDefAndNotNull(olStyle.getStroke().getWidth())) {
          oStyle.draw.size = olStyle.getStroke().getWidth();
        }
      }
      oStyle.text = this.getTextJSONStyle_(olStyle);
    }
    return oStyle;
  }

  /**
   * Get the text JSON style definition
   * @param {ol.style.Style} olStyle
   * @returns {object} JSON style definition
   */
  private getTextJSONStyle_(olStyle: Style): any {

    let oText = null;
    if (this.vitisCheck.isDefAndNotNull(olStyle.getText())) {
      oText = {};
      if (this.vitisCheck.isDefAndNotNull(olStyle.getText().getFont())) {
        const sFontWithPx = olStyle.getText().getFont();
        const aFontWithPx = sFontWithPx.split(' ');
        let sFontSize = 14;
        let sFontName = 'Arial';
        if (this.vitisCheck.isString(aFontWithPx[0])) {
          if (aFontWithPx[0].indexOf('px') !== -1) {
            sFontSize = parseInt(this.vitisClone.clone(aFontWithPx[0]), 10);
            aFontWithPx.shift();
          }
        }
        sFontName = aFontWithPx.join(' ');
        oText.font = sFontName;
        oText.size = sFontSize;
      }
      if (this.vitisCheck.isDefAndNotNull(olStyle.getText().getFill())) {
        if (this.vitisCheck.isDefAndNotNull(olStyle.getText().getFill().getColor())) {
          oText.color = olStyle.getText().getFill().getColor();
        }
      }
      if (this.vitisCheck.isDefAndNotNull(olStyle.getText().getOffsetX())) {
        oText.offsetX = olStyle.getText().getOffsetX();
      }
      if (this.vitisCheck.isDefAndNotNull(olStyle.getText().getOffsetY())) {
        oText.offsetY = (-1) * parseFloat(olStyle.getText().getOffsetY().toString());
      }
      if (this.vitisCheck.isDefAndNotNull(olStyle.getText().getStroke())) {
        if (this.vitisCheck.isDefAndNotNull(olStyle.getText().getStroke().getColor())) {
          oText.outline_color = olStyle.getText().getStroke().getColor();
        }
        if (this.vitisCheck.isDefAndNotNull(olStyle.getText().getStroke().getWidth())) {
          oText.outline_size = olStyle.getText().getStroke().getWidth();
        }
      }
      if (this.vitisCheck.isDefAndNotNull(olStyle.getText().getRotation())) {
        oText.rotation = parseFloat(olStyle.getText().getRotation().toString()) * 180 / Math.PI;
      }
      if (this.vitisCheck.isDefAndNotNull(olStyle.getText().getText())) {
        oText.text = olStyle.getText().getText();
      }
    }
    return oText;
  }

  /**
   * Set the feature style by a JSON style definition
   * @param olFeature
   * @param {StyleJSON} oJSON JSON style definition
   * @export
   */
  public setStyleByJSON(olFeature: any, oJSON: StyleJSON) {
    if (!this.vitisCheck.isDefAndNotNull(oJSON)) {
      console.error('oJSON not provided');
      return false;
    }

    if (!this.vitisCheck.isObject(oJSON)) {
      console.error('oJSON is neither JSON string nor object');
      return false;
    }

    switch (olFeature.getGeometry().getType()) {
      case 'Point':
      case 'MultiPoint':
        olFeature.setStyle(this.getPointStyleByJSON_(oJSON));
        break;
      case 'LineString':
      case 'MultiLineString':
        olFeature.setStyle(this.getLineStyleByJSON_(oJSON));
        break;
      case 'Polygon':
      case 'MultiPolygon':
        olFeature.setStyle(this.getPolygonStyleByJSON_(oJSON));
        break;
      default:
        console.error('undefined type: ', olFeature.getGeometry().getType());
        break;
    }
  }

  /**
   * Get the point style by the JSON definition
   * @param {StyleJSON} oJSON JSON style definition
   * @returns {Style}
   */
  private getPointStyleByJSON_(oJSON: StyleJSON): Style {

    oJSON = this.setDefaultValues_(oJSON, 7);

    let style = new Style({
      stroke: new Stroke({
        width: +oJSON['draw']['size'],
        color: oJSON['draw']['color']
      }),
      fill: new Fill({
        color: oJSON['draw']['outline_color']
      })
    });

    // Symbologie
    if (this.vitisCheck.isDefAndNotNull(oJSON['draw']['symbol'])) {
      // Symbole SVG
      // L'import de FontSymbol ne fonctionne pas
      // style.setImage(new FontSymbol({
      //   form: 'none', // forme du conteneur (none, circle, poi, bubble, marker, coma, shield, blazon, bookmark, hexagon, diamond, triangle, sign, ban, lozenge, square)
      //   gradient: false, // dégradé (true/false)
      //   glyph: oJSON['draw']['symbol'], // symbole à utiliser
      //   fontSize: 0.6,
      //   radius: oJSON['draw']['size'],
      //   //offsetX: -15,
      //   rotation: 0,
      //   rotateWithView: false, // true pour que l'icone bouge qand on rotate la carte
      //   offsetY: 0, // Pour mettre un offset il faut utiliser -radius
      //   color: '', // couleur du symbole ou vide pour utiliser la couleur du contour
      //   fill: new Fill({
      //     color: oJSON['draw']['color'] // couleur du conteneur
      //   }),
      //   stroke: new Stroke({
      //     color: oJSON['draw']['outline_color'], // couleur du contour
      //     width: 2 // taille du contour
      //   })
      // }));
    } else if (this.vitisCheck.isDefAndNotNull(oJSON['draw']['image'])) {

      const imgElem = document.createElement('img');
      imgElem.src = oJSON['draw']['image'];

      setTimeout(() => {
        let rotation = 0;
        if (this.vitisCheck.isDefAndNotNull(oJSON['draw']['rotation'])) {
          if (this.vitisCheck.isNumber(parseFloat(oJSON['draw']['rotation']))) {
            if (!Number.isNaN(parseFloat(oJSON['draw']['rotation']))) {
              rotation = parseFloat(oJSON['draw']['rotation']);
            }
          }
        }

        let size = 20;
        if (this.vitisCheck.isDefAndNotNull(oJSON['draw']['size'])) {
          if (this.vitisCheck.isNumber(parseFloat(oJSON['draw']['size']))) {
            if (!Number.isNaN(parseFloat(oJSON['draw']['size']))) {
              size = parseFloat(oJSON['draw']['size']);
            }
          }
        }

        const heightWidthRatio = imgElem.width / imgElem.height;
        const iconHeight = size;
        const iconWidth = size * heightWidthRatio;
        const canvasElem = document.createElement('canvas');
        const context = canvasElem.getContext('2d');
        canvasElem.height = iconHeight;
        canvasElem.width = iconWidth;
        context.drawImage(imgElem, 0, 0, iconWidth, iconHeight);

        // Image
        style.setImage(new Icon({
          img: canvasElem,
          imgSize: [iconWidth, iconHeight],
          rotation: rotation * Math.PI / 180
        }));
      });
    } else {
      // Cercle
      style.setImage(new CircleStyle({
        radius: +oJSON['draw']['size'],
        fill: new Fill({
          color: oJSON['draw']['color']
        }),
        stroke: new Stroke({
          color: oJSON['draw']['outline_color'],
          width: 2
        })
      }));
    }

    // Texte
    style = this.setStyleTextFormJSON_(style, oJSON);

    return style;
  }

  /**
   * Get the line style by the JSON definition
   * @param {StyleJSON} oJSON JSON style definition
   * @returns {Style}
   */
  private getLineStyleByJSON_(oJSON: StyleJSON): Style {

    // Valeurs par défaut
    oJSON = this.setDefaultValues_(oJSON, 2);

    // Dessin
    let style = new Style({
      fill: new Fill({
        color: oJSON['draw']['color']
      }),
      stroke: new Stroke({
        color: oJSON['draw']['color'],
        width: +oJSON['draw']['size'],
        lineDash: [+oJSON['draw']['dash']]
      })
    });

    // Texte
    style = this.setStyleTextFormJSON_(style, oJSON);

    return style;
  }

  /**
   * Get the polygon style by the JSON definition
   * @param {StyleJSON} oJSON JSON style definition
   * @returns {Style}
   */
  private getPolygonStyleByJSON_(oJSON: StyleJSON): Style {

    // Valeurs par défaut
    oJSON = this.setDefaultValues_(oJSON, 2);

    // Dessin
    let style = new Style({
      fill: new Fill({
        color: oJSON['draw']['color']
      }),
      stroke: new Stroke({
        color: oJSON['draw']['outline_color'],
        width: +oJSON['draw']['size']
      })
    });

    // Texte
    style = this.setStyleTextFormJSON_(style, oJSON);

    return style;
  }

  /**
   * Set default values
   * @param oJSON
   * @param iSize
   * @returns oJSON
   */
  private setDefaultValues_(oJSON: StyleJSON, iSize: number): any {
    // Valeurs par défaut dessin
    oJSON['draw'] = this.vitisCheck.isDefAndNotNull(oJSON['draw']) ? oJSON['draw'] : {};
    oJSON['draw']['color'] = this.vitisCheck.isDefAndNotNull(oJSON['draw']['color']) ? oJSON['draw']['color'] : 'rgba(54,184,255,0.6)';
    oJSON['draw']['outline_color'] = this.vitisCheck.isDefAndNotNull(oJSON['draw']['outline_color']) ? oJSON['draw']['outline_color'] : 'rgba(0,0,0,0.4)';
    oJSON['draw']['size'] = this.vitisCheck.isDefAndNotNull(oJSON['draw']['size']) ? oJSON['draw']['size'] : iSize.toString();
    oJSON['draw']['dash'] = this.vitisCheck.isDefAndNotNull(oJSON['draw']['dash']) ? oJSON['draw']['dash'] : '0';

    // Valeurs par défaut texte
    if (this.vitisCheck.isDefAndNotNull(oJSON['text'])) {
      oJSON['text']['size'] = this.vitisCheck.isDefAndNotNull(oJSON['text']['size']) ? oJSON['text']['size'] : '14';
      oJSON['text']['font'] = this.vitisCheck.isDefAndNotNull(oJSON['text']['font']) ? oJSON['text']['font'] : 'Arial';
      oJSON['text']['color'] = this.vitisCheck.isDefAndNotNull(oJSON['text']['color']) ? oJSON['text']['color'] : 'black';
      oJSON['text']['outline_color'] = this.vitisCheck.isDefAndNotNull(oJSON['text']['outline_color']) ? oJSON['text']['outline_color'] : 'white';
      oJSON['text']['outline_size'] = this.vitisCheck.isDefAndNotNull(oJSON['text']['outline_size']) ? oJSON['text']['outline_size'] : '1';
      oJSON['text']['offsetX'] = this.vitisCheck.isDefAndNotNull(oJSON['text']['offsetX']) ? oJSON['text']['offsetX'] : '0';
      oJSON['text']['offsetY'] = this.vitisCheck.isDefAndNotNull(oJSON['text']['offsetY']) ? oJSON['text']['offsetY'] : '0';
      oJSON['text']['rotation'] = this.vitisCheck.isDefAndNotNull(oJSON['text']['rotation']) ? oJSON['text']['rotation'] : '0';
    }

    return oJSON;
  }

  /**
   * Set text in Style
   * @param style
   * @param oJSON
   */
  private setStyleTextFormJSON_(style: Style, oJSON: StyleJSON): Style {
    // Texte
    if (this.vitisCheck.isDefAndNotNull(oJSON['text'])) {
      style.setText(new Text({
        font: oJSON['text']['size'] + 'px ' + oJSON['text']['font'],
        text: oJSON['text']['text'],
        fill: new Fill({ color: oJSON['text']['color'] }),
        stroke: new Stroke({ color: oJSON['text']['outline_color'], width: +oJSON['text']['outline_size'] }),
        offsetX: +oJSON['text']['offsetX'],
        offsetY: (-1) * parseFloat(oJSON['text']['offsetY']),
        rotation: parseFloat(oJSON['text']['rotation']) * Math.PI / 180
      }));
    }

    return style;
  }

  /**
   * Return l'url de Google Street View depuis sEWKTGeom
   * @param  {string} sEWKTGeom
   * @param  {Map} olMap
   * @returns Geometry
   */
  public getOpenStreetViewUrlFromEWKT(sEWKTGeom: string, olMap: Map, clickedMapCoordinates): string | null {

    if (!this.vitisCheck.isString(sEWKTGeom)) {
      console.error('sEWKTGeom is not a string: ', sEWKTGeom);
      return null;
    }

    if (this.isEWKTGeom(sEWKTGeom)) {

      const oWKTFormat = new WKT();
      const oWKT = this.getWKTFromEWKT(sEWKTGeom);
      const olWKTGeom = oWKT.geom;
      const sEWKTGeomProj = oWKT.proj;

      const sMapProj = olMap.getView().getProjection().getCode();

      const oGeom: any = oWKTFormat.readGeometry(olWKTGeom, {
        dataProjection: sEWKTGeomProj,
        featureProjection: sMapProj
      });

      let aCoordinates = oGeom.getCoordinates();
      let minDistance = null, distance;
      let indexCoord = 0;
      if (aCoordinates[0][0]) {
        while (typeof aCoordinates[0][0] != 'number') {
          aCoordinates = aCoordinates[0];
        }
        for (let i = 0; i < aCoordinates.length; i++) {
          if (aCoordinates[i][0] >= clickedMapCoordinates[0]) {
            distance = aCoordinates[i][0] - clickedMapCoordinates[0];
          } else if (aCoordinates[i][0] <= clickedMapCoordinates[0]) {
            distance = clickedMapCoordinates[0] - aCoordinates[i][0];
          }
          if (!minDistance) {
            minDistance = distance;
          } else if (minDistance && distance < minDistance) {
            minDistance = distance;
            indexCoord = i;
          }
        }
        aCoordinates = aCoordinates[indexCoord];
      }
      const projection = transform(aCoordinates, olMap.getView().getProjection(), 'EPSG:4326');

      return `https://www.google.com/maps?q&layer=c&cbll=${projection[1]},${projection[0]}`;

    } else {
      console.error('Geom is not an EWKT Geometry: ', sEWKTGeom);
      return null;
    }
  }

  /**
   * Retourne true/false si les coordonnées sont valides
   * @param aCoords
   * @returns
   */
  private isValidCoordinates(aCoords: any | Coordinate[]): boolean {
    for (const coord of aCoords) {
      if (this.vitisCheck.isArray(coord)) {
        if (!this.isValidCoordinates(coord)) {
          return false;
        }
      } else if (!this.vitisCheck.isNumber(coord)) {
        return false;
      }
    }
    return true;
  }
}
