
import { Component, OnInit } from '@angular/core';
import { ActivatedRoute, Router } from '@angular/router';
import { environment } from '../../../environments/environment';
import { AuthService } from '../core/services/auth/auth.service';
import { VitisCheck, SessionStorage, vitisAlert } from 'src/app/vitis/core/services/utils/vitis-utils';
import { LocalStorageService } from 'src/app/vitis/core/services/utils/localstorage/local-storage.service';
import { TranslateService } from '@ngx-translate/core';


@Component({
  selector: 'app-oauth2',
  templateUrl: './oauth2.component.html',
  styleUrls: ['./oauth2.component.scss']
})
export class OAuth2Component implements OnInit {

  // paramètre venant de al route
  private oParams: any;
  // clé temporaire pour stocker en cache ce qui doit être partagé tout au long du flow
  private params_key: string = 'oauth2_params_temp';
  private state_key: string = 'oauth2_state_temp';

  /**
   * Constructeur
   */
  constructor(
    private route: ActivatedRoute,
    private authService: AuthService,
    private vitisCheck: VitisCheck,
    private localStorageService: LocalStorageService,
    private router: Router,
    private translate: TranslateService,
    private sessionStorage: SessionStorage
  ) {
    // Initialise la taille de l'écran actuel
    this.sessionStorage.set('smallScreen', window.innerWidth <= environment.properties.smallscreen_limit);
  }

  /**
   * Chargement du composant
   */
  ngOnInit() {


    if (environment.properties.allow_oauth2_connection !== true) {
      this.redirectToClassicLoginPage();
      return;
    }

    // Survéillance du filtre dans les paramètres
    this.route.queryParams.subscribe(oParams => {
      this.oParams = oParams;
      if (!oParams.code || !oParams.session_state || !oParams.state) {
        this.lookForAuthCode();
      } else {
        this.login();
      }

    });
  }

  /**
   * Récupère un code d'authentification pour un clientId
   * @returns
   */
  private lookForAuthCode(): void {

    if (!environment.properties.oauth2_auth_endpoint) {
      this.redirectToClassicLoginPage();
      return;
    }
    if (!environment.properties.oauth2_clientId) {
      this.redirectToClassicLoginPage();
      return;
    }

    const redirectUri = environment.properties.app_client_url + '/oauth2';
    const state = this.generateStateCode();

    if (this.oParams) {
      this.localStorageService.set(this.params_key, this.oParams);
    }
    this.localStorageService.set(this.state_key, state);

    const sUrl = environment.properties.oauth2_auth_endpoint + '?client_id=' +
      environment.properties.oauth2_clientId +
      '&response_type=code&state=' + state + '&scope=profile&redirect_uri=' + redirectUri;

    // on redirige sur la page d'authentification si les ifnos ne sont pas présente
    window.location.href = sUrl;
  }

  /**
   * renvoi sur la page de connexion classique
   */
  private redirectToClassicLoginPage(): void {
    this.router.navigate(['login'], { queryParamsHandling: 'merge' });
  }

  /**
   * Génére un code d'état pour oAuth2
   * @returns
   */
  private generateStateCode(iLength: number = 14): string {
    const sPanel = 'abcdefghijklmnopqrstuvwxyz0123456789';

    let sCode = '';

    for (let i = 0; i < iLength; i++) {
      sCode += sPanel.charAt(Math.floor(Math.random() * sPanel.length));
    }


    return sCode;
  }

  /**
 * Login function
 * @returns void
 */
  public login(): void {

    const initialParams = this.localStorageService.get(this.params_key) || {};
    const lastState = this.localStorageService.get(this.state_key);
    if (this.oParams.state != lastState) {
      console.error('Bad state value');
      return;
    }

    // Mémorise la valeur de remember me
    this.localStorageService.set('login_remember_me', false);

    this.authService.OAuth2Login(this.oParams.code, this.oParams.session_state).subscribe((result) => {

      // on efface les variables temporaires
      this.localStorageService.removeItem(this.params_key);
      this.localStorageService.removeItem(this.state_key);

      if (result === true) {
        this.router.navigate(['**'], { queryParams: initialParams, queryParamsHandling: '' });
      } else {
        // Message d'erreur à afficher.
        let sErrorTranslationId = 'ERROR_LOGIN';
        if (this.vitisCheck.isDefAndNotNull(result.data) &&
          this.vitisCheck.isDefAndNotNull(result.data.errorMessage)) {
          sErrorTranslationId = result.data.errorMessage;
        }
        // Affichage du message d'erreur.
        vitisAlert.alert({
          text: this.translate.instant(sErrorTranslationId),
          type: 'error'
        });
      }
    });
  }
}
