import { Component, OnInit, Input, Output, EventEmitter, OnChanges, SecurityContext } from '@angular/core';
import { DomSanitizer } from '@angular/platform-browser';
import { VitisCheck } from 'src/app/vitis/core/services/utils/vitis-utils';

@Component({
  selector: 'app-treeview',
  templateUrl: './treeview.component.html',
  styleUrls: ['./treeview.component.scss']
})
export class TreeviewComponent implements OnInit, OnChanges {

  /**
   * defaultOpenLevel
   * multiSelect
   */
  @Input() public config: any;

  /**
   * Tableau contenant les items
   */
  @Input() public items: any;

  /**
   * Chaine de caractère contenant un filtre
   */
  @Input() public filter: String;

  /**
   * Définit si le treeview est disabled
   */
  @Input() public disabled: boolean;

  @Output() public selectedChange: EventEmitter<any> = new EventEmitter<any>();
  @Output() public filterChange: EventEmitter<any> = new EventEmitter<any>();
  @Output() public folderCollapsedChange: EventEmitter<any> = new EventEmitter<any>();

  /**
   * Configuration par défaut
   */
  private defaultConfig = {
    titleAttribute: 'text',
    nodeAttribute: 'nodes',
    pathAttribute: 'path',
    dateAttribute: 'date',
    selectFolders: false,
    showDate: false,
    iconNode: 'fa-folder',
    iconOpenNode: 'fa-folder-open',
    hideEmptyNodes: false,
    checkboxSelection: false,
    multiSelect: true
  };

  /**
   * Constructeur
   */
  constructor(
    public vitisCheck: VitisCheck,
    private _sanitizer: DomSanitizer
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit() {
    // Surcharge la configuration par défaut
    this.config = Object.assign(
      this.defaultConfig,
      this.config
    );
  }

  /**
   * Appelé lorsqu'il y a des changements sur les input
   */
  ngOnChanges() {
    this.items = this.parseItems_(this.items);
  }

  /**
   * Gère le click sur l'icône d'un node
   * @param  {any} item
   * @returns void
   */
  public collapsItem(item: any): void {
    if (item[this.config.nodeAttribute]) {
      item.collapsed = !item.collapsed;
    }
    // Événement.
    this.folderCollapsedChange.emit(item);
  }

  /**
   * Sélectionne un item
   * @param  {any} item
   * @returns void
   */
  public selectItem(item: any): void {

    // Mode disabled
    if (this.disabled) {
      return;
    }

    // La selection est un dossier ?
    if (this.vitisCheck.isDefAndNotNull(item[this.config.nodeAttribute])) {
      if (this.config.selectFolders === false) {
        this.collapsItem(item);
        return;
      }
    }

    // Toggle de selection
    const selected: boolean = item.selected == false;

    // Option multiSelect
    if (!this.config.multiSelect) {
      const aItems = this.getItemsArray_(this.items);
      for (let i = 0; i < aItems.length; i++) {
        const tmpItem = aItems[i];
        tmpItem.selected = false;
      }
    }

    // Applique selection
    item.selected = selected;

    // Événement
    this.selectedChange.emit(this.getSelectedValues());
  }

  /**
   * Retourne les valeurs selectionnées
   */
  public getSelectedValues(): Array<any> {
    const aSelectedItems = this.getSelectedItems();
    const aValues = [];
    for (let i = 0; i < aSelectedItems.length; i++) {
      if (this.config.pathAttribute) {
        if (aSelectedItems[i][this.config.pathAttribute]) {
          aValues.push(aSelectedItems[i][this.config.pathAttribute]);
        }
      } else {
        aValues.push(aSelectedItems[i]);
      }
    }
    return aValues;
  }

  /**
   * Retourne un tableau des items selectionnés
   * @returns Array
   */
  public getSelectedItems(): Array<any> {
    const selectedItems = [];
    const aItems = this.getItemsArray_(this.items);
    for (let i = 0; i < aItems.length; i++) {
      const item = aItems[i];
      if (item.selected === true) {
        selectedItems.push(item);
      }
    }
    return selectedItems;
  }

  /**
   * Set visible true/false sur les items en fonction du filtre
   * @param  {any} items
   * @param  {String} filter
   * @returns any
   */
  public getFilteredItems(items: any, filter: String): any {
    if (this.vitisCheck.isArray(items)) {
      for (let i = 0; i < items.length; i++) {
        const item = items[i];

        if (!this.isItemOnfilter_(item, filter)) {
          item.visible = false;
        } else {
          item.visible = true;
        }

        // Affichage ou non des noeuds vides
        if (this.config.hideEmptyNodes === true && (this.vitisCheck.isArray(item[this.config.nodeAttribute]) &&
          item[this.config.nodeAttribute].length == 0)) {
          item.visible = false;
        }

        // Recursivité
        if (item[this.config.nodeAttribute]) {
          item[this.config.nodeAttribute] = this.getFilteredItems(item[this.config.nodeAttribute], filter);
        }
      }
    }

    return items;
  }

  /**
   * Retourne true/false si l'élément doit être affiché
   * @param  {any} item
   * @param  {String} filter
   * @returns boolean
   */
  private isItemOnfilter_(item: any, filter: String): boolean {
    let isOnFilter = false;

    if (!this.vitisCheck.isDefAndNotNull(filter) || filter === '') {
      isOnFilter = true;
    } else if (this.vitisCheck.isString(item[this.config.titleAttribute])) {
      if (item[this.config.titleAttribute].toLowerCase().search(filter.toLowerCase()) !== -1) {
        isOnFilter = true;
      }
    }

    if (item[this.config.nodeAttribute] && !isOnFilter) {
      for (let i = 0; i < item[this.config.nodeAttribute].length; i++) {
        const child = item[this.config.nodeAttribute][i];
        const isChildOnFilter = this.isItemOnfilter_(child, filter);
        if (isChildOnFilter) {
          isOnFilter = true;
        }
      }
    }

    return isOnFilter;
  }

  /**
   * Parse les items et ajoute les paramètres (collapsed etc...)
   * @param  {any} items
   * @returns any
   */
  public parseItems_(items: any, iLvl = 0): any {
    if (this.vitisCheck.isArray(items)) {
      for (const item of items) {
        item.collapsed = item.collapsed || true;
        item.selected = item.selected;

        // Option defaultOpenLevel
        if (this.config.defaultOpenLevel >= iLvl) {
          item.collapsed = false;
        }

        // check les injection xss créer sur le serveur pour bloque l'exécution du code
        item[this.config.titleAttribute] = this._sanitizer.sanitize(SecurityContext.HTML, item[this.config.titleAttribute]);

        // Recursivité
        if (item[this.config.nodeAttribute]) {
          this.parseItems_(item[this.config.nodeAttribute], iLvl + 1);
        }
      }
    }
    return items;
  }

  /**
   * Retourne l'ensemble des items dans un tableau
   * @param  {any} items
   * @returns Array
   */
  private getItemsArray_(items: any): Array<any> {

    let aItems = [];

    if (this.vitisCheck.isArray(items)) {
      for (let i = 0; i < items.length; i++) {
        const item = items[i];

        aItems.push(item);

        // Recursivité
        if (item[this.config.nodeAttribute]) {
          const aChildrens = this.getItemsArray_(item[this.config.nodeAttribute]);

          aItems = [...aItems, ...aChildrens];
        }
      }
    }

    return aItems;
  }
}
