import { Component, OnInit, Input, OnChanges } from '@angular/core';
import { environment } from 'src/environments/environment';
import { VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { TranslateService } from '@ngx-translate/core';

@Component({
  selector: 'app-data-source-external-database',
  templateUrl: './data-source-external-database.component.html',
  styleUrls: ['./data-source-external-database.component.scss']
})
export class StudioDataSourceExternalDatabaseComponent implements OnInit, OnChanges {

  @Input() public sSourceManagerMode: string;
  @Input() public oSelectedDataSource: any;

  public aTableAttributes: any = [];
  public aTableData: any = [];
  public aExternalSchemaTables: any = [];

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private translate: TranslateService
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit() {
  }

  /**
   * Appelé lorsqu'il y a des changements sur les input
   */
  ngOnChanges() {
    if (this.oSelectedDataSource.dataType == 'externalDatabase') {
      switch (this.sSourceManagerMode) {
        case 'overview':
        case 'edition':
          // rétro-compatibilité des anciens formulaires
          if (this.oSelectedDataSource.parameters.sgbd == 'pgsql') {
            this.oSelectedDataSource.parameters.sgbd = 'pdo_pgsql';
          }
          if (this.oSelectedDataSource.parameters.sgbd == 'oci') {
            this.oSelectedDataSource.parameters.sgbd = 'pdo_oci';
          }
          break;
        case 'creation':
          // Valeurs par défaut.
          this.oSelectedDataSource.type = 'web_service';
          this.oSelectedDataSource.parameters = {
            filter: {
              relation: 'AND',
              operators: []
            }
          };
          break;
      }
    }
  }

  /**
   * Ajout d'un filtre pour la source de données.
   * @returns void
   */
  public addDataSourceFilter() {
    const aDataSourceFilter = this.oSelectedDataSource.parameters.filter.operators;
    if (aDataSourceFilter.length < 5) {

      // Supprime le dernier filtre si non rempli.
      if (aDataSourceFilter.length > 0) {
        const aLastDataSourceFilter = aDataSourceFilter[aDataSourceFilter.length - 1];
        if (aLastDataSourceFilter.column == '' || aLastDataSourceFilter.operator == '') {
          aDataSourceFilter.splice(-1, 1);
        }
      }

      // Ajoute un filtre.
      this.oSelectedDataSource.parameters.filter.operators.push({
        column: '',
        compare_operator: '',
        value: ''
      });
    }
  }

  /**
   * Suppression d'un filtre pour la source de données.
   * @param  {number} iFilterIndex
   * @returns void
   */
  public deleteDataSourceFilter(iFilterIndex: number) {
    this.oSelectedDataSource.parameters.filter.operators.splice(iFilterIndex, 1);
  }

  /**
   * Mise à jour des paramètres obligatoires.
   * @returns void
   */
  public updateDataSourceParams() {
    this.oSelectedDataSource.ressource_id = 'vitis/genericqueries';
  }

  /**
   * Chargement de la liste des tables de la source externe.
   * @returns void
   */
  public loadSchemaTables = function () {

    // Vide l'aperçu de la table.
    this.aTableAttributes.length = 0;
    this.aTableData.length = 0;

    // Vide les données.
    this.aExternalSchemaTables.length = 0;

    const oParams: any = {
      login: this.oSelectedDataSource.parameters.login,
      password: this.oSelectedDataSource.parameters.password,
      server: this.oSelectedDataSource.parameters.server,
      port: this.oSelectedDataSource.parameters.port,
      sgbd: this.oSelectedDataSource.parameters.sgbd,
      database: this.oSelectedDataSource.parameters.database,
      schema: this.oSelectedDataSource.parameters.schema,
      charset: this.oSelectedDataSource.parameters.charset,
      order_by: 'table_name'
    };

    if (this.oSelectedDataSource.parameters.charset) {
      oParams.charset = this.oSelectedDataSource.parameters.charset;
    }

    // Lance la requête
    this.vitisRequest.ajaxRequestPromise({
      method: 'GET',
      url: environment.properties.app_api_url + '/vitis/genericqueries/tables',
      httpMethodOverride: true,
      params: oParams,
    }).then((resp: any) => {
      if (resp.data) {
        if (resp.data.status === 1) {
          if (this.vitisCheck.isDefAndNotNull(resp.data)) {
            this.aExternalSchemaTables = resp.data.data;
          }
        }
      }
    });
  };

  /**
   * Teste les valeurs d'une table locale.
   * @returns void
   */
  public testExternalDatabaseValues = function () {

    // Vide les données.
    this.aTableAttributes.length = 0;
    this.aTableData.length = 0;

    // Filtres.
    const oParams: any = {
      login: this.oSelectedDataSource.parameters.login,
      password: this.oSelectedDataSource.parameters.password,
      server: this.oSelectedDataSource.parameters.server,
      port: this.oSelectedDataSource.parameters.port,
      sgbd: this.oSelectedDataSource.parameters.sgbd,
      database: this.oSelectedDataSource.parameters.database,
      schema: this.oSelectedDataSource.parameters.schema,
      table: this.oSelectedDataSource.parameters.table,
      limit: 100
    };

    if (this.oSelectedDataSource.parameters.charset) {
      oParams.charset = this.oSelectedDataSource.parameters.charset;
    }

    if (this.oSelectedDataSource.parameters.filter.operators.length > 0) {
      oParams.filter = JSON.stringify(this.oSelectedDataSource.parameters.filter);
    }

    // Lance la requête
    this.vitisRequest.ajaxRequestPromise({
      method: 'GET',
      url: environment.properties.app_api_url + '/vitis/genericqueries',
      params: oParams,
      httpMethodOverride: true,
    }).then((resp: any) => {
      if (resp.data) {
        if (resp.data.status === 1) {
          if (this.vitisCheck.isDefAndNotNull(resp.data.data)) {

            // Nom des colonnes du web service (aperçu).
            this.aTableData = resp.data.data;
            if (this.vitisCheck.isDefAndNotNull(this.aTableData)) {
              const aTableAttributes = Object.keys(this.aTableData[0]);
              for (let i = 0; i < aTableAttributes.length; i++) {

                // Pas de données tableau ou objet.
                if (!this.vitisCheck.isArray(this.aTableData[0][aTableAttributes[i]]) &&
                  !this.vitisCheck.isObject(this.aTableData[0][aTableAttributes[i]])) {
                  this.aTableAttributes.push(aTableAttributes[i]);
                }
              }
            }
          }
        }
      }
    });
  };
}
