import { Component, OnInit, ViewChild, Output, EventEmitter, Input, OnDestroy } from '@angular/core';
import { StudioService } from '../../services/studio.service';
import { FormreaderComponent } from 'src/app/vitis/formreader/formreader.component';
import { VitisCheck, VitisClone } from 'src/app/vitis/core/services/utils/vitis-utils';
import { LocalStorageService } from 'src/app/vitis/core/services/utils/localstorage/local-storage.service';
import { StudioOptions } from '../../studio.model';

@Component({
  selector: 'app-preview',
  templateUrl: './preview.component.html',
  styleUrls: ['./preview.component.scss']
})
export class StudioPreviewComponent implements OnInit, OnDestroy {

  @Input() oFormId;
  @Output() public previewModeChanges = new EventEmitter<string>();

  // Mode de prévisualisation
  public sPreviewMode: string = 'form';
  public oFormDefinition: any;
  public sFormDefinitionName: string;

  // Monaco options
  public editorOptions = {
    theme: 'vs',
    language: 'json',
    fontFamily: 'Arial',
    fontSize: 13,
    automaticLayout: true,
    formatOnPaste: true
  };
  public editorJSONCode = '';
  public editorJSCode = '';
  public editorCSSCode = '';

  public oStudioOptions: StudioOptions;
  private firstLoadCSSCode = false;

  // Rechargement auto des sources de données
  public bAutoReloadDatasources = true;

  @ViewChild(FormreaderComponent, { static: false }) oFormReader: FormreaderComponent;

  /**
   * Constructeur
   */
  constructor(
    private studioService: StudioService,
    private vitisCheck: VitisCheck,
    private vitisClone: VitisClone,
    private localStorageService: LocalStorageService
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit() {

    // options du studio
    this.studioService.getStudioOptions().subscribe((oStudioOptions) => {
      this.oStudioOptions = oStudioOptions;
    });

    // Observe la modification du formulaire
    this.studioService.getStudioForm().subscribe((oFormDefinition: any) => {
      this.oFormDefinition = oFormDefinition;

      if (this.oFormReader) {
        this.oFormReader.setFormDefinition(oFormDefinition);
      }
    });

    // Observe la modification du nom du formulaire
    this.studioService.getStudioFormName().subscribe((sFormDefinitionName: any) => {
      this.sFormDefinitionName = sFormDefinitionName;

      if (this.oFormReader) {
        this.oFormReader.setFormDefinitionName(sFormDefinitionName);
        // Rafraichit le formreader
        this.studioService.setStudioReloadFormListerner(this.oFormDefinition);
      }
    });

    // Observe la modification du js
    this.studioService.getStudioJsField().subscribe((sJsContent: any) => {
      this.editorJSCode = sJsContent;
    });

    // Observe la modification du css
    this.studioService.getStudioCssField().subscribe((sCssContent: any) => {
      this.editorCSSCode = sCssContent;
      if (!this.firstLoadCSSCode) {
        this.firstLoadCSSCode = true;
        this.loadCssPreview();
      }
    });

    // Observe le rechargement du formulaire
    this.studioService.getStudioReloadFormListerner().subscribe((oFormDefinition: any) => {
      this.reloadFormDefinition(oFormDefinition);
    });

    // Valeur de "Recharger les sources de données automatiquement" d'apprès le localstorage
    if (this.vitisCheck.isBoolean(this.localStorageService.get('studio_autoreload_datasources'))) {
      this.bAutoReloadDatasources = this.localStorageService.get('studio_autoreload_datasources');
    }
  }

  /**
   * Destruction du composant
   */
  ngOnDestroy(): void {
    // Supprime les css précédement chargés
    const oldDiv = document.getElementsByClassName('vmap-form-css');
    for (let i = 0; i < oldDiv.length; i++) {
      oldDiv[i].remove();
    }
  }

  /**
   * Change le type de prévisualisation
   * @param sPreviewMode
   */
  public changePreviewMode(sPreviewMode: string) {
    this.sPreviewMode = sPreviewMode;

    // Recharge le formulaire
    if (sPreviewMode === 'form') {
      setTimeout(() => {
        this.reloadFormDefinition(this.oFormDefinition);
      });
    } else if (sPreviewMode === 'json') {
      this.editorOptions.language = 'json';
      this.editorJSONCode = this.studioService.getJSONFormDefinition();
    } else if (sPreviewMode === 'js') {
      this.editorOptions.language = 'javascript';
    } else if (sPreviewMode === 'css') {
      this.editorOptions.language = 'css';
    }

    // Emit le type de preview
    this.previewModeChanges.emit(sPreviewMode);
  }

  /**
   * Recharge le formulaire
   *
   * @param {boolean} bLoadDatasource Par défaut false (ne charge pas les datasources)
   */
  public reloadFormDefinition(oFormDefinition: any, bLoadDatasource: boolean = false) {
    bLoadDatasource = this.bAutoReloadDatasources === true ? true : bLoadDatasource;

    if (this.sPreviewMode === 'form') {
      if (this.vitisCheck.isDefAndNotNull(this.oFormReader) &&
        this.vitisCheck.isDefAndNotNull(oFormDefinition) &&
        this.vitisCheck.isDefAndNotNull(this.sFormDefinitionName)) {
        this.sPreviewMode = '';
        setTimeout(() => {
          this.sPreviewMode = 'form';
          setTimeout(() => {
            this.oFormReader.initForm(oFormDefinition, this.sFormDefinitionName, null, bLoadDatasource, false);
          }, 50);
        }, 50);

      }
    } else if (this.sPreviewMode === 'json') {
      this.editorJSONCode = this.studioService.getJSONFormDefinition();
    }
  }

  /**
   * onEditorInit function
   * @param  {any} editor
   * @returns void
   */
  public onEditorInit(editor: any) {

    // Formatage du JSON.
    if (editor && this.sPreviewMode == 'json') {
      setTimeout(function () {
        editor.getAction('editor.action.formatDocument').run();
      }, 500);
    }
  }

  /**
   * Sauvegarde la définition
   */
  public saveForm() {
    if (this.sPreviewMode === 'json') {
      try {
        this.studioService.setStudioForm(JSON.parse(this.editorJSONCode));
        setTimeout(() => {
          this.changePreviewMode('form');
        });
      } catch (error) {
        console.error('JSON definition not valid');
      }
    }
    if (this.sPreviewMode === 'js') {
      setTimeout(() => {
        this.changePreviewMode('form');
      });
    }
    if (this.sPreviewMode === 'css') {
      this.loadCssPreview();
      this.changePreviewMode('form');
    }
  }

  /**
   * Remonte en temps réel la valeur du monacoeditor au service
   */
  public jsCodeChange() {
    this.studioService.setStudioJsField(this.editorJSCode);
  }

  /**
   * Remonte en temps réel la valeur du monacoeditor au service
   */
  public cssCodeChange() {
    this.studioService.setStudioCssField(this.editorCSSCode);
  }

  /**
   * Change la valeur de this.bAutoReloadDatasources
   * @param val
   */
  public setAutoReloadDatasources(val: boolean) {
    this.bAutoReloadDatasources = val;

    // Valeur de "Recharger les sources de données automatiquement" d'apprès le localstorage
    this.localStorageService.set('studio_autoreload_datasources', this.bAutoReloadDatasources);

    // recharge la définition
    this.reloadFormDefinition(this.oFormDefinition);
  }

  /**
   * Affiche le css pour le formulaire preview uniquement
   */
  private loadCssPreview() {
    // Supprime les css précédement chargés
    const oldDiv = document.getElementsByClassName('vmap-form-css');
    for (let i = 0; i < oldDiv.length; i++) {
      oldDiv[i].remove();
    }

    if (this.editorCSSCode) {
      const studioClass = '.studio-preview';
      const styles = this.editorCSSCode.split('\n\n').reduce((acc, curr) => {
        return acc + `${studioClass} ${curr}\n`;
      }, '');
      const oStyle = document.createElement('style');
      oStyle.textContent = styles;
      oStyle.setAttribute('class', 'vmap-form-css');
      document.getElementsByTagName('head')[0].appendChild(oStyle);
    }
  }
}
