import { Component, OnInit, Input } from '@angular/core';
import { environment } from 'src/environments/environment';
import { VitisCheck, VitisRequest } from 'src/app/vitis/core/services/utils/vitis-utils';
import { TranslateService } from '@ngx-translate/core';

@Component({
  selector: 'app-studio-select-table-database-modal',
  templateUrl: './select-table-database-modal.component.html',
  styleUrls: ['./select-table-database-modal.component.scss']
})
export class StudioSelectTableDatabaseModalComponent implements OnInit {

  public aLocalDatabases: any = [];
  public aLocalDatabasesSchemas: any = [];
  public aLocalSchemaTables: any = [];
  public aTableAttributes: any = [];
  public aTableData: any = [];

  public oDataSource: any = {
    type: 'web_service',
    ressource_id: 'vitis/genericqueries',
    parameters: {
      filter: {
        relation: 'AND',
        operators: []
      }
    }
  };

  private aFieldsList: Array<any>;

  /**
   * Constructeur
   */
  constructor(
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private translate: TranslateService
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit() {
    // Chargement de la liste des bases de données locales.
    this.loadLocalDatabases();
  }

  /**
   * Chargement de la liste des bases de données locales.
   * @returns void
   */
  public loadLocalDatabases = function () {
    // Vide l'aperçu de la table.
    this.aTableAttributes.length = 0;
    this.aTableData.length = 0;

    // Vide les données.
    this.aLocalSchemaTables.length = 0;
    this.aLocalDatabasesSchemas.length = 0;
    this.aLocalDatabases.length = 0;

    // Lancement de la requête
    this.vitisRequest.ajaxRequestPromise({
      method: 'GET',
      url: environment.properties.app_api_url + '/vitis/genericqueries/databases',
      params: {
        order_by: 'database'
      },
      httpMethodOverride: true,
    }).then((resp: any) => {
      if (resp.data) {
        if (resp.data.status === 1) {
          if (this.vitisCheck.isDefAndNotNull(resp.data)) {
            this.aLocalDatabases = resp.data.data;
          }
        }
      }
    });
  };

  /**
   * Chargement de la liste des schémas d'une base de données.
   * @param  {string} sDatabaseName
   * @returns void
   */
  public loadDatabasesSchemas = function (sDatabaseName: string) {
    // Vide l'aperçu de la table.
    this.aTableAttributes.length = 0;
    this.aTableData.length = 0;

    // Vide les données.
    this.aLocalSchemaTables.length = 0;
    this.aLocalDatabasesSchemas.length = 0;

    // Lancement de la requête
    this.vitisRequest.ajaxRequestPromise({
      method: 'GET',
      url: environment.properties.app_api_url + '/vitis/genericqueries/' + sDatabaseName + '/schemas',
      params: {
        order_by: 'schema_name'
      },
      httpMethodOverride: true,
    }).then((resp: any) => {
      if (resp.data) {
        if (resp.data.status === 1) {
          if (this.vitisCheck.isDefAndNotNull(resp.data)) {
            this.aLocalDatabasesSchemas = resp.data.data;
          }
        }
      }
    });
  };

  /**
   * Chargement de la liste des tables d'un schéma.
   * @param  {string} sDatabaseName
   * @param  {string} sSchemaName
   * @returns void
   */
  public loadSchemaTables = function (sDatabaseName: string, sSchemaName: string) {
    // Vide l'aperçu de la table.
    this.aTableAttributes.length = 0;
    this.aTableData.length = 0;

    // Vide les données.
    this.aLocalSchemaTables.length = 0;

    // Lancement de la requête
    this.vitisRequest.ajaxRequestPromise({
      method: 'GET',
      url: environment.properties.app_api_url + '/vitis/genericqueries/tables',
      params: {
        schema: sSchemaName,
        database: sDatabaseName,
        order_by: 'table_name'
      },
      httpMethodOverride: true,
    }).then((resp: any) => {
      if (resp.data) {
        if (resp.data.status === 1) {
          if (this.vitisCheck.isDefAndNotNull(resp.data) &&
            this.vitisCheck.isDefAndNotNull(resp.data.data)) {
            this.aLocalSchemaTables = resp.data.data;
          }
        }
      }
    });
  };

  /**
   * Actions si la base de données sélectionnée change.
   * @returns void
   */
  public onDatabaseChange() {
    // Vide les données du schéma et de la table sélectionnés.
    this.aLocalDatabasesSchemas.length = 0;
    this.aLocalSchemaTables.length = 0;
    delete this.oDataSource.parameters.schema;
    delete this.oDataSource.parameters.table;

    // Charge les chémas
    if (this.vitisCheck.isString(this.oDataSource.parameters.database) &&
      this.oDataSource.parameters.database !== '') {
      this.loadDatabasesSchemas(this.oDataSource.parameters.database);
    }
  }

  /**
   * Actions si le schéma sélectionné change.
   * @returns void
   */
  public onSchemaChange() {
    // Vide les données de la table sélectionnés.
    this.aLocalSchemaTables.length = 0;
    delete this.oDataSource.parameters.table;

    // Charge les tables
    if (this.vitisCheck.isString(this.oDataSource.parameters.database) &&
      this.vitisCheck.isString(this.oDataSource.parameters.schema) &&
      this.oDataSource.parameters.database !== '' &&
      this.oDataSource.parameters.schema !== '') {
      this.loadSchemaTables(this.oDataSource.parameters.database, this.oDataSource.parameters.schema);
    }
  }

  /**
   * Chargement de la liste des colonnes de la table sélectionnée.
   * @returns void
   */
  public validTableSelection() {
    // Chargement de la liste des colonnes pour le studio.
    this.vitisRequest.ajaxRequestPromise({
      method: 'GET',
      url: environment.properties.app_api_url + '/vitis/genericqueries/columns',
      params: {
        database: this.oDataSource.parameters.database,
        schema: this.oDataSource.parameters.schema,
        table: this.oDataSource.parameters.table
      },
    }).then((resp: any) => {
      if (resp.data) {
        if (resp.data.status === 1) {
          if (this.vitisCheck.isDefAndNotNull(resp.data)) {
            this.aFieldsList = resp.data.data;
            // Evènement retournant la liste des colonnes de la table sélectionnée.
            (<any>$('#studio_select_table_database_modal')).trigger( 'table.fields.list', JSON.stringify(this.aFieldsList));
          }
        }
      }
    });
  }
}
