import { Component, OnInit, ViewChild, OnDestroy, Input, OnChanges, SimpleChanges, EventEmitter } from '@angular/core';
import { DataGridComponent } from '../data-grid/data-grid.component';
import { environment } from 'src/environments/environment';
import { WorkspaceListOptions, WorkspaceListAutorefreshOptions, WorkspaceListEmbeddedOptions } from './workspacelist.model';
import { ActivatedRoute, Router } from '@angular/router';
import { VitisCheck, LocalStorage, SessionStorage, VitisRequest, vitisAlert } from 'src/app/vitis/core/services/utils/vitis-utils';
import { TranslateService } from '@ngx-translate/core';
import { AuthService } from 'src/app/vitis/core/services/auth/auth.service';
import { DataGridOptions } from 'src/app/vitis/data-grid/data-grid.model';
import moment from 'moment';
import { FormreaderComponent } from '../formreader/formreader.component';
import { fromEvent, Subscription } from 'rxjs';

declare let $: any;

@Component({
  selector: 'app-workspacelist',
  templateUrl: './workspacelist.component.html',
  styleUrls: ['./workspacelist.component.scss']
})
export class WorkspacelistComponent implements OnInit, OnDestroy, OnChanges {

  // Utilisé pour piloter workspacelist depuis un composant parent et sans passer par l'URL
  @Input() public embeddedOptions: WorkspaceListEmbeddedOptions;

  @ViewChild(DataGridComponent, { static: true }) oDataGrid: DataGridComponent;

  // Options de la grille de données.
  public oDataGridOptions: DataGridOptions = {
    datatableOptions: {},
    datagridOptions: WorkspaceListOptions
  };

  // Largeur de page
  public pageLength: Number = environment.properties.rows_per_page;

  // Mode en cours
  public mode: any;
  public oParams: any;

  // Filtre en cours
  public oFilter: any = {
    type: 'custom',
    nodes: []
  };

  // Types de filtre disponible
  public oFilterTypes: any = {
    custom: 'VITIS_WORKSPACELIST_FILTER_CUSTOM',
    form: 'VITIS_WORKSPACELIST_FILTER_FORM'
  };

  public sDefaultFilterColumn: string;

  // Résultat
  public oResults: any = {
    recordsTotal: 0,
    recordsFiltered: 0,
    data: [],
    filter: ''
  };

  // Paramètres passés dans l'URL
  public urlParams_: any;

  // Paramètres à passer à la requête
  public datatableRessourceParams_: any;

  public onRefresh = new EventEmitter<any>();

  // Modal avec un formreader
  public modalTitle: string = '';
  @ViewChild(FormreaderComponent, { static: false }) oFormReader: FormreaderComponent;

  // Modale affichée si aucun enregistrement dans la liste.
  public bDisplayNoRecordModal: boolean = false;
  public sNoRecordLabel: string = 'VITIS_WORKSPACELIST_NO_RECORD';
  public sInsertButtonLabel: string = 'VITIS_WORKSPACELIST_INSERT_BTN';

  // Variables privées
  private intervalId_: any;
  private localStorageValues_: any;
  private sessionStorageValues_: any;
  private clickedCell_: any;
  private queryCache_: any;

  private subscriptions: Subscription[] = [];

  /**
   * Constructeur
   */
  constructor(
    private route: ActivatedRoute,
    private vitisCheck: VitisCheck,
    private router: Router,
    private localStorage: LocalStorage,
    private sessionStorage: SessionStorage,
    private vitisRequest: VitisRequest,
    private translate: TranslateService,
    private auth: AuthService
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit() {

    // Nombre d'éléments à afficher sur la grille
    this.pageLength = Number(environment.properties.rows_per_page);

    // Permet à oDataGrid de connaitre les colonnes etc..
    this.oDataGrid.oDataGridOptions = this.oDataGridOptions;

    // Mode normal (non embarqué)
    if (!this.vitisCheck.isDefAndNotNull(this.embeddedOptions)) {

      // Paramètres de la route en cours d'utilisation
      this.route.params.subscribe(oParams => {
        this.oParams = oParams;
      });

      // Mode en cours d'utilisation
      this.route.data.subscribe(mode => {
        this.mode = mode;

        // options app.json
        if (mode.options) {
          this.oDataGridOptions.datagridOptions = mode.options;

          // Texte optionnel pour la modale de création (si aucun enregistrement dans la liste).
          if (this.vitisCheck.isDefAndNotNull(mode.options.no_record_modal)) {
            if (this.vitisCheck.isDefAndNotNull(mode.options.no_record_modal.label)) {
              this.sNoRecordLabel = mode.options.no_record_modal.label;
            }
            if (this.vitisCheck.isDefAndNotNull(mode.options.no_record_modal.button_label)) {
              this.sInsertButtonLabel = mode.options.no_record_modal.button_label;
            }
          }
        }

        // Récupère les paramètres en cookies --> dans data-grid ?
        this.initCookies();

        // Initialise les filtres
        this.initFilters();

        if (!this.vitisCheck.isDefAndNotNull(this.sessionStorage.get('userDisplays'))) {
          this.vitisRequest.ajaxRequest({
            method: 'GET',
            url: environment.properties.app_api_url + '/vitis/list_settings',
            success: (resp: any) => {
              this.sessionStorage.set('userDisplays', resp.data.data);
              // Initialise la grille de données
              this.initDataGrid();
            },
            catch: (e: any) => {

            }
          });
        } else {
          // Initialise la grille de données
          this.initDataGrid();
        }
      });

      // Survéillance du filtre dans les paramètres
      this.route.queryParams.subscribe(oParams => {

        // mémorise les paramètres URL
        this.urlParams_ = oParams;

        // Set la page en cours
        if (this.urlParams_ &&
          this.urlParams_.wkgstart &&
          this.urlParams_.wkglength && this.oDataGrid) {
          const page = this.urlParams_.wkgstart / this.urlParams_.wkglength;
          this.oDataGrid.onChangeParamsTable(this.urlParams_.wkglength, page + 1);
        }

        // Applique le filtre
        if (this.oFilter &&
          this.oFilter.nodes &&
          this.oFilter.nodes.length === 0) {
          if (this.vitisCheck.isDefAndNotNull(oParams && oParams.q)) {
            const { aNodes, sRelation } = this.getFilterByQuery_(oParams.q);
            if (aNodes.length > 0 &&
              this.oFilter.nodes.length === 0) {
              this.oFilter.nodes = aNodes;
              this.oFilter.relationOperator = sRelation;
              this.applyFilter();
            }
          }
        }
      });
    } else {
      if (this.embeddedOptions &&
        this.embeddedOptions.events &&
        this.embeddedOptions.events.onRefresh &&
        this.embeddedOptions.events.onRefresh.emit) {
        if (this.onRefresh &&
          this.onRefresh.subscribe) {
          this.onRefresh.subscribe(() => {
            this.embeddedOptions.events.onRefresh.emit();
          });
        }
      }
    }
  }

  /**
   * Appelé lorsqu'il y a des changements sur les input
   */
  ngOnChanges(changes: SimpleChanges) {

    if (this.vitisCheck.isDefAndNotNull(changes.embeddedOptions)) {
      if (this.embeddedOptions.mode) {
        this.mode = this.embeddedOptions.mode;

        // options app.json
        if (this.embeddedOptions.mode.options) {
          this.oDataGridOptions.datagridOptions = this.embeddedOptions.mode.options;
        }

        // Récupère les paramètres en cookies --> dans data-grid ?
        this.initCookies();

        // Initialise les filtres
        this.initFilters();

        // Initialise la grille de données
        this.initDataGrid();
      }
    }
  }

  /**
   * Destruction du composant
   */
  ngOnDestroy() {
    // Termine l'intervale de rechargement
    if (this.intervalId_) {
      clearInterval(this.intervalId_);
    }

    // Vide le filtre dans l'URL
    this.clearUrlParams_();

    // Détruit les subscribe en fermeture du composant
    this.subscriptions.forEach((subscription) => subscription.unsubscribe());
  }

  /**
   * Applique le filtre à la grille
   */
  public applyFilter(): void {

    // Filtre formaté pour la requête API
    this.oFilter.value = this.vitisRequest.getAPIFilterFromNodes(this.oFilter.nodes, this.oFilter.relationOperator);

    // Rafraichit la vue
    if (this.oDataGrid) {
      this.oDataGrid.reloadTable();
    }

    // Mémorisation du filtre dans la session
    this.sessionStorage.add(this.mode.name, {
      filter: this.oFilter
    });

    // Mémorisation du filtre dans l'url
    this.setUrlParams_();
  }

  /**
   * Initialise les filtres
   */
  private initFilters(): void {

    if (this.vitisCheck.isDefAndNotNull(this.mode)) {
      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {

        this.oFilter.type = 'custom';

        if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.search)) {

          // Recherche custom
          if (this.oDataGridOptions.datagridOptions.search.custom !== true) {
            delete this.oFilterTypes.custom;
          }

          // Recherche par formulaire
          if (this.oDataGridOptions.datagridOptions.search.form !== true) {
            delete this.oFilterTypes.form;
          }

          // Recherche par formulaire
          if (this.oDataGridOptions.datagridOptions.search.default === 'custom') {
            this.oFilter.type = 'custom';
          }
          if (this.oDataGridOptions.datagridOptions.search.default === 'form') {
            this.oFilter.type = 'form';
          }
        } else {
          delete this.oFilterTypes.form;
        }
      }
    }

    // Nombre de possibilités
    this.oFilter.nbFilterTypes = Object.keys(this.oFilterTypes).length;

    const oQueryParams: any = this.route.queryParams;
    if (!(oQueryParams && oQueryParams.value && oQueryParams.value.q)) {
      if (this.vitisCheck.isDefAndNotNull(this.sessionStorageValues_.filter)) {
        this.oFilter = this.sessionStorageValues_.filter;
      }
    }

    // Gestion du parent en cas de workspacelist dans une section
    if (this.mode.options &&
      this.mode.options.parent_id_field &&
      this.oParams &&
      this.oParams.id) {

      if (this.mode.options.disable_filter_parent !== true) {
        // Ajout d'un filtre
        if (this.vitisCheck.isArray(this.oFilter.nodes)) {
          this.oFilter.nodes.push({
            column: {
              field: this.mode.options.parent_id_field,
              translateTitle: 'Id',
            },
            compareOperator: '=',
            value: String(this.oParams.id)
          });
        }
      }
      this.oFilter.value = this.vitisRequest.getAPIFilterFromNodes(this.oFilter.nodes, this.oFilter.relationOperator);
    }

    // Gestion sDefaultFilterColumn
    this.sDefaultFilterColumn = null;
    if (this.mode.options &&
      this.mode.options.search &&
      this.mode.options.search.defaultColumn) {
      this.sDefaultFilterColumn = this.mode.options.search.defaultColumn;
    }

    // Mémorisation du filtre dans l'url
    this.setUrlParams_();
  }

  /**
   * Parse les params envoyés par DataTables pour créer des paramètres interprétables par l'API Vitis
   * @param  {any} params
   * @returns any
   */
  private getRequestParams_(params: any): any {

    const requestParams: any = {};

    // Order
    if (this.vitisCheck.isDefAndNotNull(params.order)) {
      if (this.vitisCheck.isDefAndNotNull(params.order[0])) {
        if (this.vitisCheck.isObject(params.order[0])) {
          if (this.vitisCheck.isString(params.order[0].dir)) {
            requestParams.sort_order = params.order[0].dir.toUpperCase();
          }
          if (this.vitisCheck.isNumber(params.order[0].column)) {
            const iColumnId = params.order[0].column;
            const oSortColumn = this.getOrderColumn_(iColumnId, params.columns);
            if (this.vitisCheck.isDefAndNotNull(oSortColumn)) {
              if (this.vitisCheck.isString(oSortColumn.data)) {
                requestParams.order_by = oSortColumn.data;
              }
            }
          }
        } else {
          const iColumnId = params.order[0];
          const oSortColumn = this.getOrderColumn_(iColumnId, params.columns);
          if (this.vitisCheck.isDefAndNotNull(oSortColumn)) {
            if (this.vitisCheck.isString(oSortColumn.data)) {
              requestParams.order_by = oSortColumn.data;
            }
          }
          requestParams.sort_order = params.order[1];
        }
      }
    }

    // Limit
    if (this.vitisCheck.isNumber(params.length)) {
      requestParams.limit = params.length;
    }

    // Offset
    if (this.vitisCheck.isNumber(params.start)) {
      requestParams.offset = params.start;
    }

    // Filter
    const filters = this.getFilters_();
    if (this.vitisCheck.isDefAndNotNull(filters)) {
      requestParams.filter = filters;
    }

    // Attributs
    if (this.vitisCheck.isDefAndNotNull(params.columns)) {
      // détecte la présence d'attribut supplementaire requis par la liste
      const extra_attributs = (this.oDataGridOptions.datagridOptions.extra_attributs) ? this.oDataGridOptions.datagridOptions.extra_attributs : [];
      requestParams['attributs'] = params.columns.filter(
        (aColumn: any) => this.vitisCheck.isString(aColumn.data)
      ).map(
        (aColumn: any) => aColumn.data
      ).concat(extra_attributs)
        .join('|');

    }

    return requestParams;
  }

  /**
   * Retourne check si la colonne demandée via iColumnId est sortable
   * Si elle l'est alors renvoi la colonne
   * Si elle ne l'est pas alors fait une recherche et trouve la première colonne sortable
   * @param iColumnId
   * @param aCloumns
   * @returns
   */
  private getOrderColumn_(iColumnId: number, aCloumns: any): any {
    if (aCloumns &&
      aCloumns[iColumnId] &&
      aCloumns[iColumnId].sortable !== false) {
      return aCloumns[iColumnId];
    } else {
      for (const col of aCloumns) {
        if (col.sortable !== false) {
          return col;
        }
      }
    }
  }

  /**
   * Écrit le filtre dans l'URL
   */
  public setUrlParams_() {

    /**
     * Paramètres de l'url
     */
    const queryParams: any = {};

    // Filtre
    queryParams.q = null;
    if (this.oFilter) {
      const sQuery = this.getQueryByFilter_(this.oFilter);
      if (sQuery) {
        queryParams.q = sQuery;
      }
    }

    // Page
    if (this.datatableRessourceParams_) {

      // Valeurs vides
      queryParams.wkgstart = null;
      queryParams.wkglength = null;

      // Valeurs réelles
      if (this.datatableRessourceParams_.start > 0 &&
        this.datatableRessourceParams_.length > 0) {
        queryParams.wkgstart = this.datatableRessourceParams_.start;
        queryParams.wkglength = this.datatableRessourceParams_.length;
      }
    }

    // Mode normal (non embarqué)
    if (!this.vitisCheck.isDefAndNotNull(this.embeddedOptions)) {
      this.router.navigate(
        [],
        {
          relativeTo: this.route.parent,
          queryParams,
          queryParamsHandling: 'merge',
        });
    }
  }

  /**
   * Récupère les donéées au format DataTables
   * @param  {any} params
   * @returns Promise
   */
  public getResourceData_(params: any): Promise<any> {
    return new Promise((resolve, reject) => {


      // Mémorise les paramètres
      this.datatableRessourceParams_ = params;

      // Mémorisation des paramètres de la table
      this.localStorage.add(this.mode.name, {
        datatableParams: params,
      });
      this.sessionStorage.add(this.mode.name, {
        datatableParams: params,
      });

      if (params.hasOwnProperty('filter')) {
        this.oFilter = params.filter;
      }

      // Set les paramètres dans l'url
      this.setUrlParams_();

      this.oDataGridOptions.loadingData_ = true;

      // Lance la requête
      this.vitisRequest.ajaxRequest({
        method: 'GET',
        url: this.getRessourceUrl(),
        ajaxLoader: false,
        params: this.getRequestParams_(params),
        success: (resp: any) => {

          this.oDataGridOptions.loadingData_ = false;

          const oReturn = {
            recordsTotal: 0,
            recordsFiltered: 0,
            data: [],
            filter: this.oFilter
          };

          if (this.vitisCheck.isDefAndNotNull(resp.data)) {
            if (this.vitisCheck.isDefAndNotNull(resp.data.status === 1)) {
              if (this.vitisCheck.isArray(resp.data.data)) {
                oReturn.data = resp.data.data;
              }
              if (this.vitisCheck.isDefAndNotNull(resp.data.total_row_number)) {
                oReturn.recordsTotal = resp.data.total_row_number;
                oReturn.recordsFiltered = resp.data.list_count;
              }
            }
          }

          resolve(oReturn);
        },
        error: (err: any) => {
          console.error('error while loading values :', err);

          this.oDataGridOptions.loadingData_ = false;

          const oReturn = {
            recordsTotal: 0,
            recordsFiltered: 0,
            data: [],
            filter: ''
          };

          resolve(oReturn);
        }
      });
    });
  }

  /**
   * Génére l'url pour récupérer les données
   * @returns {string}
   */
  protected getRessourceUrl(): string {
    let sRessourceUrl = this.oDataGridOptions.datagridOptions.ressource;

    // Gestion du parent en cas de workspacelist dans une section
    if (this.mode.options &&
      this.mode.options.parent_id_field &&
      this.oParams &&
      this.oParams.id) {

      if (sRessourceUrl.indexOf(`:${this.mode.options.parent_id_field}`) !== -1) {
        // Remplacement dans l'URL
        sRessourceUrl = sRessourceUrl.replace(`:${this.mode.options.parent_id_field}`, String(this.oParams.id));
      }
    }

    return environment.properties.app_api_url + '/' + sRessourceUrl;
  }

  /**
   * Initialise localStorageValues_ et sessionStorageValues_
   */
  public initCookies() {

    this.localStorageValues_ = this.localStorage.get(this.mode.name);
    this.sessionStorageValues_ = this.sessionStorage.get(this.mode.name);
    if (!this.vitisCheck.isDefAndNotNull(this.localStorageValues_)) {
      this.localStorageValues_ = {};
    }
    if (!this.vitisCheck.isDefAndNotNull(this.sessionStorageValues_)) {
      this.sessionStorageValues_ = {};
    }
  }

  /**
   * Initialise la grille de données
   */
  public initDataGrid() {

    // Paramètres de la grille de données.
    const datatableOptions: any = {
      lengthMenu: this.getLengthMenu_(),  // Tableau choix du nombre d'éléments par page
      pageLength: this.getPageLength_(),  // Nombre d'éléments par page
      order: [0, 'desc'], // Ordre
      displayStart: 0,  // Offset
      rowId: '',
      autorefresh: null,
      select: true,
      selectMode: this.mode.options.singleSelect ? 'single' : 'multi',
      colReorder: this.mode.options.colReorder || false,
      ajax: (data: any, callback: any, settings: any) => {

        // Cache utilisé pour le rafraichissement auto
        this.queryCache_ = {
          options: data,
          date: Date.now(),
        };

        this.getResourceData_(data)
          .then((oResponse: any) => {

            callback(oResponse);

            // Remplit le cache
            this.queryCache_.result = oResponse;

            // event onRefresh
            if (this.onRefresh) {
              this.onRefresh.emit();
            }

            // Si aucun enregistrement : affichage de la modale de création si l'onglet à un mode "insert".
            if (this.mode.options.insert) {
              if (oResponse.recordsTotal == 0) {
                this.bDisplayNoRecordModal = true;
              } else {
                this.bDisplayNoRecordModal = false;
              }
            }

          })
          .catch(() => {
            callback({
              recordsTotal: 0,
              recordsFiltered: 0,
              data: [],
              filter: ''
            });
          });
      }
    };

    if (this.mode.options.colReorder === true) {
      const userDisplays = this.sessionStorage.get('userDisplays');
      let currUserDisplays = [];
      let displayIdDefault = 0;
      if (this.vitisCheck.isDefAndNotNull(userDisplays)) {
        currUserDisplays = userDisplays.filter(display => display.route === this.mode.route).map(display => ({
          ...display,
          definition: JSON.parse(display.definition)
        }));
        if (currUserDisplays.find(display => display.isdefault === true)) {
          displayIdDefault = currUserDisplays.find(display => display.isdefault === true).list_settings_id;
        }
        if (!this.sessionStorage.get(this.mode.name) || !this.sessionStorage.get(this.mode.name).filter) {
          const defaultDisplay = currUserDisplays.find(display => display.list_settings_id === displayIdDefault);
          if (this.vitisCheck.isDefAndNotNull(defaultDisplay) &&
            this.vitisCheck.isDefAndNotNull(defaultDisplay.definition.filter) &&
            this.vitisCheck.isDefAndNotNull(defaultDisplay.definition.filter.value)) {
            this.oFilter = defaultDisplay.definition.filter;
          }
        }

        datatableOptions.listSettings = {
          route: this.mode.route,
          userDisplays: currUserDisplays,
          isdefault: displayIdDefault
        };
      }
    }

    // Options app.json
    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.order_by)) {
        // Si le order_by est le numéro de la colonne
        if (this.vitisCheck.isNumber(this.oDataGridOptions.datagridOptions.order_by)) {
          datatableOptions.order[0] = this.oDataGridOptions.datagridOptions.order_by;
          // Si le order_by est le nom de la colonne
        } else if (this.vitisCheck.isString(this.oDataGridOptions.datagridOptions.order_by) &&
          this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.columns)) {
          datatableOptions.order[0] = this.oDataGridOptions.datagridOptions.columns.findIndex(column => {
            return column.field === this.oDataGridOptions.datagridOptions.order_by;
          });
          // Si le order_by est un tableau
        } else if (this.vitisCheck.isArray(this.oDataGridOptions.datagridOptions.order_by)) {
          datatableOptions.order = this.oDataGridOptions.datagridOptions.order_by;
        }
      }
      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.autorefresh)) {
        datatableOptions.autorefresh = this.oDataGridOptions.datagridOptions.autorefresh;
      }
      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.id_field)) {
        datatableOptions.rowId = this.oDataGridOptions.datagridOptions.id_field;
      }
    }

    // LocalStorage
    if (this.localStorageValues_ &&
      this.localStorageValues_.datatableParams) {

      // Order
      if (this.localStorageValues_.datatableParams.order &&
        this.localStorageValues_.datatableParams.order[0] &&
        this.localStorageValues_.datatableParams.order[0].column &&
        this.localStorageValues_.datatableParams.order[0].dir) {
        datatableOptions.order =
          [this.localStorageValues_.datatableParams.order[0].column, this.localStorageValues_.datatableParams.order[0].dir];
      }
    }

    // SessionStorage
    if (this.sessionStorageValues_ &&
      this.sessionStorageValues_.datatableParams) {

      // start
      if (this.vitisCheck.isDefAndNotNull(this.sessionStorageValues_.datatableParams.start)) {
        datatableOptions.displayStart = this.sessionStorageValues_.datatableParams.start;
      }
    }

    // Paramètres URL
    if (this.urlParams_ &&
      this.urlParams_.wkgstart &&
      this.urlParams_.wkglength) {
      datatableOptions.displayStart = Number(this.urlParams_.wkgstart);
    }

    // Charge le JS
    this.loadModeJavaScript_(this.mode);

    // Rafraichissement auto
    if (datatableOptions.autorefresh &&
      datatableOptions.autorefresh.interval) {
      this.intervalId_ = setInterval(() => {
        this.autoRefresh_(datatableOptions.autorefresh);
      }, datatableOptions.autorefresh.interval);
    }

    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.columns)) {

        let oDatagridColumns = [];
        if (this.vitisCheck.isArray(this.oDataGridOptions.datagridOptions.columns)) {
          oDatagridColumns = this.oDataGridOptions.datagridOptions.columns.filter(column => {
            // Privilège(s) obligatoire(s) pour afficher la colonne.
            if (this.vitisCheck.isDefAndNotNull(column.privilege)) {
              if (this.vitisCheck.isString(column.privilege)) {
                column.privilege = [column.privilege];
              }
              if (this.vitisCheck.isArray(column.privilege)) {
                const aUserPrivileges = this.auth.getPrivileges();
                for (let i = 0; i < column.privilege.length; i++) {
                  const sPrivilege = column.privilege[i];
                  if (this.vitisCheck.isString(sPrivilege) &&
                    aUserPrivileges.indexOf(sPrivilege) !== -1) {
                    return true;
                  }
                }
              } else {
                console.error('Column privilege not string or array', column.privilege);
              }
              return false;
            }
            return true;
          });
        }

        // Traitement des colonnes
        for (let i = 0; i < oDatagridColumns.length; i++) {

          const column = oDatagridColumns[i];

          // Index par défaut de la colonne
          column.index = i;

          // Traduction
          column.translateTitle = this.translate.instant(column.title);

          column.data = column.field;

          // Visibilité de la colonne
          column.visible = column.visible === false ? false : true;

          // Colonne type
          column.type = column.template ? 'html' : 'string';

          // Alignement du contenu de la colonne.
          column.className = 'table-align-' + (column.align ? column.align : 'left');

          // Service web pour filtres
          if (column.searchWebService === undefined &&
            column.searchSuggestions !== false &&
            !this.vitisCheck.isArray(column.searchSuggestions)) {
            column.searchWebService = {
              resource: this.getRessourceUrl(),
              filterOnComplete: true,
              params: {
                attributs: column.field,
                order_by: column.field,
                distinct: true,
                sort_order: 'ASC',
                limit: 10,
              },
              label: column.field,
              value: column.field,
            };
          }
        }

        this.oDataGridOptions.datagridOptions.columns = oDatagridColumns;
      }
    }

    this.oDataGridOptions.datatableOptions = datatableOptions;

    // Initialise les boutons
    this.initButtons();

    // Initialise les boutons icons
    this.initIconButtons();

    setTimeout(() => {
      if (this.vitisCheck.isNumber(this.mode.options.height)) {
        this.oDataGridOptions.datatableOptions.tableHeight = this.mode.options.height;
      } else {
        setTimeout(() => {
          this.calcTableHeight();
        }, 100);
      }
      this.oDataGrid.initTable(this.oDataGridOptions);

      // Select event
      this.oDataGrid.selectRowEvent.subscribe(() => {
        if (this.embeddedOptions &&
          this.embeddedOptions.events &&
          this.embeddedOptions.events.onSelect &&
          this.embeddedOptions.events.onSelect.emit) {
          setTimeout(() => {
            const rows: any = this.oDataGrid.getSelectedRows_();
            rows.clickedCell_ = this.clickedCell_;
            this.embeddedOptions.events.onSelect.emit(rows);
          });
        }
      });

      // Deselect event
      this.oDataGrid.deselectRowEvent.subscribe(() => {
        if (this.embeddedOptions &&
          this.embeddedOptions.events &&
          this.embeddedOptions.events.onDeselect &&
          this.embeddedOptions.events.onDeselect.emit) {
          setTimeout(() => {
            const rows: any = this.oDataGrid.getSelectedRows_();
            rows.clickedCell_ = this.clickedCell_;
            this.embeddedOptions.events.onDeselect.emit(rows);
          });
        }
      });

      // Recalcule dynamiquement la taille du tableau au changement de taille de la fenêtre
      if (!this.vitisCheck.isNumber(this.mode.options.height)) {
        this.subscriptions.push(fromEvent(window, 'resize').subscribe(() => {
          setTimeout(() => {
            this.calcTableHeight();
            if (this.vitisCheck.isDefAndNotNull(this.oDataGrid)) {
              this.oDataGrid.ajustTableHeight();
            }
          }, 100);
        }));
      }
    });
  }

  /**
   * Calcul la hauteur du data grid
   */
  private calcTableHeight() {
    const dashboardHeight = $('#dashboard-content').height();
    const titleHeight = $('.mode-title').height();
    const filterHeight = $('.grid-filter').height();
    const menusHeight = 150;
    const tableHeight = dashboardHeight - (menusHeight + filterHeight + titleHeight) + 4;
    this.oDataGridOptions.datatableOptions.tableHeight = tableHeight;
  }

  /**
   * Retourne le filtre au format URL
   * @param oFilter
   */
  public getQueryByFilter_(oFilter: any): string {

    let sQuery: any = null;
    if (oFilter.nodes && oFilter.nodes.length > 0) {
      for (let i = 0; i < oFilter.nodes.length; i++) {
        const node = oFilter.nodes[i];
        if (i === 0) {
          sQuery = '';
        } else {
          sQuery += '||';
        }

        if (node.column.field && node.compareOperator) {
          let sNodeFilter = `${node.column.field}|${node.compareOperator}`;

          if (node.value) {
            sNodeFilter += `|${node.value}`;

            if (node.column.searchWebService && node.value != node.displayValue) {
              sNodeFilter += `|${node.displayValue}`;
            }
          }

          sQuery += encodeURIComponent(sNodeFilter);
        }
      }
      if (oFilter.relationOperator) {
        sQuery += encodeURIComponent(`|||${oFilter.relationOperator}`);
      }
    }

    return sQuery;
  }

  /**
   * Retourne les nodes du filtre en fonction d'un filtre URL
   * @param sQuery
   */
  public getFilterByQuery_(sQuery: string): any {

    const aNodes = [];
    let sRelation = 'AND';

    if (!this.vitisCheck.isDefAndNotNull(sQuery)) {
      return;
    }

    sQuery = decodeURIComponent(sQuery);

    // Relation (AND/OR)
    const aRel = sQuery.split('|||');
    if (aRel.length === 2) {
      sRelation = aRel[1];
    }
    sQuery = aRel[0];

    // Opérateurs
    const aOperators = sQuery.split('||');
    for (let i = 0; i < aOperators.length; i++) {

      // Condition
      const aNode = aOperators[i].split('|');
      if (aNode) {
        const field = aNode[0];
        const compareOperator = aNode[1];
        let value: any = aNode[2] ?? '';

        if (field &&
          compareOperator) {

          const column = this.oDataGrid.getColumnByName_(field);
          let displayValue = (aNode.length === 4) ? aNode[3] : value;

          // Gestion valeur à afficher
          if (this.vitisCheck.isArray(column.searchSuggestions)) {
            for (let ii = 0; ii < column.searchSuggestions.length; ii++) {
              const suggestion = column.searchSuggestions[ii];
              if (this.vitisCheck.isDefAndNotNull(suggestion.label) &&
                this.vitisCheck.isDefAndNotNull(suggestion.value)) {
                if (suggestion.value === value) {
                  displayValue = suggestion.label;
                }
              }
            }
          }

          // Gestion dates
          if (column.searchType === 'date') {
            let sDateFormat = 'DD/MM/YYYY';
            const oUser = this.auth.getUserInfos();
            if (this.vitisCheck.isDefAndNotNull(oUser)) {
              if (this.vitisCheck.isDefAndNotNull(oUser.formatdate_id)) {
                sDateFormat = oUser.formatdate_id;
              }
            }
            displayValue = moment(Number(value)).format(sDateFormat + ' HH:mm');
          }

          // Gestion IN
          if (compareOperator === 'IN') {
            value = value.split(',');
          }

          if (column) {
            aNodes.push({
              column,
              compareOperator,
              compareOperatorTranslate: this.translate.instant(`VITIS_WORKSPACELIST_FILTER_COMPARE_OPERATOR_${compareOperator}`),
              value,
              displayValue
            });
          }
        }
      }
    }



    return { aNodes, sRelation };
  }

  /**
   * Récupère les filtres du mode
   */
  public getParsedFilters_() {
    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.filter)) {
      const oFilter = this.oDataGridOptions.datagridOptions.filter;

      if (this.vitisCheck.isDefAndNotNull(oFilter.lang) && oFilter.lang === '') {
        oFilter.lang = environment.properties.language;
      }

      const oGivenFilter = {
        relation: 'AND',
        operators: []
      };
      for (const key in oFilter) {
        if (oFilter.hasOwnProperty(key)) {
          const sOperator: string = (Array.isArray(oFilter[key])) ? 'IN' : '=';
          oGivenFilter.operators.push({
            column: key,
            compare_operator: sOperator,
            value: oFilter[key]
          });
        }
      }

      return oGivenFilter;
    }
  }

  /**
   * Lance le rafraichissement automatique
   * @param options
   */
  private autoRefresh_(options: WorkspaceListAutorefreshOptions) {

    if (options &&
      options.interval) {

      // Ne recharge pas automatiquement si la page n'est pas visible
      if (document.hidden) {
        return;
      }

      // Ne lance pas le rechargement automatique si un chargement a déjà été lancé récemment
      if (this.queryCache_ &&
        this.queryCache_.date &&
        Date.now() - this.queryCache_.date < options.interval - 1000) {
        return;
      }

      // Lance une requête et compare les résultats avant de lancer la compilation du template
      if (options.compare_columns) {
        if (this.queryCache_ &&
          this.queryCache_.options &&
          this.queryCache_.result &&
          this.queryCache_.result.data) {
          this.getResourceData_(this.queryCache_.options).then((oResponse: any) => {

            let bSameObjects = true;
            if (oResponse &&
              oResponse.data) {

              for (let i = 0; i < oResponse.data.length; i++) {
                const cacheRow = this.queryCache_.result.data[i];
                const loadedRow = oResponse.data[i];

                for (let ii = 0; ii < options.compare_columns.length; ii++) {
                  const column = options.compare_columns[ii];

                  if (cacheRow[column] !== loadedRow[column]) {
                    bSameObjects = false;
                  }
                }
              }
            }

            if (bSameObjects === false) {
              this.oDataGrid.reloadTable();
            }
          });
        }
      } else {
        this.oDataGrid.reloadTable();
      }
    }
  }

  /**
   * Affiche le formulaire correspondant
   * @param  {any} oRow
   * @param  {String} sFormName
   */
  private displayRowForm_(oRow: any, sFormName: String) {

    // Mode embedded
    if (this.vitisCheck.isDefAndNotNull(this.embeddedOptions)) {
      if (sFormName === 'display') {
        if (this.embeddedOptions.events &&
          this.embeddedOptions.events.onDisplay &&
          this.embeddedOptions.events.onDisplay.emit) {
          this.embeddedOptions.events.onDisplay.emit(oRow);
        }
      }
      if (sFormName === 'update') {
        if (this.embeddedOptions.events &&
          this.embeddedOptions.events.onEdit &&
          this.embeddedOptions.events.onEdit.emit) {
          this.embeddedOptions.events.onEdit.emit(oRow);
        }
      }
    } else {
      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
        if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.id_field)) {
          if (this.vitisCheck.isDefAndNotNull(oRow)) {
            if (this.vitisCheck.isDefAndNotNull(oRow[this.oDataGridOptions.datagridOptions.id_field])) {
              if (this.vitisCheck.isDefAndNotNull(sFormName)) {

                const sId = oRow[this.oDataGridOptions.datagridOptions.id_field];
                const aModeRoute = this.mode.route.split('/');

                if (this.vitisCheck.isDefAndNotNull(this.mode.sections) &&
                  this.vitisCheck.isDefAndNotNull(this.mode.options.map_field)) { // Avec sections et map_field
                  // Recupère les données afin d'avoir' la valeur du map_field
                  this.vitisRequest.ajaxRequestPromise({
                    method: 'GET',
                    url: this.vitisRequest.getApiRessourceUrl(this.mode.options.ressource, sId)
                  }).then((resp) => {
                    if (resp.data) {
                      if (resp.data.status === 1) {
                        if (resp.data.data) {
                          this.mode.options.values = resp.data.data;
                          const sections = this.mode.sections[this.mode.options.values[this.mode.options.map_field]];
                          this.navigateFirstSection(sections, sId, sFormName, aModeRoute);
                        }
                      }
                    }
                  });
                } else {
                  this.navigateFirstSection(this.mode.sections, sId, sFormName, aModeRoute);
                }
              }
            }
          }
        }
      }
    }
  }

  /**
   * // Navigue vers le bon url lors de l'event sur la grid
   * @param sections // mode sections
   * @param sId // id de l'élément
   * @param sFormName // formulaire choisi 'update|insert...'
   * @param aModeRoute
   */
  private navigateFirstSection(sections, sId: number, sFormName: any, aModeRoute: any) {
    if (sections) {
      if (sections[0] && sections[0].route) {
        const aSectionRoute = sections[0].route.split('/');
        for (let i = aSectionRoute.length - 1; i >= 0; i--) {
          if (aSectionRoute[i] === ':id') {
            aSectionRoute[i] = sId;
          }
          if (aSectionRoute[i] === ':form') {
            aSectionRoute[i] = sFormName;
          }
          if (aModeRoute.indexOf(aSectionRoute[i]) !== -1) {
            aSectionRoute.splice(i, 1);
          }
        }
        this.router.navigate(aSectionRoute, { relativeTo: this.route.parent });
      }
    } else {
      this.router.navigate([sId, sFormName], { relativeTo: this.route.parent });
    }
  }

  /**
   * Nettoie le filtre dans l'URL
   */
  private clearUrlParams_() {

    this.router.navigate(
      [],
      {
        relativeTo: this.route.parent,
        queryParams: {
          wkglength: null,
          wkgstart: null,
          q: null
        },
        queryParamsHandling: 'merge',
      }
    );
  }

  /**
   * Affiche le formulaire d'insertion
   * @returns void
   */
  private displayInsertForm_(): void {

    // Mode embedded
    if (this.vitisCheck.isDefAndNotNull(this.embeddedOptions)) {
      if (this.embeddedOptions.events &&
        this.embeddedOptions.events.onCreate &&
        this.embeddedOptions.events.onCreate.emit) {
        this.embeddedOptions.events.onCreate.emit();
      }
    } else {
      if (this.vitisCheck.isDefAndNotNull(this.mode.sections)) {
        const aModeRoute = this.mode.route.split('/');
        let aSectionRoute;
        if (this.vitisCheck.isDefAndNotNull(this.mode.options.map_field)) {
          if (!this.vitisCheck.isArray(this.mode.sections)) {
            aSectionRoute = this.mode.sections[Object.keys(this.mode.sections)[0]][0].route.split('/');
          }
        } else {
          aSectionRoute = this.mode.sections[0] && this.mode.sections[0].route.split('/');
        }
        if (aSectionRoute) {
          for (let i = aSectionRoute.length - 1; i >= 0; i--) {
            if (aSectionRoute[i] === ':form') {
              aSectionRoute[i] = 'insert';
            }
            if (aSectionRoute[i] === ':id') {
              aSectionRoute.splice(i, 1);
            }
            if (aModeRoute.indexOf(aSectionRoute[i]) !== -1) {
              aSectionRoute.splice(i, 1);
            }
          }
          this.router.navigate(aSectionRoute, { relativeTo: this.route.parent });
        }
      } else {
        this.router.navigate(['insert'], { relativeTo: this.route.parent });
      }
    }
  }

  /**
   * Déclenche le processus de dupplication de la ligne sélectionnée
   */
  public cloneSelectedRow(): void {

    const aIds = this.oDataGrid.getSelectedIds_();

    if (!(aIds.length > 0)) {
      vitisAlert.alert({
        type: 'error',
        title: this.translate.instant('LIST_DELETE_NO_SELECTION'),
      });
    } else if (aIds.length > 1) {
      vitisAlert.alert({
        type: 'error',
        title: this.translate.instant('MORE_THAN_ONE_SELECTION'),
      });
    } else {
      this.localStorage.set('idClone', aIds[0]);
      this.displayInsertForm_();
    }
  }

  /**
   * Déclenche le processus de suppression des lignes sélectionnées
   */
  private deleteSelectedRows_(): void {

    const aIds = this.oDataGrid.getSelectedIds_();

    if (!(aIds.length > 0)) {
      vitisAlert.alert({
        type: 'error',
        title: this.translate.instant('LIST_DELETE_NO_SELECTION'),
      });
    } else {

      if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {
        if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.ressource)) {

          // Chemin de la ressource REST
          const sRessourceURL = this.getRessourceUrl();

          vitisAlert.alert({
            type: 'warning',
            title: this.translate.instant('LIST_DELETE_CONFIRM'),
            showCancelButton: true
          }).then((result) => {
            if (result.value) {

              // Mode embedded
              if (this.vitisCheck.isDefAndNotNull(this.embeddedOptions)) {
                if (this.embeddedOptions.events &&
                  this.embeddedOptions.events.onDelete &&
                  this.embeddedOptions.events.onDelete.emit) {
                  this.embeddedOptions.events.onDelete.emit(aIds);
                }
              } else {
                this.runDelete_(sRessourceURL, aIds);
              }
            }
          });
        }
      }
    }
  }

  /**
   * Lance une requête delete sur l'API
   * @param  {string} sRessourceURL
   * @param  {Array<string>} aIds
   * @returns Promise
   */
  private runDelete_(sRessourceURL: string, aIds: Array<string>): Promise<void> {
    return new Promise((resolve, reject) => {

      // Lancement de la requête DELETE
      this.vitisRequest.ajaxRequestPromise({
        method: 'DELETE',
        url: sRessourceURL,
        data: {
          idList: aIds
        }
      }).then((resp) => {
        if (resp.data) {
          if (resp.data.status === 1) {

            // Message succès
            this.translate.get('LIST_DELETE_OK').subscribe((translation) => {
              vitisAlert.fireToast({
                icon: 'success',
                title: translation
              });
              resolve();
            });
          } else {
            reject();
          }
          // Vide les sélections
          this.oDataGrid.clearSelection();
          // Recharge les données
          this.oDataGrid.reloadTable();
        }
      });
    });
  }

  /**
   * Initialise les boutons
   * @returns void
   */
  public initButtons(): void {
    const aButtons: any = [];

    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions)) {

      const aButtonModes: any = ['insert', 'delete', 'clone'];

      // Group bouton action
      const groupButtonAction = {
        type: 'group',
        label: 'VITIS_WORKSPACELIST_ACTIONS_BTN',
        class: 'btn-primary',
        buttons: []
      };

      for (const button of aButtonModes) {
        if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions[button]) &&
          this.hasAccessToButton(this.oDataGridOptions.datagridOptions[button])) {
          if (button === 'insert') {
            aButtons.unshift({
              type: 'button',
              class: 'btn-primary',
              label: 'LIST_BUTTON_ADD',
              icon: 'fa fa-plus',
              event: () => {
                this.displayInsertForm_();
              }
            });
          } else if (button === 'delete') {
            aButtons.unshift({
              type: 'button',
              class: 'btn-default',
              label: 'LIST_BUTTON_DELETE',
              icon: 'fa fa-trash',
              event: () => {
                this.deleteSelectedRows_();
              }
            });
          } else if (button === 'clone') {
            groupButtonAction.buttons.push({
              type: 'button',
              label: 'LIST_BUTTON_CLONE',
              event: () => {
                this.cloneSelectedRow();
              }
            });
          }
        }
      }

      if (groupButtonAction.buttons.length > 0) {
        aButtons.unshift(groupButtonAction);
      }

      // Autres boutons
      if (this.vitisCheck.isArray(this.oDataGridOptions.datagridOptions.buttons)) {
        // Ajoute les boutons
        for (let i = 0; i < this.oDataGridOptions.datagridOptions.buttons.length; i++) {
          const button = this.oDataGridOptions.datagridOptions.buttons[i];
          this.oDataGrid.initButtonEvent_(button, this);
          this.oDataGrid.initButtonDisplay_(button, this);
          aButtons.unshift(button);
        }
      }
    }

    this.oDataGridOptions.buttons = aButtons;
  }

  /**
   * Initialise les boutons présent dans chaque ligne du tableau
   */
  public initIconButtons(): void {
    const aButtons: any = [];
    let buttonUpdateExist = false;
    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.update) &&
      this.hasAccessToButton(this.oDataGridOptions.datagridOptions.update)) {
      aButtons.push({
        type: 'update',
        icon: 'icon-vitis-draw',
        event: (data: any) => {
          this.displayRowForm_(data, 'update');
        }
      });
      buttonUpdateExist = true;
    }

    if (this.vitisCheck.isDefAndNotNull(this.oDataGridOptions.datagridOptions.display) &&
      this.hasAccessToButton(this.oDataGridOptions.datagridOptions.display)) {
      // Vérifie qu'il y a un bouton update
      if (environment.properties.data_grid_update_btn_only === true && buttonUpdateExist) {
        this.oDataGridOptions.datagridOptions.display = false;
      } else {
        aButtons.push({
          type: 'display',
          icon: 'icon-vitis-search',
          event: (data: any) => {
            this.displayRowForm_(data, 'display');
          }
        });
      }
    }

    this.oDataGridOptions.iconButtons = aButtons;
  }

  /**
   * Charge le JS associé au mode
   * @param  {any} oMode
   * @returns Promise
   */
  private loadModeJavaScript_(oMode: any): Promise<boolean> {
    return new Promise((resolve, reject) => {

      // Si le mode contient un JS associé
      if (this.vitisCheck.isDefAndNotNull(oMode.options)) {
        if (this.vitisCheck.isDefAndNotNull(oMode.options.js_path)) {

          // Charge le JS du mode
          this.vitisRequest.loadExternalJs(oMode.options.js_path).then(() => {

            // Lance le JS associé
            if ((<any>window).constructor_form) {
              (<any>window).constructor_form(this, environment, {
                vitisRequest: this.vitisRequest,
                vitisCheck: this.vitisCheck,
                vitisAlert: vitisAlert,
                auth: this.auth,
                translate: this.translate
              });
            }

            resolve(true);
          });
        } else {
          resolve(false);
        }
      } else {
        resolve(false);
      }
    });
  }

  /**
   * Récupère les filtres à appliquer
   */
  private getFilters_() {
    const defaultFilter = this.getParsedFilters_();
    let currentFilter = this.oFilter.value;

    if (this.vitisCheck.isDefAndNotNull(defaultFilter) &&
      (this.vitisCheck.isDefAndNotNull(currentFilter) && Array.isArray(currentFilter.operators))) {
      currentFilter.operators = currentFilter.operators.concat(defaultFilter.operators);
    } else if (this.vitisCheck.isDefAndNotNull(defaultFilter)) {
      currentFilter = defaultFilter;
    }

    return currentFilter;
  }

  /**
   * Retourne le nombre d'éléments à afficher
   */
  private getPageLength_(): Number {
    let pageLength = this.pageLength;

    // SessionStorage
    if (this.sessionStorageValues_ &&
      this.sessionStorageValues_.datatableParams) {

      // length
      if (this.vitisCheck.isDefAndNotNull(this.sessionStorageValues_.datatableParams.length)) {
        pageLength = this.sessionStorageValues_.datatableParams.length;
      }
    }

    // Paramètres URL
    if (this.urlParams_ &&
      this.urlParams_.wkgstart &&
      this.urlParams_.wkglength) {
      pageLength = Number(this.urlParams_.wkglength);
    }

    // Concordance avec le tableau de choix
    const aLengthMenu = this.getLengthMenu_();
    if (aLengthMenu.indexOf(pageLength) === -1) {
      if (pageLength > aLengthMenu[aLengthMenu.length - 1]) {
        pageLength = aLengthMenu[aLengthMenu.length - 1];
      } else {
        pageLength = 20;
      }
    }

    return pageLength;
  }

  /**
   * Retourne le tableau contenant le nombre de lignes affichables
   */
  private getLengthMenu_(): Array<Number> {

    const lengthMenu = [5, 10, 20, 50, 100];

    if (this.oDataGridOptions &&
      this.oDataGridOptions.datagridOptions &&
      this.oDataGridOptions.datagridOptions.max_rows) {

      const iMaxRows = this.oDataGridOptions.datagridOptions.max_rows;

      for (let i = lengthMenu.length - 1; i >= 0; i--) {
        if (lengthMenu[i] > iMaxRows) {
          lengthMenu.splice(i, 1);
        }
      }
    }

    return lengthMenu;
  }

  /**
   * Charge le formulaire
   * @param oConfig
   * @param oConfig.form_path chemin vers le formulaire
   * @param oConfig.mode inser/update/display
   * @param oConfig.eventContainer eventContainer
   */
  public compileFormReaderModal(oConfig: any) {
    if (!this.vitisCheck.isDefAndNotNull(oConfig.form_path)) {
      console.error('Empty form_path for modal');
      return;
    }
    const oFormMode = this.vitisCheck.isDefAndNotNull(oConfig.mode) ? oConfig.mode : 'insert';
    // Chargement formulaire
    const url = oConfig.form_path;
    // Récupère le formulaire
    this.getJSONForm_(oConfig.form_path).then((oForm) => {
      if (this.vitisCheck.isDefAndNotNull(oConfig.eventContainer)) {
        oForm[oFormMode].eventContainer = oConfig.eventContainer;
      }
      this.oFormReader.initForm(oForm, oFormMode);
    });
  }

  /**
   * affiche la modale contenant le formreader
   */
  public showFormReaderModal() {
    $('#workspacelist-form-modal').modal('show');
  }

  /**
   * affiche la modale contenant le formreader
   */
  public hideFormReaderModal() {
    $('#workspacelist-form-modal').modal('hide');
  }

  /**
 * Récupère la définition JSON d'un formulaire
 * @param  {string} url
 * @returns Promise
 */
  private getJSONForm_(url: string): Promise<any> {
    return new Promise((resolve, reject) => {
      // Récupère le formulaire
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url
      }).then((resp) => {
        if (this.vitisCheck.isDefAndNotNull(resp.data)) {
          resolve(resp.data);
        } else {
          console.error('Error getting JSON form');
          reject(false);
        }
      });
    });
  }

  /**
   * Recherche si l'utilisateur a le droit d'utiliser le bouton
   * @param btnConfig
   * @returns
   */
  private hasAccessToButton(btnConfig: any): boolean {
    if (this.vitisCheck.isBoolean(btnConfig)) {
      return btnConfig;
    }

    if (this.vitisCheck.isString(btnConfig)) {
      btnConfig = [btnConfig];
    }

    if (this.vitisCheck.isArray(btnConfig)) {
      const aUserPrivileges = this.auth.getPrivileges();
      for (let i = 0; i < btnConfig.length; i++) {
        const sPrivilege = btnConfig[i];
        if (this.vitisCheck.isString(sPrivilege) &&
          aUserPrivileges.indexOf(sPrivilege) !== -1) {
          return true;
        }
      }
    }

    return false;
  }
}
