/**
 *                        Jobs Runner
 * 
 * Ce fichier permet de lancer les jobs déclarés dans jobs.json
 * Les jobs correspondent à des commances système, en fonction
 * du système en cours, le script lancera command_win ou
 * comman_lin
 * 
 * Les périodes sont définies de la même manière que sur cron
 * mais une option par seconde est disponible :
 * ┌────────────── second (optional)
 * │ ┌──────────── minute
 * │ │ ┌────────── hour
 * │ │ │ ┌──────── day of month
 * │ │ │ │ ┌────── month
 * │ │ │ │ │ ┌──── day of week
 * │ │ │ │ │ │
 * │ │ │ │ │ │
 * * * * * * *
 * 
 */

import { propertiesSrvc, Logger, Db, LocalFS } from "vitis-engine-lib";
import { RunnerSrvc } from "./runnerSrvc";
import { FileConfigSrvc } from "./fileconfigSrvc";
import { ScheduledTask } from "node-cron";

let oProperties: any = propertiesSrvc.getAllPropertiesSync();
const reloadInterval = oProperties.time_interval_jobs_refresh ?? 10; // Intervalle de rechargement en secondes
let aRunnningJobs: ScheduledTask[] = []

/**
 * charge les fihcier de conf pour jobs et task et exécute
 * @param bForce permet de forcer le chargement des jobs
 */
const run = (bForce : boolean) => {
  // Properties
  propertiesSrvc.getProperties().then((oPropertiesRefresh: any) => {
    oProperties = oPropertiesRefresh;
    const oLogger = new Logger(oProperties);
    // utilisation d'un ficheir spécifique pour logger via les runners
    oLogger.setLogFilePath(oProperties.vitis_runners_log_path)
    // désactive la console pour ne pas remplir le syslog
    oLogger.disableConsoleMessage();
    const oLocalFS = new LocalFS(oProperties, oLogger);
    const oDb = new Db(oProperties, {
      user: oProperties.db_superuser,
      host: oProperties.db_server,
      database: oProperties.db_name,
      password: oProperties.db_password,
      port: oProperties.db_port,
      autoSslConfig: true
    }, oLogger);

    const oFileConfigSrvc = new FileConfigSrvc(oLogger);
    const oRunnerSrvc = new RunnerSrvc(oProperties, oDb, oLogger, oLocalFS, oFileConfigSrvc);

    // Fichier de config des jobs
    const oJobsConfig = oFileConfigSrvc.getFileConfig('jobs');

    // Plannifie les jobs
    if (oJobsConfig && oJobsConfig.jobs && oJobsConfig.jobs.length) {
      // détermine si le fichier de conf a été édité
      if (bForce || (oJobsConfig.modified === true)) {
        oLogger.log(`Load Jobs at ${Date.now()} `, null, 'INFO');
        aRunnningJobs = oRunnerSrvc.startJobs(oJobsConfig.jobs, aRunnningJobs);
        // on le passe à false pour éviter de relancer à chaque fois
        oJobsConfig.modified = false;
        oFileConfigSrvc.updateFileConfig('jobs', oJobsConfig);
      }
    }

    const oTasksConfig = oFileConfigSrvc.getFileConfig('tasks');

    // Lance les tâches
    if (oTasksConfig && oTasksConfig.tasks && oTasksConfig.tasks.length) {
      oLogger.log(`Load Tasks at ${Date.now()}`, null, 'INFO');
      oRunnerSrvc.startTasks(oTasksConfig.tasks);
    }
  });
}

// premier lancement avec chargement forcé des jobs instant
run(true);
/**
 * Interval sur reloadInterval pour recharger le fichier de conf et lancer la planification
 */
setInterval(() => {
  run(false);
}, (reloadInterval * 1000) + 50);
