import fs from 'fs';
import { Logger } from "vitis-engine-lib";
import { Db } from "vitis-engine-lib";
import { LocalFS } from "vitis-engine-lib";
import { WorkbookOption, WorksheetOption, Style, Workbook, Worksheet, getExcelRowCol } from 'excel4node'

interface printExcelArgs {
  JsonTemplate: string;
  OutputFileName?: string;
  JsonScope?: string;
}

interface JsonTemplate {
  metadata?: WorkbookOption; // voir config d'un workbook dans excel4node
  sheets: { [sheetName: string]: JsonSheet; };
  styles?: { [styleName: string]: Style; }; // voir config d'un style pour colonne dans excel4node
}

interface JsonSheet {
  statics?: JsonCell[];
  options?: WorksheetOption; // voir config sheet dans excel4node
  foreach?: JsonForeach;
  freeze?: {
    row?: number; // gèle les x premières lignes
    column?: number; // gèle les x premières colonnes
  };
  width?: { [colLetter: string]: number; }; // permet de définir la taille de  cahque colonne
}

interface JsonForeach {
  row_from?: number | string;
  offset?: number;
  limit?: number;
  array: string | any[]; // scope à utiliser ou clé vers le scope à utiliser
  data?: JsonCell[];
  foreach?: JsonForeach;
}

interface JsonCell {
  type: string; //['string', 'number', 'formula', 'date', 'link', 'bool']
  value: any; // dépend du type
  cellFrom?: string;
  cellTo?: string;
  cell?: string;
  column?: string;
  columnFrom?: string;
  columnTo?: string;
  merge?: boolean;
  row?: string | number;
  style?: string | Style;
  eval?: boolean;
  condition?: string | boolean; // expression eval qui retourne true/false
  empty_val?: any
}

/**
 * Génére un excel
 */
export async function printExcel_(oProperties: any, oDb: Db, oLogger: Logger, oLocalFS: LocalFS, args: any) {

  let sOutputPath = args.OutputFileName;
  let sJsonTemplatePath = args.JsonTemplate;
  let sJsonScope = args.JsonScope;
  let oTemplate: JsonTemplate = { sheets: {} };
  let oScope: any = {};

  // lecture des fichiers json
  // Si sJsonTemplatePath est un fichier
  if (sJsonTemplatePath && fs.existsSync(sJsonTemplatePath)) {
    sJsonTemplatePath = fs.readFileSync(sJsonTemplatePath).toString();
  } else {
    return false;
  }
  // Si sJsonScope est un fichier
  if (sJsonScope && fs.existsSync(sJsonScope)) {
    sJsonScope = fs.readFileSync(sJsonScope).toString();
  } else {
    return false;
  }
  // parse des jsons
  try {
    if (sJsonTemplatePath) {
      oTemplate = JSON.parse(sJsonTemplatePath);
    }
  } catch (e) { }
  try {
    if (sJsonScope) {
      oScope = JSON.parse(sJsonScope);
    }
  } catch (e) { }

  const oRenderer: ExcelRenderer = new ExcelRenderer(oTemplate, oScope, sOutputPath);
  oRenderer.render();
  oRenderer.writeFile();
}

class ExcelRenderer {

  /**
   * liste des types connu pour les cellules
   */
  private AllowedType: string[] = ['string', 'number', 'formula', 'date', 'link', 'bool'];
  /**
   * Objet workbook excel
   */
  private wb: Workbook;
  /**
   * Librairie des différents style de cellules
   */
  private oStyles: { [key: string]: Style } = {};
  /**
   * Objets Worksheet pour chaque feuille
   */
  private oSheets: { [key: string]: Worksheet } = {};

  /**
   * 
   * @param {JsonTemplate} oTemplate Structure du template Json
   * @param {any} oScope Donnée pour remplir le excel
   * @param {string} sFilePath chemin final du fichier excel
   */
  constructor(private oTemplate: JsonTemplate, private oScope: any, private sFilePath: string) {
    // récupération des options du document
    const wbOptions: WorkbookOption = (this.oTemplate.metadata) ? this.oTemplate.metadata : {};
    // récupération de la liste des styles
    this.oStyles = (this.oTemplate.styles) ? this.oTemplate.styles : {};

    // création d'un workbook excel
    this.wb = new Workbook(wbOptions);
  }

  /**
   * Met en forme l'ensemble des feuilles définies dans le template
   * @return {void}
   */
  public render(): void {

    for (let sSheetName in this.oTemplate.sheets) {
      const oSheet: JsonSheet = this.oTemplate.sheets[sSheetName];
      const sheetOptions: WorksheetOption = oSheet.options ? oSheet.options : {};

      this.oSheets[sSheetName] = this.wb.addWorksheet(sSheetName, sheetOptions);

      this.renderExcelSheet(sSheetName);
    }

  }

  /**
   * Ecrit le workbook dans le fichier excel final
   * @return {void}
   */
  public writeFile(): void {
    // écriture du fichier excel final
    this.wb.write(this.sFilePath);
  }

  /**
   * Imprime les cartes et les ajoute en tant qu'images
   * @param {string} sSheetName nom de la feuille excel dans le template
   * @returns {void}
   */
  protected renderExcelSheet(sSheetName: string): void {
    const ws: Worksheet = this.oSheets[sSheetName];
    const sheetTpl: JsonSheet = this.oTemplate.sheets[sSheetName];
    let nb_rows: number = 0;
    // traitement des foreach en récursif
    if (sheetTpl.foreach) {
      nb_rows = this.processForeach(sSheetName, sheetTpl.foreach, this.oScope, nb_rows);
    }
    // traitement de statics
    if (sheetTpl.statics) {
      this.processStaticsCells(sSheetName, nb_rows);
    }

    if (sheetTpl.freeze) {
      if (sheetTpl.freeze.column) {
        ws.column(sheetTpl.freeze.column).freeze()
      }
      if (sheetTpl.freeze.row) {
        ws.row(sheetTpl.freeze.row).freeze()
      }
    }

    if (sheetTpl.width) {
      for (let colLetter in sheetTpl.width) {
        const width = sheetTpl.width[colLetter];
        const col = this.findCellWithCode(colLetter + '1').col;

        ws.column(col).setWidth(width);
      }
    }
  }

  /**
   * Gére les ligné à répéter, fonction récursive
   * @param {string} sSheetName nom de la feuille en cours de génération
   * @param {JsonForeach} oForeach template du foreach en cours
   * @param {any} oScope donnée à utiliser pour cette itération
   * @param {number} nb_rows position de la ligne courante
   * @returns {number} position de la ligne courante aprés le for_each
   */
  protected processForeach(sSheetName: string, oForeach: JsonForeach, oScope: any, nb_rows: number): number {
    // si la clé n'est pas défini alors on ne pourra pas boucler
    if (!oForeach.array) {
      return nb_rows;
    }
    // On récupére l'attribut pour boucler ou on le cherche dans le scope
    let aData: Array<any> = this.findForEachData(oForeach.array, oScope);
    if (oForeach.offset) {
      // si un offset est défini alors on retire les premier élément du tableau
      aData.splice(0, oForeach.offset);
    }
    if (oForeach.limit) {
      // si une limit est défini alors on ne garde que les items dans la limit
      aData = aData.slice(0, oForeach.limit);
    }

    // si un départ est donnée alors on l'utilise comme minimum
    if (oForeach.row_from) {
      nb_rows = this.findForEachStartRow(oForeach.row_from, nb_rows);
    }

    const oCellsDef: Array<JsonCell> = (oForeach.data) ? oForeach.data : [];

    for (let index in aData) {
      const oRowScope = aData[index];
      for (let oCellDef of oCellsDef) {
        // convert column in cell with nb_rows remplissage de la cellule
        if (oCellDef.column) {
          oCellDef.cell = oCellDef.column + String(nb_rows);
        }
        if (oCellDef.columnFrom) {
          oCellDef.cellFrom = oCellDef.columnFrom + String(nb_rows);
        }
        if (oCellDef.columnTo) {
          oCellDef.columnTo = oCellDef.columnTo + String(nb_rows);
        }

        if (typeof oCellDef.value === 'string') {
          oCellDef.value = oCellDef.value.replace(/@row/g, '@scope')
        }
        if (typeof oCellDef.condition === 'string') {
          oCellDef.condition = oCellDef.condition.replace(/@row/g, '@scope')
        }

        this.processCell(sSheetName, oCellDef, oRowScope, nb_rows);
      }
      nb_rows++;
      // si foreach recursif
      if (oForeach.foreach) {
        nb_rows = this.processForeach(sSheetName, oForeach.foreach, oRowScope, nb_rows);
      }
    }

    return nb_rows;
  }

  /**
   * Permet d'évaluer l'élement startRow d'un foreach
   * L'expression peut être un simple nombre auquel cas on considérera ce nombre comme le nouveau point de départ du fichier
   * Une expression évaluable utilisant la position courante (pour sauter une ligne par exemple : @actual + 1)
   * @param {any} mExpression expression à évaluer
   * @param {number} nb_rows position actuel dans le fichier excel utiliser pour la balise \@actual
   * @returns {number} nouvelle position courante dans le fichier excel
   */
  private findForEachStartRow(mExpression: any, nb_rows: number): number {
    // si un départ est donnée alors on l'utilise comme minimum
    if (typeof mExpression === 'number') {
      return mExpression;
    }

    const sExpression: string = mExpression.replace(/@actual/g, String(nb_rows));
    const eval_rows = eval(sExpression);

    return (typeof eval_rows === 'number') ? eval_rows : nb_rows;
  }

  /**
   * Permet de récupérer les donnée lié à la clé array dans la définition d'un foreach
   * @row et @scope sont équivalents
   * @param {any} mDataScope clé des données à trouver ou données directement dans le template (exemple exécutuon de formule sur une ligne)
   * @param {any} oScope Permet de cherché la clé de mDataScope si c'est un string
   * @returns {Array<any>} un tableau d'objet qui ssera ensuite exploité par une boucle (tableau de row)
   */
  private findForEachData(mDataScope: any, oScope: any): Array<any> {
    // On récupére l'attribut pour boucler ou on le cherche dans le scope
    let aData: any = mDataScope;
    if (typeof mDataScope === 'string') {
      if (mDataScope.indexOf('@scope') > -1 || mDataScope.indexOf('@row') > -1) {
        const sKey: string = mDataScope.replace(/@row/g, '@scope');
        aData = this.findValInObject(sKey, oScope);
      }
    }

    if (Array.isArray(aData)) {
      return aData;
    }

    // si c'est un objet alors on le transforme en array d'un seul objet
    if (typeof aData === 'object') {
      return [aData];
    }
    return [];
  }

  /**
   * Recherche une clé dans un objet (les clés composées fonctionnent : key.subkey)
   * @param {string} key clé à cherché
   * @param {any} oScope objet à parcourir
   * @returns {any} valeur trouvée ou null en cas d'erreur
   */
  private findValInObject(key: string, oScope: any): any {
    let sKey: string = key;
    let aSubKeys: Array<string> = [];
    // si la clé est composée alors on la découpe pour la récursivité
    if (key.indexOf('.') > -1) {
      const aTmpKey = key.split('.');
      // si le premier élément est le bind vers le scope on le retire
      if (aTmpKey[0] === '@scope') {
        aTmpKey.splice(0, 1);
      }
      sKey = aTmpKey.splice(0, 1)[0];
      aSubKeys = aTmpKey;
    }

    // si la clé n'existe pas dans le scope on stop
    if (typeof oScope[sKey] == 'undefined') {
      return null;
    }

    // si pas de sous-clé on est arrivé au bout on retourne
    if (aSubKeys.length === 0) {
      return oScope[sKey];
    }

    // si le type de donnée récupéré est incompatible avec le parcours d'objet
    if (typeof oScope[sKey] !== 'object' && !Array.isArray(oScope[sKey])) {
      return oScope[sKey];
    }
    // sinon on relance en scope restreint avec les sous-clés
    return this.findValInObject(aSubKeys.join('.'), oScope[sKey]);
  }

  /**
 * Gére les cellule statiques du fichier cette fonction est appelée aprés traitement des foreach pour un nb_rows cohérent
 * @param {string} sSheetName Nom de la feuille en cours
 * @param {number} nb_rows Position actuel dans le fichier
 * @returns {void}
 */
  protected processStaticsCells(sSheetName: string, nb_rows: number): void {

    const aStatics: JsonCell[] | undefined = this.oTemplate.sheets[sSheetName].statics;
    if (!Array.isArray(aStatics)) {
      return;
    }

    for (let oCellDef of aStatics) {
      this.processCell(sSheetName, oCellDef, this.oScope, nb_rows);
    }
  }

  /**
   * Permet de remplir une ou plusieurs cellule
   * @param {string} sSheetName nom de la feuille courante
   * @param {JsonCell} oCellDef Définition de la cellule à remplir
   * @param {any} oScope donnée à utiliser
   * @param {number} nb_rows position actuel dans le fichier
   */
  protected processCell(sSheetName: string, oCellDef: JsonCell, oScope: any, nb_rows: number): void {
    const bCondition = this.parseCondition(oScope, oCellDef, nb_rows);
    // si condition vaut false on saute la cellule
    if (!bCondition) {
      return;
    }

    const ws: Worksheet = this.oSheets[sSheetName];
    const sFunctionName = (this.AllowedType.indexOf(oCellDef.type) > -1) ? oCellDef.type : 'string';
    const oStyle = this.findStyle(oCellDef);
    const mValue = this.parseValue(oScope, sFunctionName, oCellDef, nb_rows);
    oCellDef = this.findCellsPosition(oCellDef, nb_rows);

    let cell: any; // Cell

    if (oCellDef['cell']) {
      // si une seul cellule est impacté
      const oCell: any = this.findCellWithCode(oCellDef.cell);
      cell = ws.cell(oCell.row, oCell.col);
    } else if (oCellDef.cellFrom && oCellDef.cellTo) {
      // si une plage de cellule est impacté avec fusion
      const oCellFrom: any = this.findCellWithCode(oCellDef.cellFrom);
      const oCellTo: any = this.findCellWithCode(oCellDef.cellTo);
      const bMerge: boolean = (oCellDef.merge === false) ? false : true

      cell = ws.cell(oCellFrom.row, oCellFrom.col, oCellTo.row, oCellTo.col, bMerge);
    }

    if (cell) {
      cell[sFunctionName](mValue).style(oStyle)
    }
  }

  /**
   * Récupère le style associé à une cellule
   * @param {JsonCell} oCellDef défintion de la cellule
   * @returns {any} objet à utilisé comme style de la cellule
   */
  private findStyle(oCellDef: JsonCell): any {
    if (oCellDef.style) {
      // si le style est défini en dur sur le champ on le retourne direct
      if (typeof oCellDef.style !== 'string') {
        return oCellDef.style;
      }
      // si le style correspond à une clé existante dans le template on le retourne
      if (this.oStyles[oCellDef.style]) {
        return this.oStyles[oCellDef.style];
      }
    }

    return {};
  }

  /**
   * Permet de trouver la ligne et la colonne d'une cellule à partir de son code excel
   * @param {string} sCellCode Code excel (ex: A1)
   * @returns {any} Position col/row
   */
  private findCellWithCode(sCellCode: string): any {
    return getExcelRowCol(sCellCode);
  }

  /**
   * Permet d'analyser la défintion d'une cellule pour en générer la position de la cellule si necéssaire
   * @param {JsonCell} oCellDef Défintion de la cellule
   * @param {number} nb_rows position actuelle dans le fichier
   * @returns {any} Défintion de la cellule mis en forme
   */
  private findCellsPosition(oCellDef: JsonCell, nb_rows: number): any {

    let row: number = nb_rows;
    if (oCellDef.row) {
      if (typeof oCellDef.row === 'number') {
        row = oCellDef.row;
      }
      if (typeof oCellDef.row === 'string') {
        row = eval(oCellDef.row.replace(/@last_row|@actual/g, String(nb_rows)));
      }
    }

    if (oCellDef.column) {
      oCellDef.cell = oCellDef.column + row;
    }
    if (oCellDef.columnFrom) {
      oCellDef.cellFrom = oCellDef.columnFrom + row;
    }
    if (oCellDef.columnTo) {
      oCellDef.cellTo = oCellDef.columnTo + row;
    }

    return oCellDef;
  }

  /**
   * Permet dévaluer l'expression qui est dans la défintion de la cellule 
   * @param {any} oScope Donnée à utiliser pour évaluer les données
   * @param {string} sFunctionName type de la cellule (peut jouer sur le formatage)
   * @param {JsonCell} oCellDef Définition de la cellule
   * @param {number} nb_rows Position actuel dans le fichier
   * @returns {any} valeur de la cellule
   */
  private parseValue(oScope: any, sFunctionName: string, oCellDef: JsonCell, nb_rows: number): any {
    let mValue : any = oCellDef.value;
    return this.parseItem(oScope, sFunctionName, oCellDef, nb_rows, mValue);
  }


  /**
 * Permet dévaluer l'expression qui est dans la défintion de la cellule 
 * @param {any} oScope Donnée à utiliser pour évaluer les données
 * @param {string} sFunctionName type de la cellule (peut jouer sur le formatage)
 * @param {JsonCell} oCellDef Définition de la cellule
 * @param {number} nb_rows Position actuel dans le fichier
 * @returns {any} valeur de la cellule
 */
  private parseCondition(oScope: any, oCellDef: JsonCell, nb_rows: number): boolean {
    // si condition n'est pas défini on laisse la cellule
    if (typeof oCellDef.condition == 'undefined') {
      return true;
    }
    
    let mCondition = oCellDef.condition;
    // si c'est un booleén on retourne directement
    if (typeof mCondition == 'boolean') {
      return mCondition
    }
    // on force le eval pour le if
    const oConditionDef: JsonCell = {
      ...oCellDef,
      'eval' : true,
      'empty_val' : null
    }

    mCondition = this.parseItem(oScope, 'string', oConditionDef, nb_rows, mCondition);
    if (mCondition) {
      return true;
    }
    return false;
  }


  private parseItem(oScope: any, sFunctionName: string, oCellDef: JsonCell, nb_rows: number, mValue: any): any {
    // détection d'une valeur en lien avec le scope et remplacement
    if (typeof mValue === 'string') {

      if (typeof oCellDef.empty_val == 'undefined') {
        oCellDef.empty_val = '';
      }

      const ScopeRegex = /(?<key>@scope\.(([a-zA-Z0-9_]+)\.?)+)/g;
      let aMatches = mValue.match(ScopeRegex);
      if (aMatches) {
        // on retire les clé doublons
        const aFinalMatches = aMatches.filter(function (elem, index, self) {
          return index === self.indexOf(elem);
        })
        // remplacement de chaque balise par sa valeur
        for (let sBind of aFinalMatches) {
          const mBindVal: any = this.findValInObject(sBind, oScope);
          // on remplace . par \. pour la regexp
          let sBindRegexp: string = sBind.replace(/\./g, '\\.');
          // on ajoute le \b pour ne pas remplacer des clés proches comme sum_contenance et sum_contenance_bati
          const bindRegex = new RegExp(sBindRegexp + '\\b', 'g');
          const sFinalValue: string = (mBindVal === null) ? oCellDef.empty_val : String(mBindVal);
          mValue = mValue.replace(bindRegex, sFinalValue);
        }
      }

      mValue = mValue.replace(/@last_row|@actual/g, String(nb_rows));
      if (oCellDef.eval === true) {
        mValue = eval(mValue);
      }
    }

    // gestion en fonction du type
    switch (sFunctionName) {
      case 'number':
        mValue = parseFloat(mValue);
        break;
      case 'date':
        if (mValue == '@now') {
          mValue = new Date();
        }
        break;
    }

    return mValue;
  }
}




