import puppeteer, { PaperFormat } from 'puppeteer';
import fs from 'fs';
import { Logger } from "vitis-engine-lib";
import { Db } from "vitis-engine-lib";
import { LocalFS } from "vitis-engine-lib";
import { openPage } from './utils';
import { printMaps } from './printMap';
import PDFMerger from 'pdf-merger-js';

interface printReportArgs {
    HTMLInputFileName: string;
    TmpFolder: string;
    OutputFileName: string;
    Format?: string;
    OutputFormat?: string;
    Orientation?: string;
    JsonIncludes?: string;
    JsonScope?: string;
    JsonMaps?: string;
    Properties?: string;
    MultiplesElements: boolean;
    UseStyles: boolean;
}

/**
 * Imprime un rapport
 * @param oProperties 
 * @param oDb 
 * @param oLogger 
 * @param oLocalFS 
 */
export async function printReport_(oProperties: any, oDb: Db, oLogger: Logger, oLocalFS: LocalFS, args: any) {

    const sPrintClientURL = `file:///${oProperties.vas_dir}/src/Module/Vitis/utils/printserver/client/report/index.html`;
    const sTmpFolder = args.TmpFolder;
    const sOutputFileName = args.OutputFileName;
    let sHTMLBody = args.HTMLInputFileName;
    let sFormat = args.Format ? args.Format.toLowerCase() : 'a4';
    let sOrientation = args.Orientation ? args.Orientation.toLowerCase() : 'portrait';
    let sOutputFormat = args.OutputFormat ? args.OutputFormat.toLowerCase() : 'pdf';
    let sJsonIncludes = args.JsonIncludes;
    let sJsonScopes = args.JsonScopes;
    let sJsonMaps = args.JsonMaps;
    let bPaged = args.Paged === 'true' ? true : (args.Paged === 'false' ? false : undefined);

    // Si sJsonIncludes est un fichier
    if (sJsonIncludes && fs.existsSync(sJsonIncludes)) {
        sJsonIncludes = fs.readFileSync(sJsonIncludes).toString();
    }
    // Si sJsonScopes est un fichier
    if (sJsonScopes && fs.existsSync(sJsonScopes)) {
        sJsonScopes = fs.readFileSync(sJsonScopes).toString();
    }
    // Si sJsonMaps est un fichier
    if (sJsonMaps && fs.existsSync(sJsonMaps)) {
        sJsonMaps = fs.readFileSync(sJsonMaps).toString();
    }
    //chargement du body HTML
    if (sHTMLBody && fs.existsSync(sHTMLBody)) {
        sHTMLBody = fs.readFileSync(sHTMLBody).toString();
    }

    // Récupération des arrays principaux
    let aIncludes = [];
    let aScopes = [];
    let aMaps = [];

    // Includes
    try {
        if (sJsonIncludes) {
            aIncludes = JSON.parse(sJsonIncludes);
        }
    } catch (e) { }
    // Scopes
    try {
        if (sJsonScopes) {
            aScopes = JSON.parse(sJsonScopes);
        }
    } catch (e) { }
    // Maps
    try {
        if (sJsonMaps) {
            aMaps = JSON.parse(sJsonMaps);
        }
    } catch (e) { }

    // Construction et récupération des objets rapports
    const aReports = getReportObjects(aIncludes, aScopes, aMaps);
    const aPaths = [];

    for (let i = 0; i < aReports.length; i++) {
        let aCurrentIncludes = aReports[i].aIncludes;
        let oCurrentScope = aReports[i].oScope;
        let aCurrentMaps = aReports[i].aMaps;

        if (aMaps instanceof Array && aMaps.length > 0) {
            aCurrentIncludes = await addMapsToIncludes(aCurrentIncludes, aCurrentMaps, oCurrentScope, {
                oProperties,
                oDb,
                oLogger,
                oLocalFS,
                sTmpFolder,
            });
        }

        // Ouvre la page
        const openedPage = await openPage({
            sPrintClientURL,
            sHTMLBody,
            sOrientation,
            sFormat,
            aIncludes: aCurrentIncludes,
            oScope: oCurrentScope,
            oProperties,
            bPaged,
        }, oLogger);

        const page = openedPage.page;
        const browser = openedPage.browser;

        if (sOutputFormat === 'pdf' || sOutputFormat === 'vectorpdf') {
            const sTmpFileName = sTmpFolder + "/render_" + i + ".pdf";
            await page.pdf({ path: sTmpFileName, format: sFormat, landscape: sOrientation === 'paysage' });
            aPaths.push(sTmpFileName);
        } else if (sOutputFormat === 'imagepdf') {
            await page.screenshot({ path: sTmpFolder + '/rendered.jpeg', fullPage: true });
            await screenshotToPDF(sTmpFolder + '/rendered.jpeg', sOutputFileName, sFormat, sOrientation);
        } else if (sOutputFormat === 'image') {
            await page.screenshot({ path: sOutputFileName, fullPage: true });
        }

        await browser.close();
    }

    if (sOutputFormat === 'pdf' || sOutputFormat === 'vectorpdf') {
        // Si on a plusieurs PDF en sortie
        if (aPaths.length > 1) {
            // On fusionne les PDF
            const oPdfMerger = new PDFMerger();

            for (const sPath of aPaths) {
                oPdfMerger.add(sPath);
            }

            // Sauvegarde
            oPdfMerger.save(sOutputFileName);

        } else if (aPaths.length === 1) {
            // On renomme l'unique PDF en celui que l'on souhaite avoir en sortie
            fs.renameSync(aPaths[0], sOutputFileName);
        }
    }
}

/**
 * Retourne le code Base64 d'un fichier image
 * @param file 
 * @returns 
 */
function base64Encode(file: string) {
    const contents = fs.readFileSync(file, { encoding: 'base64' });
    return contents;
}

/**
 * Transforme une image en PDF avec pagination
 * @param inputFile 
 * @param outputFile 
 * @param sFormat
 * @param sOrientation
 */
async function screenshotToPDF(inputFile: string, outputFile: string, sFormat: PaperFormat, sOrientation: string) {
    const browser = await puppeteer.launch();
    const page = await browser.newPage();
    await page.setViewport({ width: 800, height: 1126, deviceScaleFactor: 1 });
    const image = 'data:image/png;base64,' + base64Encode(inputFile);
    const sHTML = `
    <html>
        <body style="margin: 0px;">
            <img src="${image}">
        </body>
    </html>
    `;

    await page.setContent(sHTML), { waitUntil: 'networkidle0' };
    await page.pdf({
        path: outputFile,
        format: sFormat,
        landscape: sOrientation === 'paysage',
        scale: 1,
    });
    await browser.close();
}

/**
 * Imprime les cartes et les ajoute en tant qu'images
 * @param aIncludes 
 * @param aMaps 
 * @returns 
 */
async function addMapsToIncludes(aIncludes: Array<any>, aMaps: Array<any>, oScope: any, oUtils: any): Promise<any[]> {

    const aPrintedMaps = await printMaps(oUtils.oProperties, oUtils.oDb, oUtils.oLogger, oUtils.oLocalFS, aMaps, oUtils.sTmpFolder);
    aIncludes = [...aIncludes, ...aPrintedMaps];

    for (let i = 0; i < aPrintedMaps.length; i++) {
        const oPrintedMap = aPrintedMaps[i];
        if (oPrintedMap.resultScale) {
            if (oScope.resultScale) {
                if (oScope.resultScale.push) {
                    oScope.resultScale.push(oPrintedMap.resultScale);
                }
            } else {
                oScope.resultScale = [oPrintedMap.resultScale];
            }
        }
    }

    return aIncludes;
}

/**
 * Construit et retourne les objets rapports par rapport aux arrays principaux
 * @param aIncludes 
 * @param aScopes 
 * @param aMaps 
 * @returns 
 */
function getReportObjects(aIncludes: Array<any>, aScopes: Array<any>, aMaps: Array<any>): Array<any> {
    let aReports: Array<any> = [];

    // fusion des trois array en un objet par rapport
    for (let i = 0; i < aScopes.length; i++) {
        const oReport : any = {
            oScope: aScopes[i],
        }

        if (aIncludes[i]) {
            oReport.aIncludes = aIncludes[i]
        }
        if (aMaps[i]) {
            oReport.aMaps = aMaps[i]
        }

        aReports.push(oReport);
    }

    return aReports;
}