import puppeteer, { PaperFormat } from 'puppeteer';
import fs, { PathLike } from 'fs';
import { Logger } from "vitis-engine-lib";
import { Db } from "vitis-engine-lib";
import { LocalFS } from "vitis-engine-lib";
import { openPage } from './utils';
import { printMaps } from './printMap';

interface printTemplateArgs {
    HTMLInputFileName: string;
    TmpFolder: string;
    OutputFileName: string;
    Format?: string;
    OutputFormat?: string;
    Orientation?: string;
    JsonIncludes?: string;
    JsonScope?: string;
    JsonMaps?: string;
    Properties?: string;
}

/**
 * Imprime un template
 * @param oProperties 
 * @param oDb 
 * @param oLogger 
 * @param oLocalFS 
 */
export async function printTemplate_(oProperties: any, oDb: Db, oLogger: Logger, oLocalFS: LocalFS, args: any) {

    const sPrintClientURL = `file:///${oProperties.vas_dir}/src/Module/Vitis/utils/printserver/client/renderer/index.html`;
    const sTmpFolder = args.TmpFolder;
    const sOutputFileName = args.OutputFileName;
    let sHTMLBody = args.HTMLInputFileName;
    let sFormat = args.Format ? args.Format.toLowerCase() : 'a4';
    let sOrientation = args.Orientation ? args.Orientation.toLowerCase() : 'portrait';
    let sOutputFormat = args.OutputFormat ? args.OutputFormat.toLowerCase() : 'pdf';
    let sJsonIncludes = args.JsonIncludes;
    let sJsonScope = args.JsonScope;
    let sJsonMaps = args.JsonMaps;
    let bPaged = args.Paged === 'true' ? true : (args.Paged === 'false' ? false : undefined);
    let aIncludes = [];
    let oScope = {};
    let aMaps: any[] = [];

    // Si sJsonIncludes est un fichier
    if (sJsonIncludes && fs.existsSync(sJsonIncludes)) {
        sJsonIncludes = fs.readFileSync(sJsonIncludes).toString();
    }
    // Si sJsonScope est un fichier
    if (sJsonScope && fs.existsSync(sJsonScope)) {
        sJsonScope = fs.readFileSync(sJsonScope).toString();
    }
    // Si sJsonMaps est un fichier
    if (sJsonMaps && fs.existsSync(sJsonMaps)) {
        sJsonMaps = fs.readFileSync(sJsonMaps).toString();
    }
    //chargement du body HTML
    if (sHTMLBody && fs.existsSync(sHTMLBody)) {
        sHTMLBody = fs.readFileSync(sHTMLBody).toString();
    }

    try {
        if (sJsonIncludes) {
            const aTmpIncludes = JSON.parse(sJsonIncludes);
            if (aTmpIncludes instanceof Array && aTmpIncludes.length > 0) {
                aIncludes = aTmpIncludes;
            }
        }
    } catch (e) { }
    try {
        if (sJsonScope) {
            oScope = JSON.parse(sJsonScope);
        }
    } catch (e) { }
    try {
        if (sJsonMaps) {
            const aTmpMaps = JSON.parse(sJsonMaps);
            if (aTmpMaps instanceof Array && aTmpMaps.length > 0) {
                aMaps = aTmpMaps;
            }
        }
    } catch (e) { }

    if (aMaps instanceof Array && aMaps.length > 0) {
        aIncludes = await addMapsToIncludes(aIncludes, aMaps, oScope, {
            oProperties,
            oDb,
            oLogger,
            oLocalFS,
            sTmpFolder,
        });
    }

    // Ouvre la page
    const openedPage = await openPage({
        sPrintClientURL,
        sHTMLBody,
        sOrientation,
        sFormat,
        aIncludes,
        oScope,
        oProperties,
        bPaged,
    }, oLogger);
    const page = openedPage.page;
    const browser = openedPage.browser;

    // Correctif orientation de la page
    page.addStyleTag({
        'content': '@page {size: auto}'
    });

    if (sOutputFormat === 'pdf' || sOutputFormat === 'vectorpdf') {
        await page.pdf({ path: sOutputFileName, format: sFormat, landscape: sOrientation === 'paysage' });
    } else if (sOutputFormat === 'imagepdf') {
        await page.screenshot({ path: sTmpFolder + '/rendered.jpeg', fullPage: true });
        await screenshotToPDF(sTmpFolder + '/rendered.jpeg', sOutputFileName, sFormat, sOrientation)
    } else if (sOutputFormat === 'image') {
        await page.screenshot({ path: sOutputFileName, fullPage: true });
    }

    await browser.close();
}

/**
 * Retourne le code Base64 d'un fichier image
 * @param file 
 * @returns 
 */
function base64Encode(file: string) {
    const contents = fs.readFileSync(file, { encoding: 'base64' });
    return contents;
}

/**
 * Transforme une image en PDF avec pagination
 * @param inputFile 
 * @param outputFile 
 * @param sFormat
 * @param sOrientation
 */
async function screenshotToPDF(inputFile: string, outputFile: string, sFormat: PaperFormat, sOrientation: string) {
    const browser = await puppeteer.launch();
    const page = await browser.newPage();
    await page.setViewport({ width: 800, height: 1126, deviceScaleFactor: 1 });
    const image = 'data:image/png;base64,' + base64Encode(inputFile);
    const sHTML = `
    <html>
        <body style="margin: 0px;">
            <img src="${image}">
        </body>
    </html>
    `;

    await page.setContent(sHTML), { waitUntil: 'networkidle0' };
    await page.pdf({
        path: outputFile,
        format: sFormat,
        landscape: sOrientation === 'paysage',
        scale: 1,
    });
    await browser.close();
}

/**
 * Imprime les cartes et les ajoute en tant qu'images
 * @param aIncludes 
 * @param aMaps 
 * @returns 
 */
async function addMapsToIncludes(aIncludes: Array<any>, aMaps: Array<any>, oScope: any, oUtils: any): Promise<any[]> {

    const aPrintedMaps = await printMaps(oUtils.oProperties, oUtils.oDb, oUtils.oLogger, oUtils.oLocalFS, aMaps, oUtils.sTmpFolder);
    aIncludes = [...aIncludes, ...aPrintedMaps];

    for (let i = 0; i < aPrintedMaps.length; i++) {
        const oPrintedMap = aPrintedMaps[i];
        if (oPrintedMap.resultScale) {
            if (oScope.resultScale) {
                if (oScope.resultScale.push) {
                    oScope.resultScale.push(oPrintedMap.resultScale);
                }
            } else {
                oScope.resultScale = [oPrintedMap.resultScale];
            }
        }
    }

    return aIncludes;
}