import { CronJob, TaskJob } from "./models";
import cron, { ScheduledTask } from 'node-cron';
import { exec } from 'child_process';
import { Db, LocalFS, Logger, propertiesSrvc } from "vitis-engine-lib";
import { FileConfigSrvc } from "./fileconfigSrvc";
import axios from 'axios';
import https from 'https';

/**
 * Service permettant de lancer les jobs et les tasks
 */
export class RunnerSrvc {

    private isWin: boolean;
    private userid: any;
    private oProperties: any;
    private oDb: Db;
    private oLogger: Logger;
    private oLocalFS: LocalFS;
    private oFileConfigSrvc: FileConfigSrvc;

    constructor(oProperties: any, oDb: Db, oLogger: Logger, oLocalFS: LocalFS, oFileConfigSrvc: FileConfigSrvc) {
        this.oProperties = oProperties;
        this.oDb = oDb;
        this.oLogger = oLogger;
        this.oLocalFS = oLocalFS;
        this.oFileConfigSrvc = oFileConfigSrvc;

        // OS en cours
        this.isWin = process.platform === "win32";

        // Utilisateur (uniquement pour linux)
        if (!this.isWin) {
            this.userid = require('userid');
        }
    }
    /**
     * Plannifie les jobs décrits dans config et les stoppe au bout de reloadInterval
     * @param {*} aJobs 
     */
    public startJobs(aJobs: Array<CronJob>, aOldJobs: ScheduledTask[]): ScheduledTask[] {
        const aNewJobs: ScheduledTask[] = []
        if (aJobs && aJobs.length) {
            // Stoppe la plannification des anciens jobs
            for (const oOldJob of aOldJobs) {
                if (oOldJob && oOldJob.stop) {
                    oOldJob.stop();
                }
            }

            // Lance la planification des jobs déclarés
            for (let i = 0; i < aJobs.length; i++) {
                const job = aJobs[i];
                if (job.period) {

                    const oCommand = this.parseJob_(job);
                    const oAxiosInstance = this.parseApiRequest_(job);

                    // Planifie le job
                    const task = cron.schedule(job.period, () => {
                        setTimeout(() => { // Timeout permet de palier au problème de deux jobs qui se lancent simultanément
                            if (oCommand) {
                                this.runCommand_(oCommand);
                            } else if (oAxiosInstance) {
                                this.runApiRequest_(oAxiosInstance, job);
                            }
                        }, i * 200);
                    }, {});

                    aNewJobs.push(task);
                }
            }
        }

        return aNewJobs;
    }

    /**
     * Lance les tâches
     * @param aTasks 
     */
    public startTasks(aTasks: Array<TaskJob>) {
        if (aTasks && aTasks.length) {

            // Lance la planification des tâches déclarés
            for (const task of aTasks) {
                if (task) {

                    const oCommand = this.parseJob_(task);
                    const oAxiosInstance = this.parseApiRequest_(task);

                    // Donne un id à la tâche (si il n'est pas déjà donné)
                    if (!task.id_) {
                        task.id_ = this.getUniqueId();
                    }

                    // Tag la tâche comme lancée
                    task.runned_ = true;

                    if (oCommand) {

                        // Commande à lancer
                        task.command_ = oCommand.command;

                        // Planifie le task
                        this.oLogger.log(`Run task with command : ${oCommand.command}`, task, 'INFO');
                        this.runCommand_(oCommand).then((resp) => {

                            const err = resp ? resp.err : null;
                            const stdout = resp ? resp.stdout : null;
                            const stderr = resp ? resp.stderr : null;

                            // Notifie la fin de la tâche
                            this.notifyEndTask_(task, err, stdout, stderr);
                        });
                    } else if (oAxiosInstance) {
                        this.runApiRequest_(oAxiosInstance, task).then(() => {
                            // Notifie la fin de la tâche
                            this.notifyEndTask_(task);
                        });
                    }

                    // Notifie le début de la tâche
                    this.notifyStartTask_(task);
                }
            }

            // Nettoie les tâches lancées
            for (let i = aTasks.length - 1; i >= 0; i--) {
                if (aTasks[i].runned_ === true) {
                    aTasks.splice(i, 1);
                }
            }

            // Met à jour le fichier tasks
            this.oFileConfigSrvc.updateFileConfig('tasks', { tasks: aTasks });
        }
    }

    /**
     * Retourne un identifiant unique
     */
    public getUniqueId(): string {
        return `${Date.now()}-${Math.floor(Math.random() * (9999 - 1000) + 1000)}`;
    }

    /**
     * Parse un job et retourne un objet command contenant la commande a executer ainsi que les options à donner
     * @param job 
     */
    private parseJob_(job: TaskJob | CronJob): any {
        if (job) {
            const command = this.isWin ? job.command_win : job.command_lin;
            if (command) {
                // Options https://nodejs.org/api/child_process.html#child_process_child_process_exec_command_options_callback
                let options: any = {
                    env: process.env
                };
                if (job.options) {
                    options = {
                        ...options,
                        ...job.options
                    }
                }
                if (!this.isWin && this.userid && job.user_lin) {
                    try {
                        const uid = this.userid.uid(job.user_lin);
                        const gid = this.userid.gid(job.user_lin);
                        if (Number.isInteger(uid)) {
                            options.uid = uid;
                            options.env.USER = job.user_lin;
                            options.env.LOGNAME = job.user_lin;
                        }
                        if (Number.isInteger(gid)) {
                            options.gid = gid;
                        }

                        // si on récupére une valeur pour HOME
                        const sHomeEnv = propertiesSrvc.processHomeEnvForUser(job.user_lin);
                        if (typeof sHomeEnv == 'string') {
                            // modification de l'environnement
                            options.env.HOME = sHomeEnv;
                        }
                    } catch (error) {
                        this.oLogger.log('User not founded', error, 'ERROR');
                    }
                }

                return {
                    command,
                    options
                }
            }
        }
    }

    /**
     * Parse un job et retourne une instance axios
     * @param job 
     */
    private parseApiRequest_(job: TaskJob | CronJob): any {
        if (job && job.api_route && job.api_method) {
            return axios.create({
                httpsAgent: new https.Agent({
                    rejectUnauthorized: false
                })
            });
        }
    }

    /**
     * Notifie le début d'une tâche
     * @param oTask 
     */
    private notifyStartTask_(oTask: TaskJob) {
        return new Promise((resolve, reject) => {
            if (oTask.id_) {
                const sQuery = `INSERT INTO ${this.oProperties.schema_vitis}.taskshistory(task_id, status, command) VALUES ($1, $2, $3)`;
                this.oDb.query(sQuery, [oTask.id_, 'RUNNING', oTask.command_]).then((output) => {
                    resolve(output);
                }).catch((err) => {
                    console.error('could not notify start task', err);
                    reject();
                });
            }
        });
    }

    /**
     * Notifie la fin d'une tâche
     * @param oTask 
     * @param err 
     * @param stdout 
     * @param stderr 
     */
    private notifyEndTask_(oTask: TaskJob, err?: any, stdout?: any, stderr?: any) {
        return new Promise((resolve, reject) => {

            // log
            if (err) {
                this.oLogger.log('Task on error', {
                    oTask, err, stdout, stderr
                }, 'ERROR');
            } else {
                this.oLogger.log('Task runned successfully', oTask, 'INFO');
            }

            // Écriture dans la base
            if (oTask.id_) {
                return this.updateTask_(oTask.id_, {
                    status: err ? 'ERROR' : 'SUCCESS',
                    err,
                    stdout,
                    stderr
                }).then((res) => {
                    resolve(res)
                }).catch((err) => {
                    console.error('could not notify end task');
                    reject();
                });
            }
        });
    }

    /**
     * Met à jour les infos d'une tache sur la table taskshistory
     * @param iTaskId 
     * @param oValues 
     */
    private updateTask_(iTaskId: string, oValues: any): Promise<any> {
        return new Promise((resolve, reject) => {

            const aParams: any[] = [iTaskId];

            // Début de la requête
            let sQuery = `UPDATE ${this.oProperties.schema_vitis}.taskshistory SET `;

            // Ajoute les éléments à modifier
            for (const key in oValues) {
                if (oValues.hasOwnProperty(key)) {
                    const value = oValues[key];
                    const counter = aParams.length + 1;

                    // ex: length_sec=$5
                    let queryParam = `${key}=$${counter}`;
                    if (counter > 2) {
                        queryParam = ', ' + queryParam;
                    }

                    sQuery += queryParam;
                    aParams.push(value);
                }
            }

            // Fin de la requête
            sQuery += ' WHERE task_id=$1';

            this.oDb.query(sQuery, aParams).then((output) => {
                resolve(output);
            }).catch((err) => {
                console.error('could not update task history', err);
                reject();
            });
        });
    }

    /**
     * Lance une commande système
     * @param oCommand 
     * @returns 
     */
    private runCommand_(oCommand: any): Promise<any> {
        return new Promise((resolve, reject) => {
            if (oCommand) {
                this.oLogger.log('run command "' + oCommand.command + '"');
                this.oLogger.log('params : ', oCommand.options, 'DEBUG');
                exec(oCommand.command, oCommand.options, (err, stdout, stderr) => {
                    if (err) {
                        console.error(err)
                        this.oLogger.log('error occured while running "' + oCommand.command + '"', {}, 'ERROR');
                    } else if (stdout) {
                        console.log(stdout);
                    } else if (stderr) {
                        this.oLogger.log('error occured while running "' + oCommand.command + '"', {}, 'ERROR');
                        console.log(stderr);
                    }
                    resolve({ err, stdout, stderr });
                });
            }
        });
    }

    /**
     * Lance une requête API
     * @param oInstance 
     * @param oJob 
     * @returns 
     */
    private runApiRequest_(oInstance: any, oJob: TaskJob | CronJob): Promise<any> {
        return new Promise((resolve, reject) => {
            if (oJob.api_route && oJob.api_method) {

                const sMethod = oJob.api_method.toLowerCase();
                const oParams = oJob.api_params ? oJob.api_params : {};
                const sURL = this.oProperties.app_api_url + oJob.api_route;

                this.oLogger.log('run API request  "' + oJob.api_method + ' ' + sURL + '"');
                this.oLogger.log('params : ', oParams, 'DEBUG');

                switch (oJob.api_method) {
                    case 'GET':
                    case 'DELETE':
                        oInstance.get(sURL).then((res: any) => {
                            resolve(res);
                        });
                        break;
                    case 'POST':
                    case 'PUT':
                        oInstance({
                            method: sMethod,
                            url: sURL,
                            data: oParams
                        }).then((res: any) => {
                            resolve(res);
                        });
                        break;
                    default:
                        break;
                }
            }
        });
    }
}

