/**
 *                        Task Runner
 * 
 * Ce fichier permet de lancer les taches de manière synchrone
 * 
 */

import yargs from 'yargs';
import { args } from './taskRunner-yargs-config';
import { propertiesSrvc } from "vitis-engine-lib";
import { Logger } from "vitis-engine-lib";
import { Db } from "vitis-engine-lib";
import { FileConfigSrvc } from "./fileconfigSrvc";
import { RunnerSrvc } from "./runnerSrvc";
import { LocalFS } from "vitis-engine-lib";
import { TaskJob } from './models';

const reloadInterval = 2; // Intervalle de rechargement en secondes
const reloadTimeout = 300; // Intervalle maximum après timeout

/**
 * Lance une commande de manière asynchrone
 * @param oProperties 
 * @param oDb 
 * @param oLogger 
 * @param oLocalFS 
 * @param oFileConfigSrvc 
 * @param taskOptions attributs à incorporer à la tâche
 */
function runCommand(oProperties: any, oDb: Db, oLogger: Logger, oLocalFS: LocalFS, oFileConfigSrvc: FileConfigSrvc, taskOptions: TaskJob = {}) {
    if (typeof args.command == 'string') {

        let task: any = {
            command_lin: args.command,
            command_win: args.command,
        }

        if (typeof args.user == 'string') {
            task.user_lin = args.user;
        }

        if (typeof args.cwd == 'string') {
            task.options = {cwd: args.cwd};
        }

        if (taskOptions) {
            task = { ...task, ...taskOptions };
        }

        // Met à jour le fichier de tâches
        let oTasksConfig = oFileConfigSrvc.getFileConfig('tasks');
        if (oTasksConfig && oTasksConfig.tasks && oTasksConfig.tasks.length) {
            oTasksConfig.tasks.push(task);
        } else {
            oTasksConfig = { tasks: [task] }
        }
        oFileConfigSrvc.updateFileConfig('tasks', oTasksConfig);

        oLogger.log(`Run command "${args.command}"`);
    } else {
        oLogger.log(`command is not a string (Usage : ${args.$0} runCommand --command='...')`);
    }
}

/**
 * Lance une commande de manière synchrone
 * @param oProperties 
 * @param oDb 
 * @param oLogger 
 * @param oLocalFS 
 * @param oFileConfigSrvc 
 */
function runCommandSync(oProperties: any, oDb: Db, oLogger: Logger, oLocalFS: LocalFS, oFileConfigSrvc: FileConfigSrvc) {

    const oRunnerSrvc = new RunnerSrvc(oProperties, oDb, oLogger, oLocalFS, oFileConfigSrvc);
    const id_ = oRunnerSrvc.getUniqueId();
    const taskOptions = { id_ };

    oLogger.log(`Run command with id ${id_}`);

    // Lance la commande
    runCommand(oProperties, oDb, oLogger, oLocalFS, oFileConfigSrvc, taskOptions);

    // Attend le retour
    const maxIterations = reloadTimeout / reloadInterval;
    let counter = 0;
    const inter = setInterval(() => {
        counter++;

        oLogger.log(`Test if command with id ${id_} is finished (${counter})`);

        // Requête SQL
        const sQuery = `SELECT * FROM ${oProperties.schema_vitis}.taskshistory WHERE task_id='${id_}'`;
        oDb.query(sQuery).then((output: any) => {
            if (output && output[0]) {
                if (output[0].status === 'SUCCESS' || output[0].status === 'ERROR') {
                    oLogger.log(`Command with id ${id_} just finished with state ${output[0].status}`);
                    if (output[0].stderr) {
                        oLogger.log(output[0].stderr);
                    }
                    if (output[0].stdout) {
                        oLogger.log(output[0].stdout);
                    }
                    clearInterval(inter);
                    oDb.disconnect();
                }
            }
        }).catch((err) => {
            oLogger.log(`could not get task history`, err, 'ERROR');
            clearInterval(inter);
            oDb.disconnect();
        });

        // Timeout
        if (counter >= maxIterations) {
            clearInterval(inter);
            oDb.disconnect();
        }
    }, reloadInterval * 1000);
}

// Lance la commande en fonction des arguments passés
if (args._[0]) {
    const command: string = args._[0].toString();

    // Récupère les properties
    propertiesSrvc.getProperties().then((oProperties) => {

        const oLogger = new Logger(oProperties);
        const oLocalFS = new LocalFS(oProperties, oLogger);
        const oFileConfigSrvc = new FileConfigSrvc(oLogger);

        if (args.noConsole) {
            oLogger.disableConsoleMessage();
        }

        oLogger.log('Get application properties and logger initalisation', null, 'DEBUG');

        // Base de données
        const oDb = new Db(oProperties, {
            user: oProperties.db_superuser,
            host: oProperties.db_server,
            database: oProperties.db_name,
            password: oProperties.db_password,
            port: oProperties.db_port,
            autoSslConfig: true
        }, oLogger);

        if (args.timeout) {
            if (typeof args.timeout == 'number' && !Number.isNaN(args.timeout)) {
                oDb.setMaxConnectionTime(args.timeout);
            }
        }

        if (args.noTimeout) {
            oDb.disableTimeOut();
        }

        try {
            oLogger.log('Command Start', args, 'DEBUG');
            // si une fonction avec le nom de la commande existe alors elle sera appelée
            return eval(command)(oProperties, oDb, oLogger, oLocalFS, oFileConfigSrvc);
        } catch (error) {
            oLogger.log(`Command error : function not valid : ${command}`);
            return false;
        }
    });
} else {
    console.log('Veuillez spécifier une commande');
    yargs.showHelp("log");
}