<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Controller\AbstractVitisController;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\Crypto\JweCrypto;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\FileAccessor\LocalAccessor;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\VitisRequest;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpFoundation\ResponseHeaderBag;

/**
 * Controller permettant le téléchargement de ressource à travers l'API.
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */

class FileDownloader extends AbstractVitisController {
  /**
   * @var string[]
   */
  private const INLINE_DISPOSITION_MIME_TYPE = [
    'text/plain',
    'application/pdf',
    'text/html',
    'application/xml',
    'application/json',
  ];
  /**
   * @var string[]
   */
  private const ALIAS_PROPERTIES_MAP = [
    'ws_data' => 'ws_data_dir',
    'public' => 'public_dir',
    'upload' => 'upload_dir',
    'shared' => 'shared_dir',
  ];

  /**
   * @var JweCrypto
   */
  private $oJweCrypto;

  /**
   * Initialise le controler et instancie les services requis
   * @param Properties $oPropertiesSrvc Service d'accès aux properties
   * @param VitisRequest $oVitisRequest Service d'abstraction de la requête Symfony
   * @param Route $oRouteConfig Service d'accès à la configuration de la route
   * @param LoggerManager $oLoggerManager Service d'utilsation des logs
   * @param DatabaseConnection $oDatabaseConnection Service d'accès à la base de données
   * @param FileManager $oFileManager Service de gestion des fichiers
   * @param JweCrypto $oJweCrypto Service de manipulation des token JWE
   */
  public function __construct(Properties $oPropertiesSrvc, VitisRequest $oVitisRequest, Route $oRouteConfig, LoggerManager $oLoggerManager, DatabaseConnection $oDatabaseConnection, FileManager $oFileManager, JweCrypto $oJweCrypto) {
    parent::__construct($oPropertiesSrvc, $oVitisRequest, $oRouteConfig, $oLoggerManager, $oDatabaseConnection, $oFileManager);
    $this->oJweCrypto = $oJweCrypto;
  }

  /**
   * Lit un fichier et le retourne dans le buffer de sorie.
   * @param array $aReturn Retour préformaté
   * @param array $aData Données de la requête
   * @return Response réponse à retourner par l'API
   */
  private function downloadFileWithKeyAndTag(array $aReturn, array $aData): Response {
    $sPath = $this->keyAliasReplacer($aData['key']);
    if (empty($sPath)) {
      // le fichier ne peut pas être télécharger
      return $this->formatGenericError('DOWNLOADER_BAD_DIRECTORY_FOR_DOWNLOAD', '', 'Files in this directory can\'t be downloaded', Response::HTTP_UNAUTHORIZED);
    }

    return $this->getFileResponse($aReturn, $aData, $sPath);
  }
  /**
   * Lit un fichier et le retourne dans le buffer de sorie.
   * @return Response réponse à retourner par l'API
   */
  public function downloadFile(): Response {
    $aReturn = ['status' => 0];
    $aData = $this->requestToArray(true);

    // si un paramètre est manquant on coupe
    if ($this->checkRequestParameters($aData, ['key', 'eTag'])) {
      return $this->downloadFileWithKeyAndTag($aReturn, $aData);
    }

    if ($this->checkRequestParameters($aData, ['file'])) {
      return $this->downloadFileWithCryptedFile($aReturn, $aData);
    }

    // le fichier n'existe pas
    return $this->formatErrorMissingParameter();
  }
  /**
   * Lit un fichier et le retourne dans le buffer de sorie.
   * @param array $aReturn Retour préformaté
   * @param array $aData Données de la requête
   * @return Response réponse à retourner par l'API
   */
  private function downloadFileWithCryptedFile(array $aReturn, array $aData): Response {
    // decrypt
    $aCryptedData = $this->oJweCrypto->decryptData($aData['file']);
    // check si le fichier à télécharger se trouve bien dans ws_data, public, upload ou shared (ALIAS_PROPERTIES_MAP)
    if (!$this->checkRequestParameters($aCryptedData, ['key'])) {
      // le fichier n'existe pas
      return $this->formatErrorMissingParameter();
    }

    // gestion du booléen pour forcer l'interface locale
    $bForceLocalAccessor = false;

    if (isset($aCryptedData['fs'])) {
      if ($aCryptedData['fs'] === Properties::LOCAL_FS) {
        $bForceLocalAccessor = true;
      }
    }

    $aData = array_merge($aData, $aCryptedData);

    $sPath = $this->keyAliasReplacer($aData['key']);
    if (empty($sPath)) {
      $sPath = $aData['key'];
    }

    return $this->getFileResponse($aReturn, $aData, $sPath, $bForceLocalAccessor);
  }
  /**
   * Lit un fichier et le retourne dans le buffer de sorie.
   * @param array $aReturn Retour préformaté
   * @param array $aData Données de la requête
   * @param string $sPath
   * @param bool $bForceLocalAccessor force l'accesseur local
   * @return Response réponse à retourner par l'API
   */
  private function getFileResponse(array $aReturn, array $aData, string $sPath, bool $bForceLocalAccessor = false): Response {
    // récupère l'interface fichier
    $oFileAccessor = ($bForceLocalAccessor) ? $this->getLocalFileAccessor() : $this->getFileAccessor();

    if (!$oFileAccessor->file_exists($sPath)) {
      return $this->formatGenericErrorFileNotFound();
    }

    if (isset($aData['eTag'])) {
      if ($oFileAccessor->getFileEtag($sPath) !== $aData['eTag']) {
        // le fichier n'existe pas
        return $this->formatErrorMissingParameter('You can\'t download this file (Bad Etag)');
      }
    }

    $sFileName = $oFileAccessor->getFileName($sPath);
    $sContentType = $oFileAccessor->mime_content_type($sPath);

    // renvoi du nom du fichier pour le file_picker
    if (isset($aData['type'])) {
      if ($aData['type'] === 'document') {
        // on retourne direct la réponse pour ne pas lancer le téléchargement du fichier
        $aReturn['data'] = $sFileName;
        $aReturn['status'] = 1;
        $this->formatResponse($aReturn);
        // le nom du fichier est toujours récupéré depuis le header
        $this->oResponse->headers->set('Content-disposition', 'info-package; filename="' . rawurlencode($sFileName) . '"');
        // Ajout des headers de réponse considérés comme safe par le navigateur
        $this->oResponse->headers->set('Access-Control-Expose-Headers', 'Content-disposition');
        return $this->oResponse;
      }
    }

    // Récupération du thumbnail
    if (isset($aData['thumbnail'])) {
      if ($aData['thumbnail'] === 'true') {
        $sMinFilePath = LocalAccessor::thumbnailPathGenerator($sPath);
        if ($oFileAccessor->file_exists($sMinFilePath)) {
          $sPath = $sMinFilePath;
        }
      }
    }
    // si différent de fs copy local et suppression à la fin
    $bDeleteAfterSend = false;
    if ($aData['fs'] !== Properties::LOCAL_FS) {
      $oFileAccessor->copyDistantToLocal($sPath, $sPath);
      $bDeleteAfterSend = true;
    }
    $sDisposition = ResponseHeaderBag::DISPOSITION_ATTACHMENT;
    if (in_array($sContentType, self::INLINE_DISPOSITION_MIME_TYPE)) {
      $sDisposition = ResponseHeaderBag::DISPOSITION_INLINE;
    }
    $aHeaders = [
      'Content-Type-Set-By' => 'vitis/filedownloader',
      'Content-Type' => $sContentType . '; charset=UTF-8',
    ];
    if ($sContentType === "application/octet-stream") {
      $aHeaders['Content-Transfer-Encoding'] = 'Binary';
    }

    // création du blob
    return $this->getBlobResponse($sPath, $sFileName, $aHeaders, $bDeleteAfterSend, $sDisposition);
  }

  /**
   * @param array $aData Donnée à vérifier
   * @param array $aKey liste des clé à vérifer
   * @return boolean false si un clé est manquante true sinon
   */
  private function checkRequestParameters(array $aData, array $aKey): bool {
    foreach ($aKey as $sKey) {
      if (empty($aData[$sKey])) {
        return false;
      }
    }
    return true;
  }

  /**
   * remplace un alias par un properties (map de correspondance défini dans une constante ALIAS_PROPERTIES_MAP)
   * @param string $sKey Clé vers le fichier où il faut remplacer l'alias
   * @return string|null retourne le chemin correspondant à la clé si elle est touvée sinon retourne null
   */
  private function keyAliasReplacer(string $sKey): ?string {
    foreach (self::ALIAS_PROPERTIES_MAP as $sAlias => $sPropertiesKey) {
      if (strpos($sKey, $sAlias) === 0) {
        return str_replace($sAlias, $this->oProperties->get($sPropertiesKey), $sKey);
      }
    }
    return null;
  }
}
