<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Controller\AbstractVitisController;
use App\Module\Vitis\WebService\Service\DataAccessor\AbstractDataAccessor;
use App\Module\Vitis\WebService\Service\Normalizer\VitisNormalizer;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Serializer\Normalizer\AbstractNormalizer;
use App\Module\Vitis\WebService\Repository\GenericRepository;
use Doctrine\ORM\EntityManager;

/**
 * Controller Générique de Vitis.
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 * @author Alexandre Mironenkov <alexandre.mironenkov@veremes.com>
 */
class GenericController extends AbstractVitisController {
  /**
   * Récupère toutes les entités pour une route de type GET Module/Objet.
   * @return Response réponse à retourner par l'API
   */
  public function getOptions(): Response {
    return $this->oResponse;
  }

  /**
   * Récupère toutes les entités pour une route de type GET Module/Objet.
   * @return Response réponse à retourner par l'API
   */
  public function getDatas(): Response {
    $aData = $this->requestToArray();
    // chargement de la connexion à la base de donnée
    $aEntities = $this->getDataWithRequestParams($aData);
    if ($aEntities['status'] == 0) {
      return $this->formatGenericResponseForErrorOnSelect();
    }

    $aSerializerContext = $this->getSerializerContext();

    // transforme le tableau de réponse dans le résultat souhaité
    return $this->formatResponse($aEntities, Response::HTTP_OK, $aSerializerContext);
  }

  /**
   * Récupère une entité pour une route de type GET Module/Objet/{{id}}.
   * @param mixed $id id de l'entité à trouver
   * @return Response réponse à retourner par l'API
   */
  public function getData($id): Response {
    $aData = $this->requestToArray();
    // chargement de la connexion à la base de donnée
    $oEntity = $this->getDataWithRequestParams($aData, $id);
    if ($oEntity['status'] == 0) {
      return $this->formatGenericResponseForErrorOnSelect();
    }
    $aSerializerContext = $this->getSerializerContext();

    return $this->formatResponse($oEntity, Response::HTTP_OK, $aSerializerContext);
  }

  /**
   * Met à jour une entité pour une route de type PUT Module/Objet/{{id}}.
   * @param mixed $id id de l'entité à trouver
   * @return Response réponse à retourner par l'API
   */
  public function updateData($id): Response {
    $sData = $this->requestToJson();
    // Sauvegarde en base de donnée
    $oEntity = $this->updateDataInDb($id, $sData);
    if ($oEntity['status'] == 0) {
      return $this->formatGenericResponseForErrorOnUpdate();
    }

    return $this->formatResponse($oEntity);
  }

  /**
   * Créé une entité pour une route de type POST Module/Objet.
   * @return Response réponse à retourner par l'API
   */
  public function postData(): Response {
    $sData = $this->requestToJson();
    // Sauvegarde en base de donnée
    $oEntity = $this->insertDataInDb($sData);

    if ($oEntity == null) {
      return $this->formatGenericResponseForErrorOnInsert();
    }

    if ($oEntity['status'] == 0) {
      return $this->formatGenericResponseForErrorOnInsert();
    }
    return $this->formatResponse($oEntity);
  }

  /**
   * Supprime une entité pour une route de type POST Module/Objet/{{id}}.
   * @param mixed $id id de l'entité à supprimer
   * @return Response réponse à retourner par l'API
   */
  public function deleteData($id): Response {
    // Sauvegarde en base de donnée
    $oEntity = $this->deleteDataInDb($id);
    if ($oEntity['status'] == 0) {
      return $this->formatGenericResponseForErrorOnDelete();
    }
    return $this->formatResponse($oEntity);
  }

  /**
   * Supprime les entités d'une liste pour une route de type POST Module/Objet.
   * @return Response réponse à retourner par l'API
   */
  public function deleteDatas(): Response {
    $aData = $this->massiveRequestGetIdList();
    if (empty($aData)) {
      return $this->formatGenericResponseForMissingIdList();
    }
    // Sauvegarde en base de donnée
    $aEntities = $this->deleteDataInDb($aData['idList']);
    if ($aEntities['status'] == 0) {
      return $this->formatGenericResponseForErrorOnDelete();
    }
    return $this->formatResponse($aEntities);
  }

  /**
   * Met à jour plusieus entités pour une route de type PUT Module/Objet.
   * La requête doit contenir un paramètre idList
   * @return Response réponse à retourner par l'API
   */
  public function updateDatas(): Response {
    $aData = $this->massiveRequestGetIdList();
    if (empty($aData)) {
      return $this->formatGenericResponseForMissingIdList();
    }
    // Sauvegarde en base de donnée
    $oEntities = $this->updateMultipleDataInDb($aData['idList'], $aData);
    return $this->formatResponse($oEntities);
  }

  /**
   * Vérifie que la clé idList existe et la transforme  en tableau si c'est un string
   * @return array|null retourne la liste des id à mettre à jour ou null si la clé n'existe pas
   */
  protected function massiveRequestGetIdList(): ?array{
    $aData = $this->requestToArray();

    if (!isset($aData['idList'])) {
      return null;
    }

    if (!is_array($aData['idList'])) {
      $aData['idList'] = explode('|', $aData['idList']);
    }

    return $aData;
  }

  /**
   * Retourne le contexte de sérialisation pour les get
   *
   * @return array
   */
  protected function getSerializerContext(): array{
    $aData = $this->requestToArray();
    $aSerializerContext = [];
    if (isset($aData['attributs'])) {
      $aSerializerContext[AbstractNormalizer::ATTRIBUTES] = explode('|', $aData['attributs']);
    }

    if (isset($aData['distinct'])) {
      $aSerializerContext[VitisNormalizer::DISTINCT] = ($aData['distinct'] === 'true');
    }

    return $aSerializerContext;
  }

  /**
   * Fonction permettant de récupérer des données dpuis la base d emanière générique
   *
   * @param  array $aData donnée venant de la requête
   * @param  mixed $id identifiant de l'objet si null on passe ne mode listing
   * @param  AbstractDataAccessor|null $oDa 
   * @return array
   */
  protected function getDataWithRequestParams(array $aData, $id = null, ?AbstractDataAccessor $oDa = null): array{
    $oDataAccessor = $oDa ?? $this->getDataAccessor();
    if ($id !== null) {
      return $oDataAccessor->getEntity($id, $aData);
    }
    return $oDataAccessor->getEntities($aData);
  }

  /**
   * Insert une donnée
   *
   * @param  array|string     $mData string json ou un array viable pour json_encode
   * @param  AbstractDataAccessor|null $oDa 
   * @return array|null
   */
  protected function insertDataInDb($mData, ?AbstractDataAccessor $oDa = null): ?array{
    $sData = $this->convertDataForDataAccessorWriteMode($mData);
    // s'il y a une erreur dnas le sodnnées on bloque à ce niveau
    if ($sData === null) {
      return null;
    }

    $oDataAccessor = $oDa ?? $this->getDataAccessor();

    // Sauvegarde en base de donnée
    return $oDataAccessor->insertEntity($sData);
  }

  /**
   * Insert une donnée
   *
   * @param  array|string     $mData string json ou un array viable pour json_encode
   * @param  mixed            $id identifiant de l'objet à mettre à jour
   * @param  AbstractDataAccessor|null $oDa 
   * @return array|null
   */
  protected function updateDataInDb($id, $mData, ?AbstractDataAccessor $oDa = null): ?array{
    $sData = $this->convertDataForDataAccessorWriteMode($mData);
    // s'il y a une erreur dnas le sodnnées on bloque à ce niveau
    if ($sData === null) {
      return null;
    }

    $oDataAccessor = $oDa ?? $this->getDataAccessor();

    // Sauvegarde en base de donnée
    return $oDataAccessor->updateEntity($id, $sData);
  }

  /**
   * Insert une donnée
   *
   * @param  array            $ids identifiant de l'objet à mettre à jour
   * @param  array|string     $mData string json ou un array viable pour json_encode
   * @param  AbstractDataAccessor|null $oDa 
   * @return array|null
   */
  protected function updateMultipleDataInDb(array $ids, $mData, ?AbstractDataAccessor $oDa = null): ?array{
    $sData = $this->convertDataForDataAccessorWriteMode($mData);
    // s'il y a une erreur dnas le sodnnées on bloque à ce niveau
    if ($sData === null) {
      return null;
    }

    $oDataAccessor = $oDa ?? $this->getDataAccessor();

    // Sauvegarde en base de donnée
    return $oDataAccessor->updateEntities($ids, $sData);
  }

  /**
   * supprimer une ou plusieurs row de la base
   *
   * @param  mixed|mixed[] $id si c'est un array d'id on lance la suppression multiple, sinon suppression simple
   * @param  AbstractDataAccessor|null $oDa 
   * @return array
   */
  protected function deleteDataInDb($id, ?AbstractDataAccessor $oDa = null): array{
    // chargement de la connexion à la base de donnée
    $oDataAccessor = $oDa ?? $this->getDataAccessor();
    // Sauvegarde en base de donnée

    if (is_array($id)) {
      return $oDataAccessor->deleteEntities($id);
    }

    return $oDataAccessor->deleteEntity($id);
  }

  /**
   * Transforme l'entrée en json acceptable pour l'insertion/mise à jour
   *
   * @param  array|string|object $mData
   * @return string|null string json si possible null sinon
   */
  protected function convertDataForDataAccessorWriteMode($mData): ?string {
    // Vérification si la chaine est une chaine conforme json ou un objet transformable en chaine json
    if (is_string($mData)) {
      if (json_decode($mData, true) !== null) {
        return $mData;
      }
      $this->oLoggerManager->get('API_ERROR')->error('Bad Json string in input');
    } else {
      $sData = json_encode($mData);
      // si l'objet/array a pu être transformer en json on le retourne sinon problème
      if ($sData !== false) {
        return $sData;
      }
      $this->oLoggerManager->get('API_ERROR')->error('Data cannot be convert to Json');
    }
    return null;
  }

  /**
   * Renvoie le repository d'une classe
   * @param string $classEntity namespace de la classe, il suffit de faire monEntité::class
   * @param EntityManager|null $oEm 
   */
  public function getRepository(string $classEntity, ?EntityManager $oEm = null): GenericRepository|null {

    if (empty($oEm)) {
      $oEm = $this->oDatabaseConnection->getUserEntityManager();
    }

    $oRepository = $oEm->getRepository($classEntity);
    if (!$oRepository instanceof GenericRepository) {
      return null;
    }
    return $oRepository;
  }
}
