<?php

namespace App\Module\Vitis\WebService\Controller;

use Doctrine\ORM\EntityManager;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller Générique de Vitis.
 *
 * @author Ludivine Premont <ludivine.premont@veremes.com>
 */
class GenericQueriesController extends GenericController {
  /**
   * Permet de récupérer les base de données locales
   * @return Response
   */
  public function getDatabases(): Response{
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor();

    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oQueryAccessor)) {
      return $this->formatPostgresBadConnection();
    }

    $aData = $this->requestToArray();

    // Récupération des bases de données
    $oDatabases = $oQueryAccessor->getDatabases($aData);
    if (!isset($oDatabases)) {
      return $this->formatResponse(['status' => 0]);
    }

    $response = ['status' => 1, 'data' => []];
    foreach ($oDatabases as $oDatabase) {
      if (array_key_exists('attributs', $aData) && !empty($aData['attributs'])) {
        $response['data'][] = array('database' => $oDatabase['database']);
      } else {
        $response['data'][] = $oDatabase['database'];
      }
    }

    return $this->formatResponse($response);
  }

  /**
   * Permet de récupérer les schémas d'une base de données locale
   * @param mixed $databaseName Nom de la base de données pour laquelle on veut récupérer les schémas
   * @return Response réponse à retourner par l'API
   */
  public function getSchemas($databaseName) {
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($this->oDatabaseConnection->createExternalEntityManager($databaseName));

    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oQueryAccessor)) {
      return $this->formatPostgresBadConnection();
    }

    $aData = $this->requestToArray();
    // Récupération des schemas
    $oSchemas = $oQueryAccessor->getSchemas($aData);
    if (!isset($oSchemas)) {
      return $this->formatResponse(['status' => 0]);
    }

    $response = ['status' => 1, 'data' => []];
    foreach ($oSchemas as $oSchema) {
      if (array_key_exists('attributs', $aData) && !empty($aData['attributs'])) {
        $response['data'][] = array('schema' => $oSchema['schema_name']);
      } else {
        $response['data'][] = $oSchema['schema_name'];
      }
    }

    return $this->formatResponse($response);
  }

  /**
   * Permet de récupérer les schémas d'une base de données locale en passant les paramètres par le corps de la requête
   * @return Response réponse à retourner par l'API
   */
  public function getSchemasWithBody(): Response{
    $aData = $this->requestToArray();

    if (!isset($aData['database'])) {
      return $this->formatMissingParametersError();
    }

    if (empty($aData['database'])) {
      return $this->formatResponse(array('data' => [], "status" => 1));
    }

    $databaseName = $aData['database'];

    return $this->getSchemas($databaseName);
  }

  /**
   * Permet de récupérer les tables d'un schema d'une base de données locale
   * @return Response réponse à retourner par l'API
   */
  public function getTables() {
    $aData = $this->requestToArray();

    if (!isset($aData['database']) || empty($aData['database']) || !isset($aData['schema']) || empty($aData['schema'])) {
      return $this->formatMissingParametersError();
    }

    $oEntityManager = $this->getEntityManagerFromData($aData);
    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oEntityManager)) {
      return $this->formatPostgresBadConnection();
    }

    if (isset($aData['sgbd']) && !empty($aData['sgbd']) && $aData['sgbd'] == 'pdo_oci') {
      $oQueryAccessor = $this->oDatabaseConnection->getOracleAccessor($oEntityManager);
    } else {
      $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEntityManager);
    }
    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oQueryAccessor)) {
      return $this->formatPostgresBadConnection();
    }

    // Récupération des tables
    $oTables = $oQueryAccessor->getTables($aData);
    if (!isset($oTables)) {
      return $this->formatResponse(['status' => 0]);
    }

    $response = ['status' => 1, 'data' => []];
    foreach ($oTables as $oTable) {
      if (array_key_exists('attributs', $aData) && !empty($aData['attributs'])) {
        $response['data'][] = array('table' => $oTable['table_name']);
      } else {
        $response['data'][] = $oTable['table_name'];
      }
    }

    return $this->formatResponse($response);
  }

  /**
   * Permet de récupérer des données de n'importe quel base de données
   * @param array|null $aData
   * @return Response réponse à retourner par l'API
   */
  public function getGeneric( ? array $aData = null) : Response {
    if (!isset($aData)) {
      $aData = $this->requestToArray();
    }

    if (!isset($aData['database']) || empty($aData['database'])) {
      $aData['database'] = $this->oProperties->get('db_name');
    }

    if (!isset($aData['table']) || empty($aData['table']) ||
      !isset($aData['database']) || empty($aData['database']) ||
      !isset($aData['schema']) || empty($aData['schema'])) {
      return $this->formatMissingParametersError();
    }

    $oEntityManager = $this->getEntityManagerFromData($aData);
    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oEntityManager)) {
      return $this->formatPostgresBadConnection();
    }

    if (isset($aData['sgbd']) && !empty($aData['sgbd']) && $aData['sgbd'] == 'pdo_oci') {
      $oQueryAccessor = $this->oDatabaseConnection->getOracleAccessor($oEntityManager);
    } else {
      $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEntityManager);
    }
    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oQueryAccessor)) {
      return $this->formatPostgresBadConnection();
    }

    $aFields = $oQueryAccessor->getFieldsFromTable($aData);
    if ($aFields['status'] == 0) {
      return $this->formatResponse($aFields);
    }

    // Si columns_type défini, gère les données spécifiques
    if (isset($aData['columns_type'])) {
      for ($i = 0; $i < count($aFields['data']); $i++) {
        $sName = $aFields['data'][$i]['name'];
        if (isset($aData['columns_type'][$sName])) {
          switch ($aData['columns_type'][$sName]) {
          case 'image':
            if (isset($aData['file_path'])) {
              // Modifie le type
              $aFields['data'][$i]['type'] = 'file';
              $aFields['data'][$i]['image'] = true;
              $aFields['data'][$i]['multiple'] = true;
              // Ajoute l'url
              $aFields['data'][$i]['path'] = $aData['file_path'];
            }
            break;
          case 'document':
            if (isset($aData['file_path'])) {
              // Modifie le type
              $aFields['data'][$i]['type'] = 'file';
              $aFields['data'][$i]['image'] = false;
              $aFields['data'][$i]['multiple'] = true;
              // Ajoute l'url
              $aFields['data'][$i]['path'] = $aData['file_path'];
            }
            break;
          case 'geom':
            $aFields['data'][$i]['type'] = 'geom';
            break;
          default:
            break;
          }
        }
      }
    }

    $options = [
      'schema' => $aData['schema'],
      'table' => $aData['table'],
      'fields' => $aFields['data'],
    ];

    if (isset($aData['sgbd']) && !empty($aData['sgbd']) && $aData['sgbd'] == 'pdo_oci') {
      $options['field_id'] = $aFields['data'][0]['name'];
    }

    // Si on connait le champs identifiant on le passe à la config et on l'ajoute au attributs
    if (isset($aData['field_id'])) {
      $options['field_id'] = $aData['field_id'];
      if (isset($aData['attributs'])) {
        $aData['attributs'] .= '|' . $aData['field_id'];
      }
    }

    $oQueryAccessor->setOptions($options);

    // Ajout du filtre si besoin
    if (isset($aData['filter_value']) && isset($aData['filter_attr']) && isset($aData['values'])) {
      if (!isset($aData['values'][$aData['filter_value']]) || $aData['values'][$aData['filter_value']] == '') {
        return $this->formatMissingParametersError('La valeur de filter_value (' . $aData['filter_value'] . ') n\'est pas présente dans les valeurs');

      }

      // Ajouter le filtre
      $aData['filter'] = json_encode([
        'relation' => 'AND',
        'operators' => [[
          'column' => $aData['filter_attr'],
          'compare_operator' => 'IN',
          'value' => $aData['values'][$aData['filter_value']],
        ]],
      ]);
    }

    // Récupération des tables
    $response = $oQueryAccessor->getEntities($aData);

    if (array_key_exists('data', $response) &&
      array_key_exists('charset', $aData) &&
      in_array($aData['charset'], ['ISO-8859-1', 'WE8ISO8859P15'])) {
      array_walk_recursive(
        $response,
        function (mixed&$entry) {
          if (is_string($entry)) {
            $entry = mb_convert_encoding(
              $entry,
              'UTF-8',
              'ISO-8859-1'
            );
          }
        }
      );
    }

    return $this->formatResponse($response);
  }

  /**
   * Permet de récupérer les données de n'importe quel column de n'importe quel bdd
   * @param array|null $aData
   * @return Response réponse à retourner par l'API
   */
  public function getDatasColumn( ? array $aData = null) : Response {
    if (!isset($aData)) {
      $aData = $this->requestToArray();
    }

    if (!isset($aData['database']) || empty($aData['database'])) {
      $aData['database'] = $this->oProperties->get('db_name');
    }

    if (!isset($aData['table']) || empty($aData['table']) ||
      !isset($aData['database']) || empty($aData['database']) ||
      !isset($aData['schema']) || empty($aData['schema'])) {
      return $this->formatMissingParametersError();

    }

    $oEntityManager = $this->getEntityManagerFromData($aData);
    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oEntityManager)) {
      return $this->formatPostgresBadConnection();
    }

    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEntityManager);
    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oQueryAccessor)) {
      return $this->formatPostgresBadConnection();

    }

    $sDatas = $oQueryAccessor->getDatasFromColumn($aData);

    if ($sDatas['status'] == 0) {
      return $this->formatResponse($sDatas);
    }

    return $this->formatResponse($sDatas);
  }

  /**
   * Récupère l'entity manager nécessaire en fonction des paramètres.
   *
   * @param mixed $aData
   * @return EntityManager|null EntityManager
   */
  protected function getEntityManagerFromData($aData):  ? EntityManager {
    // si la clé n'existe pas on la set à null pour le reste
    if (!array_key_exists('database', $aData)) {
      $aData['database'] = null;
    }

    if (!empty($aData['login']) && !empty($aData['password']) &&
      !empty($aData['server']) && !empty($aData['port']) &&
      !empty($aData['sgbd'])) {
      $oEntityManager = $this->oDatabaseConnection->createExternalEntityManager(
        $aData['database'],
        $aData['login'],
        $aData['password'],
        $aData['sgbd'],
        $aData['server'],
        $aData['port'],
        $aData['charset'] ?? null
      );
    } else {
      $oEntityManager = $this->oDatabaseConnection->createExternalEntityManager($aData['database']);
    }

    return $oEntityManager;
  }

  /**
   * @param string $sMessage message de log
   * @return Response
   */
  protected function formatMissingParametersError(string $sMessage = '') : Response {
    return $this->formatGenericError('MISSING_PARAMS', '', $sMessage, Response::HTTP_OK);
  }

  /**
   * Récupère la liste des colonnes disponibles dans la table spécifiée
   * @return Response réponse à retourner par l'API
   */
  public function getColumnsDef(): Response{
    $aData = $this->requestToArray();
    // Connexion à la base de données sélectionnée.
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($this->oDatabaseConnection->createExternalEntityManager($aData['database']));

    // Si la connexion à la base de données n'a pas pu être récupéré.
    if (empty($oQueryAccessor)) {
      return $this->formatPostgresBadConnection();
    }

    // Récupère la liste des colonnes de la table (nom + type).
    $aColumns = $oQueryAccessor->getColumns($aData);
    if (!isset($aColumns)) {
      return $this->formatResponse(['status' => 0]);
    }

    // Formattage de la liste des colonnes en type de champ de formulaire.
    $response = ['status' => 1, 'data' => $this->getColumnProperty($aColumns)];

    return $this->formatResponse($response);
  }

  /**
   * Return les colonnes d'une table sous format formulaire / ou name&type
   * @param array $aColumns
   * @param bool $fields
   * @return array
   */
  private function getColumnProperty(array $aColumns, bool $fields = false): array{
    $response = [];
    foreach ($aColumns as $oColumn) {
      $aColumnData = [];
      if (strpos($oColumn['data_type'], 'bigint') !== false
        || strpos($oColumn['data_type'], 'integer') !== false
        || strpos($oColumn['data_type'], 'smallint') !== false
      ) {
        $data = [
          'type' => 'integer',
          'name' => $oColumn['column_name'],
        ];
        if (!$fields) {
          $data['label'] = $oColumn['column_name'];
          $data['nb_cols'] = 12;
        }
        $aColumnData = $data;
      } else if (strpos($oColumn['data_type'], 'time zone') !== false || strpos($oColumn['data_type'], 'date') !== false) {
        $data = [
          'type' => 'date',
          'name' => $oColumn['column_name'],
        ];
        if (!$fields) {
          $data['label'] = $oColumn['column_name'];
          $data['nb_cols'] = 12;
        }
        $aColumnData = $data;
      } else if (strpos($oColumn['data_type'], 'boolean') !== false) {
        $data = [
          'type' => 'radio',
          'name' => $oColumn['column_name'],
        ];
        if (!$fields) {
          $data['label'] = $oColumn['column_name'];
          $data['nb_cols'] = 12;
          $data['options'] = [
            'choices' => [
              [
                'label' => 'FORM_YES',
                'value' => true,
              ],
              [
                'label' => 'FORM_NO',
                'value' => false,
              ],
            ],
          ];
        }
        $aColumnData = $data;
      } else {
        $data = [
          'type' => 'string',
          'name' => $oColumn['column_name'],
        ];
        if (!$fields) {
          $data['type'] = 'text';
          $data['label'] = $oColumn['column_name'];
          $data['nb_cols'] = 12;
        }
        $aColumnData = $data;
      }
      $response[] = $aColumnData;
    }
    return $response;
  }

  /**
   * Permet de retourner la PRIMARY KEY d'une table, ou une clef UNIQUE, si la table en possède une
   *
   * @return Response réponse à retourner par l'API
   */
  public function getColumnUniqueIdentifier(): Response{
    $aData = $this->requestToArray();
    $response = ['status' => 0];

    if (!isset($aData['database']) && empty($aData['database'])
      || !isset($aData['schema']) && empty($aData['schema'])
      || ((!isset($aData['table']) && empty($aData['table'])) && (!isset($aData['pg_table']) && empty($aData['pg_table'])))
    ) {
      return $this->formatMissingParametersError();
    }

    // Connexion u_vitis
    $oEntityManager = $this->oDatabaseConnection->getAdminEntityManager();

    // cas ou la base de données utilisée est Oracle
    if (isset($aData['sgbd']) && !empty($aData['sgbd']) && $aData['sgbd'] == 'pdo_oci') {
      $response['status'] = 1;
      $response['data'] = null;
      return $this->formatResponse($response);
    }

    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEntityManager);

    // Si la connexion à la base de données n'a pas pu être récupérée.
    if (empty($oQueryAccessor) || $oQueryAccessor == null) {
      return $this->formatPostgresBadConnection();
    }

    $oQueryAccessor = $oQueryAccessor->createQueryBuilder();

    // On récupère les colonnes si elles ont la contrainte Primary key ou Unique
    $oQueryAccessor->select('ccu.column_name, tc.constraint_type')
      ->from('information_schema.table_constraints', 'tc')
      ->leftJoin('tc', 'information_schema.constraint_column_usage', 'ccu', 'tc.constraint_catalog = ccu.constraint_catalog and tc.constraint_name = ccu.constraint_name and tc.table_name = ccu.table_name and tc.constraint_schema = ccu.table_schema')
      ->where('(tc.table_schema = [TABLE_SCHEMA] and tc.table_name = [TABLE_NAME] and  (tc.constraint_type = [PK] or tc.constraint_type = [UK]))');

    $aParams = [
      'TABLE_SCHEMA' => ['value' => $aData['schema'], 'type' => 'string'],
      'TABLE_NAME' => ['value' => isset($aData['table']) && !empty($aData['table']) ? $aData['table'] : $aData['pg_table'], 'type' => 'string'],
      'PK' => ['value' => 'PRIMARY KEY', 'type' => 'string'],
      'UK' => ['value' => 'UNIQUE', 'type' => 'string'],
    ];

    $aColumns = $oQueryAccessor->executeWithParams($aParams)->fetchAllAssociative();

    # Retourne la Clef Primaire si la table en contient une
    foreach ($aColumns as $aColumn) {
      if ($aColumn['constraint_type'] === "PRIMARY KEY") {
        $response['status'] = 1;
        $response['data']['column_name'] = $aColumn['column_name'];
        return $this->formatResponse($response);
      }
    }

    // Retourne la première colonne UNIQUE si il y en a une
    if (isset($aColumns[0]['column_name']) && !empty($aColumns[0]['column_name'])) {
      $response['status'] = 1;
      $response['data']['column_name'] = $aColumns[0]['column_name'];
      return $this->formatResponse($response);
    }

    $response['status'] = 1;
    $response['data'] = null;
    return $this->formatResponse($response);
  }
}