<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Entity\Domain;
use App\Module\Vitis\WebService\Entity\DomainGroup;
use App\Module\Vitis\WebService\Entity\Group;
use App\Module\Vitis\WebService\Entity\User;
use App\Module\Vitis\WebService\Repository\DomainRepository;
use App\Module\Vitis\WebService\Repository\GenericRepository;
use App\Module\Vitis\WebService\Repository\UserRepository;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller Générique de Vitis.
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 * @author Alexandre Mironenkov <alexandre.mironenkov@veremes.com>
 */
class LdapController extends GenericController {
  /**
   * Créé un domaine
   * @return Response réponse à retourner par l'API
   */
  public function createDomain(): Response {
    // Création du domaine
    $sData = $this->requestToJson();
    // chargement de la connexion à la base de donnée
    $oEm = $this->oDatabaseConnection->getUserEntityManager();
    $oDataAccessor = $this->getDataAccessor($oEm);
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);
    // Sauvegarde en base de donnée
    $oEntity = $oDataAccessor->insertEntity($sData);
    if ($oEntity['status'] == 0) {
      return $this->formatGenericResponseForErrorOnInsert();
    }
    $oDomain = $oEntity['data'];

    if (!$oDomain instanceof Domain) {
      return $this->formatGenericResponseForErrorOnInsert();
    }
    // creation d'un groupe pour les utilisateurs du domaine
    $oDomain->setLoginRole('vitis_domain_' . $oDomain->getDomainId());
    $oEm->persist($oDomain);
    $oEm->flush($oDomain);
    if (!$oQueryAccessor->ifPgRoleExist($oDomain->getLoginRole())) {
      $oQueryAccessor->createRole($oDomain->getLoginRole());
    }
    return $this->formatResponse($oEntity);
  }

  /**
   * Supprime des domaines
   * @return Response réponse à retourner par l'API
   */
  public function deleteDomains(): Response {
    $aData = $this->massiveRequestGetIdList();
    if (empty($aData)) {
      return $this->formatGenericResponseForMissingIdList();
    }
    // Chargement de la connexion à la base de donnée
    $oEm = $this->oDatabaseConnection->getUserEntityManager();
    $oDataAccessor = $this->getDataAccessor($oEm);
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);
    /* @var GenericRepository $domainRepository */
    $domainRepository = $oEm->getRepository(Domain::class);

    if (!$domainRepository instanceof DomainRepository) {
      return $this->formatGenericError('', '', 'Cannot find Domain suffixes, bad repository', Response::HTTP_NOT_FOUND);
    }

    foreach ($aData['idList'] as $iDomainId) {
      /* @var $oDomain Domain */
      $oDomain = $domainRepository->find($iDomainId);
      if ($oDomain === null) {
        $this->oLoggerManager->get('API_ERROR')->warning('Cannot find domain with id ' . $iDomainId . ' to remove postgres rolegroup');
        continue;
      }
      $sRoleGroupeName = $oDomain->getLoginRole();
      if ($sRoleGroupeName === null) {
        $this->oLoggerManager->get('API_ERROR')->debug('Domain with id ' . $iDomainId . ' postgres rolegroup is null');
        continue;
      }
      $oQueryAccessor->dropRole($sRoleGroupeName);
    }
    // Sauvegarde en base de donnée
    $aEntities = $oDataAccessor->deleteEntities($aData['idList']);
    if ($aEntities['status'] == 0) {
      return $this->formatGenericResponseForErrorOnDelete();
    }
    return $this->formatResponse($aEntities);
  }

  /**
   * Récupère la liste des suffixes
   * @return Response réponse à retourner par l'API
   */
  public function getSuffixes(): Response {
    // Récupération d'UserEntityManager
    $oEm = $this->oDatabaseConnection->getAdminEntityManager();

    $domainRepository = $oEm->getRepository(Domain::class);
    /* @var $domainRepository \App\Module\Vitis\WebService\Repository\DomainRepository */

    if (!$domainRepository instanceof DomainRepository) {
      return $this->formatGenericError('', '', 'Cannot find Domain suffixes, bad repository', Response::HTTP_NOT_FOUND);
    }

    $aSuffixes = $domainRepository->getSuffixes();

    return $this->formatResponse([
      'data' => $aSuffixes,
      'total_row_number' => count($aSuffixes),
      'status' => 1,
    ]);
  }

  /**
   * Récupère les groupes d'un domaine
   * @param mixed $id id de domaine
   * @return Response réponse à retourner par l'API
   */
  public function getDomainGroups($id): Response {
    /* @var GenericRepository $domainRepository */
    $domainRepository = $this->oDatabaseConnection->getUserEntityManager()->getRepository(Domain::class);
    /* @var $oDomain Domain */
    $oDomain = $domainRepository->find($id);
    // Erreur si domaine n'existe pas
    if (!$oDomain instanceof Domain) {
      return $this->formatErrorDomainNotFound();
    }

    $aRequest = $this->requestToArray();
    $iLimit = $aRequest['limit'] ?? 20;
    $iOffset = $aRequest['offset'] ?? 0;
    $sFilter = $aRequest['filter'] ?? null;

    $aResponse = $oDomain->getDomainGroupsList($iLimit, $iOffset, $sFilter);

    if ($aResponse === null) {
      return $this->formatGenericResponseForErrorOnSelect('Cannot connect to LDAP/AD server');
    }

    return $this->formatResponse($aResponse);
  }

  /**
   * Récupère les utilisateur d'un domaine
   * @param mixed $id id de domaine
   * @return Response réponse à retourner par l'API
   */
  public function getDomainUsers($id): Response {
    /* @var GenericRepository $domainRepository */
    $domainRepository = $this->oDatabaseConnection->getUserEntityManager()->getRepository(Domain::class);
    /* @var $oDomain Domain */
    $oDomain = $domainRepository->find($id);
    // Erreur si domaine n'existe pas
    if (!$oDomain instanceof Domain) {
      return $this->formatErrorDomainNotFound();
    }
    $aRequest = $this->requestToArray();
    $iLimit = $aRequest['limit'] ?? 20;
    $iOffset = $aRequest['offset'] ?? 0;
    $sFilter = $aRequest['filter'] ?? null;

    $aResponse = $oDomain->getDomainUsersList($iLimit, $iOffset, $sFilter);

    if ($aResponse === null) {
      return $this->formatGenericResponseForErrorOnSelect('Cannot connect to LDAP/AD server');
    }

    return $this->formatResponse($aResponse);
  }

  /**
   * Importe des groups d'un domaine
   * @param mixed $id id du Domaine
   * @return Response réponse à retourner par l'API
   */
  public function importGroups($id): Response {
    // récupération des groupes à importer
    $aData = $this->requestToArray();
    // get auto_refresh
    $bAutoRefresh = $aData['auto_refresh'] ?? false;
    // Nb de groupes à importer
    $iLength = isset($aData['length']) ? intval($aData['length']) : 0;
    if ($iLength === 0) {
      return $this->formatErrorListNoSelection();
    }
    /* @var $domainRepository GenericRepository */
    $domainRepository = $this->oDatabaseConnection->getUserEntityManager()->getRepository(Domain::class);
    /* @var $oDomain Domain */
    $oDomain = $domainRepository->find($id);
    if (!$oDomain instanceof Domain) {
      return $this->formatErrorDomainNotFound();
    }
    $aDomainGroups = [];
    for ($i = 0; $i < $iLength; $i++) {
      if (isset($aData[$i])) {
        $oDomainGroup = $this->importDomainGroup($aData[$i], $oDomain, $bAutoRefresh);
        if (isset($oDomainGroup)) {
          $aDomainGroups[] = $oDomainGroup->toArray();
        }
      }
    }
    $aResponse = [
      'data' => $aDomainGroups,
      'list_count' => count($aDomainGroups),
      'total_row_number' => count($aDomainGroups) > $iLength ? count($aDomainGroups) : $iLength,
      'status' => 1,
    ];
    $this->formatResponse($aResponse);
    return $this->oResponse;
  }

  /**
   * Importe un groupe d'un domain
   * @param array $aDomainGroup
   * @param Domain $oDomain
   * @param bool $bAutoRefresh
   * @return DomainGroup|null
   */
  private function importDomainGroup(array $aDomainGroup, Domain $oDomain, bool $bAutoRefresh = false): ?DomainGroup{
    /* @var $GroupRepository GroupRepository */
    $oEm = $this->oDatabaseConnection->getUserEntityManager();
    $GroupRepository = $oEm->getRepository(Group::class);

    if (!$GroupRepository instanceof GenericRepository) {
      return null;
    }
    // get domain name
    $sName = $aDomainGroup['displayname'];
    // Check si Group exist
    $oGroup = $GroupRepository->findOneBy([
      'name' => $sName,
    ]);
    if (!isset($oGroup)) {
      // création du Group
      $oGroup = new Group();
      $oGroup->setName($sName);
      // sauvegarde dans la base de données
      $oEm->persist($oGroup);
      $oEm->flush();
      $oEm->refresh($oGroup);
    }
    $oDomainGroup = $oGroup->getDomainGroup();
    // récupération ou creation de groupe s'il n'existe pas
    if (!isset($oDomainGroup)) {
      // creation de DomainGroup
      $oDomainGroup = new DomainGroup();
    }
    // MAJ de 'dn'
    $oDomainGroup->setDomain($oDomain);
    $oDomainGroup->setDn($aDomainGroup['dn']);
    $oDomainGroup->setVitisGroup($oGroup);
    $oDomainGroup->setAutoRefresh($bAutoRefresh);
    $oEm->persist($oDomainGroup);
    $oEm->flush();
    $oEm->refresh($oDomainGroup);
    return $oDomainGroup;
  }

  /**
   * Synchronise des groupes de domaine avec un groupe Vitis
   * @return Response réponse à retourner par l'API
   */
  public function autoSyncGroups(): Response {
    // Récupération d'EntityManager
    $oEm = $this->oDatabaseConnection->getAdminEntityManager();
    /* @var $domainGroupRepository GenericRepository */
    $domainGroupRepository = $oEm->getRepository(DomainGroup::class);

    $aDomainGroups = $domainGroupRepository->findBy([
      'autoRefresh' => true,
    ]);
    $aUsers = [];
    foreach ($aDomainGroups as $oDomainGroup) {
      $aResult = $this->getDomainGroupsUsers($oDomainGroup);
      // S'il a eu une erreur retourne la
      if ($aResult === null) {
        return $this->formatGenericResponseForErrorOnSelect('Cannot connect to LDAP/AD server');
      }
      foreach ($aResult as $sKey => $aUser) {
        $aUsers[$sKey] = $aUser;
      }
    }

    return $this->formatResponse([
      'data' => $aUsers,
      'total_row_number' => count($aUsers),
      'status' => 1,
    ]);
  }

  /**
   * Synchronise un groupe de domaine avec un groupe Vitis
   * @param mixed $id id du Domaine
   * @return Response réponse à retourner par l'API
   */
  public function syncGroups($id): Response {
    // récupération des groupes à importer
    $aData = $this->requestToArray();
    // Nb de groupes à importer
    $iLength = isset($aData['length']) ? intval($aData['length']) : 0;
    if ($iLength === 0) {
      return $this->formatErrorListNoSelection();
    }
    /* @var $domainRepository GenericRepository */
    $domainRepository = $this->oDatabaseConnection->getAdminEntityManager()->getRepository(Domain::class);
    /* @var $oDomain Domain */
    $oDomain = $domainRepository->find($id);
    if (!$oDomain instanceof Domain) {
      return $this->formatErrorDomainNotFound();
    }
    $aUsers = [];
    for ($i = 0; $i < $iLength; $i++) {
      if (isset($aData[$i])) {
        $aResult = $this->getDomainGroupsUsers($oDomain->getDomainGroups()->toArray()[0]);
        // S'il a eu une erreur retourne la
        if ($aResult === null) {
          return $this->formatGenericResponseForErrorOnSelect('Cannot connect to LDAP/AD server');
        }
        foreach ($aResult as $sKey => $aUser) {
          $aUsers[$sKey] = $aUser;
        }
      }
    }
    return $this->formatResponse([
      'data' => $aUsers,
      'list_count' => count($aUsers),
      'total_row_number' => count($aUsers),
      'status' => 1,
    ]);
  }

  /**
   * Rétourne une liste des utilisateur d'un groupe de domain
   * @param DomainGroup $oDomainGroup
   * @return array|null
   */
  private function getDomainGroupsUsers(DomainGroup $oDomainGroup): ?array{
    $oDomain = $oDomainGroup->getDomain();
    $oGroup = $oDomainGroup->getVitisGroup();

    if (!$oDomain instanceof Domain || !$oGroup instanceof Group) {
      return null;
    }

    // Si le groupe n'appartiens pas au domaine
    if ($oDomain->getDomainId() != $oDomainGroup->getDomain()->getDomainId()) {
      $this->formatGenericResponseForErrorOnSelect('Group "' . $oGroup->getName() . '" not in domain "' . $oDomain->getName() . '"');
      return null;
    }

    $aUsers = $oDomain->getDomainGroupMembers($oDomainGroup->getDn());

    if ($aUsers === null) {
      return null;
    }

    foreach ($aUsers as $sKey => $aUser) {
      $oUser = $this->importUser($aUser, $oDomain, $oGroup);
      $aUsers[$sKey]['imported'] = !empty($oUser);
    }
    return $aUsers;

  }

  /**
   * Importe des utilisateurs d'un domaine
   * @param mixed $id id du Domaine
   * @return Response réponse à retourner par l'API
   */
  public function importUsers($id): Response {
    // récupération des utilisateurs à importer
    $aData = $this->requestToArray();
    // Nb d'utilisateurs à importer
    $iLength = isset($aData['length']) ? intval($aData['length']) : 0;
    if ($iLength === 0) {
      return $this->formatErrorListNoSelection();
    }
    /* @var $domainRepository GenericRepository */
    $oEm = $this->oDatabaseConnection->getAdminEntityManager();
    $domainRepository = $oEm->getRepository(Domain::class);
    /* @var $oDomain Domain */
    $oDomain = $domainRepository->find($id);
    if (!isset($oDomain)) {
      return $this->formatErrorDomainNotFound();
    }
    $aUsers = [];
    for ($i = 0; $i < $iLength; $i++) {
      if (isset($aData[$i])) {
        $oUser = $this->importUser($aData[$i], $oDomain);
        if (isset($oUser)) {
          $aUsers[] = $oUser->getUserId();
        }
      }
    }
    return $this->formatResponse([
      'data' => $aUsers,
      'list_count' => count($aUsers),
      'total_row_number' => $iLength,
      'status' => 1,
    ]);
  }

  /**
   * Importe un utilisateur d'un domain
   * @param array $aUser
   * @param Domain $oDomain
   * @param Group $oGroup
   * @return User|null
   */
  private function importUser(array $aUser, Domain $oDomain, Group $oGroup = null): ?User{
    // chargement de la connexion à la base de donnée
    $oEm = $this->oDatabaseConnection->getAdminEntityManager();
    $oQueryAccessor = $this->oDatabaseConnection->getQueryAccessor($oEm);
    // récupération du nom d'uid
    $sUID = $oDomain->getUid();
    // set login en minuscule
    $sLogin = strtolower($aUser[$sUID]);
    /* @var $userRepository UserRepository */
    $userRepository = $oEm->getRepository(User::class);

    if (!$userRepository instanceof UserRepository) {
      return null;
    }

    // Check si l'utilisateur BDD exist
    $oUser = $userRepository->findOneBy([
      'login' => $sLogin,
    ]);
    // S'il existe arrêt d'import
    if ($oUser instanceof User) {
      return null;
    } else {
      // création de l'utilisateur dans Postgres
      if (!$oQueryAccessor->createUser($sLogin)) {
        $this->oLoggerManager->get('API_ERROR')->info('Postgres role ' . $sLogin . ' already exists');
      }
      $oUser = new User();
      $oUser->setLogin($sLogin);
    }
    // Ajoute des roles au nouveau utilisateur
    $oQueryAccessor->grantRolesToRole(
      [
        $oDomain->getLoginRole(),
        'vitis_user',
      ],
      $sLogin
    );
    // Ajoute le domaine
    $oUser->setDomain($oDomain);
    // Récupère displayname d'utilisateur
    if ($oDomain->getDisplayName() != null && isset($aUser[$oDomain->getDisplayName()])) {
      $oUser->setName($aUser[$oDomain->getDisplayName()]);
    } else {
      $oUser->setName($aUser['displayname']);
    }
    // Récupère mail d'utilisateur
    if ($oDomain->getMail() != null && isset($aUser[$oDomain->getMail()])) {
      $oUser->setEmail($aUser[$oDomain->getMail()]);
    } else {
      $oUser->setEmail($aUser['mail'] ?? '');
    }
    $oEm->persist($oUser);
    $oEm->flush();
    $oEm->refresh($oUser);
    if (isset($oGroup)) {
      $qb = $oQueryAccessor->createQueryBuilder();
      $qb->insert('[SCHEMA].user_group')
        ->setValue('user_id', '[USER_ID]')
        ->setValue('group_id', '[GROUP_ID]');

      $aParams = [
        'SCHEMA' => ['value' => $this->oProperties->get('schema_vitis'), 'type' => 'schema_name'],
        'USER_ID' => ['value' => $oUser->getUserId(), 'type' => 'integer'],
        'GROUP_ID' => ['value' => $oGroup->getGroupId(), 'type' => 'integer'],
      ];
      try {
        $qb->executeWithParams($aParams);
      } catch (\Exception $e) {
        $this->oLoggerManager->get('SQL_ERROR')->error('Cannot associate group to user ' . $e->getMessage());
        return null;
      }
    }
    return $oUser;
  }

  /**
   * @return Response
   */
  protected function formatErrorDomainNotFound(): Response {
    return $this->formatGenericError('DOMAIN_NOT_FOUND', '', 'Domain not found', Response::HTTP_NOT_FOUND);
  }

  /**
   * @return Response
   */
  protected function formatErrorListNoSelection(): Response {
    return $this->formatGenericError('LIST_NO_SELECTION', '', 'Please select at least one record.', Response::HTTP_NOT_FOUND);
  }
}
