<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Controller\AbstractVitisController;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller qui permet de récupérer et d'intéragir avec les logs.
 *
 * @author Alexandre Mironenkov <alexandre.mironenkov@veremes.com>
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */
class LogsController extends AbstractVitisController {
  /**
   * @var string[]
   */
  private const EXTENSION_ALLOWED = ['log', 'zip', 'txt'];
  /**
   * @var string[]
   */
  private const FIELD_ALLOWED = ['filename', 'size_', 'lastModification_', 'lastModification'];

  /**
   * Récupère le phpinfos du serveur applicatif
   * @return Response réponse à retourner par l'API
   */
  public function getLogs(): Response {
    $oLocalFileInterface = $this->getLocalFileAccessor();
    $aFolderInfo = $oLocalFileInterface->getFolderInfos($this->oProperties->get('log_dir'), self::EXTENSION_ALLOWED, self::FIELD_ALLOWED, true);

    if ($aFolderInfo === null) {
      return $this->formatGenericError('ERROR_LOG_DIR_NOT_EXIST', '', 'This directory does not exist', Response::HTTP_NOT_FOUND);
    }

    $aFolderInfo = $this->folderInfosKeyModifier($aFolderInfo);
    $aReturn = [
      'status' => 1,
      'data' => isset($aFolderInfo['content']) ? $aFolderInfo['content'] : $aFolderInfo,
    ];
    return $this->formatResponse($aReturn);
  }

  /**
   * Modifie la strcuture de donnée
   * @param array $aFolderInfo Donnée à modifier
   * @param string $sRelativePath Permet de calculer la clé
   * @return array donnée modifiées
   */
  private function folderInfosKeyModifier(array $aFolderInfo, string $sRelativePath = ''): array{
    // utilisation de la clé en recursif
    $sRelativePath .= $aFolderInfo['filename'];

    if (isset($aFolderInfo['content'])) {
      // dossier
      $sRelativePath .= '/';
      foreach ($aFolderInfo['content'] as &$oItem) {
        $oItem = $this->folderInfosKeyModifier($oItem, $sRelativePath);
      }
    } else {
      // file
      $aFolderInfo['lastModification'] = $aFolderInfo['lastModification_'];
      $aFolderInfo['size'] = $aFolderInfo['size_'];
      $aFolderInfo['key'] = $sRelativePath;
    }

    unset($aFolderInfo['size_']);
    unset($aFolderInfo['lastModification_']);

    return $aFolderInfo;
  }

  /**
   * Récupère le chemin complet vers le fichier de log
   * @return string chemin vers le fichier de log ou null
   */
  private function getFileAccessPath(): ?string{
    // Récuperation des données d'une requete
    $aData = $this->requestToArray(true);

    // Si la requete ne contien pas "key"
    if (!isset($aData['key'])) {
      $this->formatGenericError('ERROR_MISSING_ATTRIBUTE_KEY', '', 'Missing attribute "key"', Response::HTTP_BAD_REQUEST);
      return null;
    }

    // Construit le chemin vers le fichier de log
    $sPath = $this->oProperties->get('var_dir') . '/' . $aData['key'];

    // Récupère l'interface fichier
    $oLocalFileInterface = $this->getLocalFileAccessor();

    // Construit le pattern pour la verification regex
    $pattern = '#^' . $this->oProperties->get('log_dir') . '/(?!.*\\.{2}).+$#i';

    // Si le chemin ne correspond pas au pattern
    // (ex. : "#^/var/www/vitis/vas/var/log/(?!.*\\.{2}).+$#i")
    // ou le fichier n'existe pas envoi 404
    if (preg_match($pattern, $sPath) !== 1 ||
      !$oLocalFileInterface->file_exists($sPath)) {
      $this->formatGenericErrorFileNotFound();
      return null;
    }
    return $sPath;
  }

  /**
   * Récupère le contenu d'un fichier log
   * @return Response réponse à retourner par l'API
   */
  public function getLogsContent(): Response {
    // Récupere le chemin complet vers le fichier de log
    $sPath = $this->getFileAccessPath();
    if ($sPath === null) {
      return $this->oResponse;
    }

    $oLocalFileInterface = $this->getLocalFileAccessor();
    $sContent = $oLocalFileInterface->file_get_contents($sPath);

    if (!mb_detect_encoding($sContent, 'UTF-8', true)) {
      // si la chaine n'est pas en utf-8 on encode et on remplace
      $sContent = utf8_encode($sContent);
    }

    // Construit la réponse
    $aReturn = ['status' => 1];
    $aReturn['data'] = $sContent;
    return $this->formatResponse($aReturn);
  }

  /**
   * Récupère le contenu d'un fichier log en format blob
   * @return Response réponse à retourner par l'API
   */
  public function getLogsDownload(): Response {
    // Récupere le chemin complet vers le fichier de log
    $sPath = $this->getFileAccessPath();
    if ($sPath === null) {
      return $this->oResponse;
    }
    // Création du blob du fichier à télécharger
    $aHeaders = [
      'Access-Control-Allow-Methods' => implode(', ', self::HTTP_METHODS_ALLOWED),
      'Access-Control-Allow-Headers' => implode(', ', self::HTTP_RESPONSE_ALLOWED_HEADERS),
      'Content-Type-Set-By' => 'vitis/logscontroller',
      'Content-Type' => 'application/octet-stream; charset=UTF-8',
      'Content-Transfer-Encoding' => 'Binary',
    ];

    return $this->getBlobResponse($sPath, null, $aHeaders);
  }

  /**
   * Supprime un fichier de log
   * @return Response réponse à retourner par l'API
   */
  public function deleteLogs(): Response {
    // Récupere le chemin complet vers le fichier de log
    $sPath = $this->getFileAccessPath();
    if ($sPath === null) {
      return $this->oResponse;
    }

    $aReturn = ['status' => 1];

    $oLocalFileInterface = $this->getLocalFileAccessor();
    // Construit la réponse
    if ($oLocalFileInterface->unlink($sPath) === false) {
      return $this->formatGenericError('ERROR_DELETING_LOG_FILE', '', 'Error while deleting a log file', Response::HTTP_INTERNAL_SERVER_ERROR);
    }

    // Supprime les dossiers parents si ils sont vide
    $this->deleteFoldersIfEmpty($sPath);

    $aReturn['status'] = 1;
    $this->formatResponse($aReturn);

    return $this->oResponse;
  }

  /**
   * Modifie la strcuture de donnée
   * @param array $aFolderInfo Donnée à modifier
   * @param string $sRelativePath Permet de calculer la clé
   * @return array|null donnée modifiées
   */
  private function removeLogsHistory(array $aFolderInfo, string $sRelativePath = ''): ?array{
    // Récuperation des données d'une requete
    $aData = $this->requestToArray(true);

    // utilisation de la clé en recursif
    $sRelativePath .= $aFolderInfo['filename'];

    if (isset($aFolderInfo['content'])) {
      // dossier
      $sRelativePath .= '/';
      foreach ($aFolderInfo['content'] as &$oItem) {
        $oItem = $this->removeLogsHistory($oItem, $sRelativePath);
      }
      return $aFolderInfo;
    }
    // file
    $oDate = new \DateTime();
    $iLogLifetime = $aData['log_lifetime'];
    $oLastModification = \DateTime::createFromFormat('d/m/Y H:i:s', $aFolderInfo['lastModification']);
    // Si le fichier de log n'est pas ancien
    if (intval($oDate->diff($oLastModification)->format('%a')) < $iLogLifetime) {
      return $aFolderInfo;
    }
    // Construit le chemin vers le fichier de log
    $sPath = $this->oProperties->get('var_dir') . '/' . $sRelativePath;

    // Récupère l'interface fichier
    $oLocalFileInterface = $this->getLocalFileAccessor();

    // Construit le pattern pour la verification regex
    $pattern = '#^' . $this->oProperties->get('log_dir') . '/(?!.*\\.{2}).+$#i';

    // Si le chemin ne correspond pas au pattern
    // (ex. : "#^/var/www/vitis/vas/var/log/(?!.*\\.{2}).+$#i")
    // ou le fichier n'existe pas envoi 404
    if (preg_match($pattern, $sPath) !== 1
      || !$oLocalFileInterface->file_exists($sPath)) {
      return null;
    }
    // Supprime le fichier de log
    $oLocalFileInterface->unlink($sPath);

    // Supprime les dossiers parents si ils sont vide
    $this->deleteFoldersIfEmpty($sPath);

    return $aFolderInfo;
  }

  /**
   * Purge les anciens logs
   * @return Response réponse à retourner par l'API
   */
  public function purgeLogs(): Response {
    $aReturn = ['status' => 0];

    $oLocalFileInterface = $this->getLocalFileAccessor();
    $aFolderInfo = $oLocalFileInterface->getFolderInfos($this->oProperties->get('log_dir'), self::EXTENSION_ALLOWED, self::FIELD_ALLOWED);

    if ($aFolderInfo === null) {
      return $this->formatGenericError('ERROR_LOG_DIR_NOT_EXIST', '', 'This directory does not exist', Response::HTTP_NOT_FOUND);
    }

    $aFolderInfo = $this->removeLogsHistory($aFolderInfo);
    $aReturn['data'] = isset($aFolderInfo['content']) ? $aFolderInfo['content'] : $aFolderInfo;
    $aReturn['status'] = 1;
    $this->formatResponse($aReturn);

    return $this->oResponse;
  }

  /**
   * Fonction récursive qui vérifie si le dossier parent du chemin est vide, et si c'est le cas le supprime.
   * La limite est le dossier de log
   *
   * @param string $sPath
   * @return void
   */
  private function deleteFoldersIfEmpty(string $sPath): void {
    $sDirectory = dirname($sPath);

    // Condition d'arrêt
    if ($sDirectory === $this->oProperties->get('log_dir')) {
      return;
    }

    $oLocalFileInterface = $this->getLocalFileAccessor();
    // Si le dossier est vide
    if (count($oLocalFileInterface->scandir($sDirectory)) <= 2) {
      // Supprime le dossier vide
      $oLocalFileInterface->rmdir($sDirectory);
      // Relance la fonction
      $this->deleteFoldersIfEmpty($sDirectory);
    }

  }
}
