<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Controller\TreeController;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\VitisRequest;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller qui permet de gérer le dossier partagé.
 *
 * @author Ludivine Premont <ludivine.premont@veremes.com>
 */
class SharedDirController extends TreeController {
  /**
   * Initialise le controler et instancie les services requis
   * @param Properties $oPropertiesSrvc Service d'accès aux properties
   * @param VitisRequest $oVitisRequest Service d'abstraction de la requête Symfony
   * @param Route $oRouteConfig Service d'accès à la configuration de la route
   * @param LoggerManager $oLoggerManager Service d'utilsation des logs
   * @param DatabaseConnection $oDatabaseConnection Service d'accès à la base de données
   * @param FileManager $oFileManager Service de gestion des fichiers
   */
  public function __construct(Properties $oPropertiesSrvc, VitisRequest $oVitisRequest, Route $oRouteConfig, LoggerManager $oLoggerManager, DatabaseConnection $oDatabaseConnection, FileManager $oFileManager) {
    parent::__construct($oPropertiesSrvc, $oVitisRequest, $oRouteConfig, $oLoggerManager, $oDatabaseConnection, $oFileManager);
    // création du shared_dir s'il n'existe pas
    $oFileAccessor = $oFileManager->getFileAccessor();
    if (!$oFileAccessor->file_exists($this->oProperties->get('shared_dir'))) {
      $oFileAccessor->mkdir($this->oProperties->get('shared_dir'), 0777, true);
    }
  }

  /**
   * get shared dir directory tree
   * @return Response
   */
  public function getSharedDirTree(): Response {
    $aKeepOnlyExt = $this->getKeepOnlyExt();
    $aData = $this->requestToArray();

    $iDepth = -1; // Par défaut toute la profondeur de l'arborescence
    if (!empty($aData['depth'])) {
      $iDepth = $aData['depth'];
    }

    // Chemin du dossier partagé
    $shared_dir = $this->oProperties->get('shared_dir');

    // Chemin par rapport au répertoire racine.
    $sDir = $shared_dir;
    if (!empty($aData['path'])) {
      if (strpos($aData['path'], $shared_dir) == 0) {
        // retrait du shared dir
        $aData['path'] = str_replace($shared_dir, '', $aData['path']);
      }
      
      if (!empty($aData['path'])) {
        // nettoyage du path en cas d'attaque
        $aData['path'] = str_replace('./', '/', $aData['path']);
        $aData['path'] = str_replace('../', '/', $aData['path']);
        // Chemin relatif au répertoire racine
        $sDir .= '/' . $aData['path'];
      }
    }

    // Vérifie qu'il s'agit bien du chemin de dossier partagé en index = 0
    $sDir = $this->pathChecker($sDir, $shared_dir);

    // le dossier n'est pas dans shared ou n'existe pas
    if (empty($sDir)) {
      return $this->formatGenericError('TREE_VIEW_ERROR_LOADING_FILES', '', 'Chemin non acceptable', Response::HTTP_NOT_FOUND);
    }

    $aTree = $this->oFileManager->getAccessorForPath($sDir)->getFolderInfos($sDir, $aKeepOnlyExt, [], false, (int) $iDepth);
    $aReturn = [];
    // le dossier n'a pas pu être scanner
    if ($aTree === null) {
      return $this->formatGenericError('TREE_VIEW_ERROR_LOADING_FILES', '', 'Erreur lors du scan', Response::HTTP_NOT_FOUND);
    }
    $aReturn = [
      'status' => 1,
      'data' => $aTree,
    ];

    // Si le système de fichier n'est pas local on redéfini le dossier parent (S3 n'a pas de dossier)
    if ($this->oProperties->get('file_system') !== Properties::LOCAL_FS) {
      if (!isset($aTree["content"])) {
        $aTree["content"] = [];
        $aTree["lastModification"] = "none";
        $aTree["size"] = "";
        unset($aTree["href"]);
        $aReturn['data'] = $aTree;
      }
    }

    return $this->formatResponse($aReturn);
  }

  /**
   * Supprime un fichier ou un dossier du répertoire partagé
   * @return Response
   */
  public function removeTreeDirectoryOrFile(): Response {
    return parent::deleteTreeDirectoryOrFile(null);
  }

  /**
   * Création d'un répertoire dans le répertoire partagé.
   * @return Response
   */
  public function makeTreeFolder(): Response {
    return parent::createTreeFolder(null);
  }
}
