<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Controller\AbstractVitisController;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller qui permet de récupérer et d'intéragir avec les formulaires studio.
 *
 * La route devra contenir id et form
 * id : identifiant de l'enregistrement
 * form : formulaire à retourner (default, published, custom)
 *
 * Exemple :
 * "route": "/v2/gtf/userworkspaces/{id}/form/{form}",
 *
 * Dans les options il faudra utiliser l'attribut files et y placer un tableau d'entités
 * Chaque entité est définie par name, form_file, path
 * name : nom de l'attribut en sortie
 * form_file : définit le nom de fichier pour chaque type de formulaire
 * path : chamin vers le fichier, utilisant @id et @file_name
 *
 * Exemple :
 * "files": [{
 *   "name": "json_form",
 *   "form_file": {
 *     "default": "DSubform.json",
 *     "published": "Subform.json",
 *     "custom": "WSubform.json"
 *   },
 *   "path": "@prop(ws_data_dir)/gtf/workspace/@id/form/@file_name"
 * }, {
 *   "name": "js_form",
 *   "form_file": {
 *     "default": "DSubform.js",
 *     "published": "Subform.js",
 *     "custom": "WSubform.js"
 *   },
 *   "path": "@prop(ws_data_dir)/gtf/workspace/@id/form/ressources/@file_name"
 * }, {
 *   "name": "css_form",
 *   "form_file": {
 *     "default": "DSubform.css",
 *     "published": "Subform.css",
 *     "custom": "WSubform.css"
 *   },
 *   "path": "@prop(ws_data_dir)/gtf/workspace/@id/form/ressources/@file_name"
 * }]
 */
class StudioFormController extends AbstractVitisController {
  /**
   * Clé utilisée pour l'id dans les chemins
   * @var string
   */
  private const ID_PATH_KEY = '@id';

  /**
   * Clé utilisée pour nom du fichier dans les chemins
   * @var string
   */
  private const FILE_NAME_PATH_KEY = '@file_name';

  /**
   * Récupère le contenu du formulaire
   * @param mixed $id
   * @param mixed $form
   * @return Response réponse à retourner par l'API
   */
  public function getForm($id, $form): Response {
    $aRequestOptions = $this->getRequestParameters();
    $aReturn = [];

    if (!empty($id) && !empty($form) && is_array($aRequestOptions['options'])) {
      if (array_key_exists('files', $aRequestOptions['options'])) {
        $aReturn['data'] = [];
        foreach ($aRequestOptions['options']['files'] as $i => $oFile) {
          $sFileContent = $this->getFileContent($oFile, $id, $form);
          if ($sFileContent !== null) {
            $aReturn['data'][$oFile['name']] = $sFileContent;
          }
        }
      } else if (array_key_exists('fields', $aRequestOptions['options'])) {
        $aReturn['data'] = [];
        foreach ($aRequestOptions['options']['fields'] as $i => $oField) {
          $sFieldContent = $this->getFieldContent($aRequestOptions['options'], $oField, $id, $form);
          if ($sFieldContent !== null) {
            $aReturn['data'][$oField['name']] = $sFieldContent;
          }
        }
      }
    }

    $aReturn['status'] = 1;
    return $this->formatResponse($aReturn);
  }

  /**
   * Retourne le contenu d'un fichier
   * @param mixed $oFile
   * @param mixed $id
   * @param mixed $form
   * @return string|null
   */
  private function getFileContent($oFile, $id, $form): ?string {
    if (!empty($oFile['name']) && !empty($oFile['form_file'][$form]) && !empty($oFile['path'])) {
      $sPath = $this->getFilePath($oFile, $id, $form);

      // Récupération du contenu
      $oFileInterface = $this->getFileAccessor($sPath);
      $sContent = $oFileInterface->file_get_contents($sPath);

      return $sContent;
    }
    return null;
  }

  /**
   * Retourne le contenu d'un champ de la base
   * @param mixed $oOptions
   * @param mixed $oField
   * @param mixed $id
   * @param mixed $form
   * @return string|null
   */
  private function getFieldContent($oOptions, $oField, $id, $form): ?string {
    if (!empty($oField['name']) && !empty($oField['form_field'][$form])) {
      // Requête sql
      $qb = $this->oDatabaseConnection->getQueryAccessor()->createQueryBuilder();

      try {
        $qb->select('a.[FIELD]')
          ->from('[SCHEMA].[TABLE]', 'a')
          ->where($qb->expr()->eq('a.[FIELD_ID]', '[ID_VALUE]'));

        $aParams = [
          'FIELD' => ['value' => $oField['form_field'][$form], 'type' => 'column_name'],
          'SCHEMA' => ['value' => $this->oProperties->replacePropertiesInString($oOptions['schema']), 'type' => 'schema_name'],
          'TABLE' => ['value' => $oOptions['table'], 'type' => 'table_name'],
          'FIELD_ID' => ['value' => $oOptions['field_id'], 'type' => 'column_name'],
          'ID_VALUE' => ['value' => $id, 'type' => 'mixed'],
        ];

        $oContent = $qb->executeWithParams($aParams)->fetchAllAssociative();
      } catch (\Exception$e) {
        $this->oLoggerManager->get('SQL_ERROR')->error('Studio : can\'t get form content: ' . $e->getMessage());
        return null;
      }

      return $oContent[0][$oField['form_field'][$form]] ?? '';
    }
    return null;
  }

  /**
   * Met à jour les fichiers d'un formulaire
   * @param mixed $id
   * @param mixed $form
   * @return Response
   */
  public function updateFiles($id, $form): Response {
    $aRequestOptions = $this->getRequestParameters();
    $aData = $this->requestToArray();

    $aReturn = [];

    if (!empty($id) && !empty($form) && is_array($aRequestOptions['options'])) {
      if (array_key_exists('files', $aRequestOptions['options'])) {
        $aReturn['data'] = [];
        foreach ($aRequestOptions['options']['files'] as $i => $oFile) {
          $sFilePath = $this->getFilePath($oFile, $id, $form);
          $oFileInterface = $this->getFileAccessor($sFilePath);
          $aPathInfo = pathinfo($sFilePath);
          if (!empty($sFilePath) && isset($aData[$aPathInfo['extension']])) {
            if ($oFileInterface->file_put_contents($sFilePath, $aData[$aPathInfo['extension']]) === null) {
              return $this->formatErrorCannotUpdateForm($sFilePath . ' can\'t be writed, verify the folder\'s rights');
            }
            $sFileContent = $this->getFileContent($oFile, $id, $form);
            if ($sFileContent !== null) {
              $aReturn['data'][$oFile['name']] = $sFileContent;
            }
          }
        }
      } else if (array_key_exists('fields', $aRequestOptions['options'])) {
        if ($form === 'init') {
          // Si c'est une demande init on sauvegarde le formulaire dans les différents types (Default, Published, Perso)
          $result = $this->saveInitialForms_($id);
        } else {
          $result = $this->updateFields_($id, $form);
        }
        if ($result == null) {
          $aReturn['message'] = 'Error during update form content';
          $aReturn['errorMessage'] = 'ERROR_GENERIC_CONTROLLER_UPDATE';
          return $this->formatResponse($aReturn, Response::HTTP_UNAUTHORIZED);
        }
        return $this->getForm($id, $result);
      }
    }

    $aReturn['status'] = 1;
    $this->formatResponse($aReturn);
    return $this->oResponse;
  }

  /**
   * Met à jour les différents type de formulaire (Default, Published, Perso) si ils sont vide
   * @param mixed $id
   * @return string|null
   */
  private function saveInitialForms_($id): ?string{
    $aRequestOptions = $this->getRequestParameters();

    $qb = $this->oDatabaseConnection->getQueryAccessor()->createQueryBuilder();

    if (isset($aRequestOptions['options']['fields']['json'])) {
      $sLastForm = '';
      foreach ($aRequestOptions['options']['fields']['json']['form_field'] as $sForm => $sField) {
        try {
          $qb->select('a.[FIELD]')
            ->from('[SCHEMA].[TABLE]', 'a')
            ->where($qb->expr()->eq('a.[FIELD_ID]', '[ID_VALUE]'));

          $aParams = [
            'FIELD' => ['value' => $sField, 'type' => 'column_name'],
            'SCHEMA' => ['value' => $this->oProperties->replacePropertiesInString($aRequestOptions['options']['schema']), 'type' => 'schema_name'],
            'TABLE' => ['value' => $aRequestOptions['options']['table'], 'type' => 'table_name'],
            'FIELD_ID' => ['value' => $aRequestOptions['options']['field_id'], 'type' => 'column_name'],
            'ID_VALUE' => ['value' => $id, 'type' => 'mixed'],
          ];

          $oContent = $qb->executeWithParams($aParams)->fetchAll();
        } catch (\Exception$e) {
          $this->oLoggerManager->get('SQL_ERROR')->error('Studio check empty: can\'t get form content: ' . $e->getMessage());
          return null;
        }

        // Si le chamsp Json est vide on enregistre le formulaire
        if ($oContent[0][$sField] == '') {
          $result = $this->updateFields_($id, $sForm);
          if ($result == null) {
            return null;
          }
        }
        $sLastForm = $sForm;
      }
      return $sLastForm;
    }

    return null;
  }

  /**
   * Met à jour les champs d'un formulaire
   * @param mixed $id
   * @param string $form
   * @return string|null
   */
  private function updateFields_($id, string $form): ?string{
    $aRequestOptions = $this->getRequestParameters();
    $aData = $this->requestToArray();

    // Requête sql
    $qb = $this->oDatabaseConnection->getQueryAccessor()->createQueryBuilder();

    $qb->update('[SCHEMA].[TABLE]', 'a');

    $aParams = [
      'SCHEMA' => ['value' => $this->oProperties->replacePropertiesInString($aRequestOptions['options']['schema']), 'type' => 'schema_name'],
      'TABLE' => ['value' => $aRequestOptions['options']['table'], 'type' => 'table_name'],
      'FIELD_ID' => ['value' => $aRequestOptions['options']['field_id'], 'type' => 'column_name'],
      'ID_VALUE' => ['value' => $id, 'type' => 'mixed'],
    ];
    foreach ($aRequestOptions['options']['fields'] as $key => $oField) {

      error_log('$oField :' . print_r($oField, true));

      $name = $oField['form_field'][$form];
      if (isset($aData[$key])) {
        $sFieldParamsName = 'FIELD_' . $name;
        $sFieldParamsValue = 'VALUE_' . $name;
        $aParams[$sFieldParamsName] = ['value' => $name, 'type' => 'column_name'];
        $aParams[$sFieldParamsValue] = ['value' => $aData[$key], 'type' => 'mixed'];
        $qb->set('[' . $sFieldParamsName . ']', '[' . $sFieldParamsValue . ']');
      }
    }
    $qb->where($qb->expr()->eq('a.[FIELD_ID]', '[ID_VALUE]'));

    try {
      $qb->executeWithParams($aParams);
    } catch (\Exception$e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('Studio : can\'t put form content: ' . $e->getMessage());
      return null;
    }

    return $form;
  }

  /**
   * Retourne chemin d'un fichier
   * @param mixed $oFile
   * @param mixed $id
   * @param mixed $form
   * @return string|null
   */
  private function getFilePath($oFile, $id, $form): ?string {
    if (!empty($oFile['name']) && !empty($oFile['form_file'][$form]) && !empty($oFile['path'])) {
      $sPath = $oFile['path'];

      // Remplacement des properties
      $sPath = $this->oProperties->replacePropertiesInString($sPath);

      // Remplacement de l'id
      $sPath = str_replace(self::ID_PATH_KEY, $id, $sPath);

      // Remplacement du nom du fichier
      $sPath = str_replace(self::FILE_NAME_PATH_KEY, $oFile['form_file'][$form], $sPath);

      return $sPath;
    }
    return null;
  }

/**
 * A creuser
 *
 * @param  string   $sMessage
 * @return Response
 */
  private function formatErrorCannotUpdateForm(string $sMessage = ''): Response {
    $aReturn = [
      'status' => 0, //ajout car n'y étais pas de base
      'success' => 0,
      'message' => $sMessage,
      'errorMessage' => '',
    ];
    return $this->formatResponse($aReturn, Response::HTTP_INTERNAL_SERVER_ERROR);
  }
}
