<?php

namespace App\Module\Vitis\WebService\Controller;

use App\Module\Vitis\WebService\Controller\AbstractVitisController;
use App\Module\Vitis\WebService\Entity\User;
use App\Module\Vitis\WebService\Repository\UserRepository;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\Token\JweTokenManager;
use App\Module\Vitis\WebService\Service\VitisHttpClient;
use App\Module\Vitis\WebService\Service\VitisRequest;
use Jose\Component\Signature\Serializer\CompactSerializer;
use Jose\Component\Signature\Serializer\JWSSerializerManager;
use Ramsey\Uuid\Uuid;
use Symfony\Component\HttpFoundation\Response;

/**
 * Controller permettant de gérer les token de connexion à l'application.
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 * @author Armand Bahi <armand.bahi@veremes.com>
 */
class TokenController extends AbstractVitisController {
  /**
   * @var JweTokenManager
   */
  protected $oJweTokenManager;

  /**
   * Initialise le controler et instancie les services requis
   * @param Properties $oPropertiesSrvc Service d'accès aux properties
   * @param VitisRequest $oVitisRequest Service d'abstraction de la requête Symfony
   * @param Route $oRouteConfig Service d'accès à la configuration de la route
   * @param LoggerManager $oLoggerManager Service d'utilsation des logs
   * @param DatabaseConnection $oDatabaseConnection Service d'accès à la base de données
   * @param FileManager $oFileManager Service de gestion des fichiers
   * @param JweTokenManager $oJweTokenManager Service de manipulation des token JWE
   */
  public function __construct(Properties $oPropertiesSrvc, VitisRequest $oVitisRequest, Route $oRouteConfig, LoggerManager $oLoggerManager, DatabaseConnection $oDatabaseConnection, FileManager $oFileManager, JweTokenManager $oJweTokenManager) {
    parent::__construct($oPropertiesSrvc, $oVitisRequest, $oRouteConfig, $oLoggerManager, $oDatabaseConnection, $oFileManager);
    $this->oJweTokenManager = $oJweTokenManager;
  }

  /**
   * Génère un token, utiliser par la route vitis_create_token.
   * exemple de corp de la requête :
   *      {
   *          "login" : "string",
   *          "password" : "string"
   *      }
   * @return Response réponse à retourner par l'API
   */
  public function generateToken(): Response {
    // récupération login password
    $oData = $this->requestToArray();

    return $this->generateTokenWithLoginPassword($oData['user'], $oData['password']);
  }

  /**
   * génére une réponse type token à partir d'un login/password
   *
   * @param  string   $sUserName
   * @param  string   $sPassword
   * @return Response
   */
  private function generateTokenWithLoginPassword(string $sUserName, string $sPassword): Response {
    try {
      // preset de la strcuture de réponse
      $aReturn = ['status' => 0];

      // chargement de la connexion à la base de donnée
      $oEntityManager = $this->oDatabaseConnection->getAdminEntityManager();

      // vérification de l'existence de l'utilisateur
      $oUserRepository = $oEntityManager->getRepository(User::class);

      $aUsers = [];
      if ($oUserRepository instanceof UserRepository) {
        $aUsers = $oUserRepository->findByLoginWithoutCase($sUserName);
      }

      if (count($aUsers) != 1) {
        $this->oLoggerManager->get('API_ERROR')->error('Error with token generation for ' . $sUserName . ', this user doesn\'t exist');
        // L'utilisateur n'a pas pu être trouvé
        throw new \Exception('Cannot find user');
      }
      // retransformation en objet
      $aUser = $oUserRepository->findOneBy(['login' => $aUsers[0]['login']]);
      // controle de la restriction d'IP
      if (!$this->checkIpConstraint($aUser->getIpConstraint())) {
        $sClientIp = $this->oVitisRequest->getClientIp();
        $this->oLoggerManager->get('API_ERROR')->error('Error with token generation for ' . $aUser->getLogin() . ', this user try to connect from ' . $sClientIp, [$aUser]);
        // L'utilisateur ne peut pas se connecter avec cette ip
        $sMessage = 'Connection forbidden from ip :' . $sClientIp;
        return $this->formatGenericError('ERROR_USER_IP_UNAUTHORIZED', '', $sMessage, Response::HTTP_UNAUTHORIZED);
      }

      // Initialisation de la structure de l'utilisateur
      $aPayload = $aUser->getTokenPayload();

      // récupère le mot de passe pour se conncter à postgres
      // à la connexion il n'est jamais retouché
      if (empty($sPassword)) {
        $this->oLoggerManager->get('API_ERROR')->error('Password Hash failed for ' . $aUser->getLogin());
        // Le mot de passe n'a pas pu être hash
        return $this->formatGenericError('BAD_HASH_METHOD_FOR_DB_PASSWORD', '',
          'Your password cannot be hashed, contact application\'s administrator', Response::HTTP_NOT_IMPLEMENTED);
      } else {
        // Création du token
        $aTokenStruct = $this->getJWEToken($this->getJWETokenPayload([
          'user' => $aPayload,
          'password' => $sPassword,
        ]));

        // Test de connexion avec le token
        $aUserTest = $this->getCurrentUserEntity($aTokenStruct['token']);
        if (empty($aUserTest)) {
          // L'utilisateur n'a pas pu être trouvé
          return $this->formatGenericError('ERROR_INCORRECT_LOGIN_PASSWORD', '', '', Response::HTTP_UNAUTHORIZED);
        }

        // Insertion des information du token dans les données de l'utilisateur
        $aPayload = array_merge($aPayload, $aTokenStruct);
        if (empty($aPayload['token'])) {
          // Le token n'a pas pu être généré
          $this->oLoggerManager->get('API_ERROR')->error('Token generation failed for ' . $aUser->getLogin(), $aPayload);
          return $this->formatGenericError('ERROR_EMPTY_TOKEN_PAYLOAD', '', 'Problem with Token Payload', Response::HTTP_SERVICE_UNAVAILABLE);

        } else {
          $this->oLoggerManager->get('CONNEXION_INFO')->info('Token generation for user ' . $aUser->getLogin(), $aPayload);
          // mise à jour de la date de dernière connexion
          $aUser->updateLastConnection();
          $oEntityManager->persist($aUser);
          $oEntityManager->flush();
          // Ajout du status et on retourne la requête
          $aReturn['data'] = $aPayload;
          $aReturn['status'] = 1;
          return $this->formatResponse($aReturn);
        }
      }

    } catch (\Exception $e) {
      return $this->formatGenericError('ERROR_INCORRECT_LOGIN_PASSWORD', '', 'Error occured : ' . $e->getMessage(), Response::HTTP_UNAUTHORIZED);
    }
  }

  /**
   * Génération de connexion via oauth
   *
   * @return Response
   */
  public function generateOAuthToken(): Response {
    // récupération login password
    $aData = $this->requestToArray();

    if (!array_key_exists('code', $aData) || empty($aData['code'])) {
      return $this->formatErrorMissingParameter('code parameter is missing');
    }
    if (!array_key_exists('state', $aData) || empty($aData['state'])) {
      return $this->formatErrorMissingParameter('state parameter is missing');
    }

    $oTokenData = $this->getOAuth2tokenFromProvider($aData['code']);

    if (empty($oTokenData)) {
      return $this->formatGenericError('EMPTY_OAUTH_TOKEN', '', 'Token oauth is empty see with provider');
    }

    if ($oTokenData['session_state'] !== $aData['state']) {
      return $this->formatGenericError('BAD_STATE_OAUTH_TOKEN', '', 'session identifier doesn\'t match');
    }

    $sToken = $oTokenData['access_token'];

    $oTokenInnerData = $this->validateAndExtractDataFromOAuth2Token($sToken);

    // récupération du login
    $sUsername = $oTokenInnerData['preferred_username'];

    return $this->generateTokenWithLoginPassword($sUsername, $sToken);

  }

  /**
   * Retourne les infos du token (Utilisateur, date de fin etc...)
   * @return Response réponse à retourner par l'API
   */
  public function getToken(): Response {
    $aReturn = ['status' => 0];

    // Token
    $sToken = $this->oVitisRequest->getToken();

    if (empty($sToken)) {
      return $this->formatGenericError('ERROR_MISSING_AUTHORIZATION_HEADER', '', 'Authorization header is missing', Response::HTTP_UNAUTHORIZED);
    }

    $tokenUpdated = false;
    if (Uuid::isValid($sToken)) {
      // Si la chaine token est un uuid alors on récupère le token
      $oEm = $this->oDatabaseConnection->getAdminEntityManager();
      $qb = $this->oDatabaseConnection->getQueryAccessor($oEm)->createQueryBuilder();
      $qb->select('a.token')
        ->from('[SCHEMA].login_token', 'a')
        ->where($qb->expr()->eq('a.token_id', '[TOKEN]'));

      $aParams = [
        'SCHEMA' => ['value' => $this->oProperties->get('schema_vitis'), 'type' => 'schema_name'],
        'TOKEN' => ['value' => $sToken, 'type' => 'string'],
      ];

      $oToken = $qb->executeWithParams($aParams)->fetchAllAssociative();

      $tokenUpdated = true;
      $sToken = $oToken[0]['token'];

      $aTokenJWEDecrypt = $this->oJweTokenManager->decryptToken($sToken);
      if (empty($aTokenJWEDecrypt) ||
        (!empty($aTokenJWEDecrypt['data']['ip_restriction']) &&
          !$this->checkIpConstraint($aTokenJWEDecrypt['data']['ip_restriction']))) {
        $sClientIp = $this->oVitisRequest->getClientIp();
        // L'utilisateur ne peut pas se connecter avec cette ip
        $sMessage = 'Connection forbidden from ip :' . $sClientIp;
        return $this->formatGenericError('ERROR_USER_IP_UNAUTHORIZED', '', $sMessage, Response::HTTP_UNAUTHORIZED);
      }
    }

    if (empty($sToken)) {
      return $this->formatGenericError('ERROR_MISSING_AUTHORIZATION_HEADER', '', 'Authorization header is missing', Response::HTTP_UNAUTHORIZED);
    }

    // Token décrypté
    $aTokenJWEDecrypt = $this->oJweTokenManager->decryptToken($sToken);

    // Roles si l'utilisateur est connecté
    $oUser = $this->getCurrentUserEntity($sToken);
    if ($oUser !== null) {
      $aReturn['data'] = [
        'privileges' => $oUser->getRoles(),
        'user' => $oUser->getLogin(),
        'user_id' => $oUser->getUserId(),
        'validity_date' => $aTokenJWEDecrypt['exp'],
      ];

      if ($tokenUpdated) {
        $aReturn['data']['token'] = $sToken;
      }

      $aReturn['status'] = 1;
      // log en info
      $this->oLoggerManager->get('CONNEXION_INFO')->info('Get Token for user ' . $aReturn['data']['user'], $aReturn);
      // chargement de la connexion à la base de donnée admin
      // pour passer en update toute les opérations doivent être faite avec le même EntityManager
      // or getcurrentUserEntity utilise le UserEntityManager
      // d'où la double requête
      $oEntityManager = $this->oDatabaseConnection->getAdminEntityManager();
      // vérification de l'existence de l'utilisateur
      $aUser = $oEntityManager->getRepository(User::class)->findOneBy(['login' => $oUser->getLogin()]);
      // mise à jour de la date de dernière connexion
      $aUser = $aUser->updateLastConnection();
      $oEntityManager->persist($aUser);
      $oEntityManager->flush();
      // Retour
      $this->formatResponse($aReturn);
      return $this->oResponse;
    } else {
      $this->oLoggerManager->get('API_ERROR')->error('Get Token user not found : ' . $sToken);
      // session expiré ou structure du payload non conforme
      return $this->formatGenericErrorSessionExpired();
    }
  }

  /**
   * Génére un token JWE à partir d'un payload porvenant de l'entité User
   * @param array $payload Payload pour le token jwe
   * @return array struct :
   *                  - return.token : Token JWE
   *                  - return.exp : Timestamp d'expiration en ms
   *                  - return.iat : Timestamp de création en ms
   */
  protected function getJWEToken(array $payload): array{
    return [
      'token' => $this->oJweTokenManager->encryptToken($payload), // token JWE
      'exp' => ($payload['exp'] * 1000), // timestamp d'expiration en ms
      'iat' => ($payload['iat'] * 1000), // timestamp de création en ms
    ];
  }

  /**
   * Génére payload pour jwetoken à partir d'un payload porvenant de l'entité User
   * @param array $aUserData Payload provenant de l'entité User
   * @return array struct :
   *                  - return.token : Token JWE
   *                  - return.exp : Timestamp d'expiration en ms
   *                  - return.iat : Timestamp de création en ms
   */
  protected function getJWETokenPayload(array $aUserData) {
    // The payload we want to encrypt. It MUST be a string.
    $iIssuedAt = time();
    $iExp = $iIssuedAt + $this->oProperties->get('token_ttl');
    $sIssuer = $this->oProperties->get('app_identifier');
    $sIpAudience = $this->oVitisRequest->getClientIp();
    $sUrlAudience = str_replace('/', '', $this->oVitisRequest->getRequest()->getBaseUrl());
    $sJti = Uuid::uuid1()->toString(); //ex : e4eaaaf2-d142-11e1-b3e4-080027620cdd
    $payload = [
      'iat' => $iIssuedAt,
      'nbf' => $iIssuedAt,
      'exp' => $iExp,
      'iss' => $sIssuer,
      'aud' => $sIpAudience . ':' . $sUrlAudience,
      'sub' => 'Issued by ' . $sIssuer . ' from ' . $sIpAudience . ' on ' . $sUrlAudience . ' at ' . $iIssuedAt . ' for user ' . $aUserData['user']['login'],
      'jti' => $sJti,
      'data' => $aUserData,
    ];

    return $payload;
  }

  /**
   * envoi la requête de récupération du token
   *
   * @param  string     $sCode
   * @return array|null
   */
  private function getOAuth2tokenFromProvider(string $sCode): ?array{
    try {
      $aTokenData = [
        'grant_type' => 'authorization_code',
        'client_id' => $this->oProperties->get('oauth2_clientId'),
        'client_secret' => $this->oProperties->get('oauth2_clientSecret'),
        'code' => $sCode,
        'redirect_uri' => $this->oProperties->get('app_client_url') . '/oauth2',
      ];
      $aTokenHeaders = [
        'Content-Type: application/x-www-form-urlencoded',
      ];
      // Lance la requête pour vérifier le captcha
      $oHttpFactory = new VitisHttpClient($this->oProperties);
      $client = $oHttpFactory->createClient(['verify_peer' => false]);
      $response = $client->request('POST', $this->oProperties->get('oauth2_token_endpoint'), ['body' => $aTokenData, 'headers' => $aTokenHeaders]);
      $responseContent = json_decode($response->getContent(), true);

      return (empty($responseContent)) ? null : $responseContent;

    } catch (\Exception $e) {
      $this->oLoggerManager->get('API_ERROR')->error('an error occured while reuest oauth2 providers : ' . $e->getMessage());
    }
    return null;
  }

  /**
   * valide le token et extrait la donnée
   *
   * @param  string     $sToken
   * @return array|null
   */
  private function validateAndExtractDataFromOAuth2Token(string $sToken): ?array{
    $serializerManager = new JWSSerializerManager([
      new CompactSerializer(),
    ]);

    try {
      // désérialise le token pour en récupérer le contenu
      $oJwt = $serializerManager->unserialize($sToken);

      if ($oJwt->countSignatures() === 0) {
        throw new \Exception('No signature on JWT token');
      }
    } catch (\Exception $e) {
      $this->oLoggerManager->get('API_ERROR')->error('Error occured while reading oauth2 jwt : ' . $e->getMessage());
      return null;
    }

    return json_decode($oJwt->getPayload(), true);
  }

}
