<?php

namespace App\Module\Vitis\WebService\Entity;

use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use Doctrine\ORM\EntityManager;
use Doctrine\Persistence\Mapping\ClassMetadata;
use Doctrine\Persistence\ObjectManager;

/**
 * Objet ORM Abstrait pour permettre la reprojection des geom
 *
 * La classe qui hérite de cette classe doit
 * - Utiliser le vitisEntityListener pour récupérer les services
 *
 * TODO : 
 * - Reprojection inverse pour l'insertion en base (faut jouer avec les classmetada du setter de l'em)
 * 
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */

abstract class AbstractGeomEntity extends AbstractEntity {
  /**
   * @var EntityManager|null
   */
  protected $oEntityManager;
  /**
   * @var LoggerManager|null
   */
  protected $oLoggerManager;

  /**
   * Reprojette une géométrie EWKT d'aprés ce qui est précisé dans le client
   * @param string $sGeom géométrie d'origine en EWKT
   * @return string
   */
  protected function getReprojectedGeomForAPI(string $sGeom): string {
    if (empty($this->aRequestData['result_srid'])) {
      $this->oLoggerManager->get('SQL_ERROR')->error('Cannot find SRID in request parameters');
      return $sGeom;
    }
    $aMatches = [];
    if (!\preg_match('/^SRID=(?<srid>[0-9]+);/', $sGeom, $aMatches)) {
      $this->oLoggerManager->get('SQL_ERROR')->error('cannot find SRID in origin geom');
      return $sGeom;
    }
    $sDestSrid = $this->aRequestData['result_srid'];
    $sOriginSrid = $aMatches['srid'];
    if ($sOriginSrid == $sDestSrid) {
      // same srid
      return $sGeom;
    }
    // si la géométrie n'a pas pu être casté on remonte la géométrie d'origine
    return $this->reprojectGeom($sGeom, $sDestSrid) ?? $sGeom;
  }

  /**
   * Tente de reprojeter une géométrie EWKT dans un srid souhaité via postgis
   *
   * @param  string      $sGeom
   * @param  int         $iSrid
   * @return string|null
   */
  private function reprojectGeom(string $sGeom, int $iSrid): ?string {
    if (!isset($this->oEntityManager)) {
      $this->oLoggerManager->get('SQL_ERROR')->error('Entity Manager not loaded');
      return null;
    }

    $qb = $this->oEntityManager->getConnection()->createQueryBuilder();
    // cast via postgres
    $qb->select('ST_AsEWKT(ST_Transform(ST_GeomFromEWKT(:GEOM), :SRID::integer))')
      ->setParameter('GEOM', $sGeom)
      ->setParameter('SRID', $iSrid);

    try {
      return (string) $qb->executeQuery()->fetchOne();
    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('Cannot transform geometry to SRID : ' . $iSrid . ' : ' . $e->getMessage(), [$this, $e]);
      return null;
    }

  }

  /* ---------------------------------------------------------------- */
  /* -------------------------- Interfaces -------------------------- */
  /* ---------------------------------------------------------------- */

  /**
   * Permet la récupération d'un gestionnaire d'entité pour interagir avec la base
   * @return void
   */
  public function injectObjectManager(ObjectManager $objectManager, ClassMetadata $classMetadata): void {
    if ($objectManager instanceof EntityManager) {
      $this->oEntityManager = $objectManager;
    }
  }

  /**
   * Permet au gestionnaire d'event de doctrine de charger le service des logger lors du chargement
   * @return void
   */
  public function setLoggerManagerService(LoggerManager $oService): void {
    $this->oLoggerManager = $oService;
  }
}
