<?php

namespace App\Module\Vitis\WebService\Entity;

use App\Module\Vitis\WebService\Repository\UserRepository;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use Doctrine\Common\Collections\ArrayCollection;
use Doctrine\ORM\EntityManager;
use Doctrine\ORM\Mapping\Column;
use Doctrine\ORM\Mapping\Entity;
use Doctrine\ORM\Mapping\EntityListeners;
use Doctrine\ORM\Mapping\Id;
use Doctrine\ORM\Mapping\Table;
use Doctrine\Persistence\Mapping\ClassMetadata;
use Doctrine\Persistence\ObjectManager;

/**
 * Classe entité représentant les privilèges postgres
 * @Entity(repositoryClass="App\Module\Vitis\WebService\Repository\PrivilegeRepository")
 * @EntityListeners({"App\Module\Vitis\WebService\EventListener\VitisEntityListener"})
 * @Table(schema="@prop(schema_vitis)",name="privileges")
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 * @author Armand Bahi <armand.bahi@veremes.com>
 * @author Alexandre Mironenkov <alexandre.mironenkov@veremes.com>
 */
class Privilege extends AbstractEntity {
  /**
   * @var EntityManager|null
   */
  private $oEntityManager;
  /**
   * @var LoggerManager|null
   */
  private $oLoggerManager;

  /**
   * @var string|null
   * @Id()
   * @Column(type="string", length=64)
   */
  private $rolname;

  /**
   * @var string|null
   * @Column(type="text", nullable=true)
   */
  private $description;

  /**
   * @var bool
   * @Column(type="boolean", nullable=false, options={"default"="FALSE"})
   */
  private $locked = false;

  /**
   * @var ArrayCollection<User> Liste des utilisateurs
   * S'appuie sur le pg_catalog surtout ne pas essayer d'ajouter de Many-To-Many
   */
  private $aUsers;

  /**
   * @return void
   */
  public function __construct() {
    $this->aUsers = new ArrayCollection();
  }

  /**
   * @return string|null
   */
  public function getId(): ?string {
    return $this->rolname;
  }

  /**
   * @return string|null
   */
  public function getRolname(): ?string {
    return $this->rolname;
  }

  /**
   * Ne permet la mise à jour que si le rolname n'est pas défini
   * @param string $rolname
   * @return self
   */
  public function setRolname(string $rolname): self {
    $this->rolname = $this->rolname ?? $rolname;
    return $this;
  }
  /**
   * @return string|null
   */
  public function getDescription(): ?string {
    return $this->description;
  }

  /**
   * @param string|null $description
   * @return self
   */
  public function setDescription(?string $description): self {
    $this->description = $description;
    return $this;
  }

  /**
   * retourne la liste des user_id
   * @return array
   */
  public function getUsers(): array{
    // Si l'objet viens d'être initialisé
    if (!isset($this->oEntityManager)) {
      return [];
    }
    try {
      $userRepository = $this->oEntityManager->getRepository(User::class);

      if (!$userRepository instanceof UserRepository) {
        throw new \Exception("Repository is not a UserRepository", 1);
      }

      $this->aUsers = new ArrayCollection($userRepository->findByPrivilege($this));
    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('Cannot get user list for rolename ' . $this->rolname, [$this, $e]);
      return [];
    }
    return $this->aUsers->toArray();
  }
  /**
   * @return bool
   */
  public function getLocked(): bool {
    return $this->locked;
  }

  /**
   * Protége le privilege contre la suppression
   * @return self
   */
  public function lock(): self {
    $this->locked = true;
    return $this;
  }

  /**
   * Retourne un tableau de login correspondant au tableau d'id utilisateurs.
   * @return array
   */
  public function getUsersLabel(): array{
    // Si l'objet viens d'être initialisé
    if (!isset($this->oEntityManager)) {
      return [];
    }
    try {
      $aUsers = $this->aUsers->toArray();
      // récupération du repository
      $userRepository = $this->oEntityManager->getRepository(User::class);

      if (!$userRepository instanceof UserRepository) {
        return [];
      }

      // find users
      $aEntities = $userRepository->findByIdInRange($aUsers);
    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('Cannot get user list for rolename ' . $this->rolname, [$this, $e]);
      return [];
    }
    // génère un tableau de login
    $aUserLogin = [];
    if (!empty($aEntities)) {
      foreach ($aEntities as $oUser) {
        $aUserLogin[] = $oUser->getLogin();
      }
    }
    return $aUserLogin;
  }

  /* ---------------------------------------------------------------- */
  /* -------------------------- Interfaces -------------------------- */
  /* ---------------------------------------------------------------- */

  /**
   * Permet la récupération d'un gestionnaire d'entité pour interagir avec la base
   * @return void
   */
  public function injectObjectManager(ObjectManager $objectManager, ClassMetadata $classMetadata): void {
    if ($objectManager instanceof EntityManager) {
      $this->oEntityManager = $objectManager;
    }
  }

  /**
   * Permet au gestionnaire d'event de doctrine de charger le service des logger lors du chargement
   * @return void
   */
  public function setLoggerManagerService(LoggerManager $oService): void {
    $this->oLoggerManager = $oService;
  }

}
