<?php

namespace App\Module\Vitis\WebService\Entity;

use App\Module\Vitis\WebService\Repository\PrivilegeRepository;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use Doctrine\Common\Collections\ArrayCollection;
use Doctrine\ORM\EntityManager;
use Doctrine\ORM\Mapping\Column;
use Doctrine\ORM\Mapping\Entity;
use Doctrine\ORM\Mapping\EntityListeners;
use Doctrine\ORM\Mapping\GeneratedValue;
use Doctrine\ORM\Mapping\Id;
use Doctrine\ORM\Mapping\JoinColumn;
use Doctrine\ORM\Mapping\JoinTable;
use Doctrine\ORM\Mapping\ManyToMany;
use Doctrine\ORM\Mapping\ManyToOne;
use Doctrine\ORM\Mapping\SequenceGenerator;
use Doctrine\ORM\Mapping\Table;
use Doctrine\Persistence\Mapping\ClassMetadata;
use Doctrine\Persistence\ObjectManager;

/**
 * Objet ORM pour v_user
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 * @author Armand Bahi <armand.bahi@veremes.com>
 * @author Alexandre Mironenkov <alexandre.mironenkov@veremes.com>
 *
 * @Entity(repositoryClass="App\Module\Vitis\WebService\Repository\UserRepository")
 * @EntityListeners({"App\Module\Vitis\WebService\EventListener\VitisEntityListener"})
 * @Table(schema="@prop(schema_vitis)",name="v_user")
 */
class User extends AbstractEntity {
  /**
   * @var string
   */
  private const ROLE_ADMIN = 'vitis_admin';

  /**
   * @var EntityManager|null
   */
  private $oEntityManager;
  /**
   * @var LoggerManager|null
   */
  private $oLoggerManager;

  /**
   * @var int
   * @Column(type="integer", name="user_id", nullable=false)
   * @Id
   * @GeneratedValue(strategy="SEQUENCE")
   * @SequenceGenerator(sequenceName="@prop(schema_vitis).seq_common")
   */
  private $user_id;

  /**
   * @var string|null
   * @Column(type="string", length=64)
   */
  private $login;

  /**
   * @var string|null
   * @Column(type="string", length=80, nullable=true)
   */
  private $name;

  /**
   * @var string|null
   * @Column(type="string", length=100, nullable=true)
   */
  private $email;

  /**
   * @var string|null
   * @Column(type="string", length=80, nullable=true)
   */
  private $company;

  /**
   * @var string|null
   * @Column(type="string", length=80, nullable=true)
   */
  private $department;

  /**
   * @var string|null
   * @Column(type="string", length=255, name="ip_constraint", nullable=true)
   */
  private $ip_constraint;

  /**
   * @var \DateTime|null
   * @Column(type="datetime", name="last_connection", nullable=true)
   */
  private $last_connection;

  /**
   * @var Domain|null
   * @ManyToOne(targetEntity="App\Module\Vitis\WebService\Entity\Domain", inversedBy="users")
   * @JoinColumn(name="domain_id", referencedColumnName="domain_id")
   */
  private $domain;

  /**
   * @var string
   * @Column(type="string", length=40, name="timezone_id")
   */
  private $timezone_id = 'Europe/Paris';

  /**
   * @var string
   * @Column(type="string", length=11, name="formatdate_id")
   */
  private $formatdate_id = 'YYYY-MM-DDT';

  /**
   * @var bool|null
   * @Column(type="boolean", name="dataencrypt")
   */
  private $data_encrypt = false;

  /**
   * @var string|null
   * @Column(type="string", length=100, name="secretkey")
   */
  private $secret_key;

  /**
   * @var string|null
   * @Column(type="string", name="restriction")
   */
  private $restriction;

  /**
   * @var string|null
   * @Column(type="string", nullable=true)
   */
  private $comp_data;

  /**
   * @var array|null
   */
  private $aRoles;

  /**
   * @var ArrayCollection<Group>
   * @ManyToMany(targetEntity="Group")
   * @JoinTable(
   *   name="@prop(schema_vitis).v_user_group_by_rights",
   *   joinColumns={@JoinColumn(name="user_id", referencedColumnName="user_id")},
   *   inverseJoinColumns={@JoinColumn(name="group_id", referencedColumnName="group_id")}
   * )
   */
  private $aGroups;

  /**
   * @var ArrayCollection<Restriction>
   * @ManyToMany(targetEntity="Restriction")
   * @JoinTable(
   *   name="@prop(schema_vitis).v_user_restriction",
   *   joinColumns={@JoinColumn(name="user_id", referencedColumnName="user_id")},
   *   inverseJoinColumns={@JoinColumn(name="restriction_id", referencedColumnName="restriction_id")}
   * )
   */
  private $aRestrictions;

  public function __construct() {
    $this->aRoles = [];
    $this->aGroups = new ArrayCollection();
    $this->aRestrictions = new ArrayCollection();
  }

  /**
   * @return int identifiant de l'utilisateur
   */
  public function getUserId(): int {
    return $this->user_id;
  }

  /**
   * @return string chaine formétée name (login)
   */
  public function getUserString(): string {
    return $this->name . ' (' . $this->login . ')';
  }

  /**
   * @return string login de l'utilisateur
   */
  public function getLogin(): string {
    return $this->login;
  }

  /**
   * Défini le login de l'utilisateur, réalisable uniquement lors de la création
   * @param string $login login de l'utilisateur
   * @return self
   */
  public function setLogin(string $login): self {
    if (empty($this->login)) {
      $this->login = $login;
    }
    return $this;
  }

  /**
   * @return string|null nom de l'utilisateur
   */
  public function getName(): ?string {
    return $this->name;
  }

  /**
   * @param string $name nom de l'utilisateur
   * @return self
   */
  public function setName(?string $name): self {
    $this->name = $name;
    return $this;
  }

  /**
   * @return string|null mail de l'utilisateur
   */
  public function getEmail(): ?string {
    return $this->email;
  }

  /**
   * @param string $email mail de l'utilisateur
   * @return self
   */
  public function setEmail(?string $email): self {
    $this->email = $email;
    return $this;
  }

  /**
   * @return string|null organisme de l'utilisateur
   */
  public function getCompany(): ?string {
    return $this->company;
  }

  /**
   * @return string|null timezone de l'utilisateur
   */
  public function getTimezoneId(): ?string {
    return $this->timezone_id;
  }

  /**
   * @param string|null $timezone_id timezone de l'utilisateur
   * @return self
   */
  public function setTimezoneId(?string $timezone_id): self {
    $this->timezone_id = $timezone_id;
    return $this;
  }

  /**
   * @return string|null formatdate_id de l'utilisateur
   */
  public function getFormatdateId(): ?string {
    return $this->formatdate_id;
  }

  /**
   * @param string|null $formatdate_id formatdate_id de l'utilisateur
   * @return User
   */
  public function setFormatdateId($formatdate_id): self {
    $this->formatdate_id = $formatdate_id;
    return $this;
  }

  /**
   * @param string $company organisme de l'utilisateur
   * @return self
   */
  public function setCompany(?string $company): self {
    $this->company = $company;
    return $this;
  }

  /**
   * @return string|null service de l'utilisateur
   */
  public function getDepartment(): ?string {
    return $this->department;
  }

  /**
   * @param string|null $department service de l'utilisateur
   * @return self
   */
  public function setDepartment(?string $department): self {
    $this->department = $department;
    return $this;
  }

  /**
   * @return string|null Contrainte d'ip de l'utilisateur
   */
  public function getIpConstraint(): ?string {
    return $this->ip_constraint;
  }

  /**
   * @param string|null $ip_constraint Contrainte d'ip de l'utilisateur
   * @return self
   */
  public function setIpConstraint(?string $ip_constraint): self {
    $this->ip_constraint = $ip_constraint;

    return $this;
  }

  /**
   * @return \DateTime|null Date de la dernière connexion de l'utilisateur
   */
  public function getLastConnection(): ?\DateTime {
    return $this->last_connection;
  }
  /**
   * @return bool
   */
  public function getDataEncrypt(): bool {
    return $this->data_encrypt ?? false;
  }
  /**
   * @param bool $bDataEncrypt
   * @return self
   */
  public function setDataEncrypt(bool $bDataEncrypt): self {
    $this->data_encrypt = $bDataEncrypt;
    return $this;
  }
  /**
   * @return null|string
   */
  public function getSecretKey(): ?string {
    return $this->secret_key;
  }
  /**
   * @param null|string $sSecretkey
   * @return self
   */
  public function setSecretKey(?string $sSecretkey): self {
    $this->secret_key = $sSecretkey;
    return $this;
  }
  /**
   * @return null|string
   */
  public function getRestriction(): ?string {
    return $this->restriction;
  }
  /**
   * @param null|string $sRestriction
   * @return self
   */
  public function setRestriction(?string $sRestriction): self {
    $this->restriction = $sRestriction;
    return $this;
  }

  /**
   * @return self
   */
  public function updateLastConnection(): self {
    $this->last_connection = new \DateTime();
    return $this;
  }

  public function getDomainId(): ?int {
    return $this->domain === null ? null : $this->domain->getDomainId();
  }

  public function getDomainName(): ?string {
    return $this->domain === null ? null : $this->domain->getName();
  }

  public function getDomain(): ?Domain {
    return $this->domain;
  }

  public function setDomain(Domain $domain): self {
    $this->domain = $domain;
    return $this;
  }

  /**
   * @return string|null donnée complémentaire
   */
  public function getCompData(): ?string {
    return $this->comp_data;
  }

  /**
   * @return array|null donnée complémentaire
   */
  public function getCompDataArray(): ?array{
    if (empty($this->comp_data)) {
      return null;
    }
    return json_decode($this->comp_data, true);
  }

  /**
   * @param array|string $mData donnée supplémentaire en array/json
   * @return self|null en cas d'erreur
   */
  public function setCompData($mData): ?self {
    if (\is_array($mData)) {
      $mData = json_encode($mData);
    }

    if (\is_string($mData)) {
      $this->comp_data = $mData;
      return $this;
    }

    return null;
  }

  /**
   * Retourne un tableau qui sera utilisé pour stocker les donnée de l'utilisateur dans le token JWE
   * @return array payload pour le token JWE
   */
  public function getTokenPayload(): array{
    return [
      'login' => $this->login,
      'user_id' => $this->user_id,
      'privileges' => $this->getPrivileges(),
    ];
  }

  /**
   * Retourne un tableau contenant les roles de l'utilisateur
   * @return string[] Roles de l'utilisateur
   */
  public function getRoles(): array{
    // Si l'objet viens d'être initialisé
    if (!isset($this->oEntityManager)) {
      return [];
    }

    try {
      $privilegeRepository = $this->oEntityManager->getRepository(Privilege::class);

      if (!$privilegeRepository instanceof PrivilegeRepository) {
        throw new \Exception("Repository is not a PrivilegeRepository", 1);
      }

      $this->aRoles = $privilegeRepository->getRolesByUser($this) ?? [];
      return $this->aRoles;
    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('Cannot get role list for user ' . $this->login, [$this, $e]);
      return [];
    }
  }
  /**
   * Retourne un tableau contenant les privileges de l'utilisateur
   * @return string[] privileges de l'utilisateur
   */
  public function getPrivileges(): array{
    // Si l'objet viens d'être initialisé
    if (!isset($this->oEntityManager)) {
      return [];
    }
    /* @var $privilegeRepository PrivilegeRepository */
    $privilegeRepository = $this->oEntityManager->getRepository(Privilege::class);

    if (!$privilegeRepository instanceof PrivilegeRepository) {
      return [];
    }

    $aPrivileges = $privilegeRepository->getPrivilegesByUser($this);
    return $aPrivileges;
  }
  /**
   * @return bool
   */
  public function isAdmin(): bool {
    if (empty($this->aRoles)) {
      $this->getRoles();
    }
    return in_array(self::ROLE_ADMIN, $this->aRoles);
  }

  /**
   * @return int[] liste des user_id
   */
  public function getGroups(): array{
    $aGroupId = [];
    foreach ($this->aGroups as $oGroup) {
      $aGroupId[] = $oGroup->getGroupId();
    }
    return $aGroupId;
  }

  /**
   * @return int[] liste des user_id
   */
  public function getRestrictions(): array{
    $aRestrictionId = [];
    foreach ($this->aRestrictions as $oRestriction) {
      $aRestrictionId[] = $oRestriction->getRestrictionId();
    }
    return $aRestrictionId;
  }

  /* ---------------------------------------------------------------- */
  /* -------------------------- Interfaces -------------------------- */
  /* ---------------------------------------------------------------- */

  /**
   * Permet la récupération d'un gestionnaire d'entité pour interagir avec la base
   * @return void
   */
  public function injectObjectManager(ObjectManager $objectManager, ClassMetadata $classMetadata): void {
    if ($objectManager instanceof EntityManager) {
      $this->oEntityManager = $objectManager;
    }
  }

  /**
   * Permet au gestionnaire d'event de doctrine de charger le service des logger lors du chargement
   * @return void
   */
  public function setLoggerManagerService(LoggerManager $oService): void {
    $this->oLoggerManager = $oService;
  }
}
