<?php

namespace App\Module\Vitis\WebService\EventListener;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use Doctrine\Common\EventManager;
use Doctrine\ORM\Events;
use Doctrine\ORM\Event\LoadClassMetadataEventArgs;
use Doctrine\ORM\Id\SequenceGenerator;
use Doctrine\ORM\Mapping\ClassMetadataInfo;

/**
 * EventListener Doctrine pour utiliser les properties dans les annotations
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */
class MappingListener {
  /**
   * @var Properties
   */
  private $oPropertiesSrvc;
  /**
   * @var EventManager
   */
  private $oEventManager;

  /**
   * @param Properties $PropertiesSrvc Un objet pour accéder aux properties de l'application
   * @param EventManager $oEventManager Service permettant d'émettre des evenement
   */
  public function __construct(Properties $PropertiesSrvc, EventManager $oEventManager) {
    // Sauvegarde du service de properties
    $this->oPropertiesSrvc = $PropertiesSrvc;
    $this->oEventManager = $oEventManager;
    $this->oEventManager->addEventListener(Events::loadClassMetadata, $this);
  }

  /**
   * Permet de modifier les méta-données d'une entité.
   * Utiliser pour bind des properties dans le schéma par exemple
   * @param LoadClassMetadataEventArgs $eventArgs objet contenant l'encsemple des méta-données de la classe
   * @return void
   */
  public function loadClassMetadata(LoadClassMetadataEventArgs $eventArgs): void {
    $classMetadata = $eventArgs->getClassMetadata();

    if (!$classMetadata instanceof ClassMetadataInfo) {
      return;
    }
    // remplacement du schema dans le nom de la table
    $table = $classMetadata->table;
    if (!empty($table['schema'])) {
      $table['schema'] = $this->oPropertiesSrvc->replacePropertiesInString($table['schema']);
    }
    if (!empty($table['name'])) {
      $table['name'] = $this->oPropertiesSrvc->replacePropertiesInString($table['name']);
    }
    $classMetadata->setPrimaryTable($table);
    // remplacement du schema dans les tables de jointure Many-To-Many
    foreach ($classMetadata->getAssociationMappings() as $fieldName => $mapping) {
      if ($mapping['type'] == ClassMetadataInfo::MANY_TO_MANY && !empty($mapping['isOwningSide'])) {
        $classMetadata->associationMappings[$fieldName]['joinTable']['name'] = $this->oPropertiesSrvc->replacePropertiesInString($mapping['joinTable']['name']);
      }
    }
    // remplacement dans les sequences
    if ($classMetadata->isIdGeneratorSequence()) {
      $newDefinition = $classMetadata->sequenceGeneratorDefinition;
      $newDefinition['sequenceName'] = $this->oPropertiesSrvc->replacePropertiesInString($newDefinition['sequenceName']);

      $classMetadata->setSequenceGeneratorDefinition($newDefinition);
      $em = $eventArgs->getEntityManager();
      $sequenceGenerator = new SequenceGenerator(
        $em->getConfiguration()->getQuoteStrategy()->getSequenceName(
          $newDefinition,
          $classMetadata,
          $em->getConnection()->getDatabasePlatform()),
        intval($newDefinition['allocationSize'])
      );
      $classMetadata->setIdGenerator($sequenceGenerator);
    }
  }
}
