<?php

namespace App\Module\Vitis\WebService\EventListener;

use App\Module\Vitis\WebService\EventListener\VitisEventArgs;
use App\Module\Vitis\WebService\EventListener\VitisEvents;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use Doctrine\Common\EventManager;
use Doctrine\ORM\Event\LifecycleEventArgs;
use Doctrine\Persistence\ObjectManager;

/**
 * EventListener Doctrine pour charger des services dans l'entité
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */
class VitisEntityListener {
  /**
   * Map faisant le lien entre les méthodes disponible sur les entités et les services de cette classe
   * @var array
   */
  private const SERVICE_SETTER_MAP = [
    'setPropertiesService' => 'oPropertiesSrvc',
    'setLoggerManagerService' => 'oLoggerManagerSrvc',
    'setRouteConfiguratorService' => 'oRouteConfigurator',
    'setFileManagerService' => 'oFileManager',
  ];

  /**
   * @var Properties
   */
  private $oPropertiesSrvc;
  /**
   * @var LoggerManager
   */
  private $oLoggerManagerSrvc;
  /**
   * @var Route
   */
  private $oRouteConfigurator;
  /**
   * @var FileManager
   */
  private $oFileManager;

  /**
   * @var ObjectManager|null
   */
  private $oObjectManager;

  /**
   * @param Properties $PropertiesSrvc Un objet pour accéder aux properties de l'application
   * @param LoggerManager $LoggerManagerSrvc Un objet pour accéder aux loggers de l'application
   * @param Route $oRouteConfigurator service permettant d'accéder aux informations d'une route
   */
  public function __construct(Properties $PropertiesSrvc, LoggerManager $LoggerManagerSrvc, Route $oRouteConfigurator, FileManager $oFileManager) {
    // Sauvegarde du service de properties
    $this->oPropertiesSrvc = $PropertiesSrvc;
    $this->oLoggerManagerSrvc = $LoggerManagerSrvc;
    $this->oRouteConfigurator = $oRouteConfigurator;
    $this->oFileManager = $oFileManager;
  }

  /**
   * Permet de suscrire aux event dans databaseConnection
   * @param EventManager $oEventManager Un objet pour accéder aux events
   * @return void
   */
  public function subscribeEvent(EventManager $oEventManager): void {
    $oEventManager->addEventListener(VitisEvents::preDenormalization, $this);
  }

  /**
   * Event appelé aprés la création et l'hydratation d'une entité.
   * Elle va analyser l'entité et charger les services qui lui sont necessaires
   *
   * @param object $oEntity Entité doctrine instancié dans l'entityManager
   * @param LifecycleEventArgs $args
   *
   * @return void
   */
  public function postLoad($oEntity, LifecycleEventArgs $args): void {
    $this->injectEntityService($oEntity);
  }

  /**
   * Event appelé aprés la création d'une entité lors de sa dénormalisation.
   *
   * @param VitisEventArgs $oArgs
   *
   * @return void
   */
  public function preDenormalization(VitisEventArgs $oArgs): void {
    $this->injectEntityService($oArgs->getEntity());
  }

  /**
   * Permet d'injecter des services dans une entité.
   * @param object $oEntity Entité doctrine instancié dans l'entityManager
   * @return void
   */
  private function injectEntityService($oEntity): void {
    // récupération des informations de la class
    $oEntityReflection = new \ReflectionClass($oEntity);

    foreach (self::SERVICE_SETTER_MAP as $sSetterName => $sServiceName) {
      // si le setter existe
      if ($oEntityReflection->hasMethod($sSetterName)) {
        $oReflectionMethod = $oEntityReflection->getMethod($sSetterName);
        // Si le setter est public on l'utilise
        if ($oReflectionMethod->isPublic() && isset($this->$sServiceName)) {
          $oEntity->$sSetterName($this->$sServiceName);
        }
      }
    }

    // cas particulier de l'ancienne interface ObjectManagerAware
    if (isset($this->oObjectManager)) {
      if ($oEntityReflection->hasMethod('injectObjectManager')) {
        $oReflectionMethod = $oEntityReflection->getMethod('injectObjectManager');
        if ($oReflectionMethod->isPublic()) {
          $oClass = $this->oObjectManager->getClassMetadata(\get_class($oEntity));
          $oEntity->injectObjectManager($this->oObjectManager, $oClass);
        }
      }
    }
  }

  /**
   * Set un gestionnaire d'objet pour l'ORM
   * @param ObjectManager $oObjectManager Manager d'objet à utiliser
   * @return void
   */
  public function setObjectManager(ObjectManager $oObjectManager): void {
    $this->oObjectManager = $oObjectManager;
  }
}
