<?php

// src/Module/Vitis/WebService/EventSubscriber/TokenSubscriber.php

namespace App\Module\Vitis\WebService\EventSubscriber;

use App\Module\Vitis\WebService\Controller\AbstractVitisController;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\Mail\Mail;
use App\Module\Vitis\WebService\Service\VitisRequest;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\HttpKernel\Event\ExceptionEvent;
use Symfony\Component\HttpKernel\Event\RequestEvent;
use Symfony\Component\HttpKernel\KernelEvents;
use Symfony\Component\HttpKernel\KernelInterface;

/**
 * EventSubscriber sur les événements du Kernel.
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 * @author Alexandre Mironenkov <alexandre.mironenkov@veremes.com>
 */
class RequestSubscriber extends AbstractVitisController implements EventSubscriberInterface {
  // /**
  //  * @var Request
  //  */
  // private $oRequest;
  /**
   * @var \Monolog\Logger
   */
  private $oLogger;
  /**
   * @var KernelInterface
   */
  private $oKernel;

  /**
   * Initialise le controler et instancie les services requis
   * @param Properties $oPropertiesSrvc Service d'accès aux properties
   * @param VitisRequest $oVitisRequest Service d'abstraction de la requête Symfony
   * @param Route $oRouteConfig Service d'accès à la configuration de la route
   * @param LoggerManager $oLoggerManager Service d'utilsation des logs
   * @param DatabaseConnection $oDatabaseConnection Service d'accès à la base de données
   * @param FileManager $oFileManager Service permettant d'accéder aux fichiers
   * @param KernelInterface $oKernel Service permettant d'accéder au Kernel symfony
   */
  public function __construct(Properties $oPropertiesSrvc, VitisRequest $oVitisRequest, Route $oRouteConfig, LoggerManager $oLoggerManager, DatabaseConnection $oDatabaseConnection, FileManager $oFileManager, KernelInterface $oKernel) {
    parent::__construct($oPropertiesSrvc, $oVitisRequest, $oRouteConfig, $oLoggerManager, $oDatabaseConnection, $oFileManager);
    $this->oKernel = $oKernel;

    // activation de la récupération des erreurs par mails
    if ($this->errorHandlingByMailIsEnabled()) {
      $this->oLoggerManager->get('API_ERROR')->debug('Mail error handler enabled');
      \set_error_handler($this->customErrorHandler(...), E_ERROR | E_CORE_ERROR | E_COMPILE_ERROR | E_USER_ERROR | E_RECOVERABLE_ERROR);
    }
  }

  /**
   * détermine s'il faut activer la gestion des erreurs par mail
   * @return bool
   */
  private function errorHandlingByMailIsEnabled(): bool {
    return $this->oProperties->get('enable_error_notification') === true && !empty($this->oProperties->get('admin_sign_up'));
  }

  /**
   * Fonction appelé lors de la réception d'une requête HTTP
   * Controle si l'utilisateur à le droit d'accéder à la route
   * @return void
   */
  public function checkUserRestrictions(RequestEvent $event): void {
    // si c'est une reqûete OPTION pas de controle
    if ($event->getRequest()->getMethod() === 'OPTIONS') {
      // réponse par défaut pour les requêtes options
      $event->setResponse($this->oResponse);
      return;
    }

    // récupération du logger
    $this->oLogger = $this->oLoggerManager->get('API_ERROR');
    if (!$this->userCanUseRoute()) {
      $this->oLogger->error('Cannot use this route : Authorization check failed');
      $event->setResponse($this->oResponse);
      return;
    }
  }
  /**
   * Fonction appelé lors de la réception d'une requête HTTP
   * Controle si l'utilisateur à le droit d'accéder à la route
   * @return void
   */
  public function checkRequestRestrictions(RequestEvent $event): void {
    // si c'est une reqûete OPTION pas de controle
    if ($event->getRequest()->getMethod() === 'OPTIONS') {
      // réponse par défaut pour les requêtes options
      $event->setResponse($this->oResponse);
      return;
    }
    // récupération du logger
    $this->oLogger = $this->oLoggerManager->get('API_ERROR');
    $sRegexp = $this->oRouteConfig->getOriginRegexp();

    if ($sRegexp !== null) {
      if (!$this->checkIpConstraint($sRegexp)) {
        $this->oLogger->error('Cannot use this route : Origin check failed');
        // L'utilisateur ne peut pas se connecter avec cette ip
        $aReturn = [
          'status' => '0',
          'message' => 'Connection forbidden from ip :' . $this->oVitisRequest->getClientIp(),
          'errorMessage' => 'ERROR_USER_IP_UNAUTHORIZED',
        ];
        $this->formatResponse($aReturn, Response::HTTP_UNAUTHORIZED);
        $event->setResponse($this->oResponse);
        return;
      }
    }

  }

  /**
   * Fonction appelé lors de la réception d'une requête HTTP
   * Parse les données envoyées
   * @return void
   */
  public function parseRequestData(RequestEvent $event): void {
    // Récuperation de la requête à travers
    // $this->oRequest = $event->getRequest();
    // Parsage des données de la requête si la requête n'a pas déjà été parsée
    if (!$this->oVitisRequest->alreadyParsed()) {
      $this->oVitisRequest->extractData();
    }
  }

  /**
   * Fonction appelé lors de l'émission d'une erreur 500
   * Beaucoup d'erreur de dev ne passe pas par cette fonction :
   *  - erreur de classe, héritage, typage
   *  - erreur de syntaxe
   * génére une réponse générique
   * @param ExceptionEvent $event Evénnement d'erreur symfony
   * @return void
   */
  public function genericError(ExceptionEvent $event): void {
    // récupration de l'exception
    $e = null;
    if (method_exists($event, 'getThrowable')) {
      $e = $event->getThrowable();
    } /* else if (method_exists($event, 'getException')) {
    $e = $event->getException();
    } */

    // Notification des erreurs par mail.
    if ($this->errorHandlingByMailIsEnabled()) {
      $sErrorMsg = $e->getFile() . ' on line ' . $e->getLine() . ' --> ' . $e->getMessage();
      $this->sendErrorByMail($sErrorMsg);
    }
    // on log l'erreur
    $this->oLoggerManager->get('API_ERROR')->error('Cannot use this route : ' . $e->getMessage(), [$e]);
    // si l'environnement est en dev on retourne l'erreur symfony
    if ($this->oKernel->getEnvironment() == 'dev') {
      return;
    }
    // on passe là si l'appli est en mode prod, le message n'est pas utile
    $oResponse = $this->formatGenericError('ERROR_SERVER_REQUEST_500', '', '');
    $event->setResponse($oResponse);
  }

  /**
   * utilisé pour catcher toutes les erreurs de l'application
   * @param integer $errno
   * @param string $errstr
   * @param string $errfile
   * @param integer $errline
   * @return bool
   */
  public function customErrorHandler(int $errno, string $errstr, string $errfile, int $errline): bool {
    $this->oLoggerManager->get('API_ERROR')->error('Cannot use this route : ' . $errstr);
    // Notification des erreurs par mail.
    $sErrorMsg = $errfile . " on line " . $errline . " --> " . $errstr;
    $this->sendErrorByMail($sErrorMsg);

    return true;
  }

  /**
   * @param string $sErrorMsg Message d'erreur à envoyer
   * @return void
   */
  private function sendErrorByMail(string $sErrorMsg): void {
    $oMail = new Mail($this->oProperties, $this->oLoggerManager, $this->oFileManager);
    // Ecriture de l'erreur dans les logs.
    $sNow = date('d/m/Y H:i:s');
    // Envoi du mail.
    $sErrorMsg = "[$sNow] | ERROR | " . $sErrorMsg;
    $oMail->loadFromArray([
      'subject' => '[VITIS-ERROR] Une erreur est survenue dans votre application : ' . $this->oProperties->get('app_identifier'),
      'bUseHTML' => true,
      'body' => '<p>Une nouvelle erreur est survenue dans l\'API de ' . $this->oProperties->get('app_api_url')
      . '. Retour d\'erreur (consultez le fichier error.log de l\'application pour plus d\'informations):</p>'
      . '<p><code>' . $sErrorMsg . '</code></p>'
      . '<p> Chemin du vas : ' . $this->oProperties->get('vas_dir') . '</p>',
      'aAttachments' => [],
      'to' => $this->oProperties->get('admin_sign_up'),
      'cc' => '',
      'cci' => '',
    ]);
    $oMail->send();
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    return [
      KernelEvents::REQUEST => [
        /*
         * Il ne faut pas depasser 31 à cause d'initialisation de
         * la configuration des routes dans AbstractVitisController
         */
        ['parseRequestData', 30],
        ['checkUserRestrictions', 0],
        ['checkRequestRestrictions', 0],
      ],
      KernelEvents::EXCEPTION => [
        ['genericError', -10],
      ],
    ];
  }
}
