<?php

namespace App\Module\Vitis\WebService\Repository;

use App\Module\Vitis\WebService\Entity\Privilege;
use App\Module\Vitis\WebService\Entity\User;
use Doctrine\ORM\EntityManagerInterface;
use Doctrine\ORM\Mapping\ClassMetadata;

/**
 * Repository pour les privilèges de Vitis.
 *
 * @method Privilege|null find($id, $lockMode = null, $lockVersion = null)
 * @method Privilege|null findOneBy(array $criteria, array $orderBy = null)
 * @method Privilege[]    findAll()
 * @method Privilege[]    findBy(array $criteria, array $orderBy = null, $limit = null, $offset = null)
 *
 * @author Alexandre Mironenkov   <alexandre.mironenkov@veremes.com>
 */
class PrivilegeRepository extends GenericRepository {
  /**
   * {@inheritdoc}
   */
  public function __construct(EntityManagerInterface $em, ClassMetadata $class) {
    parent::__construct($em, $class);
  }

  /**
   * @param string[] $aRolname nom de rolname
   * @return Privilege[] Returns an array of Privilege entity
   */
  public function findByRolname(array $aRolname): array{
    if (empty($aRolname)) {
      return [];
    }
    $qb = $this->createQueryBuilder('a');
    $qb->andWhere($qb->expr()->in('a.rolname', $aRolname))
      ->orderBy('a.rolname', 'ASC')
    ;
    return $qb->getQuery()->getResult();
  }

  /**
   * Retourne les roles système pour l'utilisateur
   * @param User $oUser l'utilisateur
   * @return array|null
   */
  public function getRolesByUser(User $oUser): ?array{
    /* @var $em EntityManager */
    $qb = $this->getEntityManager()->getConnection()->createQueryBuilder();
    $qb->select('groname')
      ->from('pg_user', 'a')
      ->join('a', 'pg_group', 'b', 'a.usesysid = any(b.grolist)')
      ->where('a.usename=:login')
      ->setParameter('login', $oUser->getLogin());
    try {
      $aResult = $qb->executeQuery()->fetchFirstColumn();
    } catch (\Exception $e) {
      error_log('User : can\'t get Postgres rolegroup list : ' . $e->getMessage());
      $aResult = null;
    }
    return $aResult;
  }

  /**
   * Retourne les privileges système pour l'utilisateur
   * @param User $oUser l'utilisateur
   * @return array|null
   */
  public function getPrivilegesByUser(User $oUser): ?array{
    /* @var $em EntityManager */
    $qb = $this->getEntityManager()->getConnection()->createQueryBuilder();
    $qb->select('rolname')
      ->from('pg_user', 'a')
      ->join('a', 'pg_group', 'b', 'a.usesysid = any(b.grolist)')
      ->join('a', 's_vitis.privileges', 'c', 'b.groname=c.rolname')
      ->where('a.usename=:login')
      ->setParameter('login', $oUser->getLogin());
    try {
      $aResult = $qb->executeQuery()->fetchFirstColumn();
    } catch (\Exception $e) {
      error_log('User : can\'t get Postgres rolegroup list : ' . $e->getMessage());
      $aResult = null;
    }
    return $aResult;
  }

  /**
   * Retourne les roles système pour la liste d'utilisateurs
   * @param array $aUserId l'utilisateur
   * @return array|null
   */
  public function getRolesByUsers(array $aUserId): ?array{
    $aRoles = [];
    $sUserTable = 'user';
    $sSchemaName = $this->getClassMetadata()->getSchemaName();
    if (isset($sSchemaName)) {
      $sUserTable = $sSchemaName . '.' . $sUserTable;
    }
    /* @var $em EntityManager */
    $qb = $this->getEntityManager()->getConnection()->createQueryBuilder();
    $qb->select('c.user_id, a.usename as login, array_to_json(array_agg(b.groname)) as roles')
      ->from('pg_user', 'a')
      ->leftJoin('a', 'pg_group', 'b', 'a.usesysid = any(b.grolist)')
      ->leftJoin('a', $sUserTable, 'c', 'a.usename = c.login')
      ->where($qb->expr()->in('c.user_id', $aUserId))
      ->groupBy('c.user_id, a.usename');
    try {
      $aResult = $qb->executeQuery()->fetchAllAssociative();
      foreach ($aResult as $aRole) {
        $aRole['roles'] = array_filter(json_decode($aRole['roles'], true));
        $aRoles[] = $aRole;
      }
    } catch (\Exception $e) {
      $aRoles = null;
    }
    return $aRoles;
  }

}
