<?php

namespace App\Module\Vitis\WebService\Service\Configuration;

/**
 * Permet de générer de la doc swagger à partir de la définiton des routes
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */

class Documentation extends Route {
  /**
   * basepath de l'API
   * @var string
   */
  private const DEFAULT_BASEPATH = '/v2';
  /**
   * Objet swagger par défaut
   * @var array
   */
  private const DEFAULT_SWAGGER_STRUCT = [
    'swagger' => '2.0',
    'info' => [],
    'basePath' => self::DEFAULT_BASEPATH,
    'schemes' => [
      'https',
    ],
    'produces' => [
      'application/json',
      'application/xml',
    ],
    'paths' => [],
    'definitions' => [],
    'tags' => [],
  ];

  /**
   * Objet par défaut pour la définition swagger d'un parametre
   * @var array
   */
  private const DEFAULT_SWAGGER_PARAMS = [
    'name' => '',
    'in' => 'formData',
    'required' => false,
    'type' => '',
    'description' => '',
  ];

  /**
   * Point d'entré de la génération de la doc
   *
   * @return array
   */
  public function getSwaggerDef(): array{
    if (!isset($this->aConfiguration['routes'])) {
      throw new \ErrorException('Cannot read this routes.json file, routes/configurations level is missing');
    }
    if (!isset($this->aConfiguration['documentations'])) {
      throw new \ErrorException('Cannot read this routes.json file, documentations level is missing');
    }

    $aDef = $this->getDefaultSwagDef();

    foreach ($this->aConfiguration['routes'] as $aRoute) {
      // si la route n'est pas documenté on skip
      if (!array_key_exists('documentation', $aRoute)) {
        continue;
      }

      if (!$this->documentationExists($aRoute['documentation'])) {
        throw new \ErrorException('Configuration of documentation ' . $aRoute['documentation'] . ' is not set');
      }

      $aDocumentation = $this->resolveDocumentation($aRoute['documentation']);
      $aDocSwagDef = $this->docToSwagDef($aRoute, $aDocumentation);
      $sKey = $this->prepareRouteForSwag($aRoute['route']);

      if (empty($sKey)) {
        throw new \ErrorException('Configuration of documentation for route ' . $aRoute['route'] . ' failed : cannot parse route');
      }

      foreach ($aRoute['methods'] as $sHttpMethod) {
        $sMethodKey = strtolower($sHttpMethod);
        $aDef['paths'][$sKey][$sMethodKey] = $aDocSwagDef;
      }
    }

    return $aDef;
  }

  /**
   * check si une documentation est définit dans la conf
   *
   * @param  string $sKey clé de la documentation
   * @return bool
   */
  private function documentationExists(string $sKey): bool {
    return array_key_exists($sKey, $this->aConfiguration['documentations']['definition']);
  }

  /**
   * récupère la documentation demandé et résoud les dépnedances récursivement
   *
   * @param  string     $sKey clé de la documentation à résoudre
   * @return array|null
   */
  private function resolveDocumentation(string $sKey): ?array{
    // si la défintion demandé n'existe pas ça dégage
    if (!$this->documentationExists($sKey)) {
      return null;
    }

    // on trouve la doc demandé
    $aDocumentation = $this->aConfiguration['documentations']['definition'][$sKey];

    if (array_key_exists('extends', $aDocumentation)) {
      $aParentDocumentation = $this->resolveDocumentation($aDocumentation['extends']);

      if ($aParentDocumentation !== null) {
        $aDocumentation = array_merge_recursive($aParentDocumentation, $aDocumentation);
      }
      unset($aDocumentation['extends']);
    }

    return $aDocumentation;
  }

  /**
   * Rétourne la défintion de base et met en forme certains éléments
   *
   * @return array
   */
  private function getDefaultSwagDef(): array{
    $aDef = self::DEFAULT_SWAGGER_STRUCT;

    $aDef['info']['title'] = $this->oPropertiesSrvc->get('app_identifier');
    $aDef['tags'] = $this->aConfiguration['documentations']['tags'];

    return $aDef;
  }

  /**
   * Ajoute des infos dynamiquement sur la doc de al routes en fonction de la définition
   * public, roles, variances, génréation auto
   *
   * @param  array $aRoute Route à documenté
   * @param  array $aDoc Doc de la route
   * @return array doc à jour
   */
  private function docToSwagDef(array $aRoute, array $aDoc): array{
    $aDoc['public'] = false;
    $aDoc['dynamic'] = false;

    if (array_key_exists('check_user', $aRoute)) {
      $aDoc['public'] = ($aRoute['check_user'] === false);
    }
    if (array_key_exists('role_minimum', $aRoute)) {
      $aRole = $aRoute['role_minimum'];
      if (is_string($aRoute['role_minimum'])) {
        $aRole = [$aRoute['role_minimum']];
      }

      $aDoc['roles'] = $aRole;
    }
    if (array_key_exists('options_by_role', $aRoute)) {
      $aDoc['dynamic'] = true;
    }

    if (array_key_exists('genParamsFromOptions', $aDoc)) {
      if ($aDoc['genParamsFromOptions'] !== false) {
        $aDoc = $this->genSwagParamsDefFromRoute($aRoute, $aDoc);
      }
    }

    return $aDoc;
  }

  /**
   * Génère la définition des paramètres d'après les paramètres de la doc
   *
   * @param  array $aRoute Route à documenter
   * @param  array $aDoc doc actuelle
   * @return array doc à jour
   */
  private function genSwagParamsDefFromRoute(array $aRoute, array $aDoc): array{
    $sConfKey = $aDoc['genParamsFromOptions'];

    if (!is_string($sConfKey)) {
      if (array_key_exists('options', $aRoute)) {
        $sConfKey = $aRoute['options'];
      }
    }
    if (!is_string($sConfKey)) {
      return $aDoc;
    }

    // récupération des options
    $oOptions = $this->findConfigurationWithKey($sConfKey);
    // pour chaque field if input = false on passe
    if (!array_key_exists('fields', $oOptions)) {
      return $aDoc;
    }
    $aExcludeFields = (array_key_exists('excludesParams', $aDoc)) ? $aDoc['excludesParams'] : [];
    $aRequiredFields = (array_key_exists('requiredParams', $aDoc)) ? $aDoc['requiredParams'] : [];

    if (!array_key_exists('parameters', $aDoc)) {
      $aDoc['parameters'] = [];
    }

    foreach ($oOptions['fields'] as $oField) {
      // si name dans exclude on passe au suivant
      if (in_array($oField['name'], $aExcludeFields)) {
        continue;
      }

      $aDoc['parameters'][] = $this->genSwagParamsFromField($oField, $aRequiredFields, $sConfKey);
    }

    return $aDoc;
  }

  /**
   * Transforme un paramètre de la route en paramètre pour la doc
   *
   * @param  array  $oField paramètre de la route
   * @param  array  $aRequiredFields  paramètre requis
   * @param  string $sOptKey nom de la config de la route
   * @return array
   */
  private function genSwagParamsFromField(array $oField, array $aRequiredFields, string $sOptKey): array{
    $oDef = self::DEFAULT_SWAGGER_PARAMS;

    $oDef['type'] = $oField['type'];
    $oDef['name'] = $oField['name'];
    $oDef['description'] = strtoupper('DOC_SWAGGER_' . $sOptKey . '_' . $oField['name']);

    switch ($oField['type']) {
      case 'array':
        $oDef['collectionFormat'] = 'multi';
        $oDef['name'] .= '[]';
        break;
      case 'date':
        $oDef['type'] = 'integer';
        break;
      case 'number':
        $oDef['type'] = 'integer';
        break;
    }

    $oDef['required'] = (in_array($oField['name'], $aRequiredFields));

    return $oDef;
  }

  /**
   * transforme la route de l'API en route pour la doc
   *
   * @param  string      $sRoute
   * @return string|null
   */
  private function prepareRouteForSwag(string $sRoute): ?string {
    if (strpos($sRoute, self::DEFAULT_BASEPATH) !== 0) {
      return null;
    }

    return str_replace(self::DEFAULT_BASEPATH, '', $sRoute);
  }
}
