<?php

namespace App\Module\Vitis\WebService\Service\Configuration;

/**
 * Permet d'accéder et de manipuler les properties Json de l'application.
 *
 * Les fichiers de properties sont des fichiers Json spécifique aux modules.
 * Ce sont les fichiers :
 *      -  src/Module/[Module_name]/conf/properties.json
 *      -  src/Module/[Module_name]/conf/properties_server.json
 *
 * Il est possible de dépendre d'autres properties.
 * Les dépendances ciculaire provoque une impossibilité d'accéder à la properties demandée.
 * ex :
 * {
 *     "vas_dir" : "/var/www/vitis_2020/vas",
 *     "ws_data_dir" : "@prop(vas_dir)/var/ws_data/"
 * }
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 * @author Armand Bahi <armand.bahi@veremes.com>
 */
class Properties {
  /**
   * @var array
   */
  private const PHP_INI_KEY_OVERWRITE = [
    'memory_limit' => 'memory_limit', // ajouter une properties et ajouter une fonction pour calculer sa valeur peut-être
    // impossible à travers la fonction ini_set
    //'post_max_size' => 'max_upload_file_size',
    //'upload_max_filesize' => 'max_upload_file_size',
    'upload_tmp_dir' => 'extract_dir',
    'max_execution_time' => 'max_execution_time',
    'date.timezone' => 'server_timezone',
  ];

  /**
   * @var string
   */
  private const PROPERTIES_SEPARATOR = ':';

  /**
   * @var string
   */
  public const LOCAL_FS = 'local';

  /**
   * @var mixed[]
   */
  private $aProperties;

  /**
   * @var string
   */
  private $sConfigDir;

  /**
   * @param string $sVasDir chemin jusqu'au vas pour charger la configuration
   */
  public function __construct(string $sVasDir) {
    $this->sConfigDir = $sVasDir . '/src/Module';
    $this->loadProperties();
    $this->configurePhpIni();
  }

  /**
   * Parse le fichier Json de chaque module et l'ajoute au tableau des properties
   * @return void
   */
  private function loadProperties(): void {
    $aConfigDir = array_diff(scandir($this->sConfigDir), ['.', '..']);
    $aPropertiesTmp = [];
    $aPropertiesTmp['_module_list'] = $aConfigDir;
    $aPropertiesTmp['_user_properties'] = [];
    // Si lancer en 'cli' set _server_domain sur chaine vide
    $aPropertiesTmp['_server_domain'] = '';
    if (isset($_SERVER['HTTP_HOST'])) {
      $aPropertiesTmp['_server_domain'] = explode(':', $_SERVER['HTTP_HOST'])[0];
    }
    foreach ($aConfigDir as $sConfigDirName) {
      $sConfigDirPath = $this->sConfigDir . '/' . $sConfigDirName . '/conf';
      $sPropertieFile = $sConfigDirPath . '/properties.json';
      if (file_exists($sPropertieFile)) {
        $aJsonProperties = json_decode(file_get_contents($sPropertieFile), true);
        // Gestion _server_domain
        if (!empty($aJsonProperties['_server_domain'])) {
          $aPropertiesTmp['_server_domain'] = $aJsonProperties['_server_domain'];
        }
        $aPropertiesTmp = array_merge($aPropertiesTmp, $aJsonProperties);
      }
      $sPropertieFile = $sConfigDirPath . '/user_properties.json';
      if (file_exists($sPropertieFile)) {
        $aJsonProperties = json_decode(file_get_contents($sPropertieFile), true);
        $aPropertiesTmp['_user_properties'] = array_merge_recursive($aPropertiesTmp['_user_properties'], $aJsonProperties);
      }
    }
    $this->aProperties = $aPropertiesTmp;
  }

  /**
   * Fonction permettant de récupérer les properties
   * @param string|null $sRole Role utilisé pour demander les properties (public, vitis_user, vitis_admin....)
   * @return array Tableau contenant les différents properties
   */
  public function getProperties(?string $sRole = null): array{
    $aReturn = [];
    if (isset($sRole)) {
      $aList = $this->getReadablePropertiesKey($sRole);

      foreach ($aList as $sKey) {
        $aReturn[$sKey] = $this->get($sKey);
      }
    } else {
      foreach (array_keys($this->aProperties) as $sKey) {
        $aReturn[$sKey] = $this->get($sKey);
      }
    }

    return $aReturn;
  }

  /**
   * Retourne la liste des properties visibles
   * @param string $sRole de l'utilisateur
   * @return string[] liste des properties visibles pour le role
   */
  private function getReadablePropertiesKey(string $sRole): array{
    $aList = [];
    if (array_key_exists($sRole, $this->aProperties['_user_properties'])) {
      if (isset($this->aProperties['_user_properties'][$sRole]['read'])) {
        $aList = array_merge($aList, $this->aProperties['_user_properties'][$sRole]['read']);
      }
      $aList = array_merge($aList, $this->getWritablePropertiesKey($sRole));
    }
    return $aList;
  }

  /**
   * Retourne la liste des properties éditables
   * @param string $sRole de l'utilisateur
   * @return string[] liste des properties éditables pour le role
   */
  private function getWritablePropertiesKey(string $sRole, ?array $aCustomProperties = null): array{
    $aList = [];
    $aCustomProperties = $aCustomProperties ?? $this->aProperties['_user_properties'];

    if (array_key_exists($sRole, $aCustomProperties)) {
      if (isset($aCustomProperties[$sRole]['write'])) {
        $aList = array_merge($aList, $aCustomProperties[$sRole]['write']);
      }
    }

    return $aList;
  }

  /**
   * Retourne une properties en fonction de sa clé, si la properties dépend d'autres properties les dépendances sont remplacé
   * si une dépendance circulaire est détecté la fonction retourne null
   * @param string $sPropertiesName Clé de la properties
   * @param array $aParentReplacer Clé parente pour détecter les dépendance circulaire (par défaut : [])
   * @return mixed|null valeur de la properties demandé
   */
  public function get(string $sPropertiesName, array $aParentReplacer = []) {
    if (isset($this->aProperties[$sPropertiesName])) {
      // si la properties existe on récupére toutes les balise à remplacer
      // @prop(properties_key)
      $mPropertiesValue = $this->aProperties[$sPropertiesName];
      switch (gettype($mPropertiesValue)) {
        case 'string':
          // ajout de la properties demandé pour la détection de référence circulaire
          array_push($aParentReplacer, $sPropertiesName);
          $mPropertiesValue = $this->replacePropertiesInString($mPropertiesValue, $aParentReplacer);
          break;
      }
      return $mPropertiesValue;
    } else if (strpos($sPropertiesName, self::PROPERTIES_SEPARATOR) > -1) {
      $aPropertiesPath = explode(self::PROPERTIES_SEPARATOR, $sPropertiesName);
      $aMainProperties = $this->get($aPropertiesPath[0]);
      for ($i = 1; $i < count($aPropertiesPath); $i++) {
        if (isset($aMainProperties[$aPropertiesPath[$i]])) {
          $aMainProperties = $aMainProperties[$aPropertiesPath[$i]];
        } else {
          error_log('la properties correspondant au chemin ' . $sPropertiesName . ' n\'existe pas (' . $aPropertiesPath[$i] . ' introuvable)');
          return null;
        }
      }
      return $aMainProperties;
    } else {
      error_log('la properties ' . $sPropertiesName . ' n\'existe pas');
      return null;
    }
  }

  /**
   * Retourne une properties en fonction de sa clé, si la properties dépend d'autres properties les dépendances sont remplacé
   * si une dépendance circulaire est détecté la fonction retourne null
   * @param string $sString Chaine à évaluer pour remplacement des bind
   * @param array $aParentReplacer Clé parente pour détecter les dépendance circulaire (par défaut : [])
   * @return string|null chaine avec valeur remplacé
   */
  public function replacePropertiesInString(string $sString, array $aParentReplacer = []): ?string{
    $sRegexCapture = '/@prop\(([^)]*)\)/';
    $aMatches = [];
    preg_match_all($sRegexCapture, $sString, $aMatches);
    // dédoublonnage en cas d'appel de plusieur fois la même properties dans une properties
    $aReplacer = array_unique($aMatches[1]);
    // Remplacement des balises
    foreach ($aReplacer as $sPropertiesReplacer) {
      if (!in_array($sPropertiesReplacer, $aParentReplacer)) {
        $sReplacerValue = $this->get($sPropertiesReplacer, $aParentReplacer);
        $sString = str_replace('@prop(' . $sPropertiesReplacer . ')', $sReplacerValue, $sString);
      } else {
        // logger détection d'un appel circulaire
        error_log('circular detected');
        return null;
      }
    }
    return $sString;
  }

  /**
   * remplace les valeurs du php.ini par les properties lors du chargement du service
   * @return void
   */
  private function configurePhpIni(): void {
    foreach (self::PHP_INI_KEY_OVERWRITE as $sIniKey => $sPropertiesKey) {
      $sValue = $this->get($sPropertiesKey);
      if (!empty($sValue)) {
        ini_set($sIniKey, $this->get($sPropertiesKey));
      } else {
        error_log('cannot set php.ini key ' . $sIniKey . ' with properties ' . $sPropertiesKey . '(empty value)');
      }
    }
  }

  /**
   * Filtre les properties qu'on peut écrire dans le fichier Json d'un module et met à jour l'objet
   * @param string $sModuleName Module ou écrire les properties (vitis, gtf....)
   * @param array $aProperties Tableau des "properties"
   * @param string|null $sRole nom d'un role en base
   * @return array
   */
  private function extractEditablePropertiesForModule(string $sModuleName, array $aProperties, ?string $sRole = null): array{
    $aAuthorisedProperties = [];
    $aPropertiesToWrite = [];
    // Récupére la liste de tous properties
    $sPropertieFile = $this->sConfigDir . '/' . $sModuleName . '/conf/user_properties.json';
    if (file_exists($sPropertieFile)) {
      $aJsonProperties = json_decode(file_get_contents($sPropertieFile), true);
      if (isset($sRole)) {
        $aAuthorisedProperties = $this->getWritablePropertiesKey($sRole, $aJsonProperties);
      }
      foreach ($aJsonProperties as $sTmpRole => $oProperties) {
        $aAuthorisedProperties = array_merge($aAuthorisedProperties, $this->getWritablePropertiesKey($sTmpRole, $aJsonProperties));
      }
    }
    // Construit un tableau avec les properties trouvés dans la liste
    foreach ($aProperties as $sKey => $sPropertie) {
      if (in_array($sKey, $aAuthorisedProperties)) {
        $aPropertiesToWrite[$sKey] = $sPropertie;
      }
    }
    return $aPropertiesToWrite;
  }

  /**
   * Ecrit les properties dans le fichier Json d'un module et l'ajoute au tableau des properties
   * @param string $sModuleName Module ou écrire les properties (vitis, gtf....)
   * @param array $aProperties Tableau des "properties"
   * @param string|null $sRole nom d'un role en base
   * @return bool
   */
  public function dumpProperties(string $sModuleName, array $aProperties, ?string $sRole = null): bool {
    // Récupére les properties
    $sPropertieFile = $this->sConfigDir . '/' . $sModuleName . '/conf/properties.json';
    if (file_exists($sPropertieFile)) {
      $aPropertiesToWrite = json_decode(file_get_contents($sPropertieFile), true);
    } else {
      error_log('The file  ' . $sPropertieFile . ' not found');
      return false;
    }
    // Merge les properties à écrire
    $aPropertiesToWrite = array_merge(
      $aPropertiesToWrite,
      $this->extractEditablePropertiesForModule($sModuleName, $aProperties, $sRole)
    );
    // Ecrit les properties
    $sJsonProperties = json_encode($aPropertiesToWrite, JSON_PRETTY_PRINT | JSON_UNESCAPED_SLASHES);
    file_put_contents($sPropertieFile, $sJsonProperties);
    // recharge les properties pour être cohérent
    $this->loadProperties();

    return true;
  }

}
