<?php

namespace App\Module\Vitis\WebService\Service\Configuration;

use App\Module\Vitis\WebService\Service\Configuration\AbstractConfigurationLoader;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\VitisRequest;

/**
 * Permet d'accéder et de manipuler la configuration des routes.
 *
 * Les fichiers de configuration sont des fichiers Json spécifiques aux modules.
 * Ce sont les fichiers :
 *      -  src/Module/[Module_name]/route.json
 *
 * Ce fichier contient la définition des routes et la configuration pour accéder aux données
 *
 *
 * La route est contenu dans un paramètre variable qui sert de nom à la route (cf. partie "Exemple")
 * Arguments deny :
 *      c'est simplement une liste de nom de route à interdir (pour protéger des applications ou surcharger des routes)
 * Arguments routes :
 *      - route : chemin à ajouter derrière l'alias Apache, peut permettre de passer des paramètres au controller
 *          ex : https://127.0.0.1/rest_symfony/vitis/groups/1 --> /vitis/groups/{id}
 *      - controller : namespace complet du controller à utiliser pour cette route
 *      - function : nom de la fonction à utiliser dans le controller, si des paramètres sont présent dans la route la fonction doit les attendre
 *      - methods : liste des methodes HTTP pouvant utiliser la route et la configuration
 *      - type : utilisation de la route, trois valeurs possibles [orm, query, custom]
 *      - role_minimum : role_minimum pour utiliser la route (prut être un tableau de role)
 *      - filter : array associative, utiliser les privilèges comme clé, la structure de l'objet correspond à un filtre tel qu'utilisé dans l'application (realtion/operator, column/compare_operator/value), un ou logique sera fait entre les filtres si plusieurs droits se cumule, si il faut retirer tous les filtres pour un droit il suffit de mettre le filtre null
 *      - options : string doit référencer une clé existante dans la partie configuration
 *          - orm :
 *              - option.entity : namespace complet de la classe doctrine à utiliser pour requeter la base
 *          - query : TODO
 *          - custom : options n'est pas requis dans ce cas là et si vous l'utilisez aucune structure fixe n'est à prévoir.
 *      - options_by_role : objet permettant de surcharger les options en fonction d'un privilège, la première clé est un int qui représente la priorité (une suele ocnfiguration sera conserver)
 *          - X : priorité, contient un objet privilège : configuration
 *              -  nom_du_privilège : string, clé de la configuration à charger pour le role
 * Arguments configuration orm :
 *  - clé de la configuration :
 *    - entity : string namespace complet de l'entité
 *    - field_id : string identifiant dans l'orm
 *    - fields : array d'objet field orm
 *       - type : string type du champ pour l'API, voir customFields pour un listing  des types particuliers
 *       - name : nom du champs qui sera remonté au niveau de l'API
 *       - orm_get : string nom du getter
 *       - orm_set : string nom du setter (optionnel, ça permet de faire de champs en lecture seule)
 *       - orm_col : stirng nom du champ dans la classe orm (optionnel, à utiliser si différent de name pour faire amrcher les filtres)
 * Arguments configuration query :
 *  - clé de la configuration :
 *    - schema : string nom du schéma
 *    - table : string nom de la table
 *    - field_id : string identifiant
 *    - fields : array d'objet field query
 *       - type : string type du champ pour l'API, voir customFields pour un listing  des types particuliers
 *       - name : nom du champs en base (se sera aussi le nom remonté au niveau de l'API)
 *
 * Exemples :
 * {
 *   "vitis_get_group" : {
 *       "route" : "/vitis/groups/{id}",
 *       "requirements" : {
 *          "id" : '/[0-9]+/'
 *       }
 *       "controller" : "App\\Module\\Vitis\\WebService\\Controller\\GenericController",
 *       "function" : "getData",
 *       "methods" : ["GET"],
 *       "type" : "orm",
 *       "role_minimum" : ["vitis_user", "vitis_admin],
 *       "options" : "ma_config",
 *       "filter" : {
 *         "vitis_user" : {
 *           "relation": "AND",
 *           "operators": [{
 *             "column": "mon_champ",
 *             "compare_operator": "=",
 *             "value": "ma valeur"
 *           }]
 *         },
 *         "vitis_admin" : null
 *       },
 *       "options_by_role": {
 *         "0": {
 *           "vitis_admin": "ma_config"
 *         },
 *         "1": {
 *           "vitis_user": "ma_config_for_user"
 *         }
 *       }
 *   }
 * }
 *
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 * @author Alexandre Mironenkov <alexandre.mironenkov@veremes.com>
 */

class Route extends AbstractConfigurationLoader {
  /**
   * @var string
   */
  protected const CONFIGURATION_FILE_PATH = '@prop(vas_dir)/src/Module/@Module/WebService/routes.json';

  /**
   * @var string|null
   */
  private const NO_FILTER_VALUE = null;

  /**
   * @var array|null
   */
  private $aRouteConfiguration = null;

  /**
   * @var VitisRequest
   */
  private $oVitisRequest;

  /**
   * @param Properties $oPropertiesSrvc Service permettant d'accéder aux properties
   * @param VitisRequest|null $oRequest Service Vitis contenant la requête HTTP (optionnel)
   */
  public function __construct(Properties $oPropertiesSrvc, ?VitisRequest $oRequest = null) {
    $this->oPropertiesSrvc = $oPropertiesSrvc;
    $this->loadConfiguration();

    $this->denyRoutes();
    // si la request est set on charge la configuration associée à la route
    if (isset($oRequest)) {
      $this->oVitisRequest = $oRequest;
      $this->aRouteConfiguration = $this->getRequestParameters($oRequest->getRouteName());
    }
  }

  /**
   * filtre la configuration pour interdire des routes
   *
   * @return void
   */
  private function denyRoutes(): void {
    if (!array_key_exists('deny', $this->aConfiguration) || !array_key_exists('routes', $this->aConfiguration)) {
      // rien à filtrer
      return;
    }

    if (empty($this->aConfiguration['deny']) || empty($this->aConfiguration['routes'])) {
      // rien à filtrer
      return;
    }

    foreach ($this->aConfiguration['deny'] as $sDeny) {
      if (!is_string($sDeny)) {
        // si ce n'est pas un string on en tien pas compte
        continue;
      }

      if (array_key_exists($sDeny, $this->aConfiguration['routes'])) {
        // si la route à supprimer existe alors on la supprime complétement de la configuration
        unset($this->aConfiguration['routes'][$sDeny]);
      }
    }
  }

  /**
   * @param mixed $options Options de la route
   * @return void
   */
  public function setOptionsParameter($options): void {
    $this->aRouteConfiguration['options'] = $options;
  }

  /**
   * @param string $sRouteKey nom de la route dont il faut charger la configuration (si non fournit retourne la configuration chargé dans le constructeur)
   * @return array configuration de la route
   */
  public function getRequestParameters(?string $sRouteKey = null): array{
    if (empty($sRouteKey)) {
      // s'il n'y a pas de configuration pour la route on retourne tableau vide
      return $this->aRouteConfiguration ?? [];
    }

    return $this->resolveConfiguration($sRouteKey);
  }

  /**
   * @param string $sRouteKey nom de la route dont il faut charger la configuration (si non fournit retourne la configuration chargé dans le constructeur)
   * @return array configuration de la route
   * @throws \ErrorException si la configuration n'est pas conforme
   */
  private function resolveConfiguration(string $sRouteKey): array{
    if (!isset($this->aConfiguration['routes'])) {
      throw new \ErrorException('Cannot read this routes.json file, routes/configurations level is missing');
    }

    if ($this->oVitisRequest->getMethod() === 'OPTIONS' && !isset($this->aConfiguration['routes'][$sRouteKey])) {
      return [];
    }
    $aRouteConfiguration = $this->aConfiguration['routes'][$sRouteKey];
    // si pas d'options on retourne directement la configuration
    if (!isset($aRouteConfiguration['options'])) {
      return $aRouteConfiguration;
    }

    if (is_string($aRouteConfiguration['options'])) {
      $aConfiguration = $this->findConfigurationWithKey($aRouteConfiguration['options']);
      // si option et de type string et que la configuration existe on remplace
      if ($aConfiguration !== null) {
        $aRouteConfiguration['options'] = $aConfiguration;
      } else {
        throw new \ErrorException('Configuration of route ' . $sRouteKey . ' is not set');
      }
    } else {
      throw new \ErrorException('Configuration of route ' . $sRouteKey . ' is not in the configurations part');
    }

    // options des enfants lors de l'utilisation de @DiscriminatorMap
    if (array_key_exists('discriminator_options', $aRouteConfiguration) && is_array($aRouteConfiguration['discriminator_options'])) {
      foreach ($aRouteConfiguration['discriminator_options'] as $key => $value) {
        if (isset($this->aConfiguration['configurations'][$value])) {
          $aRouteConfiguration['discriminator_options'][$key] = $this->aConfiguration['configurations'][$value];
        } else {
          throw new \ErrorException('Configuration of route ' . $value . ' is not set');
        }
      }
    }

    return $aRouteConfiguration;
  }

  /**
   * Trouve une configuration à partie de sa clé
   *
   * @param  string     $sKey
   * @return array|null
   */
  public function findConfigurationWithKey(string $sKey): ?array{
    // si option et de type string et que la configuration existe on remplace
    if (isset($this->aConfiguration['configurations'][$sKey])) {
      return $this->aConfiguration['configurations'][$sKey];
    }
    return null;

  }

  /**
   * Recharge la configuration en fonction des privilège de l'utilisateur
   * @param array $aPrivileges liste des privileges
   * @return array configuration de la route
   */
  public function reloadConfigurationForUser(array $aPrivileges = []): array{
    if (empty($this->aRouteConfiguration) || empty($aPrivileges)) {
      return [];
    }

    // si pas d'options on retourne directement la configuration
    if (!isset($this->aRouteConfiguration['options_by_role'])) {
      return $this->aRouteConfiguration;
    }

    if (is_array($this->aRouteConfiguration['options_by_role'])) {
      $aTmpConfigurations = [];
      ksort($this->aRouteConfiguration['options_by_role']);
      foreach ($this->aRouteConfiguration['options_by_role'] as $iPriority => $aConfigurationForRole) {
        // le tableau est toujours de cette forme : sPrivilege => sConfigurationName
        $sPrivilege = array_key_first($aConfigurationForRole);
        // recherche de la configuration si le privilege est dans la liste des privileges
        if (in_array($sPrivilege, $aPrivileges)) {
          if (isset($this->aConfiguration['configurations'][$aConfigurationForRole[$sPrivilege]])) {
            $aTmpConfigurations = $this->aConfiguration['configurations'][$aConfigurationForRole[$sPrivilege]];
            break;
          }
        }
      }
      // si une configuration a été trouvé on la mémorise
      if (!empty($aTmpConfigurations)) {
        $this->aRouteConfiguration['options'] = $aTmpConfigurations;
      }
    }
    return $this->aRouteConfiguration;
  }

  /**
   * Permet de savoir si la route peut changer en fonction de l'utilisateur
   * @return bool true si la route peut changer en fonction de l'utilisateur false sinon
   */
  public function configurationCanChangeWithRole(): bool {
    // si pas d'options on retourne directement la configuration
    if (isset($this->aRouteConfiguration['options_by_role'])) {
      return is_array($this->aRouteConfiguration['options_by_role']);
    }

    if (isset($this->aRouteConfiguration['filter'])) {
      return true;
    }

    return false;
  }

  /**
   * Retourne la liste des champs dans un tableau (avec ou sans option, des type peuvent être ignorés)
   * @param bool $bWithOptions défini s'il faut retourner les options (par défaut oui)
   * @param array $aIgnoreType Liste de type à ignorer
   * @return array la liste des champs concerné
   */
  public function getFieldsList(bool $bWithOptions = true, array $aIgnoreType = []): array{
    $aFields = [];

    if (isset($this->aRouteConfiguration['options'])) {
      if (isset($this->aRouteConfiguration['options']['fields'])) {
        foreach ($this->aRouteConfiguration['options']['fields'] as $aFieldConfig) {
          if ((empty($aIgnoreType) || !in_array($aFieldConfig['type'], $aIgnoreType)) && $this->fieldExists($aFieldConfig)) {
            if ($bWithOptions) {
              $aFields[] = $aFieldConfig;
            } else {
              $aFields[] = $aFieldConfig['name'];
            }
          }
        }
      }
    }

    return $aFields;
  }
/**
 * Retourne le champ utilisé comme identifiant
 * @return string|null Le champ s'il est défini, null sinon
 */
  public function getFieldId(): ?string {
    if (!isset($this->aRouteConfiguration['options'])) {
      return null;
    }

    return $this->aRouteConfiguration['options']['field_id'] ?? null;
  }
  /**
   * Détermine si un champs existe en base
   * @param array $aField configuration du champ
   * @return bool false si le champs n'existe aps true sinon
   */
  public function fieldExists(array $aField): bool {
    return $aField['col_exists'] ?? true;
  }
  /**
   * Retourne les roles minimums de la requête s'il existe
   * @return array<string>|null Le tableau de role s'il est là, null sinon
   */
  public function getRoleAccess(): ?array{
    // si la clé est défini
    if (isset($this->aRouteConfiguration['role_minimum'])) {
      // si c'est une chaine de caractère on passe en array
      return is_array($this->aRouteConfiguration['role_minimum']) ? $this->aRouteConfiguration['role_minimum'] : [$this->aRouteConfiguration['role_minimum']];
    }
    return null;
  }

  /**
   * Retourne s'il faut controller le token de la route
   * @return boolean
   */
  public function checkUser(): bool {
    return $this->aRouteConfiguration['check_user'] ?? true;
  }
  /**
   * Retourne s'il faut controller l'origine de la route
   * @return boolean
   */
  public function checkOrigin(): bool {
    $mOrigin = $this->aRouteConfiguration['check_origin'] ?? false;

    if (\is_string($mOrigin)) {
      return true;
    }
    return false;
  }
  /**
   * Retourne la regexp pour vérifier l'origne de la route ou null si rien est défini
   * @return string|null
   */
  public function getOriginRegexp(): ?string {
    if (!$this->checkOrigin()) {
      return null;
    }

    return $this->aRouteConfiguration['check_origin'] ?? null;
  }
  /**
   * Retourne la requête Vitis
   * @return VitisRequest|null
   */
  public function getRequest(): ?VitisRequest {
    return $this->oVitisRequest;
  }

  /**
   * Récupère le filtre de la route pour l'utilisateur courant
   * @param array $aPrivilege Liste des privilège de l'utilisateur
   * @return array|null Retourne le filtre format array ou null si aucun filtre n'est défini
   * si le retour est sous forme d'array ce sera une array structurer comme un filtre classique
   * en cas de plusieurs filtre on appliquera un OU entre chaque
   */
  public function getFilterForUser(array $aPrivilege): ?array{
    // pas de filtre défini sur la route
    if (!isset($this->aRouteConfiguration['filter'])) {
      return null;
    }

    $aFilters = $this->aRouteConfiguration['filter'];
    $aOperators = [];

    foreach ($aFilters as $sPrivilege => $aFilterContent) {
      // si le privilège n'est pas dans l'objet on passe au suivant
      if (!in_array($sPrivilege, $aPrivilege)) {
        continue;
      }

      // détecte les privilèges de l'utilisateurs qui annule le filtrage
      if ($aFilterContent === self::NO_FILTER_VALUE) {
        return null;
      }
      // ajout du filtre avec les autres
      $aOperators[] = $aFilterContent;
    }
    // aucun filtre n'a été trouvé
    if (empty($aOperators)) {
      return null;
    }

    return [
      'relation' => 'OR',
      'operators' => $aOperators,
    ];
  }

}
