<?php

namespace App\Module\Vitis\WebService\Service\Crypto;

/**
 * Permet d'encrypter/hash un password en scram-sha-256 pour postgres
 * Basé sur un script python similaire : https://gist.github.com/jkatz/e0a1f52f66fa03b732945f6eb94d9c21
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */

class PgScram {
  /**
   * défini le nombre d'itération pour générer le sel
   * @var integer
   */
  private const NB_ITERATION = 4096;
  /**
   * défini la longueur du sel
   * @var integer
   */
  private const SALT_LENGTH = 16;
  /**
   * défini le nom de l'algo pour postgres
   * @var string
   */
  private const HASH_ALGO_NAME = 'SCRAM-SHA-256';
  /**
   * L'algo à appliquer à la focntion hmac/hash
   * @var string
   */
  private const HMAC_ALGO = 'sha256';
  /**
   * Fin du sel pour postgres
   * @var string
   */
  private const POSTGRES_END_SALT = '00000001';
  /**
   * Chaine à utiliser pour la clé client 
   * @var string
   */
  private const POSTGRES_STR_CLIENT_KEY = 'Client Key';
  /**
   * chaine à utiliser pour la clé serveur
   * @var string
   */
  private const POSTGRES_STR_SERVER_KEY = 'Server Key';

  /**
   * Hash le mot de passe pour postgres
   * @param string $sPassword Mot de passe à ecnrypter
   * @param string|null $sSalt hash à utiliser pour encoder (doit être une chaine binaire comme produit par random_bytes)
   * @return string
   */
  public function hashPgPassword(string $sPassword, ?string $sSalt = null): string {
    // récupération du sel ou génération
    $sSalt = (empty($sSalt)) ? random_bytes(self::SALT_LENGTH) : $sSalt;
    // Hash du mot de passe avec le sel
    $sSaltedPassword = $this->generateSaltedPassword($sPassword, $sSalt);
    // génération des clé à partir du hash du mot de passe
    $sClientKey = hash_hmac(self::HMAC_ALGO, self::POSTGRES_STR_CLIENT_KEY, $sSaltedPassword, true);
    $sStoredKey = hash(self::HMAC_ALGO, $sClientKey, true);
    $sServerKey = hash_hmac(self::HMAC_ALGO, self::POSTGRES_STR_SERVER_KEY, $sSaltedPassword, true);
    // retour de la chaine pour postgres
    //'SCRAM-SHA-256$4096:salt$stored_key:server_key'
    return self::HASH_ALGO_NAME . '$' . self::NB_ITERATION . ':' . base64_encode($sSalt) . '$' . 
            base64_encode($sStoredKey) . ':' . base64_encode($sServerKey);
  }

  /**
   * Généré une chaine aléatoire de longeur défini par constante
   * @param string $sPassword à encoder
   * @param string $sSalt hash à utiliser 
   * @return string
   */
  private function generateSaltedPassword(string $sPassword, string $sSalt): string {
    // normalisation du password
    $sNormalizedPwd = trim($sPassword);
    // La signature initial est le sel + une chaine de 32bit se terminant par 1 (volonté de postgres)
    $sSalt .= hex2bin(self::POSTGRES_END_SALT);
    $sPasswordSignature = hash_hmac(self::HMAC_ALGO, $sSalt, $sNormalizedPwd, true);
    // on stock le premier hashage
    $sFinalPassword = $sPasswordSignature;

    // On réitère l'opération pour arriver au nombre d'itération voulu
    // on commence à 1 car on a déjà fait un hashage avant
    for ($i = 1; $i < self::NB_ITERATION; $i++) {
      // on hash de nouveau la signature
      $sPasswordSignature = hash_hmac(self::HMAC_ALGO, $sPasswordSignature, $sNormalizedPwd, true);
      // on merge octet par octet la chaine
      $sFinalPassword = $this->byteStringXor($sFinalPassword, $sPasswordSignature);
    }
    return $sFinalPassword;
  }

  /**
   * Fait un OU exclusif (XOR) sur deux chaines et en sort un résultat "mergé"
   * @param string $text chaine principale
   * @param string $key chaine secondaire
   * @return string chaine mergée
   */
  private function byteStringXor(string $text, string $key): string {
    // The return value and a counter for use when xoring
    $i = 0;
    $iKey = strlen($key);
    $sReturn = '';

    // Loop through each character, xoring sequentially based on the key
    foreach (str_split($text) as $sCharacter) {
      // Get the character to use from our key
      $iCharacter = $i++ % $iKey;
      $sKey = ord($key[$iCharacter]);
      $sCharacter = ord($sCharacter);

      // Perform the XOR operation and get an ASCII character
      $sReturn .= chr($sCharacter ^ $sKey);
    }

    // Return the result
    return $sReturn;
  }
}
