<?php

namespace App\Module\Vitis\WebService\Service\Crypto;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Encoder\VitisBase64;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use Monolog\Logger;

/**
 * Permet de gérer notre fichier qui contient les clés offusquées utilisées pour le cryptage
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */

class VitisKeyManager {
  /**
   * @var string
   */
  private const KEY_PATH = '@prop(vas_dir)/.vitis_db';
  /**
   * @var string
   */
  private const KEY_ALGO = 'aes-256-cbc';

  /**
   * @var Logger
   */
  private $oErrorLogger;
  /**
   * @var Properties
   */
  private $oPropertiesSrvc;

  /**
   * @var array
   */
  private $oKeyStore = [];

  /**
   * @param Properties $PropertiesSrvc Service de lecture des properties
   * @param LoggerManager $oLogger Service d'accès aux Logger
   */
  public function __construct(Properties $PropertiesSrvc, LoggerManager $oLogger) {
    // récupère le log pour les erreurs de l'API
    $this->oErrorLogger = $oLogger->get('API_ERROR');
    $this->oPropertiesSrvc = $PropertiesSrvc;

    $sKeyPath = $this->oPropertiesSrvc->replacePropertiesInString(self::KEY_PATH);
    if (!\file_exists($sKeyPath)) {
      $this->oErrorLogger->error('Cannot find the encryption key file (' . $sKeyPath . ')');
    } else {
      $sKeyStore = \file_get_contents($sKeyPath);
      $this->oKeyStore = json_decode($sKeyStore, true);
    }
  }
  /**
   * @param  string      $sKeyname
   * @return string|null
   */
  public function getKeyByName(string $sKeyname): ?string {
    if (!$this->keyExists($sKeyname)) {
      return null;
    }

    return $this->decodeKeyValue($sKeyname);
  }

  /**
   * détermine si la clé existe dans le coffre
   *
   * @param  string $sKeyname
   * @return bool
   */
  private function keyExists(string $sKeyname): bool {
    $sTrueKeyName = $this->encodeKeyName($sKeyname);

    return array_key_exists($sTrueKeyName, $this->oKeyStore);
  }

  /**
   * @param  string $sKeyname
   * @return string
   */
  private function encodeKeyName(string $sKeyname): string {
    return VitisBase64::base64encode(hash('md5', $sKeyname));
  }
  /**
   * @param  string $sKeyname
   * @return string
   */
  private function decodeKeyValue(string $sKeyname): ?string {
    $sTrueKeyName = $this->encodeKeyName($sKeyname);
    $sSecretKey = $this->secretKeyAdapter($sKeyname . '_vitis');
    return $this->readAESKey($this->oKeyStore[$sTrueKeyName], $sSecretKey);
  }

  /**
   * décrypte une donnée en AES à partir de la clé et du vecteur initial
   *
   * @param  string $data donnée à décrypter (json en b64)
   * @param  string $passphrase clé pour lire la donnée
   * @return string
   */
  private function readAESKey(string $data, string $passphrase): ?string {
    // lecture de l'objet crypté de donnée
    $json = json_decode(VitisBase64::base64decode($data), true);
    // récupération du vecteur d'initialisation
    $iv = VitisBase64::base64decode($json['iv']);
    // décryptage des données
    $encrypted_64 = $json['data'];
    $data_encrypted = VitisBase64::base64decode($encrypted_64);
    $decrypted = openssl_decrypt($data_encrypted, self::KEY_ALGO, $passphrase, OPENSSL_RAW_DATA, $iv);
    return ($decrypted == false) ? null : $decrypted;
  }

  /**
   * Adapte la clé à la bonne longueur (32char en ascii 64 en hexa)
   *
   * @param  string $passphrase clé à adapter
   * @return string clé adaptée
   */
  private function secretKeyAdapter(string $passphrase): string {

    if (strlen($passphrase) < 32) {
      while (strlen($passphrase) < 32) {
        $passphrase .= '0';
      }
      return $passphrase;
    }

    if (strlen($passphrase) > 32) {
      return substr($passphrase, 0, 32);
    }

    return $passphrase;
  }

}
