<?php

namespace App\Module\Vitis\WebService\Service\DataAccessor;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\CommonAccessorFunctions;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\CustomFieldsManager;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\Serializer\VitisSerializer;
use App\Module\Vitis\WebService\Service\Token\JweTokenManager;
use Doctrine\ORM\EntityManager;
use Monolog\Logger;

/**
 * Service abstrait pour les accés aux bases de données
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */

abstract class AbstractDataAccessor implements DataAccessorInterface {
  /**
   * @var string
   */
  protected const ENTITY_ALIAS = '';
  /**
   * @var string
   */
  protected const MODE = '';
  /**
   * @var string
   */
  protected const LOGGER_KEY = 'SQL_ERROR';

  /**
   * @var EntityManager
   */
  protected $oEntityManager;
  /**
   * @var array
   */
  protected $oOptions;
  /**
   * @var VitisSerializer
   */
  protected $oSerializer;
  /**
   * @var Route
   */
  protected $oRouteConfigurator;
  /**
   * @var Properties
   */
  protected $oPropertiesSrvc;
  /**
   * @var CustomFieldsManager
   */
  protected $oCustomFieldsManager;
  /**
   * @var Logger
   */
  protected $oErrorLogger;
  /**
   * @var CommonAccessorFunctions
   */
  protected $oAccessorFunc;

  /**
   * @param EntityManager $oEntityManager Gestionnaire d'entité Doctrine
   * @param Route $oRouteConfigurator service de configuration de la route
   * @param VitisSerializer $oSerializer Objet permettant la désearialization dans une couche ORM
   * @param LoggerManager $oLogger Service d'accès aux logger
   * @param Properties $oProperties Service d'accès aux properties
   * @param JweTokenManager $oJweTokenManager accès au Token et encryptage
   */
  public function __construct(EntityManager $oEntityManager, Route $oRouteConfigurator, VitisSerializer $oSerializer, LoggerManager $oLogger, Properties $oProperties, JweTokenManager $oJweTokenManager) {
    $this->oEntityManager = $oEntityManager;
    $this->oRouteConfigurator = $oRouteConfigurator;
    $this->oPropertiesSrvc = $oProperties;
    $this->oSerializer = $oSerializer;
    // récupération des options
    $aRouteConfig = $oRouteConfigurator->getRequestParameters();
    if (isset($aRouteConfig['options'])) {
      $this->oOptions = $aRouteConfig['options'];
    }

    $this->oCustomFieldsManager = new CustomFieldsManager($this->oPropertiesSrvc, $this->oRouteConfigurator, $oLogger, $oJweTokenManager, $oEntityManager);

    $this->oErrorLogger = $oLogger->get(static::LOGGER_KEY);

    $this->oAccessorFunc = new CommonAccessorFunctions(static::ENTITY_ALIAS, $this->oRouteConfigurator, $this->oErrorLogger, $oEntityManager, $this->oCustomFieldsManager, static::MODE);

  }

  /**
   * défini un filtre à concaténer au moment du decode
   * @param array|null $aFilter
   */
  public function setCommonFilter(?array $aFilter = null): void {
    $this->oAccessorFunc->setCommonFilter($aFilter);
  }

  /**
   * Permet de forcer de nouvelles options avant d'utiliser un accessor
   * @param array $oOptions nouvel config à appliquer
   * @return void
   */
  public function setOptions(array $oOptions): void {
    $this->oOptions = $oOptions;
    // si une liste de champs existe on configure les fonctionnalité commune avec cette liste de champ
    if (!empty($oOptions['fields'])){
      $this->oRouteConfigurator->setOptionsParameter($oOptions);
      $this->oAccessorFunc->setFieldsConfig($oOptions['fields']);
      $this->oCustomFieldsManager->setOptions($oOptions);
    }
  }

  /**
   * Provient de DataAccessor
   * {@inheritdoc}
   */
  public function getEntity($id, array $aData = [], bool $useCustomFields = true): array{
    return [
      'status' => 0,
      'message' => 'getEntity not implemented for ' . get_class($this),
    ];
  }
  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function getEntities(array $aData, bool $useCustomFields = true): array{
    return [
      'status' => 0,
      'message' => 'getEntities not implemented for ' . get_class($this),
    ];
  }
  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function insertEntity(string $sData, bool $useCustomFields = true): array{
    return [
      'status' => 0,
      'message' => 'insertEntity not implemented for ' . get_class($this),
    ];
  }
  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function insertEntities(array $aData, bool $useCustomFields = true): array{
    return [
      'status' => 0,
      'message' => 'insertEntities not implemented for ' . get_class($this),
    ];
  }
  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function updateEntity($id, string $sData, bool $useCustomFields = true): array{
    return [
      'status' => 0,
      'message' => 'updateEntity not implemented for ' . get_class($this),
    ];
  }
  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function deleteEntity($id, bool $useCustomFields = true): array{
    return [
      'status' => 0,
      'message' => 'deleteEntity not implemented for ' . get_class($this),
    ];
  }
  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function deleteEntities(array $aIds, bool $useCustomFields = true): array{
    return [
      'status' => 0,
      'message' => 'deleteEntities not implemented for ' . get_class($this),
    ];
  }

  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function updateEntities(array $aIds, string $sData, bool $useCustomFields = true): array{
    return [
      'status' => 0,
      'message' => 'updateEntities not implemented for ' . get_class($this),
    ];
  }

}
