<?php

namespace App\Module\Vitis\WebService\Service\DataAccessor;

use App\Module\Vitis\WebService\Entity\User;
use App\Module\Vitis\WebService\EventListener\MappingListener;
use App\Module\Vitis\WebService\EventListener\VitisEntityListener;
use App\Module\Vitis\WebService\Query\SortableNullValueWalker;
use App\Module\Vitis\WebService\Service\Configuration\Module;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\Crypto\PgScram;
use App\Module\Vitis\WebService\Service\Logger\CustomSqlLogger;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\Serializer\VitisSerializer;
use App\Module\Vitis\WebService\Service\Token\JweTokenManager;
use Doctrine\Common\EventManager;
use Doctrine\DBAL\Logging\Middleware;
use Doctrine\DBAL\Types\Type;
use Doctrine\ORM\Configuration;
use Doctrine\ORM\EntityManager;
use Doctrine\ORM\ORMSetup;
use Doctrine\ORM\Query;
use Jsor\Doctrine\PostGIS\Event\ORMSchemaEventSubscriber;
use Jsor\Doctrine\PostGIS\Functions\Configurator;
use Monolog\Logger;

/**
 * Permet d'instancier un objet EntityManager pour utiliser Doctrine
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */

class DatabaseConnection {
  /**
   * @var string
   */
  public const SGBD = 'pdo_pgsql';
  /**
   * @var string
   */
  private const DOCTRINE_PROXY_PATH = '@prop(extract_dir)/doctrine';

  /**
   * @var Properties
   */
  private $oPropertiesSrvc;
  /**
   * @var string[]
   */
  private $aEntityPath = [];
  /**
   * @var array
   */
  private $aModuleConfiguration = [];
  /**
   * @var Route
   */
  private $oRouteConfigurator;
  /**
   * @var EntityManager|null
   */
  private $oUserEntityManager;
  /**
   * @var EntityManager|null
   */
  private $oAdminEntityManager;
  /**
   * @var LoggerManager
   */
  private $oLoggerManager;
  /**
   * @var Logger
   */
  private $oSqlLogger;
  /**
   * @var Logger
   */
  private $oErrorLogger;
  /**
   * @var Logger
   */
  private $oConnectionLogger;
  /**
   * @var VitisSerializer
   */
  private $oSerializer;
  /**
   * @var JweTokenManager
   */
  private $oJweTokenManager;
  /**
   * @var VitisEntityListener
   */
  private $oVitisEntityListener;
  /**
   * @param Properties $PropertiesSrvc Un objet pour accéder aux properties de l'application
   * @param Module $oModuleConfigurator service d'accès à la configuration des modules, seule la partie DBAL_Functions est utile
   * ex :             "DBAL_Functions" : {
   *                      "CastGeography" : {
   *                         "class" : "App\\Module\\Vitis\\WebService\\Query\\Functions\\CastGeography",
   *                         "type" : "string"
   *                     }
   *                  }
   * @param Route $oRouteConfigurator service d'accès aux informations de la route
   * @param LoggerManager $oLogger accès au logs
   * @param JweTokenManager $oJweTokenManager accès au Token et encryptage
   * @param VitisEntityListener $oVitisEntityListener permet de charger des services dans les entités
   */
  public function __construct(Properties $PropertiesSrvc, Module $oModuleConfigurator, Route $oRouteConfigurator, LoggerManager $oLogger, JweTokenManager $oJweTokenManager, VitisEntityListener $oVitisEntityListener) {
    // Sauvegarde du service de properties
    $this->oPropertiesSrvc = $PropertiesSrvc;
    $this->oLoggerManager = $oLogger;
    $this->oSqlLogger = $oLogger->get('SQL_INFO');
    $this->oErrorLogger = $oLogger->get('API_ERROR');
    $this->oConnectionLogger = $oLogger->get('CONNEXION_INFO');
    $this->oSerializer = new VitisSerializer();
    $this->oJweTokenManager = $oJweTokenManager;
    $this->oRouteConfigurator = $oRouteConfigurator;
    $this->oVitisEntityListener = $oVitisEntityListener;
    // récupération de la liste de modules
    $aModuleList = $this->oPropertiesSrvc->get('_module_list');
    // Chargement des fichiers Doctrines
    foreach ($aModuleList as $sModuleName) {
      $this->aEntityPath[] = $this->oPropertiesSrvc->get('vas_dir') . '/src/Module/' . $sModuleName . '/WebService/Entity';
    }
    $this->aModuleConfiguration = $oModuleConfigurator->getConfiguration();
  }

  /**
   * Retourne un EntityManager pour Doctrine connecté avec l'utilisateur courant
   *
   * @param string $sUserName login de l'utilisateur
   * @param string $sPassword mot de passe de l'utilisateur
   *
   * @return EntityManager EntityManager pour l'utilisateur connecté
   */
  public function createUserEntityManager(string $sUserName, string $sPassword): EntityManager {
    $this->oUserEntityManager = (!empty($this->oUserEntityManager)) ? $this->oUserEntityManager : $this->createUserEntityManagerInstance($sUserName, $sPassword);
    return $this->oUserEntityManager;
  }

  /**
   * Retourne un EntityManager pour Doctrine connecté avec l'utilisateur courant
   *
   * @return EntityManager|null EntityManager pour l'utilisateur connecté
   */
  public function getUserEntityManager(): ?EntityManager{
    $this->oUserEntityManager = (!empty($this->oUserEntityManager)) ? $this->oUserEntityManager : $this->createUserEntityManagerFromToken();
    return $this->oUserEntityManager;
  }

  /**
   * Créer un EntityManager pour Doctrine connecté avec l'utilisateur en paramètre
   *
   * @param string $sUserName login de l'utilisateur
   * @param string $sPassword mot de passe de l'utilisateur
   *
   * @return EntityManager EntityManager pour l'utilisateur passé en paramètre
   */
  public function createUserEntityManagerInstance(string $sUserName, string $sPassword): EntityManager {
    $dbParams = $this->getLocalDbArrayParams($sUserName, $sPassword);
    $oEventManager = $this->createEventManager();
    $oConfig = $this->createConfiguration($sUserName, $oEventManager);
    $oEm = EntityManager::create($dbParams, $oConfig, $oEventManager);
    $this->oVitisEntityListener->setObjectManager($oEm);
    return $oEm;
  }

  /**
   * retourne un array conforme pour crée un entity manager Postgres
   * @param string $sUserName Login
   * @param string $sPassword Mot de passe à utiliser
   * @return array Configuration sous forme d'array assoc
   */
  private function getLocalDbArrayParams(string $sUserName, string $sPassword): array{
    $sSSlMode = $this->oPropertiesSrvc->get('db_ssl_mode');
    $aDbParam = [
      'driver' => self::SGBD,
      'host' => $this->oPropertiesSrvc->get('db_server'),
      'port' => $this->oPropertiesSrvc->get('db_port'),
      'dbname' => $this->oPropertiesSrvc->get('db_name'),
      'user' => $sUserName,
      'password' => $sPassword,
      'memory' => true,
      'sslmode' => $sSSlMode,
    ];

    // gestion des certficats clients si nécessaire
    // see : https://www.postgresql.org/docs/9.1/libpq-ssl.html
    if ($sSSlMode === 'verify-ca' || $sSSlMode === 'verify-full') {
      $sRootCertPath = $this->oPropertiesSrvc->get('db_ssl_root_cert');
      $sCertPath = $this->oPropertiesSrvc->get('db_ssl_cert');
      $sCertKeyPath = $this->oPropertiesSrvc->get('db_ssl_key');
      $sCrlCertPath = $this->oPropertiesSrvc->get('db_ssl_crl');

      if (!empty($sRootCertPath)) {
        $aDbParam['sslrootcert'] = $sRootCertPath;
      }
      if (!empty($sCertPath)) {
        $aDbParam['sslcert'] = $sCertPath;
      }
      if (!empty($sCertKeyPath)) {
        $aDbParam['sslkey'] = $sCertKeyPath;
      }
      if (!empty($sCrlCertPath)) {
        $aDbParam['sslcrl'] = $sCrlCertPath;
      }
    }

    return $aDbParam;
  }

  /**
   * Retourne un EntityManager pour Doctrine connecté avec l'utilisateur u_vitis (properties db_supersuer)
   *
   * @return EntityManager EntityManager pour l'utilisateur u_vitis (properties db_supersuer)
   */
  public function getAdminEntityManager(): EntityManager {
    $this->oAdminEntityManager = (!empty($this->oAdminEntityManager)) ? $this->oAdminEntityManager : $this->createAdminEntityManager();
    return $this->oAdminEntityManager;
  }

  /**
   * Créer un EntityManager pour Doctrine connecté avec l'utilisateur u_vitis (properties db_supersuer)
   *
   * @return EntityManager EntityManager pour l'utilisateur u_vitis (properties db_supersuer)
   */
  private function createAdminEntityManager(): EntityManager {
    $sUsername = $this->oPropertiesSrvc->get('db_superuser');
    $sPassword = $this->oPropertiesSrvc->get('db_password');
    $dbParams = $this->getLocalDbArrayParams($sUsername, $sPassword);
    $oEventManager = $this->createEventManager();
    $oConfig = $this->createConfiguration($sUsername, $oEventManager);
    $oEm = EntityManager::create($dbParams, $oConfig, $oEventManager);
    $this->oVitisEntityListener->setObjectManager($oEm);
    return $oEm;
  }

  /**
   * Créer un EventManager pour utiliser PostGis et binder les properties
   *
   * @return EventManager EventManager configuré
   */
  private function createEventManager(): EventManager {
    // Création d'un event manager pour parser les annotations
    $oEventManager = new EventManager();
    // L'instanciation du mapping listener est obligatoire même s'ul n'est pas utilisé
    /* @phpstan-ignore-next-line */
    $oMappingListener = new MappingListener($this->oPropertiesSrvc, $oEventManager);
    $oEventManager->addEventSubscriber(new ORMSchemaEventSubscriber());

    return $oEventManager;
  }

  /**
   * Créer la configuration pour un entityManager
   *
   * @param string $sUserName nom de l'utilisateur courrant
   * @param EventManager $oEventManager gestionnaire d'evenement doctrine
   * @return Configuration EventManager configuré
   */
  private function createConfiguration(string $sUserName, EventManager $oEventManager): Configuration {
    // surcharge du doctrine proxy path que windows a du mal sinon
    $sDoctrineProxyPath = $this->oPropertiesSrvc->replacePropertiesInString(self::DOCTRINE_PROXY_PATH);
    if (!file_exists($sDoctrineProxyPath)) {
      mkdir($sDoctrineProxyPath, 0777, true);
    }
    // Instanciation de la base du gestionnaire de base de donnée
    $oConfig = ORMSetup::createAnnotationMetadataConfiguration($this->aEntityPath, true, $sDoctrineProxyPath);
    // Enregistrement des fonctions SQL si besoin
    $aModuleConfig = $this->aModuleConfiguration;
    if (isset($aModuleConfig['DBAL_Functions'])) {
      foreach ($aModuleConfig['DBAL_Functions'] as $sName => $aFunctionConfig) {
        switch ($aFunctionConfig['type']) {
        case 'string':
          $oConfig->addCustomStringFunction($sName, $aFunctionConfig['class']);
          break;
        }
      }
    }
    // ajoute des type ou en surcharge
    if (isset($aModuleConfig['DBAL_Types'])) {
      foreach ($aModuleConfig['DBAL_Types'] as $sName => $aTypeConfig) {
        if (Type::hasType($sName)) {
          Type::overrideType($sName, $aTypeConfig['class']);
        } else {
          Type::addType($sName, $aTypeConfig['class']);
        }
      }
    }

    // Ajout du walker SQL pour les tri en NULLS LAST
    $oConfig->setDefaultQueryHint(Query::HINT_CUSTOM_OUTPUT_WALKER, SortableNullValueWalker::class);

    // récupération des fonctions de postgis du bundle jsor/doctrine-postgis
    Configurator::configure($oConfig);

    // Chargement de l'eventListener pour charger des services dans les entités
    $this->oVitisEntityListener->subscribeEvent($oEventManager);
    $oConfig->getEntityListenerResolver()->register($this->oVitisEntityListener);
    $oSqlMiddleware = new Middleware(new CustomSqlLogger($this->oSqlLogger, $this->oErrorLogger, $sUserName, $this->oPropertiesSrvc));
    $oConfig->setMiddlewares([$oSqlMiddleware]);

    // ajout de l'event manager dans le normalizer pour invoquer l'event de denormalization
    $this->oSerializer->getVitisNormalizer()->setEventManager($oEventManager);
    return $oConfig;
  }

  /**
   * Génère un mot de passe hasher pour postgres selon l'algorithme défini en properties :
   *      - PLAIN : Retourne le mot de passe en brut
   *      - MD5 : Retourne un hash md5 du mot de passe 'md5' . md5($sPassword . $sUserName)
   *      - SCARM-SHA256 : à venir
   *
   * @param string $sUserName login de l'utilisateur
   * @param string $sPassword mot de passe de l'utilisateur
   *
   * @return string|null hash du mot de passe ou null si l'algorithme n'est pas défini.
   */
  public function getHashedPassword(string $sUserName, string $sPassword): ?string{
    $sHashedPassword = null;
    $sEncryptMethod = $this->oPropertiesSrvc->get('db_password_hash');
    switch ($sEncryptMethod) {
    case 'PLAIN':
    case 'plain':
    case 'password':
      $sHashedPassword = $sPassword;
      break;
    case 'MD5':
    case 'md5':
      $sHashedPassword = 'md5' . md5($sPassword . $sUserName);
      break;
    case 'SCRAM-SHA256':
    case 'scram-sha-256':
      $oScram = new PgScram();
      $sHashedPassword = $oScram->hashPgPassword($sPassword);
      break;
    default:
      $this->oErrorLogger->warning('This password hash method doesn\'t exist : ' . $sEncryptMethod);
      break;
    }
    return $sHashedPassword;
  }

  /**
   * Génère une instance de data accessor permettant d'accéder aux données
   * @param EntityManager|null $oEntityManager connection à la base de donnée (par défaut celui d el'utilisateur courant)
   *
   * @return AbstractDataAccessor|null service d'accès à la base de données ou null.
   */
  public function initializeDataAccessor(?EntityManager $oEntityManager = null): ?AbstractDataAccessor{
    // initialisation
    $oDataAccessor = null;
    $aFilters = null;
    if (empty($oEntityManager)) {
      // récupération d'un entityManager pour l'utilisateur courant
      $oEntityManager = $this->getUserEntityManager();
      // charge une configuration spécifique à l'utilisateur si nécessaire
      if ($this->oRouteConfigurator->configurationCanChangeWithRole()) {
        $aUser = $this->getCurrentUser();
        // surcharge les options pour l'utilisateur si nécessaire
        $this->oRouteConfigurator->reloadConfigurationForUser($aUser->getRoles());
        // récupère des filtres s'ils sont présent
        $aFilters = $this->oRouteConfigurator->getFilterForUser($aUser->getRoles());
      }
    }
    // récupération des options de la route
    $aOptions = $this->oRouteConfigurator->getRequestParameters();

    // instanciation du bon dataAccessor (ORM ou Query)
    if (isset($aOptions['type'])) {
      switch ($aOptions['type']) {
      case 'orm':
        // retouche du normalizer pour gérer les objectManager en insert
        $this->oVitisEntityListener->setObjectManager($oEntityManager);
        //$this->oSerializer->getVitisNormalizer()->setObjectManager($oEntityManager);
        $oDataAccessor = new OrmAccessor($oEntityManager, $this->oRouteConfigurator, $this->oSerializer, $this->oLoggerManager, $this->oPropertiesSrvc, $this->oJweTokenManager);
        break;
      case 'query':
        $oDataAccessor = new QueryAccessor($oEntityManager, $this->oRouteConfigurator, $this->oSerializer, $this->oLoggerManager, $this->oPropertiesSrvc, $this->oJweTokenManager);
        break;
      }
    }

    if (!empty($aFilters) && !empty($oDataAccessor)) {
      $oDataAccessor->setCommonFilter($aFilters);
    }

    return $oDataAccessor;
  }
/**
 * Génère une instance de QueryAccessor permettant d'accéder aux données à travers le querybuilder natif
 * L'entityManager de l'utilisateur sera utiliser
 * @return QueryAccessor|null service d'accès à la base de données en natif ou null.
 */
  public function getQueryAccessor(?EntityManager $oEntityManager = null): ?QueryAccessor{
    // récupération d'un entityManager
    $oEntityManager = (!empty($oEntityManager)) ? $oEntityManager : $this->getUserEntityManager();
    if ($oEntityManager === null) {
      return null;
    }
    return new QueryAccessor($oEntityManager, $this->oRouteConfigurator, $this->oSerializer, $this->oLoggerManager, $this->oPropertiesSrvc, $this->oJweTokenManager);
  }

  /**
   * Génère une instance de OrmAccessor permettant d'accéder aux données à travers le querybuilder natif
   * L'entityManager de l'utilisateur sera utiliser
   * @return OrmAccessor|null service d'accès à la base de données en natif ou null.
   */
  public function getOrmAccessor(?EntityManager $oEntityManager = null): ?OrmAccessor{
    // récupération d'un entityManager
    $oEntityManager = (!empty($oEntityManager)) ? $oEntityManager : $this->getUserEntityManager();
    if ($oEntityManager === null) {
      return null;
    }
    // retouche du normalizer pour gérer les objectManager en insert
    $this->oVitisEntityListener->setObjectManager($oEntityManager);

    return new OrmAccessor($oEntityManager, $this->oRouteConfigurator, $this->oSerializer, $this->oLoggerManager, $this->oPropertiesSrvc, $this->oJweTokenManager);
  }

  /**
   * Retourne l'entité User pour l'utilisateur connecté
   * @param string $sToken Token de l'utilisateur connecté
   *
   * @return User|null Entité de l'utilisateur courant s'il existe, sinon retourne null.
   */
  public function getCurrentUser(string $sToken = ''): ?User{
    $sToken = (empty($sToken)) ? $this->oRouteConfigurator->getRequest()->getToken() : $sToken;

    if (empty($sToken)) {
      $this->oConnectionLogger->error('This request has not token', ['token' => $sToken]);
      return null;
    }
    // initialisation de l'entityManager depuis le token de l'utilisateur
    if ($this->createUserEntityManagerFromToken($sToken) === null) {
      return null;
    }
    // récupérations des infos du token
    $aTokenJWEDecrypt = $this->oJweTokenManager->decryptToken($sToken);

    $aUser = $this->oUserEntityManager->getRepository(User::class)
      ->findOneBy([
        'login' => $aTokenJWEDecrypt['data']['user']['login'],
      ]);

    if (isset($aUser)) {
      $this->oConnectionLogger->debug('Connection of ' . $aTokenJWEDecrypt['data']['user']['login'], ['token' => $sToken, 'user' => $aTokenJWEDecrypt['data']['user']]);
      return $aUser;
    }
    $this->oConnectionLogger->error('User not found for login ' . $aTokenJWEDecrypt['data']['user']['login'], ['token' => $sToken, 'user' => $aTokenJWEDecrypt['data']['user']]);
    return null;
  }

/**
 * Retourne l'EntityManager pour l'utilisateur connecté
 * @param string $sToken Token de l'utilisateur connecté
 *
 * @return EntityManager|null EntityManager ou null en cas d'erreur avec le token.
 */

  private function createUserEntityManagerFromToken(string $sToken = ''): ?EntityManager{
    // récupérations des infos du token
    $aTokenJWEDecrypt = $this->getUserFromToken($sToken);
    if (empty($aTokenJWEDecrypt)) {
      return null;
    } else {
      return $this->createUserEntityManager($aTokenJWEDecrypt['data']['user']['login'], $aTokenJWEDecrypt['data']['password']);
    }
  }

  /**
   * Récupère le contenu décrypté du token
   * @param string $sToken Token de l'utilisateur connecté
   *
   * @return Array|null Array ou null en cas d'erreur avec le token.
   */
  private function getUserFromToken(string $sToken = '') {
    $sToken = (empty($sToken)) ? $this->oRouteConfigurator->getRequest()->getToken() : $sToken;
    if (empty($sToken)) {
      $this->oConnectionLogger->error('This request has not token', ['token' => $sToken]);
      return null;
    }
    // récupérations des infos du token
    $aTokenJWEDecrypt = $this->oJweTokenManager->decryptToken($sToken);
    if (!empty($aTokenJWEDecrypt)) {
      return $aTokenJWEDecrypt;
    }

    $this->oConnectionLogger->error('Bad token', ['token' => $sToken]);
    return null;
  }

  /**
   * Retourne true si on arrive a connecter l'utilisateur
   *
   * @param string $sUserName login de l'utilisateur
   * @param string $sPassword mot de passe de l'utilisateur
   *
   * @return bool
   */
  public function checkUserConnection(string $sUserName, string $sPassword): bool {
    $userEntityManager = $this->createUserEntityManagerInstance($sUserName, $sPassword);
    $aUser = $userEntityManager->getRepository(User::class)
      ->findOneBy([
        'login' => $sUserName,
      ]);
    if (!empty($aUser)) {
      return true;
    }
    return false;
  }

  /**
   * Créer un EntityManager pour Doctrine connecté à une base de données externe ou une autre base de données locale
   *
   * @param string $sUserName login de l'utilisateur
   * @param string $sPassword mot de passe de l'utilisateur
   * @param string $sDbName nom de la base de données
   * @param string $sSGDB système de la base de données
   * @param string $sDbHost serveur de la base de données
   * @param string $sDbPort port de connexion à la base de données
   * @param string|null $sCharset Encodage de la base à utiliser pour la connexion (oracle only)
   * @return EntityManager EntityManager
   */
  private function createExternalEntityManagerInstance(string $sUserName, string $sPassword, string $sDbName, $sSGDB, string $sDbHost, string $sDbPort, ?string $sCharset = null): EntityManager {
    // modification lié à la rétrocompatibilité avec le studio de l'ancien framework
    switch ($sSGDB) {
    case 'pgsql':
      $sSGDB = 'pdo_pgsql';
      break;
    case 'oci':
      $sSGDB = 'pdo_oci';
      break;
    default:
      // par défaut on considère que la vairable est compatible avec pdo
      break;
    }

    $dbParams = [
      'driver' => $sSGDB,
      'host' => $sDbHost,
      'port' => $sDbPort,
      'dbname' => $sDbName,
      'user' => $sUserName,
      'password' => $sPassword,
      'memory' => true,
    ];

    if ($sSGDB == 'pdo_oci' && !empty($sCharset)) {
      $dbParams['charset'] = $sCharset;
    }

    $oEventManager = $this->createEventManager();
    $oConfig = $this->createConfiguration($sUserName, $oEventManager);
    $oEm = EntityManager::create($dbParams, $oConfig, $oEventManager);
    $this->oVitisEntityListener->setObjectManager($oEm);
    return $oEm;
  }

  /**
   * Retourne un EntityManager pour Doctrine connecté à une base de données externe
   *
   * @param string|null $sDbName nom de la base de données
   * @param string|null $sUserName login de l'utilisateur
   * @param string|null $sPassword mot de passe de l'utilisateur
   * @param string|null $sSGDB système de la base de données
   * @param string|null $sDbHost serveur de la base de données
   * @param string|null $sDbPort port de connexion à la base de données
   * @param string|null $sCharset encodage de la base (oracle only)
   * @return EntityManager|null EntityManager
   */
  public function createExternalEntityManager(?string $sDbName = null, ?string $sUserName = null, ?string $sPassword = null, ?string $sSGDB = null, ?string $sDbHost = null, ?string $sDbPort = null, ?string $sCharset = null): ?EntityManager {
    if (!isset($sUserName) && !isset($sPassword) && !isset($sSGDB) && !isset($sDbHost) && !isset($sDbPort)) {
      $aToken = $this->getUserFromToken();
      if (empty($aToken)) {
        return null;
      }

      $sUserName = $aToken['data']['user']['login'];
      $sPassword = $aToken['data']['password'];
      $sSGDB = self::SGBD;
      $sDbHost = $this->oPropertiesSrvc->get('db_server');
      $sDbPort = $this->oPropertiesSrvc->get('db_port');
      // return $this->getUserEntityManager();
    }

    return $this->createExternalEntityManagerInstance($sUserName, $sPassword, $sDbName, $sSGDB, $sDbHost, $sDbPort, $sCharset);
  }

  /**
   * Génère une instance de OracleAccessor permettant d'accéder aux données à travers le querybuilder natif
   * L'entityManager de l'utilisateur sera utiliser
   * @return OracleAccessor|null service d'accès à la base de données en natif ou null.
   */
  public function getOracleAccessor(?EntityManager $oEntityManager = null): ?OracleAccessor{
    // récupération d'un entityManager
    $oEntityManager = (!empty($oEntityManager)) ? $oEntityManager : $this->getUserEntityManager();
    if ($oEntityManager === null) {
      return null;
    }
    return new OracleAccessor($oEntityManager, $this->oRouteConfigurator, $this->oSerializer, $this->oLoggerManager, $this->oPropertiesSrvc, $this->oJweTokenManager);
  }
}
