<?php

namespace App\Module\Vitis\WebService\Service\DataAccessor\Fields;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\AbstractCustomField;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\FieldInterface;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use Doctrine\ORM\EntityManager;

/**
 * Classe pour gérer un champ de type array.
 * Met à jour une table de jointure.
 *
 *  Configuration du champ
 *    "type"          (string) : Toujours array
 *    "name"          (string) : Nom du champ
 *    "join"          (array)  :
 *    "join.schema"   (string) : nom du schema de la table de jointure
 *    "join.table"    (string) : nom de la table d ejointure,
 *    "join.field_id" (string) : nom de la colonne pour accueillir la valeur de l'id de la table
 *    "join.joined_id"(string) : nom de la colonne pour accueillir l'id de la table jointe (valeur)
 *
 *  exemple :
 *       {
 *         "type": "array",
 *         "name" : "users",
 *         "join" : {
 *           "schema" : "s_vitis",
 *           "table" : "user_group",
 *           "field_id" : "group_id",
 *           "joined_id" : "user_id"
 *         }
 *       }
 *
 *  "mdd" associé à l'exemple :
 *
 *    Entité User            Join Many-to-Many             Entité Group
 *  ________________        ____________________          _______________
 * |  s_vitis.user  |      | s_vitis.user_group |        | s_vitis.group |
 * |________________|      |____________________|        |_______________|
 * |user_id       PK|<-----|user_id        PK|FK|   |--->|group_id     PK|
 * |...             |      |group_id       PK|FK|---|    |...            |
 * |________________|      |____________________|        |_______________|
 *
 * La surcouche doctrine n'est pas obligatoire
 *
 * @author Anthony Borghi   <anthony.borghi@veremes.com>
 */
class ArrayField extends AbstractCustomField implements FieldInterface {
  /**
   * @var string
   * Valeur du champs type dans route.json
   */
  public const TYPE = 'array';

  /**
   * @var EntityManager
   */
  private $oEntityManager;

  /**
   * {@inheritdoc}
   */
  public function __construct(Properties $oPropertiesSrvc, Route $oRoute, LoggerManager $oLoggerManager, EntityManager $oEntityManager) {
    parent::__construct($oPropertiesSrvc, $oRoute, $oLoggerManager);
    $this->oEntityManager = $oEntityManager;
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    // si la donnée ne vient pas d'une vue il probable que le champ ne soit pas rempli
    // dans ce cas là on lance une requête sur la table pour récupérer la donnée
    if (!isset($aEntity[$aFieldOptions['name']])) {
      $aEntity[$aFieldOptions['name']] = $this->getDataFromJoinTable($aEntity, $aRequestData, $aFieldOptions);
    }
    return $aEntity;
  }

  /**
   * {@inheritdoc}
   */
  public function insertFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    // avant insert rien à faire au contraire
    // on ne peut pas joindre sir l'entité principale n'existe pas
    if ($this->oRouteConfig->fieldExists($aFieldOptions)) {
      unset($aEntity[$aFieldOptions['name']]);
    }
    return $aEntity;
  }
  /**
   * {@inheritdoc}
   */
  public function afterInsertFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    // si le champ n'est pas set
    if (!isset($aRequestData[$aFieldOptions['name']])) {
      return $aEntity;
    }
    // si le champ n'est pas un tableau
    if (!is_array($aRequestData[$aFieldOptions['name']])) {
      return $aEntity;
    }
    // si le champ ne contient qu'un élément vide
    if (count($aRequestData[$aFieldOptions['name']]) === 1) {
      if (!isset($aRequestData[$aFieldOptions['name']][0])) {
        return $aEntity;
      }
      // si l'élément vaut chaine vide
      if ($aRequestData[$aFieldOptions['name']][0] === '') {
        return $aEntity;
      }
    }

    if (!$this->insertJoinTable($aEntity, $aRequestData, $aFieldOptions)) {
      //si l'update n'a pas pû être fait on vide le champ
      $aEntity[$aFieldOptions['name']] = null;
    }
    return $aEntity;
  }
  /**
   * {@inheritdoc}
   */
  public function updateFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    if (!isset($aRequestData[$aFieldOptions['name']])) {
      return $aEntity;
    }
    // delete
    $aEntity = $this->deleteFieldValue($aEntity, $aRequestData, $aFieldOptions);
    // insert
    return $this->afterInsertFieldValue($aEntity, $aRequestData, $aFieldOptions);
  }
  /**
   * {@inheritdoc}
   */
  public function updateMassiveFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    // update
    return $this->updateFieldValue($aEntity, $aRequestData, $aFieldOptions);
  }

  /**
   * {@inheritdoc}
   */
  public function deleteFieldValue(array $aEntity, array $aRequestData, array $aFieldOptions): array{
    if ($this->clearJoinTable($aEntity, $aRequestData, $aFieldOptions)) {
      //si l'update a pû être fait on vide le champ
      $aEntity[$aFieldOptions['name']] = null;
    }
    return $aEntity;
  }

  /**
   * vide une table de jointure sur un id
   * @param array $aEntity données de la base
   * @param array $aRequestData données de la requête
   * @param array $aFieldOptions configuration du champ
   * @return bool true en cas de succès false en cas d'erreur
   */
  private function clearJoinTable(array $aEntity, array $aRequestData, array $aFieldOptions): bool {
    // inisitalisation du logger
    $oLogger = $this->oLoggerManager->get('SQL_ERROR');
    //initialisation de la connexion à la base
    $oConnexion = $this->oEntityManager->getConnection();
    // récupération de la config globale de la route
    $aRequestOptions = $this->oRouteConfig->getRequestParameters();
    $sTableName = $aFieldOptions['join']['schema'] . '.' . $aFieldOptions['join']['table'];
    $sTableName = $this->oProperties->replacePropertiesInString($sTableName);
    try {
      $qb = $oConnexion->createQueryBuilder();
      $qb->delete($sTableName)
        ->where($aFieldOptions['join']['field_id'] . ' = :mIdInWhere')
        ->setParameter('mIdInWhere', $aEntity[$aRequestOptions['options']['field_id']]);

      // exécution du delete
      $qb->execute();
    } catch (\Exception $e) {
      $oLogger->error('Cannot clear join table for field : ' . $aFieldOptions['name'] . ' (' . $e->getMessage() . ')', [$aEntity, $aFieldOptions, $e]);
      return false;
    }
    return true;
  }
  /**
   * remplit une table de jointure
   * @param array $aEntity données de la base
   * @param array $aRequestData données de la requête
   * @param array $aFieldOptions configuration du champ
   * @return bool true en cas de succès false en cas d'erreur
   */
  private function insertJoinTable(array $aEntity, array $aRequestData, array $aFieldOptions): bool {
    // inisitalisation du logger
    $oLogger = $this->oLoggerManager->get('SQL_ERROR');
    //initialisation de la connexion à la base
    $oConnexion = $this->oEntityManager->getConnection();
    // récupération de la config globale de la route
    $aRequestOptions = $this->oRouteConfig->getRequestParameters();
    // préparation d'un objet values
    $aValues = [
      $aFieldOptions['join']['field_id'] => $aEntity[$aRequestOptions['options']['field_id']],
    ];

    $sTableName = $aFieldOptions['join']['schema'] . '.' . $aFieldOptions['join']['table'];
    $sTableName = $this->oProperties->replacePropertiesInString($sTableName);

    try {
      // pour chaque valeur dans l'array
      foreach ($aRequestData[$aFieldOptions['name']] as $mValue) {
        $aValues[$aFieldOptions['join']['joined_id']] = "'$mValue'";

        // initialisation query builder natif
        $qb = $oConnexion->createQueryBuilder();
        $qb->insert($sTableName);
        $qb->values($aValues);
        // exécution de l'insert
        $qb->execute();
      }
    } catch (\Exception $e) {
      $oLogger->error('Cannot fill join table for field : ' . $aFieldOptions['name'] . ' (' . $e->getMessage() . ')', [$aEntity, $aFieldOptions, $e]);
      return false;
    }
    return true;
  }

  /**
   * remplit une table de jointure
   * @param array $aEntity données de la base
   * @param array $aRequestData données de la requête
   * @param array $aFieldOptions configuration du champ
   * @return array|null
   */
  private function getDataFromJoinTable(array $aEntity, array $aRequestData, array $aFieldOptions): ?array{
    // inisitalisation du logger
    $oLogger = $this->oLoggerManager->get('SQL_ERROR');
    //initialisation de la connexion à la base
    $oConnexion = $this->oEntityManager->getConnection();
    // récupération de la config globale de la route
    $aRequestOptions = $this->oRouteConfig->getRequestParameters();
    $aRows = [];
    $sTableName = $aFieldOptions['join']['schema'] . '.' . $aFieldOptions['join']['table'];
    $sTableName = $this->oProperties->replacePropertiesInString($sTableName);

    try {
      $qb = $oConnexion->createQueryBuilder();
      $qb->select($aFieldOptions['join']['joined_id'])
        ->from($sTableName)
        ->where($aFieldOptions['join']['field_id'] . ' = :mIdInWhere')
        ->setParameter('mIdInWhere', $aEntity[$aRequestOptions['options']['field_id']]);

      // exécution du delete
      $aResult = $qb->executeQuery()->fetchAllAssociative();
      foreach ($aResult as $Joined) {
        $aRows[] = $Joined[$aFieldOptions['join']['joined_id']];
      }
    } catch (\Exception $e) {
      $oLogger->error('Cannot select from join table for field : ' . $aFieldOptions['name'] . ' (' . $e->getMessage() . ')', [$aEntity, $aFieldOptions, $e]);
      return null;
    }
    return $aRows;
  }

  /**
   * {@inheritdoc}
   */
  public function fieldIsSet(array $aEntity, array $aFieldOptions): bool {
    return ($this->oRouteConfig->fieldExists($aFieldOptions) && parent::fieldIsSet($aEntity, $aFieldOptions)) || (!$this->oRouteConfig->fieldExists($aFieldOptions));
  }
}
