<?php

namespace App\Module\Vitis\WebService\Service\DataAccessor;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\Serializer\VitisSerializer;
use App\Module\Vitis\WebService\Service\Token\JweTokenManager;
use Doctrine\DBAL\Query\QueryBuilder;
use Doctrine\ORM\EntityManager;
use Monolog\Logger;

/**
 * Permet d'accéder aux données d'une connexion Oracle
 *
 * @author Ludivine Premont   <ludivine.premont@veremes.com>
 */
class OracleAccessor extends AbstractQueryAccessor implements DataAccessorInterface {
  /**
   * @var string
   */
  protected const ENTITY_ALIAS = 'GENERIC_ROW';

  /**
   * Le nom du compteur des enregistrements
   * @var string
   */
  private $sRecordsCount;

  /**
   * @param EntityManager $oEntityManager Gestionnaire d'entité Doctrine
   * @param Route $oRouteConfigurator service de configuration de la route
   * @param VitisSerializer $oSerializer Objet permettant la désearialization dans une couche ORM
   * @param LoggerManager $oLogger Service d'accès aux logger
   * @param Properties $oProperties Service d'accès aux properties
   * @param JweTokenManager $oJweTokenManager accès au Token et encryptage
   */
  public function __construct(EntityManager $oEntityManager, Route $oRouteConfigurator, VitisSerializer $oSerializer, LoggerManager $oLogger, Properties $oProperties, JweTokenManager $oJweTokenManager) {
    parent::__construct($oEntityManager, $oRouteConfigurator, $oSerializer, $oLogger, $oProperties, $oJweTokenManager);

    $this->sRecordsCount = self::ENTITY_ALIAS;
    if (isset($this->oOptions['field_id'])) {
      $this->sRecordsCount .= '.' . $this->oOptions['field_id'];
    }
  }

  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function getEntities(array $aData, bool $useCustomFields = true): array{
    // si le nom de la table n'existe pas on essaie de le recharger depuis le soptions
    if (!isset($this->sTableName)) {
      $this->reloadTableNameFromOptions();
    }

    $aEntities = null;
    $qb = $this->createQueryBuilder();

    $bDistinct = (isset($aData['distinct'])) ? ($aData['distinct'] === 'true') : false;

    // gestion de la liste des champs
    $sAttributes = (isset($aData['attributs'])) ? $aData['attributs'] : '';
    $sSelectedFields = $this->oAccessorFunc->attributeManager($sAttributes);

    // gestion de distinct
    if ($bDistinct) {
      $sSelectedFields = 'DISTINCT ' . $sSelectedFields;
    }
    // initialisation query builder
    $qb->select($sSelectedFields)
      ->from($this->sTableName, self::ENTITY_ALIAS);

    // order_by
    if (isset($aData['order_by'])) {
      if (in_array($aData['order_by'], $this->oRouteConfigurator->getFieldsList(false))) {
        $sSortOrder = (isset($aData['sort_order'])) ? $aData['sort_order'] : 'ASC';
        $qb->orderBy(self::ENTITY_ALIAS . '.' . $aData['order_by'], $sSortOrder);
      }
    }

    // filter
    $aData['filter'] = $aData['filter'] ?? null;
    if ($this->oAccessorFunc->needFilter($aData['filter'])) {
      $aFilterReturn = $this->oAccessorFunc->prepareAndDecodeFilter($aData['filter']);
      if (!empty($aFilterReturn['request'])) {
        $qb->where($aFilterReturn['request']);
        foreach ($aFilterReturn['params'] as $sKey => $aValue) {
          if (isset($aValue['type'])) {
            $qb->setParameter($sKey, $aValue['value'], $aValue['type']);
          } else {
            $qb->setParameter($sKey, $aValue['value']);
          }
        }
      }
    }

    // offset
    if (isset($aData['offset']) && $aData['offset'] > 0) {
      $qb->setFirstResult($aData['offset']);
    }

    // limit
    if (isset($aData['limit']) && $aData['limit'] > 0) {
      $qb->setMaxResults($aData['limit']);
    }

    $aEntities = (array) $qb->executeQuery()->fetchAllAssociative();

    return [
      'data' => $aEntities,
      'list_count' => count($aEntities),
      'status' => 1,
    ];
  }

  /**
   * Permet de récupérer les tables d'un schema d'une base de données locale
   * @param mixed $aData
   * @return array
   */
  public function getTables($aData) {
    $qb = $this->oEntityManager->getConnection()->createQueryBuilder();

    $qb->select('tables.OBJECT_NAME AS "table_name"')
      ->from('ALL_OBJECTS', 'tables')
      ->where($qb->expr()->eq('tables.OWNER', ':OWNER'))
      ->andWhere($qb->expr()->or(
        $qb->expr()->eq('tables.OBJECT_TYPE', ':OBJECT_TABLE'),
        $qb->expr()->eq('tables.OBJECT_TYPE', ':OBJECT_VIEW')
      )
      )
      ->setParameter('OWNER', $aData['schema'])
      ->setParameter('OBJECT_TABLE', 'TABLE')
      ->setParameter('OBJECT_VIEW', 'VIEW');

    if (isset($aData['order_by'])) {
      $qb->orderBy('"' . $aData['order_by'] . '"');
    }

    try {
      $oTables = $qb->executeQuery()->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('GenericQueries : can\'t get tables: ' . $e->getMessage());
      return ['status' => 0];
    }

    return $oTables;
  }

  /**
   * Permet de récupérer les columns d'une table pour la configuration
   * @param mixed $aData
   * @return array
   */
  public function getFieldsFromTable($aData) {
    $qb = $this->oEntityManager->getConnection()->createQueryBuilder();

    $qb->select('columns.COLUMN_NAME AS "column_name", columns.DATA_TYPE AS "data_type"')
      ->from('ALL_TAB_COLUMNS', 'columns')
      ->where($qb->expr()->eq('columns.OWNER', ':owner'))
      ->andWhere($qb->expr()->eq('columns.TABLE_NAME', ':table_name'))
      ->setParameter('owner', $aData['schema'])
      ->setParameter('table_name', $aData['table'])
      ->orderBy('"column_name"');

    try {
      $oTables = $qb->executeQuery()->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('GenericQueries : can\'t get tables: ' . $e->getMessage());
      return ['status' => 0];
    }

    $response = ['status' => 1, 'data' => []];
    foreach ($oTables as $oTable) {
      $aField = ['name' => $oTable['column_name']];

      switch ($oTable['data_type']) {
        case 'INTEGER':
        case 'SHORTINTEGER':
        case 'LONGINTEGER':
        case 'DECIMAL':
        case 'SHORTDECIMAL':
        case 'NUMBER':
          $aField['type'] = 'integer';
          break;
        case 'BOOLEAN':
          $aField['type'] = 'boolean';
          break;
        default:
          $aField['type'] = 'string';
          break;
      }

      $response['data'][] = $aField;
    }

    return $response;
  }

  public function createQueryBuilder(): QueryBuilder {
    return $this->oEntityManager->getConnection()->createQueryBuilder();
  }
}
