<?php

namespace App\Module\Vitis\WebService\Service\DataAccessor;

use App\Module\Vitis\WebService\Repository\GenericRepository;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\Normalizer\VitisNormalizer;
use App\Module\Vitis\WebService\Service\Serializer\VitisSerializer;
use App\Module\Vitis\WebService\Service\Token\JweTokenManager;
use Doctrine\ORM\EntityManager;
use Monolog\Logger;
use Symfony\Component\Serializer\Normalizer\AbstractNormalizer;

/**
 * Permet d'accéder aux données via une entité Doctrine
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 * @author Ludivine Premont   <ludivine.premont@veremes.com>
 */
class OrmAccessor extends AbstractDataAccessor implements DataAccessorInterface {
  /**
   * @var string
   */
  protected const ENTITY_ALIAS = 'generic_entity';
  /**
   * @var string
   */
  protected const MODE = 'ORM';
  /**
   * @var string
   */
  protected const LOGGER_KEY = 'DQL_ERROR';

  /**
   * @param EntityManager $oEntityManager Gestionnaire d'entité Doctrine
   * @param Route $oRouteConfigurator service de configuration de la route
   * @param VitisSerializer $oSerializer Objet permettant la désearialization dans une couche ORM
   * @param LoggerManager $oLogger Service d'accès aux logger
   * @param Properties $oProperties Service d'accès aux properties
   * @param JweTokenManager $oJweTokenManager accès au Token et encryptage
   */
  public function __construct(EntityManager $oEntityManager, Route $oRouteConfigurator, VitisSerializer $oSerializer, LoggerManager $oLogger, Properties $oProperties, JweTokenManager $oJweTokenManager) {
    parent::__construct($oEntityManager, $oRouteConfigurator, $oSerializer, $oLogger, $oProperties, $oJweTokenManager);
  }

  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function getEntity($id, array $aData = [], bool $useCustomFields = true): array{
    $oEntity = null;
    // Récupération des erreurs SQL
    try {
      $this->oEntityManager->clear();

      $oRepository = $this->oEntityManager->getRepository($this->oOptions['entity']);
      if (!$oRepository instanceof GenericRepository) {
        throw new \Exception($this->oOptions['entity'] . ' Repository is not a GenericRepository', 1);
      }
      $oEntity = $oRepository->findById($id);

    } catch (\Exception $e) {
      $this->oErrorLogger->error('ORM Accessor : ' . $e->getMessage());
      return ['status' => 0];
    }
    // si aucune entité n'est trouvé le status vaut 0 sinon 1
    if (empty($oEntity)) {
      $this->oErrorLogger->error('ORM Accessor : entity is empty');
      return ['status' => 0];
    } else {
      // passage en array filtré
      $aContext = [];
      if (isset($aData['attributs'])) {
        $aContext[AbstractNormalizer::ATTRIBUTES] = explode('|', $aData['attributs']);
      }
      if (isset($aData['distinct'])) {
        $aContext[VitisNormalizer::DISTINCT] = ($aData['distinct'] === 'true');
      }
      // on ajoute les données de base en cas de retouche sur les getter
      $oEntity[0]->setRequestData($aData);
      $aEntity = $oEntity[0]->jsonSerialize();
      $aEntity = VitisNormalizer::attributesFiltration($aEntity, $aContext);
      // gestion des champs customs
      if ($useCustomFields) {
        $aEntity = $this->oCustomFieldsManager->getCustomFields($aEntity, $aData);
      }
      return [
        'data' => $aEntity,
        'status' => 1,
      ];
    }
  }

  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function getEntities(array $aData, bool $useCustomFields = true): array{
    $aEntities = null;
    // Récupération des erreurs SQL
    try {
      $this->oEntityManager->clear();
      $oRepository = $this->oEntityManager->getRepository($this->oOptions['entity']);
      if (!$oRepository instanceof GenericRepository) {
        throw new \Exception($this->oOptions['entity'] . ' Repository is not a GenericRepository', 1);
      }
      $aEntities = $oRepository->search($aData, $this->oAccessorFunc);
    } catch (\Exception $e) {
      $this->oErrorLogger->error('ORM Accessor : ' . $e->getMessage());
      return ['status' => 0];
    }
    // si aucune entité n'est trouvé le status vaut 0 sinon 1
    if (empty($aEntities)) {
      $this->oErrorLogger->error('ORM Accessor : entity is empty');
      return ['status' => 0];
    } else {
      // préparation du contexte pour filtré
      $aContext = [];
      if (isset($aData['attributs'])) {
        $aContext[AbstractNormalizer::ATTRIBUTES] = explode('|', $aData['attributs']);
      }
      if (isset($aData['distinct'])) {
        $aContext[VitisNormalizer::DISTINCT] = ($aData['distinct'] === 'true');
      }
      // récupération des champs custom
      foreach ($aEntities['data'] as &$aEntity) {
        // on ajoute les données de base en cas de retouche sur les getter
        $aEntity->setRequestData($aData);
        $aEntity = $aEntity->jsonSerialize();
        $aEntity = VitisNormalizer::attributesFiltration($aEntity, $aContext);
        // gestion des champs customs
        if ($useCustomFields) {
          $aEntity = $this->oCustomFieldsManager->getCustomFields($aEntity, $aData);
        }
      }
      return [
        'data' => (isset($aEntities['data'])) ? $aEntities['data'] : null,
        'total_row_number' => (isset($aEntities['total_row_selected'])) ? $aEntities['total_row_selected'] : 0,
        'list_count' => (isset($aEntities['list_count'])) ? $aEntities['list_count'] : 0,
        'status' => $aEntities['status'],
      ];
    }
  }

  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function insertEntity(string $sData, bool $useCustomFields = true): array{
    $oEntity = null;
    try {
      // création de l'entité
      $oEntity = $this->oSerializer->deserialize($sData, $this->oOptions['entity']);
      $this->oEntityManager->persist($oEntity);
      if ($useCustomFields) {
        $aEntity = $this->oCustomFieldsManager->insertCustomFields($oEntity->jsonSerialize(), json_decode($sData, true));
        // mise à jour de l'entité si il y a eu des modifications
        if ($this->oCustomFieldsManager->dataIsModified()) {
          $this->oSerializer->deserialize(json_encode($aEntity), $this->oOptions['entity'], ['object_to_populate' => $oEntity]);
        }
      }
      // Sauvegarde en base de donnée
      $this->oEntityManager->persist($oEntity);
      $this->oEntityManager->flush();
      // rafraichissement de l'entité en cas de valeur set par la base de données
      $this->oEntityManager->refresh($oEntity);
    } catch (\Exception $e) {
      $this->oErrorLogger->error('ORM Accessor : ' . $e->getMessage());
      return ['status' => 0];
    }

    // gestion des champs custom
    if ($useCustomFields) {
      $aEntity = $this->oCustomFieldsManager->afterInsertCustomFields($oEntity->jsonSerialize(), json_decode($sData, true));
      // mise à jour de l'entité si il y a eu des modifications
      if ($this->oCustomFieldsManager->dataIsModified()) {
        $this->oSerializer->deserialize(json_encode($aEntity), $this->oOptions['entity'], ['object_to_populate' => $oEntity]);
        // rafraichissement de l'entité
      }
      $this->oEntityManager->refresh($oEntity);
    }
    return [
      'data' => $oEntity,
      'status' => 1,
    ];
  }

  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function updateEntity($id, string $sData, bool $useCustomFields = true): array{
    $oEntity = null;
    // récupération de l'entité à mettre à jour
    $oEntity = $this->oEntityManager->find($this->oOptions["entity"], $id);

    if (!empty($oEntity)) {
      // Mise à jour de l'entité avec les données de la requête
      $this->oSerializer->deserialize($sData, $this->oOptions['entity'], ['object_to_populate' => $oEntity]);
      // gestion des champs custom
      if ($useCustomFields) {
        $aData = json_decode($sData, true);
        $aEntity = $this->oCustomFieldsManager->updateCustomFields($oEntity->jsonSerialize(), $aData);
        $this->oSerializer->deserialize(json_encode($aEntity), $this->oOptions['entity'], ['object_to_populate' => $oEntity]);
      }
      // Sauvegarde en base de donnée
      $this->oEntityManager->persist($oEntity);
      $this->oEntityManager->flush();
      // rafraichissement en cas de modifications interne à la base
      $this->oEntityManager->refresh($oEntity);
      return [
        'data' => $oEntity,
        'status' => 1,
      ];
    } else {
      $this->oErrorLogger->error('ORM Accessor : entity is empty');
      return ['status' => 0];
    }
  }

  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function deleteEntity($id, bool $useCustomFields = true): array{
    $oEntity = null;
    // récupération de l'entité à mettre à jour
    $oEntity = $this->oEntityManager->find($this->oOptions['entity'], $id);

    if (!empty($oEntity)) {
      // gestion des champs custom
      if ($useCustomFields) {
        $aEntity = $this->oCustomFieldsManager->deleteCustomFields($oEntity->jsonSerialize(), []);
        $this->oSerializer->deserialize(json_encode($aEntity), $this->oOptions['entity'], ['object_to_populate' => $oEntity]);
      }
      // Supprime de la base de donnée
      try {
        $oRepository = $this->oEntityManager->getRepository($this->oOptions['entity']);
        if (!$oRepository instanceof GenericRepository) {
          throw new \Exception($this->oOptions['entity'] . ' Repository is not a GenericRepository', 1);
        }
        if ($oRepository->removeItemById($id)) {
          // retourne l'entité supprimée
          return [
            'data' => $id,
            'status' => 1,
          ];
        } else {
          throw new \Exception('Cannot remove item ' . $this->oOptions['entity'] . ' with removeItemById', 1);
        }
      } catch (\Exception $e) {
        $this->oErrorLogger->error('ORM Accessor : ' . $e->getMessage());
        return ['status' => 0];
      }

    } else {
      return ['status' => 0];
    }
  }

  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function deleteEntities(array $aIds, bool $useCustomFields = true): array{
    $aEntities = null;
    try {
      $oRepository = $this->oEntityManager->getRepository($this->oOptions['entity']);
      if (!$oRepository instanceof GenericRepository) {
        throw new \Exception($this->oOptions['entity'] . ' Repository is not a GenericRepository', 1);
      }
      $aEntities = $oRepository->removeList($aIds);
    } catch (\Exception $e) {
      $this->oErrorLogger->error('ORM Accessor : ' . $e->getMessage());
      return ['status' => 0];
    }
    // si aucune entité n'est trouvé le status vaut 0 sinon 1
    if (empty($aEntities)) {
      $this->oErrorLogger->error('ORM Accessor : entity is empty');
      return ['status' => 0];
    } else {
      $aResponse = [
        'status' => $aEntities['status'],
      ];
      if (isset($aEntities['total_row_removed'])) {
        $aResponse['total_row_removed'] = $aEntities['total_row_removed'];
        if (count($aIds) == $aEntities['total_row_removed']) {
          $aResponse['data'] = $aEntities['data'];
        } else {
          $aResponse['data'] = [];
          foreach ($aIds as $mId) {
            $aEntity = $this->getEntity($mId);
            if ($aEntity['status'] === 0) {
              $aResponse['data'][] = $mId;
            }
          }
        }
        if ($useCustomFields) {
          $sFieldId = $this->oRouteConfigurator->getFieldId();
          foreach ($aResponse['data'] as $mId) {
            // si la suppression a été effectuée
            if (isset($sFieldId)) {
              $this->oCustomFieldsManager->deleteCustomFields([$sFieldId => $mId], []);
            }
          }
        }
      }
      return $aResponse;
    }
  }

  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function updateEntities(array $aIds, string $sData, bool $useCustomFields = true): array{
    // Récupère les entités en fonction des ids selectionné
    $filter = [
      'column' => $this->oOptions['field_id'],
      'compare_operator' => 'IN',
      'value' => $aIds,
    ];
    $aEntities = $this->getEntities(['filter' => json_encode($filter), 'attributs' => $this->oOptions['field_id']]);

    // Créer un nouveaux tableau avec les ids qui existe en base
    $aVerifiedIds = [];
    foreach ($aEntities['data'] as $value) {
      $aVerifiedIds[] = $value[$this->oOptions['field_id']];
    }

    // Récupère les données et les formates
    $aData = json_decode($sData, true);
    // gestion des champs custom
    if ($useCustomFields) {
      foreach ($aVerifiedIds as $id) {
        $aData[$this->oOptions['field_id']] = $id;
        $aData = $this->oCustomFieldsManager->updateMassiveCustomFields($aData, $aData);
        // on ne veut surtout pas mettre à jour l'id
        unset($aData[$this->oOptions['field_id']]);
      }
    }
    $aProtectedValues = $this->oAccessorFunc->sqlValuesExtractor($aData);

    try {
      $oRepository = $this->oEntityManager->getRepository($this->oOptions['entity']);
      if (!$oRepository instanceof GenericRepository) {
        throw new \Exception($this->oOptions['entity'] . ' Repository is not a GenericRepository', 1);
      }
      $aStatus = $oRepository->updateList($aVerifiedIds, $aProtectedValues, $this->oAccessorFunc);
    } catch (\Exception $e) {
      $this->oErrorLogger->error('ORM Accessor : ' . $e->getMessage());
      return ['status' => 0];
    }

    if ($aStatus['status'] == 0) {
      $this->oErrorLogger->error('ORM Accessor : ' . $aStatus['message']);
      return ['status' => 0];
    } elseif ($aStatus['status'] == 1) {
      $filter['value'] = $aVerifiedIds;

      return $this->getEntities(['filter' => json_encode($filter)]);
    }

    return ['status' => 0];
  }

  /**
   * Créé plusieurs entités.
   * @param array $aData Tableau des données à inséré, chaque entré dans le tableau correspond à une entité
   * @param bool $useCustomFields définit s'il faut charger les champs custom
   * @return array retour la liste des entités si l'insertion s'est bien passé, le statut seulement sinon
   */
  public function insertEntities(array $aData, bool $useCustomFields = true): array{
    $aEntitiesAndData = [];
    try {
      $this->oEntityManager->getConnection()->beginTransaction(); // suspend auto-commit

      for ($i = 0; $i < count($aData); $i++) {
        // création de l'entité
        $oEntity = $this->oSerializer->deserialize(json_encode($aData[$i]), $this->oOptions['entity']);
        $this->oEntityManager->persist($oEntity);
        if ($useCustomFields) {
          $aEntity = $this->oCustomFieldsManager->insertCustomFields($oEntity->jsonSerialize(), $aData[$i]);
          // mise à jour de l'entité si il y a eu des modifications
          if ($this->oCustomFieldsManager->dataIsModified()) {
            $this->oSerializer->deserialize(json_encode($aEntity), $this->oOptions['entity'], ['object_to_populate' => $oEntity]);
          }
        }
        // Sauvegarde en base de donnée
        $this->oEntityManager->persist($oEntity);
        $this->oEntityManager->flush();
        // rafraichissement de l'entité en cas de valeur set par la base de données
        $this->oEntityManager->refresh($oEntity);
        $aEntitiesAndData[] = [
          'entity' => $oEntity,
          'data' => $aData[$i],
        ];
      }
      $this->oEntityManager->getConnection()->commit();
    } catch (\Exception $e) {
      $this->oEntityManager->getConnection()->rollBack();
      $this->oErrorLogger->error('ORM Accessor : ' . $e->getMessage());
      return ['status' => 0];
    }
    if (empty($aEntitiesAndData)) {
      $this->oErrorLogger->error('ORM Accessor : entity is empty');
      return ['status' => 0];
    } else {
      $aEntities = [];
      // gestion des champs custom
      for ($j = 0; $j < count($aEntitiesAndData); $j++) {
        if ($useCustomFields) {
          $aEntity = $this->oCustomFieldsManager->afterInsertCustomFields($aEntitiesAndData[$j]['entity']->jsonSerialize(), $aEntitiesAndData[$j]['data']);
          // mise à jour de l'entité si il y a eu des modifications
          if ($this->oCustomFieldsManager->dataIsModified()) {
            $this->oSerializer->deserialize(json_encode($aEntity), $this->oOptions['entity'], ['object_to_populate' => $aEntitiesAndData[$j]['entity']]);
          }
          // rafraichissement de l'entité
          $this->oEntityManager->refresh($aEntitiesAndData[$j]['entity']);
        }
        $aEntities[] = $aEntitiesAndData[$j]['entity'];
      }

      return [
        'data' => $aEntities,
        'status' => 1,
      ];
    }
  }
}
