<?php

namespace App\Module\Vitis\WebService\Service\DataAccessor;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Configuration\Route;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\Serializer\VitisSerializer;
use App\Module\Vitis\WebService\Service\Token\JweTokenManager;
use Doctrine\DBAL\ParameterType;
use Doctrine\DBAL\Query\QueryBuilder;
use Doctrine\ORM\EntityManager;
use Monolog\Logger;

/**
 * Permet d'accéder aux données cia une entité Doctrine
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 * @author Ludivine Premont   <ludivine.premont@veremes.com>
 */
class QueryAccessor extends AbstractQueryAccessor implements DataAccessorInterface {
  /**
   * @var string
   */
  protected const ENTITY_ALIAS = 'generic_row';

  /**
   * Le nom du compteur des enregistrements
   * @var string
   */
  private $sRecordsCount;

  /**
   * @param EntityManager $oEntityManager Gestionnaire d'entité Doctrine
   * @param Route $oRouteConfigurator service de configuration de la route
   * @param VitisSerializer $oSerializer Objet permettant la désearialization dans une couche ORM
   * @param LoggerManager $oLogger Service d'accès aux logger
   * @param Properties $oProperties Service d'accès aux properties
   * @param JweTokenManager $oJweTokenManager accès au Token et encryptage
   */
  public function __construct(EntityManager $oEntityManager, Route $oRouteConfigurator, VitisSerializer $oSerializer, LoggerManager $oLogger, Properties $oProperties, JweTokenManager $oJweTokenManager) {
    parent::__construct($oEntityManager, $oRouteConfigurator, $oSerializer, $oLogger, $oProperties, $oJweTokenManager);

    $this->sRecordsCount = self::ENTITY_ALIAS;
    if (isset($this->oOptions['field_id'])) {
      $this->sRecordsCount .= '.' . $this->oOptions['field_id'];
    }

    $this->reloadTableNameFromOptions();
  }

  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function getEntity($id, array $aData = [], bool $useCustomFields = true): array{
    $oEntity = null;
    // initialisation query builder
    $qb = $this->createQueryBuilder();

    // gestion de la liste des champs
    $sAttributes = (isset($aData['attributs'])) ? $aData['attributs'] : '';
    // Si options dans routes.json n'existe pas mettre à true
    $sSelectedFields = $this->oAccessorFunc->attributeManager($sAttributes);

    // si le nom de la table n'existe pas on essaie de le recharger depuis le soptions
    if (!isset($this->sTableName)) {
      $this->reloadTableNameFromOptions();
    }

    $qb->select($sSelectedFields)
      ->from($this->sTableName, self::ENTITY_ALIAS)
      ->where(self::ENTITY_ALIAS . '.' . $this->oOptions['field_id'] . ' = :mId')
      ->setParameter('mId', $id);
    try {
      $oEntity = $qb->execute()->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('Query Accessor : can\'t get this object : ' . $e->getMessage());
      return ['status' => 0];
    }
    // si aucune entité n'est trouvé le status vaut 0 sinon 1
    if (empty($oEntity)) {
      $this->oErrorLogger->error('Query Accessor : entity is empty (id : ' . $id . ')');
      return ['status' => 0, 'message' => 'entity is empty'];
    } else {
      $aEntity = $oEntity[0];
      if ($useCustomFields) {
        $aEntity = $this->oCustomFieldsManager->getCustomFields($aEntity, $aData);
      }
      return [
        'data' => $aEntity,
        'status' => 1,
      ];
    }
  }
  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function getEntities(array $aData, bool $useCustomFields = true): array{
    // si le nom de la table n'existe pas on essaie de le recharger depuis le soptions
    if (!isset($this->sTableName)) {
      $this->reloadTableNameFromOptions();
    }

    $aEntities = null;
    $qb = $this->createQueryBuilder();
    $qbCount = $this->createQueryBuilder();

    $bDistinct = (isset($aData['distinct'])) ? ($aData['distinct'] === 'true') : false;

    // gestion de la liste des champs
    $sAttributes = (isset($aData['attributs'])) ? $aData['attributs'] : '';
    $sSelectedFields = $this->oAccessorFunc->attributeManager($sAttributes);

    $sCountField = $this->sRecordsCount;
    // gestion de distinct
    if ($bDistinct) {
      $sSelectedFields = 'DISTINCT ' . $sSelectedFields;
      $sCountField = 'DISTINCT ' . $this->sRecordsCount;
    }
    // initialisation query builder
    $qb->select($sSelectedFields)
      ->from($this->sTableName, self::ENTITY_ALIAS);

    $qbCount->select('COUNT (' . $sCountField . ') AS "nb_rows"')
      ->from($this->sTableName, self::ENTITY_ALIAS);

    // order_by
    if (isset($aData['order_by'])) {
      if (in_array($aData['order_by'], $this->oRouteConfigurator->getFieldsList(false))) {
        // on définir un sort_order par défaut
        $sSortOrder = 'ASC';
        // si ça vient de la requête on protège et on set
        if (array_key_exists('sort_order', $aData)) {
          $sSortOrder = match(strtoupper($aData['sort_order'])) {
            'DESC' => 'DESC',
            default => 'ASC'
          };
        }

        $qb->orderBy(self::ENTITY_ALIAS . '.' . $aData['order_by'], $sSortOrder);
        $qb->nullsLast();
      }
    }

    // filter
    $aData['filter'] = $aData['filter'] ?? null;
    if ($this->oAccessorFunc->needFilter($aData['filter'])) {
      $aFilterReturn = $this->oAccessorFunc->prepareAndDecodeFilter($aData['filter']);
      if (!empty($aFilterReturn['request'])) {
        $qb->where($aFilterReturn['request']);
        $qbCount->where($aFilterReturn['request']);
        foreach ($aFilterReturn['params'] as $sKey => $aValue) {
          if (isset($aValue['type'])) {
            $qb->setParameter($sKey, $aValue['value'], $aValue['type']);
            $qbCount->setParameter($sKey, $aValue['value'], $aValue['type']);
          } else {
            $qb->setParameter($sKey, $aValue['value']);
            $qbCount->setParameter($sKey, $aValue['value']);
          }
        }
      }
    }

    // offset
    if (isset($aData['offset']) && $aData['offset'] > 0) {
      $qb->setFirstResult($aData['offset']);
    }

    // limit
    if (isset($aData['limit']) && $aData['limit'] > 0) {
      $qb->setMaxResults($aData['limit']);
    }

    $aEntities = (array) $qb->execute()->fetchAllAssociative();
    $iCount = $qbCount->execute()->fetchAllAssociative();

    // si aucune entité n'est trouvé le status vaut 0 sinon 1
    if ($useCustomFields) {
      // récupération des champs custom
      foreach ($aEntities as &$aEntity) {
        $aEntity = $this->oCustomFieldsManager->getCustomFields($aEntity, $aData);
      }
    }
    return [
      'data' => $aEntities,
      'total_row_number' => $iCount[0]['nb_rows'],
      'list_count' => count($aEntities),
      'status' => 1,
    ];
  }
  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function insertEntity(string $sData, bool $useCustomFields = true): array{
    $oReturn = ['status' => 0];
    $oConnexion = $this->oEntityManager->getConnection();
    $oConnexion->setAutoCommit(false);
    // on démarre une transaction
    $oConnexion->beginTransaction();
    try {
      $aData = json_decode($sData, true);

      if ($useCustomFields) {
        $aData = $this->oCustomFieldsManager->insertCustomFields($aData, $aData);
      }

      $qb = $this->createQueryBuilder();

      if (!isset($this->sTableName)) {
        $this->reloadTableNameFromOptions();
      }
      // initialisation query builder
      $qb->insert($this->sTableName);

      $aProtectedValues = $this->oAccessorFunc->sqlValuesExtractor($aData);

      // Valeurs
      foreach ($aProtectedValues['values'] as $sField => $sBind) {
        if (!empty($aProtectedValues['functions'][$sField])) {
          $qb->setValue($sField, $aProtectedValues['functions'][$sField]);
        } else {
          $qb->setValue($sField, $sBind);
        }
      }

      // Paramètres
      foreach ($aProtectedValues['parameters'] as $sField => $mValue) {
        // Transforme les chaines de caractère vide en null si le champ est de type integer
        if ($aProtectedValues['types'][$sField] === 'integer' && empty($mValue)) {
          $mValue = null;
        }
        $qb->setParameter($sField, $mValue);
      }

      $aStatementData = [];
      $qb->returning('*');
      $aStatementData = $qb->execute()->fetchAssociative();

      $sFieldId = $this->oRouteConfigurator->getFieldId();
      $mId = $aStatementData[$sFieldId];
      if (empty($mId)) {
        throw new \Exception('can\'t get id for last object');
      } else {
        // traitements des champs custom
        if ($useCustomFields) {
          $this->oCustomFieldsManager->afterInsertCustomFields($aStatementData, $aData);
        }
        $oConnexion->commit();
        $oReturn = $this->getEntity($mId);
      }
    } catch (\Exception $e) {
      $oConnexion->rollBack();
      $this->oErrorLogger->error('Query Accessor : can\'t insert this object : ' . $e->getMessage());
    } finally {
      // Les transaction en cours seront automatiquement commit, ne pas oublié de commit/rollback avant ce point
      $oConnexion->setAutoCommit(true);
      return $oReturn;
    }
  }
  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function updateEntity($id, string $sData, bool $useCustomFields = true): array{
    $aData = (array) json_decode($sData, true);
    // si le champ id n'est pas set on le remplis avec le paramètre
    if (!isset($aData[$this->oOptions['field_id']])) {
      $aData[$this->oOptions['field_id']] = $id;
    }
    // si les données sont déjà considérée comme modifiée c'est que l'update a été appelé par l'insertion
    // dans ce cas ce n'est pas la peine de récupérer les champs custom
    if ($useCustomFields) {
      $aData = $this->oCustomFieldsManager->updateCustomFields($aData, $aData);
    }

    $qb = $this->createQueryBuilder();
    // récupération de l'entité à mettre à jour
    $aEntity = $this->getEntity($id);
    if ($aEntity['status'] === 1) {
      // initialisation query builder
      $qb->update($this->sTableName);

      if (empty($aData)) {
        $this->oErrorLogger->error('Something is wrong on data : ' . print_r($aData, true));
        return ['status' => 0];
      }

      $aParams = [];

      $aProtectedValues = $this->oAccessorFunc->sqlValuesExtractor($aData);
      foreach ($aProtectedValues['values'] as $sField => $sBind) {
        // Requête SQL
        if (!empty($aProtectedValues['functions'][$sField])) {
          $sFunctionString = str_replace(":$sField", "[VAL_$sField]", $aProtectedValues['functions'][$sField]);
          $qb->set('[COL_' . $sField . ']', $sFunctionString);
        } else {
          $qb->set('[COL_' . $sField . ']', '[VAL_' . $sField . ']');
        }

        // Paramètres
        $aParams['COL_' . $sField] = ['value' => $sField, 'type' => 'column_name'];
        switch ($aProtectedValues['types'][$sField]) {
          case 'geom':
            $aParams['VAL_' . $sField] = ['value' => $aProtectedValues['parameters'][$sField], 'type' => 'geometry'];
            break;
          default:
            $aParams['VAL_' . $sField] = ['value' => $aProtectedValues['parameters'][$sField], 'type' => 'mixed'];
            break;
        }
      }

      // Where
      $aParams['COLUMN_ID'] = ['value' => $this->oOptions['field_id'], 'type' => 'column_name'];
      $aParams['mIdInWhere'] = ['value' => $id, 'type' => 'array'];
      $qb->where('[COLUMN_ID] = [mIdInWhere]');

      try {
        $qb->executeWithParams($aParams);
      } catch (\Exception $e) {
        $this->oErrorLogger->error('Query Accessor : can\'t update this object : ' . $e->getMessage());
        return ['status' => 0];
      }
    } else {
      return ['status' => 0];
    }
    return $this->getEntity($id);
  }
  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function deleteEntity($id, bool $useCustomFields = true): array{
    $oEntity = null;
    $qb = $this->createQueryBuilder();

    // récupération de l'entité à mettre à jour
    $aEntity = $this->getEntity($id);
    if ($aEntity['status'] === 1) {
      $oEntity = $aEntity['data'];
    }

    if (!empty($oEntity)) {
      if ($useCustomFields) {
        $oEntity = $this->oCustomFieldsManager->deleteCustomFields($oEntity, []);
      }
      // Supprime de la base de donnée
      $qb->delete($this->sTableName)
        ->where($this->oOptions['field_id'] . ' = :mIdInWhere')
        ->setParameter('mIdInWhere', $id);

      try {
        $qb->execute();
      } catch (\Exception $e) {
        $this->oErrorLogger->error('Query Accessor : can\'t delete this object : ' . $e->getMessage());
        return ['status' => 0];
      }

      return [
        'data' => $id,
        'status' => 1,
      ];
    } else {
      return ['status' => 0];
    }
  }
  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function deleteEntities(array $aIds, bool $useCustomFields = true): array{
    $aEntities = null;
    $aQueryBuilder = [
      'DELETE' => $this->createQueryBuilder()
        ->delete($this->sTableName),
      'SELECT' => $this->createQueryBuilder()
        ->select('[COLUMN_ID]')
        ->from($this->sTableName),
    ];

    $aParams = [
      'COLUMN_ID' => ['value' => $this->oOptions['field_id'], 'type' => 'column_name'],
      'aIds' => ['value' => $aIds, 'type' => 'array'],
    ];

    foreach ($aQueryBuilder as $qb) {
      $qb->where('[COLUMN_ID] IN ([aIds])');
    }

    try {
      $oResultSelectBefore = $aQueryBuilder['SELECT']->executeWithParams($aParams);
      $aQueryBuilder['DELETE']->executeWithParams($aParams);
      $oResultSelect = $aQueryBuilder['SELECT']->executeWithParams($aParams);
    } catch (\Exception $e) {
      $this->oErrorLogger->error('Query Accessor : can\'t delete this objects [' . implode(', ', $aIds) . '] : ' . $e->getMessage());
      return ['status' => 0];
    }

    $aEntitiesBefore = $oResultSelectBefore->fetchAllAssociative();
    $aEntities = $oResultSelect->fetchAllAssociative();

    if (empty($aEntitiesBefore)) {
      $aResponse = [
        'status' => 1,
        'total_row_removed' => 0,
        'data' => [],
      ];
    } else if (empty($aEntities)) {
      $aResponse = [
        'status' => 1,
        'total_row_removed' => count($aIds),
        'data' => $aIds,
      ];
    } else {
      $aResponse = [
        'status' => 0,
        'total_row_removed' => 0,
      ];
      $aResponse['data'] = [];
      foreach ($aIds as $mId) {
        $aEntity = $this->getEntity($mId);
        if ($aEntity['status'] === 0) {
          $aResponse['data'][] = $mId;
          $aResponse['total_row_removed'] = $aResponse['total_row_removed'] + 1;
        }
      }
    }
    if ($useCustomFields) {
      $sFieldId = $this->oRouteConfigurator->getFieldId();
      foreach ($aResponse['data'] as $mId) {
        // si la suppression a été effectuée
        if (isset($sFieldId)) {
          $this->oCustomFieldsManager->deleteCustomFields([$sFieldId => $mId], []);
        }
      }
    }
    return $aResponse;
  }

  /**
   * Provient de DataAccessorInterface
   * {@inheritdoc}
   */
  public function updateEntities(array $aIds, string $sData, bool $useCustomFields = true): array{
    // si le nom de la table n'existe pas on essaie de le recharger depuis le soptions
    if (!isset($this->sTableName)) {
      $this->reloadTableNameFromOptions();
    }

    // Récupère les entités en fonction des ids selectionné
    $filter = [
      'column' => $this->oOptions['field_id'],
      'compare_operator' => 'IN',
      'value' => $aIds,
    ];

    $aEntities = $this->getEntities(['filter' => json_encode($filter), 'attributs' => $this->oOptions['field_id']], false);

    // Créer un nouveaux tableau avec les ids qui existe en base
    $aVerifiedIds = [];
    foreach ($aEntities['data'] as $value) {
      $aVerifiedIds[] = $value[$this->oOptions['field_id']];
    }

    $aData = json_decode($sData, true);
    // si les données sont déjà considérée comme modifiée c'est que l'update a été appelé par l'insertion
    // dans ce cas ce n'est pas la peine de récupérer les champs custom
    if ($useCustomFields) {
      foreach ($aVerifiedIds as $id) {
        $aData[$this->oOptions['field_id']] = $id;
        $aData = $this->oCustomFieldsManager->updateMassiveCustomFields($aData, $aData);
        // on ne veut surtout pas mettre à jour l'id
        unset($aData[$this->oOptions['field_id']]);
      }
    }

    // On vérifie qu'au moins un id est valide
    if (empty($aVerifiedIds)) {
      $this->oErrorLogger->error('None of the selected ids is valid  : ' . print_r($aIds, true));
      return ['status' => 0];
    }
    // On vérifie qu'il y a des valeurs dans aData
    if (empty($aData)) {
      $this->oErrorLogger->error('Something is wrong on data : ' . print_r($aData, true));
      return ['status' => 0];
    }

    // initialisation query builder
    $qb = $this->createQueryBuilder();
    $qb->update($this->sTableName);

    $aParams = [];

    $aProtectedValues = $this->oAccessorFunc->sqlValuesExtractor($aData);
    foreach ($aProtectedValues['values'] as $sField => $sBind) {
      // Requête SQL
      if (!empty($aProtectedValues['functions'][$sField])) {
        $sFunctionString = str_replace(":$sField", "[VAL_$sField]", $aProtectedValues['functions'][$sField]);
        $qb->set('[COL_' . $sField . ']', $sFunctionString);
      } else {
        $qb->set('[COL_' . $sField . ']', '[VAL_' . $sField . ']');
      }

      // Paramètres
      $aParams['COL_' . $sField] = ['value' => $sField, 'type' => 'column_name'];
      switch ($aProtectedValues['types'][$sField]) {
        case 'geom':
          $aParams['VAL_' . $sField] = ['value' => $aProtectedValues['parameters'][$sField], 'type' => 'geometry'];
          break;
        default:
          $aParams['VAL_' . $sField] = ['value' => $aProtectedValues['parameters'][$sField], 'type' => 'mixed'];
          break;
      }
    }

    $aParams['COLUMN_ID'] = ['value' => $this->oOptions['field_id'], 'type' => 'column_name'];
    $aParams['aIds'] = ['value' => $aVerifiedIds, 'type' => 'array'];

    $qb->where('[COLUMN_ID] IN ([aIds])');

    try {
      $qb->executeWithParams($aParams);
    } catch (\Exception $e) {
      $this->oErrorLogger->error('Query Accessor : can\'t update this objects : ' . $e->getMessage());
      return ['status' => 0];
    }

    // Récupère et retourne les entités modifié
    $filter['value'] = $aVerifiedIds;
    return $this->getEntities(['filter' => json_encode($filter)]);
  }

  /**
   * Ajoute un utilisateur dans Postgres
   * @param string $sLogin login à ajouter
   * @return boolean true si l'utilisateur a été ajouté false sinon
   */
  public function createUser(string $sLogin): bool {
    return $this->createPgRole($sLogin, true);
  }
  /**
   * Ajoute un role dans Postgres
   * @param string $sRolname login à ajouter
   * @return boolean true si le role a été ajouté false sinon
   */
  public function createRole(string $sRolname): bool {
    return $this->createPgRole($sRolname, false);
  }
  /**
   * Ajoute un utilisateur/role dans Postgres
   * @param string $sRolname login/role à ajouter
   * @param bool $bCanLogin définit si on veut un utilisateur ou un rolegroup
   * @return boolean true si l'utilisateur/role a été ajouté false sinon
   */
  private function createPgRole(string $sRolname, bool $bCanLogin): bool {
    try {
      $sLoginStr = ($bCanLogin) ? 'LOGIN' : 'NOLOGIN';
      $oConnection = $this->oEntityManager->getConnection();
      $sSql = 'CREATE ROLE "' . addslashes($sRolname) . '" WITH ' . $sLoginStr . ' NOSUPERUSER INHERIT NOCREATEDB NOCREATEROLE NOREPLICATION';
      /* @var Doctrine\DBAL\Driver\ResultStatement $oStatement */
      $oStatement = $oConnection->executeQuery($sSql);
      $oStatement->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('Cannot create postgres User/Role with name ' . $sRolname . '(' . $e->getMessage() . ')');
      return false;
    }
    return true;
  }

  /**
   * Supprime un utilisateur/role Postgres s'il existe
   * Si le role est utilsé comme OWNER pour un objet postgres ou a des droits sur un objet postgres
   * Il ne pourra pas être effacer
   * @param string $sRolname nom du role à supprimer
   * @return bool
   */
  public function dropRole(string $sRolname): bool {
    try {
      $oConnection = $this->oEntityManager->getConnection();
      $sSql = 'DROP ROLE IF EXISTS "' . addslashes($sRolname) . '"';
      /* @var Doctrine\DBAL\Driver\ResultStatement $oStatement */
      $oStatement = $oConnection->executeQuery($sSql);
      $oStatement->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('Cannot delete postgres User/Role with name ' . $sRolname . '(' . $e->getMessage() . ')', [$e]);
      return false;
    }
    return true;
  }
  /**
   * Modifie le mot de passe d'utilisateur
   * @param string $sRolename
   * @param string $sPassword
   * @return bool
   */
  public function updateRolePassword(string $sRolename, string $sPassword): bool {
    try {
      $oConnection = $this->oEntityManager->getConnection();
      $sSql = 'ALTER ROLE "' . addslashes($sRolename) . '" WITH PASSWORD ' . $oConnection->quote($sPassword, ParameterType::STRING);
      /* @var Doctrine\DBAL\Driver\ResultStatement $oStatement */
      $oStatement = $oConnection->executeQuery($sSql);
      $oStatement->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('Cannot set password for role ' . $sRolename . '(' . $e->getMessage() . ')');
      return false;
    }
    return true;
  }
  /**
   * Ajoute CREATEROLE à un role postgres
   * @param string $sRolename
   * @return bool
   */
  public function addCreateRoleToRole(string $sRolename): bool {
    return $this->alterCreateRoleToRole($sRolename, true);
  }
  /**
   * supprime CREATEROLE à un role postgres
   * @param string $sRolename
   * @return bool
   */
  public function removeCreateRoleToRole(string $sRolename): bool {
    return $this->alterCreateRoleToRole($sRolename, false);
  }
  /**
   * Ajoute CREATEROLE à un role postgres
   * @param string $sRolename
   * @param bool $bCreateRole définit s'il faut ajouter ou supprimer CREATEROLE
   * @return bool
   */
  private function alterCreateRoleToRole(string $sRolename, bool $bCreateRole): bool {
    try {
      $sCreateRole = ($bCreateRole) ? 'CREATEROLE' : 'NOCREATEROLE';
      $oConnection = $this->oEntityManager->getConnection();
      $sSql = 'ALTER ROLE "' . addslashes($sRolename) . '" ' . $sCreateRole;
      $oStatement = $oConnection->executeQuery($sSql);
      $oStatement->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('Cannot add CREATEROLE to role ' . $sRolename . '(' . $e->getMessage() . ')');
      return false;
    }
    return true;
  }
  /**
   * Grant une liste de role à un role postgres
   * @param string[] $aRoles liste des role à ajouter to $sRoleName
   * @param string $sRolename
   * @return bool
   */
  public function grantRolesToRole(array $aRoles, string $sRolename): bool {
    try {
      $oConnection = $this->oEntityManager->getConnection();

      // concaténation et protection de la liste des roles
      $sRoles = $this->protectRolesArray($aRoles);

      $sSql = 'GRANT ' . $sRoles . ' TO "' . addslashes($sRolename) . '"';
      $oStatement = $oConnection->executeQuery($sSql);
      $oStatement->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('Cannot grant roles to role ' . $sRolename . '(' . $e->getMessage() . ')');
      return false;
    }
    return true;
  }
  /**
   * Grant une liste de role à un role postgres
   * @param string $sRolename
   * @param string[] $aRoles liste des role à ajouter to $sRoleName
   * @return bool
   */
  public function grantRoleToRoles(string $sRolename, array $aRoles): bool {
    try {
      $oConnection = $this->oEntityManager->getConnection();

      // concaténation et protection de la liste des roles
      $sRoles = $this->protectRolesArray($aRoles);

      $sSql = 'GRANT "' . addslashes($sRolename) . '" TO ' . $sRoles;
      $oStatement = $oConnection->executeQuery($sSql);
      $oStatement->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('Cannot grant role ' . $sRolename . ' to roles(' . $e->getMessage() . ')');
      return false;
    }
    return true;
  }
  /**
   * Revoke une liste de role à un role postgres
   * @param string[] $aRoles liste des role à ajouter to $sRoleName
   * @param string $sRolename
   * @return bool
   */
  public function revokeRolesFromRole(array $aRoles, string $sRolename): bool {
    try {
      $oConnection = $this->oEntityManager->getConnection();

      // concaténation et protection de la liste des roles
      $sRoles = $this->protectRolesArray($aRoles);

      $sSql = 'REVOKE ' . $sRoles . ' FROM "' . addslashes($sRolename) . '"';
      $oStatement = $oConnection->executeQuery($sSql);
      $oStatement->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('Cannot revoke roles to role ' . $sRolename . '(' . $e->getMessage() . ')');
      return false;
    }
    return true;
  }
  /**
   * Revoke un role à une liste de role postgres
   * @param string $sRolename
   * @param string[] $aRoles liste des role à ajouter to $sRoleName
   * @return bool
   */
  public function revokeRoleFromRoles(string $sRolename, array $aRoles): bool {
    try {
      $oConnection = $this->oEntityManager->getConnection();

      // concaténation et protection de la liste des roles
      $sRoles = $this->protectRolesArray($aRoles);

      $sSql = 'REVOKE "' . addslashes($sRolename) . '" FROM ' . $sRoles;
      $oStatement = $oConnection->executeQuery($sSql);
      $oStatement->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('Cannot revoke role ' . $sRolename . ' to roles (' . $e->getMessage() . ')');
      return false;
    }
    return true;
  }
  /**
   * protèges les roles avant de les utiliser dans des requêtes
   * @param string[] $aRoles Liste des Roles
   */
  private function protectRolesArray(array $aRoles): string {
    // protection de chaque role
    foreach ($aRoles as &$sRole) {
      $sRole = addslashes($sRole);
    }
    // concaténation de la liste des roles
    return '"' . implode('", "', $aRoles) . '"';
  }
  /**
   * Check si role existe
   * @param string $sRolename nom d'utilisateur de BDD
   * @return bool|null
   */
  public function ifUserExist(string $sRolename): ?bool {
    if (empty($sRolename)) {
      return null;
    }
    /* @var $em EntityManager */
    $qb = $this->createQueryBuilder();
    $qb->select('count(a) as "nb_user"')
      ->from('pg_user', 'a')
      ->where($qb->expr()->eq('a.usename', ':usename'))
      ->setParameter('usename', $sRolename);

    $oResult = $qb->executeWithParams([])->fetchAssociative();

    $nb_row = intval($oResult['nb_user']);
    return $nb_row > 0;
  }
  /**
   * Check si role existe
   * @param string $sRolename nom de groupe de BDD
   * @return bool|null
   */
  public function ifPgRoleExist(string $sRolename) {
    if (empty($sRolename)) {
      return null;
    }
    /* @var $em EntityManager */
    $qb = $this->createQueryBuilder();
    $qb->select('count(a) as "nb_user"')
      ->from('pg_group', 'a')
      ->where($qb->expr()->eq('a.groname', ':groname'))
      ->setParameter('groname', $sRolename);

    $oResult = $qb->executeWithParams([])->fetchAssociative();

    $nb_row = intval($oResult['nb_user']);
    return $nb_row > 0;

  }

  /**
   * Récupère les base de données de la connexion
   * @param mixed $aData
   * @return array|null
   */
  public function getDatabases($aData): ?array{
    $qb = $this->createQueryBuilder();
    $qb->select('pg_database.datname as database')
      ->from('pg_database', 'pg_database');

    // filter
    $qb = $this->setParamsFilter($aData, $qb, 'datname', 'pg_database');

    if (isset($aData['order_by'])) {
      $qb->orderBy($aData['order_by']);
    }

    try {
      $oDatabases = $qb->execute()->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('GenericQueries : can\'t get databases: ' . $e->getMessage());
      return ['status' => 0];
    }

    return $oDatabases;
  }

  /**
   * Permet de récupérer les schémas d'une base de données locale
   * @param mixed $aData
   * @return array|null
   */
  public function getSchemas($aData): ?array{
    // Ajout date d'execution si le projet est un projet en heure creuse
    $qb = $this->createQueryBuilder();
    $qb->select('DISTINCT(tables.table_schema) as schema_name')
      ->from('information_schema.tables', 'tables');

    // filter
    $qb = $this->setParamsFilter($aData, $qb, 'table_schema', 'tables');

    if (isset($aData['order_by'])) {
      $qb->orderBy($aData['order_by']);
    }

    try {
      $oSchemas = $qb->execute()->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('GenericQueries : can\'t get schemas: ' . $e->getMessage());
      return ['status' => 0];
    }

    return $oSchemas;
  }

  /**
   * Permet de récupérer les tables d'un schema d'une base de données locale
   * @param mixed $aData
   * @return array|null
   */
  public function getTables($aData): ?array{
    $qb = $this->createQueryBuilder();
    $qb->select('tables.table_name')
      ->from('information_schema.tables', 'tables')
      ->where($qb->expr()->eq('tables.table_schema', ':table_schema'))
      ->setParameter('table_schema', $aData['schema']);

    if (isset($aData['order_by'])) {
      $qb->orderBy($aData['order_by']);
    }

    try {
      $oTables = $qb->execute()->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('GenericQueries : can\'t get tables: ' . $e->getMessage());
      return ['status' => 0];
    }

    return $oTables;
  }

  /**
   * Permet de récupérer les columns d'une table pour la configuration
   * @param mixed $aData
   * @return array
   */
  public function getFieldsFromTable($aData) {
    $qb = $this->createQueryBuilder();
    $qb->select('columns.column_name, columns.data_type')
      ->from('information_schema.columns', 'columns')
      ->where($qb->expr()->eq('columns.table_schema', '[SCHEMA]'))
      ->andWhere($qb->expr()->eq('columns.table_name', '[TABLE]'))
      ->orderBy('column_name');

    $aParams = [
      'SCHEMA' => ['value' =>  $aData['schema'], 'type' => 'string'],
      'TABLE' => ['value' =>  $aData['table'], 'type' => 'string'],
    ];
    try {
      $oTables = $qb->executeWithParams($aParams)->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('GenericQueries : can\'t get tables: ' . $e->getMessage());
      return ['status' => 0];
    }

    // Si les colonnes n'ont pas pu être récupérées, essaye une autre méthode (principalement pour les vues matérialisées)
    if (empty($oTables)) {
      $qb = $this->createQueryBuilder();
      $qb->select('a.attname as column_name, format_type(a.atttypid, a.atttypmod) as data_type')
        ->from('pg_attribute', 'a')
        ->innerJoin('a', 'pg_class', 't', 'a.attrelid = t.oid')
        ->innerJoin('t', 'pg_namespace', 's', 't.relnamespace = s.oid')
        ->where($qb->expr()->eq('s.nspname', '[SCHEMA]'))
        ->andWhere($qb->expr()->eq('t.relname', '[TABLE]'))
        ->andWhere('a.attnum > 0')
        ->setParameter('table_schema', $aData['schema'])
        ->setParameter('table_name', $aData['table'])
        ->orderBy('a.attnum');

      $aParams = [
        'SCHEMA' => ['value' =>  $aData['schema'], 'type' => 'string'],
        'TABLE' => ['value' =>  $aData['table'], 'type' => 'string'],
      ];
      try {
        $oTables = $qb->executeWithParams($aParams)->fetchAllAssociative();
      } catch (\Exception $e) {
        $this->oErrorLogger->error('GenericQueries : can\'t get tables with second method: ' . $e->getMessage());
        return ['status' => 0];
      }
    }

    // Colonnes geometriques
    $qb = $this->createQueryBuilder();
    $qb->select('columns.f_geometry_column, columns.type, columns.srid, columns.coord_dimension')
      ->from('public.geometry_columns', 'columns')
      ->where($qb->expr()->eq('columns.f_table_schema', '[SCHEMA]'))
      ->andWhere($qb->expr()->eq('columns.f_table_name', '[TABLE]'))
      ->orderBy('f_geometry_column');

    $aParams = [
      'SCHEMA' => ['value' =>  $aData['schema'], 'type' => 'string'],
      'TABLE' => ['value' =>  $aData['table'], 'type' => 'string'],
    ];
    try {
      $oGeoms = $qb->executeWithParams($aParams)->fetchAllAssociative();
      foreach ($oGeoms as $oGeom) {
        foreach ($oTables as $key => $aTable) {
          if ($aTable['column_name'] === $oGeom['f_geometry_column']) {
            $aTable['data_type'] = 'geometry';
            $aTable['geometry_type'] = $oGeom['type'];
            $aTable['geometry_srid'] = $oGeom['srid'];
            $aTable['geometry_coord_dimension'] = $oGeom['coord_dimension'];
            $oTables[$key] = $aTable;
          }
        }
      }
    } catch (\Exception $e) {
      // @ignoreException
      // Ne pas mettre d'erreur au cas où postgis ne soit pas installé
    }

    $response = ['status' => 1, 'data' => []];
    foreach ($oTables as $oTable) {
      $aField = ['name' => $oTable['column_name']];

      switch ($oTable['data_type']) {
        case 'integer':
        case 'bigint':
        case 'numeric':
        case 'double':
        case 'double precision':
        case 'real':
        case 'smallint':
          $aField['type'] = 'integer';
          break;
        case 'boolean':
          $aField['type'] = 'boolean';
          break;
        case 'date':
        case 'timestamp':
        case 'timestamp with time zone':
        case 'timestamp without time zone':
          $aField['type'] = 'date';
          break;
        case 'geometry':
          $aField['type'] = 'geom';
          $aField['geometry_type'] = $oTable['geometry_type'];
          $aField['srid'] = $oTable['geometry_srid'];
          $aField['coord_dimension'] = $oTable['geometry_coord_dimension'];
          break;
        default:
          $aField['type'] = 'string';
          break;
      }

      $response['data'][] = $aField;
    }

    return $response;
  }

  /**
   * Permet de récupérer les tables d'un schema
   * @param mixed $aData
   * @return array|null
   */
  public function getOnlyTables($aData): ?array{
    $qb = $this->createQueryBuilder();
    $qb->select('tables.tablename as name')
      ->from('pg_catalog.pg_tables', 'tables')
      ->where($qb->expr()->eq('tables.schemaname', ':table_schema'))
      ->setParameter('table_schema', $aData['schema']);

    // filter
    $qb = $this->setParamsFilter($aData, $qb, 'pg_tables', 'tablename');

    if (isset($aData['order_by'])) {
      $qb->orderBy($aData['order_by']);
    }

    try {
      $oTables = $qb->execute()->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('GenericQueries : can\'t get tables: ' . $e->getMessage());
      return null;
    }

    return $oTables;
  }

  /**
   * Permet de récupérer les vues d'un schema
   * @param mixed $aData
   * @return array|null
   */
  public function getOnlyViews($aData): ?array{
    $qb = $this->createQueryBuilder();
    $qb->select('views.viewname as name')
      ->from('pg_catalog.pg_views', 'views')
      ->where($qb->expr()->eq('views.schemaname', ':schemaname'))
      ->setParameter('schemaname', $aData['schema']);

    // filter
    $qb = $this->setParamsFilter($aData, $qb, 'pg_views', 'viewname');

    if (isset($aData['order_by'])) {
      $qb->orderBy($aData['order_by']);
    }

    try {
      $oViews = $qb->execute()->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('GenericQueries : can\'t get views: ' . $e->getMessage());
      return null;
    }

    return $oViews;
  }

  /**
   * Permet de récupérer les vues matérialisé d'un schema
   * @param mixed $aData
   * @return array|null
   */
  public function getOnlyMaterializedViews($aData): ?array{
    $qb = $this->createQueryBuilder();
    $qb->select('matviews.matviewname as name')
      ->from('pg_catalog.pg_matviews', 'matviews')
      ->where($qb->expr()->eq('matviews.schemaname', ':schemaname'))
      ->setParameter('schemaname', $aData['schema']);

    // filter
    $qb = $this->setParamsFilter($aData, $qb, 'pg_matviews', 'matviewname');

    if (isset($aData['order_by'])) {
      $qb->orderBy($aData['order_by']);
    }

    try {
      $oMatviews = $qb->execute()->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('GenericQueries : can\'t get matviews: ' . $e->getMessage());
      return null;
    }

    return $oMatviews;
  }

  /**
   * Permet de récupérer les columns d'une table pour la configuration
   * @param mixed $aData
   * @return array|null
   */
  public function getColumns($aData): ?array{
    $qb = $this->createQueryBuilder();

    $qb->select('a.attname as column_name, format_type(a.atttypid, a.atttypmod) as data_type')
      ->from('pg_attribute', 'a')
      ->innerJoin('a', 'pg_class', 't', 'a.attrelid = t.oid')
      ->innerJoin('t', 'pg_namespace', 's', 't.relnamespace = s.oid')
      ->where($qb->expr()->eq('s.nspname', ':table_schema'))
      ->andWhere($qb->expr()->eq('t.relname', ':table_name'))
      ->andWhere('a.attnum > 0')
      ->andWhere('a.attisdropped = false')
      ->setParameter('table_schema', $aData['schema'])
      ->setParameter('table_name', $aData['table'])
      ->orderBy('a.attnum');

    try {
      $oColumns = $qb->execute()->fetchAllAssociative();
    } catch (\Exception $e) {
      $this->oErrorLogger->error('GenericQueries : can\'t get table columns: ' . $e->getMessage());
      return null;
    }

    return $oColumns;
  }

  /**
   * Récupère les informations sur colonne géométrie de la table
   * @param array $aData
   * @return array|null
   */
  public function getFirstGeomColumInfo(array $aData): ?array{
    // récupération qb
    $qb = $this->createQueryBuilder();
    $aGeomInfo = [];

    // Même si il y a plusieurs colonnes géométrie on ne garde que la première
    try {
      $qb->select('a.f_table_schema, a.f_table_name, a.f_geometry_column, a.srid, a.type')
        ->from('geometry_columns', 'a')
        ->where($qb->expr()->eq('a.f_table_schema', ':f_table_schema'))
        ->andWhere($qb->expr()->eq('a.f_table_name', ':f_table_name'))
        ->setParameter('f_table_schema', $aData['schema'])
        ->setParameter('f_table_name', $aData['pg_table']);

      $oResult = $qb->execute()->fetchAllAssociative();

      // On ne récupère que la première ligne trouvé
      if (empty($oResult)) {
        return null;
      }
      $aGeomInfo = $oResult[0];
    } catch (\Exception $e) {
      $this->oErrorLogger->error('getFirstGeomColumInfo : can\'t get geom column info: ' . $e->getMessage());
      return null;
    }

    return $aGeomInfo;
  }

  /**
   *
   * @param string $sColumnName
   * @param string $sEntityAlias
   * @return void
   */
  private function setSpecialDataAccessorFunc(string $sColumnName, string $sEntityAlias): void {
    // Set les columns
    $this->oAccessorFunc->setFieldsConfig([[
      'type' => 'string',
      'name' => $sColumnName,
    ]]);
    $this->oAccessorFunc->setEntityAlias($sEntityAlias);
  }

  /**
   * Ajoute à $qb les paramètres de filtre
   * @param array $aData
   * @param VitisNativeQueryBuilder $qb
   * @param string|null $sColumnName
   * @param string|null $sEntityAlias
   * @return VitisNativeQueryBuilder
   */
  private function setParamsFilter(array $aData, VitisNativeQueryBuilder $qb, string $sColumnName = null, string $sEntityAlias = null): VitisNativeQueryBuilder {
    // filter
    if (isset($aData['filter'])) {
      if (isset($sColumnName) && !empty($sColumnName) && isset($sEntityAlias) && !empty($sEntityAlias)) {
        $this->setSpecialDataAccessorFunc($sColumnName, $sEntityAlias);
      }

      // Décode et applique le filtre
      $aFilterReturn = $this->oAccessorFunc->decodeJSONFilter($aData['filter']);
      if (!empty($aFilterReturn['request'])) {
        $qb->where($aFilterReturn['request']);
        foreach ($aFilterReturn['params'] as $sKey => $aValue) {
          if (isset($aValue['type'])) {
            $qb->setParameter($sKey, $aValue['value'], $aValue['type']);
          } else {
            $qb->setParameter($sKey, $aValue['value']);
          }
        }
      }
    }

    return $qb;
  }

  /**
   * @return VitisNativeQueryBuilder QueryBuilder vitis
   */
  public function createQueryBuilder(): VitisNativeQueryBuilder {
    return new VitisNativeQueryBuilder($this->oEntityManager->getConnection(), $this->oErrorLogger);
  }

  /**
   * Undocumented function
   *
   * @param  array $aData
   * @return array
   */
  public function getDatasFromColumn(array $aData): array{
    try {
      $qb = $this->createQueryBuilder();
      $qb->select($aData['column_distinct'])
        ->from($aData['schema'] . '.' . $aData['table'])
        ->where($aData['column_distinct'] . ' is NOT NULL')
        ->groupBy($aData['column_distinct'])
        ->setMaxResults(20);
      $results = $qb->execute()->fetchAllAssociative();
      return ['status' => 1, 'data' => $results];
    } catch (\Exception $e) {
      $this->oErrorLogger->error('GenericQueries : can\'t get tables: ' . $e->getMessage());
      return ['status' => 0];
    }
  }

  /**
   * Checko les droits de via l'entityManager sur une table
   *
   * @param  string     $sSchema
   * @param  string     $sTableName
   * @return array|null
   */
  public function getTableRights(string $sSchema, string $sTableName): ?array{
    $aCheck = ['SELECT', 'INSERT', 'UPDATE', 'DELETE', 'TRUNCATE', 'REFERENCES', 'TRIGGER'];
    $aResult = [];
    try {
      foreach ($aCheck as $sRight) {
        $qb = $this->createQueryBuilder();
        $qb->select('has_table_privilege([TABLE_FULL_NAME], [RIGHT])');
        $aParams = [
          'TABLE_FULL_NAME' => ['value' =>'"' . $sSchema . '"."' . $sTableName . '"' , 'type' => 'string'],
          'RIGHT' => ['value' => $sRight, 'type' => 'string'],
        ];
        $aResult[$sRight] = $qb->executeWithParams($aParams)->fetchOne();
      }
      return $aResult;
    } catch (\Exception $e) {
      $this->oErrorLogger->error('GenericQueries : can\'t check rights: ' . $e->getMessage());
      return null;
    }
  }
}
