<?php

namespace App\Module\Vitis\WebService\Service\DataAccessor;

use Doctrine\DBAL\Connection;
use Doctrine\DBAL\ForwardCompatibility;
use Doctrine\DBAL\Query\QueryBuilder;
use Monolog\Logger;
use Ramsey\Uuid\Uuid;
use Doctrine\DBAL\Result;

/**
 * Permet d'avoir un QueryBuilder plus puissant que celui de DBAL
 * Notemment de pouvoir protéger des colonnes
 *
 * @author Anthony Borghi   <anthony.borghi@veremes.com>
 */

class VitisNativeQueryBuilder extends QueryBuilder {
  /**
   * @var string
   */
  private const VALUE_TO_PDO_BIND_TYPE = 'qb_prepared_bind';

  /**
   * @var Logger
   */
  private $oErrorLogger;

  /**
   * @var array
   */
  private $aReturningFields = [];

  /**
   * @var string|null
   */
  private $sNullPosition;

  /**
   * @param Connection $connection DBAL Connection.
   * @param Logger $oErrorLogger Logger pour les erreurs.
   */
  public function __construct(Connection $connection, Logger $oErrorLogger) {
    parent::__construct($connection);

    $this->oErrorLogger = $oErrorLogger;
  }

  /**
   * Permet de préparer des champs normalement non préparé par PDO
   * comme anciennement column_name, schema_name, table_name
   * @param array $aParams liste de paramètre à appliquer et leurs type
   * @return Result|int|string
   *
   * @throws \Exception
   */
  public function executeWithParams(array $aParams = []) {
    $sSql = $this->getSQL();
    // Utilisation de setParameter pour binder les types classiques
    // $aParams est passé par référence il pourra changer aprés cette fonction
    $sSql = $this->prepareNativeParameters($sSql, $aParams);
    // Finalisation de la requête avec les autre champs et les binds de préparation
    $sSql = $this->prepareSQLWithParams($sSql, $aParams);

    // SELECT ou utilisation de Returning
    if ($this->getType() === self::SELECT || !empty($this->aReturningFields)) {
      return $this->getConnection()->executeQuery($sSql, $this->getParameters(), $this->getParameterTypes());
    }
    // retourn le nombre de lignes modifiées
    return $this->getConnection()->executeStatement($sSql, $this->getParameters(), $this->getParameterTypes());
  }

  /**
   * On force notre execute tout le temps car il gère le returning ce que ne fait pas
   * celui du querybuilder de base
   * {@inheritdoc}
   * @return Result|int|string
   */
  public function execute() {
    return $this->executeWithParams();
  }

  /**
   * remplace certains éléments dans une partie spécifique du SQL d'un QueryBuilder
   * @param string $sSql  SQL à préparer
   * @param array $aParams paramètre à utiliser pour préparer la requête
   * @return string  sql préparé
   */
  private function prepareSQLWithParams(string $sSql, array $aParams): string {
    foreach ($aParams as $sKey => $aData) {
      $sReplacerKey = '[' . $sKey . ']';
      switch ($aData['type']) {
        case 'column_name':
        case 'table_name':
        case 'schema_name':
          // vérification de la valeur avec la regexp
          if (preg_match('/^[\p{L}\p{N}@$#_]{0,127}$/', $aData['value'])) {
            // Ajout de guillemets pour les noms de schémas et de tables qui peuvent contenir des majuscules.
            if ($aData['value'][0] !== '"' && $aData['value'][strlen($aData['value']) - 1] !== '"') {
              $aData['value'] = '"' . $aData['value'] . '"';
            }
            // Remplacement dans la reaqête
            $sSql = str_replace($sReplacerKey, $aData['value'], $sSql);
          } else {
            $this->oErrorLogger->error($aData['value'] . ' not match with column_name, schema_name or table_name');
          }
          break;
        case 'geometry':
        case 'quoted_string':
          $sValue = $this->getConnection()->quote($aData['value']);
          $sSql = str_replace($sReplacerKey, $sValue, $sSql);
          break;
        case 'double_quote':
          $sValue = '"' . addslashes($aData['value']) . '"';
          $sSql = str_replace($sReplacerKey, $sValue, $sSql);
          break;
        case self::VALUE_TO_PDO_BIND_TYPE:
          $sSql = str_replace($sReplacerKey, ':' . $sKey, $sSql);
          break;
      }
    }

    return $sSql;
  }

  /**
   * Prepare les paramètres classique pour PDO via le querybuilder
   * @param string $sSql
   * @param array $aParams passé par référence
   * @return string liste de paramètre modifié pour retouché le SQL ensuite
   */
  private function prepareNativeParameters(string $sSql, array &$aParams): string {
    $aParamsCopy = [];
    foreach ($aParams as $sKey => $aData) {
      switch ($aData['type']) {
        case 'boolean':
          $sValue = match($aData['value']) {
            true => 'TRUE',
            'true' => 'TRUE',
            1 => 'TRUE',
            false => 'FALSE',
            'false' => 'FALSE',
            0 => 'FALSE',
            default => 'NULL'
          };
          $this->setParameter($sKey, $sValue);
          $aData['type'] = self::VALUE_TO_PDO_BIND_TYPE;
          break;
        case 'integer':
        case 'number':
        case 'string':
        case 'mixed':
          $this->setParameter($sKey, $aData['value']);
          $aData['type'] = self::VALUE_TO_PDO_BIND_TYPE;
          break;
        case 'array':
          if (!is_array($aData['value'])) {
            $aData['value'] = [$aData['value']];
          }
          $aTagReplace = [];
          foreach ($aData['value'] as $mItem) {
            $sTmpKey = $sKey . '_' . str_replace('-', '', Uuid::uuid1()->toString());
            $aTagReplace[] = '[' . $sTmpKey . ']';
            $this->setParameter($sTmpKey, $mItem);
            $aParamsCopy[$sTmpKey] = ['value' => $mItem, 'type' => self::VALUE_TO_PDO_BIND_TYPE];
          }
          // remplacement du bind d'origine par une liste de bind dans la requête
          $sSql = str_replace('[' . $sKey . ']', implode(', ', $aTagReplace), $sSql);
      }
      $aParamsCopy[$sKey] = $aData;
    }
    // mise à jour $aParams pour la suite
    $aParams = $aParamsCopy;

    return $sSql;
  }

  /**
   * Specifies an item that is to be returned in the query result.
   * Replaces any previously specified selections, if any.
   *
   * USING AN ARRAY ARGUMENT IS DEPRECATED. Pass each value as an individual argument.
   *
   * <code>
   *     $qb = $conn->createQueryBuilder()
   *         ->insert(...)
   *          ...
   *         ->returning('p.field1', 'p.field2')
   * </code>
   *
   * @param string|string[]|null $fields The selection expression.
   *
   * @return $this This QueryBuilder instance.
   */
  public function returning($fields = null): self {
    // ne peut être set que pour un requête INSERT ou UPDATE
    if (!in_array($this->getType(), [self::INSERT, self::UPDATE])) {
      return $this;
    }

    if ($fields === null) {
      $fields = [];
    }

    // si $fields est un tableau on l'utilisera directements sinon on recupèrera tous les arguments sous forme de tableau
    $this->aReturningFields = is_array($fields) ? $fields : func_get_args();

    return $this;
  }

  /**
   * Permet de définir les valeurs null en premier sur le order by
   * valable uniquement pour un select
   * @return self
   */
  public function nullsFirst(): self {
    if ($this->getType() === self::SELECT) {
      $this->sNullPosition = 'FIRST';
    }
    return $this;
  }

  /**
   * Permet de définir les valeurs null en dernier sur le order by
   * valable uniquement pour un select
   * @return self
   */
  public function nullsLast(): self {
    if ($this->getType() === self::SELECT) {
      $this->sNullPosition = 'LAST';
    }
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getSQL(): string {
    // si un ordre est spécifié pour les null,
    // on retouche le order by en conséquence avant de générer la requête avec le parent
    if (!empty($this->sNullPosition)) {
      $this->updateQbOrderByWithNullPosition();
    }

    $sSql = parent::getSQL();

    // gère le returning si présent et si le type correspond
    if (!empty($this->aReturningFields)) {
      $sSql .= ' RETURNING ' . implode(', ', $this->aReturningFields);
    }

    return $sSql;
  }

  /**
   * Met à jour le orderBy de qb pour tenir compte de la position des nulls dans l'ordre
   * @return void
   */
  private function updateQbOrderByWithNullPosition(): void {
    // récupération du orderby d'origine
    $mOrderBy = $this->getQueryPart('orderBy');
    // L'orderby peut être un array ou une string
    if (is_array($mOrderBy)) {
      foreach ($mOrderBy as &$sOrderBy) {
        $sOrderBy .= ' NULLS ' . $this->sNullPosition;
      }
    } else if (is_string($mOrderBy)) {
      $mOrderBy .= ' NULLS ' . $this->sNullPosition;
    }
    // on écrase le order by parent avec le orderby set
    $this->add('orderBy', $mOrderBy, false);
  }
}
