<?php

namespace App\Module\Vitis\WebService\Service\FileAccessor;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Crypto\JweCrypto;
use App\Module\Vitis\WebService\Service\FileAccessor\FileAccessorInterface;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use Monolog\Logger;
use Ramsey\Uuid\Uuid;
use Symfony\Component\Mime\MimeTypes;

/**
 * Interface utilisé pour interagir avec les fichiers sur le disque.
 * Les retours d'erreur sont uniforme null signifie que la fonction à rencontrer un erreur
 * Pour accéder à la doc des fonction de l'interface il faut regarder l'interface FileAccessor:
 * App\Module\Vitis\WebService\Service\FileAccessor\FileAccessorInterface
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */

class LocalAccessor implements FileAccessorInterface {
  /**
   * @var int
   */
  protected const MAX_FILE_SIZE_FOR_HASH = 307200;
  /**
   * @var int
   */
  protected const MAX_FILE_TO_ARCHIVE = 255;
  /**
   * @var string
   */
  protected const API_ROUTE_DOWNLOAD = '/vitis/downloads';
  /**
   * @var int
   */
  public const DEFAULT_LINUX_PERMISSION = 0744;

  /**
   * @var Properties
   */
  protected $oProperties;
  /**
   * @var \Monolog\Logger
   */
  protected $oLogger;
  /**
   * @var JweCrypto
   */
  protected $oJweCrypto;
  /**
   * @var MimeTypes
   */
  private $oMimeTypeResolver;

  /**
   * @param Properties $properties service d'accés aux properties
   * @param LoggerManager $oLogger service d'accés aux logger
   * @param JweCrypto $oJweCrypto service de crypto
   */
  public function __construct(Properties $properties, LoggerManager $oLogger, JweCrypto $oJweCrypto) {
    $this->oProperties = $properties;
    $this->oLogger = $oLogger->get('API_ERROR');
    $this->oMimeTypeResolver = new MimeTypes();
    $this->oJweCrypto = $oJweCrypto;
  }

  /**
   * @param int $iFileSize taille du fichier à rendre lisible
   * @return string taille du fichier lisible par un humain
   */
  protected static function fileSizeToHuman(int $iFileSize): string {
    $sFileSize = $iFileSize . ' octets';
    // si plus grand que 1Ko
    if ($iFileSize > 1024) {
      // si plus grand que 1 Mo
      if ($iFileSize > 1024 * 1024) {
        // si plus grand que 1Go
        if ($iFileSize > 1024 * 1024 * 1024) {
          // Si plus grand que 1 To (on en reparlera)
          if ($iFileSize > 1024 * 1024 * 1024 * 1024) {
            $sFileSize = '-';
          } else {
            $sFileSize = (ceil($iFileSize / (1024 * 1024 * 1024))) . 'Go';
          }
        } else {
          $sFileSize = (ceil($iFileSize / (1024 * 1024))) . 'Mo';
        }
      } else {
        $sFileSize = (ceil($iFileSize / 1024)) . 'Ko';
      }
    }
    return $sFileSize;
  }
  /**
   * @param string $sFilePath nom/vhemin d'un fichier
   * @return string MIMEType du fichier
   */
  protected function fileExtToMimeType(string $sFilePath): string {
    $sExt = strtolower($this->getExtension($sFilePath));
    $aMimeTypes = $this->oMimeTypeResolver->getMimeTypes($sExt);
    if (!empty($aMimeTypes)) {
      return $aMimeTypes[0];
    } else {
      return 'application/octet-stream';
    }
  }

  /**
   * Calcul le couple largeur/hauteur pour redimensionner une image en essayant de préserver le ratio.
   * @param int $iCurrentWidth Width of the picture.
   * @param int $iCurrentHeight Height of the picture.
   * @param int $iTargetWidth Maximum final width of the picture.
   * @param int $iTargetHeight Maximum final height of the picture.
   * @return array $aSize couple width, height pour redimensionner une image.
   */
  private static function sizeCalculator(int $iCurrentWidth, int $iCurrentHeight, int $iTargetWidth, int $iTargetHeight): array{
    $iWidth = $iTargetWidth;
    $iHeight = $iTargetHeight;
    // si image plus grande on retourne les cibles
    if ($iCurrentWidth > $iTargetWidth && $iCurrentHeight > $iTargetHeight) {
      return [$iTargetWidth, $iTargetHeight];
    }
    // On détermine si on conserve la largeur ou la hauteur
    if ($iCurrentWidth > $iTargetWidth && $iCurrentHeight < $iTargetHeight) {
      $iHeight = $iCurrentHeight * $iTargetWidth / $iCurrentWidth;
      $iWidth = $iTargetWidth;
    } else if ($iCurrentWidth < $iTargetWidth && $iCurrentHeight > $iTargetHeight) {
      $iHeight = $iTargetHeight;
      $iWidth = $iCurrentWidth * $iTargetHeight / $iCurrentHeight;
    } else {
      // dans ce cas les les deux distance sont plus petite que le but on
      // calcul donc le delta pour déterminer quel coté est gardé
      $fDeltaWidth = $iCurrentWidth / $iTargetWidth;
      $fDeltaHeight = $iCurrentHeight / $iTargetHeight;
      if ($fDeltaWidth < $fDeltaHeight) {
        $iHeight = $iCurrentHeight;
        $iWidth = $iCurrentHeight * $iTargetWidth / $iTargetHeight;
      } else {
        $iWidth = $iCurrentWidth;
        $iHeight = $iCurrentWidth * $iTargetHeight / $iTargetWidth;
      }
    }

    return [$iWidth, $iHeight];
  }

  /**
   * Génére le chemin de la miniature d'une image à partir du chemin de l'image d'origine.
   * @param string $sFilePath Chemin de l'image d'origine.
   * @return string chemin de la miniature pour la génération.
   */
  public static function thumbnailPathGenerator(string $sFilePath): string {
    $sMinPath = $sFilePath;
    if (strpos($sFilePath, ".") > -1) {
      $aPath = explode(".", $sMinPath);
      array_splice($aPath, -1, 1, ['min', 'jpg']);
      $sMinPath = implode('.', $aPath);
    }
    return $sMinPath;
  }

  /****************************************************************
   *                                                              *
   *                            Files                             *
   *                     PHP native functions                     *
   *                                                              *
   ****************************************************************/

  /**
   * {@inheritdoc}
   */
  public function file_exists(string $sFilePath): bool {
    return file_exists($sFilePath);
  }
  /**
   * {@inheritdoc}
   */
  public function file_put_contents(string $sFilePath, $sData, int $iFlags = 0, $mContext = null): ?int {

    try {
      $mReturn = file_put_contents($sFilePath, $sData, $iFlags, $mContext);
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }

    return ($mReturn === false) ? null : $mReturn;
  }
  /**
   * {@inheritdoc}
   */
  public function file_get_contents(string $sFilePath, bool $bUseIncludePath = false, $mContext = null, int $iOffset = 0, int $iMaxLen = -1): ?string{
    $mReturn = false;
    try {
      if ($iMaxLen == -1) {
        $mReturn = file_get_contents($sFilePath, $bUseIncludePath, null, $iOffset);
      } else {
        $mReturn = file_get_contents($sFilePath, $bUseIncludePath, null, $iOffset, $iMaxLen);
      }
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }

    return ($mReturn === false) ? null : $mReturn;
  }
  /**
   * {@inheritdoc}
   */
  public function copy(string $sSourceFilePath, string $sDestFilePath): ?bool {
    try {
      $mReturn = copy($sSourceFilePath, $sDestFilePath);
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }

    return ($mReturn === false) ? null : $mReturn;
  }
  /**
   * {@inheritdoc}
   */
  public function filesize(string $sFilePath): ?int {
    try {
      $mReturn = filesize($sFilePath);
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
    return ($mReturn === false) ? null : $mReturn;
  }
  /**
   * {@inheritdoc}
   */
  public function readfile(string $sFilePath, bool $bUseIncludePath = false, $mContext = null): ?int {
    try {
      $mReturn = readfile($sFilePath, $bUseIncludePath, $mContext);
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
    return ($mReturn === false) ? null : $mReturn;
  }
  /**
   * {@inheritdoc}
   */
  public function rename(string $sSourceFilePath, string $sDestFilePath, $mContext = null): ?bool{
    $mReturn = false;
    try {
      if (empty($mContext)) {
        $mReturn = rename($sSourceFilePath, $sDestFilePath);
      } else {
        $mReturn = rename($sSourceFilePath, $sDestFilePath, $mContext);
      }
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }

    return ($mReturn === false) ? null : $mReturn;
  }
  /**
   * {@inheritdoc}
   */
  public function unlink(string $sFilePath, $mContext = null): ?bool{
    $mReturn = false;
    try {
      if (empty($mContext)) {
        $mReturn = unlink($sFilePath);
      } else {
        $mReturn = unlink($sFilePath, $mContext);
      }
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }

    return ($mReturn === false) ? null : $mReturn;
  }
  /**
   * {@inheritdoc}
   */
  public function filemtime(string $sFilePath): ?int {
    try {
      $mReturn = filemtime($sFilePath);
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }

    return ($mReturn === false) ? null : $mReturn;
  }
  /**
   * {@inheritdoc}
   */
  public function mime_content_type(string $sFilePath): ?string {
    if (!$this->file_exists($sFilePath)) {
      return null;
    }
    return $this->fileExtToMimeType($sFilePath);
  }
  /**
   * {@inheritdoc}
   */
  public function realpath(string $sFilePath): ?string {
    try {
      $mReturn = realpath($sFilePath);
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }

    return ($mReturn === false) ? null : str_replace('\\', '/', $mReturn);
  }
  /****************************************************************
   *                                                              *
   *                            Files                             *
   *                       Veremes functions                      *
   *                                                              *
   ****************************************************************/

  /**
   * {@inheritdoc}
   */
  public function getFileInfos(string $sFilePath, array $aKeepOnlyField = []): ?array{
    if (!$this->file_exists($sFilePath)) {
      return null;
    }

    $iFileSize = $this->filesize($sFilePath);
    $sFileSize = $this->fileSizeToHuman($iFileSize);
    $sFileExtention = $this->getExtension($sFilePath);
    $sFileName = $this->getFileName($sFilePath);

    $aFileInfos = [
      'filename' => $sFileName,
      'extention' => $sFileExtention,
      'size' => $sFileSize,
      'size_' => $iFileSize,
      'lastModification' => $this->filemtime_formated($sFilePath, 'd/m/Y H:i:s'),
      'lastModification_' => $this->filemtime_formated($sFilePath, 'Uu'),
      'href' => $this->getProxyPassUrl($sFilePath),
    ];

    return $this->getFileInfosKeyFilter($aFileInfos, $aKeepOnlyField);
  }

  /**
   * tri les infos du fichier
   * @param array $aFileInfos Infos du fichier.
   * @param array<string> $aKeepOnlyField Liste des attributs à conserver (si vide retourne tous les attributs).
   */
  protected function getFileInfosKeyFilter(array $aFileInfos, array $aKeepOnlyField = []): array{
    // tri des clés
    $aFileInfosReturn = [];
    if (!empty($aKeepOnlyField)) {
      foreach ($aKeepOnlyField as $sKey) {
        if (isset($aFileInfos[$sKey])) {
          $aFileInfosReturn[$sKey] = $aFileInfos[$sKey];
        }
      }
    } else {
      $aFileInfosReturn = $aFileInfos;
    }

    return $aFileInfosReturn;
  }

  /**
   * {@inheritdoc}
   */
  public function getExtension(string $sFilePath): string {
    $oFileInfo = new \SplFileInfo($sFilePath);
    return $oFileInfo->getExtension();
  }
  /**
   * {@inheritdoc}
   */
  public function getFileName(string $sFilePath): string {
    $oFileInfo = new \SplFileInfo($sFilePath);
    return $oFileInfo->getFilename();
  }
  /**
   * {@inheritdoc}
   */
  public function getProxyPassUrl(string $sFilePath): ?string {
    if (!$this->file_exists($sFilePath) || $this->is_dir($sFilePath)) {
      return null;
    }
    return $this->getEncryptedKeyUrl($sFilePath);
  }
  /**
   * retourne une url d'accés au fichier avec un token crypter.
   * @param string $sFilePath Chemin du fichier.
   * @return string
   */
  protected function getEncryptedKeyUrl(string $sFilePath, string $sFs = '', string $sEtag = ''): string {
    $sFs = (empty($sFs)) ? Properties::LOCAL_FS : $sFs;
    // pattern d'url à utiliser
    $sDataUrl = $this->getUrlHost() . '?file=[FILE]';
    $aFileStruct = [
      'key' => $sFilePath,
      'fs' => $sFs,
    ];
    if (!empty($sEtag)) {
      $aFileStruct['etag'] = $sEtag;
    }
    // donnée à bind
    $aData = [
      'FILE' => urlencode($this->oJweCrypto->encryptData($aFileStruct)),
    ];
    // url bindé
    return $this->generateUrl($sDataUrl, $aData);
  }
  /**
   * retourne une url d'accés au fichier avec controle du hash du fichier récupérer au préalable.
   * @param string $sFilePath Chemin du fichier.
   * @param string $sEtag Etag S3.
   * @return string
   * @deprecated
   */
  protected function getProxyPassUrlWithEtag(string $sFilePath, string $sEtag): string {
    // pattern d'url à utiliser
    $sDataUrl = $this->getUrlHost() . '?key=[KEY]&eTag=[ETAG]';
    // donnée à bind
    $aData = [
      'KEY' => urlencode($this->getUrlKey($sFilePath)),
      'ETAG' => $sEtag,
    ];
    // url bindé
    return $this->generateUrl($sDataUrl, $aData);
  }
  /**
   * retourne le host et la route vers la ressource filedownloader
   * @return string valeur de l'url
   */
  protected function generateUrl(string $sPattern, array $aData): string {
    $date = new \DateTime();
    foreach ($aData as $sKey => $sValue) {
      $sPattern = str_replace('[' . $sKey . ']', $sValue, $sPattern);
    }
    $sPattern .= '&d=' . $date->getTimestamp();
    return $sPattern;
  }
  /**
   * retourne le host et la route vers la ressource filedownloader
   * @return string valeur de l'url
   */
  protected function getUrlHost(): string {
    $sHost = $this->oProperties->get('app_api_url');
    return $sHost . self::API_ROUTE_DOWNLOAD;
  }
  /**
   * retourne la valeur à utiliser comme clé pour un fichier.
   * @param string $sFilePath Chemin du fichier.
   * @return string
   */
  protected function getUrlKey(string $sFilePath): string {
    $sKey = str_replace($this->oProperties->get('ws_data_dir'), 'ws_data', $sFilePath);
    $sKey = str_replace($this->oProperties->get('public_dir'), 'public', $sKey);
    $sKey = str_replace($this->oProperties->get('upload_dir'), 'upload', $sKey);
    $sKey = str_replace($this->oProperties->get('shared_dir'), 'shared', $sKey);
    return $sKey;
  }
  /**
   * {@inheritdoc}
   */
  public function getFileEtag(string $sFilePath): ?string {
    if (file_exists($sFilePath)) {
      return sha1(file_get_contents($sFilePath, false, null, 0, self::MAX_FILE_SIZE_FOR_HASH), false);
    } else {
      return null;
    }
  }
  /**
   * {@inheritdoc}
   */
  public function file_get_contents_with_etag(string $sFilePath, string $sEtag, bool $bUseIncludePath = false, $mContext = null, int $iOffset = 0, int $iMaxLen = -1): ?string{
    $sProcessedEtag = $this->getFileEtag($sFilePath);
    if ($sEtag == $sProcessedEtag) {
      return $this->file_get_contents($sFilePath, $bUseIncludePath, $mContext, $iOffset, $iMaxLen);
    } else {
      return null;
    }
  }
  /**
   * {@inheritdoc}
   */
  public function filemtime_formated(string $sFilePath, string $sFormat): ?string{
    $iUnixFilemtime = $this->filemtime($sFilePath);
    return (empty($iUnixFilemtime)) ? null : date($sFormat, $iUnixFilemtime);
  }
  /**
   * Alias de Copy
   * {@inheritdoc}
   */
  public function copyLocalToDistant(string $sSourceFilePath, string $sDestFilePath): ?bool{
    $sDirPath = dirname($sDestFilePath);
    if (!$this->file_exists($sDirPath)) {
      if (!$this->mkdir($sDirPath, self::DEFAULT_LINUX_PERMISSION, true)) {
        return null;
      }
    }
    return $this->copy($sSourceFilePath, $sDestFilePath);
  }
  /**
   * Alias de Copy
   * {@inheritdoc}
   */
  public function copyDistantToLocal(string $sSourceFilePath, string $sDestFilePath): ?bool {
    return $this->copyLocalToDistant($sSourceFilePath, $sDestFilePath);
  }

  /**
   * Retourne un chemin vers un dossier temporaire unique sur le disque local
   * @return string
   */
  public function getTempDir(): string {
    return $this->oProperties->get('extract_dir') . '/' . Uuid::uuid1()->toString();
  }

  /**
   * {@inheritdoc}
   */
  public function pictureResampler(string $sFilePath, int $iDstWidth, int $iDstHeight, int $iCropX = 0, int $iCropY = 0, int $iOffsetX = 0, int $iOffsetY = 0, ?string $sType = null): ?string{
    // récupération de l'extension pour ouverture du fichier
    $sType = (empty($sType)) ? strtolower($this->getExtension($sFilePath)) : $sType;

    // Récupération du format de la source
    list($iCurrentWidth, $iCurrentHeight) = \getimagesize($sFilePath);
    list($width, $height) = $this->sizeCalculator($iCurrentWidth, $iCurrentHeight, $iDstWidth, $iDstHeight);
    try {
      // Ouverture de la ressource return false si format non réconnu
      switch ($sType) {
        case 'bmp':
          $img = \imagecreatefromwbmp($sFilePath);
          break;
        case 'gif':
          $img = \imagecreatefromgif($sFilePath);
          break;
        case 'jpg':
        case 'jpeg':
          $img = \imagecreatefromjpeg($sFilePath);
          break;
        case 'png':
          $img = \imagecreatefrompng($sFilePath);
          break;
        default:
          throw new \ErrorException('pictureResampler(' . $sFilePath . '): Unknown type of picture');
      }
      //Suppression du fichier sinon problème de réouverture après si jpg et doublon si autre format
      $this->unlink($sFilePath);

      if ($img === false) {
        throw new \ErrorException('pictureResampler(' . $sFilePath . '): Cannot read this picture');
      }
      // création de l'image de sortie
      $image_p = \imagecreatetruecolor($width, $height);
      $white = \imagecolorallocate($image_p, 255, 255, 255);
      \imagefill($image_p, 0, 0, $white);
      // resizer
      if (!\imagecopyresampled($image_p, $img, $iOffsetX, $iOffsetY, $iCropX, $iCropY, $width, $height, $iCurrentWidth, $iCurrentHeight)) {
        throw new \ErrorException('pictureResampler(' . $sFilePath . '): Cannot resampling the picture');
      }
      // transformation au format jpg pour supprimer l'alpha et gagner en taille du fichier
      $dst = $sFilePath;
      if (strpos($sFilePath, ".") > -1) {
        $aPath = explode(".", $sFilePath);
        $aPath[count($aPath) - 1] = "jpg";
        $dst = implode(".", $aPath);
      }

      // Enregistrement du jpeg
      if (!\imagejpeg($image_p, $dst)) {
        throw new \ErrorException('pictureResampler(' . $sFilePath . '): Cannot write JPG picture (' . $dst . ')');
      }
      // Libération espace mémoire
      \imagedestroy($image_p);
      \imagedestroy($img);
      // retour du nouveau nom de fichier
      return $dst;
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function thumbnailResampler(string $sFilePath): ?string {
    try {
      if (!$this->file_exists($sFilePath)) {
        throw new \ErrorException('thumbnailResampler(' . $sFilePath . '): File doesn\'t exist');
      }
      $sExt = $this->getExtension($sFilePath);
      $sMinPath = $this->thumbnailPathGenerator($sFilePath);

      if (!$this->copy($sFilePath, $sMinPath)) {
        throw new \ErrorException('thumbnailResampler(' . $sFilePath . '): Cannot copy this picture');
      }
      $sMinPath = $this->pictureResampler($sMinPath, $this->oProperties->get('thumbnails_max_width'), $this->oProperties->get('thumbnails_max_height'), 0, 0, 0, 0, $sExt);

      if (empty($sMinPath)) {
        throw new \ErrorException('thumbnailResampler(' . $sFilePath . '): Cannot resample this picture');
      }
      return $sMinPath;
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function checkFileSize(string $sFilePath, int $iLimitation = -1): bool {
    $iLimitation = ($iLimitation < 0) ? $this->oProperties->get('max_upload_file_size_restricted') : $iLimitation;
    // Vérification de la taille du fichier.
    if ($this->filesize($sFilePath) > $iLimitation) {
      $this->oLogger->error('the size of the file is greater than the maximum size allowed (' . $sFilePath . ')');
      return false;
    }
    return true;
  }

  /**
   * {@inheritdoc}
   */
  public function checkExtension(string $sFilePath): bool {

    $aForbiddenExt = $this->oProperties->get('forbidden_extension');

    if (in_array($this->getExtension($sFilePath), $aForbiddenExt)) {
      $this->oLogger->error('this extension of file is not allowed (' . $sFilePath . ')');
      return false;
    }
    return true;
  }

  /****************************************************************
   *                                                              *
   *                            Folders                           *
   *                     PHP native functions                     *
   *                                                              *
   ****************************************************************/

  /**
   * {@inheritdoc}
   */
  public function scandir(string $sDirPath, int $iSortOrder = SCANDIR_SORT_ASCENDING, $mContext = null): ?array{
    $mReturn = false;
    try {
      if (empty($mContext)) {
        $mReturn = scandir($sDirPath, $iSortOrder);
      } else {
        $mReturn = scandir($sDirPath, $iSortOrder, $mContext);
      }
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }

    return ($mReturn === false) ? null : $mReturn;
  }
  /**
   * {@inheritdoc}
   */
  public function mkdir(string $sDirPath, int $iMode = self::DEFAULT_LINUX_PERMISSION, bool $bRecursive = false, $mContext = null): ?bool{
    $mReturn = false;
    try {
      if (empty($mContext)) {
        $mReturn = mkdir($sDirPath, $iMode, $bRecursive);
      } else {
        $mReturn = mkdir($sDirPath, $iMode, $bRecursive, $mContext);
      }
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }

    return ($mReturn === false) ? null : $mReturn;
  }
  /**
   * {@inheritdoc}
   */
  public function rmdir(string $sDirPath, $mContext = null): ?bool{
    $mReturn = false;
    try {
      if (empty($mContext)) {
        $mReturn = rmdir($sDirPath);
      } else {
        $mReturn = rmdir($sDirPath, $mContext);
      }
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
    return ($mReturn === false) ? null : $mReturn;
  }
  /**
   * {@inheritdoc}
   */
  public function is_dir(string $sDirPath): bool {
    return is_dir($sDirPath);
  }

  /****************************************************************
   *                                                              *
   *                            Folders                           *
   *                       Veremes functions                      *
   *                                                              *
   ****************************************************************/

  /**
   * {@inheritdoc}
   */
  public function getFolderInfos(string $sDirPath, array $aKeepOnlyExt = [], array $aKeepOnlyField = [], bool $bRemoveEmptyFolder = false, int $iDepth = -1): ?array{
    if ($this->is_dir($sDirPath)) {
      $aContentDir = $this->scandir($sDirPath);
      $aTree = $this->getFileInfos($sDirPath, $aKeepOnlyField);
      $aDir = [];
      for ($i = 0; $i < count($aContentDir); $i++) {
        if ($iDepth === -1 || $iDepth > 0) {
          if ($aContentDir[$i] !== '.' && $aContentDir[$i] !== '..') {
            $aItemInfos = $this->getFolderInfos($sDirPath . '/' . $aContentDir[$i], $aKeepOnlyExt, $aKeepOnlyField, $bRemoveEmptyFolder, ($iDepth === -1)? $iDepth : $iDepth-1);
            if (!empty($aItemInfos)) {
              if (!isset($aItemInfos['content']) || !$bRemoveEmptyFolder) {
                $aDir[] = $aItemInfos;
              }
              else if (!empty($aItemInfos['content'])) {
                $aDir[] = $aItemInfos;
              }
            }
          }
        }
      }

      $aTree['content'] = $aDir;

      if (isset($aTree['href'])) {
        unset($aTree['href']);
      }

      return $aTree;
    } else {
      if (empty($aKeepOnlyExt) || in_array(strtolower($this->getExtension($sDirPath)), $aKeepOnlyExt)) {
        return $this->getFileInfos($sDirPath, $aKeepOnlyField);
      } else {
        return null;
      }
    }
  }
  /**
   * {@inheritdoc}
   */
  public function clearDir(string $sDirPath): ?bool {
    if (!is_dir($sDirPath)) {
      return null;
    }

    $aContent = array_diff(scandir($sDirPath), ['.', '..']);

    foreach ($aContent as $sFichier) {
      $sPath = $sDirPath . '/' . $sFichier;
      if (is_dir($sPath)) {
        if (!$this->clearDir($sPath)) {
          return null;
        }
      } else {
        if (!unlink($sPath)) {
          return null;
        }
      }
    }
    // utilisation du rmdir de la classe pour avoir le retour bool|null
    return $this->rmdir($sDirPath);
  }
  /**
   * {@inheritdoc}
   */
  public function copyDirectory(string $sDirPathSrc, string $sDirPathDest): ?bool {
    if (!is_dir($sDirPathSrc)) {
      return null;
    }

    if (!is_dir($sDirPathDest)) {
      $mReturn = $this->mkdir($sDirPathDest);
      if (empty($mReturn)) {
        return null;
      }
    }

    $aContent = array_diff(scandir($sDirPathSrc), ['.', '..']);

    foreach ($aContent as $sFichier) {
      $sPath = $sDirPathSrc . '/' . $sFichier;
      if (is_dir($sPath)) {
        if (empty($this->copyDirectory($sDirPathSrc . '/' . $sFichier, $sDirPathDest . '/' . $sFichier))) {
          return null;
        }
      } else {
        if (empty(copy($sDirPathSrc . '/' . $sFichier, $sDirPathDest . '/' . $sFichier))) {
          return null;
        }
      }
    }

    return true;
  }
  /**
   * Alias de CopyDirectory
   * {@inheritdoc}
   */
  public function copyDirLocalToDistant(string $sDirPathSrc, string $sDirPathDest, array $aIgnoredItems = []): ?bool {
    return $this->copyDirectory($sDirPathSrc, $sDirPathDest);
  }
  /**
   * Alias de CopyDirectory
   * {@inheritdoc}
   */
  public function copyDirDistantToLocal(string $sDirPathSrc, string $sDirPathDest): ?bool {
    return $this->copyDirectory($sDirPathSrc, $sDirPathDest);
  }
  /**
   * {@inheritdoc}
   */
  public function createZip(string $sFolder, string $sDestination, string $sExtensionToExclude = '', string $sPassword = '', bool $bDeleteFiles = false): ?bool {
    try {
      if (!file_exists($sFolder)) {
        throw new \ErrorException('createZip(' . $sFolder . '): No such file or directory');
      }

      if (!empty($sExtensionToExclude)) {
        $aExtensionToExclude = explode('|', $sExtensionToExclude);
      } else {
        $aExtensionToExclude = [];
      }
      $sFolder = rtrim($sFolder, '/');
      $zip = new \ZipArchive();

      if ($zip->open($sDestination, \ZipArchive::CREATE | \ZipArchive::OVERWRITE) !== true) {
        throw new \ErrorException('createZip(' . $sFolder . '): An error occurred creating your ZIP file.');
      }

      $files = new \RecursiveIteratorIterator(
        new \RecursiveDirectoryIterator($sFolder, \FilesystemIterator::UNIX_PATHS),
        \RecursiveIteratorIterator::LEAVES_ONLY
      );
      $aCompressedFiles = [];
      $iTmpCounter = 0;
      foreach ($files as $name => $file) {
        if (!$file->isDir()) {
          // modification des chemin sur windows
          $filePath = str_replace('\\', '/', $file->getRealPath());

          // Récupération de l'extension
          $sExtension = $this->getExtension($filePath);

          $relativePath = substr($filePath, strlen($sFolder) + 1);
          if (in_array($sExtension, $aExtensionToExclude)) {
            continue;
          }
          // la compression ne se fait qu'au moment de la fermeture du zip
          // on ne peut pas dépasser un certain nombre de fichier en attente dans un zip
          // dans ce cas là il faut fermer le zip et le rouvrir pour lancer la compression et poursuivre l'ajout des fichiers
          // https://stackoverflow.com/a/22638907
          if ($iTmpCounter >= self::MAX_FILE_TO_ARCHIVE) {
            $zip->close();
            if ($zip->open($sDestination)) {
              $iTmpCounter = 0;
            } else {
              throw new \ErrorException('Cannot reopen Archive : ' . $sDestination . ' (number of file in archive exceeded)');
            }
          }

          if ($zip->addFile($filePath, $relativePath)) {
            $aCompressedFiles[] = $filePath;
            $iTmpCounter++;
          }
          // Mot de passe.
          if (!empty($sPassword)) {
            $zip->setEncryptionName($this->getFileName($filePath), \ZipArchive::EM_AES_256, $sPassword);
          }

        }
      }

      if ($zip->close()) {
        // Suppression des fichiers compressés.
        if ($bDeleteFiles) {
          foreach ($aCompressedFiles as $sFilePath) {
            if (file_exists($sFilePath)) {
              unlink($sFilePath);
            }

          }
        }
      }
    } catch (\Exception $e) {
      //$zip->close();
      $this->oLogger->error($e->__toString());
      return null;
    }
    $mReturn = file_exists($sDestination);
    return ($mReturn === false) ? null : $mReturn;
  }
  /**
   * {@inheritdoc}
   */
  public function encryptZip(string $sZipFilePath, string $sPassword): ?bool {
    try {
      if (!file_exists($sZipFilePath)) {
        throw new \ErrorException('encryptZip(' . $sZipFilePath . '): No such file');
      }

      $zip = new \ZipArchive();

      if ($zip->open($sZipFilePath) !== true) {
        throw new \ErrorException('encryptZip(' . $sZipFilePath . '): An error occurred while opening your ZIP file.');
      }

      for ($i = 0; $i < $zip->numFiles; $i++) {
        $zip->setEncryptionIndex($i, \ZipArchive::EM_AES_256, $sPassword);
      }

      return $zip->close();
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }
  /**
   * {@inheritdoc}
   */
  public function unZip(string $sZipFile, string $sDestination): ?bool{
    $zip = new \ZipArchive;
    $res = $zip->open($sZipFile);

    if ($this->is_dir($sDestination)) {
      $this->mkdir($sDestination, self::DEFAULT_LINUX_PERMISSION, true);
    }

    if ($res === true) {
      // extract it to the path we determined above
      $zip->extractTo($sDestination);
      $zip->close();
      $mReturn = file_exists($sDestination);
      return ($mReturn === false) ? null : $mReturn;
    } else {
      return null;
    }
  }
  /**
   * Alias de CreateZip
   * {@inheritdoc}
   */
  public function zipLocalDirToDistant(string $sFolder, string $sDestination, string $sExtensionToExclude = '', string $sPassword = '', bool $bDeleteFiles = false): ?bool {
    return $this->createZip($sFolder, $sDestination, $sExtensionToExclude, $sPassword, $bDeleteFiles);
  }
  /**
   * Alias de CreateZip
   * {@inheritdoc}
   */
  public function zipDistantDirToLocal(string $sFolder, string $sDestination, string $sExtensionToExclude = '', string $sPassword = '', bool $bDeleteFiles = false): ?bool {
    return $this->createZip($sFolder, $sDestination, $sExtensionToExclude, $sPassword, $bDeleteFiles);
  }
  /**
   * Alias de unZip
   * {@inheritdoc}
   */
  public function unzipLocalToDistantDir(string $sZipFile, string $sDestination): ?bool {
    return $this->unZip($sZipFile, $sDestination);
  }
  /**
   * Alias de unZip
   * {@inheritdoc}
   */
  public function unzipDistantToLocalDir(string $sZipFile, string $sDestination): ?bool {
    return $this->unZip($sZipFile, $sDestination);
  }
}
