<?php

namespace App\Module\Vitis\WebService\Service\FileAccessor;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Crypto\JweCrypto;
use App\Module\Vitis\WebService\Service\FileAccessor\FileAccessorInterface;
use App\Module\Vitis\WebService\Service\FileAccessor\LocalAccessor;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use Aws\Credentials\CredentialProvider;
use Aws\S3\Exception\S3Exception;
use Aws\S3\S3Client;
use Aws\S3\ObjectUploader;
use Aws\S3\MultipartUploader;
use Monolog\Logger;

/**
 * Interface utilisé pour interagir avec les fichiers sur un bucket S3.
 * Les retours d'erreur sont uniforme null signifie que la fonction à rencontrer un erreur
 * Pour accéder à la doc des fonction de l'interface il faut regarder l'interface FileAccessor:
 * App\Module\Vitis\WebService\Service\FileAccessor\FileAccessorInterface
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */

class S3Accessor extends LocalAccessor implements FileAccessorInterface {

  /**
   * L'envoi par upload transactionnel est obligatoire à partir de 5Go
   * @var int
   */
  private const S3_LARGE_FILE_THRESHOLD = 5368709120;

  /**
   * @var S3Client
   */
  private $oS3Client;
  /**
   * @var string
   */
  private $sS3Bucket = '';
  /**
   * @var string
   */
  private $sS3Prefix = '';
  /**
   * @var string
   */
  private $sMainPath = '';

  /**
   * @param Properties $properties service d'accés aux properties
   * @param LoggerManager $oLogger service d'accés aux logger
   */
  public function __construct(Properties $properties, LoggerManager $oLogger, JweCrypto $oJweCrypto) {
    parent::__construct($properties, $oLogger, $oJweCrypto);

    $oProvider = null;

    switch ($this->oProperties->get('aws_sdk_mode')) {
      case 'ec2_assume_role':
        $oProvider = CredentialProvider::defaultProvider();
        break;
      case 'credentials':
        $oProvider = CredentialProvider::ini($this->oProperties->get('aws_S3_profile'), $this->oProperties->get('aws_credentials_path'));
        $oProvider = CredentialProvider::memoize($oProvider);
        break;
      default:
        break;
    }

    if (!empty($oProvider)) {
      // Création client S3
      $this->oS3Client = new S3Client([
        'version' => 'latest',
        'region' => $this->oProperties->get('aws_S3_region'),
        //'profile'=> $properties['fileS3UploaderProfil'],
        'credentials' => $oProvider,
        'debug' => false,
      ]);
    }

    $this->extractBucketConstant();
  }

  /****************************************************************
   *                                                              *
   *                         S3 Functions                         *
   *                       Veremes functions                      *
   *                                                              *
   ****************************************************************/

  /**
   * Analyse la properties contenant le bucket pour extraire le prefixe s'il est renseigné
   * @return void
   */
  private function extractBucketConstant(): void {
    $sBucket = $this->oProperties->get('aws_S3_bucket');
    $sPrefix = '';

    if (strpos($sBucket, '/') > -1) {
      $aBucket = explode('/', $sBucket);
      $sBucket = $aBucket[0];
      $sPrefix = implode('/', array_slice($aBucket, 1));
    }

    $this->sS3Bucket = $sBucket;
    $this->sS3Prefix = $sPrefix;

    // définition du MainPath
    $sMainPath = $this->oProperties->get('vas_dir');
    $sMainPath .= '/var';
    if (empty($this->sS3Prefix)) {
      $sMainPath .= '/';
    }

    $this->sMainPath = $sMainPath;
  }
  /**
   * Remplace la racine d'un chemin fichier local par le préfixe S3 défini en properties
   * @param string $sPath Chemin sur le disque local
   * @return string clé à utiliser pour le bucket S3
   */
  private function prefixReplacer(string $sPath): string {
    $sPath = str_replace($this->sMainPath, $this->sS3Prefix, $sPath);
    // Suppression du slash de début de ligne.
    if (strpos($sPath, '/') === 0) {
      $sPath = substr($sPath, 1);
    }
    // remplacement des // dans les chemin
    $sPath = str_replace('//', '/', $sPath);

    return $sPath;
  }
  /**
   * Remplace la racine d'un chemin fichier local par le préfixe S3 défini en properties
   * @param array $aParam Paramètre pour le head
   * @return array|null résultat du head
   */
  private function headObject(array $aParam): ?array{
    try {
      $oResult = $this->oS3Client->headObject($aParam);
      return $oResult->toArray();
    } catch (S3Exception $e) {
      $this->oLogger->debug($e->__toString());
      return null;
    }
  }

  /****************************************************************
   *                                                              *
   *                            Files                             *
   *                     PHP native functions                     *
   *                                                              *
   ****************************************************************/

  /**
   * Utilise headObject pour tester l'existence du fichier
   * {@inheritdoc}
   */
  public function file_exists(string $sFilePath): bool {
    if ($this->is_dir($sFilePath)) {
      return true;
    }

    return $this->checkFileExists($sFilePath);;
  }

  /**
   * Va tester le chemin sans vérifier s'il s'agit d'un dossier
   *
   * @param  string $sFilePath
   * @return bool
   */
  protected function checkFileExists(string $sFilePath): bool {
    // redéfinition du chemin fichier
    $sFilePath = $this->prefixReplacer($sFilePath);
    $mReturn = $this->headObject(
      [
        'Bucket' => $this->sS3Bucket,
        'Key' => $sFilePath,
      ]
    );
    return !empty($mReturn);
  }

  /**
   * Créé le fichier dans un dossier temporaire avant de l'envoyer sur S3 avec putObject.
   * {@inheritdoc}
   */
  public function file_put_contents(string $sFilePath, $sData, int $iFlags = 0, $mContext = null): ?int{
    $sFilePath = $this->prefixReplacer($sFilePath);
    $sPath = '';
    // si data est vide on créer un fichier vide à envoyer sur le serveur
    if (empty($sData)) {
      $sPath = $this->getTempDir();
      $sData = fopen($sPath, 'w+');
    }

    try {
      $this->oS3Client->putObject([
        'Bucket' => $this->sS3Bucket,
        'Key' => $sFilePath,
        'Body' => $sData,
      ]
      );
    } catch (S3Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
    if (!empty($sPath)) {
      parent::unlink($sPath);
      return 1;
    } else {
      return strlen($sData);
    }
  }
  /**
   * Télécharge le fichier dans un dossier temporaire et l'ouvre avec l'interface locale
   * {@inheritdoc}
   */
  public function file_get_contents(string $sFilePath, bool $bUseIncludePath = false, $mContext = null, int $iOffset = 0, int $iMaxLen = -1): ?string{
    $sFilePath = $this->prefixReplacer($sFilePath);
    $sPath = $this->getTempDir();

    try {
      $this->oS3Client->getObject([
        'Bucket' => $this->sS3Bucket,
        'Key' => $sFilePath,
        'SaveAs' => $sPath,
      ]
      );
    } catch (S3Exception $e) {
      $this->oLogger->error($e->__toString());
      parent::unlink($sPath);
      return null;
    }

    $sFile = parent::file_get_contents($sPath, $bUseIncludePath, null, $iOffset, $iMaxLen);
    // on nettoie le fichier temporaire
    parent::unlink($sPath);

    return $sFile;
  }
  /**
   * {@inheritdoc}
   */
  public function copy(string $sSourceFilePath, string $sDestFilePath): ?bool{
    $sDestFilePath = $this->prefixReplacer($sDestFilePath);
    $sSourceFilePath = $this->prefixReplacer($sSourceFilePath);

    try {
      $this->oS3Client->copyObject([
        'Bucket' => $this->sS3Bucket,
        'Key' => $sDestFilePath,
        'CopySource' => $this->sS3Bucket . '/' . $sSourceFilePath,
      ]
      );
      return true;
    } catch (S3Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }
  /**
   * {@inheritdoc}
   */
  public function filesize(string $sFilePath): ?int {
    try {
      if ($this->is_dir($sFilePath)) {
        return 0;
      }

      $sFilePath = $this->prefixReplacer($sFilePath);

      $oResult = $this->headObject(
        [
          'Bucket' => $this->sS3Bucket,
          'Key' => $sFilePath,
        ]
      );

      if (!empty($oResult)) {
        return $oResult['ContentLength'];
      } else {
        throw new \ErrorException('filesize(' . $sFilePath . ') No such file on S3');
      }
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }

  }
  /**
   * {@inheritdoc}
   */
  public function readfile(string $sFilePath, bool $bUseIncludePath = false, $mContext = null): ?int{
    $sFilePath = $this->prefixReplacer($sFilePath);
    $sPath = $this->getTempDir();

    try {
      $this->oS3Client->getObject([
        'Bucket' => $this->sS3Bucket,
        'Key' => $sFilePath,
        'SaveAs' => $sPath,
      ]);
    } catch (S3Exception $e) {
      $this->oLogger->error($e->__toString());
      unlink($sPath);
      return null;
    }

    $iLength = parent::readfile($sPath, $bUseIncludePath, $mContext);
    unlink($sPath);

    return $iLength;
  }
  /**
   * {@inheritdoc}
   */
  public function rename(string $sSourceFilePath, string $sDestFilePath, $mContext = null): ?bool{
    $mReturn = false;

    if ($this->is_dir($sSourceFilePath)) {
      $mReturn = $this->copyDirectory($sSourceFilePath, $sDestFilePath);
      if (!empty($mReturn)) {
        $mReturn = $this->clearDir($sSourceFilePath);
      }
    } else {
      $mReturn = $this->copy($sSourceFilePath, $sDestFilePath);
      if (!empty($mReturn)) {
        $mReturn = $this->unlink($sSourceFilePath, $mContext);
      }
    }
    return ($mReturn === false) ? null : $mReturn;
  }
  /**
   * {@inheritdoc}
   */
  public function unlink(string $sFilePath, $mContext = null): ?bool {
    try {
      if (!$this->file_exists($sFilePath)) {
        throw new \ErrorException('unlink(' . $sFilePath . ') No such file on S3');
      }
      $sFilePath = $this->prefixReplacer($sFilePath);

      $this->oS3Client->deleteObject([
        'Bucket' => $this->sS3Bucket,
        'Key' => $sFilePath,
      ]
      );
      return true;
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }
  /**
   * {@inheritdoc}
   */
  public function filemtime(string $sFilePath): ?int {
    try {
      $sFilePath = $this->prefixReplacer($sFilePath);

      $oResult = $this->headObject(
        [
          'Bucket' => $this->sS3Bucket,
          'Key' => $sFilePath,
        ]
      );

      if (!empty($oResult)) {
        return $oResult['LastModified']->format('U');
      } else {
        throw new \ErrorException('filemtime(' . $sFilePath . ') No such file on S3');
      }
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }
  /**
   * {@inheritdoc}
   */
  public function mime_content_type(string $sFilePath): ?string {
    try {
      if (!$this->file_exists($sFilePath)) {
        throw new \ErrorException('mime_content_type(' . $sFilePath . ') No such file on S3');
      }
      return parent::fileExtToMimeType($sFilePath);
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function realpath(string $sFilePath): ?string{
    // passage de windows à linux
    $sPath = str_replace('\\', '/', $sFilePath);
    $aParts = explode('/', $sPath);

    $aAbsolutePath = [];

    foreach ($aParts as $sPart) {
      if ('.' == $sPart || empty($sPart)) {
        continue;
      } else if ('..' == $sPart) {
        if (count($aAbsolutePath) > 0) {
          array_pop($aAbsolutePath);
        }
      } else {
        $aAbsolutePath[] = $sPart;
      }
    }
    // réassemblage du chemin absolue
    if ($this->file_exists(implode('/', $aAbsolutePath))) {
      return implode('/', $aAbsolutePath);
    } else {
      return null;
    }
  }
  /****************************************************************
   *                                                              *
   *                            Files                             *
   *                       Veremes functions                      *
   *                                                              *
   ****************************************************************/

  /**
   * {@inheritdoc}
   */
  public function getFileInfos(string $sFilePath, array $aKeepOnlyField = []): ?array{
    try {
      $sS3FilePath = $this->prefixReplacer($sFilePath);

      $oResult = $this->headObject(
        [
          'Bucket' => $this->sS3Bucket,
          'Key' => $sS3FilePath,
        ]
      );

      if (empty($oResult)) {
        throw new \ErrorException('getFileInfos(' . $sFilePath . ') No such file on S3');
      }

      $iFileSize = $oResult['ContentLength'];
      $sFileSize = parent::fileSizeToHuman($iFileSize);

      $sFileName = $this->getFileName($sFilePath);

      $aFileInfos = [
        'filename' => $sFileName,
        'size' => $sFileSize,
        'size_' => $iFileSize,
        'lastModification' => $oResult['LastModified']->format('d/m/Y H:i:s'),
        'lastModification_' => $oResult['LastModified']->format('U'),
        'href' => $this->getProxyPassUrl($sFilePath),
      ];

      return parent::getFileInfosKeyFilter($aFileInfos, $aKeepOnlyField);
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }
  /**
   * {@inheritdoc}
   */
  public function getExtension(string $sFilePath): string {
    return parent::getExtension($sFilePath);
  }
  /**
   * {@inheritdoc}
   */
  public function getFileName(string $sFilePath): string {
    return parent::getFileName($sFilePath);
  }
  /**
   * {@inheritdoc}
   */
  public function getProxyPassUrl(string $sFilePath): ?string {
    try {
      // si le fichier est présent sur S3
      if ($this->checkFileExists($sFilePath)) {
        return parent::getEncryptedKeyUrl($sFilePath, 's3');
      }
      // si le fichier est présent en local on aura une url pour y acceder null sinon
      return parent::getProxyPassUrl($sFilePath);
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }
  /**
   * {@inheritdoc}
   */
  public function getFileEtag(string $sFilePath): ?string {
    try {
      $sFilePath = $this->prefixReplacer($sFilePath);

      $oResult = $this->headObject(
        [
          'Bucket' => $this->sS3Bucket,
          'Key' => $sFilePath,
        ]
      );

      if (empty($oResult)) {
        throw new \ErrorException('getFileEtag(' . $sFilePath . ') No such file on S3');
      } else {
        return str_replace('"', '', $oResult['ETag']);
      }
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }
  /**
   * {@inheritdoc}
   */
  public function file_get_contents_with_etag(string $sFilePath, string $sEtag, bool $bUseIncludePath = false, $mContext = null, int $iOffset = 0, int $iMaxLen = -1): ?string {
    try {
      $sProcessedEtag = $this->getFileEtag($sFilePath);
      if ($sEtag == $sProcessedEtag) {
        return $this->file_get_contents($sFilePath, $bUseIncludePath, $mContext, $iOffset, $iMaxLen);
      } else {
        throw new \ErrorException('file_get_contents_with_etag(' . $sFilePath . ',' . $sEtag . ') Processed eTag does not match.');
      }
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }
  /**
   * {@inheritdoc}
   */
  public function filemtime_formated(string $sFilePath, string $sFormat): ?string{
    $iUnixFilemtime = $this->filemtime($sFilePath);
    return (empty($iUnixFilemtime)) ? null : date($sFormat, $iUnixFilemtime);
  }
  /**
   * Copie du disque vers S3
   * {@inheritdoc}
   */
  public function copyLocalToDistant(string $sSourceFilePath, string $sDestFilePath): ?bool{
    $sDestFilePath = $this->prefixReplacer($sDestFilePath);

    $iSize = parent::filesize($sSourceFilePath);
    $source = fopen($sSourceFilePath, 'rb');

    
    // passage en multipart si dépasse le max accepté par S3 ou notre properties de seuil transactionnel
    if ($iSize > self::S3_LARGE_FILE_THRESHOLD || $iSize > $this->oProperties->get('transactional_upload_threshold')) {
      $oOptions = [
        'bucket' => $this->sS3Bucket,
        'key' => $sDestFilePath,
      ];
      $oUploader = new MultipartUploader($this->oS3Client, $source, $oOptions);
    } else {
      $oUploader = new ObjectUploader($this->oS3Client, $this->sS3Bucket, $sDestFilePath, $source);
    }

    try {
      $oUploader->upload();
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
    return true;
  }
  /**
   * Copie de S3 vers le disque
   * {@inheritdoc}
   */
  public function copyDistantToLocal(string $sSourceFilePath, string $sDestFilePath): ?bool{
    $sSourceFilePath = $this->prefixReplacer($sSourceFilePath);
    $sDirectory = dirname($sDestFilePath);

    if (!parent::is_dir($sDirectory)) {
      parent::mkdir($sDirectory, 0777, true);
    }

    try {
      $this->oS3Client->getObject([
        'Bucket' => $this->sS3Bucket,
        'Key' => $sSourceFilePath,
        'SaveAs' => $sDestFilePath,
      ]
      );
    } catch (S3Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
    return true;
  }

  /**
   * {@inheritdoc}
   */
  public function pictureResampler(string $sFilePath, int $iDstWidth, int $iDstHeight, int $iCropX = 0, int $iCropY = 0, int $iOffsetX = 0, int $iOffsetY = 0, ?string $sType = null): ?string {
    try {
      if (!$this->file_exists($sFilePath)) {
        throw new \ErrorException('pictureResampler(' . $sFilePath . ') : No such directory on S3');
      }

      // génération d'un chemin local temporaire
      $sPath = $this->getTempDir();
      $sDirName = dirname($sFilePath);
      $sFileName = $this->getFileName($sFilePath);
      $sTmpPicture = $sPath . '/' . $sFileName;

      if (parent::file_exists($sPath)) {
        parent::mkdir($sPath, 0777, true);
      }

      // copy local
      $this->copyDistantToLocal($sFilePath, $sTmpPicture);
      // resample en local
      $sDestTmpPath = parent::pictureResampler($sTmpPicture, $iDstWidth, $iDstHeight, $iCropX, $iCropY, $iOffsetX, $iOffsetY, $sType);
      if (empty($sDestTmpPath)) {
        throw new \ErrorException('pictureResampler(' . $sFilePath . ') : cannot resample this picture');
      }
      // on supprime le fichier déjà présent sur S3
      $this->unlink($sFilePath);
      $sDestFileName = parent::getFileName($sDestTmpPath);
      $sDestPath = $sDirName . '/' . $sDestFileName;

      if ($this->copyLocalToDistant($sDestTmpPath, $sDestPath)) {
        return $sDestPath;
      } else {
        throw new \ErrorException('pictureResampler(' . $sFilePath . ') : cannot copy this picture on S3');
      }

    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function thumbnailResampler(string $sFilePath): ?string {
    try {
      if (!$this->file_exists($sFilePath)) {
        throw new \ErrorException('thumbnailResampler(' . $sFilePath . '): File doesn\'t exist on S3');
      }

      $sExt = $this->getExtension($sFilePath);
      $sMinPath = parent::thumbnailPathGenerator($sFilePath);

      if (!$this->copy($sFilePath, $sMinPath)) {
        throw new \ErrorException('thumbnailResampler(' . $sFilePath . '): Cannot copy this picture on S3');
      }
      $sMinPath = $this->pictureResampler($sMinPath, $this->oProperties->get('thumbnails_max_width'), $this->oProperties->get('thumbnails_max_height'), 0, 0, 0, 0, $sExt);

      if (empty($sMinPath)) {
        throw new \ErrorException('thumbnailResampler(' . $sFilePath . '): Cannot resample this picture');
      }
      return $sMinPath;
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }

  /**
   * Cette fonction ne fait pas partie de l'interface FileAccessor
   * Copie un fichier du bucket configuré en properties dans un autre bucket.
   * @param string $sSourceFilePath Clé de l'objet à copier.
   * @param string $sDests3Key Clé de l'objet S3 dans le bucket de destination.
   * @param string $sDestBucket Bucket S3 de destination pour l'objet.
   * @return bool|null Cette fonction retourne TRUE en cas de succès ou NULL si une erreur survient.
   */
  public function copyInAnotherBucket(string $sSourceFilePath, string $sDests3Key, string $sDestBucket): ?bool{
    $sSourceFilePath = $this->prefixReplacer($sSourceFilePath);
    // Suppression du slash de début de ligne.
    if (strpos($sDests3Key, '/') === 0) {
      $sDests3Key = substr($sDests3Key, 1);
    }

    try {
      $this->oS3Client->copyObject([
        'Bucket' => $sDestBucket,
        'Key' => $sDests3Key,
        'CopySource' => $this->sS3Bucket . '/' . $sSourceFilePath,
      ]);
      return true;
    } catch (S3Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }

  /****************************************************************
   *                                                              *
   *                            Folders                           *
   *                     PHP native functions                     *
   *                                                              *
   ****************************************************************/

  /**
   * {@inheritdoc}
   */
  public function scandir(string $sDirPath, int $iSortOrder = SCANDIR_SORT_ASCENDING, $mContext = null): ?array{
    $aReturn = [];
    try {
      if (!$this->file_exists($sDirPath)) {
        throw new \ErrorException('scandir(' . $sDirPath . ') No such directory on S3');
      }
      //$aReturn = [".", ".."];
      // traitement du bucket et de ses sous-dossiers
      $sDirPath = $this->prefixReplacer($sDirPath);
      if (substr($sDirPath, -1) !== '/') {
        $sDirPath .= '/';
      }

      $aList = $this->oS3Client->listObjectsV2(
        [
          'Bucket' => $this->sS3Bucket,
          'Prefix' => $sDirPath,
        ]
      );

      if ($aList['KeyCount'] > 0) {
        foreach ($aList['Contents'] as $aObject) {
          $sObjectPath = str_replace($sDirPath, '', $aObject['Key']);
          $aObjectPath = explode('/', $sObjectPath);
          if (!empty($aObjectPath[0]) && !in_array($aObjectPath[0], $aReturn)) {
            array_push($aReturn, $aObjectPath[0]);
          }
        }
      }
      if ($iSortOrder === SCANDIR_SORT_ASCENDING) {
        sort($aReturn);
      }
      if ($iSortOrder === SCANDIR_SORT_DESCENDING) {
        rsort($aReturn);
      }
      return $aReturn;
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }
  /**
   * Crée un dossier sur S3
   * {@inheritdoc}
   */
  public function mkdir(string $sDirPath, int $iMode = 0777, bool $bRecursive = false, $mContext = null): ?bool {
    try {
      $sDirPath = $this->prefixReplacer($sDirPath);
      if (substr($sDirPath, -1) !== '/') {
        $sDirPath .= '/';
      }

      $this->oS3Client->putObject([
        'Bucket' => $this->sS3Bucket,
        'Key' => $sDirPath,
        'Body' => "",
      ]);

      return true;
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }
  /**
   * Supprime un dossier sur S3
   * {@inheritdoc}
   */
  public function rmdir(string $sDirPath, $mContext = null): ?bool {
    try {
      $sDirPath = $this->prefixReplacer($sDirPath);
      if (substr($sDirPath, -1) !== '/') {
        $sDirPath .= '/';
      }

      $this->oS3Client->deleteObject([
        'Bucket' => $this->sS3Bucket,
        'Key' => $sDirPath,
      ]);

      return true;
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }
  /**
   * {@inheritdoc}
   */
  public function is_dir(string $sDirPath): bool {
    // traitement du bucket et de ses sous-dossiers
    $sDirPath = $this->prefixReplacer($sDirPath);
    try {
      $aList = $this->oS3Client->listObjectsV2([
        'Bucket' => $this->sS3Bucket,
        'Prefix' => $sDirPath,
      ]
      );

      $aPathScan = explode('/', $sDirPath);
      $iPrefixPathCount = count($aPathScan);

      if ($aList['KeyCount'] > 1) {
        for ($i = 0; $i < $aList['KeyCount']; $i++) {
          $aPathKey = explode('/', $aList['Contents'][$i]['Key']);
          $iKeyCount = count($aPathKey);
          if ($iKeyCount >= $iPrefixPathCount) {
            if ($aPathKey[$iPrefixPathCount - 1] === $aPathScan[$iPrefixPathCount - 1] && empty(end($aPathKey))) {
              return true;
            }
          }
        }
        return false;
      } else if ($aList['KeyCount'] == 1) {
        $aPathKey = explode('/', $aList['Contents'][0]['Key']);
        $iKeyCount = count($aPathKey);
        if ($iKeyCount > $iPrefixPathCount) {
          if ($aPathKey[$iPrefixPathCount - 1] === $aPathScan[$iPrefixPathCount - 1]) {
            return true;
          } else {
            return false;
          }
        } else {
          return false;
        }
      } else {
        return false;
      }
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return false;
    }
  }

  /****************************************************************
   *                                                              *
   *                            Folders                           *
   *                       Veremes functions                      *
   *                                                              *
   ****************************************************************/

  /**
   * {@inheritdoc}
   */
  public function getFolderInfos(string $sDirPath, array $aKeepOnlyExt = [], array $aKeepOnlyField = [], bool $bRemoveEmptyFolder = false, int $iDepth = -1): ?array{
    try {

      $sDirPath = rtrim($sDirPath, '/');

      if (!$this->file_exists($sDirPath)) {
        throw new \ErrorException('getFolderInfos(' . $sDirPath . ') No such directory on S3');
      }

      if ($this->is_dir($sDirPath)) {
        $aContentDir = array_diff($this->scandir($sDirPath), ['.', '..']);
        $aDirPath = explode('/', $sDirPath);
        $sFileName = end($aDirPath);
        $aTree = [
          'filename' => $sFileName,
          'lastModification' => 'none',
          'size' => '',
          'content' => [],
        ];

        foreach ($aContentDir as $sContentElement) {
          $mReturn = null;
          $sContentElementPath = $sDirPath . '/' . $sContentElement;
          if ($this->is_dir($sContentElementPath)) {
            if ($iDepth === -1 || $iDepth > 0) {
              $mReturn = $this->getFolderInfos($sContentElementPath, $aKeepOnlyExt, $aKeepOnlyField, $bRemoveEmptyFolder, ($iDepth === -1) ? $iDepth : $iDepth - 1);
            }
          } else {
            if (empty($aKeepOnlyExt) || in_array(strtolower($this->getExtension($sDirPath)), $aKeepOnlyExt)) {
              $mReturn = $this->getFileInfos($sContentElementPath, $aKeepOnlyField);
            }
          }

          if (!empty($mReturn)) {
            array_push($aTree['content'], $mReturn);
          }
        }

        return $aTree;
      } else {
        throw new \ErrorException('getFolderInfos(' . $sDirPath . ') is not a directory');
      }
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }

  }
  /**
   * {@inheritdoc}
   */
  public function clearDir(string $sDirPath): ?bool {
    try {
      if (!$this->file_exists($sDirPath)) {
        throw new \ErrorException('clearDir(' . $sDirPath . ') No such directory on S3');
      }

      $oDeleteParams = [
        'Objects' => [],
      ];

      $sDirPath = $this->prefixReplacer($sDirPath);
      if (substr($sDirPath, -1) !== '/') {
        $sDirPath .= '/';
      }

      $aList = $this->oS3Client->listObjectsV2([
        'Bucket' => $this->sS3Bucket,
        'Prefix' => $sDirPath,
      ]
      );

      if ($aList['KeyCount'] > 0) {
        foreach ($aList['Contents'] as $aObject) {
          array_push($oDeleteParams['Objects'], ['Key' => $aObject['Key']]);
        }
      }

      $this->oS3Client->deleteObjects([
        'Bucket' => $this->sS3Bucket,
        'Delete' => $oDeleteParams,
      ]);

      return true;
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }
  /**
   * {@inheritdoc}
   */
  public function copyDirectory(string $sDirPathSrc, string $sDirPathDest): ?bool {
    try {
      if (!$this->file_exists($sDirPathSrc)) {
        throw new \ErrorException('copyDirectory(' . $sDirPathSrc . ') No such directory on S3');
      }
      // traitement du bucket et de ses sous-dossiers
      $sListPrefix = $this->prefixReplacer($sDirPathSrc);
      $sListPrefixDest = $this->prefixReplacer($sDirPathDest);

      if (substr($sListPrefix, -1) !== '/') {
        $sListPrefix .= '/';
      }
      if (substr($sListPrefixDest, -1) !== '/') {
        $sListPrefixDest .= '/';
      }

      $aList = $this->oS3Client->listObjectsV2([
        'Bucket' => $this->sS3Bucket,
        'Prefix' => $sListPrefix,
      ]
      );

      if ($aList['KeyCount'] > 0) {
        foreach ($aList['Contents'] as $aObject) {
          $this->oS3Client->copyObject([
            'Bucket' => $this->sS3Bucket,
            'Key' => str_replace($sListPrefix, $sListPrefixDest, $aObject['Key']),
            'CopySource' => $this->sS3Bucket . '/' . $aObject['Key'],
          ]
          );
        }
      }
      return true;
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }
  /**
   * {@inheritdoc}
   */
  public function copyDirLocalToDistant(string $sDirPathSrc, string $sDirPathDest, array $aIgnoredItems = []): ?bool {
    try {
      if (!parent::file_exists($sDirPathSrc)) {
        throw new \ErrorException('copyDirLocalToDistant(' . $sDirPathSrc . ') No such directory on local disk');
      }

      $sDestPrefix = $this->prefixReplacer($sDirPathDest);
      $aContentDir = array_diff(scandir($sDirPathSrc), ['..', '.']);

      foreach ($aContentDir as $sPathMember) {
        if (!in_array($sPathMember, $aIgnoredItems)) {
          if (is_dir($sDirPathSrc . '/' . $sPathMember)) {
            $bReturn = $this->copyDirLocalToDistant($sDirPathSrc . '/' . $sPathMember, $sDestPrefix . '/' . $sPathMember, $aIgnoredItems);
          } else {
            $bReturn = $this->file_put_contents($sDestPrefix . '/' . $sPathMember, file_get_contents($sDirPathSrc . '/' . $sPathMember));
          }
          if (!$bReturn) {
            throw new \ErrorException('error while sending on S3 of ' . $sDirPathSrc . '/' . $sPathMember);
          }
        }
      }
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }

    return true;
  }
  /**
   * {@inheritdoc}
   */
  public function copyDirDistantToLocal(string $sDirPathSrc, string $sDirPathDest): ?bool{
    $sSourcePrefix = $this->prefixReplacer($sDirPathSrc);
    if (substr($sDirPathDest, -1) !== '/') {
      $sDirPathDest .= '/';
    }
    if (substr($sSourcePrefix, -1) !== '/') {
      $sSourcePrefix .= '/';
    }

    try {
      if (!$this->file_exists($sDirPathSrc)) {
        throw new \ErrorException('copyDirDistantToLocal(' . $sDirPathSrc . ') No such directory on S3');
      }

      if (!parent::file_exists($sDirPathDest)) {
        parent::mkdir($sDirPathDest, 0777, true);
      }

      $aList = $this->oS3Client->listObjectsV2([
        'Bucket' => $this->sS3Bucket,
        'Prefix' => $sSourcePrefix,
      ]
      );
      if ($aList['KeyCount'] > 0) {
        foreach ($aList['Contents'] as $aObject) {
          $sDestPath = str_replace($sSourcePrefix, $sDirPathDest, $aObject['Key']);
          $sDestPath = rtrim($sDestPath, '/');
          if ($this->is_dir(rtrim($aObject['Key'], '/'))) {
            parent::mkdir($sDestPath, 0777, true);
          } else {
            $sFileContent = $this->file_get_contents($aObject['Key']);
            // création de l'arborescence si besoin
            $aPath = explode('/', $sDestPath);
            array_pop($aPath);
            $sDirectory = implode('/', $aPath);
            if (empty(parent::is_dir($sDirectory))) {
              parent::mkdir($sDirectory, 0777, true);
            }
            parent::file_put_contents($sDestPath, $sFileContent);
          }
        }
      }
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
    return true;
  }
  /**
   * {@inheritdoc}
   */
  public function createZip(string $sFolder, string $sDestination, string $sExtensionToExclude = '', string $sPassword = '', bool $bDeleteFiles = false): ?bool{
    $sPath = $this->getTempDir();
    try {
      if (!$this->file_exists($sFolder)) {
        throw new \ErrorException('createZip(' . $sFolder . ') : no such directory on S3');
      }
      if ($this->is_dir($sFolder)) {
        // get all file match with prefix $sFolder in the same arborescence
        if (!empty($this->copyDirDistantToLocal($sFolder, $sPath))) {
          // create zip in local
          if (!empty(parent::createZip($sPath, $sPath . '.zip', $sExtensionToExclude, $sPassword, true))) {
            // put result on s3
            $this->copyLocalToDistant($sPath . '.zip', $sDestination);
            // nettoyage en local
            unlink($sPath . '.zip');
            if ($bDeleteFiles) {
              if (!$this->clearDir($sFolder)) {
                throw new \ErrorException('cannnot clean the zipped directory : ' . $sFolder);
              }
            }
          } else {
            throw new \ErrorException('cannnot zip : ' . $sFolder);
          }

        } else {
          throw new \ErrorException('Error during the copy of s3 files on local file system');
        }
      } else {
        throw new \ErrorException('This prefix not match on s3 : ' . $sFolder);
      }
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
    return true;
  }
  /**
   * {@inheritdoc}
   */
  public function encryptZip(string $sZipFilePath, string $sPassword): ?bool{
    $sPath = $this->getTempDir();
    try {
      if (!$this->file_exists($sZipFilePath)) {
        throw new \ErrorException('encryptZip(' . $sZipFilePath . ') : no such file on S3');
      }
      // get zip file on S3
      if (!empty($this->copyDistantToLocal($sZipFilePath, $sPath . '.zip'))) {
        parent::encryptZip($sPath . '.zip', $sPassword);
        return $this->copyLocalToDistant($sPath . '.zip', $sZipFilePath);
      }
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
    return false;
  }
  /**
   * {@inheritdoc}
   */
  public function unZip(string $sZipFile, string $sDestination): ?bool{
    $sPath = $this->getTempDir();
    try {
      if (!$this->file_exists($sZipFile)) {
        throw new \ErrorException('unZip(' . $sZipFile . ') : no such file on S3');
      }
      // get zip file on S3
      if (!empty($this->copyDistantToLocal($sZipFile, $sPath . '.zip'))) {
        // unzip in local
        mkdir($sPath, 0777, true);
        if (!empty(parent::unZip($sPath . '.zip', $sPath))) {
          // envoi sur s3
          $this->copyDirLocalToDistant($sPath, $sDestination);
          // if deletes file cleardir
          parent::clearDir($sPath);
          // nettoyage en local
          unlink($sPath . '.zip');
        } else {
          throw new \ErrorException('Can\'t unzip : ' . $sZipFile);
        }
      } else {
        throw new \ErrorException('Can\'t do a local copy of : ' . $sZipFile);
      }
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
    return true;
  }
  /**
   * Créé une archive sur S3 en compressant un dossier local
   * {@inheritdoc}
   */
  public function zipLocalDirToDistant(string $sFolder, string $sDestination, string $sExtensionToExclude = '', string $sPassword = '', bool $bDeleteFiles = false): ?bool {
    try {
      if (!parent::file_exists($sFolder)) {
        throw new \ErrorException('zipLocalDirToDistant(' . $sFolder . ') : no such directory on local disk');
      }

      $sPath = $this->getTempDir();
      if (!empty(parent::createZip($sFolder, $sPath . '.zip', $sExtensionToExclude, $sPassword, $bDeleteFiles))) {
        // put result on s3
        $this->copyLocalToDistant($sPath . '.zip', $sDestination);
        // nettoyage en local
        unlink($sPath . '.zip');
      } else {
        throw new \ErrorException('Can\'t zip local directory : ' . $sFolder);
      }
      return true;
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }
  /**
   * Créé une archive en local en compressant un "dossier" sur S3
   * {@inheritdoc}
   */
  public function zipDistantDirToLocal(string $sFolder, string $sDestination, string $sExtensionToExclude = '', string $sPassword = '', bool $bDeleteFiles = false): ?bool {
    try {
      if (!$this->file_exists($sFolder)) {
        throw new \ErrorException('zipDistantDirToLocal(' . $sFolder . ') : no such directory on S3');
      }
      $sPath = $this->getTempDir();
      if (!empty($this->copyDirDistantToLocal($sFolder, $sPath))) {
        if (!empty(parent::createZip($sPath, $sDestination, $sExtensionToExclude, $sPassword, $bDeleteFiles))) {
          // if deletes file cleardir
          parent::clearDir($sPath);
        } else {
          throw new \ErrorException('zipLocalDirToDistant(' . $sFolder . ') :Can\'t zip local directory');
        }
      } else {
        throw new \ErrorException('zipLocalDirToDistant(' . $sFolder . ',' . $sDestination . ') : Can\'t copy the target directory in local');
      }
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
    return true;
  }
  /**
   * Dézippe une archive sur le disque et envoi le contenu sur S3 dans un "dossier"
   * {@inheritdoc}
   */
  public function unzipLocalToDistantDir(string $sZipFile, string $sDestination): ?bool {
    try {
      if (!parent::file_exists($sZipFile)) {
        throw new \ErrorException('unzipLocalToDistantDir(' . $sZipFile . ') : no such file on local disk');
      }
      $sPath = $this->getTempDir();
      if (!empty(parent::unZip($sZipFile, $sPath))) {
        if (!empty($this->copyDirLocalToDistant($sPath, $sDestination))) {
          // if deletes file cleardir
          parent::clearDir($sPath);
        } else {
          throw new \ErrorException('Can\'t copy local to S3 : ' . $sZipFile);
        }
      } else {
        throw new \ErrorException('Can\'t unzip : ' . $sZipFile);
      }
      return true;
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }
  /**
   * Dézippe une archive sur S3 dans un dossier sur le disque local
   * {@inheritdoc}
   */
  public function unzipDistantToLocalDir(string $sZipFile, string $sDestination): ?bool {
    try {
      if (!$this->file_exists($sZipFile)) {
        throw new \ErrorException('unzipDistantToLocalDir(' . $sZipFile . ') : no such file on S3');
      }
      $sPath = $this->getTempDir();
      if (!empty($this->copyDistantToLocal($sZipFile, $sPath . '.zip'))) {
        if (!empty(parent::unZip($sPath . '.zip', $sDestination))) {
          // nettoyage en local
          unlink($sPath . '.zip');
        } else {
          throw new \ErrorException('Can\'t unzip : ' . $sZipFile);
        }
      } else {
        throw new \ErrorException('Can\'t copy S3 to local : ' . $sZipFile);
      }
      return true;
    } catch (\Exception $e) {
      $this->oLogger->error($e->__toString());
      return null;
    }
  }
}
