<?php

namespace App\Module\Vitis\WebService\Service\Logger;

use App\Module\Vitis\WebService\Query\DateTimeTypeWithMicroTime;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\DataAccessor\Fields\DateField;
use Doctrine\DBAL\ParameterType;
use Monolog\Logger;
use Psr\Log\LoggerInterface;

/**
 * A SQL logger that logs to the standard Monolog/logger.
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */
class CustomSqlLogger implements LoggerInterface {
  /**
   * @var Properties
   */
  private $oPropertiesSrvc;
  /**
   * @var array
   */
  private $aMessages = [];
  /**
   * @var array
   */
  private $aTmpMessages = [];
  /**
   * @var string
   */
  private $sUsername = '';
  /**
   * @var Logger
   */
  private $oLogger;
  /**
   * @var Logger
   */
  private $oGenericLogger;
  /**
   * @var bool
   */
  private $bTransaction = false;
  /**
   * @var bool
   */
  private $bDebug = false;

  /**
   * @param Logger $oLogger logger à utiliser pour les requêtes SQL
   * @param Logger $oGenericLogger logger à utiliser pour le reste
   * @param string $sUsername nom de l'utilisateur connecté
   * @param Properties $PropertiesSrvc Un objet pour accéder aux properties de l'application
   */
  public function __construct(Logger $oLogger, Logger $oGenericLogger, string $sUsername, Properties $PropertiesSrvc) {
    $this->oLogger = $oLogger;
    $this->oGenericLogger = $oGenericLogger;
    $this->sUsername = $sUsername;
    // Sauvegarde du service de properties
    $this->oPropertiesSrvc = $PropertiesSrvc;
    $this->bDebug = ('DEBUG' === $this->oPropertiesSrvc->get('log_level'));
  }

  /**
   * Ancienne fonction de l'interface sqllogger dépréciée que l'on continue d'utiliser pour la structure des logs mais encapsuler par un logger psr
   * Logs a SQL statement somewhere.
   *
   * @param string                                                                    $sql    SQL statement
   * @param list<mixed>|array<string, mixed>|null                                     $params Statement parameters
   * @param array|null $types  Parameter types
   *
   * @return void
   */
  private function startQuery($sql, ?array $params = null, ?array $types = null): void {
    // Passage en mode transactionnel
    if ($sql == '"START TRANSACTION"' && !$this->bDebug) {
      $this->aTmpMessages = [];
      $this->bTransaction = true;
      return;
    }
    // validation de la transaction
    if ($sql == '"COMMIT"' && !$this->bDebug) {
      $this->aMessages = array_merge($this->aMessages, $this->aTmpMessages);
      $this->bTransaction = false;
      return;
    }

    $sBindSql = $sql;
    $aParams = (isset($params)) ? $params : [];
    
    // bind le sql si nécessaire pour le log
    foreach ($aParams as $i => $mParam) {
      $sParam = $mParam ?? '';

      $sType = (!empty($types) && isset($types[$i])) ? $types[$i] : gettype($mParam);

      switch ($sType) {
        case 'integer':
        case 'float':
        case ParameterType::INTEGER :
          break;
        case 'boolean':
        case 'bool':
        case ParameterType::BOOLEAN :
          break;
        case 'string':
        case ParameterType::STRING:
          if (!is_numeric($mParam) && empty($types)) {
            $sParam = "'" . $mParam . "'";
          }
          break;
        case 'array':
          $sParam = "'" . implode("','", $mParam) . "'";
          break;
        case 'datetime':
          $sParam = (isset($mParam)) ? $mParam->format(DateField::DEFAULT_FORMAT) : 'null';
          break;
        case 'udatetime':
          $sParam = (isset($mParam)) ? $mParam->format(DateTimeTypeWithMicroTime::POSTRES_DATE_FORMAT) : 'null';
          break;
        case 'json':
          $sParam = (isset($mParam)) ? json_encode($mParam) : 'null';
          break;
        default:
          $this->oGenericLogger->debug('type not known (' . $sType . ')', [
            'sql' => $sql,
            'params' => $params,
            'types' => $types,
            'proc' => $mParam
          ]);
          break;
      }

      $iPosition = strpos($sBindSql, "?");
      if ($iPosition > -1) {
        $sBindSql = substr_replace($sBindSql, $sParam, $iPosition, 1);
      } else {
        $sBindSql = str_replace(':' . $i, $sParam, $sBindSql);
      }
    }
    // définition du niveau de log
    $sLevel = 'info';
    if (strpos($sql, 'SELECT') === 0 || (in_array($sql, ['"START TRANSACTION"', '"COMMIT"']) && $this->bDebug)) {
      $sLevel = 'debug';
    }

    $aMessage = [
      'login' => $this->sUsername,
      'request' => $sBindSql,
      'parameters' => $params,
      'real_request' => $sql,
      'level' => $sLevel,
      'start' => microtime(true),
    ];

    // si une transaction est demmaré on stock de manière temporaire les requête en attendant le Commit
    if ($this->bTransaction && !$this->bDebug) {
      $this->aTmpMessages[] = $aMessage;
    } else {
      $this->aMessages[] = $aMessage;
      $this->stopQuery();
    }
  }

  /**
   * Ancienne fonction de l'interface sqllogger dépréciée que l'on continue d'utiliser pour la structure des logs mais encapsuler par un logger psr
   * Marks the last started query as stopped. This can be used for timing of queries.
   *
   * @return void
   */
  private function stopQuery(): void {
    // Ecriture de toutes les requêtes dans logs
    foreach ($this->aMessages as $aMessage) {
      $sMethod = $aMessage['level'];

      $sMessage = json_encode([
        'login' => $aMessage['login'],
        'request' => $aMessage['request'],
        'parameters' => $aMessage['parameters'],
        'real_request' => $aMessage['real_request'],
        'time_exec' => ((microtime(true) - $aMessage['start']) * 1000) . ' ms',
      ]);

      $this->oLogger->$sMethod($sMessage);
    }

    $this->aMessages = [];
  }

  /**
   * {@inheritDoc}
   */
  public function emergency(string|\Stringable $message, array $context = []): void {
    $this->log('emergency', $message, $context);
  }
  /**
   * {@inheritDoc}
   */
  public function alert(string|\Stringable $message, array $context = []): void {
    $this->log('alert', $message, $context);

  }
  /**
   * {@inheritDoc}
   */
  public function critical(string|\Stringable $message, array $context = []): void {
    $this->log('critical', $message, $context);
  }
  /**
   * {@inheritDoc}
   */
  public function error(string|\Stringable $message, array $context = []): void {
    $this->log('error', $message, $context);
  }
  /**
   * {@inheritDoc}
   */
  public function warning(string|\Stringable $message, array $context = []): void {
    $this->log('warning', $message, $context);
  }
  /**
   * {@inheritDoc}
   */
  public function notice(string|\Stringable $message, array $context = []): void {
    $this->log('notice', $message, $context);
  }
  /**
   * {@inheritDoc}
   */
  public function info(string|\Stringable $message, array $context = []): void {
    $this->log('info', $message, $context);
  }
  /**
   * {@inheritDoc}
   */
  public function debug(string|\Stringable $message, array $context = []): void {
    $this->log('debug', $message, $context);
  }
  /**
   * {@inheritDoc}
   */
  public function log(mixed $level, string|\Stringable $message, array $context = []): void {
    if (array_key_exists('sql', $context)) {
      // on protège les données cryptés
      $context = $this->protectData($context);
      $this->oGenericLogger->debug('SQL Logger initial level (' . $level . '):' . $message, $context);
      // on log une requête passage sur l'ancien logger
      $aParams = (array_key_exists('params', $context)) ? $context['params'] : null;
      $aTypes = (array_key_exists('types', $context)) ? $context['types'] : null;
      $this->startQuery($context['sql'], $aParams, $aTypes);
    } else {
      $this->oGenericLogger->log($level, $message, $context);
    }

  }

  /**
   * permet d'agir sur ce qui va être log
   * actuellement cahce les donées des colonnes cryptées
   *
   * @param  array $context
   * @return array
   */
  private function protectData(array $context) : array {
    // remplace les clé de cryptage dans les requêtes si détection des fonctions pgp
    $aMatches = [];
    while (preg_match("/pgp_sym_((en)|(de))crypt\s*\(\s*((?<replacer_data>\?)|([^',]+))\s*,\s*\'(?<replacer_key>[0-9a-z]+)\'\s*(,.*\s*)?\)/", $context['sql'], $aMatches, PREG_OFFSET_CAPTURE) > 0) {
      if (!empty($aMatches['replacer_key'])) {
        $context['sql'] = str_replace($aMatches['replacer_key'][0], '*****hidden key*****', $context['sql']);
      }

      if (!empty($aMatches['replacer_data'][0]) && array_key_exists('params', $context)) {
        // ne démarre pas à 0
        $iParamsIndex = substr_count($context['sql'], '?', 0, $aMatches['replacer_data'][1]) + 1;
        $context['params'][$iParamsIndex]  = '*****hidden value*****';
        $context['types'][$iParamsIndex]  = ParameterType::STRING;
      }
    }

    return  $context;

  }

}
