<?php

namespace App\Module\Vitis\WebService\Service\Mail\Context;

use App\Module\Vitis\WebService\Entity\MailTemplate;
use App\Module\Vitis\WebService\Entity\User;
use App\Module\Vitis\WebService\Repository\GenericRepository;
use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\DataAccessor\DatabaseConnection;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\Mail\MailContextInterface;

/**
 * Context pour le mail de mot de passe oublié
 * Utilisé pour l'envoi de mail
 *
 * @author Ludivine Premont <ludivine.premont@veremes.com>
 */
class AccountContext implements MailContextInterface {
  /**
   * @var Properties
   */
  private $oProperties;
  /**
   * @var LoggerManager
   */
  private $oLoggerManager;

  /**
   * @var DatabaseConnection
   */
  private $oDatabaseConnection;
  /**
   * @var Array
   * user.login
   * user.name
   * user.email
   * user.company
   */
  private $user;
  /**
   * @var String
   */
  private $sUrl;

  /**
   * @var Array
   */
  private $templates = [];
  /**
   * @var string
   */
  public const FORGOTTEN_PASSWORD = 'forgotten-password';
  /**
   * @var string
   */
  public const SIGN_UP = 'sign-up';
  /**
   * @var string
   */
  public const CONFIRM_SIGN_UP = 'confirm-sign-up';

  /**
   * Initialise le context
   * @param LoggerManager $oLoggerManager Service d'utilsation des logs
   * @param DatabaseConnection $oDatabaseConnection Service d'accès à la base de données
   * @param Properties $oProperties Service d'accès aux properties
   * @param Array $user Utilisateur qui veut changer de mot de passe
   * @param String $sUrl Url à remplacer dans le template de mail
   */
  public function __construct(LoggerManager $oLoggerManager, DatabaseConnection $oDatabaseConnection, Properties $oProperties, array $user, String $sUrl) {
    $this->oLoggerManager = $oLoggerManager;
    $this->oDatabaseConnection = $oDatabaseConnection;
    $this->oProperties = $oProperties;
    $this->user = $user;
    $this->sUrl = $sUrl;
    $this->templates = [
      self::FORGOTTEN_PASSWORD => $this->oProperties->get('forgotten_password_mail_model'),
      self::SIGN_UP => $this->oProperties->get('sign_up_mail_model'),
      self::CONFIRM_SIGN_UP => $this->oProperties->get('confirm_sign_up_mail_model'),
    ];
  }

  /**
   * @param String $title Titre du template dans le tableau des templates
   * @return MailTemplate|null
   */
  public function getMailTemplate($title): ?MailTemplate {
    /* @var $MailTemplateRepository GenericRepository */
    try {
      $MailTemplateRepository = $this->oDatabaseConnection->getAdminEntityManager()->getRepository(MailTemplate::class);

      if (!$MailTemplateRepository instanceof GenericRepository) {
        throw new \Exception("MailTemplate Repository is not a GenericRepository", 1);
      }

      $oTemplate = $MailTemplateRepository->findById($this->templates[$title])[0];
      return $oTemplate;
    } catch (\Exception $e) {
      $this->oLoggerManager->get('SQL_ERROR')->error('Cannot get Forgotten Password MailTemplate ', [$title, $e]);
      return null;
    }
  }
  /**
   * @return array
   */
  public function getUser(): array{
    if (empty($this->user)) {
      return [];
    }
    // return UserContext
    return [
      'login' => $this->user['login'],
      'name' => $this->user['name'],
      'email' => $this->user['email'],
      'company' => $this->user['company'],
    ];
  }
  /**
   * @return string
   */
  public function getUrl(): string {
    if (empty($this->sUrl)) {
      return '';
    }
    // passage en url
    return $this->sUrl;
  }

  /**
   * @return array
   */
  public function getProperties(): array{
    return [
      'hostname' => $this->oProperties->get('_server_domain') ?? '',
      'admin_sign_up' => $this->oProperties->get('admin_sign_up'),
      'automated_sign_up' => $this->oProperties->get('automated_sign_up'),
      'app_client_url' => $this->oProperties->get('app_client_url'),
    ];
  }

  /* ---------------------------------------------------------------- */
  /* -------------------------- Interfaces -------------------------- */
  /* ---------------------------------------------------------------- */
  /**
   * {{@inheritdoc}}
   */
  public function haveToSendMail(): bool {
    return $this->oProperties->get('advanced_login_page') ?? false;
  }
  /**
   * {{@inheritdoc}}
   */
  public function getContextData(): array{
    return [
      'user' => $this->getUser(),
      'properties' => $this->getProperties(),
      'url' => $this->getUrl(),
    ];
  }
}
