<?php

namespace App\Module\Vitis\WebService\Service\PhantomJs;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\FileAccessor\FileManager;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\PhantomJs\AbstractCaller;

class HtmlRenderer extends AbstractCaller {
  /**
   * chemin vers le script node à appeler
   * @var string
   */
  private const NODE_SCRIPT_PATH = '@prop(vas_dir)/src/Module/Vitis/utils/printserver/server/renderhtml.js';
  /**
   * chemin vers le script node à appeler
   * @var string
   */
  private const CLIENT_URL = 'file:///@prop(vas_dir)/src/Module/Vitis/utils/printserver/client/renderer/index.html';
  /**
   * @var string
   */
  private $sTempDir;
  /**
   * @var \Monolog\Logger
   */
  private $oDebugLogger;
  /**
   * @var \Monolog\Logger
   */
  private $oErrorLogger;
  /**
   * {@inheritdoc}
   */
  public function __construct(LoggerManager $oLogger, Properties $properties, FileManager $oFileManager) {
    parent::__construct($oLogger, $properties, $oFileManager);
    $this->oDebugLogger = $oLogger->get('COMMON_DEBUG');
    $this->oErrorLogger = $oLogger->get('API_ERROR');
  }

  /**
   * Evalue un code HTML en tuilisant un scope json
   * @param string $sHtmlBody HTML à évaluer
   * @param array $aScope donnée utilisé dans le template HTML
   * @param array $aIncludes images à inclure dans l'HTML
   * @return string|null HTML évalué ou null en cas d'erreur
   */
  public function render(string $sHtmlBody, array $aScope, array $aIncludes): ?string {
    try {
      // génération du tableau des arguments
      $aArguments = $this->prepareArguments($sHtmlBody, $aScope, $aIncludes);
      //exécution de la commande
      $sNodeScriptPath = $this->oProperties->replacePropertiesInString(self::NODE_SCRIPT_PATH);

      if($sNodeScriptPath === null){
        throw new \Exception('Cannot process path to node script');
      }

      $sReturn = $this->call($sNodeScriptPath, $aArguments);
      $this->oDebugLogger->debug($sReturn);
      $sHtml = $this->oLocalFileAccessor->file_get_contents($this->sTempDir . '/rendered.html');
    } catch (\Exception $e) {
      $this->oErrorLogger->error('Impossible to render HTML : ' . $e->getMessage());
      return null;
    } finally {
      if ($this->oLocalFileAccessor->file_exists($this->sTempDir)) {
        $this->oLocalFileAccessor->clearDir($this->sTempDir);
      }

    }
    return $sHtml;
  }

  /**
   * Prépare les arguments pour la CLI 
   * @param string $sHtmlBody HTML à évaluer
   * @param array $aScope donnée utilisé dans le template HTML
   * @param array $aIncludes images à inclure dans l'HTML
   * @throws \Exception si le dossier temporaire ne peut pas être créé
   * @return array tableau d'argument à utiliser pour appeler phantomjs
   */
  private function prepareArguments(string $sHtmlBody, array $aScope, array $aIncludes): array{
    $this->sTempDir = $this->oLocalFileAccessor->getTempDir();
    if ($this->oLocalFileAccessor->mkdir($this->sTempDir, 0777, true) !== true) {
      throw new \Exception('Cannot create Tempdir : ' . $this->sTempDir);
    }
    // génération d'un dossier temporaire
    $aArguments = [];
    // chemin vers l'index.html du renderer
    $aArguments[] = $this->oProperties->replacePropertiesInString(self::CLIENT_URL);
    // chemin vers l'entrée HTML
    $sHtmlPath = $this->sTempDir . '/input.html';
    $this->oLocalFileAccessor->file_put_contents($sHtmlPath, $sHtmlBody);
    $aArguments[] = $sHtmlPath;
    // chemin vers la sortie HTML
    $aArguments[] = $this->sTempDir . '/rendered.html';
    // chemin vers les includes
    $sIncludesPath = $this->sTempDir . '/includes.json';
    $this->oLocalFileAccessor->file_put_contents($sIncludesPath, json_encode($aIncludes));
    $aArguments[] = $sIncludesPath;
    // chemin vers le scope
    $sScopePath = $this->sTempDir . '/scope.json';
    $this->oLocalFileAccessor->file_put_contents($sScopePath, json_encode($aScope));
    $aArguments[] = $sScopePath;

    return $aArguments;
  }

}