<?php

namespace App\Module\Vitis\WebService\Service\PrintService;

use App\Module\Vitis\WebService\Service\VitisSystem;

class PrintReport extends AbstractPrinter {
  /**
   * Evalue un code HTML en utilisant un scope json
   * @param string $sHtmlBody HTML à évaluer
   * @param string $sFormat A4, A3, A2, A1, A0
   * @param string $sOutputFormat pdf, image, vectorpdf, imagepdf
   * @param string $sOrientation portrait, landscape
   * @param array $aScopes données utilisé dans le rapport HTML
   * @param array $aIncludes images à inclure dans l'HTML
   * @param string $sDestFilePath chemin du fichier de sortie
   * @param bool $bPaged si false alors desactive pajedjs lors de l'impression
   * @return string|null HTML évalué ou null en cas d'erreur
   */
  public function print(string $sHtmlBody, string $sFormat, string $sOutputFormat, string $sOrientation, array $aScopes, array $aIncludes, array $aMaps, string $sDestFilePath, bool $bPaged=true): ?string {
    try {
      // génération du tableau des arguments
      $aArguments = $this->prepareArguments($sHtmlBody, $sFormat, $sOutputFormat, $sOrientation, $aScopes, $aIncludes, $aMaps, $bPaged);
      //exécution de la commande
      $sReturn = $this->callPrinter($aArguments);
      $this->oLocalFileAccessor->file_put_contents($this->sTempDir . '/print.log', $sReturn);
      $this->oDebugLogger->debug($sReturn);

      if ($this->oLocalFileAccessor->file_exists($this->sTempDir . '/rendered.pdf')) {
        $this->oLocalFileAccessor->copy($this->sTempDir . '/rendered.pdf', $sDestFilePath);
      } else {
        $this->oErrorLogger->error('Error while printing');
        $this->oErrorLogger->error($sReturn);
        return null;
      }

      if (!empty($this->sTempDir)) {
        if ($this->oLocalFileAccessor->file_exists($this->sTempDir)) {
          $this->oLocalFileAccessor->clearDir($this->sTempDir);
        }
      }

    } catch (\Exception $e) {
      $this->oErrorLogger->error('Impossible to print HTML : ' . $e->getMessage());
      return null;
    }
    return $sDestFilePath;
  }

  /**
   * Prépare les arguments pour la CLI
   * @param string $sHtmlBody HTML à évaluer
   * @param string $sFormat A4, A3, A2, A1, A0
   * @param string $sOutputFormat pdf, image, vectorpdf, imagepdf
   * @param string $sOrientation portrait, landscape
   * @param array $aScopes donnée utilisé dans le template HTML
   * @param array $aIncludes images à inclure dans l'HTML
   * @param array $aMaps cartes à inclure dans l'HTML
   * @param bool $bPaged si false alors desactive pajedjs lors de l'impression
   * @throws \Exception si le dossier temporaire ne peut pas être créé
   * @return array tableau d'argument à utiliser pour appeler phantomjs
   */
  private function prepareArguments(string $sHtmlBody, string $sFormat, string $sOutputFormat, string $sOrientation, array $aScopes, array $aIncludes, array $aMaps, bool $bPaged): array{
    $this->sTempDir = $this->oLocalFileAccessor->getTempDir();
    if ($this->oLocalFileAccessor->mkdir($this->sTempDir, 0777, true) !== true) {
      throw new \Exception('Cannot create Tempdir : ' . $this->sTempDir);
    }
    // génération d'un dossier temporaire
    $aArguments = ['printReport'];
    // chemin vers l'entrée HTML
    $sHtmlPath = $this->sTempDir . '/input.html';
    $this->oLocalFileAccessor->file_put_contents($sHtmlPath, $sHtmlBody);
    $aArguments[] = '--HTMLInputFileName="' . $sHtmlPath . '"';
    // chemin vers la sortie PDF
    $aArguments[] = '--TmpFolder="' . $this->sTempDir . '"';
    $aArguments[] = '--OutputFileName="' . $this->sTempDir . '/rendered.pdf' . '"';
    // chemin vers les includes
    $sIncludesPath = $this->sTempDir . '/includes.json';
    $this->oLocalFileAccessor->file_put_contents($sIncludesPath, json_encode($aIncludes));
    $aArguments[] = '--JsonIncludes="' . $sIncludesPath . '"';
    // chemin vers les maps
    $sMapsPath = $this->sTempDir . '/maps.json';
    $this->oLocalFileAccessor->file_put_contents($sMapsPath, json_encode($aMaps));
    $aArguments[] = '--JsonMaps="' . $sMapsPath . '"';
    // chemin vers les scopes
    $sScopesPath = $this->sTempDir . '/scopes.json';
    $this->oLocalFileAccessor->file_put_contents($sScopesPath, json_encode($aScopes));
    $aArguments[] = '--JsonScopes="' . $sScopesPath . '"';
    $aArguments[] = '--Format="' . $sFormat . '"';
    $aArguments[] = '--OutputFormat="' . $sOutputFormat . '"';
    $aArguments[] = '--Orientation="' . $sOrientation . '"';
    $aArguments[] = '--Paged="' . ( $bPaged === true ? 'true' : 'false' ) . '"';

    return $aArguments;
  }
}
