<?php

namespace App\Module\Vitis\WebService\Service\Token;

use App\Module\Vitis\WebService\Service\Configuration\Properties;
use App\Module\Vitis\WebService\Service\Crypto\JweCrypto;
use App\Module\Vitis\WebService\Service\Logger\LoggerManager;
use App\Module\Vitis\WebService\Service\Token\ExpirationTimeCheckerManager;
use Jose\Component\Checker;
use Jose\Component\Checker\ClaimCheckerManager;
use Monolog\Logger;

/**
 * Le JWE est une variante du JWT où les données sont encryptées.
 * Pour plus d'informations :
 *       - JWT : https://tools.ietf.org/html/rfc7519
 *       - JWE : https://tools.ietf.org/html/rfc7516
 *
 * @author Anthony Borghi <anthony.borghi@veremes.com>
 */

class JweTokenManager extends JweCrypto {
  /**
   * @var string[]
   */
  private const CLAIMERS = ['iat', 'nbf', 'exp', 'iss', 'sub', 'aud', 'jti', 'data'];

  /**
   * @var ClaimCheckerManager
   */
  private $oClaimCheckerManager;

  /**
   * @param Properties $PropertiesSrvc Service de lecture des properties
   * @param LoggerManager $oLogger Service d'accès aux Logger
   */
  public function __construct(Properties $PropertiesSrvc, LoggerManager $oLogger) {
    parent::__construct($PropertiesSrvc, $oLogger);

    $this->oClaimCheckerManager = new ClaimCheckerManager(
      [
        new Checker\IssuedAtChecker(),
        new Checker\NotBeforeChecker(),
        new ExpirationTimeCheckerManager(),
      ]
    );
  }

  /**
   * Renvoi le token JWE contenant les données à stocker.
   * (https://tools.ietf.org/html/rfc7519#section-4.1)
   * @param array $aPayload structure de donnée à stocker dans le token JWE
   *                        aPayload.iat    (integer) : issued at, date de génération du token
   *                        aPayload.nbf    (integer) : not before, date de début de validité
   *                        aPayload.exp    (integer) : expiration time, date d'expiration
   *                        aPayload.iss    (string)  : issuer, nom de l'application à l'origine du token
   *                        aPayload.aud    (string)  : audience, nom du destinataire du token
   *                        aPayload.sub    (string)  : subject, context de génération
   *                        aPayload.jti    (string)  : JWT id, id unique pour le token
   *                        aPayload.data[] (array)   : les données à stocker (user)
   *
   * @return string|null Token encrypté ou null en cas d'erreur
   */
  public function encryptToken(array $aPayload): ?string {
    // Check if payload is conform
    try {
      $this->oClaimCheckerManager->check($aPayload, self::CLAIMERS);
    } catch (\Exception $e) {
      $this->oErrorLogger->error('Error while encrypting token : ' . $e->getMessage());
      return null;
    }

    return $this->encryptData($aPayload);
  }

  /**
   * Renvoi un tableau contenant les données dans le token.
   *
   * @param string $sToken Token JWE provenant d'une requête HTTP
   *
   * @return array|null Tableau contenant le payload du token ou null en cas d'erreur
   */
  public function decryptToken(string $sToken): ?array{
    $aPayload = $this->decryptData($sToken);

    if ($aPayload === null) {
      return null;
    }

    try {
      $this->oClaimCheckerManager->check($aPayload, self::CLAIMERS);
    } catch (\Exception $e) {
      $this->oErrorLogger->error('Error while decrypting token : ' . $e->getMessage());
      return null;
    }
    return $aPayload;
  }

}
