import { WorkspaceListOptions } from './workspacelist/workspacelist.model';
import { DirectoryOptions } from './directory/directory.model';
import { SectionSimpleformOptions, SimpleformOptions } from './simpleform/simpleform.model';
import { DisplayformOptions } from './displayform/displayform.model';
import { StudioOptions } from './studio/studio.model';
import { ApiDocOptions } from './api-doc/api-doc.model';
import { WorkspaceRelationListOptions } from './workspacerelationlist/workspacerelationlist.model';


/**
 * Structure app.json
 */
export class AppConfig {
  translate: Array<AppTranslateConfig>; // Configuration fichiers de traduction
  name: string; // Nom de l'application
  logo: AppLogoConfig; // Configuration du logo
  favicon?: string; // Favicon
  modes: Array<AppModeConfig>; // Modes du menu latéral, instancie automatiquement les routes associées
  headers?: Array<any>;
  footers?: Array<AppFooterItemConfig>; // Permet de définir des objets à faire apparaitre dans le footer
  routes?: Array<AppModeConfig>; // Permet de définir des routes additionnelles à celles créées par les modes,
  widgets?: Array<AppModeConfig>; // Permet de définir des routes additionnelles à celles créées par les modes,
  documentation?: Array<AppHeaderDocumentationConfig>; // Permet de définir des documentations pour le header
  credits?: AppHeaderCreditsConfig; // Permet de définir des crédits pour le header
  // en cas de conflit sur le path la fonction prendra en priorité celle définie dans le paramètre routes
  login?: LoginConfig; // Configuration page de login
}

/**
 * Configuration page de login
 */
export class LoginConfig {
  signup_form?: string; // Chemin vers le formulaire à utiliser
  signup_js?: string; // Chemin vers le formulaire à utiliser
  signup_route?: string; // route à exploiter pour la ressource d'inscription
  confirm_ressource?: string; // Route à utiliser pour la confirmation
}

/**
 * Structure configuration des traductions
 */
export class AppTranslateConfig {
  prefix: string; // chemin vers le dossier où se trouvent les fichiers de langue
  suffix: string; // suffixe des fichiers à charger (".json" permet de charger l'ensemble des fichiers JSON)
}

/**
 * Structure configuration du logo
 */
export class AppLogoConfig {
  type: string; // icon ou image ou image-full
  class?: string; // classe à utiliser pour le type icon
  path?: string; // chemin vers l'image à utiliser en cas de type image
}

/**
 * Types acceptés par AppModeConfig.privilege_array_mode
 */
export enum PrivilegeArrayMode {
  AND = 'AND',
  OR = 'OR'
}

/**
 * Structure configuration mode
 */
export class AppModeConfig {
  name?: string; // Nom du mode
  type?: string; // Type de mode (separator ou composant vitis ou module ou module_component ou vide si objects)
  label?: string; // Libellé ou clé de traduction
  description?: string; // Libellé ou clé de traduction pour une possible description
  privilege: string | Array<string>; // Privilège minimum ou public
  privilege_array_mode?: any; // "AND" ou "OR", permet de gérer les cas où privilege est un array (défaut vaut "OR")
  visible?: string; // Condition d'affichage de la section, peut utiliser les properties avec la syntaxe @prop(...) ou des valeurs avec {{...}} ex : {{is_geometry_layer}} === true
  icon?: string; // Icone à utilier
  route?: string; // Route utiliser
  objects?: Array<AppModeConfig>; // Permet de créer un arbre récursif, si objects est défini alors type n'est pas obligatoire
  sections?: Array<AppModeConfig> | any; // Permet de créer un arbre récursif, si objects est défini alors type n'est pas obligatoire ou table de mappage
  options?: WorkspaceListOptions | DirectoryOptions | SimpleformOptions | SectionSimpleformOptions | DisplayformOptions | StudioOptions | ApiDocOptions | ModuleComponentOptions | any;
  headless?: boolean; // Mode plein écran
}

/**
 * Structure de configuration pour utilisation de modules externes
 */
export class ModuleOptionsConfig {
  module_name: string; // Nom du module décrit dans le fichier app/modules/modules-components.ts
}

/**
 * Structure de configuration pour utilisation de composants de modules externes
 */
export class ModuleComponentOptionsConfig {
  component_name: string; // Nom du composant décrit dans le fichier app/modules/modules-components.ts
}

/**
 * Options pour composant de type WorkspaceList
 */
export class ModuleComponentOptions {
  component_name: string;
  ressource: string;
  id_field: string;
}

/**
 * Structure configuration des documentations
 */
export class AppHeaderDocumentationConfig {
  name: string; // Nom à afficher dans la modale
  url: string; // Url à ouvrir pour accéder à la doc
}

/**
 * Structure configuration des crédits
 */
export class AppHeaderCreditsConfig {
  authors: Array<string>; // Liste des auteurs de l'application
  graphics: Array<string>; // Liste des designers de l'application
  libs: Array<string>; // Liste des livrairies/techno de l'application
  contact: string; // Lien vers une page de contact de l'entreprise responable de l'application
  licence: string; // Chaine à utilisé pour définir le propriétaire de la licence ex : "© Veremes"
}

/**
 * Structure configuration des footers
 */
export class AppFooterItemConfig {
  type: string; // Type d'item ('link', 'html', 'image' ou 'modalform')
  align?: string; // Définit l'alignement de l'item dans la barre ('right' ou 'left')
  style?: string; // Style personnalisé à appliquer à l'élément
  url?: string; // Url à ouvrir au clic sur l'item (fonctionne avec 'link' et 'image')
  label?: string; // Libellé à afficher sur un item de type 'link'
  content?: string; // Contenu du html à afficher
  styleContent?: string; // Style personnalisé du conteneur HTML
  src?: string; // Url menant vers la ressource de l'image
  name?: string; // Nom alternatif de l'image, si incorrecte
  image_style?: string; // Style personnalisé à appliquer à l'élément image
  params?: any; // Modal form params
}
