import { Component, OnInit, ViewChild, OnDestroy } from '@angular/core';
import { ActivatedRoute, Params } from '@angular/router';
import { VitisCheck, VitisRequest, vitisAlert } from 'src/app/vitis/core/services/utils/vitis-utils';
import { TranslateService } from '@ngx-translate/core';
import { environment } from 'src/environments/environment';
import { FormreaderComponent } from '../formreader/formreader.component';
import { FormreaderService } from '../formreader/services/formreader.service';
import { DirectoryOptions } from './directory.model';
import { TreeviewComponent } from 'src/app/vitis/shared/components/treeview/treeview.component';
declare const $: any;

/**
 * Options:
 * parent_folder_name: '/' + nom du premier dossier de l'arborescence
 * hide_first_folder?: boolean (affiche ou non le premier dossier de l'arborescence)
 * hover_download?: boolean (affiche ou non le bouton download quand on survole un fichier dans l'aborescence)
 * hover_delete?: boolean (affiche ou non le bouton delete quand on survole un fichier dans l'aborescence)
 * buttons?: Boutons à ajouter en pied de page
 *
 * Exemple:
 * "options": {
 *   "module": "vitis",
 *   "ressource": "vitis/exploitations/shareddirectory",
 *   "parent_folder_name": "/shared",
 *   "hide_first_folder": true,
 *   "hover_download": true,
 *   "hover_delete": true,
 *   "buttons": [{
 *     "type": "group",
 *     "buttons": [{
 *       "type": "button",
 *       "name": "load-tree",
 *       "title": "FORM_RELOAD_DIRECTORY",
 *       "class": "btn-primary",
 *       "icon": "fas fa-refresh",
 *       "event": "loadResourceData()"
 *     }, {
 *       "type": "button",
 *       "name": "download-resource",
 *       "title": "FORM_DOWNLOAD_FILE",
 *       "class": "btn-primary",
 *       "icon": "fas fa-download",
 *       "event": "downloadResource()"
 *     },
 *     {
 *       "type": "button",
 *       "name": "upload-file",
 *       "title": "FORM_BTN_UPLOAD_FILE",
 *       "class": "btn-primary",
 *       "icon": "fas fa-upload",
 *       "event": "uploadFile()",
 *       "param": "FORM_ADD_FILE_TITLE"
 *     }]
 *   }, {
 *     "type": "button",
 *     "name": "create-folder",
 *     "title": "FORM_BTN_CREATE_DIRECTORY",
 *     "class": "btn-primary",
 *     "icon": "fas fa-folder",
 *     "event": "createFolder()"
 *   }, {
 *     "type": "button",
 *     "name": "delete-resource",
 *     "title": "FORM_DELETE_FILE",
 *     "class": "btn-danger",
 *     "icon": "fas fa-trash",
 *     "event": "deleteResource()"
 *   }]
 * }
 *
 * Fonctions disponible par défaut pour les boutons:
 * loadResourceData()
 * downloadResource()
 * uploadFile()
 * createFolder()
 * deleteRessource()
 *
 * Si on veut ajout une autre fonction, on peut créer un fichier js
 * et ajouter l'option js_path
 */
@Component({
  selector: 'app-directory',
  templateUrl: './directory.component.html',
  styleUrls: ['./directory.component.scss']
})
export class DirectoryComponent implements OnInit, OnDestroy {

  @ViewChild(FormreaderComponent) oFormReader: FormreaderComponent;
  @ViewChild(TreeviewComponent) oTreeview: TreeviewComponent;

  public mode: any;
  public params: any;
  public object_data: any;
  public aFileTree: Array<any>;
  public selectedFile: Array<any> = [];
  public modalTitle: string;
  public oOptions: DirectoryOptions;
  public sBaseUrl: string;

  // Paramètres par défaut du Treeview.
  public treeviewConfig: any = {
    iconFile: true,
    defaultOpenLevel: -1, // Tous les répertoires fermés.
    titleAttribute: 'filename',
    nodeAttribute: 'content',
    dateAttribute: 'lastModification',
    multiSelect: false,
    selectFolders: true,
    pathAttribute: '',
    showDate: true
  };

  /**
   * Constructeur
   */
  constructor(
    private route: ActivatedRoute,
    private vitisCheck: VitisCheck,
    private vitisRequest: VitisRequest,
    private translate: TranslateService,
    private formReaderService: FormreaderService
  ) { }

  /**
   * Chargement du composant
   */
  ngOnInit() {
    // Mode en cours d'utilisation
    this.route.data.subscribe(mode => {
      this.mode = mode;
      this.oOptions = mode.options;
      this.object_data = {};

      // Get route's params
      this.route.paramMap.subscribe((oParams: Params) => {
        this.params = oParams.params;
      });

      this.sBaseUrl = environment.properties.app_api_url + '/' + this.mode.options.ressource;
      if (this.vitisCheck.isDefAndNotNull(this.params.id)) {
        this.sBaseUrl += '/' + this.params.id;
      }

      if (this.mode.options.load_object_data === true) {
        this.vitisRequest.ajaxRequestPromise({
          method: 'GET',
          url: this.sBaseUrl,
          ajaxLoader: false,
          params: {
            depth: 1
          }
        }).then((resp: any) => {
          if (resp.data && resp.data.status == 1) {
            this.object_data = resp.data.data;
            this.initComponent();
          } else {
            this.translate.get('DIRECTORY_COMPONENT_CANNOT_GET_OBJECT_DATA').subscribe((translate) => {
              vitisAlert.alert({
                text: translate,
                type: 'error'
              });
            });
          }
        });
      } else {
        this.initComponent();
      }
    });
  }

  /**
   * Initialisation du composant
   */
  private initComponent() {
    // Initialise les boutons
    this.initButtons();

    this.initTreeviewConfig();
    // Fusionne la config par défaut et celle du app.json
    for (const sKey in this.oOptions.treeview_config) {
      if (this.oOptions.treeview_config.hasOwnProperty(sKey)) {
        this.treeviewConfig[sKey] = this.oOptions.treeview_config[sKey];
      }
    }

    // Récupère l'arborescence du projet
    this.loadResourceData();
  }

  /**
   * Change le fichier selectionné
   * @param {Array<any>} selectedFile
   */
  public selectedChange(selectedFile: Array<any>) {
    this.selectedFile = selectedFile;
  }

  /**
   * Récupère l'arborescence du projet
   * @returns Promise
   */
  private getResourceData_(): Promise<any> {
    return this.vitisRequest.ajaxRequestPromise({
      method: 'GET',
      url: this.baseUrlConcatenator(this.mode.options.subroute.getTree),
      ajaxLoader: false,
      params: {
        depth: 1
      }
    });
  }

  /**
   * Charge/Recharge l'arborescence du projet
   * @returns Promise
   */
  public loadResourceData() {
    this.getResourceData_().then((result: any) => {
      if (this.vitisCheck.isDefAndNotNull(result.data)
        && this.vitisCheck.isDefAndNotNull(result.data.data)) {
        this.aFileTree = this.aNodeConstructor([result.data.data]);
        // Quand on recharge on enlève la sélection
        this.selectedFile = [];
        // Si on doit ignorer le premier element on l'enlève
        if (this.oOptions.hide_first_folder === true) {
          this.aFileTree = this.aFileTree[0][this.treeviewConfig.nodeAttribute];
        }
      }
    });
  }

  /**
   * Télécharge la resource selectionée
   * @returns void
   */
  public downloadResource() {
    // Vérifie que ce soit un fichier en vérifiant qu'il n'y ai pas de sous-contenu
    if (this.isFile_(this.selectedFile[0])) {
      this.downloadTreeviewItem(this.selectedFile[0]);
    } else {
      // Message d'erreur
      this.translate.get('ERROR_SELECTED_FILE').subscribe((translation) => {
        vitisAlert.fireToast({
          type: 'warning',
          title: translation
        });
      });
    }
  }

  /**
   *
   */
  private getParentFolderName(): string {
    const sFolderName: string = this.oOptions.parent_folder_name;
    return sFolderName.replace(':id', this.params.id);
  }

  /**
   * Créer un fichier
   */
  public createFolder() {
    if (!this.vitisCheck.isDefAndNotNull(this.selectedFile[0]) || this.isFolder_(this.selectedFile[0])) {
      const formData = new FormData();
      formData.append('directory_path', '');
      // Si le dossier à été selectioné initialise le 'directory_path'
      if (this.vitisCheck.isDefAndNotNull(this.selectedFile[0])) {
        if (this.selectedFile[0].hasOwnProperty('content')) {
          formData.set('directory_path', this.selectedFile[0].value.replace(this.getParentFolderName(), ''));
        }
      }

      this.translate.get([
        'CONFIRM_CONFIRM',
        'CONFIRM_CANCEL',
        'FORM_FOLDER_NAME',
        'VITIS_DIRECTORY_COMPONENT_ENTER_VALID_DIRNAME',
        'ERROR_CREATE_DIRECTORY_EMPTY_NAME'
      ]).subscribe((translates) => {

        vitisAlert.alert({
          title: translates.FORM_FOLDER_NAME,
          showCancelButton: true,
          confirmButtonColor: '#3085d6',
          cancelButtonColor: '#d33',
          confirmButtonText: translates.CONFIRM_CONFIRM,
          cancelButtonText: translates.CONFIRM_CANCEL,
          html: '<div class="input-group input-group-lg">'
            + '<div class="input-group-prepend">'
            + '<span class="input-group-text">' + formData.get('directory_path') + '/</span>'
            + '</div>'
            + '<input id="creation_directory_name" type="text" class="form-control">'
            + '</div>'
        }).then((result) => {
          if (result.value === true) {

            const sPathValue: string = $('#creation_directory_name').val();
            if (sPathValue != '') {

              const pattern = /^(?!.*\.\.)[a-zA-Z0-9-_ ]+(?:\/[a-zA-Z0-9-_ ]+)*$/;
              if (pattern.test(sPathValue)) {
                formData.append('directory_name', sPathValue);

                this.vitisRequest.ajaxRequestPromise({
                  method: 'POST',
                  url: this.baseUrlConcatenator(this.mode.options.subroute.createDirectory),
                  data: formData
                }).then((resp: any) => {
                  if (resp.data && resp.data.status == 1) {
                    this.translate.get('SUCCESSFUL_OPERATION').subscribe((translate) => {
                      vitisAlert.fireToast({
                        type: 'success',
                        title: translate
                      });

                      this.loadResourceData();
                    });
                  } else {
                    this.translate.get(resp.data.errorMessage).subscribe((translate) => {
                      vitisAlert.alert({
                        text: translate,
                        type: 'error'
                      });
                    });
                  }
                });
              } else {
                vitisAlert.fireToast({
                  type: 'error',
                  title: translates.VITIS_DIRECTORY_COMPONENT_ENTER_VALID_DIRNAME
                });
              }
            } else {
              vitisAlert.fireToast({
                type: 'error',
                title: translates.ERROR_CREATE_DIRECTORY_EMPTY_NAME
              });
            }
          }
        });
      });
    } else if (this.isFile_(this.selectedFile[0])) {
      this.translate.get('ERROR_CREATE_DIRECTORY_SELECT_FILE').subscribe((translate) => {
        vitisAlert.fireToast({
          type: 'error',
          title: translate
        });
      });

    }
  }

  /**
   * Créer un fichier
   */
  public uploadFile(sTitle: string) {
    if (!this.vitisCheck.isDefAndNotNull(this.selectedFile[0]) || this.isFolder_(this.selectedFile[0])) {

      this.modalTitle = sTitle;

      const oForm: any = {
        update: {
          name: 'update_form',
          title: 'FORM_TITLE_PUBLICATION_WORKSPACE',
          eventContainer: this,
          event: 'sendForm()',
          input_size: 'xs',
          nb_cols: 12,
          rows: [{
            fields: [{
              type: 'file_wsdata',
              name: 'file',
              id: 'file',
              max_size: environment.properties.max_upload_file_size,
              label: 'FORM_ADDITIONAL_RESOURCES_PUBLICATION_WORKSPACE_LOAD_PROJECT_DIRECTORY',
              nb_cols: 12
            }]
          },
          {
            fields: [{
              type: 'checkbox',
              name: 'unzip',
              id: 'unzip',
              value: true,
              label: 'FORM_UNZIP_WORKSPACE_LOAD_PROJECT_DIRECTORY',
              default_value: true,
              nb_cols: 8
            }]
          },
          {
            fields: [{
              type: 'button',
              class: 'btn-ungroup btn-group-sm',
              nb_cols: 6,
              buttons: [{
                type: 'submit',
                name: 'btn_upload',
                label: 'CONFIRM_CONFIRM',
                class: 'btn-primary'
              }]
            }]
          }]
        }
      };
      // Lance le formulaire dans FormReader
      this.oFormReader.initForm(oForm, 'update');

      $('#directory-form-modal').modal('show');
    }
  }

  /**
   * Supprime la ressource selectionné
   */
  public deleteResource() {
    if (this.vitisCheck.isDefAndNotNull(this.selectedFile[0])) {
      this.confirmDeleteTreeviewItem(this.selectedFile[0]);
    } else {
      // Message d'erreur
      this.translate.get('ERROR_SELECTED_FILE').subscribe((translation) => {
        vitisAlert.fireToast({
          type: 'warning',
          title: translation
        });
      });
    }
  }

  /**
   * Message de confirmation avant de lancer la suppression d'un fichier/dossier du treeview
   * @param {any} item
   */
  public confirmDeleteTreeviewItem(item: any) {
    this.translate.get(['CONFIRM_DELETE_FILE',
      'CONFIRM_CONFIRM',
      'CONFIRM_CANCEL']).subscribe((translates) => {
      vitisAlert.alert({
        title: translates.CONFIRM_DELETE_FILE,
        text: item.value,
        type: 'question',
        showCancelButton: true,
        confirmButtonColor: '#3085d6',
        cancelButtonColor: '#d33',
        confirmButtonText: translates.CONFIRM_CONFIRM,
        cancelButtonText: translates.CONFIRM_CANCEL
      }).then((result) => {
        if (result.value) {
          this.deleteTreeviewItem(item);
        }
      });
    });
  }

  /**
   * Initialise l'événement d'un bouton
   * @param {any} button
   */
  public initButtonEvent(button: any) {

    // Transforme la chaine de caractères en vrai fonction
    if (this.vitisCheck.isString(button.event)) {
      button.sEvent = button.event;
    }
    if (this.vitisCheck.isString(button.sEvent)) {
      button.event = () => {

        // Nom de la fonction sans les paranthèses
        let sFunction = button.sEvent;
        if (sFunction.indexOf('(') !== -1) {
          sFunction = sFunction.substring(0, sFunction.indexOf('('));
        }

        // Lance la fonction depuis le composant
        if (this.vitisCheck.isFunction(this[sFunction])) {
          if (this.vitisCheck.isDefAndNotNull(button.param)) {
            return this[sFunction].call(this, button.param);
          }
          return this[sFunction].call(this);
        }
      };
    }

    // Recursivité
    if (this.vitisCheck.isArray(button.buttons)) {
      for (let i = 0; i < button.buttons.length; i++) {
        const buttonChild = button.buttons[i];
        this.initButtonEvent(buttonChild);
      }
    }
  }

  /**
   * Charge le JS associé au mode
   * @param  {any} oMode
   * @returns Promise
   */
  private loadModeJavaScript_(oMode: any): Promise<boolean> {
    return new Promise((resolve, reject) => {

      // Si le mode contient un JS associé
      if (this.vitisCheck.isDefAndNotNull(oMode.options)) {
        if (this.vitisCheck.isDefAndNotNull(oMode.options.js_path)) {

          // Charge le JS du mode
          this.vitisRequest.loadExternalJs(oMode.options.js_path).then(() => {

            // Lance le JS associé
            if ((<any>window).constructor_form) {
              (<any>window).constructor_form(this, environment, {
                vitisRequest: this.vitisRequest,
                vitisCheck: this.vitisCheck,
                vitisAlert: vitisAlert,
                translate: this.translate
              });
            }

            resolve(true);
          });
        } else {
          resolve(false);
        }
      } else {
        resolve(false);
      }
    });
  }

  /**
   * Initialise les boutons
   * @returns void
   */
  public initButtons(): void {
    const aButtons: any = [];

    if (this.vitisCheck.isDefAndNotNull(this.oOptions)) {

      if (this.vitisCheck.isArray(this.oOptions.buttons)) {
        // Ajoute les boutons
        for (let i = 0; i < this.oOptions.buttons.length; i++) {
          const button = this.oOptions.buttons[i];
          this.initButtonEvent(button);
          aButtons.push(button);
        }

        // Charge le JS
        this.loadModeJavaScript_(this.mode);
      }
    }

    this.oOptions.buttons = aButtons;
  }

  /**
   * initialise la configuration du treeview
   * @param aList
   * @param sPath
   */
  public initTreeviewConfig() {
    const hoverButtons = [];

    if (this.oOptions.hover_download === true) {
      hoverButtons.push({
        icon: 'fa fa-download',
        event: (item: any) => {
          this.downloadTreeviewItem(item);
        }
      });
    }

    if (this.oOptions.hover_delete === true) {
      hoverButtons.push({
        icon: 'fa fa-trash',
        event: (item: any) => {
          this.confirmDeleteTreeviewItem(item);
        }
      });
    }

    this.treeviewConfig.hoverButtons = hoverButtons;

    // Fusionne la config par défaut et celle du app.json
    for (const sKey in this.oOptions.treeview_config) {
      if (this.oOptions.treeview_config.hasOwnProperty(sKey)) {
        this.treeviewConfig[sKey] = this.oOptions.treeview_config[sKey];
      }
    }
  }

  /**
   * Construit l'arborescence.
   * @param {Array<any>} aList
   * @param {string} sPath?
   **/
  public aNodeConstructor(aList: Array<any>, sPath?: string): Array<any> {
    if (!this.vitisCheck.isDefAndNotNull(sPath)) {
      sPath = '/';
    } else {
      sPath += '/';
    }
    const aNodes = [];
    for (let i = 0; i < aList.length; i++) {
      const aNode: any = aList[i];

      aNode.value = sPath + aList[i][this.treeviewConfig.titleAttribute];
      aNode.tags = [aList[i][this.treeviewConfig.dateAttribute]];

      if (this.vitisCheck.isDefAndNotNull(aList[i][this.treeviewConfig.nodeAttribute])) {
        aNode[this.treeviewConfig.nodeAttribute] = this.aNodeConstructor(
          aList[i][this.treeviewConfig.nodeAttribute],
          sPath + aList[i][this.treeviewConfig.titleAttribute]);
      }

      aNodes.push(aNode);
    }

    return this.orderFoldersFirst_(aNodes);
  }

  /**
   * Télécharge un fichier du treeview
   * @param  {any} item
   * @returns void
   */
  public downloadTreeviewItem(item: any) {
    // Charge le contenu du fichier sélectionné
    const sFileName = item[this.treeviewConfig.titleAttribute];
    const sUrl = item.href;
    this.vitisRequest.ajaxDownloadFile(sUrl, sFileName);
  }

  /**
   * Supprime un fichier/dossier du treeview
   * @param {any} item
   */
  public deleteTreeviewItem(item: any) {
    this.vitisRequest.ajaxRequestPromise({
      method: 'DELETE',
      url: this.baseUrlConcatenator(this.mode.options.subroute.deleteItem),
      params: {
        'idList': item.value.replace(this.getParentFolderName(), '')
      }
    }).then((resp: any) => {
      if (resp.data && resp.data.status == 1) {
        this.loadResourceData();
      } else {
        this.translate.get(resp.data.errorMessage).subscribe((translate) => {
          vitisAlert.alert({
            text: translate,
            type: 'error'
          });
        });
      }
    });
  }

  /**
   * On vérifie que ce soit un dossier en vérifiant qu'il y ai du sous-contenu
   * @param file
   */
  private isFolder_(file: any) {
    return this.vitisCheck.isDefAndNotNull(file) &&
      this.vitisCheck.isDefAndNotNull(file[this.treeviewConfig.nodeAttribute]);
  }

  /**
   * On vérifie que ce soit un fichier en vérifiant qu'il n'y ai pas de sous-contenu
   * @param file
   */
  private isFile_(file: any) {
    return this.vitisCheck.isDefAndNotNull(file) &&
      !this.vitisCheck.isDefAndNotNull(file[this.treeviewConfig.nodeAttribute]);
  }


  /**
   * Envoi le formulaire
   * @param {string} sEndUrl
   */
  public sendForm() {
    const values = this.oFormReader.getFormValues();

    this.formReaderService.manageLargesFiles(values).then((fileValues: any) => {
      // Récupération des données du form
      const formData = this.formReaderService.transformToDataForm(fileValues);
      formData.set('directory_path', '/');
      // Si un dossier à été selectioné initialise le 'directory_path'
      if (this.vitisCheck.isDefAndNotNull(this.selectedFile[0])) {
        if (this.selectedFile[0].hasOwnProperty('content')) {
          formData.set('directory_path', this.selectedFile[0].value.replace(this.getParentFolderName(), ''));
        }
      }

      const sUrl = this.baseUrlConcatenator(this.mode.options.subroute.sendFile);

      this.vitisRequest.ajaxRequestPromise({
        method: 'PUT',
        url: sUrl,
        data: formData
      }).then((resp: any) => {
        if (resp.data && resp.data.status == 1) {
          this.translate.get('SUCCESSFUL_OPERATION').subscribe((translate) => {
            vitisAlert.fireToast({
              type: 'success',
              title: translate
            });

            this.loadResourceData();
            $('#directory-form-modal').modal('hide');
          });
        }
      });
    });

  }

  /**
   * Si le bouton doit être désactivé
   * @param {any} button
   * @returns {boolean}
   */
  public isDisabled(button: any): boolean {
    if (this.vitisCheck.isDefAndNotNull(button.disabled)) {
      return eval(button.disabled);
    }
    return false;
  }
  /**
   * Si le bouton doit être désactivé
   * @param {any} button
   * @returns {boolean}
   */
  public isVisible(button: any): boolean {
    if (this.vitisCheck.isDefAndNotNull(button.visible)) {
      return eval(button.visible);
    }
    return true;
  }

  /**
   * Tri les dossiers en premier
   * @param aNodes
   */
  private orderFoldersFirst_(aNodes: Array<any>): Array<any> {
    if (this.oOptions.sort_folders_first !== false) {
      aNodes.sort((a, b) => {
        if (this.vitisCheck.isDefAndNotNull(a[this.treeviewConfig.nodeAttribute]) &&
          this.vitisCheck.isDefAndNotNull(b[this.treeviewConfig.nodeAttribute])) {
          return 0;
        }
        if (this.vitisCheck.isDefAndNotNull(a[this.treeviewConfig.nodeAttribute]) &&
          !this.vitisCheck.isDefAndNotNull(b[this.treeviewConfig.nodeAttribute])) {
          return -1;
        }
        if (!this.vitisCheck.isDefAndNotNull(a[this.treeviewConfig.nodeAttribute]) &&
          this.vitisCheck.isDefAndNotNull(b[this.treeviewConfig.nodeAttribute])) {
          return 1;
        }
      });
    }

    return aNodes;
  }

  /**
   * Teste s'il faut modifier l'url de base et fait la modification
   * @param {string} sEndUrl chaine à concaténer
   */
  private baseUrlConcatenator(sEndUrl: string): string {
    let sUrl = this.sBaseUrl;
    if (this.vitisCheck.isDefAndNotNull(sEndUrl) && sEndUrl !== '') {
      sUrl += '/' + sEndUrl;
    }
    return sUrl;
  }

  /**
   * Change le fichier selectionné
   * @param {any} item
   */
  public folderCollapsedChange(item: any) {
    if (item.collapsed === false && this.vitisCheck.isDefAndNotNull(item.value)) {
      // Chargement de l'arboresence du répertoire.
      this.vitisRequest.ajaxRequestPromise({
        method: 'GET',
        url: this.baseUrlConcatenator(this.mode.options.subroute.getTree),
        ajaxLoader: true,
        params: {
          depth: 1,
          path: String(item.value).replace(this.getParentFolderName(), '')
        }
      }).then((resp: any) => {
        if (resp.data && resp.data.status == 1) {
          if (this.vitisCheck.isDefAndNotNull(resp.data.data) &&
            this.vitisCheck.isDefAndNotNull(resp.data.data.content)) {
            item.content = resp.data.data.content;
            // Ajout des paramètres.
            this.oTreeview.parseItems_(item.content);
            // Construction de l'arboresence.
            for (let i = 0; i < item.content.length; i++) {
              item.content[i].value = item.value + '/' + item.content[i][this.treeviewConfig.titleAttribute];
              item.content[i].tags = item.content[i][this.treeviewConfig.dateAttribute];
            }
          }
        } else {
          this.translate.get('DIRECTORY_COMPONENT_CANNOT_GET_OBJECT_DATA').subscribe((translate) => {
            vitisAlert.alert({
              text: translate,
              type: 'error'
            });
          });
        }
      });
    }
  }

  /**
   * Destruction du composant
   */
  ngOnDestroy() {
    if ((<any>window).constructor_form) {
      (<any>window).constructor_form = () => { };
    }
  }
}
